Package: did
Title: Treatment Effects with Multiple Periods and Groups
Version: 2.5.1
Authors@R: c(person("Brantly", "Callaway", email = "brantly.callaway@uga.edu", role = c("aut", "cre")), person("Pedro H. C.", "Sant'Anna", email="pedro.santanna@emory.edu", role = c("aut")))
URL: https://bcallaway11.github.io/did/,
        https://github.com/bcallaway11/did/
Description: The standard Difference-in-Differences (DID) setup involves two periods and two groups -- a treated group and untreated group.  Many applications of DID methods involve more than two periods and have individuals that are treated at different points in time.  This package contains tools for computing average treatment effect parameters in Difference in Differences setups with more than two periods and with variation in treatment timing using the methods developed in Callaway and Sant'Anna (2021) <doi:10.1016/j.jeconom.2020.12.001>.  The main parameters are group-time average treatment effects which are the average treatment effect for a particular group at a particular time.  These can be aggregated into a fewer number of treatment effect parameters, and the package deals with the cases where there is selective treatment timing, dynamic treatment effects, calendar time effects, or combinations of these.  There are also functions for testing the Difference in Differences assumption, and plotting group-time average treatment effects.
Depends: R (>= 4.1.0)
License: GPL-3
Encoding: UTF-8
LazyData: true
Imports: BMisc (>= 1.4.4), Matrix, pbapply, ggplot2, DRDID (>= 1.3.0),
        generics, methods, tidyr, fastglm, data.table (>= 1.15.4),
        dreamerr (>= 1.4.0)
VignetteBuilder: knitr
Suggests: rmarkdown, plm, here, knitr, covr, backports, broom, testthat
        (>= 3.0.0), withr, remotes, callr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-08 15:28:03 UTC; bmc43193
Author: Brantly Callaway [aut, cre],
  Pedro H. C. Sant'Anna [aut]
Maintainer: Brantly Callaway <brantly.callaway@uga.edu>
Repository: CRAN
Date/Publication: 2026-07-08 16:20:02 UTC

Package: did
Title: Treatment Effects with Multiple Periods and Groups
Version: 2.5.0
Authors@R: c(person("Brantly", "Callaway", email = "brantly.callaway@uga.edu", role = c("aut", "cre")), person("Pedro H. C.", "Sant'Anna", email="pedro.santanna@emory.edu", role = c("aut")))
URL: https://bcallaway11.github.io/did/,
        https://github.com/bcallaway11/did/
Description: The standard Difference-in-Differences (DID) setup involves two periods and two groups -- a treated group and untreated group.  Many applications of DID methods involve more than two periods and have individuals that are treated at different points in time.  This package contains tools for computing average treatment effect parameters in Difference in Differences setups with more than two periods and with variation in treatment timing using the methods developed in Callaway and Sant'Anna (2021) <doi:10.1016/j.jeconom.2020.12.001>.  The main parameters are group-time average treatment effects which are the average treatment effect for a particular group at a particular time.  These can be aggregated into a fewer number of treatment effect parameters, and the package deals with the cases where there is selective treatment timing, dynamic treatment effects, calendar time effects, or combinations of these.  There are also functions for testing the Difference in Differences assumption, and plotting group-time average treatment effects.
Depends: R (>= 4.1.0)
License: GPL-3
Encoding: UTF-8
LazyData: true
Imports: BMisc (>= 1.4.4), Matrix, pbapply, ggplot2, DRDID (>= 1.3.0),
        generics, methods, tidyr, fastglm, data.table (>= 1.15.4),
        dreamerr (>= 1.4.0)
VignetteBuilder: knitr
Suggests: rmarkdown, plm, here, knitr, covr, backports, broom, testthat
        (>= 3.0.0), withr, remotes, callr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-13 15:06:15 UTC; bmc43193
Author: Brantly Callaway [aut, cre],
  Pedro H. C. Sant'Anna [aut]
Maintainer: Brantly Callaway <brantly.callaway@uga.edu>
Repository: CRAN
Date/Publication: 2026-06-13 15:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(ggdid,AGGTEobj)
S3method(ggdid,MP)
S3method(glance,AGGTEobj)
S3method(glance,MP)
S3method(nobs,AGGTEobj)
S3method(nobs,MP)
S3method(print,AGGTEobj)
S3method(print,MP)
S3method(summary,AGGTEobj)
S3method(summary,MP)
S3method(summary,MP.TEST)
S3method(tidy,AGGTEobj)
S3method(tidy,MP)
export(AGGTEobj)
export(DIDparams)
export(MP)
export(MP.TEST)
export(aggte)
export(att_gt)
export(build_sim_dataset)
export(compute.aggte)
export(compute.att_gt)
export(compute.att_gt2)
export(conditional_did_pretest)
export(ggdid)
export(glance)
export(gplot)
export(indicator)
export(mboot)
export(pre_process_did)
export(pre_process_did2)
export(process_attgt)
export(reset.sim)
export(sim)
export(splot)
export(test.mboot)
export(tidy)
export(trimmer)
import(data.table)
import(fastglm)
import(ggplot2)
importFrom(BMisc,TorF)
importFrom(BMisc,get_list_element)
importFrom(BMisc,makeBalancedPanel)
importFrom(BMisc,multiplier_bootstrap)
importFrom(BMisc,rhs_vars)
importFrom(BMisc,toformula)
importFrom(DRDID,drdid_panel)
importFrom(DRDID,drdid_rc)
importFrom(DRDID,reg_did_panel)
importFrom(DRDID,reg_did_rc)
importFrom(DRDID,std_ipw_did_panel)
importFrom(DRDID,std_ipw_did_rc)
importFrom(dreamerr,check_set_arg)
importFrom(generics,glance)
importFrom(generics,tidy)
importFrom(methods,as)
importFrom(methods,is)
importFrom(stats,aggregate)
importFrom(stats,binomial)
importFrom(stats,complete.cases)
importFrom(stats,cov)
importFrom(stats,ecdf)
importFrom(stats,glm)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,na.pass)
importFrom(stats,nobs)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(tidyr,gather)
importFrom(utils,globalVariables)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Difference-in-Differences <img src="man/figures/logo.png" align="right" alt="" width="155" />

[![](http://cranlogs.r-pkg.org/badges/grand-total/did?color=blue)](https://cran.r-project.org/package=did)
[![](http://cranlogs.r-pkg.org/badges/last-month/did?color=blue)](https://cran.r-project.org/package=did)
[![](https://www.r-pkg.org/badges/version/did?color=blue)](https://cran.r-project.org/package=did)
[![](https://img.shields.io/badge/devel%20version-2.5.0-blue.svg)](https://github.com/bcallaway11/did)
[![CRAN
checks](https://badges.cranchecks.info/summary/did.svg)](https://cran.r-project.org/web/checks/check_results_did.html)
[![](https://img.shields.io/github/last-commit/bcallaway11/did.svg)](https://github.com/bcallaway11/did/commits/master)

<!-- README.md is generated from README.Rmd. Please edit that file -->

The **did** package contains tools for computing average treatment
effect parameters in a Difference-in-Differences setup allowing for

- More than two time periods

- Variation in treatment timing (i.e., units can become treated at
  different points in time)

- Treatment effect heterogeneity (i.e, the effect of participating in
  the treatment can vary across units and exhibit potentially complex
  dynamics, selection into treatment, or time effects)

- The parallel trends assumption holds only after conditioning on
  covariates

The main parameters are **group-time average treatment effects**. These
are the average treatment effect for a particular group (group is
defined by treatment timing) in a particular time period. These
parameters are a natural generalization of the average treatment effect
on the treated (ATT) which is identified in the textbook case with two
periods and two groups to the case with multiple periods.

Group-time average treatment effects are also natural building blocks
for more aggregated treatment effect parameters such as overall
treatment effects or event-study-type estimands.

## Getting Started

There has been some recent work on DiD with multiple time periods. The
**did** package implements the framework put forward in

- [Callaway, Brantly and Pedro H.C. Sant’Anna.
  "Difference-in-Differences with Multiple Time Periods." Journal of
  Econometrics, Vol. 225, No. 2, pp. 200-230,
  2021.](https://doi.org/10.1016/j.jeconom.2020.12.001) or
  [arXiv](https://arxiv.org/abs/1803.09015)

**Higher level discussions of issues are available in**

- [Our approach to DiD with multiple time
  periods](https://bcallaway11.github.io/did/articles/multi-period-did.html)

## Installation

You can install **did** from CRAN with:

``` r
install.packages("did")
```

or get the latest version from github with:

``` r
# install.packages("devtools")
devtools::install_github("bcallaway11/did")
```

## A short example

The following is a simplified example of the effect of states increasing
their minimum wages on county-level teen employment rates which comes
from [Callaway and Sant’Anna
(2021)](https://doi.org/10.1016/j.jeconom.2020.12.001).

- [More detailed examples are also
  available](https://bcallaway11.github.io/did/articles/did-basics.html)

A subset of the data is available in the package and can be loaded by

``` r
library(did)
data(mpdta)
```

The dataset contains 500 observations of county-level teen employment
rates from 2003-2007. Some states are first treated in 2004, some in
2006, and some in 2007 (see the paper for more details). The important
variables in the dataset are

- **lemp** This is the log of county-level teen employment. It is the
  outcome variable

- **first.treat** This is the period when a state first increases its
  minimum wage. It can be 2004, 2006, or 2007. It is the variable that
  defines *group* in this application

- **year** This is the year and is the *time* variable

- **countyreal** This is an id number for each county and provides the
  individual identifier in this panel data context

To estimate group-time average treatment effects, use the **att_gt**
function

``` r
out <- att_gt(
  yname = "lemp",
  gname = "first.treat",
  idname = "countyreal",
  tname = "year",
  xformla = ~1,
  data = mpdta,
  est_method = "reg"
)
```

**att_gt** returns a class **MP** object. This has a lot of information,
but most importantly is has estimates of the group-time average
treatment effects and their standard errors. To see these, we can call
the **summary** function

``` r
summary(out)
#> 
#> Call:
#> att_gt(yname = "lemp", tname = "year", idname = "countyreal", 
#>     gname = "first.treat", xformla = ~1, data = mpdta, est_method = "reg")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> Group-Time Average Treatment Effects:
#>  Group Time ATT(g,t) Std. Error [95% Simult.  Conf. Band]  
#>   2004 2004  -0.0105     0.0246       -0.0752      0.0542  
#>   2004 2005  -0.0704     0.0332       -0.1580      0.0171  
#>   2004 2006  -0.1373     0.0397       -0.2419     -0.0326 *
#>   2004 2007  -0.1008     0.0357       -0.1947     -0.0069 *
#>   2006 2004   0.0065     0.0229       -0.0537      0.0668  
#>   2006 2005  -0.0028     0.0197       -0.0546      0.0491  
#>   2006 2006  -0.0046     0.0179       -0.0519      0.0427  
#>   2006 2007  -0.0412     0.0192       -0.0918      0.0094  
#>   2007 2004   0.0305     0.0154       -0.0100      0.0710  
#>   2007 2005  -0.0027     0.0164       -0.0459      0.0405  
#>   2007 2006  -0.0311     0.0184       -0.0795      0.0173  
#>   2007 2007  -0.0261     0.0170       -0.0708      0.0187  
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> P-value for pre-test of parallel trends assumption:  0.16812
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

This provides estimates of group-time average treatment effects for all
groups in all time periods. Group-time average treatment effects are
identified when `t >= g` (these are post-treatment time periods for each
group), but **summary** reports them even in periods when `t < g` –
these can be used to pre-test for the parallel trends assumption. The
`P-value for pre-test of parallel trends assumption` is for a Wald
pre-test of the parallel trends assumption. Here the parallel trends
assumption would not be rejected at conventional significance levels.

It is often also convenient to plot the group-time average treatment
effects. This can be done using the **ggdid** command:

``` r
ggdid(out, ylim = c(-.25, .1))
```

<img src="man/figures/README-unnamed-chunk-9-1.png" alt="" width="90%" style="display: block; margin: auto;" />

The red dots in the plot are pre-treatment group-time average treatment
effects . Here they are provided with 95% simultaneous confidence
intervals. These are the estimates that can be interpreted as a pre-test
of the parallel trends assumption. The blue dots are post-treatment
group-time average treatment effects. Under the parallel trends
assumption, these can be interpreted as policy effects – here the effect
of the minimum wage on county-level teen employment due to increasing
the minimum wage.

**Event Studies**

Although in the current example it is pretty easy to directly interpret
the group-time average treatment effects, there are many cases where it
is convenient to aggregate the group-time average treatment effects into
a small number of parameters. A main type of aggregation is into an
*event study* plot.

To make an event study plot in the **did** package, one can use the
**aggte** function

``` r
es <- aggte(out, type = "dynamic")
```

Just like for group-time average treatment effects, these can be
summarized and plotted. First, the summary

``` r
summary(es)
#> 
#> Call:
#> aggte(MP = out, type = "dynamic")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on event-study/dynamic aggregation:  
#>      ATT    Std. Error     [ 95%  Conf. Int.]  
#>  -0.0772        0.0219    -0.1201     -0.0344 *
#> 
#> 
#> Dynamic Effects:
#>  Event time Estimate Std. Error [95% Simult.  Conf. Band]  
#>          -3   0.0305     0.0154       -0.0100      0.0710  
#>          -2  -0.0006     0.0132       -0.0352      0.0341  
#>          -1  -0.0245     0.0153       -0.0646      0.0157  
#>           0  -0.0199     0.0123       -0.0522      0.0124  
#>           1  -0.0510     0.0167       -0.0948     -0.0071 *
#>           2  -0.1373     0.0413       -0.2456     -0.0289 *
#>           3  -0.1008     0.0351       -0.1931     -0.0085 *
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

The column `event time` is for each group relative to when they first
participate in the treatment. To give some examples, `event time=0`
corresponds to the *on impact* effect, and `event time=-1` is the
*effect* in the period before a unit becomes treated (checking that this
is equal to 0 is potentially useful as a pre-test).

To plot the event study, use **ggdid**

``` r
ggdid(es)
```

<img src="man/figures/README-unnamed-chunk-12-1.png" alt="" width="90%" style="display: block; margin: auto;" />

The figure here is very similar to the group-time average treatment
effects. Red dots are pre-treatment periods, blue dots are
post-treatment periods. The difference is that the x-axis is in event
time.

**Overall Effect of Participating in the Treatment**

The event study above reported an overall effect of participating in the
treatment. This was computed by averaging the average effects computed
at each length of exposure.

In many cases, a more general purpose overall treatment effect parameter
is give by computing the average treatment effect for each group, and
then averaging across groups. This sort of procedure provides an average
treatment effect parameter with a very similar interpretation to the
Average Treatment Effect on the Treated (ATT) in the two period and two
group case.

To compute this overall average treatment effect parameter, use

``` r
group_effects <- aggte(out, type = "group")
summary(group_effects)
#> 
#> Call:
#> aggte(MP = out, type = "group")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on group/cohort aggregation:  
#>     ATT    Std. Error     [ 95%  Conf. Int.]  
#>  -0.031        0.0126    -0.0558     -0.0062 *
#> 
#> 
#> Group Effects:
#>  Group Estimate Std. Error [95% Simult.  Conf. Band]  
#>   2004  -0.0797     0.0291       -0.1434     -0.0161 *
#>   2006  -0.0229     0.0176       -0.0613      0.0155  
#>   2007  -0.0261     0.0175       -0.0642      0.0121  
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

Of particular interest is the `Overall ATT` in the results. Here, we
estimate that increasing the minimum wage decreased teen employment by
3.1% and the effect is marginally statistically significant.

## Additional Resources

We have provided several more vignettes that may be helpful for using
the **did** package

- [Getting Started with the did
  Package](https://bcallaway11.github.io/did/articles/did-basics.html)

- [Introduction to DiD with Multiple Time
  Periods](https://bcallaway11.github.io/did/articles/multi-period-did.html)

- [Pre-Testing in a DiD Setup using the did
  Package](https://bcallaway11.github.io/did/articles/pre-testing.html)

- [Writing Extensions to the did
  Package](https://bcallaway11.github.io/did/articles/extensions.html)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Difference-in-Differences <img src="man/figures/logo.png" align="right" alt="" width="155" />

[![](http://cranlogs.r-pkg.org/badges/grand-total/did?color=blue)](https://cran.r-project.org/package=did)
[![](http://cranlogs.r-pkg.org/badges/last-month/did?color=blue)](https://cran.r-project.org/package=did)
[![](https://www.r-pkg.org/badges/version/did?color=blue)](https://cran.r-project.org/package=did)
[![](https://img.shields.io/badge/devel%20version-2.5.1-blue.svg)](https://github.com/bcallaway11/did)
[![CRAN
checks](https://badges.cranchecks.info/summary/did.svg)](https://cran.r-project.org/web/checks/check_results_did.html)
[![](https://img.shields.io/github/last-commit/bcallaway11/did.svg)](https://github.com/bcallaway11/did/commits/master)

<!-- README.md is generated from README.Rmd. Please edit that file -->

The **did** package contains tools for computing average treatment
effect parameters in a Difference-in-Differences setup allowing for

- More than two time periods

- Variation in treatment timing (i.e., units can become treated at
  different points in time)

- Treatment effect heterogeneity (i.e, the effect of participating in
  the treatment can vary across units and exhibit potentially complex
  dynamics, selection into treatment, or time effects)

- The parallel trends assumption holds only after conditioning on
  covariates

The main parameters are **group-time average treatment effects**. These
are the average treatment effect for a particular group (group is
defined by treatment timing) in a particular time period. These
parameters are a natural generalization of the average treatment effect
on the treated (ATT) which is identified in the textbook case with two
periods and two groups to the case with multiple periods.

Group-time average treatment effects are also natural building blocks
for more aggregated treatment effect parameters such as overall
treatment effects or event-study-type estimands.

## Getting Started

There has been some recent work on DiD with multiple time periods. The
**did** package implements the framework put forward in

- [Callaway, Brantly and Pedro H.C. Sant’Anna.
  "Difference-in-Differences with Multiple Time Periods." Journal of
  Econometrics, Vol. 225, No. 2, pp. 200-230,
  2021.](https://doi.org/10.1016/j.jeconom.2020.12.001) or
  [arXiv](https://arxiv.org/abs/1803.09015)

**Higher level discussions of issues are available in**

- [Our approach to DiD with multiple time
  periods](https://bcallaway11.github.io/did/articles/multi-period-did.html)

## Installation

You can install **did** from CRAN with:

``` r
install.packages("did")
```

or get the latest version from github with:

``` r
# install.packages("devtools")
devtools::install_github("bcallaway11/did")
```

## A short example

The following is a simplified example of the effect of states increasing
their minimum wages on county-level teen employment rates which comes
from [Callaway and Sant’Anna
(2021)](https://doi.org/10.1016/j.jeconom.2020.12.001).

- [More detailed examples are also
  available](https://bcallaway11.github.io/did/articles/did-basics.html)

A subset of the data is available in the package and can be loaded by

``` r
library(did)
data(mpdta)
```

The dataset contains 500 observations of county-level teen employment
rates from 2003-2007. Some states are first treated in 2004, some in
2006, and some in 2007 (see the paper for more details). The important
variables in the dataset are

- **lemp** This is the log of county-level teen employment. It is the
  outcome variable

- **first.treat** This is the period when a state first increases its
  minimum wage. It can be 2004, 2006, or 2007. It is the variable that
  defines *group* in this application

- **year** This is the year and is the *time* variable

- **countyreal** This is an id number for each county and provides the
  individual identifier in this panel data context

To estimate group-time average treatment effects, use the **att_gt**
function

``` r
out <- att_gt(
  yname = "lemp",
  gname = "first.treat",
  idname = "countyreal",
  tname = "year",
  xformla = ~1,
  data = mpdta,
  est_method = "reg"
)
```

**att_gt** returns a class **MP** object. This has a lot of information,
but most importantly is has estimates of the group-time average
treatment effects and their standard errors. To see these, we can call
the **summary** function

``` r
summary(out)
#> 
#> Call:
#> att_gt(yname = "lemp", tname = "year", idname = "countyreal", 
#>     gname = "first.treat", xformla = ~1, data = mpdta, est_method = "reg")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> Group-Time Average Treatment Effects:
#>  Group Time ATT(g,t) Std. Error [95% Simult.  Conf. Band]  
#>   2004 2004  -0.0105     0.0246       -0.0752      0.0542  
#>   2004 2005  -0.0704     0.0332       -0.1580      0.0171  
#>   2004 2006  -0.1373     0.0397       -0.2419     -0.0326 *
#>   2004 2007  -0.1008     0.0357       -0.1947     -0.0069 *
#>   2006 2004   0.0065     0.0229       -0.0537      0.0668  
#>   2006 2005  -0.0028     0.0197       -0.0546      0.0491  
#>   2006 2006  -0.0046     0.0179       -0.0519      0.0427  
#>   2006 2007  -0.0412     0.0192       -0.0918      0.0094  
#>   2007 2004   0.0305     0.0154       -0.0100      0.0710  
#>   2007 2005  -0.0027     0.0164       -0.0459      0.0405  
#>   2007 2006  -0.0311     0.0184       -0.0795      0.0173  
#>   2007 2007  -0.0261     0.0170       -0.0708      0.0187  
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> P-value for pre-test of parallel trends assumption:  0.16812
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

This provides estimates of group-time average treatment effects for all
groups in all time periods. Group-time average treatment effects are
identified when `t >= g` (these are post-treatment time periods for each
group), but **summary** reports them even in periods when `t < g` –
these can be used to pre-test for the parallel trends assumption. The
`P-value for pre-test of parallel trends assumption` is for a Wald
pre-test of the parallel trends assumption. Here the parallel trends
assumption would not be rejected at conventional significance levels.

It is often also convenient to plot the group-time average treatment
effects. This can be done using the **ggdid** command:

``` r
ggdid(out, ylim = c(-.25, .1))
```

<img src="man/figures/README-unnamed-chunk-9-1.png" alt="" width="90%" style="display: block; margin: auto;" />

The red dots in the plot are pre-treatment group-time average treatment
effects . Here they are provided with 95% simultaneous confidence
intervals. These are the estimates that can be interpreted as a pre-test
of the parallel trends assumption. The blue dots are post-treatment
group-time average treatment effects. Under the parallel trends
assumption, these can be interpreted as policy effects – here the effect
of the minimum wage on county-level teen employment due to increasing
the minimum wage.

**Event Studies**

Although in the current example it is pretty easy to directly interpret
the group-time average treatment effects, there are many cases where it
is convenient to aggregate the group-time average treatment effects into
a small number of parameters. A main type of aggregation is into an
*event study* plot.

To make an event study plot in the **did** package, one can use the
**aggte** function

``` r
es <- aggte(out, type = "dynamic")
```

Just like for group-time average treatment effects, these can be
summarized and plotted. First, the summary

``` r
summary(es)
#> 
#> Call:
#> aggte(MP = out, type = "dynamic")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on event-study/dynamic aggregation:  
#>      ATT    Std. Error     [ 95%  Conf. Int.]  
#>  -0.0772        0.0219    -0.1201     -0.0344 *
#> 
#> 
#> Dynamic Effects:
#>  Event time Estimate Std. Error [95% Simult.  Conf. Band]  
#>          -3   0.0305     0.0154       -0.0100      0.0710  
#>          -2  -0.0006     0.0132       -0.0352      0.0341  
#>          -1  -0.0245     0.0153       -0.0646      0.0157  
#>           0  -0.0199     0.0123       -0.0522      0.0124  
#>           1  -0.0510     0.0167       -0.0948     -0.0071 *
#>           2  -0.1373     0.0413       -0.2456     -0.0289 *
#>           3  -0.1008     0.0351       -0.1931     -0.0085 *
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

The column `event time` is for each group relative to when they first
participate in the treatment. To give some examples, `event time=0`
corresponds to the *on impact* effect, and `event time=-1` is the
*effect* in the period before a unit becomes treated (checking that this
is equal to 0 is potentially useful as a pre-test).

To plot the event study, use **ggdid**

``` r
ggdid(es)
```

<img src="man/figures/README-unnamed-chunk-12-1.png" alt="" width="90%" style="display: block; margin: auto;" />

The figure here is very similar to the group-time average treatment
effects. Red dots are pre-treatment periods, blue dots are
post-treatment periods. The difference is that the x-axis is in event
time.

**Overall Effect of Participating in the Treatment**

The event study above reported an overall effect of participating in the
treatment. This was computed by averaging the average effects computed
at each length of exposure.

In many cases, a more general purpose overall treatment effect parameter
is give by computing the average treatment effect for each group, and
then averaging across groups. This sort of procedure provides an average
treatment effect parameter with a very similar interpretation to the
Average Treatment Effect on the Treated (ATT) in the two period and two
group case.

To compute this overall average treatment effect parameter, use

``` r
group_effects <- aggte(out, type = "group")
summary(group_effects)
#> 
#> Call:
#> aggte(MP = out, type = "group")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on group/cohort aggregation:  
#>     ATT    Std. Error     [ 95%  Conf. Int.]  
#>  -0.031        0.0126    -0.0558     -0.0062 *
#> 
#> 
#> Group Effects:
#>  Group Estimate Std. Error [95% Simult.  Conf. Band]  
#>   2004  -0.0797     0.0291       -0.1434     -0.0161 *
#>   2006  -0.0229     0.0176       -0.0613      0.0155  
#>   2007  -0.0261     0.0175       -0.0642      0.0121  
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Outcome Regression
```

Of particular interest is the `Overall ATT` in the results. Here, we
estimate that increasing the minimum wage decreased teen employment by
3.1% and the effect is marginally statistically significant.

## Additional Resources

We have provided several more vignettes that may be helpful for using
the **did** package

- [Getting Started with the did
  Package](https://bcallaway11.github.io/did/articles/did-basics.html)

- [Introduction to DiD with Multiple Time
  Periods](https://bcallaway11.github.io/did/articles/multi-period-did.html)

- [Pre-Testing in a DiD Setup using the did
  Package](https://bcallaway11.github.io/did/articles/pre-testing.html)

- [Writing Extensions to the did
  Package](https://bcallaway11.github.io/did/articles/extensions.html)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results="hide", warning=FALSE, message=FALSE-----------------
library(did)
# Source the currently version of the did package (based on our Dropbox)
# fldr <- here::here("R/")
# sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
source(here::here("vignettes/setup_sims.R"))

## ----echo=FALSE---------------------------------------------------------------
time.periods <- 5
reset.sim()
te <- 0
te.e <- c(-1,rep(1,time.periods-1))
bett <- betu <- rep(0,time.periods)
set.seed(12345)

# generate data set with these parameters
dta <- build_sim_dataset()
dta$G <- ifelse(dta$G==0, 0, dta$G+1)
dta$G <- ifelse(dta$G==6, 0, dta$G)

## -----------------------------------------------------------------------------
nrow(dta)
head(dta)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# estimate group-time average treatment effects using att_gt method
# (and ignoring pre-treatment "dip")
attgt.ignoredip <- att_gt(yname = "Y",
                          tname = "period",
                          idname = "id",
                          gname = "G",
                          xformla = ~1,
                          data = dta,
                          )

# summarize the results
summary(attgt.ignoredip)

# make dynamic effects plot
p <- ggdid(aggte(attgt.ignoredip, "dynamic"))

# add actual treatment effects to the plot
library(ggplot2)
truth <- cbind.data.frame(e = seq(-3,2), att.e = c(0,0,-1,1,1,1))
p <- p + geom_line(data = truth, aes(x = e, y = att.e), inherit.aes = FALSE, color = "blue")
p <- p + geom_point(data = truth, aes(x = e, y = att.e), inherit.aes = FALSE, color = "blue")
p

## -----------------------------------------------------------------------------
compute.attgt <- function(dta) {
  # pick up all groups
  groups <- unique(dta$G)

  # pick up all time periods
  time.periods <- unique(dta$period)

  # sort the groups and drop the untreated group
  groups <- sort(groups)[-1]

  # sort the time periods and drop the first two
  # (can't compute treatment effects for these two
  # periods with one-period anticipation -- w/o anticipation
  # we would just drop one period here)
  time.periods <- sort(time.periods)[-c(1,2)]

  # drop last time period (because we don't know if
  # these units are affected by anticipation or not
  # and we are being very careful)
  # (if you were worried about more than one anticipation
  # period here, would need to drop more time periods
  # from the end)
  time.periods <- time.periods[-length(time.periods)]

  # list to store all group-time average treatment effects
  # that we calculate
  attgt.list <- list()
  counter <- 1

  # loop over all groups
  for (g in groups) {

    # get the correct "base" period for this group
    # (subtract 2 to avoid anticipation)
    main.base.period <- g - 2
    
    # loop over all time periods
    for (tp in time.periods) {

      #----------------------------------------------------
      # if it's a pre-treatment time period (used for the
      # pre-test, we need to adjust the base period)

      # group not treated yet
      if (tp < g) {
        # move two periods before
        base.period <- tp - 2
      } else {
        # this is a post-treatment period
        base.period <- main.base.period
      }
      #----------------------------------------------------



      #----------------------------------------------------
      # now, all we need to do is collect the right subset
      # of the data and estimate a 2x2 DiD

      # get group g and untreated group
      this.data <- subset(dta, G==g | G==0)

      # get current period and base period data
      this.data <- subset(this.data, period==tp | period==base.period)

      # set up to compute 2x2 estimator
      Ypost <- subset(this.data, period==tp)$Y
      Ypre <- subset(this.data, period==base.period)$Y

      # dummy variable being in group g
      G <- 1*(subset(this.data, period==tp)$G == g)

      # compute 2x2 estimator using DRDID package
      # (in this unconditional case, it would be straightforward
      # to calculate the 2x2 att just using averages, but we
      # like the DRDID package as it will work for more complicated
      # cases as well)
      attgt <- DRDID::reg_did_panel(Ypost, Ypre, G, covariates=NULL)$ATT

      # save results
      attgt.list[[counter]] <- list(att=attgt, group=g, time.period=tp)
      counter <- counter+1
      #----------------------------------------------------

    }
  }

  #-----------------------------------------------------------------------------
  # aggregate into dynamic effects
  
  # turn results into a data.frame
  attgt.results <- do.call("rbind.data.frame", attgt.list)

  # add event time to the results
  attgt.results$e <- attgt.results$time.period - attgt.results$group
  
  # calculate relative sizes of each group
  # (will be used as weights)
  n.group <- sapply(groups, function(gg) nrow(subset(dta, G==gg)))
  # merge in group sizes
  ngroup.mat <- cbind(groups, n.group)
  attgt.results <- merge(attgt.results, ngroup.mat, by.x = "group", by.y = "groups")

  # event times to calculate dynamic effects
  eseq <- unique(attgt.results$e) 
  eseq <- sort(eseq)

  # calculate average effects by event time
  att.e <- c()
  counter <- 1
  for (this.e in eseq) {
    # get subset of results at this event time
    res.e <- subset(attgt.results, e==this.e)

    # calculate weights by group size
    res.e$weight <- res.e$n.group / sum(res.e$n.group)

    # calculate dynamic effect as weighted average
    att.e[counter] <- sum(res.e$att * res.e$weight)

    # on to the next one
    counter <- counter+1
  }

  # store dynamic effects results
  dyn.results <- cbind.data.frame(e = eseq, att.e = att.e)

  # return group-time average treatment effects and dynamic effects
  return(list(attgt.results=attgt.results[,c("group","att","time.period")],
              dyn.results=dyn.results))
}

## -----------------------------------------------------------------------------
anticipation.results <- compute.attgt(dta)
anticipation.results

## ----bootstrap, eval=FALSE----------------------------------------------------
# # the number of bootstrap iterations
# biters <- 100
# 
# # list to store bootstrap results
# boot.res <- list()
# 
# # loop for each nonparametric bootstrap iteration
# for (b in 1:biters) {
#   # draw a bootstrap sample; here, we'll call an outside function
#   bdata <- BMisc::blockBootSample(dta, "id")
# 
#   # call our function for estimating dynamic effects on the
#   # bootstrapped data
#   boot.res[[b]] <- compute.attgt(bdata)$dyn.results$att.e
# }
# 
# # use the bootstrapped results to compute standard errors
# boot.res <- t(simplify2array(boot.res))
# boot.se <- apply(boot.res, 2, sd)
# 
# # add the standard errors to the main results
# dyn.results <- anticipation.results$dyn.results
# dyn.results$att.se <- boot.se

## ----eval=FALSE---------------------------------------------------------------
# p <- ggplot(data=dyn.results, aes(x=e, y=att.e)) +
#   geom_line() +
#   geom_point() +
#   geom_errorbar(aes(ymin=(att.e-1.96*att.se), ymax=(att.e+1.96*att.se)), width=0.1) +
#   theme_bw()
# 
# p <- p + geom_line(data=truth, aes(x=e, y=att.e), inherit.aes=FALSE, color="blue")
# p <- p + geom_point(data=truth, aes(x=e, y=att.e), inherit.aes=FALSE, color="blue")
# p
# 

## ----setup2, include = FALSE--------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  echo = TRUE,
  eval = FALSE
)

## ----DGP_viz, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # Load libraries and set baseline parameters
# library(tidyverse)
# library(lfe)
# library(fastDummies)
# library(ggthemes)
# library(did)
# theme_set(theme_clean() + theme(plot.background = element_blank()))
# #----------------------------------------------------------------------------
# iseed  = 20201221
# nrep <- 100
# true_mu <- 1
# set.seed(iseed)
# 
# #----------------------------------------------------------------------------
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobservd heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into four treatment groups G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse(year >= cohort_year, 1, 0),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error)
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data()
# 
# # plot
# plot1 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value", color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))  +
#   ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with all groups being eventually treated")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot1
# 
# #ggsave("plot_dgp1.png", plot1, width = 10, height = 5, dpi = 100)

## ----ES1, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse(rel_year < -5, 1, 0),
#            Post = ifelse(rel_year > 5, 1, 0))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical <- map_dfr(1:nrep, run_ES_DiD)
# 
# colors <- c("True Effect" = "red", "Estimated Effect" = "blue")
# 
# ES_plot_classical <- data_classical %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#   geom_line(aes(color = 'Estimated Effect'), size = 1) +
#   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("TWFE event-study regression with binned end-points")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical
# 
# #ggsave("es_plot_classical.png", ES_plot_classical, width = 10, height = 5, dpi = 150)

## ----ES3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year")
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat <- map_dfr(1:nrep, run_ES_DiD_sat)
# 
# ES_plot_sat <- data_sat %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# 
# ES_plot_sat
# 
# #ggsave("es_plot_sat.png", ES_plot_sat, width = 10, height = 5, dpi = 150)
# 

## ----CS, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS <- map_dfr(1:nrep, run_CS)
# 
# ES_plot_CS <- data_CS %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated")+
#     scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS
# 
# #ggsave("es_plot_CS.png", ES_plot_CS, width = 10, height = 5, dpi = 150)
# 

## ----DGP_viz2, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data2 <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobservd heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into 1\{t \ge g \}G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
#     # Relabel 2004 cohort as never-treated
#     mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data2()
# 
# # plot
# plot2 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value",  color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
#                      values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3"))+
#   ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with a never-treated group")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot2
# 
# #ggsave("plot_dgp_never_treated.png",plot2, width = 10, height = 5, dpi = 100)

## ----ES1_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>%
#     dummy_cols(select_columns = "rel_year") %>%
#     mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
#            Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
#     mutate(Pre = ifelse(is.na(Pre), 0, Pre),
#            Post = ifelse(is.na(Post), 0, Post))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#  # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical_never <- map_dfr(1:nrep, run_ES_DiD_never)
# 
# ES_plot_classical_never <- data_classical_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with binned end-points")+
#     scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical_never
# 
# #ggsave("es_plot_classical_never.png", ES_plot_classical_never, width = 10, height = 5, dpi = 150)

## ----ES3_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year2 = rel_year) %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     select(-("rel_year_-Inf"))
# 
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#  # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat_never <- map_dfr(1:nrep, run_ES_DiD_sat_never)
# 
# ES_plot_sat_never <- data_sat_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_sat_never
# 
# #ggsave("es_plot_sat_never.png", ES_plot_sat_never, width = 10, height = 5, dpi = 150)
# 

## ----CS_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "never_treated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_never <- map_dfr(1:nrep, run_CS_never)
# 
# ES_plot_CS_never <- data_CS_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_never
# 
# #ggsave("es_plot_CS_never.png", ES_plot_CS_never, width = 10, height = 5, dpi = 150)

## ----CS_ny, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# # function to run ES DID
# run_CS_ny <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_ny <- map_dfr(1:nrep, run_CS_ny)
# 
# ES_plot_CS_ny <- data_CS_ny %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_ny
# 
# #ggsave("es_plot_CS_ny.png", ES_plot_CS_ny, width = 10, height = 5, dpi = 150)

## ----DGP_viz3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data3 <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobservd heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into 1\{t \ge g \}G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
#            mu = ifelse(cohort_year==1992, 2, ifelse(cohort_year==1998, 1, 3)),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
#     # Relabel 2004 cohort as never-treated
#     mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data3()
# 
# # plot
# plot3 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value",  color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
#                      values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3")) +
#   ggtitle("One draw of the DGP with heterogeneous treatment effect dynamics across cohorts \n and with a never-treated group")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot3
# 
# #ggsave("plot_dgp_never_treated_het.png", plot3, width = 10, height = 5, dpi = 100)

## ----ES1_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>%
#     dummy_cols(select_columns = "rel_year") %>%
#     mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
#            Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
#     mutate(Pre = ifelse(is.na(Pre), 0, Pre),
#            Post = ifelse(is.na(Post), 0, Post))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
# # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical_never_het <- map_dfr(1:nrep, run_ES_DiD_never_het)
# 
# ES_plot_classical_never_het <- data_classical_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with binned end-points")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical_never_het
# 
# #ggsave("es_plot_classical_never_het.png", ES_plot_classical_never_het, width = 10, height = 5, dpi = 150)

## ----ES3_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year2 = rel_year) %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     select(-("rel_year_-Inf"))
# 
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
# # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat_never_het <- map_dfr(1:nrep, run_ES_DiD_sat_never_het)
# 
# ES_plot_sat_never_het <- data_sat_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_sat_never_het
# 
# #ggsave("es_plot_sat_never_het.png", ES_plot_sat_never_het, width = 10, height = 5, dpi = 150)
# 

## ----CS_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "never_treated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_never_het <- map_dfr(1:nrep, run_CS_never_het)
# 
# ES_plot_CS_never_het <- data_CS_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_never_het
# 
# #ggsave("es_plot_CS_never_het.png", ES_plot_CS_never_het, width = 10, height = 5)

## ----CS_ny_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# # function to run ES DID
# run_CS_ny_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_ny_het <- map_dfr(1:nrep, run_CS_ny_het)
# 
# ES_plot_CS_ny_het <- data_CS_ny_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_ny_het
# 
# #ggsave("es_plot_CS_ny_het.png", ES_plot_CS_ny_het, width = 10, height = 5, dpi = 150)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results="hide", warning=FALSE, message=FALSE-----------------
library(did)
# Source the currently version of the did package (based on our Dropbox)
#fldr <- here::here("R/")
#sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
#source(here::here("vignettes/setup_sims.R"))

## ----message = FALSE, warning = FALSE-----------------------------------------
# set seed so everything is reproducible
set.seed(1814)

# generate dataset with 4 time periods
sp <- reset.sim()
sp$te <- 0
time.periods <- 4

# add dynamic effects
sp$te.e <- 1:time.periods

# generate data set with these parameters
# here, we dropped all units who are treated in time period 1 as they do not help us recover ATT(g,t)'s.
dta <- build_sim_dataset(sp)

# How many observations remained after dropping the ``always-treated'' units
nrow(dta)
#This is what the data looks like
head(dta)

## -----------------------------------------------------------------------------
# estimate group-time average treatment effects using att_gt method
example_attgt <- att_gt(yname = "Y",
                        tname = "period",
                        idname = "id",
                        gname = "G",
                        xformla = ~X,
                        data = dta
                        )

# summarize the results
summary(example_attgt)

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# plot the results
ggdid(example_attgt)

## -----------------------------------------------------------------------------
agg.simple <- aggte(example_attgt, type = "simple")
summary(agg.simple)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
agg.es <- aggte(example_attgt, type = "dynamic")
summary(agg.es)
ggdid(agg.es)

## ----fig.width=6, fig.align='center', out.width="90%", dpi = 200--------------
agg.gs <- aggte(example_attgt, type = "group")
summary(agg.gs)
ggdid(agg.gs)

## ----fig.width=8, fig.height=5, fig.align='center', out.width="90%", dpi = 200----
agg.ct <- aggte(example_attgt, type = "calendar")
summary(agg.ct)
ggdid(agg.ct)

## ----eval=FALSE---------------------------------------------------------------
# example_attgt_altcontrol <- att_gt(yname = "Y",
#                                    tname = "period",
#                                    idname = "id",
#                                    gname = "G",
#                                    xformla = ~X,
#                                    data = dta,
#                                    control_group = "notyettreated"			
#                                    )
# summary(example_attgt_altcontrol)

## ----eval=FALSE---------------------------------------------------------------
# example_attgt_reg <- att_gt(yname = "Y",
#                             tname = "period",
#                             idname = "id",
#                             gname = "G",
#                             xformla = ~X,
#                             data = data,
#                             est_method = "reg"
#                             )
# summary(example_attgt_reg)

## -----------------------------------------------------------------------------
data(mpdta)

## -----------------------------------------------------------------------------
head(mpdta)

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-time average treatment effects without covariates
mw.attgt <- att_gt(yname = "lemp",
                   gname = "first.treat",
                   idname = "countyreal",
                   tname = "year",
                   xformla = ~1,
                   data = mpdta
                   )

# summarize the results
summary(mw.attgt)

# plot the results
# set ylim so that all plots have the same scale along y-axis
ggdid(mw.attgt, ylim = c(-.3, .3))

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate the group-time average treatment effects
mw.dyn <- aggte(mw.attgt, type = "dynamic")
summary(mw.dyn)
ggdid(mw.dyn, ylim = c(-.3, .3))

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
mw.dyn.balance <- aggte(mw.attgt, type = "dynamic", balance_e = 1)
summary(mw.dyn.balance)
ggdid(mw.dyn.balance, ylim = c(-.3, .3))

## ----eval=FALSE---------------------------------------------------------------
# mw.attgt.X <- att_gt(yname = "lemp",
#                      gname = "first.treat",
#                      idname = "countyreal",
#                      tname = "year",
#                      xformla = ~lpop,
#                      data = mpdta
#                      )

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results="hide", warning=FALSE, message=FALSE-----------------
library(did)
# Source the currently version of the did package (based on our Dropbox)
# fldr <- here::here("R/")
# sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
source(here::here("vignettes/setup_sims.R"))

## ----echo=FALSE---------------------------------------------------------------
time.periods <- 4
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)

## -----------------------------------------------------------------------------
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()

head(data)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

head(data)

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)

# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways",
          index = c("id", "period"))

summary(es)

#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs=coefs, ses=ses, exposure=exposure)

library(ggplot2)

ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-group time average treatment effects
did_att_gt <- att_gt(yname = "Y",
                     tname = "period",
                     idname = "id",
                     gname = "G",
                     data = data,
                     bstrap = FALSE,
                     cband = FALSE)
summary(did_att_gt)

# plot them
ggdid(did_att_gt)


## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate them into event study plot
did_es <- aggte(did_att_gt, type = "dynamic")

# plot the event study
ggdid(did_es)

## ----echo=FALSE---------------------------------------------------------------
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)

## -----------------------------------------------------------------------------
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate selective treatment timing
# (*** this is what is different here ***)
te.bet.ind <- time.periods:1 / (time.periods/2)

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()

## -----------------------------------------------------------------------------
# run through same code as in earlier example...omitted

## ----echo=FALSE---------------------------------------------------------------
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)

## -----------------------------------------------------------------------------
# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways", 
          index = c("id", "period"))

summary(es)

## ----echo=FALSE---------------------------------------------------------------
#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs = coefs, ses = ses, exposure = exposure)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# run through same code as before...omitted

# new event study plot
ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-group time average treatment effects
did.att.gt <- att_gt(yname = "Y",
                     tname = "period",
                     idnam = "id",
                     gname = "G",
                     data = data
                     )
summary(did.att.gt)

# plot them
ggdid(did.att.gt)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate them into event study plot
did.es <- aggte(did.att.gt, type = "dynamic")

# plot the event study
ggdid(did.es)

## ----eval=FALSE---------------------------------------------------------------
# # not run (this code can be substantially slower)
# reset.sim()
# set.seed(1814)
# nt <- 1000
# nu <- 1000
# cdp <- conditional_did_pretest("Y", "period", "id", "G", xformla = ~X, data = data)
# cdp

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results="hide", warning=FALSE, message=FALSE-----------------
library(did)

## ----message = FALSE, warning = FALSE-----------------------------------------
# set seed so everything is reproducible
set.seed(1814)

# generate dataset with 4 time periods
sp <- reset.sim()
sp$te <- 0
time.periods <- 4

# add dynamic effects
sp$te.e <- 1:time.periods

# generate data set with these parameters
# here, we dropped all units who are treated in time period 1 as they do not help us recover ATT(g,t)'s.
dta <- build_sim_dataset(sp)

# How many observations remained after dropping the ``always-treated'' units
nrow(dta)
#This is what the data looks like
head(dta)

## -----------------------------------------------------------------------------
# estimate group-time average treatment effects using att_gt method
example_attgt <- att_gt(yname = "Y",
                        tname = "period",
                        idname = "id",
                        gname = "G",
                        xformla = ~X,
                        data = dta
                        )

# summarize the results
summary(example_attgt)

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# plot the results
ggdid(example_attgt)

## -----------------------------------------------------------------------------
agg.simple <- aggte(example_attgt, type = "simple")
summary(agg.simple)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
agg.es <- aggte(example_attgt, type = "dynamic")
summary(agg.es)
ggdid(agg.es)

## ----fig.width=6, fig.align='center', out.width="90%", dpi = 200--------------
agg.gs <- aggte(example_attgt, type = "group")
summary(agg.gs)
ggdid(agg.gs)

## ----fig.width=8, fig.height=5, fig.align='center', out.width="90%", dpi = 200----
agg.ct <- aggte(example_attgt, type = "calendar")
summary(agg.ct)
ggdid(agg.ct)

## ----eval=FALSE---------------------------------------------------------------
# example_attgt_altcontrol <- att_gt(yname = "Y",
#                                    tname = "period",
#                                    idname = "id",
#                                    gname = "G",
#                                    xformla = ~X,
#                                    data = dta,
#                                    control_group = "notyettreated"			
#                                    )
# summary(example_attgt_altcontrol)

## ----eval=FALSE---------------------------------------------------------------
# example_attgt_reg <- att_gt(yname = "Y",
#                             tname = "period",
#                             idname = "id",
#                             gname = "G",
#                             xformla = ~X,
#                             data = dta,
#                             est_method = "reg"
#                             )
# summary(example_attgt_reg)

## -----------------------------------------------------------------------------
data(mpdta)

## -----------------------------------------------------------------------------
head(mpdta)

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-time average treatment effects without covariates
mw.attgt <- att_gt(yname = "lemp",
                   gname = "first.treat",
                   idname = "countyreal",
                   tname = "year",
                   xformla = ~1,
                   data = mpdta
                   )

# summarize the results
summary(mw.attgt)

# plot the results
# set ylim so that all plots have the same scale along y-axis
ggdid(mw.attgt, ylim = c(-.3, .3))

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate the group-time average treatment effects
mw.dyn <- aggte(mw.attgt, type = "dynamic")
summary(mw.dyn)
ggdid(mw.dyn, ylim = c(-.3, .3))

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
mw.dyn.balance <- aggte(mw.attgt, type = "dynamic", balance_e = 1)
summary(mw.dyn.balance)
ggdid(mw.dyn.balance, ylim = c(-.3, .3))

## ----eval=FALSE---------------------------------------------------------------
# mw.attgt.X <- att_gt(yname = "lemp",
#                      gname = "first.treat",
#                      idname = "countyreal",
#                      tname = "year",
#                      xformla = ~lpop,
#                      data = mpdta
#                      )

#' @title MP
#'
#' @description Multi-period objects that hold results for group-time average treatment effects
#'
#' @param group which group (defined by period first treated) an group-time average treatment effect is for
#' @param t which time period a group-time average treatment effect is for
#' @param att the group-average treatment effect for group \code{group} and time
#'  period \code{t}
#' @param c simultaneous critical value if one is obtaining simultaneous confidence
#'  bands. Otherwise it reports the critical value based on pointwise normal
#'  approximation.
#' @param V_analytical Analytical estimator for the asymptotic variance-covariance matrix for group-time average treatment effects
#' @param se standard errors for group-time average treatment effects. If bootstrap is set to TRUE, this provides bootstrap-based se.
#' @param inffunc the influence function for estimating group-time average
#'  treatment effects: one column per ATT(g,t) and one row per cross-sectional
#'  unit (one row per observation with repeated cross sections). The rownames
#'  hold the unit ids and are the authoritative link between rows and units --
#'  the row order differs across `faster_mode = TRUE` (internal
#'  (period, cohort, id) ordering) and `faster_mode = FALSE` (id-sorted), so
#'  align rows by rowname, never by position.
#' @param n the number of unique cross-sectional units (unique values of idname)
#' @param W the Wald statistic for pre-testing the common trends assumption
#' @param Wpval the p-value of the Wald statistic for pre-testing the
#'  common trends assumption
#' @param aggte an aggregate treatment effects object
#' @param alp the significance level, default is 0.05
#' @param DIDparams a [`DIDparams`] object.  A way to optionally return the parameters
#'  of the call to [att_gt()] or [conditional_did_pretest()].
#'
#' @return MP object
#' @export
MP <- function(group, t, att, V_analytical, se, c, inffunc, n=NULL, W=NULL, Wpval=NULL, aggte=NULL, alp = 0.05, DIDparams=NULL) {
  out <- list(group=group, t=t, att=att, V_analytical=V_analytical, se=se, c=c,
  inffunc=inffunc, n=n, W=W, Wpval=Wpval, aggte=aggte, alp = alp,
  DIDparams=DIDparams, call=DIDparams$call)
  class(out) <- "MP"
  out
}

#' @title summary.MP
#'
#' @description prints a summary of a \code{MP} object
#'
#' @param object an \code{MP} object
#' @param ... extra arguments
#'
#' @export
summary.MP <- function(object, ...) {
  mpobj <- object

  # call
  cat("\n")
  cat("Call:\n")
  print(mpobj$DIDparams$call)
  cat("\n")

  # citation
  citation()
  cat("\n")

  # group time average treatment effects
  cat("Group-Time Average Treatment Effects:\n")

  cband_text1a <- paste0(100*(1-mpobj$alp),"% ")
  cband_text1b <- if (mpobj$DIDparams$bstrap) {
    if (mpobj$DIDparams$cband) "Simult. " else "Pointwise "
  } else {
    "Pointwise "
  }
  cband_text1 <- paste0("[", cband_text1a, cband_text1b)

  cband_lower <- mpobj$att - mpobj$c*mpobj$se
  cband_upper <- mpobj$att + mpobj$c*mpobj$se

  sig <- (cband_upper < 0) | (cband_lower > 0)
  sig[is.na(sig)] <- FALSE
  sig_text <- ifelse(sig, "*", "")

  out <- cbind.data.frame(mpobj$group, mpobj$t, mpobj$att, mpobj$se, cband_lower, cband_upper)
  out <- round(out,4)
  out <- cbind.data.frame(out, sig_text)


  colnames(out) <- c("Group", "Time", "ATT(g,t)","Std. Error", cband_text1, "Conf. Band]", "")
  print(out, row.names=FALSE)
  cat("---\n")
  cat("Signif. codes: `*' confidence band does not cover 0")
  cat("\n\n")

  # report pre-test
  if (!is.null(mpobj$Wpval)) {
    cat("P-value for pre-test of parallel trends assumption:  ")
    cat(as.character(mpobj$Wpval))
    cat("\n")
  }



  # set control group text
  control_group <- mpobj$DIDparams$control_group
  control_group_text <- NULL
  if (control_group == "nevertreated") {
    control_group_text <- "Never Treated"
  } else if (control_group == "notyettreated") {
    control_group_text <- "Not Yet Treated"
  }

  if (!is.null(control_group)) {
    cat("Control Group:  ")
    cat(control_group_text)
    cat(",  ")
  }

  # anticipation periods
  cat("Anticipation Periods:  ")
  cat(mpobj$DIDparams$anticipation)
  cat("\n")

  # estimation method text
  est_method <- mpobj$DIDparams$est_method
  if ( is(est_method,"character") ) {
    est_method_text <- est_method
    if (est_method == "dr") {
      est_method_text <- "Doubly Robust"
    } else if (est_method == "ipw") {
      est_method_text <- "Inverse Probability Weighting"
    } else if (est_method == "reg") {
      est_method_text <- "Outcome Regression"
    }

    cat("Estimation Method:  ")
    cat(est_method_text)
    cat("\n")
  }
}

#' @title print.MP
#'
#' @description prints value of a \code{MP} object
#'
#' @param x a \code{MP} object
#' @param ... extra arguments
#'
#' @export
print.MP <- function(x,...) {
  summary.MP(x,...)
}

#' @title Pre-Test of Conditional Parallel Trends Assumption
#'
#' @description An integrated moments test for the conditional parallel trends
#'  assumption holding in all pre-treatment time periods for all groups
#'
#' @inheritParams att_gt
#' @param weightsname The name of the column containing the sampling weights.
#'  If not set, all observations have same weight.
#' @param est_method the method to compute group-time average treatment
#'  effects.  The default for `conditional_did_pretest` is "ipw" for
#'  inverse probability weighting.  Other built-in methods include "dr"
#'  which uses the doubly robust approach in the `DRDID` package and "reg"
#'  for first step regression estimators.
#' @param bstrap Not used by the pre-test.  Critical values and p-values
#'  for the Cramer von Mises and Kolmogorov-Smirnov test statistics are
#'  always computed with the multiplier bootstrap (using `biters`
#'  iterations), regardless of this argument, and no standard errors are
#'  reported.
#' @param biters The number of multiplier bootstrap iterations used to
#'  simulate the limiting distribution of the test statistics.  The
#'  default is 1000.
#' @param clustervars A vector of variables names to cluster on (the
#'  multiplier bootstrap then draws cluster-level multipliers).  At most,
#'  there can be two variables (otherwise will throw an error) and one of
#'  these must be the same as idname which allows for clustering at the
#'  individual level.
#' @param cores The number of cores to use for parallel processing. This
#'  parallelizes Step 1 (computing the test statistic); Step 2's multiplier
#'  bootstrap is vectorized and runs in a single process.
#'
#' @examples
#' \dontrun{
#' data(mpdta)
#' pre.test <- conditional_did_pretest(yname="lemp",
#'                                     tname="year",
#'                                     idname="countyreal",
#'                                     gname="first.treat",
#'                                     xformla=~lpop,
#'                                     data=mpdta)
#' summary(pre.test)
#' }
#'
#' @references Callaway, Brantly and Sant'Anna, Pedro H. C.  "Difference-in-Differences with Multiple Time Periods and an Application on the Minimum Wage and Employment." Working Paper <https://arxiv.org/abs/1803.09015v2> (2018).
#'
#' @return an \code{\link{MP.TEST}} object
#' @export
conditional_did_pretest <- function(yname,
                                    tname,
                                    idname=NULL,
                                    gname,
                                    xformla=NULL,
                                    data,
                                    panel=TRUE,
                                    allow_unbalanced_panel=FALSE,
                                    control_group=c("nevertreated","notyettreated"),
                                    weightsname=NULL,
                                    alp=0.05,
                                    bstrap=TRUE,
                                    cband=TRUE,
                                    biters=1000,
                                    clustervars=NULL,
                                    est_method="ipw",
                                    print_details=FALSE,
                                    pl=FALSE,
                                    cores=1) {

  message("We are no longer updating this function.  It should continue to work, but most users find the pre-tests already reported by the `att_gt` function to be sufficient for most empirical applications.")

  # this is a DIDparams object
  dp <- pre_process_did(yname=yname,
                        tname=tname,
                        idname=idname,
                        gname=gname,
                        xformla=xformla,
                        data=data,
                        panel=panel,
                        allow_unbalanced_panel=allow_unbalanced_panel,
                        control_group=control_group,
                        weightsname=weightsname,
                        alp=alp,
                        bstrap=bstrap,
                        cband=cband,
                        biters=biters,
                        clustervars=clustervars,
                        est_method=est_method,
                        print_details=print_details,
                        pl=pl,
                        cores=cores
  )


  data <- dp$data
  tlist <- dp$tlist
  glist <- dp$glist
  n <- dp$n

  # check if possible to do test
  # note: tlist[2] contains the 2nd time period
  # (which is the first period where able to calculate ATT(g,t)'s)
  if ( max(glist) <= tlist[2] ) {
    stop("There are no pre-treatment periods available to conduct the conditional pre-test. Ensure your data has sufficient pre-treatment periods.")
  }


  # set which weight function to use
  # the only option that will work with current setup
  # is indicator so hard-code it here
  weightfun <- indicator


  if (allow_unbalanced_panel) {
    stop("The conditional pre-test is not supported for unbalanced panel data. Set `allow_unbalanced_panel = FALSE` to balance the panel, or use the pre-test reported by att_gt() instead.")
  }

  # create dataset with n observations;
  # recover covariates from this dataset
  if (panel & (allow_unbalanced_panel == FALSE)) {
    dta <- data[ data[,tname]==tlist[1], ]
  } else {
    dta <- data
  }

  # need X twice, once to loop over, and once to calculate weighting function
  X <- model.matrix(xformla, dta)
  X1 <- X

  # Internal compute.att_gt speedups for the per-unit loop below:
  #  (1) pretreatment_cells_only -- post-treatment (g,t) cells are dropped
  #      right after the loop (the keepers filter retains only group > t), so
  #      compute.att_gt never computes them.
  #  (2) setup_precomp / y_override -- compute.att_gt's per-call setup (data
  #      copy, design matrix, per-period slicing, alignment check) depends on
  #      the outcome only through the per-period outcome slices, and the loop
  #      below only modifies the outcome column; build the y-invariant bundle
  #      once and pass each unit's modified outcome as a plain vector instead
  #      of copying the whole data.frame n times. Used only when the balanced,
  #      id-aligned panel precompute applies; otherwise the legacy data-copy
  #      path inside the loop is kept.
  # Both are bit-identical to the legacy loop (same kept cells, same inputs to
  # every 2x2 estimator, and compute.att_gt consumes no RNG).
  basedp <- dp
  basedp$print_details <- FALSE
  basedp$pretreatment_cells_only <- TRUE
  use_y_override <- FALSE
  if (dp$panel) {
    setupdp <- basedp
    setupdp$return_setup <- TRUE
    setup_precomp <- compute.att_gt(setupdp)
    if (!is.null(setup_precomp) && isTRUE(setup_precomp$use_precompute_panel)) {
      basedp$setup_precomp <- setup_precomp
      use_y_override <- TRUE
      id_col <- data[, idname]
      y_col <- data[, yname]
    }
  }

  cat("Step 1 of 2: Computing test statistic....\n")
  out <- pbapply::pblapply(1:nrow(X), function(i) {
    # these are the weights for the conditional moment test
    # indicator weights
    www <- as.numeric(weightfun(X1, X[i,]))
    # for indicator weights, just choose rows where weights = 1
    rightids <- dta[,idname][www==1]

    # set new parameters to pass to call to compute.att_gt
    thisdp <- basedp

    # set the outcome to be the outcomes multiplied by the weighting function
    # (***this *trick* is only going to work for indicator weights***
    # (otherwise you will multiply by www twice))
    if (use_y_override) {
      # pass only the modified outcome vector; the y-invariant setup bundle
      # built above is reused by compute.att_gt
      ynew <- numeric(length(y_col))
      sel <- id_col %in% rightids
      ynew[sel] <- y_col[sel]
      thisdp$y_override <- ynew
    } else {
      # legacy path: copy the whole dataset and overwrite the outcome column
      thisdata <- data
      thisdata[,yname] <- 0
      thisdata[ thisdata[,idname] %in% rightids,yname] <- data[ thisdata[,idname] %in% rightids,yname]
      thisdata$y <- thisdata[,yname]
      thisdp$data <- thisdata
    }

    # compute the test statistic with call to compute.att_gt
    Jres <- compute.att_gt(thisdp)

    # turn in into more usable format
    J.results <- process_attgt(Jres$attgt.list)
    group <- J.results$group
    att <- J.results$att
    tt <- J.results$tt
    inf.func <- Jres$inffunc

    # return the results for this weighting function
    list(J=att, group=group, t=tt, inf.func=inf.func)
  }, cl=cores)

  # drop post-treatment (g,t). compute.att_gt now skips these cells entirely
  # (pretreatment_cells_only above), so this filter is a no-op retained as a
  # safety net in case a code path reintroduces post-treatment cells.
  out <- lapply(out, function(Js) {
    # which elements of results to keep
    keepers <- Js$group > Js$t
    this.group <- Js$group[keepers]
    this.t <- Js$t[keepers]
    this.J <- as.matrix(Js$J[keepers])
    this.inf.func <- as.matrix(Js$inf.func[,keepers])
    list(J=this.J, group=this.group, t=this.t, inf.func=this.inf.func)
  })

  # grab an array of the influence function for the test statistic
  Jinf.func <- simplify2array(BMisc::get_list_element(out, "inf.func"))

  # get the test statistic for all values of g,t,x
  J.inner <- sapply(out, function(o) o$J)

  # Orient J as (nX rows = covariate threshold values) x (n_gt cols = pre-treatment
  # (g,t) cells) so the observed CvM statistic below uses the SAME (1/nX)
  # normalization as the bootstrap null distribution in test.mboot() (which forms
  # Jb = t(apply(Ub*inf.func, c(2,3), mean)) and is therefore (nX x n_gt)).
  #
  # sapply() returns an (n_gt x nX) MATRIX when there are multiple pre-treatment
  # cells, but a length-nX VECTOR when there is only one -- so transpose the matrix
  # case and column-ify the vector case.
  #
  # BUG FIX (gh issue: pre-test spuriously rejects): this was previously
  #   ifelse(class(J.inner) == "matrix", J <- t(J.inner), J <- as.matrix(J.inner))
  # which was correct under R < 4.0 but silently broke in R >= 4.0, where
  # class(<matrix>) became c("matrix","array") (length 2). ifelse() then evaluated
  # BOTH assignment branches and the second (as.matrix, no transpose) always won, so
  # for the multi-cell case J stayed (n_gt x nX). Because sum(apply(J^2, 2, mean)) =
  # (1/nrow(J)) * sum(J^2), that mis-orientation scaled the observed CvM by
  # nX/n_gt = n/n_gt relative to the bootstrap, driving the p-value to ~0 and making
  # the pre-test over-reject whenever there was more than one pre-treatment (g,t)
  # cell. Use is.matrix() to restore the intended orientation.
  if (is.matrix(J.inner)) {
    J <- t(J.inner)
  } else {
    J <- as.matrix(J.inner)
  }

  # compute CvM test statistic by averaging over X, and summing over g and t
  CvM <- n*sum(apply(J^2, 2, mean))


  #-----------------------------------------------------------------------------
  # use the multiplier bootstrap to simulate limiting distribution of
  # test statistic
  cat("Step 2 of 2: Simulating limiting distribution of test statistic....\n")
  boot.res <- test.mboot(Jinf.func, dp, cores=cores)


  #-----------------------------------------------------------------------------
  ## # some debugging code
  ## # keeping in case helpful later on
  ## ddd <- 1
  ## bout <- sapply(1:100, function(b) {

  ##   Jstar <- apply(sample(c(-1,1), size=n, replace=TRUE)*Jinf.func[,ddd,], 2, mean)
  ##   mean((sqrt(n)*Jstar)^2)
  ## })

  ## ts <- mean((sqrt(n)*J[,ddd])^2)
  ## 1-ecdf(bout)(ts)


  ## ts2.inner <- apply( (sqrt(n)*J)^2, 2, mean)
  ## ts2 <- sum(ts2.inner)
  ## bout2 <- sapply(1:100, function(b) {

  ##   Jstar <- t(apply(sample(c(-1,1), size=n, replace=TRUE)*Jinf.func[,,], c(2,3), mean))
  ##   cvm.inner <- apply((sqrt(n)*Jstar)^2, 2, mean)
  ##   sum(cvm.inner)
  ## })
  ## 1-ecdf(bout2)(ts2)
  ## #-----------------------------------------------------------------------------

  # bootstrap results
  CvMb <- boot.res$bres
  # bootstrap critical value
  CvM.crit.val <- boot.res$crit.val
  # bootstrap p-value
  CvMpval <- 1-ecdf(CvMb)(CvM)

  #-----------------------------------------------------------------------------
  # KS Test - not reporting any more
  #
  # KSb <- unlist(lapply(bout, function(b1) b1$KSb))
  # KSocval <- quantile(KSb, probs=(1-alp), type=1)
  # KSpval <- 1-ecdf(KSb)(KS)
  #-----------------------------------------------------------------------------

  # return test results
  out <- MP.TEST(CvM=CvM, CvMb=CvMb, CvMcval=CvM.crit.val, CvMpval=CvMpval, clustervars=clustervars, xformla=xformla)
  out
}



## #' @title expf
## #'
## #' @description exponential weighting function
## #'
## #' @param X matrix of X's from the data
## #' @param u a particular value to multiply times the X's
## #'
## #' @return numeric vector
## #' @examples
## #' data(mpdta)
## #' dta <- subset(mpdta, year==2007)
## #' X <- model.matrix(~lpop, data=dta)
## #' X <- expf(X, X[1,])
## #'
## #' @export
## expf <- function(X, u) {
##     exp(X%*%u)
## }

#' @title indicator
#'
#' @description indicator weighting function
#'
#' @param X matrix of X's from the data
#' @param u a particular value to compare X's to
#'
#' @return numeric vector
#'
#' @examples
#' data(mpdta)
#' dta <- subset(mpdta, year==2007)
#' X <- model.matrix(~lpop, data=dta)
#' X <- indicator(X, X[1,])
#'
#' @export
indicator <- function(X, u) {
  # Row i is 1 iff X[i, j] <= u[j] for EVERY column j. Vectorized: rep(u, each =
  # nrow(X)) lines u up with X's column-major layout, so (X <= .) is the elementwise
  # comparison and a row passes iff none of its entries exceed u (rowSums == ncol).
  # Bit-identical to the prior t(apply(X, 1, `<=`)) + apply(all) for finite X
  # (model.matrix output is finite), but O(n*p) without the per-row closure.
  1 * (rowSums(X <= rep(u, each = nrow(X))) == ncol(X))
}



#' @title Multiplier Bootstrap for Conditional Moment Test
#'
#' @description A slightly modified multiplier bootstrap procedure
#'  for the pre-test of the conditional parallel trends assumption
#'
#' @inheritParams mboot
#' @param cores Unused; retained for backward compatibility. The
#'  multiplier bootstrap is computed with vectorized matrix operations
#'  in a single process, so this argument has no effect here. In
#'  \code{conditional_did_pretest}, \code{cores} parallelizes only
#'  Step 1 (computing the test statistic).
#'
#' @return list
#' \item{bres}{CvM test statistics for each bootstrap iteration}
#' \item{crit.val}{critical value for CvM test statistic}
#'
#' @export
test.mboot <- function(inf.func, DIDparams, cores=1) {

  # setup needed variables
  data <- DIDparams$data
  idname <- DIDparams$idname
  clustervars <- DIDparams$clustervars
  biters <- DIDparams$biters
  tname <- DIDparams$tname
  tlist <- unique(data[,tname])[order(unique(data[,tname]))]
  alp <- DIDparams$alp
  panel <- DIDparams$panel

  # just get n obsevations (for clustering below...)
  if (panel) {
    dta <- data[ data[,tname]==tlist[1], ]
  } else {
    dta <- data
  }
  n <- nrow(dta)

  # if include id as variable to cluster on
  # drop it as we do this automatically
  if (idname %in% clustervars) {
    clustervars <- clustervars[-which(clustervars==idname)]
  }

  # we can only handle up to 2-way clustering
  # (in principle could do more, but not high priority now)
  if (length(clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # ---------------------------------------------------------------------------
  # Vectorized multiplier bootstrap (replaces a biters-long pbsapply loop that
  # multiplied a full n x k x nX influence array by fresh Rademacher weights and
  # ran apply(., c(2,3), mean) on every draw -- O(n*k*nX) compute AND an O(n*k*nX)
  # transient array allocation per draw, ~1.2 GB x biters at n in the thousands).
  #
  # The per-draw statistic is, writing inf[i,j,x] for the influence array,
  #   Jb[x,j] = (1/n) sum_i U_b[i] inf[i,j,x]
  #   CvMb_b  = n * mean_x( sum_j Jb[x,j]^2 ) = (n/nX) * sum_{x,j} Jb[x,j]^2.
  # Stacking the weights of all draws into V (n x biters) and reshaping inf to
  # M (n x k*nX), Jb for every draw is crossprod(V, M)/n in one BLAS call, and
  # CvMb is (n/nX) * rowSums(.^2). Drawing all biters*n (or biters*n1) Rademacher
  # values in a single sample() call consumes the RNG stream identically to the
  # per-draw loop (each value is one R_unif_index(2)), so results match the old
  # loop up to floating-point summation order (~1e-14).
  # ---------------------------------------------------------------------------
  dms <- dim(inf.func)
  k <- dms[2]; nX <- dms[3]

  # Build the n x biters multiplier matrix; columns are independent draws.
  if (length(clustervars) > 0) {
    # Rademacher weights are constant within cluster; map each unit to its cluster's
    # weight via match() exactly as the original loop did.
    uc <- unique(dta[, clustervars])
    n1 <- length(uc)
    Vc <- matrix(sample(c(-1, 1), as.numeric(n1) * biters, replace = TRUE), nrow = n1)
    V <- Vc[match(dta[, clustervars], uc), , drop = FALSE]
  } else {
    V <- matrix(sample(c(-1, 1), as.numeric(n) * biters, replace = TRUE), nrow = n)
  }

  # Tile over the X dimension so peak memory is n*k*chunk doubles instead of the
  # full n*k*nX reshape. CvMb is a plain sum over X, so chunking is exact (the only
  # difference is summation order). Chunk targets ~5e6 doubles (~40 MB) per slice.
  chunk <- max(1L, as.integer(5e6 %/% (as.numeric(n) * k)))
  bout <- numeric(biters)
  for (start in seq.int(1L, nX, by = chunk)) {
    xs <- start:min(start + chunk - 1L, nX)
    Mc <- matrix(inf.func[, , xs, drop = FALSE], nrow = n)   # n x (k*length(xs))
    Jf <- crossprod(V, Mc) / n                                # biters x (k*length(xs))
    bout <- bout + rowSums(Jf^2)
  }
  bout <- (n / nX) * bout

  crit.val <- quantile(bout, 1-alp, type=1)

  list(bres=bout, crit.val=crit.val)
}

# =============================================================================
# Conditional pre-test internals: the vectorized indicator() and the vectorized
# multiplier bootstrap test.mboot() must reproduce their explicit (loop) reference
# definitions. test.mboot() is exercised both unclustered and clustered (the
# clustered path was previously untested).
# =============================================================================

test_that("indicator() vectorization matches the row-wise definition exactly", {
  old_indicator <- function(X, u) {
    cond <- t(apply(X, 1, function(x) x <= u))
    1 * apply(cond, 1, all)
  }
  # X always carries an intercept column in the pre-test (model.matrix output), so
  # use >= 2 columns -- exactly the shape indicator() is called with in practice.
  set.seed(3)
  X <- matrix(rnorm(500), 100, 5)
  for (i in 1:20) expect_identical(indicator(X, X[i, ]), old_indicator(X, X[i, ]))
  X2 <- cbind(1, matrix(rnorm(200), 100, 2))   # intercept + 2 covariates
  for (i in 1:10) expect_identical(indicator(X2, X2[i, ]), old_indicator(X2, X2[i, ]))
})

test_that("test.mboot equals the explicit per-draw multiplier bootstrap (unclustered)", {
  set.seed(1)
  n <- 120; k <- 2; nX <- 120
  arr <- array(rnorm(n * k * nX), c(n, k, nX))
  dp <- list(data = data.frame(id = 1:n, period = 1L), idname = "id",
             clustervars = NULL, biters = 200, tname = "period", alp = 0.05, panel = TRUE)
  ref_boot <- function(inf.func, n, biters) {
    sapply(1:biters, function(b) {
      Ub <- sample(c(-1, 1), n, replace = TRUE)
      Jb <- t(apply(Ub * inf.func, c(2, 3), mean))
      n * sum(apply(Jb^2, 2, mean))
    })
  }
  set.seed(42); ref <- ref_boot(arr, n, 200)
  set.seed(42); got <- test.mboot(arr, dp)$bres
  expect_equal(unname(got), ref, tolerance = 1e-9)
})

test_that("test.mboot equals the explicit clustered multiplier bootstrap", {
  set.seed(2)
  n <- 100; k <- 2; nX <- 100; ncl <- 20
  arr <- array(rnorm(n * k * nX), c(n, k, nX))
  cl <- rep(1:ncl, length.out = n)
  dp <- list(data = data.frame(id = 1:n, period = 1L, cl = cl), idname = "id",
             clustervars = "cl", biters = 200, tname = "period", alp = 0.05, panel = TRUE)
  ref_boot_cl <- function(inf.func, cl, n, biters) {
    uc <- unique(cl)
    sapply(1:biters, function(b) {
      Vb <- sample(c(-1, 1), length(uc), replace = TRUE)
      Ub <- Vb[match(cl, uc)]
      Jb <- t(apply(Ub * inf.func, c(2, 3), mean))
      n * sum(apply(Jb^2, 2, mean))
    })
  }
  set.seed(7); ref <- ref_boot_cl(arr, cl, n, 200)
  set.seed(7); got <- test.mboot(arr, dp)$bres
  expect_equal(unname(got), ref, tolerance = 1e-9)
})

test_that("test.mboot multi-chunk tiling accumulates across chunks correctly", {
  # test.mboot tiles over the X dimension with chunk = max(1L, 5e6 %/% (n*k))
  # and accumulates bout across iterations. Every other test in the suite has
  # n*k small enough that the loop runs exactly once; this is the only test
  # where the multi-chunk accumulation (chunk = 1, three iterations) is hit --
  # the regime real large-n pre-test runs land in.
  skip_on_cran()   # ~60 MB transient influence array
  set.seed(123)
  n <- 50; k <- 50010; nX <- 3
  expect_identical(max(1L, as.integer(5e6 %/% (as.numeric(n) * k))), 1L)  # multi-chunk path is exercised
  arr <- array(rnorm(n * k * nX), c(n, k, nX))
  dp <- list(data = data.frame(id = 1:n, period = 1L), idname = "id",
             clustervars = NULL, biters = 50, tname = "period", alp = 0.05, panel = TRUE)
  # Per-draw reference: Jb[x,j] = mean_i(Ub[i] * inf[i,j,x]); CvMb = (n/nX) * sum(Jb^2).
  # colMeans on the n x (k*nX) reshape is the same statistic as the apply-style
  # reference above, just fast enough for k = 50010.
  M <- matrix(arr, nrow = n)
  ref_boot <- function(M, n, nX, biters) {
    sapply(1:biters, function(b) {
      Ub <- sample(c(-1, 1), n, replace = TRUE)
      (n / nX) * sum(colMeans(Ub * M)^2)
    })
  }
  set.seed(42); ref <- ref_boot(M, n, nX, 50)
  set.seed(42); got <- test.mboot(arr, dp)$bres
  expect_equal(unname(got), ref, tolerance = 1e-9)
})

utils::globalVariables(c("year","att","att.se","post","group","x"))

#' @title citation
#'
#' @description print the citation for the relevant paper
#'
#' @keywords internal
citation <- function() {
  cat("Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  \"Difference-in-Differences with Multiple Time Periods.\" Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> \n")
}

#' @title Process `did` Function Arguments
#'
#' @description Function to process arguments passed to the main methods in the
#'  `did` package as well as conducting some tests to make sure
#'  data is in proper format / try to throw helpful error messages.
#'
#' @inheritParams att_gt
#' @param call Function call to att_gt
#'
#' @return a [`DIDparams`] object
#'
#' @export
pre_process_did <- function(yname,
                            tname,
                            idname,
                            gname,
                            xformla = NULL,
                            data,
                            panel = TRUE,
                            allow_unbalanced_panel,
                            control_group = c("nevertreated","notyettreated"),
                            anticipation = 0,
                            weightsname = NULL,
                            fix_weights = NULL,
                            alp = 0.05,
                            bstrap = FALSE,
                            cband = FALSE,
                            biters = 1000,
                            clustervars = NULL,
                            est_method = "dr",
                            base_period = "varying",
                            print_details = TRUE,
                            faster_mode = FALSE,
                            pl = FALSE,
                            cores = 1,
                            call = NULL) {
  #-----------------------------------------------------------------------------
  # Data pre-processing and error checking
  #-----------------------------------------------------------------------------
  # set control group
  control_group <- control_group[1]
  if(!(control_group %in% c("nevertreated","notyettreated"))){
    stop("control_group must be either 'nevertreated' or 'notyettreated'")
  }
  base_period <- base_period[1]
  if (!(base_period %in% c("universal", "varying"))) {
    stop("base_period must be either 'universal' or 'varying'.")
  }
  # Check if anticipation is numeric and non-negative (same contract as the fast path)
  if (!is.numeric(anticipation)) {
    stop("anticipation must be numeric. Please convert it.")
  }
  if (anticipation < 0) {
    stop("anticipation must be non-negative. Please check your arguments.")
  }
  check_reserved_did_names(yname = yname, tname = tname, idname = idname,
                           gname = gname, xformla = xformla,
                           weightsname = weightsname,
                           clustervars = clustervars)
  # make sure dataset is a data.frame
  # this gets around RStudio's default of reading data as tibble
  if (!all( class(data) == "data.frame")) {
    data <- as.data.frame(data)
  }

  # validate that all required column names exist in the data
  required_cols <- c(yname, tname, idname, gname, weightsname, clustervars)
  missing_cols <- setdiff(required_cols, colnames(data))
  if (length(missing_cols) > 0) {
    stop("The following column(s) are not found in the data: ",
         paste(missing_cols, collapse = ", "), ". ",
         "Please check the spelling of yname, tname, idname, gname, weightsname, and clustervars.")
  }

  # At most one cluster variable beyond idname is supported (clustering at the
  # unit level via idname is implicit). Enforced here -- mirroring the fast path
  # (pre_process_did2) and mboot(), with identical wording -- so every
  # faster_mode x bstrap combination rejects the input up front; the analytical
  # (bstrap = FALSE) path used to silently cluster on the first extra variable only.
  cv_check <- clustervars
  if (!is.null(cv_check) && !is.null(idname) && (idname %in% cv_check)) {
    cv_check <- setdiff(cv_check, idname)
  }
  if (length(cv_check) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # make sure time periods are numeric
  if (! (is.numeric(data[, tname])) ) stop("The time variable '", tname, "' must be numeric. Please convert it.")

  #  make sure gname is numeric
  if (! (is.numeric(data[, gname])) ) stop("The group variable '", gname, "' must be numeric. Please convert it.")

  #  make sure the outcome is numeric (logical 0/1 outcomes are also allowed)
  if (! (is.numeric(data[, yname]) || is.logical(data[, yname])) ) stop("The outcome variable '", yname, "' must be numeric. Please convert it.")

  # put in blank xformla if no covariates or check whether all variables are in data
  if (is.null(xformla)) {
    xformla <- ~1
  } else {
    # extract variable names from the formula
    formula_vars <- all.vars(xformla)

    # identify variables in xformla not in data
    missing_vars <- setdiff(formula_vars, names(data))

    # error checking for missing variables in data
    if (length(missing_vars) > 0) {
      stop(paste("The following variables are not in data:", paste(missing_vars, collapse = ", ")), call. = FALSE)
    }
  }

  # drop irrelevant columns from data. Keep the RAW covariate variables (all.vars of
  # xformla), NOT the evaluated model.frame, so model.matrix(xformla, .) can be
  # rebuilt downstream. This is what makes transform formulae work (e.g. ~I(X^2),
  # ~poly(X, 2), ~log(X)): the evaluated model.frame would store columns named by the
  # transformed expression -- losing the raw variable that model.matrix needs -- and,
  # for matrix-valued terms like poly(), a matrix column that later breaks the
  # data.table coercion in compute.att_gt(). For bare-variable formulae (e.g. ~X,
  # ~X1+X2) and factor covariates the kept columns are identical to before.
  xvars <- all.vars(xformla)
  keep_cols <- unique(c(idname, tname, yname, gname, weightsname, clustervars, xvars))
  data <- data[, keep_cols, drop = FALSE]

  # check if any covariates were missing
  n_orig <- nrow(data)
  # drop rows with any missing id / time / outcome / group / weight / cluster or any
  # missing RAW covariate value
  data <- data[complete.cases(data), ]
  # also drop rows whose EVALUATED design is non-finite (e.g. log of a non-positive
  # covariate), preserving the previous model.frame-based row dropping. We use
  # model.frame (NOT model.matrix) with na.action = na.pass: model.frame keeps EVERY
  # row -- including those where a term evaluates to NA/NaN -- so complete.cases()
  # flags them and the indicator stays aligned with `data`. (model.matrix would
  # instead silently drop the NaN rows, making the mask shorter than `data` and the
  # offending rows survive.) Inf-valued terms are kept, matching the prior behavior.
  # Safe to evaluate now that raw-covariate NAs have been removed (so poly()/ns()/...
  # will not error on NA input).
  if (length(xvars) > 0L && nrow(data) > 0L) {
    mf_check <- suppressWarnings(model.frame(xformla, data = data, na.action = na.pass))
    finite_rows <- complete.cases(mf_check)
    if (!all(finite_rows)) data <- data[finite_rows, ]
  }
  n_diff <- n_orig - nrow(data)
  if (n_diff != 0) {
    warning(paste0("dropped ", n_diff, " rows from original data due to missing data"))
  }

  # weights if null
  if (is.null(weightsname)) w <- rep(1, nrow(data)) else w <- data[,weightsname]

  # Validate user-supplied weights: negative weights flip signs and a non-positive
  # mean divides by ~0 during normalization, silently producing NA/NaN ATTs.
  if (!is.null(weightsname) && (any(w < 0, na.rm = TRUE) || isTRUE(mean(w, na.rm = TRUE) <= 0)))
    stop("The weights variable '", weightsname, "' must be non-negative with a positive mean.")

  if (".w" %in% colnames(data)) stop("Your data already contains a column named '.w', which is reserved for internal use by `did`. Please rename this column before calling att_gt().")
  data$.w <- w

  # Check for time-varying weights in panel data. Grouped max/min via data.table
  # (GForce-optimized) is much faster than tapply() with a per-unit closure and
  # produces the same per-unit ranges as diff(range(x)) on the raw weights.
  if (!is.null(weightsname) && panel) {
    dtw <- data.table(id = data[[idname]], .w = data[[weightsname]])
    w_rng <- dtw[, .(mx = max(.w), mn = min(.w)), by = "id"]
    if (any((w_rng$mx - w_rng$mn) > .Machine$double.eps^0.5, na.rm = TRUE)) {
      message(
        "Time-varying weights detected. For balanced panel data, the default ",
        "behavior uses the weight from the earlier of the two time periods in ",
        "each 2x2 comparison (the base period for post-treatment cells). ",
        "Use the 'fix_weights' argument to control this behavior. ",
        "See ?att_gt for details."
      )
    }
  }

  # Outcome variable will be denoted by y
  # data$.y <- data[, yname]

  # figure out the dates
  # list of dates from smallest to largest
  tlist <- sort(unique(data[,tname]))

  # Groups with treatment time bigger than max time period + anticipation are considered to be never treated.
  # We account for anticipation because units treated shortly after the last observed period
  # may already exhibit anticipatory effects during the observed time window, and should
  # not be used as controls.
  asif_never_treated <- (data[,gname] > max(tlist, na.rm = TRUE) + anticipation)
  asif_never_treated[is.na(asif_never_treated)] <- FALSE
  data[asif_never_treated, gname] <- 0

  # list of treated groups (by time) from smallest to largest
  glist <- sort(unique(data[,gname]))


  # Check if there is a never treated group

  # if ( length(glist[glist==0]) == 0) {
  #   if(control_group=="nevertreated"){
  #     stop("There is no available never-treated group")
  #   } else {
  #     # Drop all time periods with time periods >= latest treated
  #     data <- subset(data,(data[,tname] < (max(glist)-anticipation)))
  #     # Replace last treated time with zero
  #     # lines.gmax <- data[,gname]==max(glist, na.rm = TRUE)
  #     # data[lines.gmax,gname] <- 0
  #
  #     tlist <- sort(unique(data[,tname]))
  #     glist <- sort(unique(data[,gname]))
  #
  #     # don't comput ATT(g,t) for groups that are only treated at end
  #     # and only play a role as a comparison group
  #     glist <- glist[ glist < max(glist)]
  #   }
  # }

  if (!any(glist == 0)) {
    # Compute latest treated cohort once, and the cutoff time
    latest_g <- max(glist, na.rm = TRUE)
    cutoff_t  <- latest_g - anticipation

    if (control_group == "nevertreated") {
      # Warn the user
      warning(
        "No never-treated group is available. ",
        "The last treated cohort is being coerced as 'never-treated' units, and data from periods after that is being filtered out (no available comparison groups)."
      )

      # Drop all periods ≥ (latest_g - anticipation)
      data <- data[ data[[ tname ]] < cutoff_t, , drop = FALSE ]

      # For any row where gname == latest_g, set gname := 0
      lines.gmax <- data[, gname]==latest_g
      data[lines.gmax, gname] <- 0
    } else {
      # If "notyettreated", we simply drop those periods and leave gnames alone
      data <- data[ data[[ tname ]] < cutoff_t, , drop = FALSE ]
    }

    # Recompute tlist and glist from the filtered/modified data
    tlist <- sort(unique(data[,tname]))
    glist <- sort(unique(data[,gname]))


    # If control_group != "nevertreated", drop the max cohort from glist
    if (control_group != "nevertreated") {
      glist <- glist[glist < latest_g]
    }
  }

  # Only the treated groups
  glist <- glist[glist>0]

  # drop groups treated in the first period or before
  first.period <- tlist[1]
  glist <- glist[glist > first.period + anticipation]

  # check for groups treated in the first period (accounting for anticipation) and drop these
  # nfirstperiod <- length(unique(data[ !((data[,gname] > first.period) | (data[,gname]==0)), ] )[,idname])
  treated_first_period <- ( data[,gname] <= first.period + anticipation ) & ( !(data[,gname]==0) )
  treated_first_period[is.na(treated_first_period)] <- FALSE
  # if/else (not ifelse) so the data subset is built only for the relevant branch;
  # the result is identical to the previous ifelse(panel, ...) expression.
  nfirstperiod <- if (panel) length(unique(data[treated_first_period, idname])) else sum(treated_first_period)
  if ( nfirstperiod > 0 ) {
    warning(paste0("Dropped ", nfirstperiod, " units that were already treated in the first period",
                    if (anticipation > 0) paste0(" (accounting for anticipation = ", anticipation, ")") else "",
                    "."))
    data <- data[ data[,gname] %in% c(0,glist), ]
    # update tlist and glist
    tlist <- sort(unique(data[,tname]))
    glist <- sort(unique(data[,gname]))
    glist <- glist[glist>0]

    # drop groups treated in the first period or before
    first.period <- tlist[1]
    glist <- glist[glist > first.period + anticipation]

  }

  #  make sure id is numeric
  if (! is.null(idname)){
    #  make sure id is numeric
    if (! (is.numeric(data[, idname])) ) stop("The id variable '", idname, "' must be numeric. Please convert it.")

    # Validate treatment irreversibility and (id, period) uniqueness with a single
    # radix order over (id, t) plus O(n) adjacent-element scans, instead of the much
    # slower unique.data.frame()/anyDuplicated.data.frame() row-key machinery. After
    # ordering, each unit's rows are contiguous, so g varying within a unit shows up
    # as adjacent rows with equal id but differing g, and a duplicated (id, period)
    # pair as adjacent rows with equal id and equal t. All three columns are numeric
    # and NA-free at this point, so the comparisons are exact.
    idv <- data[, idname]
    nn <- length(idv)
    if (nn >= 2L) {
      tv <- data[, tname]
      o <- order(idv, tv, method = "radix")
      io <- idv[o]
      same_id <- io[-1L] == io[-nn]

      # Check that gname is time-invariant within each unit (treatment irreversibility).
      go <- data[, gname][o]
      if (any(same_id & (go[-1L] != go[-nn]))) {
        stop("The value of gname (treatment variable) must be the same across all periods for each particular unit. The treatment must be irreversible.")
      }

      # Check that (idname, tname) is unique: each unit observed at most once per
      # period. Mirrors the fast path (pre_process_did2.R) so both code paths reject
      # duplicated (id, period) rows identically -- without this guard the slow path
      # silently produced incorrect estimates on long-format data with duplicates.
      to <- tv[o]
      if (any(same_id & (to[-1L] == to[-nn]))) {
        stop("The value of idname must be unique (by tname). Some units are observed more than once in a period.")
      }

      # Check that cluster variables are time-invariant within each unit, mirroring
      # the fast path (validate_args) and mboot() -- with identical wording -- so
      # invalid clustering inputs are rejected up front regardless of bstrap.
      # Without this, the analytical (bstrap = FALSE) path fell back to i.i.d.
      # standard errors with a warning advising bstrap = TRUE, advice that then
      # fails in mboot() for the very same input. Reuses the (id, t) radix order
      # and adjacency scan above.
      for (cvar in setdiff(clustervars, idname)) {
        cvo <- data[o, cvar]
        if (any(same_id & (cvo[-1L] != cvo[-nn]))) {
          stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
        }
      }
    }
  }



  # if user specifies repeated cross sections,
  # set that it really is repeated cross sections
  true_repeated_cross_sections <- FALSE
  if (!panel) {
    true_repeated_cross_sections <- TRUE
  }

  #-----------------------------------------------------------------------------
  # setup data in panel case
  #-----------------------------------------------------------------------------
  # Check if data is a balanced panel if panel = TRUE and allow_unbalanced_panel = TRUE
  bal_panel_test <- panel*allow_unbalanced_panel
  if (bal_panel_test) {
    # data is already complete-case filtered above and (id, period) uniqueness has
    # been validated, so the panel is balanced iff every unit appears in every
    # period, i.e. nrow(data) equals (number of units) x (number of periods). This
    # replaces a BMisc::makeBalancedPanel() round-trip that built a balanced copy
    # of the data just to compare unit counts and then threw it away.
    allow_unbalanced_panel <-
      nrow(data) != length(unique(data[[idname]])) * as.numeric(length(unique(data[[tname]])))
  }



  if (panel) {

    # check for unbalanced panel
    if (allow_unbalanced_panel) {

      # Flag for true repeated cross sections
      panel <- FALSE
      true_repeated_cross_sections <- FALSE

    } else {

      # this is the case where we coerce balanced panel

      # check for complete cases (rows with missing data were already dropped
      # above, so anyNA() short-circuits the redundant complete.cases() pass and
      # full-table subset copies in the common case)
      if (anyNA(data)) {
        keepers <- complete.cases(data)
        n <- length(unique(data[,idname]))
        n.keep <- length(unique(data[keepers,idname]))
        if (!all(keepers)) {
          warning(paste0("Dropped ", (n-n.keep), " observations that had missing data."))
          data <- data[keepers,]
        }
      }

      # make it a balanced data set: keep only units observed in every period.
      # A unit is fully observed iff its row count equals the number of distinct
      # periods ((id, period) uniqueness was validated above), so this
      # count-and-filter keeps exactly the same rows as BMisc::makeBalancedPanel()
      # without the per-group .SD materialization; any row-order difference is
      # normalized by the (id, period) sort below.
      uid <- unique(data[[idname]])
      n.old <- length(uid)
      cnt <- tabulate(match(data[[idname]], uid))
      keep_bal <- data[[idname]] %in% uid[cnt == length(unique(data[[tname]]))]
      if (!all(keep_bal)) data <- data[keep_bal, , drop = FALSE]
      n <- length(unique(data[,idname]))
      if (n < n.old) {
        # n.old - n counts dropped UNITS (unique ids), not unit-time rows; word the
        # warning accordingly, with the same text as the fast path (pre_process_did2)
        warning(n.old - n, " units are missing in some periods. Converting to balanced panel by dropping them.")
      }

      # If drop all data, you do not have a panel.
      if (nrow(data)==0) {
        stop("All observations dropped while converting data to balanced panel. Consider setting `panel = FALSE` and/or revisiting 'idname'.")
      }

      n <- nrow(data[ data[,tname]==tlist[1], ])

      # slow, repeated check here...
      ## # check that first.treat doesn't change across periods for particular individuals
      ## if (!all(sapply( split(data, data[,idname]), function(df) {
      ##   length(unique(df[,gname]))==1
      ## }))) {
      ##   stop("The value of gname must be the same across all periods for each particular individual.")
      ## }

    }
  }

  #-----------------------------------------------------------------------------
  # code for setting up repeated cross sections (and unbalanced panel)
  #-----------------------------------------------------------------------------
  if (!panel) {

    # check for complete cases (rows with missing data were already dropped
    # above, so anyNA() short-circuits the redundant complete.cases() pass and
    # full-table subset copies in the common case)
    if (anyNA(data)) {
      keepers <- complete.cases(data)
      if (!all(keepers)) {
        warning(paste0("Dropped ", sum(!keepers), " observations that had missing data."))
        data <- data[keepers,]
      }
    }

    # If drop all data, you do not have a panel.
    if (nrow(data)==0) {
      stop("All observations were dropped due to missing data. Check your outcome, group, time, and covariate variables for missing values.")
    }

    # n-row data.frame to hold the influence function
    if (true_repeated_cross_sections) {
      data$.rowid <- seq(1:nrow(data))
      idname <- ".rowid"
    } else {
      # set rowid to idname for repeated cross section/unbalanced
      data$.rowid <- data[, idname]
    }

    # n is unique number of cross section observations
    # this is different for repeated cross sections and unbalanced panel
    n <- length(unique(data[,idname]))
  }

  ## # Update tlist and glist because of data handling
  ## # figure out the dates
  ## # list of dates from smallest to largest
  ## tlist <- unique(data[,tname])[order(unique(data[,tname]))]
  ## # list of treated groups (by time) from smallest to largest
  ## glist <- unique(data[,gname])[order(unique(data[,gname]))]

  ## # Only the treated groups
  ## glist <- glist[glist>0]

  ## # drop groups treated in the first period or before
  ## first.period <- tlist[1]
  ## glist <- glist[glist > first.period + anticipation]

  # Check if groups is empty (usually a problem with the way people defined groups)
  if(length(glist)==0){
    stop("No valid groups. The variable in 'gname' should be expressed as the time a unit is first treated (0 if never-treated).")
  }

  # if there are only two time periods, then uniform confidence
  # bands are the same as pointwise confidence intervals
  if (length(tlist)==2) {
    cband <- FALSE
  }

  #-----------------------------------------------------------------------------
  # more error handling after we have balanced the panel

  # check against very small groups. tabulate(match(.)) yields the same per-group
  # row counts as the previous aggregate() call without invoking an R closure per
  # group; sorted gvals reproduces aggregate()'s ascending group order, and the
  # Group.1/x column names are kept for the subset()/paste logic below.
  gvals <- sort(unique(data[[gname]]))
  gcnt <- tabulate(match(data[[gname]], gvals), nbins = length(gvals))
  gsize <- data.frame(Group.1 = gvals, x = gcnt / length(tlist))

  # how many in each group before give warning
  # 5 is just a buffer, could pick something else, but seems to work fine
  reqsize <- length(BMisc::rhs_vars(xformla)) + 5

  # which groups to warn about
  gsize <- subset(gsize, x < reqsize) # x is name of column from aggregate

  # warn if some groups are small
  if (nrow(gsize) > 0) {
    gpaste <-  paste(gsize[,1], collapse=",")
    warning(paste0("Some groups in your dataset have very few observations, which may cause estimation problems.\n  Check groups: ", gpaste, "."))

    if ( (0 %in% gsize[,1]) & (control_group == "nevertreated") ) {
      stop("The never-treated group is too small to serve as a reliable control. Try setting `control_group = 'notyettreated'` to include not-yet-treated units as controls.")
    }
  }
  #----------------------------------------------------------------------------

  # How many time periods
  nT <- length(tlist)
  # How many treated groups
  nG <- length(glist)

  # order dataset wrt idname and tname, in place. The sort keys are tie-free
  # ((id, period) uniqueness was validated above; .rowid is 1:n for repeated cross
  # sections), so the permutation is unique and the row order matches the previous
  # order()-based subset without a full-table copy. `data` is function-local (it
  # was subset/copied above), so the by-reference conversion cannot touch the
  # user's input; setDF() returns a plain data.frame for downstream consumers
  # (DIDparams, compute.att_gt).
  setDT(data)
  setorderv(data, c(idname, tname))
  setDF(data)

  # store parameters for passing around later
  dp <- DIDparams(yname=yname,
                  tname=tname,
                  idname=idname,
                  gname=gname,
                  xformla=xformla,
                  data=data,
                  control_group=control_group,
                  anticipation=anticipation,
                  weightsname=weightsname,
                  fix_weights=fix_weights,
                  alp=alp,
                  bstrap=bstrap,
                  biters=biters,
                  clustervars=clustervars,
                  cband=cband,
                  print_details=print_details,
                  faster_mode=faster_mode,
                  pl=pl,
                  cores=cores,
                  est_method=est_method,
                  base_period=base_period,
                  panel=panel,
                  true_repeated_cross_sections=true_repeated_cross_sections,
                  n=n,
                  nG=nG,
                  nT=nT,
                  tlist=tlist,
                  glist=glist,
                  call=call)
}

# Analytical (no-bootstrap) cluster-robust standard errors. With a cluster variable and bstrap = FALSE,
# att_gt() and aggte() report the cluster-robust variance (cluster sums of the influence function), the
# same cluster-sum aggregation as the multiplier bootstrap (Callaway & Sant'Anna 2021, Remark 10).
# panel in which units within a cluster are dependent: they share a common shock each period (e.g. a
# state-by-year shock), so the cluster -- not the individual unit -- is the independent sampling unit
.make_clustered_shocks <- function(seed, G = 50L) {
  set.seed(seed)
  sz <- rep(c(1L, 2L, 4L, 10L), length.out = G); N <- sum(sz); cl <- rep(seq_len(G), times = sz)
  alpha <- rnorm(G, 0, 1)[cl]; nu <- rnorm(N, 0, 1); per <- 1:4
  eta <- matrix(rnorm(G * 4L, 0, 1.5), G, 4L)                     # common shock per cluster, per period
  g <- sample(c(2L, 3L, 0L), G, replace = TRUE, prob = c(.34, .33, .33))[cl]  # cluster-level treatment
  d <- data.frame(id = rep(seq_len(N), each = 4L), t = rep(per, N), cl = rep(cl, each = 4L),
                  g = rep(g, each = 4L), a = rep(alpha, each = 4L), nu = rep(nu, each = 4L))
  d$y <- d$a + d$nu + 0.3 * d$t + eta[cbind(d$cl, d$t)] + (d$g != 0L & d$t >= d$g) + rnorm(nrow(d))
  d[order(d$id, d$t), ]
}

test_that("analytical (no-bootstrap) clustered SE for att_gt equals the cluster-sum CRVE (faster_mode TRUE and FALSE)", {
  d <- .make_clustered_shocks(404L)
  for (fm in c(TRUE, FALSE)) {
    res <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                  control_group = "nevertreated", bstrap = FALSE, clustervars = "cl",
                  base_period = "varying", faster_mode = fm)
    cv <- res$DIDparams$cluster_vector
    # the per-unit cluster vector must be available in BOTH modes (derived from data in slower mode)
    expect_true(!is.null(cv) && length(cv) == nrow(res$inffunc))
    n <- nrow(res$inffunc)
    S <- rowsum(as.matrix(res$inffunc), cv)            # n_clusters x k cluster sums
    se_target <- sqrt(colSums(S^2)) / n                 # cluster-sum CRVE per ATT(g,t)
    ok <- is.finite(res$se) & is.finite(se_target)
    # analytical => deterministic => exact match (no Monte Carlo tolerance needed)
    expect_equal(unname(res$se[ok]), unname(se_target[ok]), tolerance = 1e-8)

    # and it differs from the i.i.d. SE when the units within a cluster are dependent
    res_iid <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                      control_group = "nevertreated", bstrap = FALSE, base_period = "varying", faster_mode = fm)
    k <- which(res$group == 2L & res$t == 2L)
    expect_gt(abs(res$se[k] - res_iid$se[k]) / res_iid$se[k], 0.05)
  }
})

test_that("analytical cluster SE agrees with the bootstrap cluster SE (multiple DGPs)", {
  skip_on_cran()  # bootstrap-heavy
  # units within a cluster share a common shock; treatment at the cluster level (within = FALSE) or within clusters (TRUE).
  mk <- function(seed, within, G = 60L) {
    set.seed(seed); sz <- rep(c(1L, 2L, 4L, 10L), length.out = G); N <- sum(sz); cl <- rep(seq_len(G), times = sz)
    alpha <- rnorm(G, 0, 1)[cl]; nu <- rnorm(N, 0, 1); eta <- matrix(rnorm(G * 4L, 0, 1.5), G, 4L)
    g <- if (within) sample(c(2L, 3L, 0L), N, replace = TRUE, prob = c(.34, .33, .33))
         else sample(c(2L, 3L, 0L), G, replace = TRUE, prob = c(.34, .33, .33))[cl]
    d <- data.frame(id = rep(seq_len(N), each = 4L), t = rep(1:4, N), cl = rep(cl, each = 4L),
                    g = rep(g, each = 4L), a = rep(alpha, each = 4L), nu = rep(nu, each = 4L))
    d$y <- d$a + d$nu + 0.3 * d$t + eta[cbind(d$cl, d$t)] + (d$g != 0L & d$t >= d$g) + rnorm(nrow(d))
    d[order(d$id, d$t), ]
  }
  for (within in c(FALSE, TRUE)) {
    d <- mk(11L, within)
    ra <- att_gt(yname="y", tname="t", idname="id", gname="g", data=d, control_group="nevertreated",
                 bstrap=FALSE, clustervars="cl", base_period="varying")
    set.seed(11L)
    rb <- att_gt(yname="y", tname="t", idname="id", gname="g", data=d, control_group="nevertreated",
                 bstrap=TRUE, biters=3000L, clustervars="cl", pl=FALSE, cband=FALSE, base_period="varying")
    k <- which(ra$group == 2L & ra$t == 2L)
    # the reported bootstrap SE (IQR scale) is within a few percent of the analytical at this G
    expect_lt(abs(rb$se[k] - ra$se[k]) / ra$se[k], 0.15)
    # the bootstrap's own SD scale matches the analytical closely (same cluster variance)
    dp <- ra$DIDparams; dp$biters <- 3000L; dp$pl <- FALSE
    set.seed(11L); bo <- mboot(as.matrix(ra$inffunc)[, k, drop = FALSE], dp)
    sd_se <- sd(as.numeric(bo$bres)) * sqrt(length(unique(dp$cluster_vector))) / nrow(ra$inffunc)
    expect_lt(abs(sd_se - ra$se[k]) / ra$se[k], 0.06)
  }
})

# true repeated cross-sections: a fresh cross-section is drawn each period within each cluster, the cluster
# shares a common per-period shock, and treatment is assigned at the cluster level. The cluster -- not the
# (non-existent) panel unit -- is the independent sampling unit.
.make_rcs <- function(seed, G = 60L, per_cell = 6L) {
  set.seed(seed); per <- 1:4
  gC <- sample(c(2L, 3L, 0L), G, replace = TRUE, prob = c(.34, .33, .33)); aC <- rnorm(G, 0, 1)
  eta <- matrix(rnorm(G * 4L, 0, 1.5), G, 4L); rows <- list(); k <- 0L
  for (cl in seq_len(G)) for (tt in per) {
    y <- aC[cl] + 0.3 * tt + eta[cl, tt] + (gC[cl] != 0L & tt >= gC[cl]) + rnorm(per_cell)
    k <- k + 1L; rows[[k]] <- data.frame(t = tt, cl = cl, g = gC[cl], y = y)
  }
  do.call(rbind, rows)[sample(G * 4L * per_cell), ]   # shuffle so row order is non-trivial
}

test_that("analytical clustered SE for repeated cross-sections (idname omitted/provided, faster_mode TRUE/FALSE)", {
  d    <- .make_rcs(909L)
  d_id <- d; d_id$uid <- seq_len(nrow(d_id))
  se_om <- list()
  for (fm in c(TRUE, FALSE)) {
    # idname OMITTED: pre_process_did() creates an internal .rowid; clustering must still aggregate to
    # cluster sums via that internal id (the function argument idname is NULL here).
    r_om <- att_gt(yname = "y", tname = "t", gname = "g", data = d, control_group = "nevertreated",
                   bstrap = FALSE, clustervars = "cl", panel = FALSE, base_period = "varying", faster_mode = fm)
    # idname PROVIDED (a per-row observation id)
    r_id <- att_gt(yname = "y", tname = "t", idname = "uid", gname = "g", data = d_id,
                   control_group = "nevertreated", bstrap = FALSE, clustervars = "cl",
                   panel = FALSE, base_period = "varying", faster_mode = fm)
    for (r in list(r_om, r_id)) {
      cv <- r$DIDparams$cluster_vector
      expect_true(!is.null(cv) && length(cv) == nrow(r$inffunc))   # must be present and aligned
      n <- nrow(r$inffunc); S <- rowsum(as.matrix(r$inffunc), cv)
      se_target <- sqrt(colSums(S^2)) / n                          # cluster-sum CRVE per ATT(g,t)
      ok <- is.finite(r$se) & is.finite(se_target)
      expect_equal(unname(r$se[ok]), unname(se_target[ok]), tolerance = 1e-8)
    }
    se_om[[as.character(fm)]] <- r_om$se[which(r_om$group == 2L & r_om$t == 2L)]
  }
  # the two code paths derive the cluster vector independently, yet must give the same analytical SE
  expect_equal(unname(se_om[["TRUE"]]), unname(se_om[["FALSE"]]), tolerance = 1e-6)
})

test_that("clustered bootstrap and analytical SE agree for repeated cross-sections (idname omitted)", {
  skip_on_cran()  # bootstrap-heavy
  d <- .make_rcs(910L)
  for (fm in c(TRUE, FALSE)) {
    ra <- att_gt(yname = "y", tname = "t", gname = "g", data = d, control_group = "nevertreated",
                 bstrap = FALSE, clustervars = "cl", panel = FALSE, base_period = "varying", faster_mode = fm)
    rb <- att_gt(yname = "y", tname = "t", gname = "g", data = d, control_group = "nevertreated",
                 bstrap = TRUE, biters = 5000L, clustervars = "cl", pl = FALSE, cband = FALSE,
                 panel = FALSE, base_period = "varying", faster_mode = fm)
    k <- which(ra$group == 2L & ra$t == 2L)
    expect_true(is.finite(rb$se[k]))
    expect_lt(abs(rb$se[k] - ra$se[k]) / ra$se[k], 0.12)
  }
})

test_that("analytical clustered SE flows through aggte at all levels (simple/group/dynamic)", {
  d <- .make_clustered_shocks(505L)
  res_cl  <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                    control_group = "nevertreated", bstrap = FALSE, clustervars = "cl", base_period = "varying")
  res_iid <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                    control_group = "nevertreated", bstrap = FALSE, base_period = "varying")
  expect_true(!is.null(res_cl$DIDparams$cluster_vector))  # required for clustered SEs to propagate through aggte()
  for (ty in c("simple", "group", "dynamic")) {
    a_cl  <- suppressWarnings(aggte(res_cl,  type = ty, bstrap = FALSE))
    a_iid <- suppressWarnings(aggte(res_iid, type = ty, bstrap = FALSE))
    expect_true(is.finite(a_cl$overall.se) && a_cl$overall.se > 0)
    # clustered overall SE differs from the i.i.d. one when units within a cluster are dependent
    expect_gt(abs(a_cl$overall.se - a_iid$overall.se) / a_iid$overall.se, 0.02)
  }
})

# Regression tests for the latest-control-cohort deletion bug.
#
# Bug (did <= 2.5.0): with control_group = "notyettreated" and NO never-treated
# group, the latest cohort is correctly removed from glist so it can serve as a
# not-yet-treated control. But when some unit is treated in the first period
# (accounting for anticipation), the "drop first-period-treated units" step
# filtered the data by cohort membership in glist (plus the never-treated
# sentinel -- 0 in the slow path, Inf in the fast path) -- and because the latest
# cohort is no longer in glist, that filter ALSO deleted the entire latest
# control cohort, silently corrupting ATT(g,t) for every other group.
#
# Invariant A (MUST): att_gt output must be INVARIANT to the presence of
# effectively-always-treated units (treated <= first period, accounting for
# anticipation). Three trigger pathways are exercised:
#   P1  a literal always-treated cohort (g <= first period), anticipation = 0
#   P2  anticipation >= 1 promotes the EARLIEST cohort to first-period-treated
#       (fires with NO nominal always-treated unit)
#   P3  anticipation un-coerces a late (g = T + 1) cohort from never-treated,
#       removing the sole never-treated group

# --- helpers ----------------------------------------------------------------

# build a balanced panel from a vector of cohorts; `het` controls level spread
.mk_design <- function(cohorts, n_periods, n = 25, seed = 1, het = 1.2) {
  set.seed(seed)
  rows <- data.table::rbindlist(lapply(cohorts, function(g)
    data.table::data.table(g = g,
                           b = exp(stats::rnorm(n, log(1e5), het)),
                           uid = NA_integer_)))
  rows[, uid := .I]
  d <- merge(data.table::CJ(uid = rows$uid, t = 1:n_periods),
             rows[, .(uid, g, b)], by = "uid")
  d[, y := b + 50 * (t - 1) + 200 * (t >= g & g > 0) + stats::rnorm(.N, 0, 5)]
  as.data.frame(d[])
}

.att_keyed <- function(data, drop_uid = integer(0), ...) {
  if (length(drop_uid)) data <- data[!(data$uid %in% drop_uid), ]
  cs <- suppressWarnings(suppressMessages(
    att_gt(yname = "y", tname = "t", idname = "uid", gname = "g",
           data = data, control_group = "notyettreated",
           bstrap = FALSE, ...)))
  stats::setNames(cs$att, paste(cs$group, cs$t, sep = "_"))
}

# assert: ATT(g,t) on cells common to full vs (full minus EAT units) is identical,
# including the NA pattern
.expect_invariant <- function(data, eat_uid, ..., tol = 1e-8) {
  full <- .att_keyed(data, ...)
  drop <- .att_keyed(data, drop_uid = eat_uid, ...)
  common <- intersect(names(full), names(drop))
  expect_true(length(common) > 0)
  expect_equal(is.na(full[common]), is.na(drop[common]))   # NA pattern matches
  ok <- !is.na(full[common])
  expect_equal(unname(full[common][ok]), unname(drop[common][ok]), tolerance = tol)
}

# --- P1: literal always-treated cohort, anticipation = 0 --------------------

test_that("Invariant A holds under P1 (always-treated cohort, no never-treated)", {
  d <- .mk_design(c(1, 2, 3, 4), n_periods = 4, seed = 2024)
  eat <- unique(d$uid[d$g == 1])
  for (em in c("dr", "reg", "ipw")) {
    .expect_invariant(d, eat, est_method = em, faster_mode = TRUE)
    .expect_invariant(d, eat, est_method = em, faster_mode = FALSE)
  }
})

# --- P2: anticipation promotes earliest cohort (no always-treated) ----------

test_that("Invariant A holds under P2 (anticipation>=1 promotes earliest cohort)", {
  d <- .mk_design(c(2, 3, 5), n_periods = 5, seed = 11)
  eat <- unique(d$uid[d$g == 2])          # earliest cohort, promoted at anticipation = 1
  .expect_invariant(d, eat, anticipation = 1, faster_mode = TRUE)
  .expect_invariant(d, eat, anticipation = 1, faster_mode = FALSE)
})

# --- P3: anticipation removes the sole (coerced) never-treated group ---------

test_that("Invariant A holds under P3 (anticipation un-coerces a late cohort)", {
  d <- .mk_design(c(2, 3, 6), n_periods = 5, seed = 7)   # cohort 6 = T+1, coerced-to-never at anti=0
  eat <- unique(d$uid[d$g == 2])
  .expect_invariant(d, eat, anticipation = 1, faster_mode = TRUE)
  .expect_invariant(d, eat, anticipation = 1, faster_mode = FALSE)
})

# --- scaling oracle: effectively-always-treated outcomes are never read ------

test_that("always-treated outcomes are never read (scaling oracle)", {
  d <- .mk_design(c(1, 2, 3, 4), n_periods = 4, seed = 2024)
  d_scaled <- d
  d_scaled$y[d_scaled$g == 1] <- d_scaled$y[d_scaled$g == 1] * 1e6
  a1 <- .att_keyed(d)
  a2 <- .att_keyed(d_scaled)
  expect_equal(is.na(a1), is.na(a2))
  ok <- !is.na(a1)
  expect_equal(unname(a1[ok]), unname(a2[ok]), tolerance = 1e-8)
})

# --- fast/slow parity in the vulnerable regime ------------------------------

test_that("fast and slow paths agree in the no-never + always-treated regime", {
  d <- .mk_design(c(1, 2, 3, 4, 5), n_periods = 5, seed = 99)
  fast <- .att_keyed(d, faster_mode = TRUE)
  slow <- .att_keyed(d, faster_mode = FALSE)
  expect_equal(is.na(fast), is.na(slow))
  ok <- !is.na(fast)
  expect_equal(unname(fast[ok]), unname(slow[ok]), tolerance = 1e-10)
})

# --- structural: latest cohort kept as control, not an estimated group ------

test_that("latest cohort is retained as a control, not deleted, when always-treated present", {
  d <- .mk_design(c(1, 2, 3, 4, 5), n_periods = 5, seed = 5)
  dp <- suppressWarnings(suppressMessages(
    pre_process_did(yname = "y", tname = "t", idname = "uid", gname = "g",
                    data = d, panel = TRUE, allow_unbalanced_panel = FALSE,
                    control_group = "notyettreated", print_details = FALSE)))
  kept <- sort(unique(dp$data$g))
  cs <- suppressWarnings(suppressMessages(
    att_gt(yname = "y", tname = "t", idname = "uid", gname = "g",
           data = d, control_group = "notyettreated", bstrap = FALSE)))
  expect_true(5 %in% kept)                 # latest cohort kept in the data (as control)
  expect_false(1 %in% kept)                # always-treated cohort dropped
  expect_false(5 %in% unique(cs$group))    # latest cohort gets no ATT of its own
})

# --- nevertreated branch was always immune: confirm it stays invariant -------

test_that("control_group='nevertreated' is invariant to always-treated presence", {
  d <- .mk_design(c(1, 2, 3, 4), n_periods = 4, seed = 2024)
  eat <- unique(d$uid[d$g == 1])
  full <- suppressWarnings(suppressMessages(att_gt("y", "t", "uid", "g", data = d,
            control_group = "nevertreated", bstrap = FALSE)))
  drop <- suppressWarnings(suppressMessages(att_gt("y", "t", "uid", "g",
            data = d[!(d$uid %in% eat), ], control_group = "nevertreated", bstrap = FALSE)))
  fk <- stats::setNames(full$att, paste(full$group, full$t, sep = "_"))
  dk <- stats::setNames(drop$att, paste(drop$group, drop$t, sep = "_"))
  common <- intersect(names(fk), names(dk))
  expect_true(length(common) > 0)
  expect_equal(is.na(fk[common]), is.na(dk[common]))
  ok <- !is.na(fk[common])
  expect_equal(unname(fk[common][ok]), unname(dk[common][ok]), tolerance = 1e-8)
})

#' @title Pre-Test of Conditional Parallel Trends Assumption
#'
#' @description An integrated moments test for the conditional parallel trends
#'  assumption holding in all pre-treatment time periods for all groups
#'
#' @inheritParams att_gt
#' @param weightsname The name of the column containing the sampling weights.
#'  If not set, all observations have same weight.
#' @param est_method the method to compute group-time average treatment
#'  effects.  The default for `conditional_did_pretest` is "ipw" for
#'  inverse probability weighting.  Other built-in methods include "dr"
#'  which uses the doubly robust approach in the `DRDID` package and "reg"
#'  for first step regression estimators.
#' @param bstrap Not used by the pre-test.  Critical values and p-values
#'  for the Cramer von Mises and Kolmogorov-Smirnov test statistics are
#'  always computed with the multiplier bootstrap (using `biters`
#'  iterations), regardless of this argument, and no standard errors are
#'  reported.
#' @param biters The number of multiplier bootstrap iterations used to
#'  simulate the limiting distribution of the test statistics.  The
#'  default is 1000.
#' @param clustervars A vector of variables names to cluster on (the
#'  multiplier bootstrap then draws cluster-level multipliers).  At most,
#'  there can be two variables (otherwise will throw an error) and one of
#'  these must be the same as idname which allows for clustering at the
#'  individual level.
#' @param cores The number of cores to use for parallel processing. This
#'  parallelizes Step 1 (computing the test statistic); Step 2's multiplier
#'  bootstrap is vectorized and runs in a single process.
#'
#' @examples
#' \dontrun{
#' data(mpdta)
#' pre.test <- conditional_did_pretest(yname="lemp",
#'                                     tname="year",
#'                                     idname="countyreal",
#'                                     gname="first.treat",
#'                                     xformla=~lpop,
#'                                     data=mpdta)
#' summary(pre.test)
#' }
#'
#' @references Callaway, Brantly and Sant'Anna, Pedro H. C.  "Difference-in-Differences with Multiple Time Periods and an Application on the Minimum Wage and Employment." Working Paper <https://arxiv.org/abs/1803.09015v2> (2018).
#'
#' @return an \code{\link{MP.TEST}} object
#' @export
conditional_did_pretest <- function(yname,
                                    tname,
                                    idname=NULL,
                                    gname,
                                    xformla=NULL,
                                    data,
                                    panel=TRUE,
                                    allow_unbalanced_panel=FALSE,
                                    control_group=c("nevertreated","notyettreated"),
                                    weightsname=NULL,
                                    alp=0.05,
                                    bstrap=TRUE,
                                    cband=TRUE,
                                    biters=1000,
                                    clustervars=NULL,
                                    est_method="ipw",
                                    print_details=FALSE,
                                    pl=FALSE,
                                    cores=1) {

  message("We are no longer updating this function.  It should continue to work, but most users find the pre-tests already reported by the `att_gt` function to be sufficient for most empirical applications.")

  if (missing(control_group)) control_group <- "nevertreated"

  # this is a DIDparams object
  dp <- pre_process_did(yname=yname,
                        tname=tname,
                        idname=idname,
                        gname=gname,
                        xformla=xformla,
                        data=data,
                        panel=panel,
                        allow_unbalanced_panel=allow_unbalanced_panel,
                        control_group=control_group,
                        weightsname=weightsname,
                        alp=alp,
                        bstrap=bstrap,
                        cband=cband,
                        biters=biters,
                        clustervars=clustervars,
                        est_method=est_method,
                        print_details=print_details,
                        pl=pl,
                        cores=cores
  )


  data <- dp$data
  tlist <- dp$tlist
  glist <- dp$glist
  n <- dp$n

  # check if possible to do test
  # note: tlist[2] contains the 2nd time period
  # (which is the first period where able to calculate ATT(g,t)'s)
  if ( max(glist) <= tlist[2] ) {
    stop("There are no pre-treatment periods available to conduct the conditional pre-test. Ensure your data has sufficient pre-treatment periods.")
  }


  # set which weight function to use
  # the only option that will work with current setup
  # is indicator so hard-code it here
  weightfun <- indicator


  if (allow_unbalanced_panel) {
    stop("The conditional pre-test is not supported for unbalanced panel data. Set `allow_unbalanced_panel = FALSE` to balance the panel, or use the pre-test reported by att_gt() instead.")
  }

  # create dataset with n observations;
  # recover covariates from this dataset
  if (panel & (allow_unbalanced_panel == FALSE)) {
    dta <- data[ data[,tname]==tlist[1], ]
  } else {
    dta <- data
  }

  # need X twice, once to loop over, and once to calculate weighting function
  X <- model.matrix(xformla, dta)
  X1 <- X

  # Internal compute.att_gt speedups for the per-unit loop below:
  #  (1) pretreatment_cells_only -- post-treatment (g,t) cells are dropped
  #      right after the loop (the keepers filter retains only group > t), so
  #      compute.att_gt never computes them.
  #  (2) setup_precomp / y_override -- compute.att_gt's per-call setup (data
  #      copy, design matrix, per-period slicing, alignment check) depends on
  #      the outcome only through the per-period outcome slices, and the loop
  #      below only modifies the outcome column; build the y-invariant bundle
  #      once and pass each unit's modified outcome as a plain vector instead
  #      of copying the whole data.frame n times. Used only when the balanced,
  #      id-aligned panel precompute applies; otherwise the legacy data-copy
  #      path inside the loop is kept.
  # Both are bit-identical to the legacy loop (same kept cells, same inputs to
  # every 2x2 estimator, and compute.att_gt consumes no RNG).
  basedp <- dp
  basedp$print_details <- FALSE
  basedp$pretreatment_cells_only <- TRUE
  use_y_override <- FALSE
  if (dp$panel) {
    setupdp <- basedp
    setupdp$return_setup <- TRUE
    setup_precomp <- compute.att_gt(setupdp)
    if (!is.null(setup_precomp) && isTRUE(setup_precomp$use_precompute_panel)) {
      basedp$setup_precomp <- setup_precomp
      use_y_override <- TRUE
      id_col <- data[, idname]
      y_col <- data[, yname]
    }
  }

  cat("Step 1 of 2: Computing test statistic....\n")
  out <- pbapply::pblapply(1:nrow(X), function(i) {
    # these are the weights for the conditional moment test
    # indicator weights
    www <- as.numeric(weightfun(X1, X[i,]))
    # for indicator weights, just choose rows where weights = 1
    rightids <- dta[,idname][www==1]

    # set new parameters to pass to call to compute.att_gt
    thisdp <- basedp

    # set the outcome to be the outcomes multiplied by the weighting function
    # (***this *trick* is only going to work for indicator weights***
    # (otherwise you will multiply by www twice))
    if (use_y_override) {
      # pass only the modified outcome vector; the y-invariant setup bundle
      # built above is reused by compute.att_gt
      ynew <- numeric(length(y_col))
      sel <- id_col %in% rightids
      ynew[sel] <- y_col[sel]
      thisdp$y_override <- ynew
    } else {
      # legacy path: copy the whole dataset and overwrite the outcome column
      thisdata <- data
      thisdata[,yname] <- 0
      thisdata[ thisdata[,idname] %in% rightids,yname] <- data[ thisdata[,idname] %in% rightids,yname]
      thisdata$y <- thisdata[,yname]
      thisdp$data <- thisdata
    }

    # compute the test statistic with call to compute.att_gt
    Jres <- compute.att_gt(thisdp)

    # turn in into more usable format
    J.results <- process_attgt(Jres$attgt.list)
    group <- J.results$group
    att <- J.results$att
    tt <- J.results$tt
    inf.func <- Jres$inffunc

    # return the results for this weighting function
    list(J=att, group=group, t=tt, inf.func=inf.func)
  }, cl=cores)

  # drop post-treatment (g,t). compute.att_gt now skips these cells entirely
  # (pretreatment_cells_only above), so this filter is a no-op retained as a
  # safety net in case a code path reintroduces post-treatment cells.
  out <- lapply(out, function(Js) {
    # which elements of results to keep
    keepers <- Js$group > Js$t
    this.group <- Js$group[keepers]
    this.t <- Js$t[keepers]
    this.J <- as.matrix(Js$J[keepers])
    this.inf.func <- as.matrix(Js$inf.func[,keepers])
    list(J=this.J, group=this.group, t=this.t, inf.func=this.inf.func)
  })

  # grab an array of the influence function for the test statistic
  Jinf.func <- simplify2array(BMisc::get_list_element(out, "inf.func"))

  # get the test statistic for all values of g,t,x
  J.inner <- sapply(out, function(o) o$J)

  # Orient J as (nX rows = covariate threshold values) x (n_gt cols = pre-treatment
  # (g,t) cells) so the observed CvM statistic below uses the SAME (1/nX)
  # normalization as the bootstrap null distribution in test.mboot() (which forms
  # Jb = t(apply(Ub*inf.func, c(2,3), mean)) and is therefore (nX x n_gt)).
  #
  # sapply() returns an (n_gt x nX) MATRIX when there are multiple pre-treatment
  # cells, but a length-nX VECTOR when there is only one -- so transpose the matrix
  # case and column-ify the vector case.
  #
  # BUG FIX (gh issue: pre-test spuriously rejects): this was previously
  #   ifelse(class(J.inner) == "matrix", J <- t(J.inner), J <- as.matrix(J.inner))
  # which was correct under R < 4.0 but silently broke in R >= 4.0, where
  # class(<matrix>) became c("matrix","array") (length 2). ifelse() then evaluated
  # BOTH assignment branches and the second (as.matrix, no transpose) always won, so
  # for the multi-cell case J stayed (n_gt x nX). Because sum(apply(J^2, 2, mean)) =
  # (1/nrow(J)) * sum(J^2), that mis-orientation scaled the observed CvM by
  # nX/n_gt = n/n_gt relative to the bootstrap, driving the p-value to ~0 and making
  # the pre-test over-reject whenever there was more than one pre-treatment (g,t)
  # cell. Use is.matrix() to restore the intended orientation.
  if (is.matrix(J.inner)) {
    J <- t(J.inner)
  } else {
    J <- as.matrix(J.inner)
  }

  # compute CvM test statistic by averaging over X, and summing over g and t
  CvM <- n*sum(apply(J^2, 2, mean))


  #-----------------------------------------------------------------------------
  # use the multiplier bootstrap to simulate limiting distribution of
  # test statistic
  cat("Step 2 of 2: Simulating limiting distribution of test statistic....\n")
  boot.res <- test.mboot(Jinf.func, dp, cores=cores)


  #-----------------------------------------------------------------------------
  ## # some debugging code
  ## # keeping in case helpful later on
  ## ddd <- 1
  ## bout <- sapply(1:100, function(b) {

  ##   Jstar <- apply(sample(c(-1,1), size=n, replace=TRUE)*Jinf.func[,ddd,], 2, mean)
  ##   mean((sqrt(n)*Jstar)^2)
  ## })

  ## ts <- mean((sqrt(n)*J[,ddd])^2)
  ## 1-ecdf(bout)(ts)


  ## ts2.inner <- apply( (sqrt(n)*J)^2, 2, mean)
  ## ts2 <- sum(ts2.inner)
  ## bout2 <- sapply(1:100, function(b) {

  ##   Jstar <- t(apply(sample(c(-1,1), size=n, replace=TRUE)*Jinf.func[,,], c(2,3), mean))
  ##   cvm.inner <- apply((sqrt(n)*Jstar)^2, 2, mean)
  ##   sum(cvm.inner)
  ## })
  ## 1-ecdf(bout2)(ts2)
  ## #-----------------------------------------------------------------------------

  # bootstrap results
  CvMb <- boot.res$bres
  # bootstrap critical value
  CvM.crit.val <- boot.res$crit.val
  # bootstrap p-value
  CvMpval <- 1-ecdf(CvMb)(CvM)

  #-----------------------------------------------------------------------------
  # KS Test - not reporting any more
  #
  # KSb <- unlist(lapply(bout, function(b1) b1$KSb))
  # KSocval <- quantile(KSb, probs=(1-alp), type=1)
  # KSpval <- 1-ecdf(KSb)(KS)
  #-----------------------------------------------------------------------------

  # return test results
  out <- MP.TEST(CvM=CvM, CvMb=CvMb, CvMcval=CvM.crit.val, CvMpval=CvMpval, clustervars=clustervars, xformla=xformla)
  out
}



## #' @title expf
## #'
## #' @description exponential weighting function
## #'
## #' @param X matrix of X's from the data
## #' @param u a particular value to multiply times the X's
## #'
## #' @return numeric vector
## #' @examples
## #' data(mpdta)
## #' dta <- subset(mpdta, year==2007)
## #' X <- model.matrix(~lpop, data=dta)
## #' X <- expf(X, X[1,])
## #'
## #' @export
## expf <- function(X, u) {
##     exp(X%*%u)
## }

#' @title indicator
#'
#' @description indicator weighting function
#'
#' @param X matrix of X's from the data
#' @param u a particular value to compare X's to
#'
#' @return numeric vector
#'
#' @examples
#' data(mpdta)
#' dta <- subset(mpdta, year==2007)
#' X <- model.matrix(~lpop, data=dta)
#' X <- indicator(X, X[1,])
#'
#' @export
indicator <- function(X, u) {
  # Row i is 1 iff X[i, j] <= u[j] for EVERY column j. Vectorized: rep(u, each =
  # nrow(X)) lines u up with X's column-major layout, so (X <= .) is the elementwise
  # comparison and a row passes iff none of its entries exceed u (rowSums == ncol).
  # Bit-identical to the prior t(apply(X, 1, `<=`)) + apply(all) for finite X
  # (model.matrix output is finite), but O(n*p) without the per-row closure.
  1 * (rowSums(X <= rep(u, each = nrow(X))) == ncol(X))
}



#' @title Multiplier Bootstrap for Conditional Moment Test
#'
#' @description A slightly modified multiplier bootstrap procedure
#'  for the pre-test of the conditional parallel trends assumption
#'
#' @inheritParams mboot
#' @param cores Unused; retained for backward compatibility. The
#'  multiplier bootstrap is computed with vectorized matrix operations
#'  in a single process, so this argument has no effect here. In
#'  \code{conditional_did_pretest}, \code{cores} parallelizes only
#'  Step 1 (computing the test statistic).
#'
#' @return list
#' \item{bres}{CvM test statistics for each bootstrap iteration}
#' \item{crit.val}{critical value for CvM test statistic}
#'
#' @export
test.mboot <- function(inf.func, DIDparams, cores=1) {
  validate_positive_whole_number(cores, "cores")
  if (!is.numeric(inf.func) || length(dim(inf.func)) != 3L ||
      any(dim(inf.func) <= 0L)) {
    stop("inf.func must be a numeric three-dimensional array with positive dimensions.")
  }
  if (!is.list(DIDparams)) {
    stop("DIDparams must be a list or DIDparams object.")
  }

  # setup needed variables
  data <- DIDparams$data
  idname <- DIDparams$idname
  clustervars <- DIDparams$clustervars
  biters <- DIDparams$biters
  tname <- DIDparams$tname
  if (!is.data.frame(data)) {
    stop("DIDparams$data must be a data.frame.")
  }
  validate_column_name(idname, "DIDparams$idname", names(data))
  validate_column_name(tname, "DIDparams$tname", names(data))
  validate_column_names(clustervars, "DIDparams$clustervars", names(data), allow_null = TRUE)
  tlist <- unique(data[,tname])[order(unique(data[,tname]))]
  alp <- DIDparams$alp
  panel <- DIDparams$panel
  validate_positive_whole_number(biters, "biters")
  validate_alp(alp)
  validate_logical_scalar(panel, "DIDparams$panel")

  # just get n obsevations (for clustering below...)
  if (panel) {
    dta <- data[ data[,tname]==tlist[1], ]
  } else {
    dta <- data
  }
  n <- nrow(dta)
  if (dim(inf.func)[1] != n) {
    stop("inf.func first dimension must match the number of bootstrap observations.")
  }

  # if include id as variable to cluster on
  # drop it as we do this automatically
  if (idname %in% clustervars) {
    clustervars <- clustervars[-which(clustervars==idname)]
  }

  # we can only handle up to 2-way clustering
  # (in principle could do more, but not high priority now)
  if (length(clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # ---------------------------------------------------------------------------
  # Vectorized multiplier bootstrap (replaces a biters-long pbsapply loop that
  # multiplied a full n x k x nX influence array by fresh Rademacher weights and
  # ran apply(., c(2,3), mean) on every draw -- O(n*k*nX) compute AND an O(n*k*nX)
  # transient array allocation per draw, ~1.2 GB x biters at n in the thousands).
  #
  # The per-draw statistic is, writing inf[i,j,x] for the influence array,
  #   Jb[x,j] = (1/n) sum_i U_b[i] inf[i,j,x]
  #   CvMb_b  = n * mean_x( sum_j Jb[x,j]^2 ) = (n/nX) * sum_{x,j} Jb[x,j]^2.
  # Stacking the weights of all draws into V (n x biters) and reshaping inf to
  # M (n x k*nX), Jb for every draw is crossprod(V, M)/n in one BLAS call, and
  # CvMb is (n/nX) * rowSums(.^2). Drawing all biters*n (or biters*n1) Rademacher
  # values in a single sample() call consumes the RNG stream identically to the
  # per-draw loop (each value is one R_unif_index(2)), so results match the old
  # loop up to floating-point summation order (~1e-14).
  # ---------------------------------------------------------------------------
  dms <- dim(inf.func)
  k <- dms[2]; nX <- dms[3]

  # Build the n x biters multiplier matrix; columns are independent draws.
  if (length(clustervars) > 0) {
    # Rademacher weights are constant within cluster; map each unit to its cluster's
    # weight via match() exactly as the original loop did.
    uc <- unique(dta[, clustervars])
    n1 <- length(uc)
    Vc <- matrix(sample(c(-1, 1), as.numeric(n1) * biters, replace = TRUE), nrow = n1)
    V <- Vc[match(dta[, clustervars], uc), , drop = FALSE]
  } else {
    V <- matrix(sample(c(-1, 1), as.numeric(n) * biters, replace = TRUE), nrow = n)
  }

  # Tile over the X dimension so peak memory is n*k*chunk doubles instead of the
  # full n*k*nX reshape. CvMb is a plain sum over X, so chunking is exact (the only
  # difference is summation order). Chunk targets ~5e6 doubles (~40 MB) per slice.
  chunk <- max(1L, as.integer(5e6 %/% (as.numeric(n) * k)))
  bout <- numeric(biters)
  for (start in seq.int(1L, nX, by = chunk)) {
    xs <- start:min(start + chunk - 1L, nX)
    Mc <- matrix(inf.func[, , xs, drop = FALSE], nrow = n)   # n x (k*length(xs))
    Jf <- crossprod(V, Mc) / n                                # biters x (k*length(xs))
    bout <- bout + rowSums(Jf^2)
  }
  bout <- (n / nX) * bout

  crit.val <- quantile(bout, 1-alp, type=1)

  list(bres=bout, crit.val=crit.val)
}

#' @title Process `did` Function Arguments
#'
#' @description Function to process arguments passed to the main methods in the
#'  `did` package as well as conducting some tests to make sure
#'  data is in proper format / try to throw helpful error messages.
#'
#' @inheritParams att_gt
#' @param call Function call to att_gt
#'
#' @return a [`DIDparams`] object
#'
#' @export
pre_process_did <- function(yname,
                            tname,
                            idname,
                            gname,
                            xformla = NULL,
                            data,
                            panel = TRUE,
                            allow_unbalanced_panel,
                            control_group = c("nevertreated","notyettreated"),
                            anticipation = 0,
                            weightsname = NULL,
                            fix_weights = NULL,
                            alp = 0.05,
                            bstrap = FALSE,
                            cband = FALSE,
                            biters = 1000,
                            clustervars = NULL,
                            est_method = "dr",
                            base_period = "varying",
                            print_details = TRUE,
                            faster_mode = FALSE,
                            pl = FALSE,
                            cores = 1,
                            call = NULL) {
  #-----------------------------------------------------------------------------
  # Data pre-processing and error checking
  #-----------------------------------------------------------------------------
  # set control group
  if (missing(control_group)) control_group <- "nevertreated"
  validate_choice_scalar(
    control_group,
    "control_group",
    c("nevertreated", "notyettreated"),
    "control_group must be either 'nevertreated' or 'notyettreated'"
  )
  validate_choice_scalar(
    base_period,
    "base_period",
    c("universal", "varying"),
    "base_period must be either 'universal' or 'varying'."
  )
  validate_logical_scalar(panel, "panel")
  validate_logical_scalar(allow_unbalanced_panel, "allow_unbalanced_panel")
  validate_logical_scalar(bstrap, "bstrap")
  validate_logical_scalar(cband, "cband")
  validate_logical_scalar(faster_mode, "faster_mode")
  validate_logical_scalar(print_details, "print_details")
  validate_logical_scalar(pl, "pl")
  validate_positive_whole_number(cores, "cores")
  validate_anticipation(anticipation)
  validate_alp(alp)
  if (bstrap) validate_positive_whole_number(biters, "biters")
  validate_xformla(xformla)
  # make sure dataset is a data.frame
  # this gets around RStudio's default of reading data as tibble
  if (!all( class(data) == "data.frame")) {
    data <- as.data.frame(data)
  }

  data_names <- names(data)
  validate_column_name(yname, "yname", data_names)
  validate_column_name(tname, "tname", data_names)
  validate_column_name(gname, "gname", data_names)
  validate_column_name(idname, "idname", data_names, allow_null = !panel)
  validate_column_name(weightsname, "weightsname", data_names, allow_null = TRUE)
  validate_column_names(clustervars, "clustervars", data_names, allow_null = TRUE)

  check_reserved_did_names(yname = yname, tname = tname, idname = idname,
                           gname = gname, xformla = xformla,
                           weightsname = weightsname,
                           clustervars = clustervars)

  # validate that all required column names exist in the data
  required_cols <- c(yname, tname, idname, gname, weightsname, clustervars)
  missing_cols <- setdiff(required_cols, colnames(data))
  if (length(missing_cols) > 0) {
    stop("The following column(s) are not found in the data: ",
         paste(missing_cols, collapse = ", "), ". ",
         "Please check the spelling of yname, tname, idname, gname, weightsname, and clustervars.")
  }

  # At most one cluster variable beyond idname is supported (clustering at the
  # unit level via idname is implicit). Enforced here -- mirroring the fast path
  # (pre_process_did2) and mboot(), with identical wording -- so every
  # faster_mode x bstrap combination rejects the input up front; the analytical
  # (bstrap = FALSE) path used to silently cluster on the first extra variable only.
  cv_check <- clustervars
  if (!is.null(cv_check) && !is.null(idname) && (idname %in% cv_check)) {
    cv_check <- setdiff(cv_check, idname)
  }
  if (length(cv_check) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # make sure time periods are numeric
  if (! (is.numeric(data[, tname])) ) stop("The time variable '", tname, "' must be numeric. Please convert it.")

  #  make sure gname is numeric
  if (! (is.numeric(data[, gname])) ) stop("The group variable '", gname, "' must be numeric. Please convert it.")

  # gname must be 0 (never-treated) or a positive treatment-timing value.
  # Negative codes are not supported: 0 is reserved for never-treated, so a
  # non-positive time scale is ambiguous. Reject them up front so both code
  # paths behave identically (the fast path previously accepted negative codes
  # silently while this slow path later errored with "No valid groups").
  if (any(data[, gname] < 0, na.rm = TRUE)) {
    stop("The group variable '", gname, "' must be 0 (never-treated) or a ",
         "positive treatment-timing value; negative values are not supported. ",
         "If your time periods are non-positive, shift them so the earliest ",
         "period is >= 1.")
  }

  #  make sure the outcome is numeric (logical 0/1 outcomes are also allowed)
  if (! (is.numeric(data[, yname]) || is.logical(data[, yname])) ) stop("The outcome variable '", yname, "' must be numeric. Please convert it.")

  # put in blank xformla if no covariates or check whether all variables are in data
  if (is.null(xformla)) {
    xformla <- ~1
  } else {
    # extract variable names from the formula
    formula_vars <- all.vars(xformla)

    # identify variables in xformla not in data
    missing_vars <- setdiff(formula_vars, names(data))

    # error checking for missing variables in data
    if (length(missing_vars) > 0) {
      stop(paste("The following variables are not in data:", paste(missing_vars, collapse = ", ")), call. = FALSE)
    }
  }

  # drop irrelevant columns from data. Keep the RAW covariate variables (all.vars of
  # xformla), NOT the evaluated model.frame, so model.matrix(xformla, .) can be
  # rebuilt downstream. This is what makes transform formulae work (e.g. ~I(X^2),
  # ~poly(X, 2), ~log(X)): the evaluated model.frame would store columns named by the
  # transformed expression -- losing the raw variable that model.matrix needs -- and,
  # for matrix-valued terms like poly(), a matrix column that later breaks the
  # data.table coercion in compute.att_gt(). For bare-variable formulae (e.g. ~X,
  # ~X1+X2) and factor covariates the kept columns are identical to before.
  xvars <- all.vars(xformla)
  keep_cols <- unique(c(idname, tname, yname, gname, weightsname, clustervars, xvars))
  data <- data[, keep_cols, drop = FALSE]

  # check if any covariates were missing
  n_orig <- nrow(data)
  # drop rows with any missing or non-finite id / time / outcome / weight /
  # cluster or RAW covariate value. gname is excluded from the finite check
  # because Inf is a valid never-treated code there (see complete_finite_cases);
  # missing/NaN gname is still dropped via complete.cases().
  data <- data[complete_finite_cases(data, finite_exclude = gname), ]
  # also drop rows whose EVALUATED design is missing/non-finite (e.g. log of a non-positive
  # covariate), preserving the previous model.frame-based row dropping. We use
  # model.frame (NOT model.matrix) with na.action = na.pass: model.frame keeps EVERY
  # row -- including those where a term evaluates to NA/NaN -- so complete.cases()
  # flags them and the indicator stays aligned with `data`. (model.matrix would
  # instead silently drop the NaN rows, making the mask shorter than `data` and the
  # offending rows survive.)
  # Safe to evaluate now that raw-covariate NAs have been removed (so poly()/ns()/...
  # will not error on NA input).
  if (length(xvars) > 0L && nrow(data) > 0L) {
    mf_check <- suppressWarnings(model.frame(xformla, data = data, na.action = na.pass))
    finite_rows <- complete_finite_cases(mf_check)
    if (!all(finite_rows)) data <- data[finite_rows, ]
  }
  n_diff <- n_orig - nrow(data)
  if (n_diff != 0) {
    warning(paste0("dropped ", n_diff, " rows from original data due to missing or non-finite data"))
  }

  # weights if null
  if (is.null(weightsname)) w <- rep(1, nrow(data)) else w <- data[,weightsname]

  # Validate user-supplied weights: negative weights flip signs and a non-positive
  # mean divides by ~0 during normalization, silently producing NA/NaN ATTs.
  if (!is.null(weightsname) && (any(w < 0, na.rm = TRUE) || isTRUE(mean(w, na.rm = TRUE) <= 0)))
    stop("The weights variable '", weightsname, "' must be non-negative with a positive mean.")

  if (".w" %in% colnames(data)) stop("Your data already contains a column named '.w', which is reserved for internal use by `did`. Please rename this column before calling att_gt().")
  data$.w <- w

  # Check for time-varying weights in panel data. Grouped max/min via data.table
  # (GForce-optimized) is much faster than tapply() with a per-unit closure and
  # produces the same per-unit ranges as diff(range(x)) on the raw weights.
  if (!is.null(weightsname) && panel) {
    dtw <- data.table(id = data[[idname]], .w = data[[weightsname]])
    w_rng <- dtw[, .(mx = max(.w), mn = min(.w)), by = "id"]
    if (any((w_rng$mx - w_rng$mn) > .Machine$double.eps^0.5, na.rm = TRUE)) {
      message(
        "Time-varying weights detected. For balanced panel data, the default ",
        "behavior uses the weight from the earlier of the two time periods in ",
        "each 2x2 comparison (the base period for post-treatment cells). ",
        "Use the 'fix_weights' argument to control this behavior. ",
        "See ?att_gt for details."
      )
    }
  }

  # Outcome variable will be denoted by y
  # data$.y <- data[, yname]

  # figure out the dates
  # list of dates from smallest to largest
  tlist <- sort(unique(data[,tname]))

  # Groups with treatment time bigger than max time period + anticipation are considered to be never treated.
  # We account for anticipation because units treated shortly after the last observed period
  # may already exhibit anticipatory effects during the observed time window, and should
  # not be used as controls.
  asif_never_treated <- (data[,gname] > max(tlist, na.rm = TRUE) + anticipation)
  asif_never_treated[is.na(asif_never_treated)] <- FALSE
  data[asif_never_treated, gname] <- 0

  # list of treated groups (by time) from smallest to largest
  glist <- sort(unique(data[,gname]))


  # Check if there is a never treated group

  # if ( length(glist[glist==0]) == 0) {
  #   if(control_group=="nevertreated"){
  #     stop("There is no available never-treated group")
  #   } else {
  #     # Drop all time periods with time periods >= latest treated
  #     data <- subset(data,(data[,tname] < (max(glist)-anticipation)))
  #     # Replace last treated time with zero
  #     # lines.gmax <- data[,gname]==max(glist, na.rm = TRUE)
  #     # data[lines.gmax,gname] <- 0
  #
  #     tlist <- sort(unique(data[,tname]))
  #     glist <- sort(unique(data[,gname]))
  #
  #     # don't comput ATT(g,t) for groups that are only treated at end
  #     # and only play a role as a comparison group
  #     glist <- glist[ glist < max(glist)]
  #   }
  # }

  if (!any(glist == 0)) {
    # Compute latest treated cohort once, and the cutoff time
    latest_g <- max(glist, na.rm = TRUE)
    cutoff_t  <- latest_g - anticipation

    if (control_group == "nevertreated") {
      # Warn the user
      warning(
        "No never-treated group is available. ",
        "The last treated cohort is being coerced as 'never-treated' units, and data from periods after that is being filtered out (no available comparison groups)."
      )

      # Drop all periods ≥ (latest_g - anticipation)
      data <- data[ data[[ tname ]] < cutoff_t, , drop = FALSE ]

      # For any row where gname == latest_g, set gname := 0
      lines.gmax <- data[, gname]==latest_g
      data[lines.gmax, gname] <- 0
    } else {
      # If "notyettreated", we simply drop those periods and leave gnames alone
      data <- data[ data[[ tname ]] < cutoff_t, , drop = FALSE ]
    }

    # Recompute tlist and glist from the filtered/modified data
    tlist <- sort(unique(data[,tname]))
    glist <- sort(unique(data[,gname]))


    # If control_group != "nevertreated", drop the max cohort from glist
    if (control_group != "nevertreated") {
      glist <- glist[glist < latest_g]
    }
  }

  # Only the treated groups
  glist <- glist[glist>0]

  # drop groups treated in the first period or before
  first.period <- tlist[1]
  glist <- glist[glist > first.period + anticipation]

  # check for groups treated in the first period (accounting for anticipation) and drop these
  # nfirstperiod <- length(unique(data[ !((data[,gname] > first.period) | (data[,gname]==0)), ] )[,idname])
  treated_first_period <- ( data[,gname] <= first.period + anticipation ) & ( !(data[,gname]==0) )
  treated_first_period[is.na(treated_first_period)] <- FALSE
  # if/else (not ifelse) so the data subset is built only for the relevant branch;
  # the result is identical to the previous ifelse(panel, ...) expression.
  nfirstperiod <- if (panel) length(unique(data[treated_first_period, idname])) else sum(treated_first_period)
  if ( nfirstperiod > 0 ) {
    warning(paste0("Dropped ", nfirstperiod, " units that were already treated in the first period",
                    if (anticipation > 0) paste0(" (accounting for anticipation = ", anticipation, ")") else "",
                    "."))
    # Drop ONLY the first-period-treated units, by row identity. The previous
    # `data[gname %in% c(0, glist)]` dropped by cohort membership in glist, which --
    # when there is no never-treated group -- also deleted the latest cohort that was
    # deliberately removed from glist above (the `glist[glist < latest_g]` trim) so it
    # could serve as a not-yet-treated control. That silently deleted a valid
    # comparison cohort and corrupted ATT(g,t) for the other groups; the
    # treated_first_period mask removes exactly the already-treated units, nothing else.
    data <- data[ !treated_first_period, , drop = FALSE ]
    # update tlist and glist
    tlist <- sort(unique(data[,tname]))
    glist <- sort(unique(data[,gname]))
    glist <- glist[glist>0]

    # drop groups treated in the first period or before
    first.period <- tlist[1]
    glist <- glist[glist > first.period + anticipation]

    # The latest cohort stays in the data as a not-yet-treated control but, when there
    # is still no never-treated group, must remain excluded from glist (it gets no ATT
    # of its own) -- mirroring the exclusion above for the nfirstperiod == 0 case.
    if (control_group != "nevertreated" && !any(data[,gname] == 0)) {
      glist <- glist[glist < latest_g]
    }

  }

  #  make sure id is numeric
  if (! is.null(idname)){
    #  make sure id is numeric
    if (! (is.numeric(data[, idname])) ) stop("The id variable '", idname, "' must be numeric. Please convert it.")

    # Validate treatment irreversibility and (id, period) uniqueness with a single
    # radix order over (id, t) plus O(n) adjacent-element scans, instead of the much
    # slower unique.data.frame()/anyDuplicated.data.frame() row-key machinery. After
    # ordering, each unit's rows are contiguous, so g varying within a unit shows up
    # as adjacent rows with equal id but differing g, and a duplicated (id, period)
    # pair as adjacent rows with equal id and equal t. All three columns are numeric
    # and NA-free at this point, so the comparisons are exact.
    idv <- data[, idname]
    nn <- length(idv)
    if (nn >= 2L) {
      tv <- data[, tname]
      o <- order(idv, tv, method = "radix")
      io <- idv[o]
      same_id <- io[-1L] == io[-nn]

      # Check that gname is time-invariant within each unit (treatment irreversibility).
      go <- data[, gname][o]
      if (any(same_id & (go[-1L] != go[-nn]))) {
        stop("The value of gname (treatment variable) must be the same across all periods for each particular unit. The treatment must be irreversible.")
      }

      # Check that (idname, tname) is unique: each unit observed at most once per
      # period. Mirrors the fast path (pre_process_did2.R) so both code paths reject
      # duplicated (id, period) rows identically -- without this guard the slow path
      # silently produced incorrect estimates on long-format data with duplicates.
      to <- tv[o]
      if (any(same_id & (to[-1L] == to[-nn]))) {
        stop("The value of idname must be unique (by tname). Some units are observed more than once in a period.")
      }

      # Check that cluster variables are time-invariant within each unit, mirroring
      # the fast path (validate_args) and mboot() -- with identical wording -- so
      # invalid clustering inputs are rejected up front regardless of bstrap.
      # Without this, the analytical (bstrap = FALSE) path fell back to i.i.d.
      # standard errors with a warning advising bstrap = TRUE, advice that then
      # fails in mboot() for the very same input. Reuses the (id, t) radix order
      # and adjacency scan above.
      for (cvar in setdiff(clustervars, idname)) {
        cvo <- data[o, cvar]
        if (any(same_id & (cvo[-1L] != cvo[-nn]))) {
          stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
        }
      }
    }
  }



  # if user specifies repeated cross sections,
  # set that it really is repeated cross sections
  true_repeated_cross_sections <- FALSE
  if (!panel) {
    true_repeated_cross_sections <- TRUE
  }

  #-----------------------------------------------------------------------------
  # setup data in panel case
  #-----------------------------------------------------------------------------
  # Check if data is a balanced panel if panel = TRUE and allow_unbalanced_panel = TRUE
  bal_panel_test <- panel*allow_unbalanced_panel
  if (bal_panel_test) {
    # data is already complete-case filtered above and (id, period) uniqueness has
    # been validated, so the panel is balanced iff every unit appears in every
    # period, i.e. nrow(data) equals (number of units) x (number of periods). This
    # replaces a BMisc::makeBalancedPanel() round-trip that built a balanced copy
    # of the data just to compare unit counts and then threw it away.
    allow_unbalanced_panel <-
      nrow(data) != length(unique(data[[idname]])) * as.numeric(length(unique(data[[tname]])))
  }



  if (panel) {

    # check for unbalanced panel
    if (allow_unbalanced_panel) {

      # Flag for true repeated cross sections
      panel <- FALSE
      true_repeated_cross_sections <- FALSE

    } else {

      # this is the case where we coerce balanced panel

      # check for complete cases (rows with missing data were already dropped
      # above, so anyNA() short-circuits the redundant complete.cases() pass and
      # full-table subset copies in the common case)
      if (anyNA(data)) {
        keepers <- complete.cases(data)
        n <- length(unique(data[,idname]))
        n.keep <- length(unique(data[keepers,idname]))
        if (!all(keepers)) {
          warning(paste0("Dropped ", (n-n.keep), " observations that had missing data."))
          data <- data[keepers,]
        }
      }

      # make it a balanced data set: keep only units observed in every period.
      # A unit is fully observed iff its row count equals the number of distinct
      # periods ((id, period) uniqueness was validated above), so this
      # count-and-filter keeps exactly the same rows as BMisc::makeBalancedPanel()
      # without the per-group .SD materialization; any row-order difference is
      # normalized by the (id, period) sort below.
      uid <- unique(data[[idname]])
      n.old <- length(uid)
      cnt <- tabulate(match(data[[idname]], uid))
      keep_bal <- data[[idname]] %in% uid[cnt == length(unique(data[[tname]]))]
      if (!all(keep_bal)) data <- data[keep_bal, , drop = FALSE]
      n <- length(unique(data[,idname]))
      if (n < n.old) {
        # n.old - n counts dropped UNITS (unique ids), not unit-time rows; word the
        # warning accordingly, with the same text as the fast path (pre_process_did2)
        warning(n.old - n, " units are missing in some periods. Converting to balanced panel by dropping them.")
      }

      # If drop all data, you do not have a panel.
      if (nrow(data)==0) {
        stop("All observations dropped while converting data to balanced panel. Consider setting `panel = FALSE` and/or revisiting 'idname'.")
      }

      n <- nrow(data[ data[,tname]==tlist[1], ])

      # slow, repeated check here...
      ## # check that first.treat doesn't change across periods for particular individuals
      ## if (!all(sapply( split(data, data[,idname]), function(df) {
      ##   length(unique(df[,gname]))==1
      ## }))) {
      ##   stop("The value of gname must be the same across all periods for each particular individual.")
      ## }

    }
  }

  #-----------------------------------------------------------------------------
  # code for setting up repeated cross sections (and unbalanced panel)
  #-----------------------------------------------------------------------------
  if (!panel) {

    # check for complete cases (rows with missing data were already dropped
    # above, so anyNA() short-circuits the redundant complete.cases() pass and
    # full-table subset copies in the common case)
    if (anyNA(data)) {
      keepers <- complete.cases(data)
      if (!all(keepers)) {
        warning(paste0("Dropped ", sum(!keepers), " observations that had missing data."))
        data <- data[keepers,]
      }
    }

    # If drop all data, you do not have a panel.
    if (nrow(data)==0) {
      stop("All observations were dropped due to missing data. Check your outcome, group, time, and covariate variables for missing values.")
    }

    # n-row data.frame to hold the influence function
    if (true_repeated_cross_sections) {
      data$.rowid <- seq(1:nrow(data))
      idname <- ".rowid"
    } else {
      # set rowid to idname for repeated cross section/unbalanced
      data$.rowid <- data[, idname]
    }

    # n is unique number of cross section observations
    # this is different for repeated cross sections and unbalanced panel
    n <- length(unique(data[,idname]))
  }

  ## # Update tlist and glist because of data handling
  ## # figure out the dates
  ## # list of dates from smallest to largest
  ## tlist <- unique(data[,tname])[order(unique(data[,tname]))]
  ## # list of treated groups (by time) from smallest to largest
  ## glist <- unique(data[,gname])[order(unique(data[,gname]))]

  ## # Only the treated groups
  ## glist <- glist[glist>0]

  ## # drop groups treated in the first period or before
  ## first.period <- tlist[1]
  ## glist <- glist[glist > first.period + anticipation]

  # Check if groups is empty (usually a problem with the way people defined groups)
  if(length(glist)==0){
    stop("No valid groups. The variable in 'gname' should be expressed as the time a unit is first treated (0 if never-treated).")
  }

  # if there are only two time periods, then uniform confidence
  # bands are the same as pointwise confidence intervals
  if (length(tlist)==2) {
    cband <- FALSE
  }

  #-----------------------------------------------------------------------------
  # more error handling after we have balanced the panel

  # check against very small groups. tabulate(match(.)) yields the same per-group
  # row counts as the previous aggregate() call without invoking an R closure per
  # group; sorted gvals reproduces aggregate()'s ascending group order, and the
  # Group.1/x column names are kept for the subset()/paste logic below.
  gvals <- sort(unique(data[[gname]]))
  gcnt <- tabulate(match(data[[gname]], gvals), nbins = length(gvals))
  gsize <- data.frame(Group.1 = gvals, x = gcnt / length(tlist))

  # how many in each group before give warning
  # 5 is just a buffer, could pick something else, but seems to work fine
  reqsize <- length(BMisc::rhs_vars(xformla)) + 5

  # which groups to warn about
  gsize <- subset(gsize, x < reqsize) # x is name of column from aggregate

  # warn if some groups are small
  if (nrow(gsize) > 0) {
    gpaste <-  paste(gsize[,1], collapse=",")
    warning(paste0("Some groups in your dataset have very few observations, which may cause estimation problems.\n  Check groups: ", gpaste, "."))

    if ( (0 %in% gsize[,1]) & (control_group == "nevertreated") ) {
      stop("The never-treated group is too small to serve as a reliable control. Try setting `control_group = 'notyettreated'` to include not-yet-treated units as controls.")
    }
  }
  #----------------------------------------------------------------------------

  # How many time periods
  nT <- length(tlist)
  # How many treated groups
  nG <- length(glist)

  # order dataset wrt idname and tname, in place. The sort keys are tie-free
  # ((id, period) uniqueness was validated above; .rowid is 1:n for repeated cross
  # sections), so the permutation is unique and the row order matches the previous
  # order()-based subset without a full-table copy. `data` is function-local (it
  # was subset/copied above), so the by-reference conversion cannot touch the
  # user's input; setDF() returns a plain data.frame for downstream consumers
  # (DIDparams, compute.att_gt).
  setDT(data)
  setorderv(data, c(idname, tname))
  setDF(data)

  # store parameters for passing around later
  dp <- DIDparams(yname=yname,
                  tname=tname,
                  idname=idname,
                  gname=gname,
                  xformla=xformla,
                  data=data,
                  control_group=control_group,
                  anticipation=anticipation,
                  weightsname=weightsname,
                  fix_weights=fix_weights,
                  alp=alp,
                  bstrap=bstrap,
                  biters=biters,
                  clustervars=clustervars,
                  cband=cband,
                  print_details=print_details,
                  faster_mode=faster_mode,
                  pl=pl,
                  cores=cores,
                  est_method=est_method,
                  base_period=base_period,
                  panel=panel,
                  true_repeated_cross_sections=true_repeated_cross_sections,
                  n=n,
                  nG=nG,
                  nT=nT,
                  tlist=tlist,
                  glist=glist,
                  call=call)
}

#' @title Aggregate Group-Time Average Treatment Effects
#'
#' @description A function to take group-time average treatment effects
#'  and aggregate them into a smaller number of parameters.  There are
#'  several possible aggregations including "simple", "dynamic", "group",
#'  and "calendar."
#'
#' @param MP an MP object (i.e., the results of the [att_gt()] method)
#' @param type Which type of aggregated treatment effect parameter to compute.
#'   One option is "simple" (this just computes a weighted average of all
#'   group-time average treatment effects with weights proportional to group
#'   size).  Other options are "dynamic" (this computes average effects across
#'   different lengths of exposure to the treatment and is similar to an
#'   "event study"; here the overall effect averages the effect of the
#'   treatment across all positive lengths of exposure); "group" (this
#'   is the default option and
#'   computes average treatment effects across different groups; here
#'   the overall effect averages the effect across different groups); and
#'   "calendar" (this computes average treatment effects across different
#'   time periods; here the overall effect averages the effect across each
#'   time period).
#' @param balance_e If set (and if one computes dynamic effects), it balances
#'  the sample with respect to event time.  For example, if `balance_e=2`,
#'  `aggte` will drop groups that are not exposed to treatment for
#'  at least three periods. (the initial period when `e=0` as well as the
#'  next two periods when `e=1` and the `e=2`).  This ensures that
#'  the composition of groups does not change when event time changes.
#' @param min_e For event studies, this is the smallest event time to compute
#'  dynamic effects for.  By default, `min_e = -Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param max_e For event studies, this is the largest event time to compute
#'  dynamic effects for.  By default, `max_e = Inf` so that effects at
#'  all lengths of exposure are computed.
#' @param na.rm Logical value if we are to remove missing values from analyses. Default is FALSE.
#' @param bstrap Boolean for whether or not to compute standard errors using
#'  the multiplier bootstrap.  Default is value set in the MP object.  If `bstrap=FALSE`, then analytical
#'  standard errors are reported; these are cluster-robust when a cluster variable was supplied to `att_gt`.
#' @param biters The number of bootstrap iterations to use.  The default is the value set in the MP object,
#'  and this is only applicable if `bstrap=TRUE`.
#'
#' @param cband Boolean for whether or not to compute a uniform confidence
#'  band that covers all of the group-time average treatment effects
#'  with fixed probability `1-alp`.  In order to compute uniform confidence
#'  bands, `bstrap` must also be set to `TRUE`.  The default is
#'  the value set in the MP object
#' @param alp the significance level, default is value set in the MP object.
#' @param clustervars A vector of variables to cluster on.  The default is
#'  the cluster variables set in the MP object (i.e., the ones passed to
#'  [att_gt()]).  Clustering at the aggregation stage reuses the per-unit
#'  cluster information stored by [att_gt()], so `aggte` can only cluster
#'  on `idname` and/or the variable that was passed to [att_gt()] via its
#'  `clustervars` argument.  If a requested cluster variable cannot be
#'  honored -- because `clustervars` was not supplied to [att_gt()], or
#'  because it names a different variable than the one [att_gt()] clustered
#'  on -- `aggte` issues a warning and reports standard errors that do NOT
#'  account for clustering.  To obtain clustered standard errors for the
#'  aggregated parameters, re-run [att_gt()] with the desired `clustervars`.

#'
#' @return An [`AGGTEobj`] object that holds the results from the
#'  aggregation
#'
#' @section Examples:
#'
#
#' Initial ATT(g,t) estimates from [att_gt()]
#'
#' ```{r, comment = "#>", collapse = TRUE}
#' data(mpdta)
#' set.seed(09152024)
#' out <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=NULL,
#'                data=mpdta)
#' ```
#'
#' You can aggregate the ATT(g,t) in many ways.
#'
#' **Overall ATT:**
#' ```{r, comment = "#>", collapse = TRUE}
#' aggte(out, type = "simple")
#' ```
#'
#' **Dynamic ATT (Event-Study):**
#' ```{r, comment = "#>", collapse = TRUE}
#' aggte(out, type = "dynamic")
#' ```
#'
#' **ATT for each group:**
#' ```{r, comment = "#>", collapse = TRUE}
#' aggte(out, type = "group")
#' ```
#'
#' **ATT for each calendar year:**
#' ```{r, comment = "#>", collapse = TRUE}
#' aggte(out, type = "calendar")
#' ```
#'
#'
#' @export
aggte <- function(MP,
                  type = "group",
                  balance_e = NULL,
                  min_e = -Inf,
                  max_e = Inf,
                  na.rm = FALSE,
                  bstrap = NULL,
                  biters = NULL,
                  cband = NULL,
                  alp = NULL,
                  clustervars = NULL
                  ) {

  call <- match.call()

  compute.aggte(MP = MP,
                type = type,
                balance_e = balance_e,
                min_e = min_e,
                max_e = max_e,
                na.rm = na.rm,
                bstrap = bstrap,
                biters = biters,
                cband = cband,
                alp = alp,
                clustervars = clustervars,
                call = call)
}

# Regression tests for two faster_mode standard-error bugs on UNBALANCED panels with clustering.
#
# (1) att_gt(): the analytical clustered SE (bstrap = FALSE, clustervars coarser than idname) silently
#     fell back to the i.i.d. SE under faster_mode = TRUE, because the stored DIDparams$cluster_vector was
#     observation-length (the unbalanced time-invariant data keeps every row) rather than unit-length, so
#     it no longer aligned with the influence function. It is now rebuilt from the data to align with the
#     influence-function rows whenever it is missing or mis-sized.
#
# (2) aggte(): the estimated-weight influence term (wif) was built from aggregate(data, by id, mean),
#     which is sorted by id, but the influence function is in the data's first-appearance unit order.
#     get_agg_inf_func() adds the two row-wise without realigning, so under faster_mode = TRUE on an
#     unbalanced panel the wif was added to the WRONG units and the aggregated SE was wrong (the point
#     estimate is unaffected because wif is mean-zero). The aggregation data is now ordered to match the
#     influence function.
#
# In both cases faster_mode = TRUE must reproduce faster_mode = FALSE (the correct reference) to
# numerical precision, for balanced and unbalanced panels alike.
# Units nested in unbalanced clusters that share a common shock each period (so the cluster, not the unit,
# is the independent sampling unit). Optionally drop rows to create an unbalanced panel and shuffle the
# row order, so that the internal first-appearance unit order differs from the sorted-by-id order -- the
# condition that triggered both bugs under faster_mode.
.make_ub_clustered <- function(seed, G = 40L, unbalanced = TRUE) {
  set.seed(seed)
  sz <- 1L + rpois(G, 4); N <- sum(sz); cl <- rep(seq_len(G), times = sz)
  alpha <- rnorm(G, 0, 1)[cl]; nu <- rnorm(N, 0, 1); Tt <- 6L
  eta <- matrix(rnorm(G * Tt, 0, 1), G, Tt)              # common shock per cluster, per period
  g <- sample(c(3L, 4L, 5L, 0L), N, replace = TRUE, prob = c(.2, .2, .2, .4))
  d <- data.frame(id = rep(seq_len(N), each = Tt), period = rep(seq_len(Tt), N),
                  cluster = rep(cl, each = Tt), g = rep(g, each = Tt),
                  a = rep(alpha, each = Tt), nu = rep(nu, each = Tt))
  d$y <- d$a + d$nu + 0.5 * d$period + eta[cbind(d$cluster, d$period)] +
    2 * (d$g != 0L & d$period >= d$g) + rnorm(nrow(d))
  d <- d[order(d$id, d$period), ]
  if (unbalanced) d <- d[sample(nrow(d), floor(0.85 * nrow(d))), ]   # unbalanced + shuffled
  d
}

.fit <- function(d, fm, clus, unb = TRUE) {
  suppressWarnings(suppressMessages(att_gt(
    yname = "y", tname = "period", idname = "id", gname = "g", data = d,
    control_group = "nevertreated", base_period = "universal", bstrap = FALSE, cband = FALSE,
    allow_unbalanced_panel = unb, faster_mode = fm,
    clustervars = if (clus) "cluster" else NULL)))
}

test_that("att_gt analytical clustered SE: faster_mode works (not i.i.d. fallback) on an unbalanced panel", {
  d <- .make_ub_clustered(202L)
  cT <- .fit(d, TRUE,  TRUE)
  cF <- .fit(d, FALSE, TRUE)
  iT <- .fit(d, TRUE,  FALSE)   # i.i.d. (no clustervars)

  # cluster vector now aligns with the influence function under faster_mode (the bug: length mismatch)
  cv <- cT$DIDparams$cluster_vector
  expect_true(!is.null(cv) && length(cv) == nrow(cT$inffunc))

  # analytical clustered SE equals the cluster-sum CRVE closed form (1/n) sqrt(sum_c S_c^2)
  n <- nrow(cT$inffunc)
  S <- rowsum(as.matrix(cT$inffunc), cv)
  se_cf <- sqrt(colSums(S^2)) / n
  ok <- is.finite(cT$se) & is.finite(se_cf)
  expect_equal(unname(cT$se[ok]), unname(se_cf[ok]), tolerance = 1e-8)

  # clustering is actually applied: differs from the i.i.d. SE (was a silent fallback to i.i.d.)
  expect_gt(max(abs(cT$se - iT$se), na.rm = TRUE), 1e-6)

  # faster_mode TRUE == FALSE
  expect_equal(unname(cT$se), unname(cF$se), tolerance = 1e-8)
})

test_that("aggte: faster_mode TRUE == FALSE on an unbalanced clustered panel (all aggregation types)", {
  d <- .make_ub_clustered(202L)
  for (clus in c(TRUE, FALSE)) {
    mT <- .fit(d, TRUE,  clus)
    mF <- .fit(d, FALSE, clus)
    for (ty in c("simple", "group", "dynamic", "calendar")) {
      aT <- suppressWarnings(suppressMessages(aggte(mT, type = ty, bstrap = FALSE, cband = FALSE)))
      aF <- suppressWarnings(suppressMessages(aggte(mF, type = ty, bstrap = FALSE, cband = FALSE)))
      lab <- paste(ty, if (clus) "clustered" else "iid")
      expect_equal(aT$overall.att, aF$overall.att, tolerance = 1e-10, label = paste(lab, "overall.att"))
      expect_equal(aT$overall.se,  aF$overall.se,  tolerance = 1e-8,  label = paste(lab, "overall.se"))
      expect_equal(unname(aT$se.egt), unname(aF$se.egt), tolerance = 1e-8, label = paste(lab, "se.egt"))
    }
  }
})

test_that("no regression: balanced panel still has faster_mode TRUE == FALSE (att_gt + aggte, clustered)", {
  d <- .make_ub_clustered(55L, unbalanced = FALSE)
  cT <- .fit(d, TRUE,  TRUE, unb = FALSE)
  cF <- .fit(d, FALSE, TRUE, unb = FALSE)
  expect_equal(unname(cT$se), unname(cF$se), tolerance = 1e-8)
  for (ty in c("simple", "group", "dynamic", "calendar")) {
    aT <- suppressWarnings(suppressMessages(aggte(cT, type = ty, bstrap = FALSE, cband = FALSE)))
    aF <- suppressWarnings(suppressMessages(aggte(cF, type = ty, bstrap = FALSE, cband = FALSE)))
    expect_equal(aT$overall.se, aF$overall.se, tolerance = 1e-8, label = paste(ty, "overall.se"))
    expect_equal(unname(aT$se.egt), unname(aF$se.egt), tolerance = 1e-8, label = paste(ty, "se.egt"))
  }
})

#-----------------------------------------------------------------------------
#
# Tests for tidy() and nobs() S3 methods
#
#-----------------------------------------------------------------------------

skip_if_not_installed("broom")

data(mpdta, package = "did")

set.seed(1234)
mp <- att_gt(
  yname   = "lemp",
  gname   = "first.treat",
  idname  = "countyreal",
  tname   = "year",
  xformla = ~1,
  data    = mpdta,
  bstrap  = FALSE
)

agg_dyn      <- aggte(mp, type = "dynamic",  bstrap = FALSE, cband = FALSE)
agg_group    <- aggte(mp, type = "group",    bstrap = FALSE, cband = FALSE)
agg_calendar <- aggte(mp, type = "calendar", bstrap = FALSE, cband = FALSE)
agg_simple   <- aggte(mp, type = "simple",   bstrap = FALSE, cband = FALSE)

# ---------------------------------------------------------------------------
# tidy.MP
# ---------------------------------------------------------------------------

test_that("tidy.MP returns expected columns", {
  td <- broom::tidy(mp)
  expect_s3_class(td, "data.frame")
  expect_true(all(c("term", "group", "time", "estimate", "std.error",
                    "statistic", "p.value",
                    "conf.low", "conf.high",
                    "point.conf.low", "point.conf.high") %in% names(td)))
})

test_that("tidy.MP statistic equals estimate / std.error", {
  td <- broom::tidy(mp)
  expect_equal(td$statistic, td$estimate / td$std.error)
})

test_that("tidy.MP p.value matches normal approximation", {
  td <- broom::tidy(mp)
  expected <- 2 * (1 - pnorm(abs(td$statistic)))
  expect_equal(td$p.value, expected)
})

test_that("tidy.MP p.value is between 0 and 1", {
  td <- broom::tidy(mp)
  expect_true(all(td$p.value >= 0 & td$p.value <= 1, na.rm = TRUE))
})

# ---------------------------------------------------------------------------
# tidy.AGGTEobj
# ---------------------------------------------------------------------------

test_that("tidy.AGGTEobj (dynamic) returns expected columns", {
  td <- broom::tidy(agg_dyn)
  expect_true(all(c("type", "term", "event.time", "estimate", "std.error",
                    "statistic", "p.value",
                    "conf.low", "conf.high",
                    "point.conf.low", "point.conf.high") %in% names(td)))
})

test_that("tidy.AGGTEobj (group) returns expected columns", {
  td <- broom::tidy(agg_group)
  expect_true(all(c("type", "term", "group", "estimate", "std.error",
                    "statistic", "p.value",
                    "conf.low", "conf.high",
                    "point.conf.low", "point.conf.high") %in% names(td)))
})

test_that("tidy.AGGTEobj (calendar) returns expected columns", {
  td <- broom::tidy(agg_calendar)
  expect_true(all(c("type", "term", "estimate", "std.error",
                    "statistic", "p.value",
                    "conf.low", "conf.high",
                    "point.conf.low", "point.conf.high") %in% names(td)))
})

test_that("tidy.AGGTEobj (simple) returns expected columns", {
  td <- broom::tidy(agg_simple)
  expect_true(all(c("type", "term", "estimate", "std.error",
                    "statistic", "p.value",
                    "conf.low", "conf.high",
                    "point.conf.low", "point.conf.high") %in% names(td)))
})

test_that("tidy.AGGTEobj statistic and p.value are consistent", {
  for (agg in list(agg_dyn, agg_group, agg_calendar, agg_simple)) {
    td <- broom::tidy(agg)
    expect_equal(td$statistic, td$estimate / td$std.error)
    expect_equal(td$p.value, 2 * (1 - pnorm(abs(td$statistic))))
  }
})

# ---------------------------------------------------------------------------
# nobs
# ---------------------------------------------------------------------------

test_that("nobs.MP returns number of unique units", {
  n <- stats::nobs(mp)
  expect_type(n, "integer")
  expect_equal(n, length(unique(mpdta$countyreal)))
})

test_that("nobs.AGGTEobj returns number of unique units", {
  for (agg in list(agg_dyn, agg_group, agg_calendar, agg_simple)) {
    n <- stats::nobs(agg)
    expect_equal(n, length(unique(mpdta$countyreal)))
  }
})

test_that("nobs.MP and nobs.AGGTEobj agree", {
  expect_equal(stats::nobs(mp), stats::nobs(agg_dyn))
})

# Regression tests for the 2.5.1 audit fixes:
#   A. fix_weights = "varying" on a balanced panel reported SEs exactly 2x too
#      large (force_rc influence-function fold was not divided by 2).
#   B. aggte() crashed with cryptic errors on empty / all-NA aggregation
#      selections (calendar na.rm, simple/dynamic empty post-treatment windows).
#   C. negative gname codes were silently accepted on the fast path but errored
#      on the slow path (now rejected consistently up front).

# small balanced panel with a never-treated group; constant (uniform) weights
.bal_panel <- function(seed = 1, n = 200, sigma = 0.5) {
  set.seed(seed)
  g  <- sample(c(0L, 0L, 2L, 3L), n, replace = TRUE)
  fe <- stats::rnorm(n)
  d  <- data.table::CJ(id = 1:n, t = 1:3)
  d[, g := g[id]]; d[, fe := fe[id]]
  d[, y := fe + 0.2 * t + 1 * (g != 0 & t >= g) + stats::rnorm(.N, 0, sigma)]
  as.data.frame(d[, .(id, t, g, y)])
}
.qgt <- function(df, ...) suppressWarnings(suppressMessages(
  att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = df,
         control_group = "nevertreated", bstrap = FALSE, ...)))

# ---- A. fix_weights = "varying" SE normalization -------------------------------

test_that("fix_weights='varying' SE is not 2x too large on a balanced panel", {
  df <- .bal_panel(1)
  r0 <- .qgt(df, fix_weights = NULL)
  rv <- .qgt(df, fix_weights = "varying")
  ord <- match(paste(r0$group, r0$t), paste(rv$group, rv$t))
  # point estimates were always correct
  expect_equal(r0$att, rv$att[ord], tolerance = 1e-8)
  # with constant weights, the varying (RC) SE must equal the panel SE. Before
  # the fix it was EXACTLY 2x larger (the force_rc fold lacked the 1/2). This is
  # the property the Monte Carlo confirmed (varying SE == empirical sampling SD).
  expect_equal(r0$se, rv$se[ord], tolerance = 1e-8)
  # explicit guard against regression to the 2x bug
  expect_true(all(abs(rv$se[ord] / r0$se - 1) < 1e-6))
})

test_that("fix_weights='varying' fast and slow paths agree (balanced panel)", {
  df <- .bal_panel(2)
  rf <- .qgt(df, fix_weights = "varying", faster_mode = TRUE)
  rs <- .qgt(df, fix_weights = "varying", faster_mode = FALSE)
  expect_equal(rf$att, rs$att, tolerance = 1e-9)
  expect_equal(rf$se, rs$se, tolerance = 1e-9)
})

test_that("fix_weights='varying' aggregations inherit the corrected SE", {
  df <- .bal_panel(3)
  cs0 <- .qgt(df, fix_weights = NULL)
  csv <- .qgt(df, fix_weights = "varying")
  a0 <- suppressWarnings(suppressMessages(aggte(cs0, type = "simple", na.rm = TRUE)))
  av <- suppressWarnings(suppressMessages(aggte(csv, type = "simple", na.rm = TRUE)))
  expect_equal(a0$overall.att, av$overall.att, tolerance = 1e-8)
  expect_equal(a0$overall.se,  av$overall.se,  tolerance = 1e-8)
})

# ---- B. aggte() empty / all-NA selection guards --------------------------------

test_that("aggte() normal aggregations still work (no regression)", {
  data(mpdta, package = "did")
  cs <- suppressWarnings(suppressMessages(att_gt("lemp", "year", "countyreal",
        "first.treat", data = mpdta, bstrap = FALSE)))
  for (ty in c("simple", "dynamic", "group", "calendar")) {
    a <- suppressWarnings(suppressMessages(aggte(cs, type = ty, na.rm = TRUE)))
    expect_true(is.finite(a$overall.att))
    expect_true(is.finite(a$overall.se))
  }
})

test_that("aggte(type='calendar', na.rm=TRUE) drops all-NA calendar periods instead of crashing", {
  data(mpdta, package = "did")
  cs <- suppressWarnings(suppressMessages(att_gt("lemp", "year", "countyreal",
        "first.treat", data = mpdta, bstrap = FALSE)))
  # NA out the lone post-treatment cell of calendar year 2005 (att_gt itself
  # legitimately produces NA cells on unbalanced data; this mimics that).
  w <- which(cs$t == 2005 & cs$group <= cs$t)
  cs$att[w] <- NA; cs$inffunc[, w] <- NA
  expect_no_error(
    res <- suppressWarnings(suppressMessages(aggte(cs, type = "calendar", na.rm = TRUE)))
  )
  expect_false(2005 %in% res$egt)        # the all-NA period was dropped
  expect_true(is.finite(res$overall.att))
})

test_that("aggte() empty post-treatment windows give clean errors, not cryptic crashes", {
  data(mpdta, package = "did")
  cs <- suppressWarnings(suppressMessages(att_gt("lemp", "year", "countyreal",
        "first.treat", data = mpdta, bstrap = FALSE)))
  # simple / dynamic with no e >= 0 -> clean "no valid estimates", NOT
  # "non-numeric argument..." or "...report this as a bug."
  expect_error(suppressWarnings(suppressMessages(aggte(cs, type = "simple", max_e = -1))),
               "No valid att_gt")
  expect_error(suppressWarnings(suppressMessages(aggte(cs, type = "dynamic", max_e = -1))),
               "No valid att_gt")
  # event-time window that excludes every period -> clean "no event times".
  expect_error(suppressWarnings(suppressMessages(aggte(cs, type = "dynamic", min_e = 100))),
               "No event times")
})

# ---- C. negative gname rejected consistently -----------------------------------

test_that("negative gname is rejected with a clear error in both code paths", {
  set.seed(5); n <- 120
  g  <- sample(c(0L, -1L), n, replace = TRUE)
  d  <- data.table::CJ(id = 1:n, t = c(-3L, -2L, -1L, 0L))
  d[, g := g[id]]
  d[, y := stats::rnorm(.N) + (g != 0 & t >= g)]
  df <- as.data.frame(d)
  for (fm in c(TRUE, FALSE)) {
    expect_error(
      suppressWarnings(suppressMessages(att_gt("y", "t", "id", "g", data = df,
        bstrap = FALSE, faster_mode = fm))),
      "negative values are not supported"
    )
  }
})

# ---- D. parallel multiplier bootstrap: reproducibility + chunking guard ---------

test_that("parallel multiplier bootstrap is reproducible under a fixed seed", {
  skip_on_os("windows")          # parallel path is force-disabled on Windows
  inf <- matrix(stats::rnorm(2600 * 4), 2600, 4)   # n > 2500 triggers the parallel branch
  pre_kind <- RNGkind()                            # whatever the runner's RNG kind is
  set.seed(7); a <- did:::run_multiplier_bootstrap(inf, 300, pl = TRUE, cores = 2)
  set.seed(7); b <- did:::run_multiplier_bootstrap(inf, 300, pl = TRUE, cores = 2)
  expect_identical(a, b)                            # same seed -> bit-identical draws
  expect_identical(RNGkind(), pre_kind)             # caller's RNG kind restored (whatever it was)
})

test_that("parallel bootstrap chunking never produces negative chunks (biters < cores)", {
  # the fix: chunks are non-negative and sum to biters for any biters/cores
  # (the old rep(ceiling(biters/cores), cores) + correction went negative when
  # biters < cores, e.g. biters=2,cores=4 -> [-1,1,1,1] -> crash).
  for (biters in c(1, 2, 3, 7, 1000)) {
    for (cores in c(1, 2, 4, 8)) {
      chunks <- diff(round(seq(0, biters, length.out = cores + 1)))
      chunks <- chunks[chunks > 0]
      expect_true(all(chunks > 0))
      expect_equal(sum(chunks), biters)
    }
  }
  # and the parallel branch itself no longer errors with biters < cores
  skip_on_os("windows")
  inf <- matrix(stats::rnorm(2600 * 3), 2600, 3)
  expect_no_error(did:::run_multiplier_bootstrap(inf, biters = 1, pl = TRUE, cores = 2))
})

# ---- E. calendar aggregation warns that min_e/max_e/balance_e are ignored ------

test_that("aggte(type='calendar') warns that min_e/max_e/balance_e are ignored, returns unrestricted", {
  data(mpdta, package = "did")
  cs <- suppressWarnings(suppressMessages(att_gt("lemp", "year", "countyreal",
        "first.treat", data = mpdta, bstrap = FALSE)))
  ws <- testthat::capture_warnings(suppressMessages(aggte(cs, type = "calendar", max_e = 2, na.rm = TRUE)))
  expect_true(any(grepl("ignored for type", ws)))
  a_win <- suppressWarnings(suppressMessages(aggte(cs, type = "calendar", max_e = 2, na.rm = TRUE)))
  a_unr <- suppressWarnings(suppressMessages(aggte(cs, type = "calendar", na.rm = TRUE)))
  expect_equal(a_win$overall.att, a_unr$overall.att)   # max_e is a no-op for calendar (correct)
  # no spurious warning at the defaults
  ws0 <- testthat::capture_warnings(suppressMessages(aggte(cs, type = "calendar", na.rm = TRUE)))
  expect_false(any(grepl("ignored for type", ws0)))
})

# aggte()'s clustervars argument can only honor clustering that att_gt() actually set up (att_gt retains
# the cluster column and builds the per-unit cluster_vector only for the clustervars passed to it). When
# aggte() is asked to cluster on a variable that att_gt() did not use, the cluster information is
# unavailable, so aggte() must NOT silently return the i.i.d. SE (analytic path) or error in mboot()
# (bootstrap path) -- it warns and falls back to non-clustered standard errors. Inheriting clustering from
# the att_gt object, or overriding to the same variable att_gt used, work normally and emit no warning.
.mk_two_cluster <- function(seed) {
  set.seed(seed)
  G <- 30L; sz <- rep(c(2L, 4L, 6L), length.out = G); N <- sum(sz); cl <- rep(seq_len(G), sz)
  region <- rep(sample(1:5, G, TRUE), times = sz)            # a second time-invariant grouping
  alpha <- rnorm(G, 0, 1)[cl]; nu <- rnorm(N); Tt <- 5L
  eta <- matrix(rnorm(G * Tt, 0, 1.5), G, Tt)                # shared within-cluster shock
  g <- sample(c(0L, 2L, 3L, 4L), N, TRUE, c(.4, .2, .2, .2))
  d <- data.frame(id = rep(seq_len(N), each = Tt), period = rep(seq_len(Tt), N),
                  cluster = rep(cl, each = Tt), region = rep(region, each = Tt),
                  g = rep(g, each = Tt), a = rep(alpha, each = Tt), nu = rep(nu, each = Tt))
  d$y <- d$a + d$nu + 0.4 * d$period + eta[cbind(d$cluster, d$period)] +
    (d$g != 0L & d$period >= d$g) + rnorm(nrow(d))
  d[order(d$id, d$period), ]
}

# did aggte() emit the "clustering unavailable" warning?
.warned <- function(expr) {
  flag <- FALSE
  withCallingHandlers(force(expr), warning = function(c) {
    if (grepl("cluster information needed is not available", conditionMessage(c))) flag <<- TRUE
    invokeRestart("muffleWarning")
  })
  flag
}
.agg <- function(m, ty, ...) suppressMessages(aggte(m, type = ty, bstrap = FALSE, cband = FALSE, ...))

test_that("inherited clustering and same-variable override are honored without warning", {
  d <- .mk_two_cluster(11)
  mc <- suppressWarnings(suppressMessages(att_gt(yname = "y", tname = "period", idname = "id", gname = "g",
        data = d, clustervars = "cluster", control_group = "nevertreated", bstrap = FALSE)))
  for (ty in c("simple", "dynamic")) {
    w_inh <- .warned(ai <- .agg(mc, ty))
    w_exp <- .warned(ae <- .agg(mc, ty, clustervars = "cluster"))
    expect_false(w_inh)
    expect_false(w_exp)
    expect_equal(ai$overall.se, ae$overall.se, tolerance = 1e-10)
  }
})

test_that("aggte warns and falls back to i.i.d. when att_gt was not clustered (analytic)", {
  d <- .mk_two_cluster(11)
  mi <- suppressWarnings(suppressMessages(att_gt(yname = "y", tname = "period", idname = "id", gname = "g",
        data = d, control_group = "nevertreated", bstrap = FALSE)))
  for (ty in c("simple", "dynamic", "group", "calendar")) {
    a_iid  <- .agg(mi, ty)                                    # no clustervars -> i.i.d.
    w <- .warned(a_over <- .agg(mi, ty, clustervars = "cluster"))
    expect_true(w)                                            # warning fired
    expect_equal(a_over$overall.se, a_iid$overall.se, tolerance = 1e-10)   # degraded to i.i.d.
  }
})

test_that("aggte bootstrap override warns and falls back instead of erroring", {
  d <- .mk_two_cluster(11)
  mi <- suppressWarnings(suppressMessages(att_gt(yname = "y", tname = "period", idname = "id", gname = "g",
        data = d, control_group = "nevertreated", bstrap = TRUE, biters = 200)))
  ab <- NULL
  w <- .warned(ab <- suppressMessages(aggte(mi, type = "dynamic", bstrap = TRUE, biters = 200,
                                            cband = FALSE, clustervars = "cluster")))
  expect_true(w)
  expect_false(is.null(ab))
  expect_true(all(is.finite(ab$se.egt)))                      # no "undefined columns selected" crash
})

test_that("aggte refuses to switch to a different cluster variable than att_gt used", {
  d <- .mk_two_cluster(11)
  mc <- suppressWarnings(suppressMessages(att_gt(yname = "y", tname = "period", idname = "id", gname = "g",
        data = d, clustervars = "cluster", control_group = "nevertreated", bstrap = FALSE)))
  a_inh <- .agg(mc, "dynamic")                                # correctly clustered on 'cluster'
  w <- .warned(a_sw <- .agg(mc, "dynamic", clustervars = "region"))
  expect_true(w)                                              # cannot switch to 'region'
  expect_false(isTRUE(all.equal(a_sw$overall.se, a_inh$overall.se)))   # degraded -> differs from clustered
})

# Replication tests for the Callaway & Sant'Anna JEL article
# Data source: https://github.com/pedrohcgs/JEL-DiD
# These tests verify that `did` package results match the published article.
# Tests are skipped if the JEL data file is not available.


# Helper: resolve JEL data path, downloading from GitHub if needed (only called after skip_on_cran)
get_jel_data_path <- function() {
  path <- Sys.getenv("JEL_DID_DATA_PATH",
                     unset = file.path(path.expand("~"), "JEL-DiD", "data", "county_mortality_data.csv"))
  if (file.exists(path)) return(path)

  path <- file.path(tempdir(), "county_mortality_data.csv")
  if (file.exists(path) && file.info(path)$size > 0) return(path)

  download_ok <- FALSE
  tryCatch({
    res <- download.file(
      "https://raw.githubusercontent.com/pedrohcgs/JEL-DiD/main/data/county_mortality_data.csv",
      destfile = path,
      quiet = TRUE
    )
    download_ok <- identical(res, 0L)
  }, error = function(e) {
    # download failed; skip_if_not will handle missing file
  })

  # Remove empty/corrupt files so file.exists() correctly triggers skip
  if (!download_ok || !file.exists(path) || file.info(path)$size <= 0) {
    if (file.exists(path)) unlink(path)
  }
  path
}

# Helper to load and clean JEL data
load_jel_data <- function(path, filter_2xt = TRUE) {
  mydata <- read.csv(path, stringsAsFactors = FALSE)
  mydata$state <- substr(mydata$county, nchar(mydata$county) - 1, nchar(mydata$county))

  # Drop DC and pre-2014 adoption states
  mydata <- mydata[!(mydata$state %in% c("DC", "DE", "MA", "NY", "VT")), ]

  if (filter_2xt) {
    # For 2x2 and 2xT: keep only 2014-adopters and never-adopters
    mydata <- mydata[mydata$yaca == 2014 | is.na(mydata$yaca) | mydata$yaca > 2019, ]
  }

  # Compute covariates
  mydata$perc_white <- mydata$population_20_64_white / mydata$population_20_64 * 100
  mydata$perc_hispanic <- mydata$population_20_64_hispanic / mydata$population_20_64 * 100
  mydata$perc_female <- mydata$population_20_64_female / mydata$population_20_64 * 100
  mydata$unemp_rate <- mydata$unemp_rate * 100
  mydata$median_income <- mydata$median_income / 1000

  # Keep only needed columns
  keep_cols <- c("county_code", "year", "population_20_64", "yaca", "crude_rate_20_64",
                 "perc_female", "perc_white", "perc_hispanic", "unemp_rate", "poverty_rate", "median_income")
  mydata <- mydata[, keep_cols]

  # Drop rows with NA in non-yaca columns
  non_yaca <- setdiff(keep_cols, "yaca")
  mydata <- mydata[complete.cases(mydata[, non_yaca]), ]

  # Keep counties with data in both 2013 and 2014
  county_year_counts <- tapply(mydata$year %in% c(2013, 2014), mydata$county_code, sum)
  valid_counties <- names(county_year_counts[county_year_counts == 2])
  mydata <- mydata[mydata$county_code %in% valid_counties, ]

  # Keep counties with all 11 years of data
  county_counts <- table(mydata$county_code)
  valid_counties <- names(county_counts[county_counts == 11])
  mydata <- mydata[mydata$county_code %in% valid_counties, ]

  mydata
}


# ===========================================================================
# Table 7: 2x2 CS-DiD with Covariates (Sant'Anna-Zhao)
# ===========================================================================
test_that("JEL Table 7: 2x2 CS-DiD point estimates match", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = TRUE)

  # Make 2x2 dataset (2013-2014 only)
  short_data <- mydata[mydata$year %in% c(2013, 2014), ]
  short_data$treat_year <- ifelse(!is.na(short_data$yaca) & short_data$yaca == 2014, 2014, 0)
  short_data$county_code <- as.numeric(short_data$county_code)

  # Population weights from 2013
  wt_2013 <- short_data[short_data$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  short_data <- merge(short_data, wt_2013, by = "county_code")

  covs_formula <- ~perc_female + perc_white + perc_hispanic + unemp_rate + poverty_rate + median_income

  # Expected point estimates from JEL article (Table 7)
  expected <- list(
    reg_unweighted = -1.6154372119,
    ipw_unweighted = -0.8585625501,
    dr_unweighted  = -1.2256473242,
    reg_weighted   = -3.4592200594,
    ipw_weighted   = -3.8416966846,
    dr_weighted    = -3.7561045985
  )

  for (method in c("reg", "ipw", "dr")) {
    for (wt_info in list(list(name = NULL, label = "unweighted"),
                         list(name = "set_wt", label = "weighted"))) {
      res <- suppressWarnings(att_gt(
        yname = "crude_rate_20_64", tname = "year", idname = "county_code",
        gname = "treat_year", xformla = covs_formula,
        data = short_data, panel = TRUE, control_group = "nevertreated",
        bstrap = FALSE, est_method = method, weightsname = wt_info$name,
        base_period = "universal"
      ))
      agg <- suppressWarnings(aggte(res, na.rm = TRUE, bstrap = FALSE))

      key <- paste0(method, "_", wt_info$label)
      expect_equal(agg$overall.att, expected[[key]], tolerance = 1e-6,
                   label = paste("Table 7:", key))
    }
  }
})


# ===========================================================================
# 2xT Event Study: No covariates, weighted, reg, nevertreated
# ===========================================================================
test_that("JEL 2xT: event study ATT(g,t) point estimates match", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = TRUE)
  mydata$treat_year <- ifelse(!is.na(mydata$yaca) & mydata$yaca == 2014, 2014, 0)
  mydata$county_code <- as.numeric(mydata$county_code)

  wt_2013 <- mydata[mydata$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  mydata <- merge(mydata, wt_2013, by = "county_code")

  mod <- att_gt(
    yname = "crude_rate_20_64", tname = "year", idname = "county_code",
    gname = "treat_year", xformla = NULL, data = mydata, panel = TRUE,
    control_group = "nevertreated", bstrap = FALSE,
    est_method = "reg", weightsname = "set_wt", base_period = "universal"
  )

  # Expected ATT(g,t) for g=2014
  expected_att <- c(
    4.1292044043,   # t=2009
    -0.5016807242,  # t=2010
    2.7531791360,   # t=2011
    2.7804626426,   # t=2012
    0.0,            # t=2013 (base period)
    -2.5628745138,  # t=2014
    -1.6973291127,  # t=2015
    0.2189009815,   # t=2016
    -0.8133358354,  # t=2017
    -1.1532954495,  # t=2018
    1.7866564429    # t=2019
  )

  expect_equal(mod$att, expected_att, tolerance = 1e-6)

  # Dynamic aggregation
  es <- suppressWarnings(aggte(mod, type = "dynamic", bstrap = FALSE))
  expect_equal(es$att.egt, expected_att, tolerance = 1e-6)

  # Overall ATT for e in {0,5}
  agg <- suppressWarnings(aggte(mod, type = "dynamic", min_e = 0, max_e = 5, bstrap = FALSE))
  expect_equal(agg$overall.att, -0.7035462478, tolerance = 1e-6)
})


# ===========================================================================
# 2xT with covariates: 3 estimation methods
# ===========================================================================
test_that("JEL 2xT: event study with covariates matches across methods", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = TRUE)
  mydata$treat_year <- ifelse(!is.na(mydata$yaca) & mydata$yaca == 2014, 2014, 0)
  mydata$county_code <- as.numeric(mydata$county_code)

  wt_2013 <- mydata[mydata$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  mydata <- merge(mydata, wt_2013, by = "county_code")

  covs_formula <- ~perc_female + perc_white + perc_hispanic + unemp_rate + poverty_rate + median_income

  for (method in c("reg", "ipw", "dr")) {
    res <- att_gt(
      yname = "crude_rate_20_64", tname = "year", idname = "county_code",
      gname = "treat_year", xformla = covs_formula,
      data = mydata, panel = TRUE, control_group = "nevertreated",
      bstrap = FALSE, est_method = method, weightsname = "set_wt",
      base_period = "universal"
    )

    es <- suppressWarnings(aggte(res, type = "dynamic", na.rm = TRUE, bstrap = FALSE))

    # Base period (e = -1) should be exactly 0
    base_idx <- which(es$egt == -1)
    expect_equal(es$att.egt[base_idx], 0, label = paste("2xT covs", method, "e=-1"))

    # All estimates should be finite
    expect_true(all(is.finite(es$att.egt[!is.na(es$att.egt)])),
                label = paste("2xT covs", method, "finite"))
  }
})


# ===========================================================================
# GxT: No covariates, weighted, notyettreated
# ===========================================================================
test_that("JEL GxT: staggered event study without covariates matches", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = FALSE)  # GxT keeps all groups
  mydata$treat_year <- ifelse(!is.na(mydata$yaca) & mydata$yaca <= 2019, mydata$yaca, 0)
  mydata$county_code <- as.numeric(mydata$county_code)

  wt_2013 <- mydata[mydata$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  mydata <- merge(mydata, wt_2013, by = "county_code")

  mod <- att_gt(
    yname = "crude_rate_20_64", tname = "year", idname = "county_code",
    gname = "treat_year", xformla = NULL, data = mydata, panel = TRUE,
    control_group = "notyettreated", bstrap = FALSE,
    weightsname = "set_wt", base_period = "universal"
  )

  # Dynamic aggregation
  es <- suppressWarnings(aggte(mod, type = "dynamic", bstrap = FALSE))

  # Expected dynamic ATT at key event times
  expected_dynamic <- c(
    `e=-5`  =  2.2186783578,
    `e=-4`  =  0.8579225109,
    `e=-3`  =  1.9161499399,
    `e=-2`  =  2.5644742860,
    `e=-1`  =  0.0,
    `e=0`   = -1.6545648988,
    `e=1`   = -0.2616435324,
    `e=2`   =  1.7055625922,
    `e=3`   = -0.5405232028,
    `e=4`   = -0.5148819184,
    `e=5`   =  1.7866564429
  )

  # Match e = -5 to e = 5
  for (e_val in -5:5) {
    idx <- which(es$egt == e_val)
    key <- paste0("e=", e_val)
    expect_equal(es$att.egt[idx], expected_dynamic[[key]], tolerance = 1e-6,
                 label = paste("GxT no covs", key))
  }

  # Overall ATT for e in {0,5}
  agg <- suppressWarnings(aggte(mod, type = "dynamic", min_e = 0, max_e = 5, bstrap = FALSE))
  expect_equal(agg$overall.att, 0.0867675805, tolerance = 1e-6,
               label = "GxT no covs overall ATT (e=0:5)")
})


# ===========================================================================
# GxT: With covariates, DR, weighted, notyettreated
# ===========================================================================
test_that("JEL GxT: staggered event study with DR covariates matches", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = FALSE)
  mydata$treat_year <- ifelse(!is.na(mydata$yaca) & mydata$yaca <= 2019, mydata$yaca, 0)
  mydata$county_code <- as.numeric(mydata$county_code)

  wt_2013 <- mydata[mydata$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  mydata <- merge(mydata, wt_2013, by = "county_code")

  covs_formula <- ~perc_female + perc_white + perc_hispanic + unemp_rate + poverty_rate + median_income

  mod <- suppressWarnings(att_gt(
    yname = "crude_rate_20_64", tname = "year", idname = "county_code",
    gname = "treat_year", xformla = covs_formula,
    data = mydata, panel = TRUE, control_group = "notyettreated",
    bstrap = FALSE, est_method = "dr", weightsname = "set_wt",
    base_period = "universal"
  ))

  # Overall ATT for e in {0,5}
  agg <- suppressWarnings(aggte(mod, type = "dynamic", min_e = 0, max_e = 5, bstrap = FALSE))
  expect_equal(agg$overall.att, -2.2469982988, tolerance = 1e-6,
               label = "GxT DR covs overall ATT (e=0:5)")

  # Dynamic aggregation at key event times
  es <- suppressWarnings(aggte(mod, type = "dynamic", bstrap = FALSE))

  expected_dynamic <- c(
    `e=-5`  =  2.6684811691,
    `e=-4`  =  2.1333836537,
    `e=-3`  =  2.8574389179,
    `e=-2`  =  2.9129673584,
    `e=-1`  =  0.0,
    `e=0`   = -1.4929553032,
    `e=1`   = -1.9693922262,
    `e=2`   = -2.7250659699,
    `e=3`   = -5.0556625460,
    `e=4`   = -4.7161630370,
    `e=5`   =  2.4772492894
  )

  for (e_val in -5:5) {
    idx <- which(es$egt == e_val)
    key <- paste0("e=", e_val)
    expect_equal(es$att.egt[idx], expected_dynamic[[key]], tolerance = 1e-6,
                 label = paste("GxT DR covs", key))
  }
})


# ===========================================================================
# Consistency: faster_mode should match regular mode for all JEL analyses
# ===========================================================================
test_that("JEL: faster_mode matches regular mode", {
  skip_on_cran()
  jel_data_path <- get_jel_data_path()
  skip_if_not(file.exists(jel_data_path), "JEL-DiD data not available")

  mydata <- load_jel_data(jel_data_path, filter_2xt = TRUE)

  # 2x2 data
  short_data <- mydata[mydata$year %in% c(2013, 2014), ]
  short_data$treat_year <- ifelse(!is.na(short_data$yaca) & short_data$yaca == 2014, 2014, 0)
  short_data$county_code <- as.numeric(short_data$county_code)

  wt_2013 <- short_data[short_data$year == 2013, c("county_code", "population_20_64")]
  names(wt_2013)[2] <- "set_wt"
  short_data <- merge(short_data, wt_2013, by = "county_code")

  covs_formula <- ~perc_female + perc_white + perc_hispanic + unemp_rate + poverty_rate + median_income

  # Test DR weighted
  res_slow <- suppressWarnings(att_gt(
    yname = "crude_rate_20_64", tname = "year", idname = "county_code",
    gname = "treat_year", xformla = covs_formula,
    data = short_data, panel = TRUE, control_group = "nevertreated",
    bstrap = FALSE, est_method = "dr", weightsname = "set_wt",
    base_period = "universal", faster_mode = FALSE
  ))
  res_fast <- suppressWarnings(att_gt(
    yname = "crude_rate_20_64", tname = "year", idname = "county_code",
    gname = "treat_year", xformla = covs_formula,
    data = short_data, panel = TRUE, control_group = "nevertreated",
    bstrap = FALSE, est_method = "dr", weightsname = "set_wt",
    base_period = "universal", faster_mode = TRUE
  ))

  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10,
               label = "2x2 DR weighted: ATTs match")

  agg_slow <- suppressWarnings(aggte(res_slow, na.rm = TRUE, bstrap = FALSE))
  agg_fast <- suppressWarnings(aggte(res_fast, na.rm = TRUE, bstrap = FALSE))
  expect_equal(agg_slow$overall.att, agg_fast$overall.att, tolerance = 1e-10,
               label = "2x2 DR weighted: aggregate ATT matches")
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ---- echo=FALSE, results="hide", warning=FALSE, message=FALSE----------------
# source("setup_sims.R")
# fldr <- "~/Dropbox/did/R/"
#fldr <- "/Users/santanph/Dropbox/Co-authored Projects/did/R/"
# sapply(paste0(fldr,list.files(fldr)), source)
library(DRDID)
library(BMisc)
# library(ggplot2)
# library(ggpubr)


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, ipw model is incorectly specified here
#-----------------------------------------------------------------------------
set.seed(09142024)
time.periods <- 4
did::reset.sim()
data <- did::build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res

#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_ipw_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test if 2 period case works (possible to introduce bugs that affect this
# case only)
# Expected results: warning about no pre-treatment periods to test
#-----------------------------------------------------------------------------
time.periods <- 2
did::reset.sim()
data <- did::build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res

summary(aggte(res, type="simple"))
summary(aggte(res, type="group"))
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="calendar"))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test no covariates case
# Expected Result: te=1, p-value for pre-test uniformly distributed,
#  identical results for different estimation methods
#-----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
bett <- betu <- rep(0,time.periods)
data <- did::build_sim_dataset()

res <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

res <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test repeated cross sections, regression sims
# Expected result: te=1, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_sim_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg", panel=FALSE)
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", panel=FALSE)
res

#-----------------------------------------------------------------------------
# test repeated cross sections, ipw sims
# Expected result: te=1, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_ipw_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg", panel=FALSE)
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", panel=FALSE)
res

#-----------------------------------------------------------------------------
# test repeated cross sections, test aggregations
# Expected result: te=length of exposure, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
did::reset.sim()
te.e <- 1:time.periods
data <- did::build_sim_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

summary(aggte(res))
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# these are same test cases as for panel data
# but estimate using allow_unbalanced_panel = TRUE
# but setting an id which gives a way to incorporate
# unbalanced panel
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test uniform[0,1]
#-----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
data <- did::build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", allow_unbalanced_panel=TRUE)
res


#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_ipw_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", allow_unbalanced_panel=TRUE)
res

#-----------------------------------------------------------------------------
# try it with an actual unbalanced panel
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_ipw_dataset()
data <- data[sample(1:nrow(data),  size=floor(.9*nrow(data))),]

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=TRUE, allow_unbalanced_panel=TRUE)
res

#-----------------------------------------------------------------------------
# version that should error
# have to have an idname if you use an unbalanced panel
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_sim_dataset()
data <- data[sample(1:nrow(data),  size=floor(.9*nrow(data))),]

tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname=NULL,
                       gname="G", est_method="reg", panel=TRUE, allow_unbalanced_panel=TRUE),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         })



## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test not yet treated as control
# Expected result: te=1, p-value for pre-test U[0,1]
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_ipw_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated",
              gname="G", est_method="dr", panel=FALSE)
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test not yet treated as control in case w/o never treated group
# Expected result: te=1, p-value for pre-test U[0,1]
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_sim_dataset()
data <- subset(data, G > 0) # drop nevertreated

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated",
              gname="G", est_method="dr", panel=FALSE)
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test nevertreated as control in case w/o never treated group
# Expected result: te=1, p-value for pre-test U[0,1], error on no never treated
#  units
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_sim_dataset()
data <- subset(data, G > 0) # drop nevertreated

# dr
tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         }
         )


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
did::reset.sim()
te <- 0
te.e <- 1:time.periods
data <- did::build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic"))


#-----------------------------------------------------------------------------
# test group treatment timing
# Expected result: te constant within group / varies across groups
#-----------------------------------------------------------------------------
did::reset.sim()
te <- 0
te.bet.ind <- 1:time.periods
data <- did::build_ipw_dataset(panel=FALSE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="ipw", panel=FALSE)
res
summary(aggte(res, type="group"))


#-----------------------------------------------------------------------------
# test calendar time effects
# expected result: te=time
#-----------------------------------------------------------------------------
did::reset.sim()
te <- 0
te.t <- thet + 1:time.periods
data <- did::build_sim_dataset(panel=FALSE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="calendar"))

#-----------------------------------------------------------------------------
# test balancing with respect to length of exposure
# expected result: balancing fixes treatment effect dynamics
#-----------------------------------------------------------------------------
did::reset.sim()
te <- 0
te.e <- 1:time.periods
te.bet.ind <- 1:time.periods
data <- did::build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="dynamic", balance_e=1))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test that att_gt and aggte work with unequally spaced periods
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
time.periods <- 8
did::reset.sim()
te <- 0
te.e <- 1:time.periods
data <- did::build_sim_dataset()
keep.periods <- c(1,2,5,7)
data <- subset(data, G %in% c(0, keep.periods))
data <- subset(data, period %in% keep.periods)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test that att_gt and aggte work with unequally spaced groups
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
time.periods <- 5
did::reset.sim()
te <- 0
te.e <- 1:time.periods
data <- did::build_sim_dataset()
keep.groups <- c(3,5)
data <- subset(data, G %in% c(0, keep.groups))

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic", balance_e=0))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test that att_gt works when some units are treated in first period
# expected result: te=length of exposure, code runs with warning message about
#  dropped units
#-----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
te <- 1
data <- did::build_sim_dataset()
data <- subset(data, period >= 2)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
did::reset.sim()
te <- 0
te.e <- 1:time.periods
data <- did::build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="dynamic", min_e=-1, max_e=1))


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure - 1 (w/ one period -1 anticipation)
#-----------------------------------------------------------------------------
time.periods <- 5
did::reset.sim()
te <- 0
te.e <- -1:(time.periods-2)
data <- did::build_sim_dataset()
data$G <- data$G + 1 # add anticipation

#-----------------------------------------------------------------------------
# will get results wrong due to anticipation effect for g=6
# which shows up in the comparison group
# this will only affect g=5 with 1 period anticipation
#-----------------------------------------------------------------------------
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=1
              )
res
summary(aggte(res, type="dynamic"))

#-----------------------------------------------------------------------------
# drop last time period and results should be correct
#-----------------------------------------------------------------------------
data <- subset(data, period < time.periods)
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=1
              )
res
summary(aggte(res, type="dynamic"))

#-----------------------------------------------------------------------------
# incorrectly ignore anticipation, dynamic effects are incorrect due to ignoring
# anticipation
#-----------------------------------------------------------------------------
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=0
              )
res
summary(aggte(res, type="dynamic"))


## -----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
data <- did::build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", alp=0.01)
res


## -----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
data <- did::build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", alp=0.01, cband=FALSE)
res


## -----------------------------------------------------------------------------
## some groups later than last treated period
## plus missing groups
time.periods <- 7
did::reset.sim()
data <- did::build_sim_dataset()
data <- subset(data, period <= 4)
missingG_ids <- sample(unique(data$id), size=10)
data[data$id %in% missingG_ids,"G"] <- NA

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", cband=FALSE)
res


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# incorrectly specified id
#-----------------------------------------------------------------------------
time.periods <- 4
did::reset.sim()
data <- did::build_sim_dataset()

# dr
tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="brant",
                       gname="G", est_method="dr"),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         }
         )


#-----------------------------------------------------------------------------
# incorrectly specified time period
#-----------------------------------------------------------------------------


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# custom estimation method
# Expected results: te=1, pre-test p-value uniformly distributed, code runs
#-----------------------------------------------------------------------------
did::reset.sim()
data <- did::build_sim_dataset(panel=TRUE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method=DRDID::drdid_imp_panel, panel=TRUE)
res

#' Difference in Differences
#'
#' `did` implements Difference in Differences with multiple periods and variation in
#' treatment timing.
#'
#' @keywords internal
"_PACKAGE"

#' @importFrom stats pnorm qnorm pchisq quantile cov aggregate setNames
#' @importFrom stats model.frame model.matrix na.pass complete.cases binomial
#' @importFrom stats rnorm var ecdf glm predict
#' @importFrom utils globalVariables
#' @import ggplot2
#' @importFrom BMisc toformula rhs_vars makeBalancedPanel get_list_element
#' @importFrom BMisc multiplier_bootstrap TorF
#' @import data.table
#' @import fastglm
#' @importFrom tidyr gather
#' @importFrom methods is as
#' @importFrom dreamerr check_set_arg
#' @importFrom DRDID drdid_panel reg_did_panel std_ipw_did_panel std_ipw_did_rc reg_did_rc drdid_rc
NULL
utils::globalVariables(c(
    ".", ".G", ".w", "asif_never_treated", "treated_first_period", "count", "constant", ".rowid",
    "V1", "control_group", "cohort", "cohort_size", "period", "period_size", "y1", "y0",
    "D", "N", "post", "weights",
    "i.weights", "y", "cluster", "id", "..cols_to_keep", "..g", "..keep_cols",
    "inf_func_long", "inf_func_agg"
))

#' @title DIDparams
#'
#' @description Object to hold did parameters that are passed across functions
#'
#' @inheritParams att_gt
#' @inheritParams pre_process_did
#' @param n The number of observations.  This is equal to the
#'  number of units (which may be different from the number
#'  of rows in a panel dataset).
#' @param nG The number of groups
#' @param nT The number of time periods
#' @param tlist a vector containing each time period
#' @param glist a vector containing each group
#' @param true_repeated_cross_sections Whether or not the data really
#'  is repeated cross sections.  (We include this because unbalanced
#'  panel code runs through the repeated cross sections code)
#'
#' @export
DIDparams <- function(yname,
                   tname,
                   idname=NULL,
                   gname,
                   xformla=NULL,
                   data,
                   control_group,
                   anticipation=0,
                   weightsname=NULL,
                   fix_weights=NULL,
                   alp=0.05,
                   bstrap=TRUE,
                   biters=1000,
                   clustervars=NULL,
                   cband=TRUE,
                   print_details=TRUE,
                   faster_mode=FALSE,
                   pl=FALSE,
                   cores=1,
                   est_method="dr",
                   base_period="varying",
                   panel=TRUE,
                   true_repeated_cross_sections,
                   n=NULL,
                   nG=NULL,
                   nT=NULL,
                   tlist=NULL,
                   glist=NULL,
                   call=NULL) {

  out <- list(yname=yname,
              tname=tname,
              idname=idname,
              gname=gname,
              xformla=xformla,
              data=data,
              control_group=control_group,
              anticipation=anticipation,
              weightsname=weightsname,
              fix_weights=fix_weights,
              alp=alp,
              bstrap=bstrap,
              biters=biters,
              clustervars=clustervars,
              cband=cband,
              print_details=print_details,
              faster_mode=faster_mode,
              pl=pl,
              cores=cores,
              est_method=est_method,
              base_period=base_period,
              panel=panel,
              true_repeated_cross_sections=true_repeated_cross_sections,
              n=n,
              nG=nG,
              nT=nT,
              tlist=tlist,
              glist=glist,
              call=call)
  class(out) <- "DIDparams"
  out
}

#' @title trimmer
#'
#' @description A utility function to find observations that appear
#'  to violate support conditions.  This function is not called
#'  anywhere in the code, but it is just useful for debugging some
#'  common issues that users run into.
#'
#' @param g is a particular group (below I pass in 2009)
#' @inheritParams att_gt
#' @param threshold the cutoff for which observations are flagged as
#'  likely violators of the support condition.
#'
#' @return list of ids of observations that likely violate support conditions
#'
#' @export
trimmer <- function(g, tname, idname, gname, xformla, data, control_group="notyettreated", threshold=.999) {

  time.period <- data[,tname]
  this.data <- data[time.period == (g-1),]
  if (control_group == "notyettreated") {
    # not yet treated
    this.data <- this.data[(this.data[,gname] >= g) |
                           (this.data[,gname] == 0), ]
  } else {
    # never treated
    this.data <- this.data[(this.data[,gname] == g) |
                           (this.data[,gname] == 0), ]
  }
  this.data$D <- 1*this.data[,gname]==g
  this.pscore_reg <- glm(BMisc::toformula("D", BMisc::rhs_vars(xformla)),
                         data=this.data,
                         family=binomial(link="logit"))
  this.pscore <- predict(this.pscore_reg, type="response")
  dropper <- (this.pscore >  threshold) & (this.data$D==1)
  if (sum(dropper) > 0) {
    print("hard to match treated observations: ")
    print(this.data[dropper,idname])
    return(this.data[dropper,idname])
  }
}

#' @title overlap_logit_fit
#' @description Fit the per-(g,t) overlap-check propensity logit via fastglm's
#'  low-level entry point (`fastglmPure`), skipping the `fastglm()` wrapper's
#'  per-call input coercion and family/deviance bookkeeping. Passing fastglm's own
#'  defaults (`method = 0`, `tol = 1e-8`, `maxit = 100`) makes the fitted values --
#'  and hence the overlap decision (`max(fitted) >= 0.999`) -- bit-identical to the
#'  previous `fastglm::fastglm(x, y, family = binomial())` call, while ~1.4x faster.
#' @param x covariate matrix
#' @param y treatment indicator
#' @return a fastglmPure fit list (with `$fitted.values`)
#' @noRd
overlap_logit_fit <- function(x, y) {
  fastglm::fastglmPure(x = x, y = y, family = stats::binomial(),
                       method = 0L, tol = 1e-8, maxit = 100L)
}

#' @title overlap_check_fail
#' @description Evaluate the per-(g,t) overlap guard, i.e. whether the
#'  preliminary propensity logit gives `max(fitted) >= 0.999`. For an
#'  intercept-only design (`xformla = ~1`, the default) the unweighted logit MLE
#'  fits every unit at `mean(y)`, so the guard reduces to `mean(y) >= 0.999`
#'  with no fit at all -- except within `1e-6` of the 0.999 knife edge, where
#'  the IRLS iterate (tol = 1e-8; observed error up to ~1.5e-9) can land on
#'  either side of the cutoff. There we fall back to the real fit, keeping the
#'  decision bit-identical to always fitting.
#' @param x covariate matrix
#' @param y treatment indicator
#' @param intercept_only whether `x` is a single column of ones
#' @return TRUE if the overlap condition is violated
#' @noRd
overlap_check_fail <- function(x, y, intercept_only = FALSE) {
  if (intercept_only) {
    pbar <- mean(y)
    if (abs(pbar - 0.999) > 1e-6) return(pbar >= 0.999)
    # knife edge: defer to the actual fit below
  }
  max(overlap_logit_fit(x, y)$fitted.values) >= 0.999
}

#' @title rcond_check_fail
#' @description Evaluate the per-(g,t) regression-feasibility guard, i.e.
#'  whether `rcond(crossprod(x[controls, ]))` is below machine epsilon. For an
#'  intercept-only design the control-unit Gram matrix is the scalar count of
#'  control units (n0): `rcond` is exactly 1 when n0 > 0 and exactly 0 when
#'  n0 == 0, so the test reduces to "no control units" with no Gram matrix.
#' @param x covariate matrix
#' @param y treatment indicator (controls are `y == 0`)
#' @param intercept_only whether `x` is a single column of ones
#' @return TRUE if the control-unit design is singular or ill-conditioned
#' @noRd
rcond_check_fail <- function(x, y, intercept_only = FALSE) {
  if (intercept_only) {
    return(!any(y == 0))
  }
  control_covs <- x[y == 0, , drop = FALSE]
  rcond(crossprod(control_covs)) < .Machine$double.eps
}

#' @title Check User-Supplied Names Against Internal Names
#' @description Stop if user arguments reference names that `did` creates and
#' mutates internally.
#' @noRd
check_reserved_did_names <- function(yname, tname, idname, gname, xformla,
                                     weightsname = NULL, clustervars = NULL) {
  reserved_names <- c(".w", ".rowid", ".G", ".C", "post",
                      "asif_never_treated", "treated_first_period")
  xvars <- if (is.null(xformla)) character(0) else all.vars(xformla)
  used_names <- unique(c(yname, tname, idname, gname, weightsname,
                         clustervars, xvars))
  used_names <- used_names[!is.na(used_names) & nzchar(used_names)]
  bad_names <- intersect(used_names, reserved_names)
  if (length(bad_names) > 0L) {
    stop("The following variable name(s) are reserved for internal use by `did`: ",
         paste(bad_names, collapse = ", "), ". Please rename these column(s) ",
         "before calling att_gt().")
  }
}

#' @title get_wide_data
#' @description A utility function to convert long data to wide data, i.e., takes a 2 period dataset and turns it into a cross sectional dataset.
#'
#' @param data data.table used in function
#' @param yname name of outcome variable that can change over time
#' @param idname unique id
#' @param tname time period name
#'
#' @return data from first period with .y0 (outcome in first period), .y1 (outcome in second period),
#' and .dy (change in outcomes over time) appended to it
#' @noRd
get_wide_data <- function(data, yname, idname, tname) {
  # check if data is data.table
  if (!is.data.table(data)) {
    stop("data must be a data.table")
  }

  # check that only 2 periods of data are available
  if (length(unique(data[[tname]])) != 2) {
    stop("data must have exactly 2 periods")
  }

  # making computations — use set() to avoid copy triggered by $<-
  y_vals <- data[[yname]]
  set(data, j = ".y1", value = data.table::shift(y_vals, -1))
  set(data, j = ".y0", value = y_vals)
  set(data, j = ".dy", value = data[[".y1"]] - y_vals)

  # Subset to first period's rows
  # Pre-extract to avoid data.table .checkTypos when column name matches variable name
  .time_vals <- data[[tname]]
  first.period <- min(.time_vals)
  data <- data[.time_vals == first.period]

  return(data)
}

#' @title Check balanced panel data
#' @description A utility function to check if your dataset is a balanced panel dataset.
#'
#' Precondition: at most one row per (id_col, time_col) combination -- enforced
#' upstream by validate_args(). Under that uniqueness, every unit has at most one
#' row per period, so the panel is balanced iff nrow == n_ids * n_periods.
#'
#' @param data data.table used in function
#' @param id_col name of id column in the dataset
#' @param time_col name of time column in the dataset
#'
#' @return Boolean indicating if the dataset is balanced or not.
#' @noRd
check_balance <- function(data, id_col, time_col) {

  # Check if every unit is observed in every time period (no grouping needed
  # given the (id, time) uniqueness precondition above)
  is_balanced <- nrow(data) == data.table::uniqueN(data[[id_col]]) * data.table::uniqueN(data[[time_col]])

  return(is_balanced)
}

# =============================================================================
# Coverage for output/summary methods and exported utilities that were
# previously unexercised: summary.AGGTEobj / print.AGGTEobj (all aggregation
# types + control-group / est-method / cband branches), summary.MP.TEST,
# trimmer(), and the get_wide_data() input guards. These are behaviour tests
# of existing code paths (no new behaviour).
# =============================================================================

test_that("summary.AGGTEobj / print.AGGTEobj cover every aggregation type", {
  set.seed(20260609)
  sp <- did::reset.sim(time.periods = 4)
  data <- did::build_sim_dataset(sp)
  mp <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
           gname = "G", bstrap = TRUE, biters = 100, cband = TRUE,
           control_group = "nevertreated")))
  for (ty in c("simple", "group", "dynamic", "calendar")) {
    agg <- suppressWarnings(aggte(mp, type = ty))
    expect_s3_class(agg, "AGGTEobj")
    expect_output(print(agg))      # print.AGGTEobj delegates to summary.AGGTEobj
    expect_output(summary(agg))    # exercises overall block + per-type egt table
  }
})

test_that("summary.AGGTEobj covers pointwise (non-cband) and not-yet-treated / est_method branches", {
  set.seed(20260609)
  sp <- did::reset.sim(time.periods = 4)
  data <- did::build_sim_dataset(sp)
  # bstrap = FALSE -> the "Pointwise" cband_text1b branch; notyettreated control;
  # cycle est_method to hit the dr / ipw / reg text branches.
  for (em in c("dr", "ipw", "reg")) {
    mp <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, est_method = em,
             control_group = "notyettreated")))
    agg <- suppressWarnings(aggte(mp, type = "dynamic"))
    expect_output(summary(agg))
  }
})

test_that("summary.MP.TEST prints the conditional pre-test results", {
  skip_on_cran()
  sp <- did::reset.sim(time.periods = 3, n = 200)
  data <- did::build_sim_dataset(sp)
  cdp <- suppressWarnings(suppressMessages(conditional_did_pretest(
    yname = "Y", tname = "period", idname = "id", gname = "G",
    xformla = ~X, data = data)))
  expect_s3_class(cdp, "MP.TEST")
  expect_output(summary(cdp), "Cramer von Mises")
})

test_that("trimmer() runs for both control-group definitions", {
  set.seed(20260609)
  sp <- did::reset.sim(time.periods = 3)
  data <- did::build_sim_dataset(sp)
  g <- 2  # an early treatment group present in the sim
  # trimmer uses print() for any flagged units; swallow it. It returns the
  # flagged ids or NULL when none violate the support condition.
  out_ny <- capture.output(res_ny <- trimmer(g = g, tname = "period", idname = "id",
    gname = "G", xformla = ~X, data = data, control_group = "notyettreated"))
  out_nt <- capture.output(res_nt <- trimmer(g = g, tname = "period", idname = "id",
    gname = "G", xformla = ~X, data = data, control_group = "nevertreated"))
  expect_true(is.null(res_ny) || is.numeric(res_ny) || is.data.frame(res_ny) || is.atomic(res_ny))
  expect_true(is.null(res_nt) || is.numeric(res_nt) || is.data.frame(res_nt) || is.atomic(res_nt))
})

test_that("get_wide_data() guards reject non-data.table and non-2-period input", {
  dt2 <- data.table::data.table(id = c(1, 1, 2, 2), period = c(1, 2, 1, 2), Y = rnorm(4))
  dt3 <- data.table::data.table(id = rep(1:2, each = 3), period = rep(1:3, 2), Y = rnorm(6))
  expect_error(get_wide_data(as.data.frame(dt2), "Y", "id", "period"),
               "data must be a data.table")
  expect_error(get_wide_data(dt3, "Y", "id", "period"),
               "data must have exactly 2 periods")
})

# Overlap/rcond guard short-circuit and per-(g, covariate-period) cache:
# the intercept-only closed forms and the cached guard booleans must reproduce
# the per-cell preliminary-fit decisions, results, and warnings bit-identically.

test_that("intercept-only overlap short-circuit matches the real fit's decision, including the 0.999 knife edge", {
  decisions_match <- function(D) {
    X <- as.matrix(rep(1, length(D)))
    full <- suppressWarnings(max(did:::overlap_logit_fit(X, D)$fitted.values) >= 0.999)
    shortcut <- suppressWarnings(did:::overlap_check_fail(X, D, intercept_only = TRUE))
    identical(full, shortcut)
  }
  # interior, just-below-cutoff, the exact 0.999 knife edge (where the IRLS
  # iterate lands just BELOW the cutoff and the cell must keep passing),
  # just-above-cutoff, and the degenerate no-control / no-treated designs
  expect_true(decisions_match(rep(c(1, 0), c(500, 500))))   # mean = 0.5
  expect_true(decisions_match(rep(c(1, 0), c(998, 2))))     # mean = 0.998
  expect_true(decisions_match(rep(c(1, 0), c(999, 1))))     # mean = 0.999 exactly
  expect_false(did:::overlap_check_fail(as.matrix(rep(1, 1000)),
                                        rep(c(1, 0), c(999, 1)),
                                        intercept_only = TRUE))
  expect_true(decisions_match(rep(c(1, 0), c(9991, 9))))    # mean = 0.9991
  expect_true(decisions_match(rep(c(1, 0), c(1500, 1))))    # mean ~ 0.99933
  expect_true(decisions_match(rep(1, 100)))                 # mean = 1 (no controls)
  expect_true(decisions_match(rep(0, 100)))                 # mean = 0 (no treated)
})

test_that("intercept-only rcond short-circuit matches the Gram-matrix test", {
  X <- as.matrix(rep(1, 50))
  ref <- function(D) {
    cc <- X[D == 0, , drop = FALSE]
    rcond(crossprod(cc)) < .Machine$double.eps
  }
  D_some_controls <- rep(c(1, 0), c(40, 10))
  D_no_controls <- rep(1, 50)
  expect_identical(did:::rcond_check_fail(X, D_some_controls, intercept_only = TRUE),
                   ref(D_some_controls))
  expect_identical(did:::rcond_check_fail(X, D_no_controls, intercept_only = TRUE),
                   ref(D_no_controls))
  expect_false(did:::rcond_check_fail(X, D_some_controls, intercept_only = TRUE))
  expect_true(did:::rcond_check_fail(X, D_no_controls, intercept_only = TRUE))
})

test_that("guard cache returns bit-identical results to per-cell checks (fast and slow paths)", {
  set.seed(20260610)
  sp <- did::reset.sim(time.periods = 5, n = 500)
  d <- did::build_sim_dataset(sp)
  run_one <- function(faster, disable_cache) {
    old <- options(did.disable_check_cache = disable_cache)
    on.exit(options(old), add = TRUE)
    suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", est_method = "dr",
      control_group = "nevertreated", faster_mode = faster, bstrap = FALSE)))
  }
  for (faster in c(TRUE, FALSE)) {
    cached <- run_one(faster, disable_cache = FALSE)
    uncached <- run_one(faster, disable_cache = TRUE)
    expect_identical(cached$att, uncached$att)
    expect_identical(cached$se, uncached$se)
    expect_identical(as.matrix(cached$inffunc), as.matrix(uncached$inffunc))
  }
})

test_that("cached guard failures still warn once per cell with unchanged text", {
  set.seed(20260610)
  d <- expand.grid(id = 1:200, period = 1:4)
  gvals <- c(0, 2, 3)
  d$G <- gvals[(d$id - 1) %% 3 + 1]
  d$Xsep <- 1 * (d$G > 0)  # separates treated from never-treated: overlap fails in all 6 cells
  d$Y <- 0.1 * d$period + (d$G > 0 & d$period >= d$G) + rnorm(nrow(d), 0, 0.5)
  run_one <- function(faster, disable_cache) {
    old <- options(did.disable_check_cache = disable_cache)
    on.exit(options(old), add = TRUE)
    capture_warnings(suppressMessages(att_gt(yname = "Y", xformla = ~Xsep,
      data = d, tname = "period", idname = "id", gname = "G", est_method = "dr",
      faster_mode = faster, bstrap = FALSE)))
  }
  # cache on/off must produce identical warning vectors (count, text, order)
  w_fast <- run_one(TRUE, disable_cache = FALSE)
  expect_identical(w_fast, run_one(TRUE, disable_cache = TRUE))
  w_slow <- run_one(FALSE, disable_cache = FALSE)
  expect_identical(w_slow, run_one(FALSE, disable_cache = TRUE))
  expect_identical(w_fast, w_slow)
  # one overlap warning per failed cell, even when the boolean came from the cache
  expect_identical(sum(grepl("overlap condition violated for group", w_fast)), 6L)
})

#' @title AGGTEobj
#'
#' @description Objects of this class hold results on aggregated
#'  group-time average treatment effects
#'
#' @title Aggregate Treatment Effect Parameters Object
#'
#' @description An object for holding aggregated treatment effect parameters.
#'
#' @inheritParams aggte
#' @inheritParams compute.aggte
#' @param overall.att The estimated overall ATT
#' @param overall.se Standard error for overall ATT
#' @param egt Holds the length of exposure (for dynamic effects), the
#'  group (for selective treatment timing), or the time period (for calendar
#'  time effects)
#' @param att.egt The ATT specific to egt
#' @param se.egt The standard error specific to egt
#' @param crit.val.egt A critical value for computing uniform confidence
#'  bands for dynamic effects, selective treatment timing, or time period
#'  effects.
#' @param inf.function The influence function of the chosen aggregated parameters
#' @param DIDparams A DIDparams object
#'
#' @return an AGGTEobj
#' @export
AGGTEobj <- function(overall.att = NULL,
                     overall.se = NULL,
                     type = "simple",
                     egt = NULL,
                     att.egt = NULL,
                     se.egt = NULL,
                     crit.val.egt = NULL,
                     inf.function = NULL,
                     min_e = NULL,
                     max_e = NULL,
                     balance_e = NULL,
                     call=NULL,
                     DIDparams=NULL) {

  out <- list(overall.att = overall.att,
              overall.se = overall.se,
              type = type,
              egt = egt,
              att.egt = att.egt,
              se.egt = se.egt,
              crit.val.egt = crit.val.egt,
              inf.function = inf.function,
              min_e = min_e,
              max_e = max_e,
              balance_e = balance_e,
              call = call,
              DIDparams = DIDparams)

  class(out)  <- "AGGTEobj"
  out
}

#' @title Summary Aggregate Treatment Effect Parameter Objects
#'
#' @description A function to summarize aggregated treatment effect parameters.
#'
#' @param object an \code{AGGTEobj} object
#' @param ... other arguments
#'
#' @export
summary.AGGTEobj <- function(object, ...) {

  # call
  cat("\n")
  cat("Call:\n")
  print(object$call)
  cat("\n")

  #citation
  citation()
  cat("\n")

  # overall estimates
  alp <- object$DIDparams$alp
  pointwise_cval <- qnorm(1-alp/2)
  overall_cband_upper <- object$overall.att + pointwise_cval*object$overall.se
  overall_cband_lower <- object$overall.att - pointwise_cval*object$overall.se
  out1 <- cbind.data.frame(object$overall.att, object$overall.se, overall_cband_lower, overall_cband_upper)
  out1 <- round(out1, 4)
  overall_sig <- (overall_cband_upper < 0) | (overall_cband_lower > 0)
  overall_sig[is.na(overall_sig)] <- FALSE
  overall_sig_text <- ifelse(overall_sig, "*", "")
  out1 <- cbind.data.frame(out1, overall_sig_text)
  cat("\n")
  #cat("Overall ATT:  \n")
  if (object$type=="dynamic") cat("Overall summary of ATT\'s based on event-study/dynamic aggregation:  \n")
  if (object$type=="group") cat("Overall summary of ATT\'s based on group/cohort aggregation:  \n")
  if (object$type=="calendar") cat("Overall summary of ATT\'s based on calendar time aggregation:  \n")
  colnames(out1) <- c("ATT","   Std. Error", paste0("    [ ",100*(1-object$DIDparams$alp),"% "), "Conf. Int.]","")
  print(out1, row.names=FALSE)
  cat("\n\n")

  # handle cases depending on type
  if (object$type %in% c("group","dynamic","calendar")) {

    # header
    if (object$type=="dynamic") { c1name <- "Event time"; cat("Dynamic Effects:") }
    if (object$type=="group") { c1name <- "Group"; cat("Group Effects:") }
    if (object$type=="calendar") { c1name <- "Time"; cat("Time Effects:") }

    cat("\n")
    cband_text1a <- paste0(100*(1-object$DIDparams$alp),"% ")
    cband_text1b <- if (object$DIDparams$bstrap) {
      if (object$DIDparams$cband) "Simult. " else "Pointwise "
    } else {
      "Pointwise "
    }
    cband_text1 <- paste0("[", cband_text1a, cband_text1b)

    cband_lower <- object$att.egt - object$crit.val.egt*object$se.egt
    cband_upper <- object$att.egt + object$crit.val.egt*object$se.egt

    sig <- (cband_upper < 0) | (cband_lower > 0)
    sig[is.na(sig)] <- FALSE
    sig_text <- ifelse(sig, "*", "")

    out2 <- cbind.data.frame(object$egt, object$att.egt, object$se.egt, cband_lower, cband_upper)
    out2 <- round(out2, 4)
    out2 <- cbind.data.frame(out2, sig_text)

    colnames(out2) <- c(c1name, "Estimate","Std. Error", cband_text1, "Conf. Band]", "")
    print(out2, row.names=FALSE, justify = "centre")
  }
  cat("---\n")
  cat("Signif. codes: `*' confidence band does not cover 0")
  cat("\n\n")

  # set control group text
  control_group <- object$DIDparams$control_group
  control_group_text <- NULL
  if (control_group == "nevertreated") {
    control_group_text <- "Never Treated"
  } else if (control_group == "notyettreated") {
    control_group_text <- "Not Yet Treated"
  }

  if (!is.null(control_group)) {
    cat("Control Group:  ")
    cat(control_group_text)
    cat(",  ")
  }

  # anticipation periods
  cat("Anticipation Periods:  ")
  cat(object$DIDparams$anticipation)
  cat("\n")

  # estimation method text
  est_method <- object$DIDparams$est_method
  if ( is(est_method,"character") ) {
    est_method_text <- est_method
    if (est_method == "dr") {
      est_method_text <- "Doubly Robust"
    } else if (est_method == "ipw") {
      est_method_text <- "Inverse Probability Weighting"
    } else if (est_method == "reg") {
      est_method_text <- "Outcome Regression"
    }

    cat("Estimation Method:  ")
    cat(est_method_text)
    cat("\n")
  }
}

#' @title print.AGGTEobj
#'
#' @description prints value of a \code{AGGTEobj} object
#'
#' @param x a \code{AGGTEobj} object
#' @param ... extra arguments
#'
#' @export
print.AGGTEobj <- function(x,...) {
  summary.AGGTEobj(x,...)
}

#' @title Compute Aggregated Treatment Effect Parameters
#'
#' @description Does the heavy lifting on computing aggregated group-time
#'  average treatment effects
#'
#' @inheritParams att_gt
#' @inheritParams aggte
#' @param call The function call to aggte
#'
#' @return [`AGGTEobj`] object
#'
#' @keywords internal
#'
#' @export
compute.aggte <- function(MP,
                          type = "group",
                          balance_e = NULL,
                          min_e = -Inf,
                          max_e = Inf,
                          na.rm = FALSE,
                          bstrap = NULL,
                          biters = NULL,
                          cband = NULL,
                          alp = NULL,
                          clustervars = NULL,
                          call = NULL) {
  #-----------------------------------------------------------------------------
  # unpack MP object
  #-----------------------------------------------------------------------------
  # load parameters
  group <- MP$group
  t <- MP$t
  att <- MP$att
  dp <- MP$DIDparams
  inffunc1 <- MP$inffunc
  n <- MP$n

  # aggte() needs the influence functions to aggregate and to compute standard errors.
  # They are absent when att_gt() was run with compute_inffunc = FALSE (point estimates only).
  if (is.null(inffunc1)) {
    stop("This att_gt() result was produced with compute_inffunc = FALSE (point estimates ",
         "only), so it has no influence functions and cannot be aggregated by aggte(). ",
         "Re-run att_gt() with compute_inffunc = TRUE (the default) to use aggte().")
  }


  gname <- dp$gname
  tname <- dp$tname
  idname <- dp$idname
  panel <- dp$panel
  if (is.null(clustervars)) {
    clustervars <- dp$clustervars
  }

  # Aggregation-level clustering reuses the per-unit cluster information that att_gt() stored when
  # clustervars was supplied to it. If clustering is requested here (inherited from the att_gt object or
  # via an aggte-level clustervars override) but that information is unavailable -- because the cluster
  # variable was not passed to att_gt() (so neither the cluster column nor cluster_vector were retained),
  # or because an override names a different variable than att_gt clustered on -- it cannot be honored.
  # Warn and fall back to non-clustered standard errors, instead of silently returning the i.i.d. SE
  # (analytic path) or erroring in mboot() (bootstrap path). To make this a hard error instead, replace
  # warning() with stop() below.
  extra_cl_req <- clustervars[!(clustervars %in% c(idname, ""))]
  if (length(extra_cl_req) > 0) {
    cv <- dp$cluster_vector
    can_cluster <- !is.null(cv) && length(cv) == nrow(inffunc1) &&
      identical(extra_cl_req, dp$cluster_vector_var)
    if (!can_cluster) {
      warning(paste0(
        "Clustered standard errors were requested in aggte() (clustervars = '",
        paste(extra_cl_req, collapse = "', '"), "'), but the cluster information needed is not available ",
        "in this att_gt object",
        if (!is.null(dp$cluster_vector_var)) {
          paste0(" (att_gt() clustered on '", paste(dp$cluster_vector_var, collapse = "', '"),
                 "', and aggte() cannot switch to a different cluster variable)")
        } else {
          " (clustervars was not supplied to att_gt(), so aggte() cannot introduce clustering)"
        },
        ". Reporting standard errors that do NOT account for clustering; re-run att_gt() with ",
        "clustervars = '", paste(extra_cl_req, collapse = "', '"), "' for clustered inference."))
      clustervars <- NULL
    }
  }
  if (is.null(bstrap)) {
    bstrap <- dp$bstrap
  }
  if (is.null(biters)) {
    biters <- dp$biters
  }
  if (is.null(alp)) {
    alp <- dp$alp
  }
  if (is.null(cband)) {
    cband <- dp$cband
  }
  if (isTRUE(dp$faster_mode)) {
    tlist <- dp$time_periods
    glist <- dp$treated_groups
  } else {
    data <- as.data.frame(dp$data)
    tlist <- dp$tlist
    glist <- dp$glist
  }

  # In faster_mode, the full long data is only needed on the two branches below that
  # actually consume `data` (most faster-mode panel calls use time_invariant_data
  # instead). Materialize it lazily there, recoding gname Inf -> 0 (the old
  # convention) on the converted copy so the data.table stored inside the user's
  # MP object is never modified by reference.
  faster_mode_long_data <- function() {
    data <- as.data.frame(dp$data)
    data[data[, gname] == Inf, gname] <- 0
    data
  }

  # overwrite MP objects (so we can actually compute bootstrap)
  MP$DIDparams$clustervars <- clustervars
  MP$DIDparams$bstrap <- bstrap
  MP$DIDparams$biters <- biters
  MP$DIDparams$alp <- alp
  MP$DIDparams$cband <- cband
  dp <- MP$DIDparams

  if (!(type %in% c("simple", "dynamic", "group", "calendar"))) {
    stop('`type` must be one of c("simple", "dynamic", "group", "calendar")')
  }

  if (na.rm) {
    notna <- !is.na(att)
    if (!any(notna)) {
      stop("All att_gt() estimates are NA. Cannot compute aggregated treatment effects. ",
           "Check your att_gt() results for estimation problems (e.g., too few observations, ",
           "singular matrices, or overlap violations).")
    }
    group <- group[notna]
    t <- t[notna]
    att <- att[notna]
    inffunc1 <- inffunc1[, notna, drop = FALSE]
    # tlist <- sort(unique(t))
    glist <- sort(unique(group))

    # If aggte is of the group type, ensure we have non-missing post-treatment ATTs for each group
    if (type == "group") {
      # Get the groups that have some non-missing ATT(g,t) in post-treatmemt periods
      gnotna <- sapply(glist, function(g) {
        # look at post-treatment periods for group g, restricted to the SAME
        # max_e window used by the group-specific estimate below (selective.att.g /
        # selective.se.inner). Without the (t <= group + max_e) condition a group
        # whose only non-NA ATT(g,t) lies PAST max_e would pass this filter but then
        # have an all-NA (hence, under na.rm, empty) selection in the estimate,
        # erroring in get_agg_inf_func(). (max_e defaults to Inf, so this is a no-op
        # unless the user sets a finite max_e.)
        whichg <- which((group == g) & (g <= t) & (t <= (group + max_e)))
        attg <- att[whichg]
        group_select <- !is.na(mean(attg))
        return(group_select)
      })
      gnotna <- glist[gnotna]
      # indicator for not all post-treatment ATT(g,t) missing
      not_all_na <- group %in% gnotna
      if (!any(not_all_na)) {
        stop("No groups have non-missing post-treatment att_gt() estimates. ",
             "Cannot compute group aggregation. Check your att_gt() results.")
      }
      # Re-do the na.rm thing to update the groups
      group <- group[not_all_na]
      t <- t[not_all_na]
      att <- att[not_all_na]
      inffunc1 <- inffunc1[, not_all_na, drop = FALSE]
      # tlist <- sort(unique(t))
      glist <- sort(unique(group))
    }
  }

  if ((na.rm == FALSE) && base::anyNA(att)) stop("Missing values found in att_gt() estimates. To remove them, set `na.rm = TRUE`.")

  # data from first period
  # ifelse(panel,
  #       dta <- data[ data[,tname]==tlist[1], ],
  #       dta <- data
  #       )
  if (panel) {
    if (isTRUE(dp$faster_mode) && !is.null(dp$time_invariant_data)) {
      # use pre-computed time-invariant data (one row per unit) instead of subsetting
      dta <- as.data.frame(dp$time_invariant_data)
      # map Inf back to 0 for gname to match the old convention
      dta[dta[, gname] == Inf, gname] <- 0
    } else {
      if (isTRUE(dp$faster_mode)) data <- faster_mode_long_data()
      # data from first period
      dta <- data[data[, tname] == tlist[1], ]
    }
  } else {
    if (isTRUE(dp$faster_mode)) data <- faster_mode_long_data()
    # Aggregate to one row per unit. aggregate() returns rows sorted by idname, but the influence
    # function (inffunc1) is in the data's *first-appearance* unit order -- which differs from sorted for
    # unbalanced panels under faster_mode. Restore that order so the estimated-weight influence term (wif)
    # is added to the matching units in get_agg_inf_func(); otherwise wif lands on the wrong units and the
    # aggregated standard error is wrong (the point estimate is unaffected because wif is mean-zero).
    dta_agg <- base::suppressWarnings(stats::aggregate(data, list((data[, idname])), mean))
    dta <- dta_agg[match(unique(data[, idname]), dta_agg[, 1L]), -1L, drop = FALSE]
  }

  #-----------------------------------------------------------------------------
  # data organization and recoding
  #-----------------------------------------------------------------------------

  # do some recoding to make sure time periods are 1 unit apart
  # and then put these back together at the end
  originalt <- t
  originalgroup <- group
  originalglist <- glist
  originaltlist <- tlist
  # In case g's are not part of tlist
  originalgtlist <- sort(unique(c(originaltlist, originalglist)))
  uniquet <- seq(1, length(unique(originalgtlist)))
  # lookup vectors for vectorized mapping between original and new t values
  orig_with_zero <- c(originalgtlist, 0)
  new_with_zero <- c(uniquet, 0)
  # vectorized function to switch from "new" t values to original t values
  t2orig <- function(t) {
    orig_with_zero[match(t, new_with_zero)]
  }
  # vectorized function to switch from original t values to new sequential t values
  orig2t_vec <- function(orig) {
    new_with_zero[match(orig, orig_with_zero)]
  }
  t <- orig2t_vec(originalt)
  group <- orig2t_vec(originalgroup)
  glist <- orig2t_vec(originalglist)
  tlist <- unique(t)
  maxT <- max(t)

  # Set the weights
  if (".w" %in% names(dta)) {
    weights.ind <- dta$.w
  } else if (isTRUE(dp$faster_mode)) {
    weights.ind <- dta$weights
  } else {
    weights.ind <- dta$.w
  }

  # we can work in overall probabilities because conditioning will cancel out
  # cause it shows up in numerator and denominator
  gvar <- dta[[gname]]  # extract the group column once, reused below
  pg <- sapply(originalglist, function(g) mean(weights.ind * (gvar == g)))

  # length of this is equal to number of groups
  pgg <- pg

  # same but length is equal to the number of ATT(g,t)
  pg <- pg[match(group, glist)]

  # which group time average treatment effects are post-treatment
  keepers <- which(group <= t & t <= (group + max_e)) ### added second condition to allow for limit on longest period included in att

  # n x 1 vector of group variable
  G <- orig2t_vec(gvar)

  #-----------------------------------------------------------------------------
  # Compute the simple ATT summary
  #-----------------------------------------------------------------------------

  if (type == "simple") {
    # simple att
    # averages all post-treatment ATT(g,t) with weights
    # given by group size
    simple.att <- sum(att[keepers] * pg[keepers]) / (sum(pg[keepers]))
    if (is.nan(simple.att)) simple.att <- NA

    # get the part of the influence function coming from estimated weights
    simple.wif <- wif(keepers, pg, weights.ind, G, group)

    # get the overall influence function
    simple.if <- get_agg_inf_func(
      att = att,
      inffunc1 = inffunc1,
      whichones = keepers,
      weights.agg = pg[keepers] / sum(pg[keepers]),
      wif = simple.wif
    )
    # Make it as vector
    simple.if <- as.numeric(simple.if)

    # get standard errors from overall influence function
    simple.se <- getSE(simple.if, dp)
    if (!is.na(simple.se)) {
      if (simple.se <= sqrt(.Machine$double.eps) * 10) simple.se <- NA
    }


    return(AGGTEobj(
      overall.att = simple.att,
      overall.se = simple.se,
      type = type,
      inf.function = list(simple.att = simple.if),
      call = call,
      DIDparams = dp
    ))
  }

  #-----------------------------------------------------------------------------
  # Compute the group (i.e., selective) treatment timing estimators
  #-----------------------------------------------------------------------------

  if (type == "group") {
    # get group specific ATTs
    # note: there are no estimated weights here
    selective.att.g <- sapply(glist, function(g) {
      # look at post-treatment periods for group g
      whichg <- which((group == g) & (g <= t) & (t <= (group + max_e))) ### added last condition to allow for limit on longest period included in att
      attg <- att[whichg]
      mean(attg)
    })
    selective.att.g[is.nan(selective.att.g)] <- NA


    # get standard errors for each group specific ATT
    selective.se.inner <- lapply(glist, function(g) {
      whichg <- which((group == g) & (g <= t) & (t <= (group + max_e))) ### added last condition to allow for limit on longest period included in att
      inf.func.g <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whichg,
        weights.agg = pg[whichg] / sum(pg[whichg]),
        wif = NULL
      ))
      se.g <- getSE(inf.func.g, dp)
      list(inf.func = inf.func.g, se = se.g)
    })

    # recover standard errors separately by group
    selective.se.g <- unlist(BMisc::get_list_element(selective.se.inner, "se"))
    selective.se.g[selective.se.g <= sqrt(.Machine$double.eps) * 10] <- NA

    # recover influence function separately by group
    selective.inf.func.g <- simplify2array(BMisc::get_list_element(selective.se.inner, "inf.func"))

    # use multiplier bootstrap (across groups) to get critical value
    # for constructing uniform confidence bands
    selective.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      selective.crit.val <- mboot(selective.inf.func.g, dp, return_V = FALSE)$crit.val

      if (is.na(selective.crit.val) | is.infinite(selective.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        selective.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (selective.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        selective.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (selective.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall att under selective treatment timing
    # (here use pgg instead of pg because we can just look at each group)
    selective.att <- sum(selective.att.g * pgg) / sum(pgg)

    # account for having to estimate pgg in the influence function
    selective.wif <- wif(
      keepers = 1:length(glist),
      pg = pgg,
      weights.ind = weights.ind,
      G = G,
      group = glist
    )

    # get overall influence function
    selective.inf.func <- get_agg_inf_func(
      att = selective.att.g,
      inffunc1 = selective.inf.func.g,
      whichones = (1:length(glist)),
      weights.agg = pgg / sum(pgg),
      wif = selective.wif
    )


    selective.inf.func <- as.numeric(selective.inf.func)
    # get overall standard error
    selective.se <- getSE(selective.inf.func, dp)
    if (!is.na(selective.se)) {
      if ((selective.se <= sqrt(.Machine$double.eps) * 10)) selective.se <- NA
    }

    return(AGGTEobj(
      overall.att = selective.att,
      overall.se = selective.se,
      type = type,
      egt = originalglist,
      att.egt = selective.att.g,
      se.egt = selective.se.g,
      crit.val.egt = selective.crit.val,
      inf.function = list(
        selective.inf.func.g = selective.inf.func.g,
        selective.inf.func = selective.inf.func
      ),
      call = call,
      DIDparams = dp
    ))
  }



  #-----------------------------------------------------------------------------
  # Compute the event-study estimators
  #-----------------------------------------------------------------------------

  if (type == "dynamic") {
    # event times
    # this looks at all available event times
    # note: event times can be negative here.
    # note: event time = 0 corresponds to "on impact"
    # eseq <- unique(t-group)
    eseq <- unique(originalt - originalgroup)
    eseq <- eseq[order(eseq)]

    # if the user specifies balance_e, then we are going to
    # drop some event times and some groups; if not, we just
    # keep everything (that is what this variable is for)
    include.balanced.gt <- rep(TRUE, length(originalgroup))

    # if we balance the sample with resepect to event time
    if (!is.null(balance_e)) {
      include.balanced.gt <- (t2orig(maxT) - originalgroup >= balance_e)

      eseq <- unique(originalt[include.balanced.gt] - originalgroup[include.balanced.gt])
      eseq <- eseq[order(eseq)]

      eseq <- eseq[(eseq <= balance_e) & (eseq >= balance_e - t2orig(maxT) + t2orig(1))]
    }

    # only looks at some event times
    eseq <- eseq[(eseq >= min_e) & (eseq <= max_e)]

    # compute atts that are specific to each event time
    dynamic.att.e <- sapply(eseq, function(e) {
      # keep att(g,t) for the right g&t as well as ones that
      # are not trimmed out from balancing the sample
      whiche <- which((originalt - originalgroup == e) & (include.balanced.gt))
      atte <- att[whiche]
      pge <- pg[whiche] / (sum(pg[whiche]))
      sum(atte * pge)
    })

    # compute standard errors for dynamic effects
    dynamic.se.inner <- lapply(eseq, function(e) {
      whiche <- which((originalt - originalgroup == e) & (include.balanced.gt))
      pge <- pg[whiche] / (sum(pg[whiche]))
      wif.e <- wif(whiche, pg, weights.ind, G, group)
      inf.func.e <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whiche,
        weights.agg = pge,
        wif = wif.e
      ))
      se.e <- getSE(inf.func.e, dp)
      list(inf.func = inf.func.e, se = se.e)
    })

    dynamic.se.e <- unlist(BMisc::get_list_element(dynamic.se.inner, "se"))
    dynamic.se.e[dynamic.se.e <= sqrt(.Machine$double.eps) * 10] <- NA

    dynamic.inf.func.e <- simplify2array(BMisc::get_list_element(dynamic.se.inner, "inf.func"))

    dynamic.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      dynamic.crit.val <- mboot(dynamic.inf.func.e, dp, return_V = FALSE)$crit.val

      if (is.na(dynamic.crit.val) | is.infinite(dynamic.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        dynamic.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (dynamic.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        dynamic.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (dynamic.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall average treatment effect
    # by averaging over positive dynamics
    epos <- eseq >= 0
    dynamic.att <- mean(dynamic.att.e[epos])
    dynamic.inf.func <- get_agg_inf_func(
      att = dynamic.att.e[epos],
      inffunc1 = as.matrix(dynamic.inf.func.e[, epos]),
      whichones = (1:sum(epos)),
      weights.agg = (rep(1 / sum(epos), sum(epos))),
      wif = NULL
    )

    dynamic.inf.func <- as.numeric(dynamic.inf.func)
    dynamic.se <- getSE(dynamic.inf.func, dp)
    if (!is.na(dynamic.se)) {
      if (dynamic.se <= sqrt(.Machine$double.eps) * 10) dynamic.se <- NA
    }

    return(AGGTEobj(
      overall.att = dynamic.att,
      overall.se = dynamic.se,
      type = type,
      egt = eseq,
      att.egt = dynamic.att.e,
      se.egt = dynamic.se.e,
      crit.val.egt = dynamic.crit.val,
      inf.function = list(
        dynamic.inf.func.e = dynamic.inf.func.e,
        dynamic.inf.func = dynamic.inf.func
      ),
      call = call,
      min_e = min_e,
      max_e = max_e,
      balance_e = balance_e,
      DIDparams = dp
    ))
  }

  #-----------------------------------------------------------------------------
  # calendar time effects
  #-----------------------------------------------------------------------------

  if (type == "calendar") {
    # drop time periods where no one is treated yet
    # (can't get treatment effects in those periods)
    minG <- min(group)
    calendar.tlist <- tlist[tlist >= minG]

    # calendar time specific atts
    calendar.att.t <- sapply(calendar.tlist, function(t1) {
      # look at post-treatment periods for group g
      whicht <- which((t == t1) & (group <= t))
      attt <- att[whicht]
      pgt <- pg[whicht] / (sum(pg[whicht]))
      sum(pgt * attt)
    })

    # get standard errors and influence functions
    # for each time specific att
    calendar.se.inner <- lapply(calendar.tlist, function(t1) {
      whicht <- which((t == t1) & (group <= t))
      pgt <- pg[whicht] / (sum(pg[whicht]))
      wif.t <- wif(
        keepers = whicht,
        pg = pg,
        weights.ind = weights.ind,
        G = G,
        group = group
      )
      inf.func.t <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whicht,
        weights.agg = pgt,
        wif = wif.t
      ))
      se.t <- getSE(inf.func.t, dp)
      list(inf.func = inf.func.t, se = se.t)
    })

    # recover standard errors separately by time
    calendar.se.t <- unlist(BMisc::get_list_element(calendar.se.inner, "se"))
    calendar.se.t[calendar.se.t <= sqrt(.Machine$double.eps) * 10] <- NA
    # recover influence function separately by time
    calendar.inf.func.t <- simplify2array(BMisc::get_list_element(calendar.se.inner, "inf.func"))

    # use multiplier boostrap (across groups) to get critical value
    # for constructing uniform confidence bands
    calendar.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      calendar.crit.val <- mboot(calendar.inf.func.t, dp, return_V = FALSE)$crit.val

      if (is.na(calendar.crit.val) | is.infinite(calendar.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        calendar.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (calendar.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        calendar.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (calendar.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall att under calendar time effects
    # this is just average over all time periods
    calendar.att <- mean(calendar.att.t)

    # get overall influence function
    calendar.inf.func <- get_agg_inf_func(
      att = calendar.att.t,
      inffunc1 = calendar.inf.func.t,
      whichones = (1:length(calendar.tlist)),
      weights.agg = rep(1 / length(calendar.tlist), length(calendar.tlist)),
      wif = NULL
    )
    calendar.inf.func <- as.numeric(calendar.inf.func)
    # get overall standard error
    calendar.se <- getSE(calendar.inf.func, dp)
    if (!is.na(calendar.se)) {
      if (calendar.se <= sqrt(.Machine$double.eps) * 10) calendar.se <- NA
    }
    return(AGGTEobj(
      overall.att = calendar.att,
      overall.se = calendar.se,
      type = type,
      egt = t2orig(calendar.tlist),
      att.egt = calendar.att.t,
      se.egt = calendar.se.t,
      crit.val.egt = calendar.crit.val,
      inf.function = list(
        calendar.inf.func.t = calendar.inf.func.t,
        calendar.inf.func = calendar.inf.func
      ),
      call = call,
      DIDparams = dp
    ))
  }
}

#-----------------------------------------------------------------------------
# Internal functions for getting standard errors
#-----------------------------------------------------------------------------

#' @title Compute extra term in influence function due to estimating weights
#'
#' @description A function to compute the extra term that shows up in the
#'  influence function for aggregated treatment effect parameters
#'  due to estimating the weights
#'
#' @param keepers a vector of indices for which group-time average
#'  treatment effects are used to compute a particular aggregated parameter
#' @param pg a vector with same length as total number of group-time average
#'  treatment effects that contains the probability of being in particular group
#' @param weights.ind additional sampling weights (nx1)
#' @param G vector containing which group a unit belongs to (nx1)
#' @param group vector of groups
#'
#' @return nxk influence function matrix
#'
#' @keywords internal
wif <- function(keepers, pg, weights.ind, G, group) {
  # note: weights are all of the form P(G=g|cond)/sum_cond(P(G=g|cond))
  # this is equal to P(G=g)/sum_cond(P(G=g)) which simplifies things here

  # The (n x k) matrix [ weights.ind * 1{G == group[k]} - pg[k] ] appears in both
  # the numerator and the denominator terms below; build it once and reuse it.
  # This is identical to the previous code, which constructed it twice via two
  # separate sapply() calls. sum(pg[keepers]) is likewise hoisted out of the loop.
  Spg <- sum(pg[keepers])
  centered <- sapply(keepers, function(k) {
    weights.ind * 1 * BMisc::TorF(G == group[k]) - pg[k]
  })

  # effect of estimating weights in the numerator
  if1 <- centered / Spg
  # effect of estimating weights in the denominator
  if2 <- base::rowSums(centered) %*% t(pg[keepers] / (Spg^2))

  # return the influence function for the weights
  if1 - if2
}


#' @title Get an influence function for particular aggregate parameters
#'
#' @title This is a generic internal function for combining influence
#'  functions across ATT(g,t)'s to return an influence function for
#'  various aggregated treatment effect parameters.
#'
#' @param att vector of group-time average treatment effects
#' @param inffunc1 influence function for all group-time average treatment effects
#'  (matrix)
#' @param whichones which elements of att will be used to compute the aggregated
#'  treatment effect parameter
#' @param weights.agg the weights to apply to each element of att(whichones);
#'  should have the same dimension as att(whichones)
#' @param wif extra influence function term coming from estimating the weights;
#'  should be n x k matrix where k is dimension of whichones
#'
#' @return nx1 influence function
#'
#' @keywords internal
get_agg_inf_func <- function(att, inffunc1, whichones, weights.agg, wif = NULL) {
  # enforce weights are in matrix form
  weights.agg <- as.matrix(weights.agg)

  # check dimensions
  if (length(whichones) == 0) {
    stop("No valid att_gt() estimates found for this aggregation. ",
         "This may happen if all estimates for a particular group or time period are NA.")
  }
  if (max(whichones) > ncol(inffunc1)) {
    stop("Internal error: influence function column indices (", max(whichones),
         ") exceed available columns (", ncol(inffunc1), "). Please report this as a bug.")
  }

  # multiplies influence function times weights and sums to get vector of weighted IF (of length n)
  thisinffunc <- inffunc1[, whichones, drop = FALSE] %*% weights.agg

  # Incorporate influence function of the weights
  if (!is.null(wif)) {
    thisinffunc <- thisinffunc + wif %*% as.matrix(att[whichones])
  }

  # return influence function
  return(thisinffunc)
}


#' @title Take influence function and return standard errors
#'
#' @description Function to take an nx1 influence function and return
#'  a standard error
#'
#' @param thisinffunc An influence function
#' @inheritParams compute.aggte
#'
#' @return scalar standard error
#'
#' @keywords internal
getSE <- function(thisinffunc, DIDparams = NULL) {
  alp <- .05
  bstrap <- FALSE
  n <- length(thisinffunc)
  if (!is.null(DIDparams)) {
    bstrap <- DIDparams$bstrap
    alp <- DIDparams$alp
    cband <- DIDparams$cband
  }

  if (bstrap) {
    bout <- mboot(thisinffunc, DIDparams, return_V = FALSE)
    return(bout$se)
  } else {
    # Analytical cluster-robust SE (no bootstrap): cluster sums of the aggregated influence function,
    # matching the cluster-sum aggregation of the multiplier bootstrap (CS 2021, Remark 10). Falls back to
    # the i.i.d. form when there is no cluster variable beyond idname or cluster identifiers are unavailable.
    # dp$cluster_vector is set by att_gt() for both faster_mode = TRUE and FALSE.
    cv <- if (!is.null(DIDparams)) DIDparams$cluster_vector else NULL
    extra_cl <- if (!is.null(DIDparams)) DIDparams$clustervars[!(DIDparams$clustervars %in% c(DIDparams$idname, ""))] else character(0)
    if (length(extra_cl) > 0 && !is.null(cv) && length(cv) == length(thisinffunc)) {
      S <- rowsum(thisinffunc, cv)
      return(sqrt(sum(S^2)) / n)
    }
    return(sqrt(mean((thisinffunc)^2) / n))
  }
}

#' @title trimmer
#'
#' @description A utility function to find observations that appear
#'  to violate support conditions.  This function is not called
#'  anywhere in the code, but it is just useful for debugging some
#'  common issues that users run into.
#'
#' @param g is a particular group (below I pass in 2009)
#' @inheritParams att_gt
#' @param threshold the cutoff for which observations are flagged as
#'  likely violators of the support condition.
#'
#' @return list of ids of observations that likely violate support conditions
#'
#' @export
trimmer <- function(g, tname, idname, gname, xformla, data, control_group="notyettreated", threshold=.999) {
  if (!all(class(data) == "data.frame")) {
    data <- as.data.frame(data)
  }
  validate_finite_numeric_scalar(g, "g")
  validate_column_name(tname, "tname", names(data))
  validate_column_name(idname, "idname", names(data))
  validate_column_name(gname, "gname", names(data))
  validate_xformla(xformla)
  validate_choice_scalar(
    control_group,
    "control_group",
    c("nevertreated", "notyettreated"),
    "control_group must be either 'nevertreated' or 'notyettreated'."
  )
  validate_probability_scalar(threshold, "threshold")

  time.period <- data[[tname]]
  this.data <- data[time.period == (g-1),]
  if (control_group == "notyettreated") {
    # not yet treated
    this.data <- this.data[(this.data[[gname]] >= g) |
                           (this.data[[gname]] == 0), ]
  } else {
    # never treated
    this.data <- this.data[(this.data[[gname]] == g) |
                           (this.data[[gname]] == 0), ]
  }
  this.data$D <- 1 * this.data[[gname]] == g
  this.pscore_reg <- glm(BMisc::toformula("D", BMisc::rhs_vars(xformla)),
                         data=this.data,
                         family=binomial(link="logit"))
  this.pscore <- predict(this.pscore_reg, type="response")
  dropper <- (this.pscore >  threshold) & (this.data$D==1)
  if (sum(dropper) > 0) {
    print("hard to match treated observations: ")
    print(this.data[dropper, idname, drop = FALSE])
    return(this.data[dropper, idname, drop = FALSE])
  }
}

#' @title overlap_logit_fit
#' @description Fit the per-(g,t) overlap-check propensity logit via fastglm's
#'  low-level entry point (`fastglmPure`), skipping the `fastglm()` wrapper's
#'  per-call input coercion and family/deviance bookkeeping. Passing fastglm's own
#'  defaults (`method = 0`, `tol = 1e-8`, `maxit = 100`) makes the fitted values --
#'  and hence the overlap decision (`max(fitted) >= 0.999`) -- bit-identical to the
#'  previous `fastglm::fastglm(x, y, family = binomial())` call, while ~1.4x faster.
#' @param x covariate matrix
#' @param y treatment indicator
#' @return a fastglmPure fit list (with `$fitted.values`)
#' @noRd
overlap_logit_fit <- function(x, y) {
  fastglm::fastglmPure(x = x, y = y, family = stats::binomial(),
                       method = 0L, tol = 1e-8, maxit = 100L)
}

#' @title overlap_check_fail
#' @description Evaluate the per-(g,t) overlap guard, i.e. whether the
#'  preliminary propensity logit gives `max(fitted) >= 0.999`. For an
#'  intercept-only design (`xformla = ~1`, the default) the unweighted logit MLE
#'  fits every unit at `mean(y)`, so the guard reduces to `mean(y) >= 0.999`
#'  with no fit at all -- except within `1e-6` of the 0.999 knife edge, where
#'  the IRLS iterate (tol = 1e-8; observed error up to ~1.5e-9) can land on
#'  either side of the cutoff. There we fall back to the real fit, keeping the
#'  decision bit-identical to always fitting.
#' @param x covariate matrix
#' @param y treatment indicator
#' @param intercept_only whether `x` is a single column of ones
#' @return TRUE if the overlap condition is violated
#' @noRd
overlap_check_fail <- function(x, y, intercept_only = FALSE) {
  if (intercept_only) {
    pbar <- mean(y)
    if (abs(pbar - 0.999) > 1e-6) return(pbar >= 0.999)
    # knife edge: defer to the actual fit below
  }
  max(overlap_logit_fit(x, y)$fitted.values) >= 0.999
}

#' @title rcond_check_fail
#' @description Evaluate the per-(g,t) regression-feasibility guard, i.e.
#'  whether `rcond(crossprod(x[controls, ]))` is below machine epsilon. For an
#'  intercept-only design the control-unit Gram matrix is the scalar count of
#'  control units (n0): `rcond` is exactly 1 when n0 > 0 and exactly 0 when
#'  n0 == 0, so the test reduces to "no control units" with no Gram matrix.
#' @param x covariate matrix
#' @param y treatment indicator (controls are `y == 0`)
#' @param intercept_only whether `x` is a single column of ones
#' @return TRUE if the control-unit design is singular or ill-conditioned
#' @noRd
rcond_check_fail <- function(x, y, intercept_only = FALSE) {
  if (intercept_only) {
    return(!any(y == 0))
  }
  control_covs <- x[y == 0, , drop = FALSE]
  rcond(crossprod(control_covs)) < .Machine$double.eps
}

#' @title Check User-Supplied Names Against Internal Names
#' @description Stop if user arguments reference names that `did` creates and
#' mutates internally.
#' @noRd
check_reserved_did_names <- function(yname, tname, idname, gname, xformla,
                                     weightsname = NULL, clustervars = NULL) {
  reserved_names <- c(".w", ".rowid", ".G", ".C", "post",
                      "asif_never_treated", "treated_first_period")
  xvars <- if (is.null(xformla)) character(0) else all.vars(xformla)
  used_names <- unique(c(yname, tname, idname, gname, weightsname,
                         clustervars, xvars))
  used_names <- used_names[!is.na(used_names) & nzchar(used_names)]
  bad_names <- intersect(used_names, reserved_names)
  if (length(bad_names) > 0L) {
    stop("The following variable name(s) are reserved for internal use by `did`: ",
         paste(bad_names, collapse = ", "), ". Please rename these column(s) ",
         "before calling att_gt().")
  }
}

validate_xformla <- function(xformla) {
  if (!is.null(xformla) && !inherits(xformla, "formula")) {
    stop("xformla must be NULL or a formula.")
  }
  invisible(xformla)
}

validate_character_scalar <- function(x, name, allow_null = FALSE) {
  if (allow_null && is.null(x)) return(invisible(x))
  if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x)) {
    stop(name, " must be a single non-missing character string.")
  }
  invisible(x)
}

validate_column_name <- function(x, name, data_names, allow_null = FALSE) {
  validate_character_scalar(x, name, allow_null = allow_null)
  if (is.null(x)) return(invisible(x))
  if (!(x %in% data_names)) {
    stop(name, " must be a character scalar and a name of a column from the dataset.")
  }
  invisible(x)
}

validate_column_names <- function(x, name, data_names, allow_null = FALSE) {
  if (allow_null && is.null(x)) return(invisible(x))
  if (!is.character(x) || length(x) == 0L || anyNA(x) || any(!nzchar(x))) {
    stop(name, " must be NULL or a character vector naming column(s) from the dataset.")
  }
  missing <- setdiff(x, data_names)
  if (length(missing) > 0L) {
    stop(name, " contains column name(s) not found in the dataset: ",
         paste(missing, collapse = ", "), ".")
  }
  invisible(x)
}

complete_finite_cases <- function(data, finite_exclude = character(0)) {
  keep <- stats::complete.cases(data)
  if (!length(keep)) return(keep)
  for (nm in names(data)) {
    # Columns in finite_exclude still get the NA/NaN check via complete.cases()
    # above, but skip the is.finite() check below. This preserves Inf in such
    # columns -- notably gname, where Inf is a documented never-treated code
    # ("group status 0 or Inf") that must NOT be dropped as if it were bad data.
    if (nm %in% finite_exclude) next
    x <- data[[nm]]
    if (is.numeric(x)) {
      finite_x <- is.finite(x)
      if (!is.null(dim(finite_x)) && NROW(finite_x) == length(keep)) {
        finite_x <- rowSums(!finite_x) == 0L
      }
      keep <- keep & as.vector(finite_x)
    }
  }
  keep
}

validate_anticipation <- function(anticipation) {
  if (!is.numeric(anticipation)) {
    stop("anticipation must be numeric. Please convert it.")
  }
  if (length(anticipation) != 1L || is.na(anticipation) ||
      !is.finite(anticipation)) {
    stop("anticipation must be a single finite non-missing number. Please check your arguments.")
  }
  if (anticipation < 0 || anticipation != round(anticipation)) {
    stop("anticipation must be a non-negative whole number. Please check your arguments.")
  }
  invisible(anticipation)
}

validate_logical_scalar <- function(x, name) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    stop(name, " must be a single logical (TRUE or FALSE).")
  }
  invisible(x)
}

validate_numeric_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x)) {
    stop(name, " must be a single non-missing number.")
  }
  invisible(x)
}

validate_finite_numeric_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || !is.finite(x)) {
    stop(name, " must be a single finite non-missing number.")
  }
  invisible(x)
}

validate_finite_numeric_vector <- function(x, name, len) {
  if (!is.numeric(x) || length(x) != len || anyNA(x) || any(!is.finite(x))) {
    stop(name, " must be a numeric vector of length ", len,
         " with only finite non-missing values.")
  }
  invisible(x)
}

validate_positive_numeric_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x <= 0) {
    stop(name, " must be a single positive finite number.")
  }
  invisible(x)
}

validate_alp <- function(alp, name = "alp") {
  if (!is.numeric(alp) || length(alp) != 1 || is.na(alp) || alp <= 0 || alp >= 1) {
    stop(name, " must be a single number strictly between 0 and 1.")
  }
  invisible(alp)
}

validate_probability_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x <= 0 || x >= 1) {
    stop(name, " must be a single finite number strictly between 0 and 1.")
  }
  invisible(x)
}

validate_positive_whole_number <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1 || is.na(x) ||
      !is.finite(x) || x < 1 || x != round(x)) {
    stop(name, " must be a single positive whole number.")
  }
  invisible(x)
}

validate_nonnegative_whole_number <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1 || is.na(x) ||
      !is.finite(x) || x < 0 || x != round(x)) {
    stop(name, " must be a single non-negative whole number.")
  }
  invisible(x)
}

validate_choice_scalar <- function(x, name, choices, message = NULL) {
  if (!is.character(x) || length(x) != 1L || is.na(x) || !(x %in% choices)) {
    if (is.null(message)) {
      message <- paste0(name, " must be one of: ", paste(choices, collapse = ", "), ".")
    }
    stop(message)
  }
  invisible(x)
}

validate_optional_choice_scalar <- function(x, name, choices, message = NULL) {
  if (is.null(x)) return(invisible(x))
  validate_choice_scalar(x, name, choices, message)
}

validate_optional_numeric_scalar <- function(x, name) {
  if (is.null(x)) return(invisible(x))
  validate_numeric_scalar(x, name)
}

#' @title get_wide_data
#' @description A utility function to convert long data to wide data, i.e., takes a 2 period dataset and turns it into a cross sectional dataset.
#'
#' @param data data.table used in function
#' @param yname name of outcome variable that can change over time
#' @param idname unique id
#' @param tname time period name
#'
#' @return data from first period with .y0 (outcome in first period), .y1 (outcome in second period),
#' and .dy (change in outcomes over time) appended to it
#' @noRd
get_wide_data <- function(data, yname, idname, tname) {
  # check if data is data.table
  if (!is.data.table(data)) {
    stop("data must be a data.table")
  }

  # check that only 2 periods of data are available
  if (length(unique(data[[tname]])) != 2) {
    stop("data must have exactly 2 periods")
  }

  # making computations — use set() to avoid copy triggered by $<-
  y_vals <- data[[yname]]
  set(data, j = ".y1", value = data.table::shift(y_vals, -1))
  set(data, j = ".y0", value = y_vals)
  set(data, j = ".dy", value = data[[".y1"]] - y_vals)

  # Subset to first period's rows
  # Pre-extract to avoid data.table .checkTypos when column name matches variable name
  .time_vals <- data[[tname]]
  first.period <- min(.time_vals)
  data <- data[.time_vals == first.period]

  return(data)
}

#' @title Check balanced panel data
#' @description A utility function to check if your dataset is a balanced panel dataset.
#'
#' Precondition: at most one row per (id_col, time_col) combination -- enforced
#' upstream by validate_args(). Under that uniqueness, every unit has at most one
#' row per period, so the panel is balanced iff nrow == n_ids * n_periods.
#'
#' @param data data.table used in function
#' @param id_col name of id column in the dataset
#' @param time_col name of time column in the dataset
#'
#' @return Boolean indicating if the dataset is balanced or not.
#' @noRd
check_balance <- function(data, id_col, time_col) {

  # Check if every unit is observed in every time period (no grouping needed
  # given the (id, time) uniqueness precondition above)
  is_balanced <- nrow(data) == data.table::uniqueN(data[[id_col]]) * data.table::uniqueN(data[[time_col]])

  return(is_balanced)
}

# =============================================================================
# Robustness guards: malformed inputs must be rejected with a clear, identical
# message in BOTH the fast (faster_mode = TRUE) and slow (faster_mode = FALSE)
# paths. Regression tests for two confirmed silent-failure bugs:
#   - duplicate (id, period) rows (slow path used to return wrong numbers)
#   - negative / zero-mean weights (used to silently produce NA)
# =============================================================================

test_that("duplicate (id, period) rows are rejected identically by both modes", {
  set.seed(1)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  ddup <- rbind(d, d[d$id == d$id[1] & d$period == 2, ])  # one duplicated (id, period)
  msg <- "must be unique \\(by tname\\)"
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
    tname = "period", idname = "id", gname = "G", faster_mode = TRUE)), msg)
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
    tname = "period", idname = "id", gname = "G", faster_mode = FALSE)), msg)
  # The guard is intentionally unconditional on panel: with idname supplied it
  # also rejects duplicated (id, period) rows on repeated cross sections and
  # unbalanced panels (the slow path used to silently accept them there).
  for (fm in c(TRUE, FALSE)) {
    expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      faster_mode = fm)), msg)
    expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
      tname = "period", idname = "id", gname = "G", allow_unbalanced_panel = TRUE,
      faster_mode = fm)), msg)
  }
})

test_that("negative / zero-mean weights are rejected identically by both modes", {
  set.seed(1)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$wgt <- ifelse(d$id %% 5 == 0, -1, 1)   # some negative weights
  msg <- "must be non-negative with a positive mean"
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wgt", faster_mode = TRUE)), msg)
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wgt", faster_mode = FALSE)), msg)

  d$wzero <- 0   # all-zero -> non-positive mean
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wzero")), msg)
})

test_that("valid positive weights are still accepted in both modes", {
  set.seed(2)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$wgt <- runif(nrow(d), 0.5, 2)
  for (fm in c(TRUE, FALSE)) {
    r <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", weightsname = "wgt",
      faster_mode = fm, bstrap = FALSE)))
    expect_s3_class(r, "MP")
    expect_true(any(is.finite(r$att)))
  }
})

test_that("fast RC path returns NA instead of crashing when a cell has no treated post observations", {
  set.seed(43)
  ids <- 1:90
  d <- expand.grid(id = ids, period = 1:4)
  d$G <- ifelse(d$id <= 40, 0, ifelse(d$id <= 65, 2, 3))
  d <- d[!(d$G == 2 & d$period >= 3), ]
  d$X <- rnorm(nrow(d))
  d$Y <- 0.5 * d$X + 0.2 * d$period + (d$G > 0 & d$period >= d$G) +
    rnorm(nrow(d), 0, 0.2)

  expect_warning(
    res <- suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      faster_mode = TRUE, est_method = "dr", bstrap = FALSE)),
    "No units in group 2 in time period 3")
  expect_s3_class(res, "MP")
  expect_true(any(is.na(res$att[res$group == 2])))
})

test_that("never-treated small-group stop fires in both modes when multiple groups are small", {
  set.seed(44)
  d <- expand.grid(id = 1:6, period = 1:3)
  d$G <- ifelse(d$id <= 3, 0, 2)
  d$Y <- rnorm(nrow(d))
  for (j in 1:6) d[[paste0("X", j)]] <- rnorm(nrow(d))
  xf <- ~X1 + X2 + X3 + X4 + X5 + X6
  msg <- "never-treated group is too small"

  expect_error(suppressWarnings(att_gt(yname = "Y", xformla = xf, data = d,
    tname = "period", idname = "id", gname = "G", faster_mode = FALSE,
    bstrap = FALSE)), msg)
  expect_error(suppressWarnings(att_gt(yname = "Y", xformla = xf, data = d,
    tname = "period", idname = "id", gname = "G", faster_mode = TRUE,
    bstrap = FALSE)), msg)
})

test_that("slow unbalanced influence aggregation handles fractional numeric ids", {
  set.seed(45)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d <- d[-sample(seq_len(nrow(d)), floor(0.05 * nrow(d))), ]
  d$id <- d$id + 0.123456789012345

  slow <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    allow_unbalanced_panel = TRUE, faster_mode = FALSE, bstrap = FALSE)))
  fast <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    allow_unbalanced_panel = TRUE, faster_mode = TRUE, bstrap = FALSE)))

  expect_equal(slow$att, fast$att, tolerance = 1e-10)
  expect_equal(slow$se, fast$se, tolerance = 1e-8)
})

test_that("fast path preserves user columns named weights", {
  set.seed(46)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)

  d$weights <- d$Y
  slow_y <- suppressWarnings(suppressMessages(att_gt(yname = "weights", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = FALSE, bstrap = FALSE)))
  fast_y <- suppressWarnings(suppressMessages(att_gt(yname = "weights", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = TRUE, bstrap = FALSE)))
  expect_equal(slow_y$att, fast_y$att, tolerance = 1e-10)
  expect_false(all(abs(fast_y$att) < 1e-12))

  d$weights <- d$X
  slow_x <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~weights,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = FALSE, bstrap = FALSE)))
  fast_x <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~weights,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = TRUE, bstrap = FALSE)))
  expect_equal(slow_x$att, fast_x$att, tolerance = 1e-10)
  expect_equal(slow_x$se, fast_x$se, tolerance = 1e-8)
})

test_that("transformed non-finite covariates are dropped before RC overlap checks", {
  # Regression test: a -Inf covariate (log(0), reachable via transform-formula
  # support) used to reach overlap_logit_fit() and degrade affected cells to NA.
  # Preprocessing now drops those rows before either implementation builds 2x2
  # cells, so both modes should warn once, estimate the remaining cells, and
  # stay numerically aligned.
  set.seed(20260609)
  sp <- did::reset.sim(time.periods = 4, n = 400)
  d <- did::build_sim_dataset(sp)
  d$Xpos <- exp(d$X)
  d$Xpos[d$id == unique(d$id)[1]] <- 0  # log(0) = -Inf for one unit

  w_slow <- capture_warnings(
    slow <- suppressMessages(att_gt(yname = "Y", xformla = ~log(Xpos), data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      est_method = "dr", faster_mode = FALSE, bstrap = FALSE)))
  w_fast <- capture_warnings(
    fast <- suppressMessages(att_gt(yname = "Y", xformla = ~log(Xpos), data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      est_method = "dr", faster_mode = TRUE, bstrap = FALSE)))

  expect_identical(w_slow, "dropped 4 rows from original data due to missing or non-finite data")
  expect_identical(w_fast, w_slow)
  expect_false(any(grepl("Error computing internal 2x2 DiD", w_slow)))
  expect_false(any(is.na(slow$att)))
  expect_equal(slow$att, fast$att, tolerance = 1e-10)
  expect_equal(slow$se, fast$se, tolerance = 1e-10)
})

test_that("overlap_logit_fit matches fastglm::fastglm exactly, including the 0.999 decision", {
  # overlap_logit_fit() (fastglmPure with method = 0L, tol = 1e-8, maxit = 100L)
  # replaced fastglm::fastglm(x, y, family = binomial()) at every overlap-check
  # call site with a docstring claim of bit-identical fitted values. Lock that
  # contract -- and the max(fitted) >= 0.999 decision that drives the
  # att = NA-with-warning behavior -- on the degenerate designs where IRLS is
  # most fragile: near separation, complete separation, rank deficiency, and an
  # all-zero column (the sparse-factor-level case).
  set.seed(20260610)
  n <- 200
  x <- rnorm(n)
  designs <- list(
    near_sep = list(X = cbind(1, x), D = rbinom(n, 1, plogis(8 * x))),
    comp_sep = list(X = cbind(1, x), D = as.numeric(x > 0)),
    rank_def = list(X = cbind(1, x, x), D = rbinom(n, 1, plogis(2 * x))),
    zero_col = list(X = cbind(1, x, 0), D = rbinom(n, 1, plogis(2 * x)))
  )
  for (nm in names(designs)) {
    X <- designs[[nm]]$X; D <- designs[[nm]]$D
    # expected "fitted probabilities numerically 0 or 1" noise under separation
    f1 <- suppressWarnings(fastglm::fastglm(X, D, family = binomial()))
    f2 <- suppressWarnings(did:::overlap_logit_fit(X, D))
    expect_identical(f1$fitted.values, f2$fitted.values)
    expect_identical(max(f1$fitted.values) >= 0.999, max(f2$fitted.values) >= 0.999)
  }
})

test_that("slow panel path converts estimator NaN cells to NA like fast mode", {
  set.seed(47)
  d <- expand.grid(id = 1:90, period = 1:4)
  d$G <- ifelse(d$id <= 30, 0, ifelse(d$id <= 60, 2, 3))
  unit_x <- rnorm(90)
  d$X <- unit_x[d$id]
  d$Y <- 0.5 * d$X + 0.2 * d$period + (d$G > 0 & d$period >= d$G) +
    rnorm(nrow(d), 0, 0.2)
  d$w <- ifelse(d$G == 3, 0, 1)

  slow <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    weightsname = "w", faster_mode = FALSE, bstrap = FALSE)))
  fast <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    weightsname = "w", faster_mode = TRUE, bstrap = FALSE)))

  expect_true(any(is.na(slow$att[slow$group == 3])))
  expect_false(any(is.nan(slow$att)))
  expect_equal(is.na(slow$att), is.na(fast$att))
})

test_that("never-treated units coded as gname = Inf are NOT dropped (Inf is a valid never-treated code)", {
  # Regression test: complete_finite_cases() must exclude gname from its
  # is.finite() filter. Inf is a documented never-treated code (att_gt() reports
  # "group status 0 or Inf"); on master gname = Inf produces results identical to
  # gname = 0. A naive "drop all non-finite numeric rows" filter silently deletes
  # every never-treated unit -- erroring or, worse (control_group="nevertreated"),
  # returning plausible-looking but WRONG estimates with no error.
  data(mpdta, package = "did")
  d_inf <- mpdta
  d_inf$first.treat[d_inf$first.treat == 0] <- Inf

  for (fm in c(TRUE, FALSE)) {
    for (cg in c("nevertreated", "notyettreated")) {
      ref <- suppressWarnings(suppressMessages(att_gt(yname = "lemp", tname = "year",
        idname = "countyreal", gname = "first.treat", xformla = ~lpop, data = mpdta,
        control_group = cg, faster_mode = fm, bstrap = FALSE, cband = FALSE)))
      inf <- suppressWarnings(suppressMessages(att_gt(yname = "lemp", tname = "year",
        idname = "countyreal", gname = "first.treat", xformla = ~lpop, data = d_inf,
        control_group = cg, faster_mode = fm, bstrap = FALSE, cband = FALSE)))
      # never-treated units must survive: same number of influence-function rows
      expect_equal(nrow(inf$inffunc), nrow(ref$inffunc))
      expect_equal(inf$att, ref$att, tolerance = 1e-10)
      expect_equal(inf$inffunc, ref$inffunc, tolerance = 1e-10)
    }
  }
})

test_that("NA / NaN gname rows are still dropped while Inf is preserved", {
  # The finite-exclude carve-out for gname must not also disable the NA/NaN check:
  # complete.cases() still removes missing/NaN gname rows.
  df <- data.frame(g = c(0, Inf, NA, NaN, 5),
                   y = c(1, 2, 3, 4, Inf),
                   x = c(1, 2, 3, 4, 5))
  expect_equal(complete_finite_cases(df, finite_exclude = "g"),
               c(TRUE, TRUE, FALSE, FALSE, FALSE))
  # without the carve-out, the Inf-coded never-treated row would also be dropped
  expect_equal(complete_finite_cases(df),
               c(TRUE, FALSE, FALSE, FALSE, FALSE))
})

#' @title Process Results from [compute.att_gt()]
#'
#' @param attgt.list list of results from [compute.att_gt()]
#'
#' @return list with elements:
#' \item{group}{which group a set of results belongs to}
#' \item{tt}{which time period a set of results belongs to}
#' \item{att}{the group time average treatment effect}
#'
#' @export
process_attgt <- function(attgt.list) {
  group <- vapply(attgt.list, function(x) as.numeric(x[["group"]]), numeric(1))
  att   <- vapply(attgt.list, function(x) as.numeric(x[["att"]]), numeric(1))
  tt    <- vapply(attgt.list, function(x) as.numeric(x[["year"]]), numeric(1))

  list(group=group, att=att, tt=tt)
}

#' @title Multiplier Bootstrap
#'
#' @description A function to take an influence function and use the
#'  multiplier bootstrap to compute standard errors and critical values for
#'  uniform confidence bands.
#'
#' @param inf.func an influence function
#' @param DIDparams DIDparams object
#' @param pl whether or not to use parallel processing in the multiplier
#'  bootstrap, default=FALSE
#' @param cores the number of cores to use with parallel processing,
#'  default=1
#' @param return_V whether to compute and return the bootstrap variance
#'  matrix `V`. Default is `TRUE`. Internal callers that only consume
#'  `bres`, `se`, or `crit.val` set this to `FALSE` to skip the computation
#'  (it is the only O(biters x k^2) step in the function).
#'
#' @return list with elements
#' \item{bres}{results from each bootstrap iteration}
#' \item{V}{variance matrix (`NULL` when `return_V = FALSE`)}
#' \item{se}{standard errors}
#' \item{crit.val}{a critical value for computing uniform confidence bands}
#'
#' @export
mboot <- function(inf.func, DIDparams, pl = FALSE, cores = 1, return_V = TRUE) {

  # setup needed variables according to faster_mode; This returns different type of objects
  # depending on whether we are in faster_mode or not that has to be handled in the code below
  idname <- DIDparams$idname
  clustervars <- DIDparams$clustervars
  biters <- DIDparams$biters
  tname <- DIDparams$tname
  alp <- DIDparams$alp
  panel <- DIDparams$panel
  true_repeated_cross_sections <- DIDparams$true_repeated_cross_sections
  unbalanced_panel <- DIDparams$allow_unbalanced_panel
  # Reuse the per-unit cluster vector that att_gt() stored in DIDparams when it
  # aligns with the influence-function rows: it was already derived and validated
  # (time invariance) there, so the data load, the aggregate() check, and the
  # unique() derivation below can all be skipped. Fall back to deriving the
  # clusters from the data for (older) DIDparams objects lacking an aligned
  # cluster_vector -- the fallback keeps the time-varying-cluster stop.
  cluster_vector <- DIDparams$cluster_vector
  use_cluster_vector <- !is.null(cluster_vector) &&
    length(cluster_vector) == NROW(inf.func)
  # Avoid full data copy when possible; only needed for clustering without an
  # aligned cluster_vector
  need_data <- length(clustervars) > 0 && !use_cluster_vector
  if (need_data) {
    if (isTRUE(DIDparams$faster_mode) && !is.null(DIDparams$time_invariant_data)) {
      dta <- as.data.frame(DIDparams$time_invariant_data)
    } else {
      data <- as.data.frame(DIDparams$data)
      tlist <- if (isTRUE(DIDparams$faster_mode)) DIDparams$time_periods else sort(unique(data[,tname]))
      if (panel) {
        dta <- data[data[,tname] == tlist[1], ]
      } else {
        dta <- data
      }
    }
  }

  # Convert sparse matrix to dense for bootstrap computation
  inf.func <- as.matrix(inf.func)

  # set correct number of units
  n <- nrow(inf.func)

  # if include id as variable to cluster on
  # drop it as we do this automatically
  if (idname %in% clustervars) {
    clustervars <- clustervars[-which(clustervars==idname)]
  }

  if(!is.null(clustervars)){
    if(is.numeric(clustervars)){
      stop("'clustervars' must be a character string specifying the name of the clustering variable, not a numeric value.")
    }
  }
  # we can only handle up to 2-way clustering
  # (in principle could do more, but not high priority now)
  if (length(clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  if (length(clustervars) > 0 && !use_cluster_vector) {
    if(!DIDparams$faster_mode){
      # check that cluster variable does not vary over time within unit
      # reuse 'data' already loaded above (need_data is TRUE on this fallback path)
      clust_tv <- aggregate(data[,clustervars], by=list(data[,idname]), function(rr) length(unique(rr))==1)
      if (!all(clust_tv[,2])) {
        stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
      }
    }
  }

  # multiplier bootstrap
  n_clusters <- n
  if (length(clustervars)==0) {
    bres <- sqrt(n) * run_multiplier_bootstrap(inf.func, biters, pl, cores)
  } else {
    # Cluster-robust multiplier bootstrap following Callaway & Sant'Anna (2021, Remark 10): draw
    # cluster-specific multipliers and apply them to the influence function aggregated to cluster
    # *sums*, consistent with the (1/n) empirical average that defines the estimator. For
    # equal-sized clusters this coincides with the previous aggregation; for unbalanced clusters and
    # repeated cross-sections it keeps the bootstrap aligned with the cluster-robust variance.
    if (use_cluster_vector) {
      # att_gt() already built this vector aligned with the inf.func rows (and
      # validated time invariance when doing so); reusing it avoids reloading
      # the data and re-deriving the clusters on every call.
      cluster <- cluster_vector
      n_clusters <- length(unique(cluster))
    } else {
      n_clusters <- length(unique(dta[,clustervars]))
      cluster <- unique(dta[,c(idname,clustervars)])[,2]
    }
    cluster_sum_if <- rowsum(inf.func, cluster, reorder=TRUE)
    bres <- sqrt(n_clusters) * run_multiplier_bootstrap(cluster_sum_if, biters, pl, cores)
  }


  # handle vector and matrix case differently, so you get nxk matrix
  # ifelse(class(bres)=="matrix", bres <- t(bres), bres <- as.matrix(bres))

  if (isTRUE(class(bres) == "numeric")) bres <- as.matrix(bres)

  # Non-degenerate dimensions
  # ndg.dim <- (base::colSums(bres) != 0)
  ndg.dim <- (!is.na(colSums(bres))) & (base::colSums(bres^2) > sqrt(.Machine$double.eps)*10)
  # If NA, set it to false
  #ndg.dim[is.na(ndg.dim)] <- FALSE
  bres <- as.matrix(bres[ , ndg.dim])

  # bootstrap variance matrix (this matrix can be defective because of degenerate cases).
  # Computed only on request: no internal caller consumes it, and it is the only
  # O(biters * k^2) term in mboot.
  V <- if (return_V) cov(bres) else NULL
  # bootstrap standard error: IQR-based scale normalized to the SD of a normal.
  # The qnorm difference is a constant (hoisted out of the per-column loop), and
  # the type = 1 quantiles are exactly the order statistics x_(ceil(B * p)), so a
  # single sort per column replaces the quantile() calls bit-identically (the
  # post-sort length is the per-column non-NA count, preserving na.rm = TRUE;
  # sort.int() drops NAs by default (na.last = NA)).
  iqr_norm <- qnorm(.75) - qnorm(.25)
  bSigma <- apply(bres, 2, function(b) {
    b <- sort.int(b)
    nb <- length(b)
    (b[ceiling(.75 * nb)] - b[ceiling(.25 * nb)]) / iqr_norm
  })

  # critical value for uniform confidence band: per bootstrap draw, the max over
  # columns of |b / bSigma|. Vectorized as a column-fold of pmax (equivalent to the
  # previous row-wise apply(max, na.rm = TRUE): NA entries are ignored via -Inf, and
  # an all-NA row yields -Inf, which is dropped by the is.finite filter below).
  if (ncol(bres) == 0L) {
    # All bootstrap dimensions degenerate: every row has no finite t-statistic.
    # This matches the previous apply(max, na.rm = TRUE), which returned -Inf for
    # every row and then dropped them all via the is.finite filter.
    bT <- numeric(0)
  } else {
    scaled <- abs(bres / rep(bSigma, each = nrow(bres)))
    scaled[is.na(scaled)] <- -Inf
    bT <- scaled[, 1L]
    for (j in seq_len(ncol(scaled))[-1L]) bT <- pmax(bT, scaled[, j])
    bT <- bT[is.finite(bT)]
  }
  crit.val <- quantile(bT, 1-alp, type=1, na.rm = TRUE)

  #se <- rep(0, length(ndg.dim))
  se <- rep(NA, length(ndg.dim))
  # Standard error from the bootstrap interquartile-range scale. With cluster sums (above) this is the
  # cluster-robust SE (1/n) * sqrt(sum_c S_c^2); without clustering n_clusters = n and it reduces to the
  # i.i.d. form bSigma / sqrt(n). (Equivalent to the previous bSigma / sqrt(n_clusters) when clusters are
  # equal-sized; the change matters only for unbalanced clusters / repeated cross-sections.)
  se[ndg.dim] <- as.numeric(bSigma) * sqrt(n_clusters) / n
  #se[se==0] <- NA

  list(bres = bres, V = V, se = se, crit.val = crit.val)
}

run_multiplier_bootstrap <- function(inf.func, biters, pl = FALSE, cores = 1) {
  ngroups = ceiling(biters/cores)
  chunks = rep(ngroups, cores)
  # Round down so you end up with the right number of biters
  chunks[1] = chunks[1] + biters - sum(chunks)

  n <- nrow(inf.func)
  parallel.function <- function(biters) {
    BMisc::multiplier_bootstrap(inf.func, biters)
  }
  # From tests, this is about where it becomes worth it to parallelize
  if (.Platform$OS.type == "windows" && pl == TRUE && cores > 1) {
    warning("Parallel processing (pl=TRUE) is not supported on Windows. Using sequential processing instead.")
    pl <- FALSE
  }
  if(n > 2500 & pl == TRUE & cores > 1) {
    results = parallel::mclapply(
      chunks,
      FUN = parallel.function,
      mc.cores = cores
    )
    results = do.call(rbind, results)
  } else {
    results = BMisc::multiplier_bootstrap(inf.func, biters)
  }
  return(results)
}

# =============================================================================
# mboot() post-processing: the rewritten critical-value computation (bSigma via
# a single two-probability quantile call, bT via an abs/NA->-Inf/pmax column
# fold) must reproduce the previous row-wise apply() reference, and the new
# all-degenerate branch (ncol(bres) == 0L after dropping degenerate dimensions)
# must return NA crit.val / all-NA se without error -- matching the old
# behavior where every row's max was -Inf and got dropped by is.finite().
# mboot() runs directly on a DIDparams-like list: with clustervars = NULL it
# never touches the data slot (see test-pretest-vectorization.R).
# =============================================================================

dp_mboot <- list(idname = "id", clustervars = NULL, biters = 100,
                 tname = "period", alp = 0.05, panel = TRUE,
                 true_repeated_cross_sections = FALSE,
                 allow_unbalanced_panel = FALSE, faster_mode = FALSE)

test_that("mboot on an all-degenerate influence function returns NA crit.val and all-NA se", {
  n <- 80
  ifz <- Matrix::sparseMatrix(i = integer(0), j = integer(0), x = numeric(0),
                              dims = c(n, 3))   # all-zero sparse n x 3
  set.seed(101)
  out <- mboot(ifz, dp_mboot)
  expect_identical(ncol(out$bres), 0L)          # every dimension dropped as degenerate
  expect_true(is.na(out$crit.val))
  expect_length(out$se, 3)
  expect_true(all(is.na(out$se)))
})

test_that("mboot pmax-fold critical value equals the old row-wise apply reference", {
  n <- 80
  set.seed(202)
  infm <- cbind(matrix(rnorm(n * 3), n, 3), 0)  # 3 finite columns + one all-zero
  set.seed(303)
  out <- mboot(infm, dp_mboot)

  # Degenerate column dropped before post-processing; its se slot stays NA.
  expect_identical(ncol(out$bres), 3L)
  expect_true(all(is.finite(out$se[1:3])))
  expect_true(is.na(out$se[4]))

  # Old reference (pre-rewrite mboot) recomputed on the same returned bres:
  # per-column IQR scale, then per-row max(abs(b / bSigma), na.rm = TRUE).
  bres <- out$bres
  iqr_norm <- qnorm(.75) - qnorm(.25)
  bSigma_old <- apply(bres, 2, function(b) {
    (quantile(b, .75, type = 1, na.rm = TRUE) -
       quantile(b, .25, type = 1, na.rm = TRUE)) / iqr_norm
  })
  bT_old <- suppressWarnings(apply(bres, 1, function(b) max(abs(b / bSigma_old), na.rm = TRUE)))
  bT_old <- bT_old[is.finite(bT_old)]
  crit_old <- unname(quantile(bT_old, 1 - dp_mboot$alp, type = 1, na.rm = TRUE))

  expect_equal(unname(out$crit.val), crit_old, tolerance = 1e-12)
  expect_equal(unname(out$se[1:3]), unname(bSigma_old) / sqrt(n), tolerance = 1e-12)
})

#' @title Compute Aggregated Treatment Effect Parameters
#'
#' @description Does the heavy lifting on computing aggregated group-time
#'  average treatment effects
#'
#' @inheritParams att_gt
#' @inheritParams aggte
#' @param call The function call to aggte
#'
#' @return [`AGGTEobj`] object
#'
#' @keywords internal
#'
#' @export
compute.aggte <- function(MP,
                          type = "group",
                          balance_e = NULL,
                          min_e = -Inf,
                          max_e = Inf,
                          na.rm = FALSE,
                          bstrap = NULL,
                          biters = NULL,
                          cband = NULL,
                          alp = NULL,
                          clustervars = NULL,
                          call = NULL) {
  if (!inherits(MP, "MP")) {
    stop("MP must be an MP object produced by att_gt().")
  }

  #-----------------------------------------------------------------------------
  # unpack MP object
  #-----------------------------------------------------------------------------
  # load parameters
  group <- MP$group
  t <- MP$t
  att <- MP$att
  dp <- MP$DIDparams
  inffunc1 <- MP$inffunc
  n <- MP$n

  validate_logical_scalar(na.rm, "na.rm")
  validate_choice_scalar(
    type,
    "type",
    c("simple", "dynamic", "group", "calendar"),
    '`type` must be one of c("simple", "dynamic", "group", "calendar")'
  )
  validate_numeric_scalar(min_e, "min_e")
  validate_numeric_scalar(max_e, "max_e")
  if (!is.null(balance_e)) validate_nonnegative_whole_number(balance_e, "balance_e")

  # aggte() needs the influence functions to aggregate and to compute standard errors.
  # They are absent when att_gt() was run with compute_inffunc = FALSE (point estimates only).
  if (is.null(inffunc1)) {
    stop("This att_gt() result was produced with compute_inffunc = FALSE (point estimates ",
         "only), so it has no influence functions and cannot be aggregated by aggte(). ",
         "Re-run att_gt() with compute_inffunc = TRUE (the default) to use aggte().")
  }
  if (length(group) != length(t) || length(att) != length(group)) {
    stop("MP object has inconsistent group, time, and att lengths.")
  }
  if (NCOL(inffunc1) != length(att)) {
    stop("MP object has inconsistent influence-function columns and att estimates.")
  }
  if (!is.null(n) && NROW(inffunc1) != n) {
    stop("MP object has inconsistent influence-function rows and n.")
  }


  gname <- dp$gname
  tname <- dp$tname
  idname <- dp$idname
  panel <- dp$panel
  if (is.null(clustervars)) {
    clustervars <- dp$clustervars
  }

  # Aggregation-level clustering reuses the per-unit cluster information that att_gt() stored when
  # clustervars was supplied to it. If clustering is requested here (inherited from the att_gt object or
  # via an aggte-level clustervars override) but that information is unavailable -- because the cluster
  # variable was not passed to att_gt() (so neither the cluster column nor cluster_vector were retained),
  # or because an override names a different variable than att_gt clustered on -- it cannot be honored.
  # Warn and fall back to non-clustered standard errors, instead of silently returning the i.i.d. SE
  # (analytic path) or erroring in mboot() (bootstrap path). To make this a hard error instead, replace
  # warning() with stop() below.
  extra_cl_req <- clustervars[!(clustervars %in% c(idname, ""))]
  if (length(extra_cl_req) > 0) {
    cv <- dp$cluster_vector
    can_cluster <- !is.null(cv) && length(cv) == nrow(inffunc1) &&
      identical(extra_cl_req, dp$cluster_vector_var)
    if (!can_cluster) {
      warning(paste0(
        "Clustered standard errors were requested in aggte() (clustervars = '",
        paste(extra_cl_req, collapse = "', '"), "'), but the cluster information needed is not available ",
        "in this att_gt object",
        if (!is.null(dp$cluster_vector_var)) {
          paste0(" (att_gt() clustered on '", paste(dp$cluster_vector_var, collapse = "', '"),
                 "', and aggte() cannot switch to a different cluster variable)")
        } else {
          " (clustervars was not supplied to att_gt(), so aggte() cannot introduce clustering)"
        },
        ". Reporting standard errors that do NOT account for clustering; re-run att_gt() with ",
        "clustervars = '", paste(extra_cl_req, collapse = "', '"), "' for clustered inference."))
      clustervars <- NULL
    }
  }
  if (is.null(bstrap)) {
    bstrap <- dp$bstrap
  }
  if (is.null(biters)) {
    biters <- dp$biters
  }
  if (is.null(alp)) {
    alp <- dp$alp
  }
  if (is.null(cband)) {
    cband <- dp$cband
  }
  validate_logical_scalar(bstrap, "bstrap")
  validate_logical_scalar(cband, "cband")
  validate_alp(alp)
  if (bstrap || cband) validate_positive_whole_number(biters, "biters")
  if (isTRUE(dp$faster_mode)) {
    tlist <- dp$time_periods
    glist <- dp$treated_groups
  } else {
    data <- as.data.frame(dp$data)
    tlist <- dp$tlist
    glist <- dp$glist
  }

  # In faster_mode, the full long data is only needed on the two branches below that
  # actually consume `data` (most faster-mode panel calls use time_invariant_data
  # instead). Materialize it lazily there, recoding gname Inf -> 0 (the old
  # convention) on the converted copy so the data.table stored inside the user's
  # MP object is never modified by reference.
  faster_mode_long_data <- function() {
    data <- as.data.frame(dp$data)
    data[data[, gname] == Inf, gname] <- 0
    data
  }

  # overwrite MP objects (so we can actually compute bootstrap)
  MP$DIDparams$clustervars <- clustervars
  MP$DIDparams$bstrap <- bstrap
  MP$DIDparams$biters <- biters
  MP$DIDparams$alp <- alp
  MP$DIDparams$cband <- cband
  dp <- MP$DIDparams

  if (na.rm) {
    notna <- !is.na(att)
    if (!any(notna)) {
      stop("All att_gt() estimates are NA. Cannot compute aggregated treatment effects. ",
           "Check your att_gt() results for estimation problems (e.g., too few observations, ",
           "singular matrices, or overlap violations).")
    }
    group <- group[notna]
    t <- t[notna]
    att <- att[notna]
    inffunc1 <- inffunc1[, notna, drop = FALSE]
    # tlist <- sort(unique(t))
    glist <- sort(unique(group))

    # If aggte is of the group type, ensure we have non-missing post-treatment ATTs for each group
    if (type == "group") {
      # Get the groups that have some non-missing ATT(g,t) in post-treatmemt periods
      gnotna <- sapply(glist, function(g) {
        # look at post-treatment periods for group g, restricted to the SAME
        # max_e window used by the group-specific estimate below (selective.att.g /
        # selective.se.inner). Without the (t <= group + max_e) condition a group
        # whose only non-NA ATT(g,t) lies PAST max_e would pass this filter but then
        # have an all-NA (hence, under na.rm, empty) selection in the estimate,
        # erroring in get_agg_inf_func(). (max_e defaults to Inf, so this is a no-op
        # unless the user sets a finite max_e.)
        whichg <- which((group == g) & (g <= t) & (t <= (group + max_e)))
        attg <- att[whichg]
        group_select <- !is.na(mean(attg))
        return(group_select)
      })
      gnotna <- glist[gnotna]
      # indicator for not all post-treatment ATT(g,t) missing
      not_all_na <- group %in% gnotna
      if (!any(not_all_na)) {
        stop("No groups have non-missing post-treatment att_gt() estimates. ",
             "Cannot compute group aggregation. Check your att_gt() results.")
      }
      # Re-do the na.rm thing to update the groups
      group <- group[not_all_na]
      t <- t[not_all_na]
      att <- att[not_all_na]
      inffunc1 <- inffunc1[, not_all_na, drop = FALSE]
      # tlist <- sort(unique(t))
      glist <- sort(unique(group))
    }
  }

  if ((na.rm == FALSE) && base::anyNA(att)) stop("Missing values found in att_gt() estimates. To remove them, set `na.rm = TRUE`.")

  # data from first period
  # ifelse(panel,
  #       dta <- data[ data[,tname]==tlist[1], ],
  #       dta <- data
  #       )
  if (panel) {
    if (isTRUE(dp$faster_mode) && !is.null(dp$time_invariant_data)) {
      # use pre-computed time-invariant data (one row per unit) instead of subsetting
      dta <- as.data.frame(dp$time_invariant_data)
      # map Inf back to 0 for gname to match the old convention
      dta[dta[, gname] == Inf, gname] <- 0
    } else {
      if (isTRUE(dp$faster_mode)) data <- faster_mode_long_data()
      # data from first period
      dta <- data[data[, tname] == tlist[1], ]
    }
  } else {
    if (isTRUE(dp$faster_mode)) data <- faster_mode_long_data()
    # Aggregate to one row per unit. aggregate() returns rows sorted by idname, but the influence
    # function (inffunc1) is in the data's *first-appearance* unit order -- which differs from sorted for
    # unbalanced panels under faster_mode. Restore that order so the estimated-weight influence term (wif)
    # is added to the matching units in get_agg_inf_func(); otherwise wif lands on the wrong units and the
    # aggregated standard error is wrong (the point estimate is unaffected because wif is mean-zero).
    dta_agg <- base::suppressWarnings(stats::aggregate(data, list((data[, idname])), mean))
    dta <- dta_agg[match(unique(data[, idname]), dta_agg[, 1L]), -1L, drop = FALSE]
  }

  #-----------------------------------------------------------------------------
  # data organization and recoding
  #-----------------------------------------------------------------------------

  # do some recoding to make sure time periods are 1 unit apart
  # and then put these back together at the end
  originalt <- t
  originalgroup <- group
  originalglist <- glist
  originaltlist <- tlist
  # In case g's are not part of tlist
  originalgtlist <- sort(unique(c(originaltlist, originalglist)))
  uniquet <- seq(1, length(unique(originalgtlist)))
  # lookup vectors for vectorized mapping between original and new t values
  orig_with_zero <- c(originalgtlist, 0)
  new_with_zero <- c(uniquet, 0)
  # vectorized function to switch from "new" t values to original t values
  t2orig <- function(t) {
    orig_with_zero[match(t, new_with_zero)]
  }
  # vectorized function to switch from original t values to new sequential t values
  orig2t_vec <- function(orig) {
    new_with_zero[match(orig, orig_with_zero)]
  }
  t <- orig2t_vec(originalt)
  group <- orig2t_vec(originalgroup)
  glist <- orig2t_vec(originalglist)
  tlist <- unique(t)
  maxT <- max(t)

  # Set the weights
  if (".w" %in% names(dta)) {
    weights.ind <- dta$.w
  } else if (isTRUE(dp$faster_mode)) {
    weights.ind <- dta$weights
  } else {
    weights.ind <- dta$.w
  }

  # we can work in overall probabilities because conditioning will cancel out
  # cause it shows up in numerator and denominator
  gvar <- dta[[gname]]  # extract the group column once, reused below
  pg <- sapply(originalglist, function(g) mean(weights.ind * (gvar == g)))

  # length of this is equal to number of groups
  pgg <- pg

  # same but length is equal to the number of ATT(g,t)
  pg <- pg[match(group, glist)]

  # which group time average treatment effects are post-treatment
  keepers <- which(group <= t & t <= (group + max_e)) ### added second condition to allow for limit on longest period included in att

  # n x 1 vector of group variable
  G <- orig2t_vec(gvar)

  #-----------------------------------------------------------------------------
  # Compute the simple ATT summary
  #-----------------------------------------------------------------------------

  if (type == "simple") {
    # simple att
    # averages all post-treatment ATT(g,t) with weights
    # given by group size
    simple.att <- sum(att[keepers] * pg[keepers]) / (sum(pg[keepers]))
    if (is.nan(simple.att)) simple.att <- NA

    # get the part of the influence function coming from estimated weights
    simple.wif <- wif(keepers, pg, weights.ind, G, group)

    # get the overall influence function
    simple.if <- get_agg_inf_func(
      att = att,
      inffunc1 = inffunc1,
      whichones = keepers,
      weights.agg = pg[keepers] / sum(pg[keepers]),
      wif = simple.wif
    )
    # Make it as vector
    simple.if <- as.numeric(simple.if)

    # get standard errors from overall influence function
    simple.se <- getSE(simple.if, dp)
    if (!is.na(simple.se)) {
      if (simple.se <= sqrt(.Machine$double.eps) * 10) simple.se <- NA
    }


    return(AGGTEobj(
      overall.att = simple.att,
      overall.se = simple.se,
      type = type,
      inf.function = list(simple.att = simple.if),
      call = call,
      DIDparams = dp
    ))
  }

  #-----------------------------------------------------------------------------
  # Compute the group (i.e., selective) treatment timing estimators
  #-----------------------------------------------------------------------------

  if (type == "group") {
    # get group specific ATTs
    # note: there are no estimated weights here
    selective.att.g <- sapply(glist, function(g) {
      # look at post-treatment periods for group g
      whichg <- which((group == g) & (g <= t) & (t <= (group + max_e))) ### added last condition to allow for limit on longest period included in att
      attg <- att[whichg]
      mean(attg)
    })
    selective.att.g[is.nan(selective.att.g)] <- NA


    # get standard errors for each group specific ATT
    selective.se.inner <- lapply(glist, function(g) {
      whichg <- which((group == g) & (g <= t) & (t <= (group + max_e))) ### added last condition to allow for limit on longest period included in att
      inf.func.g <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whichg,
        weights.agg = pg[whichg] / sum(pg[whichg]),
        wif = NULL
      ))
      se.g <- getSE(inf.func.g, dp)
      list(inf.func = inf.func.g, se = se.g)
    })

    # recover standard errors separately by group
    selective.se.g <- unlist(BMisc::get_list_element(selective.se.inner, "se"))
    selective.se.g[selective.se.g <= sqrt(.Machine$double.eps) * 10] <- NA

    # recover influence function separately by group
    selective.inf.func.g <- simplify2array(BMisc::get_list_element(selective.se.inner, "inf.func"))

    # use multiplier bootstrap (across groups) to get critical value
    # for constructing uniform confidence bands
    selective.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      selective.crit.val <- mboot(selective.inf.func.g, dp, return_V = FALSE)$crit.val

      if (is.na(selective.crit.val) | is.infinite(selective.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        selective.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (selective.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        selective.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (selective.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall att under selective treatment timing
    # (here use pgg instead of pg because we can just look at each group)
    selective.att <- sum(selective.att.g * pgg) / sum(pgg)

    # account for having to estimate pgg in the influence function
    selective.wif <- wif(
      keepers = 1:length(glist),
      pg = pgg,
      weights.ind = weights.ind,
      G = G,
      group = glist
    )

    # get overall influence function
    selective.inf.func <- get_agg_inf_func(
      att = selective.att.g,
      inffunc1 = selective.inf.func.g,
      whichones = (1:length(glist)),
      weights.agg = pgg / sum(pgg),
      wif = selective.wif
    )


    selective.inf.func <- as.numeric(selective.inf.func)
    # get overall standard error
    selective.se <- getSE(selective.inf.func, dp)
    if (!is.na(selective.se)) {
      if ((selective.se <= sqrt(.Machine$double.eps) * 10)) selective.se <- NA
    }

    return(AGGTEobj(
      overall.att = selective.att,
      overall.se = selective.se,
      type = type,
      egt = originalglist,
      att.egt = selective.att.g,
      se.egt = selective.se.g,
      crit.val.egt = selective.crit.val,
      inf.function = list(
        selective.inf.func.g = selective.inf.func.g,
        selective.inf.func = selective.inf.func
      ),
      call = call,
      DIDparams = dp
    ))
  }



  #-----------------------------------------------------------------------------
  # Compute the event-study estimators
  #-----------------------------------------------------------------------------

  if (type == "dynamic") {
    # event times
    # this looks at all available event times
    # note: event times can be negative here.
    # note: event time = 0 corresponds to "on impact"
    # eseq <- unique(t-group)
    eseq <- unique(originalt - originalgroup)
    eseq <- eseq[order(eseq)]

    # if the user specifies balance_e, then we are going to
    # drop some event times and some groups; if not, we just
    # keep everything (that is what this variable is for)
    include.balanced.gt <- rep(TRUE, length(originalgroup))

    # if we balance the sample with resepect to event time
    if (!is.null(balance_e)) {
      include.balanced.gt <- (t2orig(maxT) - originalgroup >= balance_e)

      eseq <- unique(originalt[include.balanced.gt] - originalgroup[include.balanced.gt])
      eseq <- eseq[order(eseq)]

      eseq <- eseq[(eseq <= balance_e) & (eseq >= balance_e - t2orig(maxT) + t2orig(1))]
    }

    # only looks at some event times
    eseq <- eseq[(eseq >= min_e) & (eseq <= max_e)]

    # Guard the empty window (e.g. min_e/max_e exclude every event time):
    # downstream sapply(eseq, ...) would otherwise return a list and fail later
    # with a cryptic "Not compatible with requested type" error.
    if (length(eseq) == 0) {
      stop("No event times fall within the requested window. ",
           "Adjust 'min_e'/'max_e' (and 'balance_e') so at least one event ",
           "time is included.")
    }

    # compute atts that are specific to each event time
    dynamic.att.e <- sapply(eseq, function(e) {
      # keep att(g,t) for the right g&t as well as ones that
      # are not trimmed out from balancing the sample
      whiche <- which((originalt - originalgroup == e) & (include.balanced.gt))
      atte <- att[whiche]
      pge <- pg[whiche] / (sum(pg[whiche]))
      sum(atte * pge)
    })

    # compute standard errors for dynamic effects
    dynamic.se.inner <- lapply(eseq, function(e) {
      whiche <- which((originalt - originalgroup == e) & (include.balanced.gt))
      pge <- pg[whiche] / (sum(pg[whiche]))
      wif.e <- wif(whiche, pg, weights.ind, G, group)
      inf.func.e <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whiche,
        weights.agg = pge,
        wif = wif.e
      ))
      se.e <- getSE(inf.func.e, dp)
      list(inf.func = inf.func.e, se = se.e)
    })

    dynamic.se.e <- unlist(BMisc::get_list_element(dynamic.se.inner, "se"))
    dynamic.se.e[dynamic.se.e <= sqrt(.Machine$double.eps) * 10] <- NA

    dynamic.inf.func.e <- simplify2array(BMisc::get_list_element(dynamic.se.inner, "inf.func"))

    dynamic.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      dynamic.crit.val <- mboot(dynamic.inf.func.e, dp, return_V = FALSE)$crit.val

      if (is.na(dynamic.crit.val) | is.infinite(dynamic.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        dynamic.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (dynamic.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        dynamic.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (dynamic.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall average treatment effect
    # by averaging over positive dynamics
    epos <- eseq >= 0
    dynamic.att <- mean(dynamic.att.e[epos])
    dynamic.inf.func <- get_agg_inf_func(
      att = dynamic.att.e[epos],
      inffunc1 = as.matrix(dynamic.inf.func.e[, epos]),
      whichones = seq_len(sum(epos)),
      weights.agg = (rep(1 / sum(epos), sum(epos))),
      wif = NULL
    )

    dynamic.inf.func <- as.numeric(dynamic.inf.func)
    dynamic.se <- getSE(dynamic.inf.func, dp)
    if (!is.na(dynamic.se)) {
      if (dynamic.se <= sqrt(.Machine$double.eps) * 10) dynamic.se <- NA
    }

    return(AGGTEobj(
      overall.att = dynamic.att,
      overall.se = dynamic.se,
      type = type,
      egt = eseq,
      att.egt = dynamic.att.e,
      se.egt = dynamic.se.e,
      crit.val.egt = dynamic.crit.val,
      inf.function = list(
        dynamic.inf.func.e = dynamic.inf.func.e,
        dynamic.inf.func = dynamic.inf.func
      ),
      call = call,
      min_e = min_e,
      max_e = max_e,
      balance_e = balance_e,
      DIDparams = dp
    ))
  }

  #-----------------------------------------------------------------------------
  # calendar time effects
  #-----------------------------------------------------------------------------

  if (type == "calendar") {
    # min_e / max_e / balance_e have no effect on calendar-time aggregation;
    # warn if the user set them so the (correct) unrestricted result is not
    # mistaken for a windowed one.
    if (is.finite(max_e) || is.finite(min_e) || !is.null(balance_e)) {
      warning("`min_e`, `max_e`, and `balance_e` are ignored for type = \"calendar\"; ",
              "returning the unrestricted calendar-time effects.")
    }
    # drop time periods where no one is treated yet
    # (can't get treatment effects in those periods)
    minG <- min(group)
    calendar.tlist <- tlist[tlist >= minG]

    # Drop calendar periods with no non-missing post-treatment ATT(g,t) cell
    # (e.g. after na.rm removed all of a period's cells, common with unbalanced
    # panels that have genuinely-NA cells). Analogous to the `gnotna` guard in
    # the group branch above; without it wif()/get_agg_inf_func() are called on
    # an empty selection and error with a cryptic message.
    has_post <- vapply(calendar.tlist,
                       function(t1) any((t == t1) & (group <= t)), logical(1))
    calendar.tlist <- calendar.tlist[has_post]
    if (length(calendar.tlist) == 0) {
      stop("No calendar periods have non-missing post-treatment att_gt() ",
           "estimates. Cannot compute calendar aggregation. Check your ",
           "att_gt() results.")
    }

    # calendar time specific atts
    calendar.att.t <- sapply(calendar.tlist, function(t1) {
      # look at post-treatment periods for group g
      whicht <- which((t == t1) & (group <= t))
      attt <- att[whicht]
      pgt <- pg[whicht] / (sum(pg[whicht]))
      sum(pgt * attt)
    })

    # get standard errors and influence functions
    # for each time specific att
    calendar.se.inner <- lapply(calendar.tlist, function(t1) {
      whicht <- which((t == t1) & (group <= t))
      pgt <- pg[whicht] / (sum(pg[whicht]))
      wif.t <- wif(
        keepers = whicht,
        pg = pg,
        weights.ind = weights.ind,
        G = G,
        group = group
      )
      inf.func.t <- as.numeric(get_agg_inf_func(
        att = att,
        inffunc1 = inffunc1,
        whichones = whicht,
        weights.agg = pgt,
        wif = wif.t
      ))
      se.t <- getSE(inf.func.t, dp)
      list(inf.func = inf.func.t, se = se.t)
    })

    # recover standard errors separately by time
    calendar.se.t <- unlist(BMisc::get_list_element(calendar.se.inner, "se"))
    calendar.se.t[calendar.se.t <= sqrt(.Machine$double.eps) * 10] <- NA
    # recover influence function separately by time
    calendar.inf.func.t <- simplify2array(BMisc::get_list_element(calendar.se.inner, "inf.func"))

    # use multiplier boostrap (across groups) to get critical value
    # for constructing uniform confidence bands
    calendar.crit.val <- stats::qnorm(1 - alp / 2)
    if (dp$cband == TRUE) {
      if (dp$bstrap == FALSE) {
        warning("Used bootstrap procedure to compute simultaneous confidence band")
      }
      calendar.crit.val <- mboot(calendar.inf.func.t, dp, return_V = FALSE)$crit.val

      if (is.na(calendar.crit.val) | is.infinite(calendar.crit.val)) {
        warning("Simultaneous critical value is NA, likely because the standard errors could not be computed. Falling back to pointwise confidence intervals.")
        calendar.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (calendar.crit.val < stats::qnorm(1 - alp / 2)) {
        warning("The simultaneous confidence band is narrower than the pointwise confidence interval, which is unexpected. Falling back to pointwise confidence intervals.")
        calendar.crit.val <- stats::qnorm(1 - alp / 2)
        dp$cband <- FALSE
      }

      if (calendar.crit.val >= 7) {
        warning("Simultaneous critical value is very large, suggesting it may be unreliable. This typically happens when the number of observations per group is small and/or there is not much variation in outcomes. Consider using pointwise confidence intervals instead (set `cband = FALSE`).")
      }
    }

    # get overall att under calendar time effects
    # this is just average over all time periods
    calendar.att <- mean(calendar.att.t)

    # get overall influence function
    calendar.inf.func <- get_agg_inf_func(
      att = calendar.att.t,
      inffunc1 = calendar.inf.func.t,
      whichones = (1:length(calendar.tlist)),
      weights.agg = rep(1 / length(calendar.tlist), length(calendar.tlist)),
      wif = NULL
    )
    calendar.inf.func <- as.numeric(calendar.inf.func)
    # get overall standard error
    calendar.se <- getSE(calendar.inf.func, dp)
    if (!is.na(calendar.se)) {
      if (calendar.se <= sqrt(.Machine$double.eps) * 10) calendar.se <- NA
    }
    return(AGGTEobj(
      overall.att = calendar.att,
      overall.se = calendar.se,
      type = type,
      egt = t2orig(calendar.tlist),
      att.egt = calendar.att.t,
      se.egt = calendar.se.t,
      crit.val.egt = calendar.crit.val,
      inf.function = list(
        calendar.inf.func.t = calendar.inf.func.t,
        calendar.inf.func = calendar.inf.func
      ),
      call = call,
      DIDparams = dp
    ))
  }
}

#-----------------------------------------------------------------------------
# Internal functions for getting standard errors
#-----------------------------------------------------------------------------

#' @title Compute extra term in influence function due to estimating weights
#'
#' @description A function to compute the extra term that shows up in the
#'  influence function for aggregated treatment effect parameters
#'  due to estimating the weights
#'
#' @param keepers a vector of indices for which group-time average
#'  treatment effects are used to compute a particular aggregated parameter
#' @param pg a vector with same length as total number of group-time average
#'  treatment effects that contains the probability of being in particular group
#' @param weights.ind additional sampling weights (nx1)
#' @param G vector containing which group a unit belongs to (nx1)
#' @param group vector of groups
#'
#' @return nxk influence function matrix
#'
#' @keywords internal
wif <- function(keepers, pg, weights.ind, G, group) {
  # note: weights are all of the form P(G=g|cond)/sum_cond(P(G=g|cond))
  # this is equal to P(G=g)/sum_cond(P(G=g)) which simplifies things here

  # The (n x k) matrix [ weights.ind * 1{G == group[k]} - pg[k] ] appears in both
  # the numerator and the denominator terms below; build it once and reuse it.
  # This is identical to the previous code, which constructed it twice via two
  # separate sapply() calls. sum(pg[keepers]) is likewise hoisted out of the loop.
  # No keepers (e.g. an empty post-treatment selection under na.rm or a
  # min_e/max_e window that excludes all cells): return a 0-column influence
  # matrix so the caller's get_agg_inf_func() emits its clean "no valid
  # estimates" error instead of `list() / 0` throwing a cryptic
  # "non-numeric argument to binary operator" here.
  if (length(keepers) == 0L) {
    return(matrix(numeric(0), nrow = length(weights.ind), ncol = 0L))
  }
  Spg <- sum(pg[keepers])
  centered <- sapply(keepers, function(k) {
    weights.ind * 1 * BMisc::TorF(G == group[k]) - pg[k]
  })

  # effect of estimating weights in the numerator
  if1 <- centered / Spg
  # effect of estimating weights in the denominator
  if2 <- base::rowSums(centered) %*% t(pg[keepers] / (Spg^2))

  # return the influence function for the weights
  if1 - if2
}


#' @title Get an influence function for particular aggregate parameters
#'
#' @title This is a generic internal function for combining influence
#'  functions across ATT(g,t)'s to return an influence function for
#'  various aggregated treatment effect parameters.
#'
#' @param att vector of group-time average treatment effects
#' @param inffunc1 influence function for all group-time average treatment effects
#'  (matrix)
#' @param whichones which elements of att will be used to compute the aggregated
#'  treatment effect parameter
#' @param weights.agg the weights to apply to each element of att(whichones);
#'  should have the same dimension as att(whichones)
#' @param wif extra influence function term coming from estimating the weights;
#'  should be n x k matrix where k is dimension of whichones
#'
#' @return nx1 influence function
#'
#' @keywords internal
get_agg_inf_func <- function(att, inffunc1, whichones, weights.agg, wif = NULL) {
  # enforce weights are in matrix form
  weights.agg <- as.matrix(weights.agg)

  # check dimensions
  if (length(whichones) == 0) {
    stop("No valid att_gt() estimates found for this aggregation. ",
         "This may happen if all estimates for a particular group or time period are NA.")
  }
  if (max(whichones) > ncol(inffunc1)) {
    stop("Internal error: influence function column indices (", max(whichones),
         ") exceed available columns (", ncol(inffunc1), "). Please report this as a bug.")
  }

  # multiplies influence function times weights and sums to get vector of weighted IF (of length n)
  thisinffunc <- inffunc1[, whichones, drop = FALSE] %*% weights.agg

  # Incorporate influence function of the weights
  if (!is.null(wif)) {
    thisinffunc <- thisinffunc + wif %*% as.matrix(att[whichones])
  }

  # return influence function
  return(thisinffunc)
}


#' @title Take influence function and return standard errors
#'
#' @description Function to take an nx1 influence function and return
#'  a standard error
#'
#' @param thisinffunc An influence function
#' @inheritParams compute.aggte
#'
#' @return scalar standard error
#'
#' @keywords internal
getSE <- function(thisinffunc, DIDparams = NULL) {
  alp <- .05
  bstrap <- FALSE
  n <- length(thisinffunc)
  if (!is.null(DIDparams)) {
    bstrap <- DIDparams$bstrap
    alp <- DIDparams$alp
    cband <- DIDparams$cband
  }

  if (bstrap) {
    bout <- mboot(thisinffunc, DIDparams, return_V = FALSE)
    return(bout$se)
  } else {
    # Analytical cluster-robust SE (no bootstrap): cluster sums of the aggregated influence function,
    # matching the cluster-sum aggregation of the multiplier bootstrap (CS 2021, Remark 10). Falls back to
    # the i.i.d. form when there is no cluster variable beyond idname or cluster identifiers are unavailable.
    # dp$cluster_vector is set by att_gt() for both faster_mode = TRUE and FALSE.
    cv <- if (!is.null(DIDparams)) DIDparams$cluster_vector else NULL
    extra_cl <- if (!is.null(DIDparams)) DIDparams$clustervars[!(DIDparams$clustervars %in% c(DIDparams$idname, ""))] else character(0)
    if (length(extra_cl) > 0 && !is.null(cv) && length(cv) == length(thisinffunc)) {
      S <- rowsum(thisinffunc, cv)
      return(sqrt(sum(S^2)) / n)
    }
    return(sqrt(mean((thisinffunc)^2) / n))
  }
}

#' @title Process Results from [compute.att_gt()]
#'
#' @param attgt.list list of results from [compute.att_gt()]
#'
#' @return list with elements:
#' \item{group}{which group a set of results belongs to}
#' \item{tt}{which time period a set of results belongs to}
#' \item{att}{the group time average treatment effect}
#'
#' @export
process_attgt <- function(attgt.list) {
  if (!is.list(attgt.list) || length(attgt.list) == 0L) {
    stop("attgt.list must be a non-empty list of group-time result objects.")
  }
  get_cell_value <- function(x, field, allow_na = FALSE) {
    if (!is.list(x)) {
      stop("Each attgt.list element must be a list-like group-time result object.")
    }
    value <- x[[field]]
    if (allow_na && length(value) == 1L && is.na(value)) {
      return(as.numeric(value))
    }
    if (!is.numeric(value) || length(value) != 1L ||
        (!allow_na && is.na(value))) {
      stop("Each attgt.list element must contain a numeric scalar '", field, "'.")
    }
    value
  }
  group <- vapply(attgt.list, get_cell_value, numeric(1), field = "group")
  att   <- vapply(attgt.list, get_cell_value, numeric(1), field = "att", allow_na = TRUE)
  tt    <- vapply(attgt.list, get_cell_value, numeric(1), field = "year")

  list(group=group, att=att, tt=tt)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results="hide", warning=FALSE, message=FALSE-----------------
library(did)
# Source the current version of the did package (based on our Dropbox)
# fldr <- here::here("R/")
# sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
source(here::here("vignettes/setup_sims.R"))

## ----echo=FALSE---------------------------------------------------------------
time.periods <- 4
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)

## -----------------------------------------------------------------------------
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()

head(data)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

head(data)

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)

# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways",
          index = c("id", "period"))

summary(es)

#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs=coefs, ses=ses, exposure=exposure)

library(ggplot2)

ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-time average treatment effects
did_att_gt <- att_gt(yname = "Y",
                     tname = "period",
                     idname = "id",
                     gname = "G",
                     data = data,
                     bstrap = FALSE,
                     cband = FALSE)
summary(did_att_gt)

# plot them
ggdid(did_att_gt)


## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate them into event study plot
did_es <- aggte(did_att_gt, type = "dynamic")

# plot the event study
ggdid(did_es)

## ----echo=FALSE---------------------------------------------------------------
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)

## -----------------------------------------------------------------------------
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate selective treatment timing
# (*** this is what is different here ***)
te.bet.ind <- time.periods:1 / (time.periods/2)

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()

## -----------------------------------------------------------------------------
# run through same code as in earlier example...omitted

## ----echo=FALSE---------------------------------------------------------------
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)

## -----------------------------------------------------------------------------
# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways", 
          index = c("id", "period"))

summary(es)

## ----echo=FALSE---------------------------------------------------------------
#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs = coefs, ses = ses, exposure = exposure)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# run through same code as before...omitted

# new event study plot
ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()

## ----fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200----
# estimate group-time average treatment effects
did.att.gt <- att_gt(yname = "Y",
                     tname = "period",
                     idname = "id",
                     gname = "G",
                     data = data
                     )
summary(did.att.gt)

# plot them
ggdid(did.att.gt)

## ----fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# aggregate them into event study plot
did.es <- aggte(did.att.gt, type = "dynamic")

# plot the event study
ggdid(did.es)

## ----eval=FALSE---------------------------------------------------------------
# # not run (this code can be substantially slower)
# reset.sim()
# set.seed(1814)
# nt <- 1000
# nu <- 1000
# cdp <- conditional_did_pretest("Y", "period", "id", "G", xformla = ~X, data = data)
# cdp

#' @title Multiplier Bootstrap
#'
#' @description A function to take an influence function and use the
#'  multiplier bootstrap to compute standard errors and critical values for
#'  uniform confidence bands.
#'
#' @param inf.func an influence function
#' @param DIDparams DIDparams object
#' @param pl whether or not to use parallel processing in the multiplier
#'  bootstrap, default=FALSE
#' @param cores the number of cores to use with parallel processing,
#'  default=1
#' @param return_V whether to compute and return the bootstrap variance
#'  matrix `V`. Default is `TRUE`. Internal callers that only consume
#'  `bres`, `se`, or `crit.val` set this to `FALSE` to skip the computation
#'  (it is the only O(biters x k^2) step in the function).
#'
#' @return list with elements
#' \item{bres}{results from each bootstrap iteration}
#' \item{V}{variance matrix (`NULL` when `return_V = FALSE`)}
#' \item{se}{standard errors}
#' \item{crit.val}{a critical value for computing uniform confidence bands}
#'
#' @export
mboot <- function(inf.func, DIDparams, pl = FALSE, cores = 1, return_V = TRUE) {
  validate_logical_scalar(pl, "pl")
  validate_positive_whole_number(cores, "cores")
  validate_logical_scalar(return_V, "return_V")

  # setup needed variables according to faster_mode; This returns different type of objects
  # depending on whether we are in faster_mode or not that has to be handled in the code below
  idname <- DIDparams$idname
  clustervars <- DIDparams$clustervars
  biters <- DIDparams$biters
  tname <- DIDparams$tname
  alp <- DIDparams$alp
  panel <- DIDparams$panel
  validate_positive_whole_number(biters, "biters")
  validate_alp(alp)
  validate_logical_scalar(panel, "DIDparams$panel")
  true_repeated_cross_sections <- DIDparams$true_repeated_cross_sections
  unbalanced_panel <- DIDparams$allow_unbalanced_panel
  # Reuse the per-unit cluster vector that att_gt() stored in DIDparams when it
  # aligns with the influence-function rows: it was already derived and validated
  # (time invariance) there, so the data load, the aggregate() check, and the
  # unique() derivation below can all be skipped. Fall back to deriving the
  # clusters from the data for (older) DIDparams objects lacking an aligned
  # cluster_vector -- the fallback keeps the time-varying-cluster stop.
  cluster_vector <- DIDparams$cluster_vector
  use_cluster_vector <- !is.null(cluster_vector) &&
    length(cluster_vector) == NROW(inf.func)
  # Avoid full data copy when possible; only needed for clustering without an
  # aligned cluster_vector
  need_data <- length(clustervars) > 0 && !use_cluster_vector
  if (need_data) {
    if (isTRUE(DIDparams$faster_mode) && !is.null(DIDparams$time_invariant_data)) {
      dta <- as.data.frame(DIDparams$time_invariant_data)
    } else {
      data <- as.data.frame(DIDparams$data)
      tlist <- if (isTRUE(DIDparams$faster_mode)) DIDparams$time_periods else sort(unique(data[,tname]))
      if (panel) {
        dta <- data[data[,tname] == tlist[1], ]
      } else {
        dta <- data
      }
    }
    validate_column_names(clustervars, "clustervars", names(dta), allow_null = TRUE)
  }

  # Convert sparse matrix to dense for bootstrap computation
  inf.func <- as.matrix(inf.func)
  if (!is.numeric(inf.func) || nrow(inf.func) < 1L || ncol(inf.func) < 1L) {
    stop("inf.func must be a numeric matrix with at least one row and one column.")
  }

  # set correct number of units
  n <- nrow(inf.func)

  # if include id as variable to cluster on
  # drop it as we do this automatically
  if (!is.null(idname) && idname %in% clustervars) {
    clustervars <- clustervars[-which(clustervars==idname)]
  }

  if(!is.null(clustervars)){
    if(is.numeric(clustervars)){
      stop("'clustervars' must be a character string specifying the name of the clustering variable, not a numeric value.")
    }
  }
  # we can only handle up to 2-way clustering
  # (in principle could do more, but not high priority now)
  if (length(clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  if (length(clustervars) > 0 && !use_cluster_vector) {
    if(!DIDparams$faster_mode){
      # check that cluster variable does not vary over time within unit
      # reuse 'data' already loaded above (need_data is TRUE on this fallback path)
      clust_tv <- aggregate(data[,clustervars], by=list(data[,idname]), function(rr) length(unique(rr))==1)
      if (!all(clust_tv[,2])) {
        stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
      }
    }
  }

  # multiplier bootstrap
  n_clusters <- n
  if (length(clustervars)==0) {
    bres <- sqrt(n) * run_multiplier_bootstrap(inf.func, biters, pl, cores)
  } else {
    # Cluster-robust multiplier bootstrap following Callaway & Sant'Anna (2021, Remark 10): draw
    # cluster-specific multipliers and apply them to the influence function aggregated to cluster
    # *sums*, consistent with the (1/n) empirical average that defines the estimator. For
    # equal-sized clusters this coincides with the previous aggregation; for unbalanced clusters and
    # repeated cross-sections it keeps the bootstrap aligned with the cluster-robust variance.
    if (use_cluster_vector) {
      # att_gt() already built this vector aligned with the inf.func rows (and
      # validated time invariance when doing so); reusing it avoids reloading
      # the data and re-deriving the clusters on every call.
      cluster <- cluster_vector
      n_clusters <- length(unique(cluster))
    } else {
      n_clusters <- length(unique(dta[,clustervars]))
      cluster <- unique(dta[,c(idname,clustervars)])[,2]
    }
    if (length(cluster) != n) {
      stop("cluster vector length must match the number of influence-function rows.")
    }
    cluster_sum_if <- rowsum(inf.func, cluster, reorder=TRUE)
    bres <- sqrt(n_clusters) * run_multiplier_bootstrap(cluster_sum_if, biters, pl, cores)
  }


  # handle vector and matrix case differently, so you get nxk matrix
  # ifelse(class(bres)=="matrix", bres <- t(bres), bres <- as.matrix(bres))

  if (isTRUE(class(bres) == "numeric")) bres <- as.matrix(bres)

  # Non-degenerate dimensions
  # ndg.dim <- (base::colSums(bres) != 0)
  ndg.dim <- (!is.na(colSums(bres))) & (base::colSums(bres^2) > sqrt(.Machine$double.eps)*10)
  # If NA, set it to false
  #ndg.dim[is.na(ndg.dim)] <- FALSE
  bres <- as.matrix(bres[ , ndg.dim])

  # bootstrap variance matrix (this matrix can be defective because of degenerate cases).
  # Computed only on request: no internal caller consumes it, and it is the only
  # O(biters * k^2) term in mboot.
  V <- if (return_V) cov(bres) else NULL
  # bootstrap standard error: IQR-based scale normalized to the SD of a normal.
  # The qnorm difference is a constant (hoisted out of the per-column loop), and
  # the type = 1 quantiles are exactly the order statistics x_(ceil(B * p)), so a
  # single sort per column replaces the quantile() calls bit-identically (the
  # post-sort length is the per-column non-NA count, preserving na.rm = TRUE;
  # sort.int() drops NAs by default (na.last = NA)).
  iqr_norm <- qnorm(.75) - qnorm(.25)
  bSigma <- apply(bres, 2, function(b) {
    b <- sort.int(b)
    nb <- length(b)
    (b[ceiling(.75 * nb)] - b[ceiling(.25 * nb)]) / iqr_norm
  })

  # critical value for uniform confidence band: per bootstrap draw, the max over
  # columns of |b / bSigma|. Vectorized as a column-fold of pmax (equivalent to the
  # previous row-wise apply(max, na.rm = TRUE): NA entries are ignored via -Inf, and
  # an all-NA row yields -Inf, which is dropped by the is.finite filter below).
  if (ncol(bres) == 0L) {
    # All bootstrap dimensions degenerate: every row has no finite t-statistic.
    # This matches the previous apply(max, na.rm = TRUE), which returned -Inf for
    # every row and then dropped them all via the is.finite filter.
    bT <- numeric(0)
  } else {
    scaled <- abs(bres / rep(bSigma, each = nrow(bres)))
    scaled[is.na(scaled)] <- -Inf
    bT <- scaled[, 1L]
    for (j in seq_len(ncol(scaled))[-1L]) bT <- pmax(bT, scaled[, j])
    bT <- bT[is.finite(bT)]
  }
  crit.val <- quantile(bT, 1-alp, type=1, na.rm = TRUE)

  #se <- rep(0, length(ndg.dim))
  se <- rep(NA, length(ndg.dim))
  # Standard error from the bootstrap interquartile-range scale. With cluster sums (above) this is the
  # cluster-robust SE (1/n) * sqrt(sum_c S_c^2); without clustering n_clusters = n and it reduces to the
  # i.i.d. form bSigma / sqrt(n). (Equivalent to the previous bSigma / sqrt(n_clusters) when clusters are
  # equal-sized; the change matters only for unbalanced clusters / repeated cross-sections.)
  se[ndg.dim] <- as.numeric(bSigma) * sqrt(n_clusters) / n
  #se[se==0] <- NA

  list(bres = bres, V = V, se = se, crit.val = crit.val)
}

run_multiplier_bootstrap <- function(inf.func, biters, pl = FALSE, cores = 1) {
  validate_positive_whole_number(biters, "biters")
  validate_logical_scalar(pl, "pl")
  validate_positive_whole_number(cores, "cores")

  # Split biters into per-core chunks that are always non-negative and sum to
  # biters. The previous rep(ceiling(biters/cores), cores) + correction made
  # chunks[1] negative when biters < cores (e.g. biters=2, cores=4 -> [-1,1,1,1]),
  # which crashed BMisc::multiplier_bootstrap(). This split never goes negative
  # and drops empty chunks.
  chunks <- diff(round(seq(0, biters, length.out = cores + 1)))
  chunks <- chunks[chunks > 0]

  n <- nrow(inf.func)
  parallel.function <- function(biters) {
    BMisc::multiplier_bootstrap(inf.func, biters)
  }
  # From tests, this is about where it becomes worth it to parallelize
  if (.Platform$OS.type == "windows" && pl == TRUE && cores > 1) {
    warning("Parallel processing (pl=TRUE) is not supported on Windows. Using sequential processing instead.")
    pl <- FALSE
  }
  if (n > 2500 && pl && cores > 1) {
    # Use parallel-safe RNG streams so a fixed set.seed() makes the parallel
    # bootstrap reproducible run-to-run. Under the default Mersenne-Twister,
    # mclapply()'s forked children re-seed non-deterministically, so the
    # bootstrap draws -- and hence the SE and the uniform-band critical value --
    # would drift between identical-seed runs. Restore the caller's RNGkind on exit.
    old_kind <- RNGkind("L'Ecuyer-CMRG")
    on.exit(RNGkind(old_kind[1L], old_kind[2L], old_kind[3L]), add = TRUE)
    results <- parallel::mclapply(
      chunks,
      FUN = parallel.function,
      mc.cores = min(cores, length(chunks))   # no more workers than non-empty chunks
    )
    results <- do.call(rbind, results)
  } else {
    results <- BMisc::multiplier_bootstrap(inf.func, biters)
  }
  return(results)
}

#' tidy results
#' @export
generics::tidy

#' tidy results
#' @export
generics::glance

#' Number of unique cross-sectional units in an MP object
#'
#' @importFrom stats nobs
#' @param object a model of class MP produced by the [att_gt()] function
#' @param ... additional arguments (ignored)
#' @return Integer. The number of unique cross-sectional units in the data.
#' @export
nobs.MP <- function(object, ...) {
  as.integer(object$n)
}

#' Number of unique cross-sectional units in an AGGTEobj object
#'
#' @importFrom stats nobs
#' @param object a model of class AGGTEobj produced by the [aggte()] function
#' @param ... additional arguments (ignored)
#' @return Integer. The number of unique cross-sectional units in the data.
#' @export
nobs.AGGTEobj <- function(object, ...) {
  if (isTRUE(object$DIDparams$faster_mode)) {
    as.integer(object$DIDparams$id_count)
  } else {
    as.integer(object$DIDparams$n)
  }
}

#' Tidy an MP object into a data frame
#'
#' Returns a tidy data frame of group-time average treatment effect estimates
#' from an [att_gt()] result.
#'
#' @importFrom generics tidy
#' @param x a model of class MP produced by the [att_gt()] function
#' @param ... Additional arguments to tidying method.
#'
#' @return A data frame with one row per ATT(g,t) estimate and columns:
#' \describe{
#'   \item{term}{ATT(g,t) label}
#'   \item{group}{the treatment cohort g}
#'   \item{time}{the time period t}
#'   \item{estimate}{the ATT(g,t) point estimate}
#'   \item{std.error}{standard error}
#'   \item{statistic}{t-statistic (\code{estimate / std.error})}
#'   \item{p.value}{two-sided pointwise p-value
#'     (\code{2 * (1 - pnorm(abs(statistic)))}).
#'     Marginal per-estimate; does \strong{not} account for multiple testing
#'     across ATT(g,t) cells.}
#'   \item{conf.low, conf.high}{simultaneous confidence band limits, using the
#'     bootstrap uniform critical value when \code{bstrap=TRUE} and
#'     \code{cband=TRUE}, otherwise pointwise}
#'   \item{point.conf.low, point.conf.high}{pointwise confidence interval limits
#'     using \code{qnorm(1 - alp/2)}, always}
#' }
#'
#' @export
tidy.MP <- function(x, ...) {
  out <- data.frame(
    term      = paste0('ATT(',x$group,",",x$t, ")"),
    group     = x$group,
    time      = x$t,
    estimate  = x$att,
    std.error = x$se,
    statistic = x$att / x$se,
    p.value   = 2 * (1 - stats::pnorm(abs(x$att / x$se))),
    conf.low  = x$att - x$c * x$se,
    conf.high = x$att + x$c * x$se,
    point.conf.low  = x$att - stats::qnorm(1 - x$alp/2) * x$se,
    point.conf.high = x$att + stats::qnorm(1 - x$alp/2) * x$se)
  out
}

#' glance model characteristics from MP objects
#'
#' @importFrom generics glance
#' @param x a model of class MP produced by the [att_gt()] function
#' @param ... other arguments passed to methods
#' @export
glance.MP <- function(x, ...) {
  dp <- x$DIDparams
  est.method <- if (is.function(dp$est_method)) "custom" else dp$est_method
  if (isTRUE(dp$faster_mode)) {
    out <- data.frame(
      nobs          = x$n,
      ngroup        = dp$treated_groups_count,
      ntime         = dp$time_periods_count,
      control.group = dp$control_group,
      est.method    = est.method)
  } else {
    out <- data.frame(
      nobs          = x$n,
      ngroup        = dp$nG,
      ntime         = dp$nT,
      control.group = dp$control_group,
      est.method    = est.method)
  }
  out
}

#' Tidy an AGGTEobj into a data frame
#'
#' Returns a tidy data frame of aggregated treatment effect estimates from an
#' [aggte()] result.
#'
#' @importFrom generics tidy
#' @param x a model of class AGGTEobj produced by the [aggte()] function
#' @param ... Additional arguments to tidying method.
#'
#' @return A data frame whose columns depend on \code{type}:
#' \describe{
#'   \item{type}{the aggregation type: \code{"simple"}, \code{"dynamic"},
#'     \code{"group"}, or \code{"calendar"}}
#'   \item{term}{label for each estimate}
#'   \item{estimate}{point estimate}
#'   \item{std.error}{standard error}
#'   \item{statistic}{t-statistic (\code{estimate / std.error})}
#'   \item{p.value}{two-sided pointwise p-value
#'     (\code{2 * (1 - pnorm(abs(statistic)))}).
#'     Marginal per-estimate; does \strong{not} account for multiple testing
#'     across event times or groups.}
#'   \item{conf.low, conf.high}{simultaneous confidence band limits.  When
#'     \code{bstrap=TRUE} and \code{cband=TRUE} these use the bootstrap uniform
#'     critical value (\code{crit.val.egt}); otherwise they equal the pointwise
#'     intervals.  For \code{type="simple"} and the overall average row of
#'     \code{type="group"}, a single scalar is returned so simultaneous and
#'     pointwise coincide.}
#'   \item{point.conf.low, point.conf.high}{pointwise confidence interval limits
#'     always using \code{qnorm(1 - alp/2)}.}
#' }
#'
#' @details
#' The key distinction between \code{conf.low}/\code{conf.high} and
#' \code{point.conf.low}/\code{point.conf.high} is that the former accounts for
#' multiple testing across all estimates (simultaneous coverage), while the
#' latter provides marginal (per-estimate) coverage only.  Use the simultaneous
#' bands when you want to make joint inferences across all event times or groups.
#'
#' @export
tidy.AGGTEobj<- function(x, ...) {
  if(x$type == "dynamic"){
    out <- data.frame(
      type          = x$type,
      term = paste0('ATT(', x$egt, ")"),
      event.time= x$egt,
      estimate  = x$att.egt,
      std.error = x$se.egt,
      statistic = x$att.egt / x$se.egt,
      p.value   = 2 * (1 - stats::pnorm(abs(x$att.egt / x$se.egt))),
      conf.low  = x$att.egt - x$crit.val.egt * x$se.egt,
      conf.high = x$att.egt + x$crit.val.egt * x$se.egt,
      point.conf.low  = x$att.egt - stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt,
      point.conf.high = x$att.egt + stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt)
  }
  if(x$type == "group"){
    out <- data.frame(
      type     = x$type,
      term = c(paste0('ATT(Average)'), paste0('ATT(', x$egt, ")")),
      group    = c('Average', x$egt),
      estimate  = c(x$overall.att, x$att.egt),
      std.error = c(x$overall.se, x$se.egt),
      statistic = c(x$overall.att, x$att.egt) / c(x$overall.se, x$se.egt),
      p.value   = 2 * (1 - stats::pnorm(abs(c(x$overall.att, x$att.egt) / c(x$overall.se, x$se.egt)))),
      conf.low  = c(x$overall.att - stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se, x$att.egt - x$crit.val.egt * x$se.egt),
      conf.high = c(x$overall.att + stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se, x$att.egt + x$crit.val.egt * x$se.egt),
      point.conf.low  = c(x$overall.att - stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se, x$att.egt - stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt),
      point.conf.high = c(x$overall.att + stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se,x$att.egt + stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt))
     }
  
  if(x$type == "calendar"){
    out <- data.frame(
      type      = x$type,
      time      = x$egt,
      term = paste0('ATT(', x$egt, ")"),
      estimate  = x$att.egt,
      std.error = x$se.egt,
      statistic = x$att.egt / x$se.egt,
      p.value   = 2 * (1 - stats::pnorm(abs(x$att.egt / x$se.egt))),
      conf.low  = x$att.egt - x$crit.val.egt * x$se.egt,
      conf.high = x$att.egt + x$crit.val.egt * x$se.egt,
      point.conf.low  = x$att.egt - stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt,
      point.conf.high = x$att.egt + stats::qnorm(1 - x$DIDparams$alp/2) * x$se.egt)
  }

  if(x$type == "simple"){
    out <- data.frame(
      type      = x$type,
      term      = 'ATT(simple average)',
      estimate  = x$overall.att,
      std.error = x$overall.se,
      statistic = x$overall.att / x$overall.se,
      p.value   = 2 * (1 - stats::pnorm(abs(x$overall.att / x$overall.se))),
      conf.low  = x$overall.att - stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se,
      conf.high = x$overall.att + stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se,
      point.conf.low  = x$overall.att - stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se,
      point.conf.high = x$overall.att + stats::qnorm(1 - x$DIDparams$alp/2) * x$overall.se)
  }
  
  out
}

#' glance model characteristics from AGGTEobj objects
#'
#' @importFrom generics glance
#' @param x a model of class AGGTEobj produced by the [aggte()] function
#' @param ... other arguments passed to methods
#' @export
glance.AGGTEobj<- function(x, ...) {
  est.method <- if (is.function(x$DIDparams$est_method)) "custom" else x$DIDparams$est_method
  if(isTRUE(x$DIDparams$faster_mode)) {
    out <- data.frame(
      type          = x$type,
      nobs          = x$DIDparams$id_count,
      ngroup        = x$DIDparams$treated_groups_count,
      ntime         = x$DIDparams$time_periods_count,
      control.group = x$DIDparams$control_group,
      est.method    = est.method)
  } else {
    out <- data.frame(
      type          = x$type,
      nobs          = x$DIDparams$n,
      ngroup        = x$DIDparams$nG,
      ntime         = x$DIDparams$nT,
      control.group = x$DIDparams$control_group,
      est.method    = est.method)
  }

  out
}

#' @title County Teen Employment Dataset
#'
#' @description A dataset containing (the log of) teen employment in 500 counties
#'  in the U.S. from 2003 to 2007.  This is a subset of the dataset used in Callaway and
#'  Sant'Anna (2021).  See that paper for additional descriptions.
#'
#' @format A data frame with 2500 rows and 6 variables:
#' \describe{
#'   \item{year}{the year of the observation}
#'   \item{countyreal}{a unique identifier for a particular county}
#'   \item{lpop}{the log of 1000s of population for the county}
#'   \item{lemp}{the log of teen employment in the county}
#'   \item{first.treat}{the year that the state where the county is located
#'    raised its minimum wage, it is set equal to 0 for counties that have
#'    minimum wages equal to the federal minimum wage over the entire
#'    period.}
#'   \item{treat}{whether or not a particular county is ever treated
#'    during the sample period.  It is time-invariant within each county
#'    and equal to 1 exactly when first.treat is positive; it is not a
#'    year-specific treatment indicator.}
#' }
#' @source Callaway and Sant'Anna (2021)
"mpdta"

#' @title DIDparams
#'
#' @description Object to hold DiD parameters that are passed across functions
#'
#' @inheritParams att_gt2
#' @inheritParams pre_process_did2
#' @param did_tensor list of outcome tensors that are used in the estimation
#' @param args list of arguments that are used in the estimation
#' @noRd
DIDparams2 <- function(did_tensors, args, call=NULL) {
  # get the arguments from args
  yname <- args$yname
  tname <- args$tname
  idname <- args$idname
  gname <- args$gname
  xformla <- args$xformla # formula of covariates
  panel <- args$panel
  est_method <- args$est_method
  bstrap <- args$bstrap
  biters <- args$biters
  cband <- args$cband
  anticipation <- args$anticipation
  control_group <- args$control_group
  allow_unbalanced_panel <- args$allow_unbalanced_panel
  weightsname <- args$weightsname
  base_period <- args$base_period
  clustervars <- args$clustervars
  cores <- args$cores
  pl <- args$pl
  print_details <- args$print_details
  faster_mode <- args$faster_mode
  alp <- args$alp
  true_repeated_cross_sections <- args$true_repeated_cross_sections
  time_periods_count <- args$time_periods_count
  time_periods <- args$time_periods
  treated_groups_count <- args$treated_groups_count
  treated_groups <- args$treated_groups
  id_count <- args$id_count

  # get the arguments from did_tensors
  outcomes_tensor <- did_tensors$outcomes_tensor
  data <- did_tensors$data
  time_invariant_data <- did_tensors$time_invariant_data
  cohort_counts <- did_tensors$cohort_counts
  period_counts <- did_tensors$period_counts
  crosstable_counts <- did_tensors$crosstable_counts
  # covariates <- did_tensors$covariates # matrix of covariates
  covariates_tensor <- did_tensors$covariates_tensor # matrix of covariates
  covariates_matrix <- did_tensors$covariates_matrix
  cluster_vector <- did_tensors$cluster
  weights_vector <- did_tensors$weights
  weights_tensor <- did_tensors$weights_tensor
  fix_weights <- args$fix_weights


  out <- list(yname=yname,
              tname=tname,
              idname=idname,
              gname=gname,
              xformla=xformla,
              panel=panel,
              est_method=est_method,
              bstrap=bstrap,
              biters=biters,
              cband=cband,
              anticipation=anticipation,
              control_group=control_group,
              allow_unbalanced_panel=allow_unbalanced_panel,
              weightsname=weightsname,
              base_period=base_period,
              clustervars=clustervars,
              cores=cores,
              pl = pl,
              print_details=print_details,
              faster_mode=faster_mode,
              alp=alp,
              true_repeated_cross_sections=true_repeated_cross_sections,
              time_periods_count=time_periods_count,
              time_periods=time_periods,
              treated_groups_count=treated_groups_count,
              treated_groups=treated_groups,
              id_count=id_count,
              outcomes_tensor=outcomes_tensor,
              data=data,
              time_invariant_data=time_invariant_data,
              cohort_counts=cohort_counts,
              period_counts=period_counts,
              crosstable_counts=crosstable_counts,
              # covariates=covariates,
              covariates_tensor = covariates_tensor,
              covariates_matrix = covariates_matrix,
              cluster_vector=cluster_vector,
              weights_vector=weights_vector,
              weights_tensor=weights_tensor,
              fix_weights=fix_weights,
              call=call)
  class(out) <- "DIDparams"
  return(out)
}

#' @title reset.sim
#' @description a function to create a "reasonable" set of parameters
#'  to create simulated panel data that obeys a parallel trends assumption.
#'  In particular, it provides parameters where the the effect of participating
#'  in the treatment is equal to one in all post-treatment time periods.
#' 
#'  After calling this function, the user can change particular values of the
#'  parameters in order to generate dynamics, heterogeneous effects across
#'  groups, etc.
#'
#' @param time.periods The number of time periods to include
#' @param n The total number of observations
#' @param ipw If TRUE, sets parameters so that DGP is
#'  compatible with recovering ATT(g,t)'s using IPW (i.e.,
#'  where logit that just includes a linear term in X works).  If
#'  FALSE, sets parameters that will be incompatible with IPW.
#'  Either way, these parameters can be specified by the user
#'  if so desired.
#' @param reg If TRUE, sets parameters so that DGP is compatible
#'  with recovering ATT(g,t)'s using regressions on untreated
#'  untreated potential outcomes.  If FALSE, sets parameters that
#'  will be incompatible with using regressions (i.e., regressions
#'  that include only linear term in X).  Either way, these
#'  parameters can be specified by the user if so desired.
#'
#' @return list of simulation parameters
#' 
#' @export
reset.sim <- function(time.periods=4, n=5000, ipw=TRUE, reg=TRUE) {
  #-----------------------------------------------------------------------------
  # set parameters
  #-----------------------------------------------------------------------------
  # coefficient on X 
  bett <- seq(1:time.periods)
  # time fixed effect
  thet <- seq(1:time.periods)
  # time fixed effect
  theu <- thet # changing this creates violations of parallel trends
  # covariate effect
  betu <- bett # changing this creates violations of conditional parallel trends
  #-----------------------------------------------------------------------------
  # parameters for treated potential outcomes
  #-----------------------------------------------------------------------------
  te.bet.ind <- rep(1,time.periods) # no selective treatment timing
  te.bet.X <- bett #no heterogeneous effects by X
  te.t <- thet # no calendar time effects
  te.e <- rep(0,time.periods) # no dynamic effects
  te <- 1 # overall basic effect
  # parameters in generalized propensity score
  # don't make them too big otherwise can get divide by 0
  gamG <- c(0,1:time.periods)/(2*time.periods)

  # return list of parameters
  list(time.periods=time.periods,
       bett=bett,
       thet=thet,
       theu=theu,
       betu=betu,
       te.bet.ind=te.bet.ind,
       te.bet.X=te.bet.X,
       te.t=te.t,
       te.e=te.e,
       te=te,
       n=n,
       gamG=gamG,
       ipw=ipw,
       reg=reg
       )
}

#' @title build_sim_dataset
#'
#' @description A function for building simulated data
#'
#' @param sp_list A list of simulation parameters.  See `reset.sim` to generate
#'  some default values for parameters
#' @param panel whether to construct panel data (the default) or repeated
#'  cross sections data
#'
#' @return a data.frame with the following columns
#'   \itemize{
#'     \item G observations group
#'     \item X value of covariate
#'     \item id observation's id
#'     \item cluster observation's cluster (by construction there is no within-cluster correlation)
#'     \item period time period for current observation
#'     \item Y outcome
#'     \item treat whether or not this unit is ever treated
#'   }
#'
#' @export
build_sim_dataset <- function(sp_list, panel=TRUE) {
  #-----------------------------------------------------------------------------
  # build dataset
  #-----------------------------------------------------------------------------
  time.periods <- sp_list$time.periods
  nt <- sp_list$nt
  bett <- sp_list$bett
  thet=sp_list$thet
  nu <- sp_list$nu
  theu <- sp_list$theu
  betu <- sp_list$betu
  te.bet.ind <- sp_list$te.bet.ind
  te.bet.X <- sp_list$te.bet.X
  te.t <- sp_list$te.t
  te.e <- sp_list$te.e
  te <- sp_list$te
  n <- sp_list$n
  gamG <- sp_list$gamG
  ipw <- sp_list$ipw
  reg <- sp_list$reg

  X <- rnorm(n)

  # Build the exp(outer(...)) propensity term once and reuse it for the row-sum
  # normalization (it was previously computed twice). apply(., 1, sum) is kept
  # (rather than rowSums) so `pr` is bit-identical to before -- this matters because
  # the sample() draw for G below depends on the exact values of `pr`.
  if (ipw) {
    e <- exp(outer(X, gamG))
  } else {
    e <- exp(outer((pnorm(X) + 0.5)^2, gamG))
  }
  pr <- e / apply(e, 1, sum)

  G <- apply(pr, 1, function(pvec) sample(seq(0,time.periods), size=1, prob=pvec))

  Gt <- G[G>0]
  nt <- length(Gt)

  if (reg) {
    Xmodel <- X
  } else {
    Xmodel <- X^2
  }

  Xt <- Xmodel[G>0]

  # draw individual fixed effect
  Ct <- rnorm(nt, mean=G)

  # generate untreated potential outcomes in each time period
  Ynames <- paste0("Y",1:time.periods)
  #Ynames <- paste0(1:time.periods)
  Y0tmat <- sapply(1:time.periods, function(t) {
    thet[t] + Ct + Xt*bett[t] + rnorm(nt)
  })
  Y0tdf <- as.data.frame(Y0tmat)
  
  # generate treated potential outcomes
  # te.e[pmax(t-Gt+1, 1)] is the event-time treatment effect: event time
  # t - Gt is stored 1-based, so te.e[1] is the on-impact (t = Gt) effect.
  # pmax() clamps the index at 1 so it stays valid in pre-treatment periods
  # (t < Gt), where the (Gt <= t) factor zeroes the term anyway.
  Y1tdf <- sapply(1:time.periods, function(t) {
    te.t[t] + te.bet.ind[Gt]*Ct + Xt*te.bet.X[t] + (Gt <= t)*te.e[pmax(t-Gt+1, 1)] + te + rnorm(nt)
  })

  # generate observed data
  Ytdf <- sapply(1:time.periods, function(t) {
    (Gt<=t)*Y1tdf[,t] + (Gt>t)*Y0tdf[,t]
  })
  colnames(Ytdf) <- Ynames

  # store observed data for treated group
  dft <- cbind.data.frame(G=Gt,X=X[G>0],Ytdf)

  # untreated units

  # draw untreated covariate
  nu <- sum(G==0)
  Xu <- Xmodel[G==0]

  # draw untreated fixed effect
  Cu <- rnorm(nu, mean=0)


  # generate untreated potential outcomes
  Y0umat <- sapply(1:time.periods, function(t) {
    theu[t] + Cu + rnorm(nu) + Xu*betu[t]
  })
  Y0udf <- as.data.frame(Y0umat)
  colnames(Y0udf) <- Ynames

  # store dataset of observed outcomes for untreated units
  dfu <- cbind.data.frame(G=0,X=X[G==0],Y0udf)

  # store overall dataset
  df <- rbind.data.frame(dft, dfu)

  # generate id variable
  df$id <- 1:nrow(df)
  # generate clusters (there's no actual within-cluster correlation)
  df$cluster <- sample(1:50, size=nrow(df), replace=TRUE)

  # convert data from wide to long format
  ddf <- tidyr::pivot_longer(df,
                             cols=tidyr::starts_with("Y"),
                             names_to="period",
                             names_prefix="Y",
                             values_to="Y")
  ddf$period <- as.numeric(ddf$period)
  ddf$treat <- 1*(ddf$G > 0)
  ddf <- ddf[order(ddf$id, ddf$period),] # reorder data

  if (!panel) { # repeated cross sections
    n <- nt+nu
    Time <- sample(1:time.periods, size=n, replace=TRUE, prob=rep(1/time.periods, time.periods))
    right.row <- sapply( unique(ddf$id), function(i) {
      which(ddf$id==i & ddf$period==Time[i])
    })
    ddf <- ddf[right.row,]
  }

  ddf <- subset(ddf, G != 1)
  ddf
}


#' @title sim
#' @description An internal function that builds simulated data, computes
#'  ATT(g,t)'s and some aggregations.  It is useful for testing the inference
#'  procedures in the `did` function.
#'
#' @inheritParams reset.sim
#' @inheritParams build_sim_dataset
#'
#' @param ret which type of results to return.  The options are `Wpval` (returns
#'  1 if the p-value from a Wald test that all pre-treatment ATT(g,t)'s are equal
#'  is less than .05),
#'  `cband`  (returns 1 if a uniform confidence band covers 0 for groups and times),
#'  `simple` (returns 1 if, using the simple treatment effect aggregation results
#'  in rejecting that this aggregated treatment effect parameter is equal to 0),
#'  `dynamic` (returns 1 if the uniform confidence band from the dynamic treatment
#'  effect aggregation covers 0 in all pre- and post-treatment periods).  The default
#'  value is NULL, and in this case the function will just return the results from
#'  the call to `att_gt`.
#' @param bstrap whether or not to use the bootstrap to conduct inference (default is TRUE)
#' @param cband whether or not to compute uniform confidence bands in the call to `att_gt`
#'  (the default is TRUE)
#' @param control_group Whether to use the "nevertreated" comparison group (the default)
#'  or the "notyettreated" as the comparison group
#' @param xformla Formula for covariates in `att_gt` (default is `~X`)
#' @param est_method Which estimation method to use in `att_gt` (default is "dr")
#' @param clustervars Any additional variables which should be clustered on
#' @param panel whether to simulate panel data (the default) or otherwise repeated
#'  cross sections data
#' 
#' @return When `ret=NULL`, returns the results of the call to `att_gt`, otherwise returns
#'  1 if the specified test rejects or 0 if not.
#' 
#' @export
sim <- function(sp_list,
                ret=NULL,
                bstrap=TRUE,
                cband=TRUE,
                control_group="nevertreated",
                xformla=~X,
                est_method="dr",
                clustervars=NULL,
                panel=TRUE) {

  ddf <- build_sim_dataset(sp_list=sp_list,
                           panel=panel)

  time.periods <- sp_list$time.periods
  te.e <- sp_list$te.e
  te <- sp_list$te
  
  # get results
  res <- att_gt(yname="Y", xformla=xformla, data=ddf, tname="period", idname="id",
                gname="G", 
                bstrap=bstrap, cband=cband, control_group=control_group,
                est_method=est_method,
                clustervars=clustervars,
                panel=panel)


  if (is.null(ret)) {
    return(res)
  } else if (ret=="Wpval") {
    rej <- 1*(res$Wpval < .05)
    return(rej)
  } else if (ret=="cband") {
    cu <- res$att + res$c*res$se 
    cl <- res$att - res$c*res$se
    covers0 <- 1*(all( (cu > 0) & (cl < 0)))
    return(covers0)
  } else if (ret=="simple") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else if (ret=="dynamic") {
    truth <- c(rep(0,(time.periods-2)),te+te.e[1:(time.periods-1)])
    agg <- aggte(res, type="dynamic")
    cu <- agg$att.egt + agg$crit.val.egt * agg$se.egt
    cl <- agg$att.egt - agg$crit.val.egt * agg$se.egt
    coverstruth <- 1*(all( (cu > truth) & (cl < truth)))
    return(coverstruth)
  } else if (ret=="notyettreated") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else {
    return(res)
  }

}

## pretest_sim <- function(ret=NULL, bstrap=FALSE, cband=FALSE,
##                         control.group="nevertreated", panel=TRUE, xformla=~X, cores=1) {

##   ddf <- build_ipw_dataset(panel=panel)

##   # get results
##   res <- conditional_did_pretest(yname="Y", xformla=xformla, data=ddf,
##                                  tname="period", idname="id",
##                                  first.treat.name="G", estMethod="ipw",
##                                  printdetails=FALSE,
##                                  bstrap=bstrap, cband=cband,
##                                  control.group=control.group,
##                                  panel=panel, 
##                                  pl=TRUE, cores=cores)

##   res$CvMpval
## }

# =============================================================================
# Robustness guards: malformed inputs must be rejected with a clear, identical
# message in BOTH the fast (faster_mode = TRUE) and slow (faster_mode = FALSE)
# paths. Regression tests for two confirmed silent-failure bugs:
#   - duplicate (id, period) rows (slow path used to return wrong numbers)
#   - negative / zero-mean weights (used to silently produce NA)
# =============================================================================

test_that("duplicate (id, period) rows are rejected identically by both modes", {
  set.seed(1)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  ddup <- rbind(d, d[d$id == d$id[1] & d$period == 2, ])  # one duplicated (id, period)
  msg <- "must be unique \\(by tname\\)"
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
    tname = "period", idname = "id", gname = "G", faster_mode = TRUE)), msg)
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
    tname = "period", idname = "id", gname = "G", faster_mode = FALSE)), msg)
  # The guard is intentionally unconditional on panel: with idname supplied it
  # also rejects duplicated (id, period) rows on repeated cross sections and
  # unbalanced panels (the slow path used to silently accept them there).
  for (fm in c(TRUE, FALSE)) {
    expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      faster_mode = fm)), msg)
    expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = ddup,
      tname = "period", idname = "id", gname = "G", allow_unbalanced_panel = TRUE,
      faster_mode = fm)), msg)
  }
})

test_that("negative / zero-mean weights are rejected identically by both modes", {
  set.seed(1)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$wgt <- ifelse(d$id %% 5 == 0, -1, 1)   # some negative weights
  msg <- "must be non-negative with a positive mean"
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wgt", faster_mode = TRUE)), msg)
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wgt", faster_mode = FALSE)), msg)

  d$wzero <- 0   # all-zero -> non-positive mean
  expect_error(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", weightsname = "wzero")), msg)
})

test_that("valid positive weights are still accepted in both modes", {
  set.seed(2)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$wgt <- runif(nrow(d), 0.5, 2)
  for (fm in c(TRUE, FALSE)) {
    r <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", weightsname = "wgt",
      faster_mode = fm, bstrap = FALSE)))
    expect_s3_class(r, "MP")
    expect_true(any(is.finite(r$att)))
  }
})

test_that("fast RC path returns NA instead of crashing when a cell has no treated post observations", {
  set.seed(43)
  ids <- 1:90
  d <- expand.grid(id = ids, period = 1:4)
  d$G <- ifelse(d$id <= 40, 0, ifelse(d$id <= 65, 2, 3))
  d <- d[!(d$G == 2 & d$period >= 3), ]
  d$X <- rnorm(nrow(d))
  d$Y <- 0.5 * d$X + 0.2 * d$period + (d$G > 0 & d$period >= d$G) +
    rnorm(nrow(d), 0, 0.2)

  expect_warning(
    res <- suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      faster_mode = TRUE, est_method = "dr", bstrap = FALSE)),
    "No units in group 2 in time period 3")
  expect_s3_class(res, "MP")
  expect_true(any(is.na(res$att[res$group == 2])))
})

test_that("never-treated small-group stop fires in both modes when multiple groups are small", {
  set.seed(44)
  d <- expand.grid(id = 1:6, period = 1:3)
  d$G <- ifelse(d$id <= 3, 0, 2)
  d$Y <- rnorm(nrow(d))
  for (j in 1:6) d[[paste0("X", j)]] <- rnorm(nrow(d))
  xf <- ~X1 + X2 + X3 + X4 + X5 + X6
  msg <- "never-treated group is too small"

  expect_error(suppressWarnings(att_gt(yname = "Y", xformla = xf, data = d,
    tname = "period", idname = "id", gname = "G", faster_mode = FALSE,
    bstrap = FALSE)), msg)
  expect_error(suppressWarnings(att_gt(yname = "Y", xformla = xf, data = d,
    tname = "period", idname = "id", gname = "G", faster_mode = TRUE,
    bstrap = FALSE)), msg)
})

test_that("slow unbalanced influence aggregation handles fractional numeric ids", {
  set.seed(45)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d <- d[-sample(seq_len(nrow(d)), floor(0.05 * nrow(d))), ]
  d$id <- d$id + 0.123456789012345

  slow <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    allow_unbalanced_panel = TRUE, faster_mode = FALSE, bstrap = FALSE)))
  fast <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    allow_unbalanced_panel = TRUE, faster_mode = TRUE, bstrap = FALSE)))

  expect_equal(slow$att, fast$att, tolerance = 1e-10)
  expect_equal(slow$se, fast$se, tolerance = 1e-8)
})

test_that("fast path preserves user columns named weights", {
  set.seed(46)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)

  d$weights <- d$Y
  slow_y <- suppressWarnings(suppressMessages(att_gt(yname = "weights", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = FALSE, bstrap = FALSE)))
  fast_y <- suppressWarnings(suppressMessages(att_gt(yname = "weights", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = TRUE, bstrap = FALSE)))
  expect_equal(slow_y$att, fast_y$att, tolerance = 1e-10)
  expect_false(all(abs(fast_y$att) < 1e-12))

  d$weights <- d$X
  slow_x <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~weights,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = FALSE, bstrap = FALSE)))
  fast_x <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~weights,
    data = d, tname = "period", idname = "id", gname = "G",
    faster_mode = TRUE, bstrap = FALSE)))
  expect_equal(slow_x$att, fast_x$att, tolerance = 1e-10)
  expect_equal(slow_x$se, fast_x$se, tolerance = 1e-8)
})

test_that("slow RC path NA-cells a throwing preliminary logit instead of aborting att_gt", {
  # Regression test: a -Inf covariate (log(0), reachable via transform-formula
  # support) makes overlap_logit_fit() throw. The slow RC branch used to run the
  # overlap/rcond guards OUTSIDE the per-cell tryCatch, hard-aborting the whole
  # att_gt() call while the fast path and the slow panel path degraded to NA
  # cells with a warning. Both modes must now fail identically, cell by cell.
  set.seed(20260609)
  sp <- did::reset.sim(time.periods = 4, n = 400)
  d <- did::build_sim_dataset(sp)
  d$Xpos <- exp(d$X)
  d$Xpos[d$id == unique(d$id)[1]] <- 0  # log(0) = -Inf for one unit

  w_slow <- capture_warnings(
    slow <- suppressMessages(att_gt(yname = "Y", xformla = ~log(Xpos), data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      est_method = "dr", faster_mode = FALSE, bstrap = FALSE)))
  w_fast <- capture_warnings(
    fast <- suppressMessages(att_gt(yname = "Y", xformla = ~log(Xpos), data = d,
      tname = "period", idname = "id", gname = "G", panel = FALSE,
      est_method = "dr", faster_mode = TRUE, bstrap = FALSE)))

  expect_true(any(grepl("Error computing internal 2x2 DiD", w_slow)))
  expect_true(any(grepl("Error computing internal 2x2 DiD", w_fast)))
  expect_true(any(is.na(slow$att)))      # affected cells degrade to NA
  expect_true(any(is.finite(slow$att)))  # healthy cells still estimated
  expect_equal(is.na(slow$att), is.na(fast$att))
  expect_equal(slow$att, fast$att, tolerance = 1e-10)
  expect_equal(slow$se, fast$se, tolerance = 1e-10)
})

test_that("overlap_logit_fit matches fastglm::fastglm exactly, including the 0.999 decision", {
  # overlap_logit_fit() (fastglmPure with method = 0L, tol = 1e-8, maxit = 100L)
  # replaced fastglm::fastglm(x, y, family = binomial()) at every overlap-check
  # call site with a docstring claim of bit-identical fitted values. Lock that
  # contract -- and the max(fitted) >= 0.999 decision that drives the
  # att = NA-with-warning behavior -- on the degenerate designs where IRLS is
  # most fragile: near separation, complete separation, rank deficiency, and an
  # all-zero column (the sparse-factor-level case).
  set.seed(20260610)
  n <- 200
  x <- rnorm(n)
  designs <- list(
    near_sep = list(X = cbind(1, x), D = rbinom(n, 1, plogis(8 * x))),
    comp_sep = list(X = cbind(1, x), D = as.numeric(x > 0)),
    rank_def = list(X = cbind(1, x, x), D = rbinom(n, 1, plogis(2 * x))),
    zero_col = list(X = cbind(1, x, 0), D = rbinom(n, 1, plogis(2 * x)))
  )
  for (nm in names(designs)) {
    X <- designs[[nm]]$X; D <- designs[[nm]]$D
    # expected "fitted probabilities numerically 0 or 1" noise under separation
    f1 <- suppressWarnings(fastglm::fastglm(X, D, family = binomial()))
    f2 <- suppressWarnings(did:::overlap_logit_fit(X, D))
    expect_identical(f1$fitted.values, f2$fitted.values)
    expect_identical(max(f1$fitted.values) >= 0.999, max(f2$fitted.values) >= 0.999)
  }
})

test_that("slow panel path converts estimator NaN cells to NA like fast mode", {
  set.seed(47)
  d <- expand.grid(id = 1:90, period = 1:4)
  d$G <- ifelse(d$id <= 30, 0, ifelse(d$id <= 60, 2, 3))
  unit_x <- rnorm(90)
  d$X <- unit_x[d$id]
  d$Y <- 0.5 * d$X + 0.2 * d$period + (d$G > 0 & d$period >= d$G) +
    rnorm(nrow(d), 0, 0.2)
  d$w <- ifelse(d$G == 3, 0, 1)

  slow <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    weightsname = "w", faster_mode = FALSE, bstrap = FALSE)))
  fast <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d, tname = "period", idname = "id", gname = "G",
    weightsname = "w", faster_mode = TRUE, bstrap = FALSE)))

  expect_true(any(is.na(slow$att[slow$group == 3])))
  expect_false(any(is.nan(slow$att)))
  expect_equal(is.na(slow$att), is.na(fast$att))
})

#library(ggplot2)
#library(ggpubr)


## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, ipw model is incorrectly specified here
#-----------------------------------------------------------------------------
test_that("att_gt works w/o dynamics, time effects, or group effects", {
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$ipw <- FALSE
  data <- did::build_sim_dataset(sp)

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
  # reg
  res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg")


  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_reg$att[1], 1, tol=.5)
})


test_that("att_gt works using ipw", {
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$reg <- FALSE
  data <- did::build_sim_dataset(sp)

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                   gname="G", est_method="dr")


  # ipw
  res_ipw <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
                gname="G", est_method="ipw")

  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_ipw$att[1], 1, tol=.5)
})

test_that("two period case", {
  set.seed(09142024)
  sp <- did::reset.sim(time.periods=2)
  sp$ipw <- FALSE
  sp$n <- 10000
  data <- did::build_sim_dataset(sp)

  res <- suppressWarnings(
    att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
           gname="G", est_method="reg")
  )
  res

  agg_simple <- suppressWarnings(aggte(res, type="simple"))
  agg_group <- suppressWarnings(aggte(res, type="group"))
  agg_dynamic <- suppressWarnings(aggte(res, type="dynamic"))
  agg_calendar <- suppressWarnings(aggte(res, type="calendar"))

  expect_equal(agg_simple$overall.att, 1, tol=.5)
  expect_equal(agg_group$overall.att, 1, tol=.5)
  expect_equal(agg_dynamic$overall.att, 1, tol=.5)
  expect_equal(agg_calendar$overall.att, 1, tol=.5)
})

test_that("no covariates case", {
  set.seed(09142024)
  time.periods <- 4
  sp <- did::reset.sim(time.periods=time.periods)

  # no effect of covariates
  sp$bett <- sp$betu <- rep(0,time.periods)
  data <- did::build_sim_dataset(sp)

  res_dr <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
                gname="G", est_method="dr")

  res_reg <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
                gname="G", est_method="reg")

  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_reg$att[1], 1, tol=.5)
})

test_that("repeated cross section", {
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp, panel=FALSE)

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                   gname="G", est_method="dr", panel=FALSE)

  # reg
  res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                    gname="G", est_method="reg", panel=FALSE)

  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_reg$att[1], 1, tol=.5)
})


test_that("ipw repeated cross sections", {
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$reg <- FALSE
  sp$n <- 20000 # these are noisy
  data <- did::build_sim_dataset(sp, panel=FALSE)

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method="dr", panel=FALSE)


  #ipw
  res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", panel=FALSE)

  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_ipw$att[1], 1, tol=.5)
})


test_that("repeated cross sections dynamic effects", {
  set.seed(09142024)
  time.periods <- 4
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te.e <- 1:time.periods
  data <- did::build_sim_dataset(sp, panel=FALSE)

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                   gname="G", est_method="dr", panel=FALSE)


  agg_dynamic <- aggte(res_dr, type="dynamic")
  agg_idx <- agg_dynamic$egt==2

  expect_equal(agg_dynamic$att.egt[agg_idx], 3, tol=.5)
})

test_that("unbalanced panel", {
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  # drop second row to create unbalanced panel
  data <- data[-2,]

  # dr
  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method="dr", allow_unbalanced_panel=TRUE)

  expect_equal(res_dr$att[1], 1, tol=.5)

  expect_warning(att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method="dr", allow_unbalanced_panel=FALSE))

  # ipw version
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$reg <- FALSE
  data <- did::build_sim_dataset(sp)
  data <- data[-2,]

  res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method="ipw", allow_unbalanced_panel=TRUE)

  expect_equal(res_dr$att[1], 1, tol=.5)

  # unbalanced paenl without providing id, should error
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data <- data[sample(1:nrow(data),  size=floor(.9*nrow(data))),]

  expect_error(att_gt(yname="Y", xformla=~X, data=data, tname="period", idname=NULL,
                      gname="G", est_method="reg", panel=TRUE, allow_unbalanced_panel=TRUE))
})

test_that("not yet treated comparison group", {
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$reg <- FALSE
  data <- did::build_sim_dataset(sp, panel=FALSE)

  # dr
  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="notyettreated",
                gname="G", est_method="dr", panel=FALSE)

  expect_equal(res$att[1], 1, tol=.5)

  # no never treated group
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data <- subset(data, G > 0) # drop nevertreated

  # dr
  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="notyettreated",
                gname="G", est_method="dr", panel=FALSE)
  expect_equal(res$att[1], 1, tol=.5)


  # try to use never treated group as comparison group, should warn
  expect_warning(nonev_orig <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                      control_group="nevertreated",
                      gname="G", est_method="dr", panel=FALSE, faster_mode = FALSE))

  # try to use never treated group as comparison group with faster mode, should warn
  expect_warning(nonev_faster <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                        control_group="nevertreated",
                        gname="G", est_method="dr", panel=FALSE, faster_mode = TRUE))

  # make sure both methods give same ATT(g,t) with no never treated group
  expect_equal(nonev_orig$att, nonev_faster$att)

})

test_that("aggregations", {
  set.seed(09142024)
  # dynamic effects
  time.periods <- 4
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.e <- 1:time.periods
  data <- did::build_sim_dataset(sp)

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="nevertreated",
                gname="G", est_method="reg", panel=FALSE)

  agg_dynamic <- aggte(res, type="dynamic")
  agg_idx <- agg_dynamic$egt==2

  expect_equal(agg_dynamic$att.egt[agg_idx], 2, tol=.5)


  # group effects
  set.seed(09142024)
  time.periods <- 4
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.bet.ind <- 1:time.periods
  sp$reg <- FALSE
  data <- did::build_sim_dataset(sp, panel=FALSE)

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="notyettreated",
                gname="G", est_method="ipw", panel=FALSE)

  agg_group <- aggte(res, type="group")

  expect_equal(agg_group$att.egt[2], 2*2, tol=.5)


  # calendar time effects
  set.seed(09142024)
  time.periods <- 4
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.t <- sp$thet + 1:time.periods
  data <- did::build_sim_dataset(sp, panel=FALSE)

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="nevertreated",
                gname="G", est_method="dr", panel=FALSE)

  agg_calendar <- aggte(res, type="calendar")
  expect_equal(agg_calendar$att.egt[2], 2, tol=.5)


  # balancing with respect to event time
  set.seed(09142024)
  sp <- did::reset.sim()
  sp$te <- 0
  sp$te.e <- 1:time.periods
  sp$te.bet.ind <- 1:time.periods
  data <- did::build_sim_dataset(sp)

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="nevertreated",
                gname="G", est_method="dr", panel=FALSE)


  agg_dynamic <- aggte(res, type="dynamic")
  agg_dynamic_balance <- aggte(res, type="dynamic", balance_e=1)

  ad_idx <- which(agg_dynamic$egt == 1)
  adb_idx <- which(agg_dynamic_balance$egt == 1)

  expect_equal(agg_dynamic_balance$att.egt[adb_idx] - agg_dynamic_balance$att.egt[adb_idx-1], 1, tol=.5)
})

test_that("unequally spaced groups", {
  set.seed(09142024)
  time.periods <- 8
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.e <- 1:time.periods
  data <- did::build_sim_dataset(sp)
  keep.periods <- c(1,2,5,7)
  data <- subset(data, G %in% c(0, keep.periods))
  data <- subset(data, period %in% keep.periods)

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                control_group="nevertreated",
                gname="G", est_method="reg", panel=FALSE)

  agg_dynamic <- aggte(res, type="dynamic")
  agg_idx <- agg_dynamic$egt==2

  expect_equal(agg_dynamic$att.egt[agg_idx], 3, tol=.5)

  agg_dynamic_balance <- aggte(res, type="dynamic", balance_e=0)
  agg_idx2 <- which(agg_dynamic_balance$egt==0)
  expect_equal(agg_dynamic_balance$att.egt[agg_idx2], 1, tol=.5)
})

test_that("some units treated in first period", {
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data <- subset(data, period >= 2)

  expect_warning(att_gt(yname="Y", xformla=~X, data=data, tname="period",
                        control_group="nevertreated",
                        gname="G", est_method="reg", panel=FALSE))
})

test_that("min and max length of exposures", {
  set.seed(09142024)
  sp <- did::reset.sim()
  time.periods <- 4
  sp$te <- 0
  sp$te.e <- 1:time.periods
  sp$bett <- sp$betu <- rep(0,time.periods)
  data <- did::build_sim_dataset(sp)

  res <- att_gt(yname="Y", xformla=~1, data=data, tname="period",
                idname="id",
                control_group="nevertreated",
                gname="G", est_method="reg", panel=TRUE,
                base_period="varying")

  agg_dynamic <- aggte(res, type="dynamic", min_e=-1, max_e=1)
  agg_idx <- which(agg_dynamic$egt == 1)
  expect_equal(agg_dynamic$att.egt[agg_idx], 2, tol=.5)
})


test_that("anticipation", {
  set.seed(09142024)
  time.periods <- 5
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.e <- -1:(time.periods-2)
  data <- did::build_sim_dataset(sp)
  data$G <- ifelse(data$G==0, 0, data$G + 1) # add anticipation
  data <- subset(data, G <= time.periods) # drop last period (due to way data is constructed)
  # this will have an anticipation effect=-1, no effect at exposure,
  # and treatment effects increasing by one in subsequent periods

  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                idname="id",
                control_group="nevertreated",
                gname="G", est_method="dr",
                anticipation=1
                )

  agg_dynamic <- aggte(res, type="dynamic")
  agg_idx <- which(agg_dynamic$egt==2)
  expect_equal(agg_dynamic$att.egt[agg_idx], 2, tol=.5)

  # incorrectly ignore anticipation
  # causes over-stating treatment effects
  # due to using incorrect base-period
  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                idname="id",
                control_group="nevertreated",
                gname="G", est_method="dr",
                anticipation=0
                )

  agg_dynamic <- aggte(res, type="dynamic")
  agg_idx <- which(agg_dynamic$egt==2)
  expect_equal(agg_dynamic$att.egt[agg_idx], 3, tol=.5)


  # test for previous bug when using anticipation and
  # a notyettreated comparison group
  data <- subset(data, G != 0)
  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                idname="id",
                control_group="notyettreated",
                gname="G", est_method="dr",
                anticipation=1
                )

  agg_dynamic <- aggte(res, type="dynamic")
  agg_idx <- which(agg_dynamic$egt==0)
  expect_equal(agg_dynamic$att.egt[agg_idx], 0, tol=.3)
})

test_that("significance level and uniform confidence bands", {
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  # 5% significance level
  set.seed(1234)
  res05 <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                  gname="G", est_method="dr", alp=0.05)
  # 1% significance level
  set.seed(1234)
  res01 <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                  gname="G", est_method="dr", alp=0.01)
  # 5% pointwise
  set.seed(1234)
  res_pw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                  gname="G", est_method="dr", alp=0.05, cband=FALSE)

  expect_lte(res05$att[1] + res05$c*res05$se[1],
             res01$att[1] + res01$c*res01$se[1])
  expect_gte(res05$att[1] + res05$c*res05$se[1],
             res_pw$att[1] + res_pw$c*res_pw$se[1])

})

test_that("malformed data", {
  set.seed(09142024)
  # some groups later than last treated period
  # plus missing groups
  time.periods <- 7
  sp <- did::reset.sim(time.periods=time.periods)
  data <- did::build_sim_dataset(sp)
  data <- subset(data, period <= 4)
  missingG_ids <- sample(unique(data$id), size=10)
  data[data$id %in% missingG_ids,"G"] <- NA

  expect_warning(att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method="dr"))


  #-----------------------------------------------------------------------------
  # incorrectly specified id
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  expect_error(att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="brant",
                      gname="G", est_method="dr"))
})

test_that("varying or universal base period", {
  set.seed(09142024)
  time.periods <- 8
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.e <- 1:time.periods
  data <- did::build_sim_dataset(sp)
  data <- subset(data, (G<=5) | G==0 )
  # add pre-treatment effects
  data$G <- ifelse(data$G==0, 0, data$G+3)


  # dr
  res_varying <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", base_period="varying")

  agg_dynamic_varying <- aggte(res_varying, type="dynamic")

  res_universal <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                          gname="G", est_method="dr", base_period="universal")

  agg_dynamic_universal <- aggte(res_universal, type="dynamic")

  agg_idx <- which(agg_dynamic_varying$egt == -3)

  expect_equal(agg_dynamic_varying$att.egt[agg_idx], 1, tol=.5)
  expect_equal(agg_dynamic_universal$att.egt[agg_idx], -2, tol=.5)
})


test_that("small groups", {
  # code should still compute in this case (as comparison
  # group is large, but should give a warning about small groups)
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  # keep only one observation from group 2
  G2_keep_id <- unique(subset(data, G==2)$id)[1]
  data <- subset(data, (G != 2) | (id == G2_keep_id))

  # dr
  expect_warning(res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr"), "very few observations")
  # reg
  expect_warning(res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg"), "very few observations")
  # ipw
  expect_warning(res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw"), "very few observations")

  # estimates will be imprecise for group 2
  idx <- which(res_dr$group == 3 & res_dr$t==3)
  expect_equal(res_dr$att[idx], 1, tol=.5)
  expect_equal(res_reg$att[idx], 1, tol=.5)
})


test_that("small comparison group", {
  set.seed(09142024)
  # code doesn't run here if use never treated comparison group
  # but should run for all groups except the last one when
  # the not-yet-treated comparison group
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  # keep only one observation from untreated group
  G0_keep_id <- unique(subset(data, G==0)$id)[1]
  data <- subset(data, (G != 0) | (id == G0_keep_id))

  #-----------------------------------------------------------------------------
  # never treated comparison group
  #-----------------------------------------------------------------------------
  # dr
  expect_error(expect_warning(res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr"), "very few observations"), "never-treated group is too small")
  # reg
  expect_error(expect_warning(res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg"), "very few observations"), "never-treated group is too small")
  # ipw
  expect_error(expect_warning(res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw"), "very few observations"), "never-treated group is too small")

  #-----------------------------------------------------------------------------
  # not-yet-treated comparison group with faster_mode = TRUE
  #-----------------------------------------------------------------------------

  # dr
  expect_warning(res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                                               gname="G", est_method="dr", faster_mode = TRUE), "very few observations")
  # reg
  expect_warning(res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                                                gname="G", est_method="reg", faster_mode = TRUE), "very few observations")
  # ipw
  expect_warning(res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                                   gname="G", est_method="ipw", faster_mode = TRUE), "very few observations")

  #-----------------------------------------------------------------------------
  # not-yet-treated comparison group
  #-----------------------------------------------------------------------------
  # dr
  expect_warning(res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
              gname="G", est_method="dr", faster_mode = FALSE), "very few observations")
  # reg
  expect_warning(res_reg <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
              gname="G", est_method="reg", faster_mode = FALSE), "singular or numerically ill-conditioned")
  # ipw
  expect_warning(res_ipw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
              gname="G", est_method="ipw", faster_mode = FALSE), "overlap condition violated")

  # code should still work for some (g,t)'s
  expect_equal(res_dr$att[1], 1, tol=.5)
  expect_equal(res_reg$att[1], 1, tol=.5)
  expect_equal(res_ipw$att[1], 1, tol=.5)

  #-----------------------------------------------------------------------------
  # aggregations
  #-----------------------------------------------------------------------------

  agg_dyn <- aggte(res_dr, type="dynamic", na.rm=TRUE)
  agg_group <- aggte(res_reg, type="group", na.rm=TRUE)

  expect_equal(agg_dyn$att.egt[3], 1, tol=.5)
  expect_equal(agg_group$att.egt[1], 1, tol=.5)

  # make sure that standard errors are computed too
  expect_false(is.na(agg_dyn$se.egt[3]))
  expect_false(is.na(agg_group$se.egt[1]))

  agg_cal <- aggte(res_ipw, type="calendar", na.rm=TRUE)
  expect_equal(agg_cal$att.egt[1], 1, tol=.5)
  expect_false(is.na(agg_cal$se.egt[1]))
})

test_that("custom estimation method", {
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                gname="G", est_method=DRDID::drdid_imp_panel, panel=TRUE)
  expect_equal(res$att[1], 1, tol=.5)
})



test_that("sampling weights", {
  set.seed(09142024)
  # the idea here is that we can re-weight and should
  # get the same thing as if we subset
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data2 <- data
  keepids <- sample(unique(data$id), length(unique(data$id)))
  data$w <- 1*(data$id %in% keepids) # weights shouldn't have to have mean/sum 1
  data2 <- subset(data, id %in% keepids)

  res_weights <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                        gname="G", est_method="reg", weightsname="w")

  res_subset <- att_gt(yname="Y", xformla=~X, data=data2, tname="period", idname="id", control_group="notyettreated",
                        gname="G", est_method="reg")

  # test for same att's
  expect_equal(res_weights$att[1], res_subset$att[1])
  # test for same standard errors
  expect_equal(res_weights$se[1], res_subset$se[1], tol=.02)

})

# =============================================================================
# Column naming: user columns named gname/tname/idname should not crash
# =============================================================================

test_that("works when user column is literally named 'gname'", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  # Rename columns to match parameter names exactly
  names(data)[names(data) == "G"] <- "gname"
  names(data)[names(data) == "period"] <- "tname"
  names(data)[names(data) == "id"] <- "idname"

  mod <- att_gt(yname="Y", xformla=~X, data=data, tname="tname", idname="idname",
                gname="gname", est_method="reg", bstrap=FALSE)
  expect_false(all(is.na(mod$att)))

  # aggte should also work (this was the specific dreamerr bug)
  agg <- suppressWarnings(aggte(mod, type="simple"))
  expect_false(is.na(agg$overall.att))

  agg_dyn <- suppressWarnings(aggte(mod, type="dynamic"))
  expect_false(is.na(agg_dyn$overall.att))
})

test_that("works when user column is literally named 'gname' with faster_mode", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  names(data)[names(data) == "G"] <- "gname"
  names(data)[names(data) == "period"] <- "tname"
  names(data)[names(data) == "id"] <- "idname"

  mod <- att_gt(yname="Y", xformla=~X, data=data, tname="tname", idname="idname",
                gname="gname", est_method="reg", bstrap=FALSE, faster_mode=TRUE)
  expect_false(all(is.na(mod$att)))

  agg <- aggte(mod, type="simple")
  expect_false(is.na(agg$overall.att))
})

# =============================================================================
# Time-varying weights: fix_weights tests
# =============================================================================

test_that("time-varying weights: faster_mode matches slow mode (default fix_weights=NULL)", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), -0.1, 0.1)

  for (em in c("reg", "dr", "ipw")) {
    for (bp in c("varying", "universal")) {
      res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                         gname="G", est_method=em, weightsname="tv_weight",
                         base_period=bp, faster_mode=FALSE, bstrap=FALSE)
      res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                         gname="G", est_method=em, weightsname="tv_weight",
                         base_period=bp, faster_mode=TRUE, bstrap=FALSE)

      expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                   label=paste("ATT match:", em, bp))
    }
  }
})

test_that("fix_weights options: faster_mode matches slow mode (balanced panel)", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), -0.1, 0.1)

  for (fw in c("varying", "base_period", "first_period")) {
    res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                       gname="G", est_method="dr", weightsname="tv_weight",
                       fix_weights=fw, faster_mode=FALSE, bstrap=FALSE)
    res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                       gname="G", est_method="dr", weightsname="tv_weight",
                       fix_weights=fw, faster_mode=TRUE, bstrap=FALSE)

    expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                 label=paste("ATT match:", fw))
  }
})

test_that("time-invariant weights: all fix_weights options produce identical ATTs", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  n_ids <- length(unique(data$id))
  n_periods <- length(unique(data$period))
  data$const_weight <- rep(runif(n_ids, 1, 10), each = n_periods)

  res_default <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                        gname="G", est_method="reg", weightsname="const_weight",
                        bstrap=FALSE)

  for (fw in c("base_period", "first_period")) {
    res_fw <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                     gname="G", est_method="reg", weightsname="const_weight",
                     fix_weights=fw, bstrap=FALSE)
    expect_equal(res_default$att, res_fw$att, tolerance=1e-10,
                 label=paste("same ATT for", fw))
  }
})

test_that("message emitted for time-varying weights in balanced panel", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period * 1.0 + runif(nrow(data), 0, 0.5)

  expect_message(
    att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
           gname="G", weightsname="tv_weight", bstrap=FALSE),
    "Time-varying weights detected"
  )
})

test_that("no message for time-invariant weights", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  n_ids <- length(unique(data$id))
  n_periods <- length(unique(data$period))
  data$const_weight <- rep(runif(n_ids, 1, 10), each = n_periods)

  expect_no_message(
    att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
           gname="G", weightsname="const_weight", bstrap=FALSE)
  )
})

test_that("notyettreated with time-varying weights: faster_mode matches", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), 0, 0.5)

  res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                     gname="G", est_method="dr", weightsname="tv_weight",
                     control_group="notyettreated", faster_mode=FALSE, bstrap=FALSE)
  res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                     gname="G", est_method="dr", weightsname="tv_weight",
                     control_group="notyettreated", faster_mode=TRUE, bstrap=FALSE)

  expect_equal(res_slow$att, res_fast$att, tolerance=1e-10)
})

test_that("RC with time-varying weights: faster_mode matches", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period * 1.0 + runif(nrow(data), 0, 0.5)

  res_slow <- att_gt(yname="Y", data=data, tname="period", idname="id",
                     gname="G", est_method="reg", weightsname="tv_weight",
                     panel=FALSE, faster_mode=FALSE, bstrap=FALSE)
  res_fast <- att_gt(yname="Y", data=data, tname="period", idname="id",
                     gname="G", est_method="reg", weightsname="tv_weight",
                     panel=FALSE, faster_mode=TRUE, bstrap=FALSE)

  expect_equal(res_slow$att, res_fast$att, tolerance=1e-10)
})

test_that("fix_weights validation", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  expect_error(
    att_gt(yname="Y", data=data, tname="period", idname="id",
           gname="G", fix_weights="invalid_option", bstrap=FALSE),
    "fix_weights must be NULL"
  )

  # base_period and first_period not supported for repeated cross sections
  expect_error(
    att_gt(yname="Y", data=data, tname="period", idname="id",
           gname="G", fix_weights="base_period", panel=FALSE, bstrap=FALSE),
    "not supported for repeated cross sections"
  )
  expect_error(
    att_gt(yname="Y", data=data, tname="period", idname="id",
           gname="G", fix_weights="first_period", panel=FALSE, bstrap=FALSE),
    "not supported for repeated cross sections"
  )

  # varying not supported with custom est_method when panel = TRUE
  my_panel_est <- function(y1, y0, D, covariates, i.weights, inffunc, ...) {
    list(ATT = mean(y1 - y0), att.inf.func = rep(0, length(y1)))
  }
  expect_error(
    att_gt(yname="Y", data=data, tname="period", idname="id",
           gname="G", fix_weights="varying", est_method=my_panel_est,
           panel=TRUE, bstrap=FALSE),
    "not currently supported with custom est_method"
  )

  # varying IS supported with custom est_method when panel = FALSE (RC signature)
  my_rc_est <- function(y, post, D, covariates, i.weights, inffunc, ...) {
    n_obs <- length(y)
    post_c <- post[D==0]
    y_c <- y[D==0]
    w_c <- i.weights[D==0]
    att <- mean(y_c[post_c==1] * w_c[post_c==1]) / mean(w_c[post_c==1]) -
           mean(y_c[post_c==0] * w_c[post_c==0]) / mean(w_c[post_c==0])
    list(ATT = att, att.inf.func = rep(0, n_obs))
  }
  # Wald pre-test warning is expected with this small sim dataset (group 2
  # has only one pre-treatment period), but the key thing is: no error and
  # no recycling warnings from mismatched influence-function length.
  rc_result <- expect_no_error(
    withCallingHandlers(
      att_gt(yname="Y", data=data, tname="period", idname="id",
             gname="G", fix_weights="varying", est_method=my_rc_est,
             panel=FALSE, bstrap=FALSE),
      warning = function(w) {
        if (grepl("not a multiple of replacement length", conditionMessage(w)))
          stop("IF length mismatch: ", conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    )
  )
  expect_true(inherits(rc_result, "MP"))
  expect_false(anyNA(rc_result$att))
})

test_that("unbalanced panel fix_weights with units missing from reference period", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  # Drop some treated units from first period so fix_weights="first_period" must drop them
  first_p <- min(data$period)
  drop_ids <- unique(data$id[data$G > 0])[1:10]
  data <- data[!(data$id %in% drop_ids & data$period == first_p), ]
  data$w <- runif(nrow(data), 1, 5)

  for (fw in c("first_period", "base_period")) {
    res_slow <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", data = data, tname = "period", idname = "id",
             gname = "G", allow_unbalanced_panel = TRUE,
             fix_weights = fw, weightsname = "w",
             bstrap = FALSE, faster_mode = FALSE)
    ))
    res_fast <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", data = data, tname = "period", idname = "id",
             gname = "G", allow_unbalanced_panel = TRUE,
             fix_weights = fw, weightsname = "w",
             bstrap = FALSE, faster_mode = TRUE)
    ))
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10,
                 label = paste("unbalanced", fw, "ATT match"))
  }
})

# =============================================================================
# Influence function consistency: slow vs fast mode ATT AND SE must match
# =============================================================================

test_that("IF consistency: balanced panel, all fix_weights x est_method x base_period", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), -0.1, 0.1)

  for (fw in c(NA, "varying", "base_period", "first_period")) {
    fw_arg <- if (is.na(fw)) NULL else fw
    for (em in c("dr", "ipw", "reg")) {
      for (bp in c("varying", "universal")) {
        label <- paste("panel", if (is.na(fw)) "NULL" else fw, em, bp)

        res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                           idname="id", gname="G", est_method=em,
                           weightsname="tv_weight", fix_weights=fw_arg,
                           base_period=bp, faster_mode=FALSE,
                           bstrap=FALSE, cband=FALSE)
        res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                           idname="id", gname="G", est_method=em,
                           weightsname="tv_weight", fix_weights=fw_arg,
                           base_period=bp, faster_mode=TRUE,
                           bstrap=FALSE, cband=FALSE)

        expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                     label=paste("ATT", label))
        expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
                     label=paste("SE", label))
      }
    }
  }
})

test_that("IF consistency: balanced panel, notyettreated control group", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), -0.1, 0.1)

  for (fw in c(NA, "varying", "base_period", "first_period")) {
    fw_arg <- if (is.na(fw)) NULL else fw
    label <- paste("notyettreated", if (is.na(fw)) "NULL" else fw)

    res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                       idname="id", gname="G", est_method="dr",
                       weightsname="tv_weight", fix_weights=fw_arg,
                       control_group="notyettreated",
                       faster_mode=FALSE, bstrap=FALSE, cband=FALSE)
    res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                       idname="id", gname="G", est_method="dr",
                       weightsname="tv_weight", fix_weights=fw_arg,
                       control_group="notyettreated",
                       faster_mode=TRUE, bstrap=FALSE, cband=FALSE)

    expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                 label=paste("ATT", label))
    expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
                 label=paste("SE", label))
  }
})

test_that("IF consistency: repeated cross-sections, default weights x est_method", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tv_weight <- data$period + runif(nrow(data), -0.1, 0.1)

  # RC with default weights (fix_weights=NULL); fixed weight options tested separately
  for (em in c("dr", "ipw", "reg")) {
    label <- paste("RC NULL", em)

    res_slow <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                       idname="id", gname="G", est_method=em,
                       weightsname="tv_weight",
                       panel=FALSE, faster_mode=FALSE,
                       bstrap=FALSE, cband=FALSE)
    res_fast <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
                       idname="id", gname="G", est_method=em,
                       weightsname="tv_weight",
                       panel=FALSE, faster_mode=TRUE,
                       bstrap=FALSE, cband=FALSE)

    expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                 label=paste("ATT", label))
    expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
                 label=paste("SE", label))
  }
})

test_that("IF consistency: unbalanced panel, default weights x est_method", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  # Create unbalanced panel by dropping some observations
  set.seed(42)
  drop_idx <- sample(nrow(data), size = floor(nrow(data) * 0.05))
  data_unbal <- data[-drop_idx, ]

  # Default weights (fix_weights=NULL); fixed weight options for unbalanced panels
  # have known edge cases with unit availability across periods
  for (em in c("dr", "reg")) {
    label <- paste("unbalanced NULL", em)

    res_slow <- att_gt(yname="Y", xformla=~X, data=data_unbal, tname="period",
                       idname="id", gname="G", est_method=em,
                       allow_unbalanced_panel=TRUE,
                       faster_mode=FALSE, bstrap=FALSE, cband=FALSE)
    res_fast <- att_gt(yname="Y", xformla=~X, data=data_unbal, tname="period",
                       idname="id", gname="G", est_method=em,
                       allow_unbalanced_panel=TRUE,
                       faster_mode=TRUE, bstrap=FALSE, cband=FALSE)

    expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                 label=paste("ATT", label))
    expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
                 label=paste("SE", label))
  }
})

test_that("IF consistency: no covariates (xformla=~1), all data types", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  # Balanced panel, no covariates
  for (fw in c(NA, "varying")) {
    fw_arg <- if (is.na(fw)) NULL else fw
    label <- paste("no-covar panel", if (is.na(fw)) "NULL" else fw)

    res_slow <- att_gt(yname="Y", data=data, tname="period", idname="id",
                       gname="G", fix_weights=fw_arg,
                       faster_mode=FALSE, bstrap=FALSE, cband=FALSE)
    res_fast <- att_gt(yname="Y", data=data, tname="period", idname="id",
                       gname="G", fix_weights=fw_arg,
                       faster_mode=TRUE, bstrap=FALSE, cband=FALSE)

    expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
                 label=paste("ATT", label))
    expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
                 label=paste("SE", label))
  }

  # RC, no covariates
  res_slow <- att_gt(yname="Y", data=data, tname="period", idname="id",
                     gname="G", panel=FALSE,
                     faster_mode=FALSE, bstrap=FALSE, cband=FALSE)
  res_fast <- att_gt(yname="Y", data=data, tname="period", idname="id",
                     gname="G", panel=FALSE,
                     faster_mode=TRUE, bstrap=FALSE, cband=FALSE)

  expect_equal(res_slow$att, res_fast$att, tolerance=1e-10,
               label="ATT RC no-covar")
  expect_equal(res_slow$se, res_fast$se, tolerance=1e-10,
               label="SE RC no-covar")
})

test_that("clustered standard errors", {
  set.seed(09142024)
  # check that we can compute when clustered standard errors are supplied
  # either as numeric or as factor
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  data$cluster <- as.numeric(data$cluster)
  res_numeric <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                   gname="G", est_method="dr", clustervars="cluster")

  data$cluster <- as.factor(data$cluster)
  res_factor <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                        gname="G", est_method="dr", clustervars="cluster")

  # test for same att's
  expect_equal(res_factor$att[1], res_numeric$att[1])
  # test for same standard errors
  expect_equal(res_factor$se[1], res_numeric$se[1], tol=.02)

  #-----------------------------------------------------------------------------
  # clustered standard errors with unbalanced panel
  data <- data[-3,] # drop one observation
  res_ub <- att_gt(yname="Y",
              tname="period",
              idname="id",
              gname="G",
              xformla=~X,
              data=data,
              panel=TRUE,
              allow_unbalanced_panel=TRUE,
              faster_mode = FALSE,
              clustervars="cluster")
  expect_equal(res_ub$att[1], 1, tol=.5)

  # clustered standard errors with unbalanced panel with faster mode
  res_ub_faster <- att_gt(yname="Y",tname="period",idname="id",gname="G",
              xformla=~X,data=data,panel=TRUE, allow_unbalanced_panel=TRUE,
              faster_mode = TRUE, clustervars="cluster")
  expect_equal(res_ub_faster$att[1], 1, tol=.5)

  #-----------------------------------------------------------------------------
  # also, check that we error when clustering variable varies within unit
  # over time -- identically in both modes (the slow path used to accept this
  # input and fall back to i.i.d. SEs with bstrap = FALSE)
  set.seed(09142024)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$cluster <- as.numeric(data$cluster)
  data[1,]$cluster <- data[1,]$cluster+1

  for (fm in c(TRUE, FALSE)) {
    expect_error(res_vc <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                          gname="G", est_method="dr", clustervars="cluster", faster_mode=fm),
                 "Time-varying cluster variables are not supported")
  }

  #-----------------------------------------------------------------------------
  # clustered standard errors with repeated cross sections data
  data <- did::build_sim_dataset(sp, panel=FALSE)
  res_rc <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id", control_group="notyettreated",
                   gname="G", est_method="dr", clustervars="cluster", panel=FALSE)
  expect_equal(res_rc$att[1], 1, tol=.5)
})

test_that("faster mode enabled for panel data", {
  data <- did::mpdta
  out <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "universal",
                control_group = "nevertreated", est_method = "dr", faster_mode = FALSE)
  out2 <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "universal",
                control_group = "nevertreated", est_method = "dr", faster_mode = TRUE)

  # check if results are equal.
  expect_equal(out$att, out2$att)
  expect_equal(out$se, as.numeric(out2$se))
  # --------------------------------------------------------------------------------------------------------
  out <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "varying",
                control_group = "nevertreated", est_method = "dr", faster_mode = FALSE)
  out2 <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                 xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "varying",
                 control_group = "nevertreated", est_method = "dr", faster_mode = TRUE)

  # check if results are equal.
  expect_equal(out$att, out2$att)
  expect_equal(out$se, as.numeric(out2$se))

  # --------------------------------------------------------------------------------------------------------
  out <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "varying",
                control_group = "notyettreated", est_method = "dr", faster_mode = FALSE)
  out2 <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                 xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "varying",
                 control_group = "notyettreated", est_method = "dr", faster_mode = TRUE)

  # check if results are equal.
  expect_equal(out$att, out2$att)
  expect_equal(out$se, as.numeric(out2$se))

  # --------------------------------------------------------------------------------------------------------
  out <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "universal",
                control_group = "notyettreated", est_method = "dr", faster_mode = FALSE)
  out2 <- att_gt(yname = "lemp", gname = "first.treat", idname = "countyreal", tname = "year",
                 xformla = ~1, data = data, bstrap = FALSE, cband = FALSE, base_period = "universal",
                 control_group = "notyettreated", est_method = "dr", faster_mode = TRUE)

  # check if results are equal.
  expect_equal(out$att, out2$att)
  expect_equal(out$se, as.numeric(out2$se))

})

test_that("faster model enabled for repeated cross sectional data", {

  data_rcs <- as.data.table(did::build_sim_dataset(reset.sim(time.periods=4, n=1000), panel=FALSE))
  data_rcs$period <- as.integer(data_rcs$period)
  data_rcs[G == 0, G := Inf]

  # ----------------------------------------------------------------------------------------------------------------------------------
  out_rcs <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                    bstrap = FALSE, cband = FALSE, base_period = "universal", control_group = "nevertreated",
                    est_method = "dr", faster_mode = FALSE)

  out_rcs2 <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                      bstrap = FALSE, cband = FALSE, base_period = "universal", control_group = "nevertreated",
                      est_method = "dr", faster_mode = TRUE )

  # check if results are equal.
  expect_equal(out_rcs$att, out_rcs2$att)
  expect_equal(out_rcs$se, as.numeric(out_rcs2$se))

  # ----------------------------------------------------------------------------------------------------------------------------------
  out_rcs <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                    bstrap = FALSE, cband = FALSE, base_period = "varying", control_group = "nevertreated",
                    est_method = "dr", faster_mode = FALSE)

  out_rcs2 <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                      bstrap = FALSE, cband = FALSE, base_period = "varying", control_group = "nevertreated",
                      est_method = "dr", faster_mode = TRUE )

  # check if results are equal.
  expect_equal(out_rcs$att, out_rcs2$att)
  expect_equal(out_rcs$se, as.numeric(out_rcs2$se))

  # ----------------------------------------------------------------------------------------------------------------------------------
  out_rcs <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                    bstrap = FALSE, cband = FALSE, base_period = "varying", control_group = "notyettreated",
                    est_method = "dr", faster_mode = FALSE)

  out_rcs2 <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                      bstrap = FALSE, cband = FALSE, base_period = "varying", control_group = "notyettreated",
                      est_method = "dr", faster_mode = TRUE )

  # check if results are equal.
  expect_equal(out_rcs$att, out_rcs2$att)
  expect_equal(out_rcs$se, as.numeric(out_rcs2$se))

  # ----------------------------------------------------------------------------------------------------------------------------------
  out_rcs <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                    bstrap = FALSE, cband = FALSE, base_period = "universal", control_group = "notyettreated",
                    est_method = "dr", faster_mode = FALSE)

  out_rcs2 <- att_gt(yname = "Y", gname = "G", tname = "period", xformla = ~1, data = data_rcs, panel = FALSE,
                      bstrap = FALSE, cband = FALSE, base_period = "universal", control_group = "notyettreated",
                      est_method = "dr", faster_mode = TRUE )

  # check if results are equal.
  expect_equal(out_rcs$att, out_rcs2$att)
  expect_equal(out_rcs$se, as.numeric(out_rcs2$se))

})

test_that("faster model enabled for unbalanced panel data and time-varying covariates", {

  set.seed(05202025)
  # balanced panel dimensions
  N  <- 1000        # individuals
  TT <- 6           # time periods (t = 1, …, 6)
  # create id–time grid
  dt <- CJ(id = 1:N, t = 1:TT)   # CJ:=Cartesian join == balanced panel
  # assign cohort (first time treated) or never-treated
  # 40% treated in period 4, 30% in 5, 10% in 6, rest never treated
  cohort_vals  <- c(0, 4, 5, 6)                     # 0 = never treated
  cohort_probs <- c(1 - .40 - .30 - .10, .40, .30, .10)

  dt[, g := sample(
    cohort_vals,
    size = 1,          # one draw per id → time-invariant
    prob = cohort_probs),
    by = id]  # group by id

  # treatment indicator: D = 1 if t ≥ g and g > 0
  dt[, D := as.integer(t >= g & g > 0)]

  # four time-varying covariates
  dt[, `:=`(
    x1 = rnorm(.N),
    x2 = runif(.N, 0, 10) + 0.1 * t,
    x3 = rnorm(N)[id] + 0.2 * t + rnorm(.N),
    x4 = sin(2 * pi * t / TT) + rnorm(.N)
  )]

  # outcome model: baseline + covariate effects + treatment effect
  # note: just for debugging purposes, not a well-defined DGP for theoretical results!!
  beta  <- c(1.5, -0.8,  0.4, 2.0)          # coefficients on x1–x4
  tau   <- 3                                # true treatment effect
  alpha_i <- rnorm(N)[dt$id]                # id fixed effect
  gamma_t <- seq(-1, 1, length.out = TT)[dt$t]  # common time trend

  dt[, y := alpha_i + gamma_t +
       beta[1]*x1 + beta[2]*x2 + beta[3]*x3 + beta[4]*x4 +
       tau * D + rnorm(.N, sd = 2)]

  # outcome model: baseline + covariate effects + treatment effect
  # note: just for debugging purposes, not a well-defined DGP for theoretical results!!
  beta  <- c(1.5, -0.8,  0.4, 2.0)          # coefficients on x1–x4
  tau   <- 3                                # true treatment effect
  alpha_i <- rnorm(N)[dt$id]                # id fixed effect
  gamma_t <- seq(-1, 1, length.out = TT)[dt$t]  # common time trend

  dt[, y := alpha_i + gamma_t +
       beta[1]*x1 + beta[2]*x2 + beta[3]*x3 + beta[4]*x4 +
       tau * D + rnorm(.N, sd = 2)]

  #################################################################
  ##  Make the panel UNBALANCED
  ##  drop 25 % of rows at random, but leave ≥1 row per id
  #################################################################
  drop_rate <- 0.25
  dt_unbal <- dt[, .SD[runif(.N) > drop_rate], by = id]     # keep rows with prob 0.75
  dt_unbal  <- dt_unbal[, if (.N > 0) .SD, by = id]         # (safety) ids with ≥1 obs

  # ----------------------------------------------------------------------------------------------------------------------------------
  # let's run did with covariates using faster_mode=FALSE
  att_slower <- att_gt(
    yname   = "y",
    tname   = "t",
    idname  = "id",
    gname   = "g",
    xformla = ~ x1 + x2 + x3 + x4,   # four time-varying covariates
    data    = dt,
    base_period = "universal", # "universal" or "varying"
    panel   = TRUE,
    faster_mode = FALSE,
    bstrap = FALSE,
  )

  # now let's run did with covariates using faster_mode=TRUE
  att_faster <- att_gt(
    yname   = "y",
    tname   = "t",
    idname  = "id",
    gname   = "g",
    xformla = ~ x1 + x2 + x3 + x4,   # four time-varying covariates
    data    = dt,
    panel   = TRUE,
    base_period = "universal", # "universal" or "varying"
    faster_mode = TRUE,
    bstrap = FALSE,
  )

  # check if results are equal.
  expect_equal(att_slower$att, att_faster$att)
  expect_equal(att_slower$se, as.numeric(att_faster$se))

  # get event study estimates
  out1 = att_slower |>
    aggte(type = "dynamic",  cband = FALSE, bstrap = FALSE)

  out2 = att_faster |>
    aggte(type = "dynamic", cband = FALSE, bstrap = FALSE)

  # check if results are equal.
  expect_equal(out1$att.egt, out2$att.egt)
  expect_equal(out1$se.egt, as.numeric(out2$se.egt))

  # ----------------------------------------------------------------------------------------------------------------------------------
  # running the same but with unbalanced panel data
  att_slow <- att_gt(
    yname   = "y",
    tname   = "t",
    idname  = "id",
    gname   = "g",
    xformla = ~ x1 + x2 + x3 + x4,
    data    = dt_unbal,
    panel   = TRUE,
    allow_unbalanced_panel = TRUE,
    faster  = FALSE,
    bstrap = FALSE
  )

  expect_message(att_fast <- att_gt(
    yname   = "y",
    tname   = "t",
    idname  = "id",
    gname   = "g",
    xformla = ~ x1 + x2 + x3 + x4,
    data    = dt_unbal,
    panel   = TRUE,
    allow_unbalanced_panel = TRUE,
    faster  = TRUE,
    bstrap = FALSE
  ), "unbalanced panel")

  # check if results are equal.
  expect_equal(att_slow$att, att_fast$att)
  expect_equal(att_slow$se, as.numeric(att_fast$se))


  # get event study estimates
  out1 = att_slow |>
    aggte(type = "dynamic",  cband = FALSE, bstrap = FALSE)


  out2 = att_fast |>
    aggte(type = "dynamic", cband = FALSE, bstrap = FALSE)

  # check if results are equal.
  expect_equal(out1$att.egt, out2$att.egt)
  expect_equal(out1$se.egt, out2$se.egt, tol=.0005)

})


test_that("faster_mode = TRUE matches baseline on filtered sim dataset when there are not subsequent cohort and time periods", {
  # simulate full panel
  set.seed(09142024)
  sp <- reset.sim()
  dt <- build_sim_dataset(sp)

  # filter down to two periods
  # here we know build_sim_dataset() has periods 1:4, so pick 2 & 4
  # cohorts -> [3,4], periods -> [2,4]
  dt2 <- dt[dt$period %in% c(2, 4), ]

  # run att_gt with both modes (no errors)
  expect_warning({
    res_slow <- att_gt(
      yname         = "Y",
      tname         = "period",
      idname        = "id",
      gname         = "G",
      data          = dt2,
      panel         = TRUE,
      control_group = "nevertreated",
      xformla       = NULL,
      est_method    = "dr",
      base_period   = "universal",
      faster_mode   = FALSE
    )
    res_fast <- att_gt(
      yname         = "Y",
      tname         = "period",
      idname        = "id",
      gname         = "G",
      data          = dt2,
      panel         = TRUE,
      control_group = "nevertreated",
      xformla       = NULL,
      est_method    = "dr",
      base_period   = "universal",
      faster_mode   = TRUE
    )
  }, "Dropped 999 units that were already treated in the first period")


  # they should have the same length and (within tol) the same values
  expect_length(res_slow$att, 4)
  expect_length(res_fast$att, 4)
  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-8)
})


#-----------------------------------------------------------------------------
# Regression tests for time indexing bug (Issue: faster_mode returns wrong time indices)
# These tests ensure that faster_mode=TRUE returns the same time periods (t field)
# as faster_mode=FALSE, not just the same ATT estimates
#-----------------------------------------------------------------------------

test_that("faster_mode time indexing matches baseline with repeated cross-sections", {
  set.seed(12345)

  # Create data with non-consecutive time periods to catch indexing bugs
  # Time periods: 2, 3, 4, 5, 6, 7, 8, 9 (starting at 2, not 1)
  n_groups <- 50
  n_periods <- 9

  df <- data.frame(
    g = rep(1:n_groups, each = n_periods),
    t = rep(1:n_periods, times = n_groups),
    gfe = rep(rnorm(n_groups, 0, 1), each = n_periods)
  )

  # Add time fixed effects
  tfe_vec <- 0.1 * rnorm(n_periods, 0, 1)
  df$tfe <- tfe_vec[df$t]

  # Assign treatment cohorts (using time values, not indices)
  df$cohort <- 0
  df$cohort[df$g >= 21 & df$g <= 30] <- 4
  df$cohort[df$g >= 11 & df$g <= 20] <- 6
  df$cohort[df$g >= 1 & df$g <= 10] <- 8

  # Generate outcomes
  df$y0 <- 0.1 * df$g + df$gfe + 0.1 * df$t + df$tfe + rnorm(nrow(df), 0, 1)
  df$y1 <- df$y0

  # Add treatment effects for each cohort
  treated_c4 <- df$cohort == 4 & df$t >= 4
  df$y1[treated_c4] <- df$y1[treated_c4] + 1.5 + 1.0 * (df$t[treated_c4] - 4)

  treated_c6 <- df$cohort == 6 & df$t >= 6
  df$y1[treated_c6] <- df$y1[treated_c6] + 1.0 + 0.7 * (df$t[treated_c6] - 6)

  treated_c8 <- df$cohort == 8 & df$t >= 8
  df$y1[treated_c8] <- df$y1[treated_c8] + 0.5 + 0.4 * (df$t[treated_c8] - 8)

  # Expand data (simulate varying sample sizes)
  set.seed(123)
  rep_by_g <- sapply(1:n_groups, function(j) floor(5 + 2 * (1.1)^j * runif(1)))
  df$rep_g <- rep_by_g[df$g]
  df <- df[rep(1:nrow(df), df$rep_g), ]

  rep_by_t <- sapply(1:n_periods, function(j) floor(2 + 3 * (1.2)^(n_periods - j) * runif(1)))
  df$rep_t <- rep_by_t[df$t]
  df <- df[rep(1:nrow(df), df$rep_t), ]

  df <- df[, !(names(df) %in% c("rep_g", "rep_t"))]

  # Treatment indicator and observed outcome
  df$tx <- (df$cohort > 0) & (df$t >= df$cohort)
  df$y <- ifelse(df$tx, df$y1, df$y0)
  df$id <- 1:nrow(df)

  # Estimate with both modes
  res_slow <- att_gt(
    yname = "y",
    tname = "t",
    gname = "cohort",
    idname = "id",
    data = df,
    est_method = "reg",
    panel = FALSE,
    bstrap = FALSE,
    cband = FALSE,
    faster_mode = FALSE
  )

  res_fast <- att_gt(
    yname = "y",
    tname = "t",
    gname = "cohort",
    idname = "id",
    data = df,
    est_method = "reg",
    panel = FALSE,
    bstrap = FALSE,
    cband = FALSE,
    faster_mode = TRUE
  )

  # Test that time indexing is identical (this was the bug)
  expect_equal(res_slow$t, res_fast$t,
               info = "Time periods (t field) should match between faster_mode=TRUE and FALSE")

  # Also verify group indexing
  expect_equal(res_slow$group, res_fast$group,
               info = "Group values should match between faster_mode=TRUE and FALSE")

  # Verify ATT estimates match
  expect_equal(res_slow$att, res_fast$att,
               info = "ATT estimates should match between faster_mode=TRUE and FALSE")

  # Verify standard errors match
  expect_equal(res_slow$se, as.numeric(res_fast$se),
               info = "Standard errors should match between faster_mode=TRUE and FALSE")
})


test_that("faster_mode time indexing matches baseline with panel data and varying base period", {
  set.seed(54321)

  # Use the package's built-in data with non-standard time periods
  data <- did::mpdta

  # Test with varying base period (default, where the bug was most obvious)
  res_slow <- att_gt(
    yname = "lemp",
    gname = "first.treat",
    idname = "countyreal",
    tname = "year",
    xformla = ~1,
    data = data,
    bstrap = FALSE,
    cband = FALSE,
    base_period = "varying",
    control_group = "nevertreated",
    est_method = "dr",
    faster_mode = FALSE
  )

  res_fast <- att_gt(
    yname = "lemp",
    gname = "first.treat",
    idname = "countyreal",
    tname = "year",
    xformla = ~1,
    data = data,
    bstrap = FALSE,
    cband = FALSE,
    base_period = "varying",
    control_group = "nevertreated",
    est_method = "dr",
    faster_mode = TRUE
  )

  # Critical test: time periods must match exactly
  expect_identical(res_slow$t, res_fast$t,
                   info = "Time periods must be identical between modes with varying base period")

  # Verify other components
  expect_identical(res_slow$group, res_fast$group)
  expect_equal(res_slow$att, res_fast$att)
  expect_equal(res_slow$se, as.numeric(res_fast$se))
})


test_that("faster_mode time indexing with non-consecutive time periods", {
  set.seed(99999)

  # Create data with gaps in time periods (e.g., 2000, 2002, 2005, 2007, 2010)
  # This tests that we use actual calendar times, not sequential indices
  time_periods <- c(2000, 2002, 2005, 2007, 2010)
  n_units <- 100

  # Create balanced panel
  df <- expand.grid(id = 1:n_units, year = time_periods)
  df$id <- as.integer(df$id)

  # Assign cohorts
  df$cohort <- 0
  df$cohort[df$id <= 30] <- 2005
  df$cohort[df$id > 30 & df$id <= 60] <- 2007
  # Rest are never treated (cohort = 0)

  # Generate outcomes
  df$y0 <- df$id * 0.1 + rnorm(nrow(df), 0, 0.5)
  df$y1 <- df$y0

  # Add treatment effects
  treated <- df$cohort > 0 & df$year >= df$cohort
  df$y1[treated] <- df$y1[treated] + 2.0

  # Observed outcome
  df$y <- ifelse(treated, df$y1, df$y0)

  # Estimate with both modes
  res_slow <- att_gt(
    yname = "y",
    tname = "year",
    idname = "id",
    gname = "cohort",
    data = df,
    panel = TRUE,
    bstrap = FALSE,
    cband = FALSE,
    est_method = "reg",
    faster_mode = FALSE
  )

  res_fast <- att_gt(
    yname = "y",
    tname = "year",
    idname = "id",
    gname = "cohort",
    data = df,
    panel = TRUE,
    bstrap = FALSE,
    cband = FALSE,
    est_method = "reg",
    faster_mode = TRUE
  )

  # The critical test: years should be actual calendar years, not indices
  expect_true(all(res_slow$t %in% time_periods),
              info = "Time values should be actual calendar years")
  expect_true(all(res_fast$t %in% time_periods),
              info = "Time values should be actual calendar years in faster_mode")

  # Time periods must match between modes
  expect_equal(res_slow$t, res_fast$t,
               info = "Time periods must match with non-consecutive calendar times")

  # Verify the minimum time period is correct (should be 2002, not 1 or 2000)
  # With varying base period, first time should be period after first period
  expect_true(min(res_slow$t) > min(time_periods),
              info = "With varying base period, should skip first time period")
  expect_equal(min(res_slow$t), min(res_fast$t))

  # Verify other components
  expect_equal(res_slow$group, res_fast$group)
  expect_equal(res_slow$att, res_fast$att)
  expect_equal(res_slow$se, as.numeric(res_fast$se))
})


test_that("faster_mode time indexing with universal base period", {
  set.seed(11111)

  # Simpler test with universal base period
  sp <- did::reset.sim(time.periods = 5)
  data <- did::build_sim_dataset(sp)

  res_slow <- att_gt(
    yname = "Y",
    xformla = ~X,
    data = data,
    tname = "period",
    idname = "id",
    gname = "G",
    est_method = "dr",
    base_period = "universal",
    bstrap = FALSE,
    faster_mode = FALSE
  )

  res_fast <- att_gt(
    yname = "Y",
    xformla = ~X,
    data = data,
    tname = "period",
    idname = "id",
    gname = "G",
    est_method = "dr",
    base_period = "universal",
    bstrap = FALSE,
    faster_mode = TRUE
  )

  # Time indexing must match
  expect_equal(res_slow$t, res_fast$t,
               info = "Time periods must match with universal base period")

  # Verify results are consistent
  expect_equal(res_slow$group, res_fast$group)
  expect_equal(res_slow$att, res_fast$att)
  expect_equal(res_slow$se, as.numeric(res_fast$se))
})

# =============================================================================
# Comprehensive tests for aggte() aggregation types
# =============================================================================

# Shared setup: known DGP with treatment effect = 1
set.seed(20260401)
sp <- did::reset.sim()
sp$te <- 1  # constant treatment effect
data_agg <- did::build_sim_dataset(sp)

mp_agg <- suppressWarnings(suppressMessages(
  att_gt(yname = "Y", xformla = ~X, data = data_agg, tname = "period",
         idname = "id", gname = "G", est_method = "dr",
         bstrap = FALSE)
))

# =============================================================================
# type = "simple"
# =============================================================================

test_that("aggte simple returns valid overall ATT", {
  agg <- suppressWarnings(aggte(mp_agg, type = "simple"))
  expect_s3_class(agg, "AGGTEobj")
  expect_false(is.na(agg$overall.att))
  expect_equal(agg$overall.att, 1, tolerance = 0.5)
})

test_that("aggte simple returns valid SE", {
  agg <- suppressWarnings(aggte(mp_agg, type = "simple"))
  expect_true(agg$overall.se > 0)
  expect_false(is.na(agg$overall.se))
})

test_that("aggte simple has no egt component", {
  agg <- suppressWarnings(aggte(mp_agg, type = "simple"))
  expect_null(agg$egt)
})

test_that("aggte simple influence function has correct length", {
  agg <- suppressWarnings(aggte(mp_agg, type = "simple"))
  expect_equal(length(agg$inf.function$simple.att), nobs(mp_agg))
})

# =============================================================================
# type = "dynamic"
# =============================================================================

test_that("aggte dynamic returns event-time specific ATTs", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  expect_false(is.null(agg$egt))
  expect_true(length(agg$egt) > 0)
  # Should have both negative and non-negative event times
  expect_true(any(agg$egt < 0))
  expect_true(any(agg$egt >= 0))
})

test_that("aggte dynamic event times are sorted", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  expect_equal(agg$egt, sort(agg$egt))
})

test_that("aggte dynamic overall.att averages post-treatment event times", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  expect_false(is.na(agg$overall.att))
  expect_equal(agg$overall.att, 1, tolerance = 0.5)
})

test_that("aggte dynamic min_e filters event times", {
  agg_full <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  agg_filt <- suppressWarnings(aggte(mp_agg, type = "dynamic", min_e = -1))
  expect_true(min(agg_filt$egt) >= -1)
  expect_true(length(agg_filt$egt) <= length(agg_full$egt))
})

test_that("aggte dynamic max_e filters event times", {
  agg_full <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  agg_filt <- suppressWarnings(aggte(mp_agg, type = "dynamic", max_e = 1))
  expect_true(max(agg_filt$egt) <= 1)
  expect_true(length(agg_filt$egt) <= length(agg_full$egt))
})

test_that("aggte dynamic min_e and max_e together", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic", min_e = -1, max_e = 1))
  expect_true(min(agg$egt) >= -1)
  expect_true(max(agg$egt) <= 1)
})

test_that("aggte dynamic balance_e filters groups", {
  agg_unbal <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  agg_bal <- suppressWarnings(aggte(mp_agg, type = "dynamic", balance_e = 1))
  # Balanced version may have fewer event times or same
  expect_true(length(agg_bal$egt) <= length(agg_unbal$egt))
})

test_that("aggte dynamic SEs are positive where ATT is not NA", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic"))
  non_na <- !is.na(agg$att.egt)
  expect_true(all(agg$se.egt[non_na] > 0))
})

# =============================================================================
# type = "group"
# =============================================================================

test_that("aggte group returns per-group ATTs", {
  agg <- suppressWarnings(aggte(mp_agg, type = "group"))
  expect_false(is.null(agg$egt))
  # egt should contain the group values
  expect_true(all(agg$egt %in% unique(mp_agg$group)))
})

test_that("aggte group overall.att is reasonable", {
  agg <- suppressWarnings(aggte(mp_agg, type = "group"))
  expect_false(is.na(agg$overall.att))
  expect_equal(agg$overall.att, 1, tolerance = 0.5)
})

test_that("aggte group SEs are positive for each group", {
  agg <- suppressWarnings(aggte(mp_agg, type = "group"))
  non_na <- !is.na(agg$att.egt)
  expect_true(all(agg$se.egt[non_na] > 0))
})

# =============================================================================
# type = "calendar"
# =============================================================================

test_that("aggte calendar returns per-period ATTs", {
  agg <- suppressWarnings(aggte(mp_agg, type = "calendar"))
  expect_false(is.null(agg$egt))
  expect_true(length(agg$egt) > 0)
})

test_that("aggte calendar overall.att is reasonable", {
  agg <- suppressWarnings(aggte(mp_agg, type = "calendar"))
  expect_false(is.na(agg$overall.att))
  expect_equal(agg$overall.att, 1, tolerance = 0.5)
})

test_that("aggte calendar SEs are positive for each period", {
  agg <- suppressWarnings(aggte(mp_agg, type = "calendar"))
  non_na <- !is.na(agg$att.egt)
  expect_true(all(agg$se.egt[non_na] > 0))
})

test_that("aggte calendar only includes post-treatment periods", {
  agg <- suppressWarnings(aggte(mp_agg, type = "calendar"))
  min_group <- min(mp_agg$group)
  # All calendar periods should be at or after the earliest treatment
  expect_true(all(agg$egt >= min_group))
})

# =============================================================================
# na.rm behavior
# =============================================================================

test_that("aggte with na.rm=TRUE drops NA ATTs and proceeds", {
  # Create MP with some NA ATTs manually
  mp_tmp <- mp_agg
  mp_tmp$att[1] <- NA
  # Should work with na.rm=TRUE
  agg <- suppressWarnings(aggte(mp_tmp, type = "dynamic", na.rm = TRUE))
  expect_s3_class(agg, "AGGTEobj")
  expect_false(is.na(agg$overall.att))
})

# =============================================================================
# Cross-type consistency
# =============================================================================

test_that("all aggte types return AGGTEobj class", {
  for (tp in c("simple", "dynamic", "group", "calendar")) {
    agg <- suppressWarnings(aggte(mp_agg, type = tp))
    expect_s3_class(agg, "AGGTEobj")
  }
})

test_that("all aggte types have non-NULL DIDparams", {
  for (tp in c("simple", "dynamic", "group", "calendar")) {
    agg <- suppressWarnings(aggte(mp_agg, type = tp))
    expect_false(is.null(agg$DIDparams), label = tp)
  }
})

test_that("aggte preserves overridden bstrap/alp settings", {
  agg <- suppressWarnings(aggte(mp_agg, type = "dynamic", bstrap = FALSE, alp = 0.01))
  expect_equal(agg$DIDparams$bstrap, FALSE)
  expect_equal(agg$DIDparams$alp, 0.01)
})

# =============================================================================
# cband with bootstrap
# =============================================================================

test_that("aggte with cband=TRUE and bstrap=TRUE returns simultaneous critical value", {
  mp_boot <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data_agg, tname = "period",
           idname = "id", gname = "G", est_method = "reg",
           bstrap = TRUE, biters = 100, cband = TRUE)
  ))
  agg <- aggte(mp_boot, type = "dynamic", bstrap = TRUE, biters = 100, cband = TRUE)
  # Simultaneous critical value should be at least as large as pointwise z
  expect_true(agg$crit.val.egt >= qnorm(0.975))
})

# =============================================================================
# Regression: type = "group" with na.rm = TRUE and a finite max_e
# =============================================================================

test_that("aggte group with na.rm and finite max_e excludes (not errors on) groups whose only non-NA cell is past max_e", {
  # A group whose only non-NA ATT(g,t) lies PAST max_e used to pass the group filter
  # (which ignored max_e) but then have an all-NA, na.rm-emptied selection in the
  # estimate -> stop("No valid att_gt() estimates ..."). The group should instead be
  # dropped from the group aggregation.
  mp <- mp_agg
  g_first <- sort(unique(mp$group))[1]
  # NA-out that group's earliest post-treatment cells, leaving only a late one (e large)
  post_g <- which(mp$group == g_first & mp$t >= g_first)
  # Hard precondition (not a silent if-guard): if the fixture ever drifts below
  # two post-treatment cells, the regression coverage must fail loudly.
  expect_gte(length(post_g), 2)
  mp$att[post_g[-length(post_g)]] <- NA   # keep only the last (largest event-time) cell
  last_e <- (mp$t[post_g[length(post_g)]] - g_first)
  me <- max(0, last_e - 1)                # max_e below the only surviving cell's event-time
  expect_error(
    suppressWarnings(suppressMessages(aggte(mp, type = "group", na.rm = TRUE, max_e = me))),
    NA)                                   # must NOT error
  agg <- suppressWarnings(suppressMessages(aggte(mp, type = "group", na.rm = TRUE, max_e = me)))
  expect_s3_class(agg, "AGGTEobj")
  expect_false(g_first %in% agg$egt)      # the offending group is excluded
})

test_that("aggte group default (max_e = Inf) is unchanged by the gnotna max_e filter", {
  a1 <- suppressWarnings(suppressMessages(aggte(mp_agg, type = "group")))
  expect_s3_class(a1, "AGGTEobj")
  expect_false(is.na(a1$overall.att))
})

# =============================================================================
# Tests for glance.MP and glance.AGGTEobj S3 methods
# =============================================================================

# Shared setup: generate MP and AGGTEobj results
set.seed(20260401)
sp <- did::reset.sim()
data_gl <- did::build_sim_dataset(sp)

mp_slow <- suppressWarnings(suppressMessages(
  att_gt(yname = "Y", xformla = ~X, data = data_gl, tname = "period",
         idname = "id", gname = "G", est_method = "dr",
         bstrap = FALSE, faster_mode = FALSE)
))

mp_fast <- suppressWarnings(suppressMessages(
  att_gt(yname = "Y", xformla = ~X, data = data_gl, tname = "period",
         idname = "id", gname = "G", est_method = "dr",
         bstrap = FALSE, faster_mode = TRUE)
))

agg_types <- c("simple", "dynamic", "group", "calendar")
agg_slow <- lapply(setNames(agg_types, agg_types), function(tp) {
  suppressWarnings(aggte(mp_slow, type = tp))
})
agg_fast <- lapply(setNames(agg_types, agg_types), function(tp) {
  suppressWarnings(aggte(mp_fast, type = tp))
})

# =============================================================================
# glance.MP tests
# =============================================================================

test_that("glance.MP returns a one-row data.frame", {
  gl <- glance(mp_slow)
  expect_equal(nrow(gl), 1)
  expect_s3_class(gl, "data.frame")
})

test_that("glance.MP has expected columns", {
  gl <- glance(mp_slow)
  expected_cols <- c("nobs", "ngroup", "ntime", "control.group", "est.method")
  expect_true(all(expected_cols %in% names(gl)))
})

test_that("glance.MP values are reasonable", {
  gl <- glance(mp_slow)
  expect_true(gl$nobs > 0)
  expect_true(gl$ngroup > 0)
  expect_true(gl$ntime > 0)
  expect_equal(gl$control.group, "nevertreated")
  expect_equal(gl$est.method, "dr")
})

test_that("glance.MP nobs matches nobs.MP", {
  gl <- glance(mp_slow)
  expect_equal(gl$nobs, nobs(mp_slow))
})

test_that("glance.MP works with faster_mode = TRUE", {
  gl <- glance(mp_fast)
  expect_equal(nrow(gl), 1)
  expect_true(gl$nobs > 0)
  expect_true(gl$ngroup > 0)
  expect_true(gl$ntime > 0)
})

test_that("glance.MP agrees between faster_mode settings", {
  gl_slow <- glance(mp_slow)
  gl_fast <- glance(mp_fast)
  expect_equal(gl_slow$nobs, gl_fast$nobs)
  expect_equal(gl_slow$ngroup, gl_fast$ngroup)
  expect_equal(gl_slow$ntime, gl_fast$ntime)
  expect_equal(gl_slow$control.group, gl_fast$control.group)
  expect_equal(gl_slow$est.method, gl_fast$est.method)
})

# =============================================================================
# glance.AGGTEobj tests
# =============================================================================

test_that("glance.AGGTEobj returns a one-row data.frame for all 4 types", {
  for (tp in agg_types) {
    gl <- glance(agg_slow[[tp]])
    expect_equal(nrow(gl), 1, label = paste("nrow for", tp))
    expect_s3_class(gl, "data.frame")
  }
})

test_that("glance.AGGTEobj has expected columns", {
  gl <- glance(agg_slow[["dynamic"]])
  expected_cols <- c("type", "nobs", "ngroup", "ntime", "control.group", "est.method")
  expect_true(all(expected_cols %in% names(gl)))
})

test_that("glance.AGGTEobj type column matches requested type", {
  for (tp in agg_types) {
    gl <- glance(agg_slow[[tp]])
    expect_equal(gl$type, tp, label = paste("type for", tp))
  }
})

test_that("glance.AGGTEobj values are not NULL or NA", {
  for (tp in agg_types) {
    gl <- glance(agg_slow[[tp]])
    expect_false(any(sapply(gl, is.null)), label = paste("no NULLs for", tp))
    expect_false(any(is.na(gl)), label = paste("no NAs for", tp))
  }
})

test_that("glance.AGGTEobj works with faster_mode = TRUE", {
  for (tp in agg_types) {
    gl <- glance(agg_fast[[tp]])
    expect_equal(nrow(gl), 1, label = paste("nrow for", tp, "fast"))
    expect_false(any(sapply(gl, is.null)), label = paste("no NULLs for", tp, "fast"))
  }
})

test_that("glance.MP and glance.AGGTEobj agree on nobs", {
  gl_mp <- glance(mp_slow)
  for (tp in agg_types) {
    gl_agg <- glance(agg_slow[[tp]])
    expect_equal(gl_mp$nobs, gl_agg$nobs, label = paste("nobs for", tp))
  }
})

# =============================================================================
# Backwards compatibility: DIDparams lacking faster_mode (pre-2.2 saved objects)
# =============================================================================

test_that("glance.MP works when DIDparams lacks faster_mode", {
  mp_old <- mp_slow
  mp_old$DIDparams$faster_mode <- NULL
  expect_identical(glance(mp_old), glance(mp_slow))
})

test_that("glance.AGGTEobj works when DIDparams lacks faster_mode", {
  agg_old <- agg_slow[["dynamic"]]
  agg_old$DIDparams$faster_mode <- NULL
  expect_identical(glance(agg_old), glance(agg_slow[["dynamic"]]))
})

test_that("nobs.AGGTEobj works when DIDparams lacks faster_mode", {
  agg_old <- agg_slow[["dynamic"]]
  agg_old$DIDparams$faster_mode <- NULL
  expect_identical(stats::nobs(agg_old), stats::nobs(agg_slow[["dynamic"]]))
})

# =============================================================================
# Custom est_method (user-supplied function): glance reports "custom"
# =============================================================================

mp_custom <- suppressWarnings(suppressMessages(
  att_gt(yname = "Y", xformla = ~X, data = data_gl, tname = "period",
         idname = "id", gname = "G",
         est_method = function(y1, y0, D, covariates, i.weights, inffunc, ...) {
           DRDID::drdid_panel(y1 = y1, y0 = y0, D = D, covariates = covariates,
                              i.weights = i.weights, boot = FALSE,
                              inffunc = inffunc)
         },
         bstrap = FALSE, faster_mode = FALSE)
))

test_that("glance.MP works with a custom est_method function", {
  gl <- glance(mp_custom)
  expect_s3_class(gl, "data.frame")
  expect_equal(nrow(gl), 1)
  expect_equal(gl$est.method, "custom")
})

test_that("glance.AGGTEobj works with a custom est_method function", {
  agg_custom <- suppressWarnings(aggte(mp_custom, type = "simple"))
  gl <- glance(agg_custom)
  expect_s3_class(gl, "data.frame")
  expect_equal(nrow(gl), 1)
  expect_equal(gl$est.method, "custom")
})

#-----------------------------------------------------------------------------
#
# These are tests for the inference procedure for computing ATT(g,t)'s and
# different aggregation methods.  These tests are meant to run fast ---
# they only confirm that standard errors are not changing across different
# iterations of the code.  More extensive/direct tests are available at
# did/tests/att_gt_inference_tests.Rmd.
#
# IMPORTANT: For these tests to run, version 2.1.2 of the did package should
# be available.
#
# The following code will install version 2.1.2 of the did package into
# the directory above:
# install.packages("did", version="2.1.2", lib=old_did_path)
#
#-----------------------------------------------------------------------------


same_matrix_elem <- function(A, B) {
  stopifnot(identical(dim(A), dim(B)))
  colnames(A) <- NULL
  colnames(B) <- NULL
  dense_sort <- function(M) apply(as(M, "matrix"), 2, sort, na.last = TRUE)
  all.equal(dense_sort(A), dense_sort(B))
}

temp_lib <- tempfile()
dir.create(temp_lib)
withr::defer(unlink(temp_lib, recursive = TRUE), teardown_env())

old_did_available <- FALSE
if (!identical(Sys.getenv("NOT_CRAN"), "false")) {
  old_did_available <- tryCatch({
    # The install can fail with warnings (not errors), e.g. when the dependency
    # chain cannot compile from source, so suppress them and verify the result
    # directly on disk: requireNamespace(lib.loc=) would spuriously return TRUE
    # because the dev `did` namespace is already loaded by devtools::test().
    suppressWarnings(suppressMessages(
      remotes::install_version("did", version = "2.1.2", lib = temp_lib,
                               repos = "https://cloud.r-project.org", quiet = TRUE)
    ))
    desc <- file.path(temp_lib, "did", "DESCRIPTION")
    file.exists(desc) &&
      identical(as.character(read.dcf(desc, "Version")), "2.1.2")
  }, error = function(e) FALSE)
}

test_that("inference with balanced panel data and aggregations", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  # tryCatch(detach("package:did"), error=function(e) "")

  # library(did, lib.loc="~/R/old_packages/")

  # first: make sure that we are using right version of package for these estimates
  # expect_true(packageVersion("did") == "2.1.2", "wrong version of package")

  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  expect_true(packageVersion("did") != "2.1.2", "wrong version of package")
  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr"
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg"
  )
  # reg
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw"
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))
  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc, dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc, ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .01)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .01)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .01)

  # checks for aggregations
  set.seed(1234)
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw"
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg"
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr"
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  dyn_new <- aggte(ipw_new, type = "dynamic")
  group_new <- aggte(reg_new, type = "group")
  cal_new <- aggte(dr_new, type = "calendar")


  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .01)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .01)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .01)
})


test_that("inference with clustering", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster"
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )


  expect_true(packageVersion("did") != "2.1.2", "wrong version of package")
  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", clustervars = "cluster"
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg", clustervars = "cluster"
  )
  # reg
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw", clustervars = "cluster"
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))

  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc, dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc, ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .01)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .01)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .01)

  # aggregations
  set.seed(1234)
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster"
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster"
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster"
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  dyn_new <- aggte(ipw_new, type = "dynamic")
  group_new <- aggte(reg_new, type = "group")
  cal_new <- aggte(dr_new, type = "calendar")

  # checks for aggregations
  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .01)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .01)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .01)
})

test_that("same inference with unbalanced panel and panel data", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)

  res_factor <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", clustervars = "cluster"
  )

  #-----------------------------------------------------------------------------
  # clustered standard errors with unbalanced panel
  res_ub <- att_gt(
    yname = "Y",
    tname = "period",
    idname = "id",
    gname = "G",
    xformla = ~X,
    data = data,
    panel = TRUE,
    allow_unbalanced_panel = TRUE,
    clustervars = "cluster"
  )

  # check that influence function is the same if we use unbalanced panel approach
  # vs. balanced panel approach
  # Note: differences are showing up in never-treated units (that are towards end of sample)
  expect_true(all(res_factor$inffunc == res_ub$inffunc))
})


test_that("inference with repeated cross sections", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp, panel = FALSE)

  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  expect_true(packageVersion("did") != "2.1.2", "wrong version of package")
  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", panel = FALSE
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg", panel = FALSE
  )
  # reg
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw", panel = FALSE
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))
  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc, dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc, ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  # upping the tolerance because repeated cross sections
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .05)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .05)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .05)

  # aggregations
  set.seed(1234)
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", panel = FALSE
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", panel = FALSE
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", panel = FALSE
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  dyn_new <- aggte(ipw_new, type = "dynamic")
  group_new <- aggte(reg_new, type = "group")
  cal_new <- aggte(dr_new, type = "calendar")


  # checks for aggregations
  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .05)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .05)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .05)
})


test_that("inference with repeated cross sections and clustering", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp, panel = FALSE)

  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster", panel = FALSE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", clustervars = "cluster", panel = FALSE
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg", clustervars = "cluster", panel = FALSE
  )
  # reg
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw", clustervars = "cluster", panel = FALSE
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))
  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc, dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc, ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  # upping the tolerance here because of repeated cross sections & clustering
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .05)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .05)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .05)

  # aggregations
  set.seed(1234)
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster", panel = FALSE
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster", panel = FALSE
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster", panel = FALSE
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  dyn_new <- aggte(ipw_new, type = "dynamic")
  group_new <- aggte(reg_new, type = "group")
  cal_new <- aggte(dr_new, type = "calendar")


  # checks for aggregations
  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .05)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .05)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .05)
})



test_that("inference with unbalanced panel", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data <- data[-3, ]

  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", panel = TRUE, allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", panel = TRUE, allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", panel = TRUE, allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  expect_true(packageVersion("did") != "2.1.2", "wrong version of package")
  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", panel = TRUE, allow_unbalanced_panel = TRUE
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg", panel = TRUE, allow_unbalanced_panel = TRUE
  )
  # reg
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw", panel = TRUE, allow_unbalanced_panel = TRUE
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))
  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc, dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc, ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .01)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .01)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .01)

  # aggregations — set seed inside each subprocess for reproducible bootstrap SEs
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", panel = TRUE, allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", panel = TRUE, allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", panel = TRUE, allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  set.seed(1234)
  dyn_new <- aggte(ipw_new, type = "dynamic")
  set.seed(1234)
  group_new <- aggte(reg_new, type = "group")
  set.seed(1234)
  cal_new <- aggte(dr_new, type = "calendar")


  # checks for aggregations
  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .01)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .01)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .01)
})

test_that("inference with unbalanced panel and clustering", {
  skip_on_cran()
  skip_if(!old_did_available, "did v2.1.2 not available from CRAN")
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data <- data[-3, ]


  set.seed(1234)
  # dr
  dr_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  # reg
  reg_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  # ipw
  ipw_2.1.2 <- callr::r(
    function(data, temp_lib) {
      library(did, lib.loc = temp_lib)
      att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  expect_true(packageVersion("did") != "2.1.2", "wrong version of package")
  set.seed(1234)
  # dr
  dr_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "dr", clustervars = "cluster", allow_unbalanced_panel = TRUE
  )
  # reg
  reg_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "reg", clustervars = "cluster", allow_unbalanced_panel = TRUE
  )
  # ipw
  ipw_new <- att_gt(
    yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
    gname = "G", est_method = "ipw", clustervars = "cluster", allow_unbalanced_panel = TRUE
  )

  # checks for ATT(g,t)'s
  expect_true(all.equal(dr_2.1.2$att, dr_new$att))
  expect_true(all.equal(reg_2.1.2$att, reg_new$att))
  expect_true(all.equal(ipw_2.1.2$att, ipw_new$att))
  # check that the influence function is the same
  expect_true(same_matrix_elem(dr_new$inffunc,  dr_2.1.2$inffunc))
  expect_true(same_matrix_elem(reg_new$inffunc, reg_2.1.2$inffunc))
  expect_true(same_matrix_elem(ipw_new$inffunc,  ipw_2.1.2$inffunc))

  # standard errors should be close
  # not totally sure, but I think slight differences are expected
  # perhaps from implementing the multiplier on the C++ side
  # in newer versions of the code
  expect_equal(dr_2.1.2$se[1], dr_new$se[1], tol = .01)
  expect_equal(reg_2.1.2$se[1], reg_new$se[1], tol = .01)
  expect_equal(ipw_2.1.2$se[1], ipw_new$se[1], tol = .01)

  # aggregations — set seed inside each subprocess for reproducible bootstrap SEs
  dyn_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "ipw", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "dynamic")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  group_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "reg", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "group")
    },
    args = list(data = data, temp_lib = temp_lib)
  )
  cal_2.1.2 <- callr::r(
    function(data, temp_lib) {
      set.seed(1234)
      library(did, lib.loc = temp_lib)
      res <- att_gt(
        yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
        gname = "G", est_method = "dr", clustervars = "cluster", allow_unbalanced_panel = TRUE
      )
      aggte(res, type = "calendar")
    },
    args = list(data = data, temp_lib = temp_lib)
  )

  set.seed(1234)
  dyn_new <- aggte(ipw_new, type = "dynamic")
  set.seed(1234)
  group_new <- aggte(reg_new, type = "group")
  set.seed(1234)
  cal_new <- aggte(dr_new, type = "calendar")

  # checks for aggregations
  expect_true(all(dyn_2.1.2$inffunc == dyn_new$inffunc))
  expect_true(all(group_2.1.2$inffunc == group_new$inffunc))
  expect_true(all(cal_2.1.2$inffunc == cal_new$inffunc))

  # standard errors for aggregations
  expect_equal(dyn_2.1.2$se[1], dyn_new$se[1], tol = .01)
  expect_equal(group_2.1.2$se[1], group_new$se[1], tol = .01)
  expect_equal(cal_2.1.2$se[1], cal_new$se[1], tol = .01)
})

# =============================================================================
# Tests for stop(), warning(), and message() conditions
# =============================================================================

# Shared setup
set.seed(20260401)
sp <- did::reset.sim()
data_eh <- did::build_sim_dataset(sp)

# =============================================================================
# att_gt() validation errors
# =============================================================================

test_that("att_gt errors on invalid est_method string", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = "bad", bstrap = FALSE),
    "must be one of"
  )
})

test_that("att_gt errors on non-character non-function est_method", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = 42, bstrap = FALSE),
    "must be a character string"
  )
})

test_that("att_gt errors on invalid fix_weights value", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", fix_weights = "bad", bstrap = FALSE),
    "must be NULL or one of"
  )
})

test_that("att_gt rejects non-exact control_group and base_period values in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", control_group = "NotYetTreated", faster_mode = fm,
             bstrap = FALSE),
      "control_group must be either"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", base_period = "Universal", faster_mode = fm,
             bstrap = FALSE),
      "base_period must be either"
    )
  }
})

test_that("att_gt rejects negative or non-numeric anticipation in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = -1, faster_mode = fm, bstrap = FALSE),
      "anticipation must be non-negative"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = "1", faster_mode = fm, bstrap = FALSE),
      "anticipation must be numeric"
    )
  }
})

test_that("att_gt rejects argument-referenced internal variable names in both modes", {
  reserved_names <- c(".w", ".rowid", ".G", ".C", "post",
                      "asif_never_treated", "treated_first_period")
  reserved_data <- data_eh
  for (nm in reserved_names) {
    reserved_data[[nm]] <- seq_len(nrow(reserved_data))
  }
  msg <- "reserved for internal use"

  for (fm in c(FALSE, TRUE)) {
    for (nm in reserved_names) {
      expect_error(
        att_gt(yname = nm, data = reserved_data, tname = "period",
               idname = "id", gname = "G", faster_mode = fm,
               bstrap = FALSE),
        msg
      )
      expect_error(
        att_gt(yname = "Y", xformla = as.formula(paste("~", nm)),
               data = reserved_data, tname = "period", idname = "id",
               gname = "G", faster_mode = fm, bstrap = FALSE),
        msg
      )
      expect_error(
        att_gt(yname = "Y", data = reserved_data, tname = "period",
               idname = "id", gname = "G", weightsname = nm,
               faster_mode = fm, bstrap = FALSE),
        msg
      )
    }
  }
})

test_that("att_gt errors on panel=TRUE without idname in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "Must provide idname when panel = TRUE"
    )
  }
})

test_that("att_gt still runs with panel=FALSE and no idname in both modes", {
  res <- list()
  for (fm in c(FALSE, TRUE)) {
    res[[as.character(fm)]] <- suppressWarnings(
      att_gt(yname = "Y", data = data_eh, tname = "period", gname = "G",
             panel = FALSE, bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res[[as.character(fm)]]$att)))
  }
  expect_equal(res[["FALSE"]]$att, res[["TRUE"]]$att, tolerance = 1e-10)
})

test_that("att_gt errors on invalid alp", {
  for (bad_alp in list(1.5, 0, 1, -0.05, c(0.05, 0.1), "0.05", NA_real_)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, alp = bad_alp),
      "alp must be a single number strictly between 0 and 1"
    )
  }
})

test_that("att_gt errors on invalid biters when bootstrapping", {
  for (bad_biters in list(-5, 0, 2.5, c(100, 200), "100", NA_real_)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", bstrap = TRUE, biters = bad_biters),
      "biters must be a single positive whole number"
    )
  }
  # biters is not used (and so not validated) when bstrap = FALSE
  res <- suppressWarnings(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE, biters = -5)
  )
  expect_s3_class(res, "MP")
})

test_that("att_gt errors on fix_weights with panel=FALSE", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", fix_weights = "base_period", panel = FALSE,
           bstrap = FALSE),
    "not supported for repeated cross sections"
  )
})

test_that("att_gt warns on extra args with built-in est_method", {
  expect_warning(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = "reg", bstrap = FALSE, extra_arg = 1),
    "Extra arguments"
  )
})

test_that("att_gt messages about anticipation", {
  expect_message(
    suppressWarnings(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = 1, bstrap = FALSE)
    ),
    "anticipation ="
  )
})

test_that("att_gt computes clustered SEs without the bootstrap (no 'requires bootstrap' warning)", {
  # Clustering at the individual level (idname) without the bootstrap returns the unit-level analytical
  # standard errors -- no "requires bootstrap" warning. A coarser cluster variable yields the analytical
  # cluster-robust SE (covered in test-cluster-analytic.R).
  w <- capture_warnings(
    res <- att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
                  gname = "G", clustervars = "id", bstrap = FALSE)
  )
  expect_false(any(grepl("Clustered standard errors require", w)))
  expect_false(any(grepl("could not be computed analytically", w)))
  expect_true(any(is.finite(res$se)))
})

# =============================================================================
# pre_process_did / pre_process_did2 validation errors
# =============================================================================

test_that("att_gt errors on missing column name (slower mode)", {
  expect_error(
    att_gt(yname = "nonexistent", data = data_eh, tname = "period",
           idname = "id", gname = "G", bstrap = FALSE, faster_mode = FALSE),
    "not found"
  )
})

test_that("att_gt errors on missing column name (faster_mode)", {
  expect_error(
    att_gt(yname = "nonexistent", data = data_eh, tname = "period",
           idname = "id", gname = "G", bstrap = FALSE, faster_mode = TRUE),
    "nonexistent"
  )
})

test_that("att_gt errors on non-numeric tname", {
  bad_data <- data_eh
  bad_data$period <- as.character(bad_data$period)
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "must be numeric"
  )
})

test_that("att_gt errors on non-numeric gname", {
  bad_data <- data_eh
  bad_data$G <- as.character(bad_data$G)
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "must be numeric"
  )
})

test_that("att_gt errors on non-numeric outcome variable in both modes", {
  bad_chr <- data_eh
  bad_chr$Y <- as.character(bad_chr$Y)
  bad_fac <- data_eh
  bad_fac$Y <- factor(ifelse(bad_fac$Y > 0, "hi", "lo"))
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = bad_chr, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "The outcome variable 'Y' must be numeric"
    )
    expect_error(
      att_gt(yname = "Y", data = bad_fac, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "The outcome variable 'Y' must be numeric"
    )
  }
})

test_that("att_gt still accepts logical and integer outcomes in both modes", {
  log_data <- data_eh
  log_data$Y <- log_data$Y > median(log_data$Y)
  int_data <- data_eh
  int_data$Y <- as.integer(round(int_data$Y))
  for (fm in c(FALSE, TRUE)) {
    res_log <- suppressWarnings(
      att_gt(yname = "Y", data = log_data, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res_log$att)))
    res_int <- suppressWarnings(
      att_gt(yname = "Y", data = int_data, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res_int$att)))
  }
})

test_that("att_gt errors on treatment reversals (faster_mode)", {
  bad_data <- data_eh
  # Make one unit switch groups across time
  target_id <- bad_data$id[1]
  periods <- sort(unique(bad_data$period))
  bad_data$G[bad_data$id == target_id & bad_data$period == periods[1]] <- 3
  bad_data$G[bad_data$id == target_id & bad_data$period == periods[2]] <- 4
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE, faster_mode = TRUE),
    "same across all periods|time-invariant|must be the same"
  )
})

test_that("att_gt warns on missing data dropped", {
  bad_data <- data_eh
  bad_data$Y[1:5] <- NA
  expect_warning(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "dropped|missing"
  )
})

test_that("att_gt warns on small groups", {
  # Create data with a very small treated group
  small_data <- data_eh
  treated_ids <- unique(small_data$id[small_data$G > 0])
  # Keep only 2 treated units and some controls
  keep_ids <- c(treated_ids[1:2], unique(small_data$id[small_data$G == 0])[1:50])
  small_data <- small_data[small_data$id %in% keep_ids, ]
  expect_warning(
    att_gt(yname = "Y", data = small_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "very few observations"
  )
})

test_that("att_gt handles data with .w column (slower mode)", {
  # The .w column gets dropped during column selection in pre_process_did
  # before the collision check, so this should run without error
  bad_data <- data.frame(data_eh)
  bad_data$.w <- 1
  result <- att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE, faster_mode = FALSE)
  expect_s3_class(result, "MP")
})

test_that("att_gt messages on time-varying weights (panel)", {
  tv_data <- data_eh
  tv_data$tw <- tv_data$period + runif(nrow(tv_data))
  expect_message(
    att_gt(yname = "Y", data = tv_data, tname = "period", idname = "id",
           gname = "G", weightsname = "tw", bstrap = FALSE),
    "Time-varying weights"
  )
})

test_that("clustervars contract enforced in all faster_mode x bstrap combinations", {
  # ?att_gt promises an error for more than two cluster variables (one of which
  # must be idname). Both pre-processors must reject the input with identical
  # wording (also matching mboot()) regardless of bstrap -- the analytical
  # (bstrap = FALSE) slow path used to silently cluster on the first extra
  # variable only.
  d <- data_eh
  d$cl1 <- d$id %% 10  # time-invariant within unit
  d$cl2 <- d$id %% 7
  msg <- "At most one cluster variable \\(beyond 'idname'\\) is supported"
  for (fm in c(FALSE, TRUE)) {
    for (bs in c(FALSE, TRUE)) {
      # three cluster variables (idname + two extras)
      expect_error(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("id", "cl1", "cl2"),
               bstrap = bs, faster_mode = fm),
        msg
      )
      # two cluster variables, neither of which is idname
      expect_error(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("cl1", "cl2"),
               bstrap = bs, faster_mode = fm),
        msg
      )
    }
  }
})

test_that("one extra cluster variable still works in all faster_mode x bstrap combinations", {
  # clustered DGP: units within a cluster share a common per-period shock, so
  # the clustered SE genuinely differs from the unclustered one
  set.seed(20260609)
  n_cl <- 40L
  sz <- rep(c(2L, 4L), length.out = n_cl)
  N <- sum(sz)
  cl <- rep(seq_len(n_cl), times = sz)
  gC <- sample(c(2L, 3L, 0L), n_cl, replace = TRUE)[cl]
  eta <- matrix(rnorm(n_cl * 4L, 0, 1.5), n_cl, 4L)
  d <- data.frame(id = rep(seq_len(N), each = 4L), period = rep(1:4, N),
                  cl = rep(cl, each = 4L), G = rep(gC, each = 4L))
  d$Y <- eta[cbind(d$cl, d$period)] + 0.3 * d$period +
    (d$G != 0 & d$period >= d$G) + rnorm(nrow(d))
  for (fm in c(FALSE, TRUE)) {
    se_cl <- NULL
    for (bs in c(FALSE, TRUE)) {
      res <- suppressWarnings(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("id", "cl"), bstrap = bs,
               biters = 100, cband = FALSE, faster_mode = fm)
      )
      expect_s3_class(res, "MP")
      expect_true(any(is.finite(res$se)))
      if (!bs) se_cl <- res$se
    }
    # the analytical (bstrap = FALSE) path actually clusters: the SE differs
    # from the unclustered analytical SE
    res_iid <- att_gt(yname = "Y", data = d, tname = "period", idname = "id",
                      gname = "G", bstrap = FALSE, faster_mode = fm)
    ok <- is.finite(se_cl) & is.finite(res_iid$se)
    expect_true(any(ok))
    expect_gt(max(abs(se_cl[ok] - res_iid$se[ok])), 1e-8)
  }
})

# =============================================================================
# aggte() validation errors
# =============================================================================

test_that("aggte errors on invalid type", {
  mp_tmp <- att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE)
  expect_error(
    aggte(mp_tmp, type = "invalid"),
    "must be one of"
  )
})

test_that("aggte errors when ATTs contain NA and na.rm=FALSE", {
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  mp_na <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  ))
  # Inject NA directly — no need to rely on estimation failure
  mp_na$att[1] <- NA
  expect_error(
    aggte(mp_na, type = "dynamic", na.rm = FALSE),
    "Missing values"
  )
})

# =============================================================================
# Wald pre-test warnings
# =============================================================================

test_that("att_gt handles singular covariance for Wald test gracefully", {
  # When covariance matrix is singular, W and Wpval should be NULL
  # and the result should still be a valid MP object
  small_sp <- did::reset.sim()
  small_sp$n <- 50
  small_data <- did::build_sim_dataset(small_sp)
  result <- suppressWarnings(
    att_gt(yname = "Y", data = small_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
  # W should be either a valid number or NULL (when singular)
  expect_true(is.null(result$W) || is.numeric(result$W))
})

# =============================================================================
# compute.att_gt warnings
# =============================================================================

test_that("att_gt warns on overlap violations", {
  # Create data where propensity score is near 1 (near-perfect separation)
  sep_data <- data_eh
  # Make covariate perfectly predict treatment for some cells
  sep_data$X_sep <- ifelse(sep_data$G > 0, 100, -100)
  expect_warning(
    att_gt(yname = "Y", xformla = ~X_sep, data = sep_data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE),
    "overlap condition"
  )
})

test_that("slow path warns once per failed cell, matching fast-path wording, with accurate Wald diagnosis", {
  # Each failed (g,t) cell must produce exactly ONE warning with identical text in
  # both modes: the tryCatch wrapper must not re-warn the internal overlap/singular
  # sentinels after the diagnostic warning already fired. And when pre-treatment
  # cells existed but were all dropped for NA/zero variance, the Wald warning must
  # say so instead of claiming no pre-treatment cells exist.
  set.seed(20260610)
  d <- expand.grid(id = 1:200, period = 1:3)
  gvals <- c(0, 2, 3)
  d$G <- gvals[(d$id - 1) %% 3 + 1]
  d$Xsep <- 1 * (d$G > 0)  # separates treated from never-treated: overlap fails in all 4 cells
  d$Y <- 0.1 * d$period + (d$G > 0 & d$period >= d$G) + rnorm(nrow(d), 0, 0.5)

  w_slow <- capture_warnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~Xsep, data = d, tname = "period", idname = "id",
           gname = "G", est_method = "dr", faster_mode = FALSE, bstrap = FALSE)))
  w_fast <- capture_warnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~Xsep, data = d, tname = "period", idname = "id",
           gname = "G", est_method = "dr", faster_mode = TRUE, bstrap = FALSE)))

  # identical warning sets across modes (one per failed cell + the Wald warning)
  expect_identical(w_slow, w_fast)
  # exactly one warning per failed cell, with the fast-path wording
  expect_identical(sum(grepl("overlap condition violated for group", w_slow)), 4L)
  # no leaked internal sentinel via the tryCatch wrapper
  expect_false(any(grepl("Error computing internal 2x2 DiD", w_slow)))
  # accurate Wald diagnosis: pre-treatment cells existed but had NA/zero variance
  expect_true(any(grepl("missing or zero variance", w_slow)))
  expect_false(any(grepl("treated early in the panel", w_slow)))
})

test_that("att_gt warns when no pre-treatment periods for Wald test", {
  # Create data where a group is first treated in period 2 (only 1 pre-treatment period)
  # With base_period="varying", there may be no pre-treatment ATTs for the Wald test
  early_data <- data.frame(
    id = rep(1:100, each = 3),
    period = rep(1:3, 100),
    G = rep(c(rep(2, 50), rep(0, 50)), each = 3),
    Y = rnorm(300)
  )
  expect_warning(
    att_gt(yname = "Y", data = early_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "pre-treatment|Wald"
  )
})

test_that("empty-cell warning names the actually-empty base period under base_period='universal' in both modes", {
  # Group 3's universal base period is 2; group 3 has observations in every
  # OTHER period. The slow path used to name the cell's current period
  # (1/3/4) -- periods where the group does have units -- instead of the empty
  # base period (2), which the fast path reported correctly.
  set.seed(20260610)
  d <- expand.grid(rep = 1:30, period = 1:4, G = c(0, 3))
  d$Y <- rnorm(nrow(d)) + 0.1 * d$period + (d$G > 0 & d$period >= d$G)
  d <- d[!(d$G == 3 & d$period == 2), ]  # group 3 empty only in its base period
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", gname = "G",
             panel = FALSE, base_period = "universal", bstrap = FALSE,
             faster_mode = fm)
    ))
    # one warning per non-base cell (t = 1, 3, 4), each naming period 2
    expect_identical(
      sum(grepl("No units in group 3 in time period 2", w, fixed = TRUE)), 3L,
      info = paste("faster_mode =", fm)
    )
    expect_false(any(grepl("No units in group 3 in time period [134]", w)),
                 info = paste("faster_mode =", fm))
  }
})

# =============================================================================
# pre-processing warning parity across modes
# =============================================================================

test_that("'no never-treated group' warning is identical across modes and discloses the period filtering", {
  # Both modes drop all periods >= (latest cohort - anticipation); the fast-mode
  # warning used to omit that a chunk of the data was discarded.
  set.seed(20260609)
  d <- data.frame(
    id = rep(1:60, each = 4),
    period = rep(1:4, 60),
    G = rep(c(rep(2, 30), rep(4, 30)), each = 4)
  )
  d$Y <- rnorm(nrow(d))
  msgs <- list()
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", idname = "id",
             gname = "G", control_group = "nevertreated", bstrap = FALSE,
             faster_mode = fm)
    ))
    m <- grep("No never-treated group", w, value = TRUE)
    expect_length(m, 1L)
    expect_match(m, "filtered out", fixed = TRUE)
    msgs[[as.character(fm)]] <- m
  }
  expect_identical(msgs[["FALSE"]], msgs[["TRUE"]])
})

test_that("balanced-panel coercion warning counts dropped units identically in both modes", {
  # 5 units each miss one period's row: 5 UNITS (20 rows) are dropped while
  # coercing to a balanced panel. The slow path used to report the unit count
  # as 'observations' (a 4x undercount of the rows actually removed).
  drop_ids <- unique(data_eh$id)[1:5]
  d <- data_eh[!(data_eh$id %in% drop_ids & data_eh$period == max(data_eh$period)), ]
  expected <- "5 units are missing in some periods. Converting to balanced panel by dropping them."
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", idname = "id",
             gname = "G", panel = TRUE, allow_unbalanced_panel = FALSE,
             bstrap = FALSE, faster_mode = fm)
    ))
    expect_true(any(w == expected), info = paste("faster_mode =", fm))
    expect_false(any(grepl("observations while converting", w)),
                 info = paste("faster_mode =", fm))
  }
})

# =============================================================================
# Tests for stop(), warning(), and message() conditions
# =============================================================================

# Shared setup
set.seed(20260401)
sp <- did::reset.sim()
data_eh <- did::build_sim_dataset(sp)

# =============================================================================
# att_gt() validation errors
# =============================================================================

test_that("att_gt errors on invalid est_method string", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = "bad", bstrap = FALSE),
    "must be one of"
  )
})

test_that("att_gt errors on non-character non-function est_method", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = 42, bstrap = FALSE),
    "must be a character string"
  )
})

test_that("att_gt errors on invalid fix_weights value", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", fix_weights = "bad", bstrap = FALSE),
    "must be NULL or one of"
  )
})

test_that("att_gt rejects non-exact control_group and base_period values in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", control_group = "NotYetTreated", faster_mode = fm,
             bstrap = FALSE),
      "control_group must be either"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", control_group = c("nevertreated", "notyettreated"),
             faster_mode = fm, bstrap = FALSE),
      "control_group must be either"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", base_period = "Universal", faster_mode = fm,
             bstrap = FALSE),
      "base_period must be either"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", base_period = c("varying", "universal"),
             faster_mode = fm, bstrap = FALSE),
      "base_period must be either"
    )
  }
})

test_that("att_gt rejects negative or non-numeric anticipation in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = -1, faster_mode = fm, bstrap = FALSE),
      "anticipation must be a non-negative whole number"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = "1", faster_mode = fm, bstrap = FALSE),
      "anticipation must be numeric"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = c(0, 1), faster_mode = fm, bstrap = FALSE),
      "anticipation must be a single finite non-missing number"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = NA_real_, faster_mode = fm, bstrap = FALSE),
      "anticipation must be a single finite non-missing number"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = Inf, faster_mode = fm, bstrap = FALSE),
      "anticipation must be a single finite non-missing number"
    )
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = 1.5, faster_mode = fm, bstrap = FALSE),
      "anticipation must be a non-negative whole number"
    )
  }
})

test_that("att_gt rejects invalid scalar logical controls before base R errors", {
  bad_args <- list(
    panel = NA,
    allow_unbalanced_panel = NA,
    bstrap = NA,
    cband = NA,
    faster_mode = NA,
    print_details = NA,
    pl = "yes"
  )

  for (nm in names(bad_args)) {
    args <- list(yname = "Y", data = data_eh, tname = "period",
                 idname = "id", gname = "G", bstrap = FALSE)
    args[[nm]] <- bad_args[[nm]]
    expect_error(
      do.call(att_gt, args),
      paste0(nm, " must be a single logical"),
      info = nm
    )
  }
})

test_that("att_gt rejects invalid cores before parallel code sees it", {
  for (bad_cores in list(0, -1, 1.5, c(1, 2), "2", NA_real_, Inf)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", cores = bad_cores, bstrap = FALSE),
      "cores must be a single positive whole number"
    )
  }
})

test_that("att_gt rejects argument-referenced internal variable names in both modes", {
  reserved_names <- c(".w", ".rowid", ".G", ".C", "post",
                      "asif_never_treated", "treated_first_period")
  reserved_data <- data_eh
  for (nm in reserved_names) {
    reserved_data[[nm]] <- seq_len(nrow(reserved_data))
  }
  msg <- "reserved for internal use"

  for (fm in c(FALSE, TRUE)) {
    for (nm in reserved_names) {
      expect_error(
        att_gt(yname = nm, data = reserved_data, tname = "period",
               idname = "id", gname = "G", faster_mode = fm,
               bstrap = FALSE),
        msg
      )
      expect_error(
        att_gt(yname = "Y", xformla = as.formula(paste("~", nm)),
               data = reserved_data, tname = "period", idname = "id",
               gname = "G", faster_mode = fm, bstrap = FALSE),
        msg
      )
      expect_error(
        att_gt(yname = "Y", data = reserved_data, tname = "period",
               idname = "id", gname = "G", weightsname = nm,
               faster_mode = fm, bstrap = FALSE),
        msg
      )
    }
  }
})

test_that("att_gt rejects invalid xformla before formula internals in both modes", {
  for (fm in c(FALSE, TRUE)) {
    for (bad_xformla in list(NA, 1, "~X", list(~X))) {
      expect_error(
        att_gt(yname = "Y", data = data_eh, tname = "period",
               idname = "id", gname = "G", xformla = bad_xformla,
               faster_mode = fm, bstrap = FALSE),
        "xformla must be NULL or a formula",
        info = paste("faster_mode", fm)
      )
    }
  }
})

test_that("slow path rejects malformed column-name arguments before ambiguous indexing", {
  bad_args <- list(
    yname = c("Y", "X"),
    tname = NA_character_,
    idname = c("id", "id"),
    gname = c("G", "G"),
    weightsname = c("w1", "w2"),
    clustervars = NA_character_
  )

  for (nm in names(bad_args)) {
    args <- list(yname = "Y", data = data_eh, tname = "period",
                 idname = "id", gname = "G", bstrap = FALSE,
                 faster_mode = FALSE)
    args[[nm]] <- bad_args[[nm]]
    expect_error(
      do.call(att_gt, args),
      paste0(nm, " must|", nm, " contains"),
      info = nm
    )
  }
})

test_that("att_gt drops rows with missing gname and non-finite numeric inputs in both modes", {
  cases <- list(
    gname_missing = within(data_eh, G[1] <- NA_real_),
    outcome_infinite = within(data_eh, Y[1] <- Inf),
    weight_infinite = within(data_eh, {
      w <- rep(1, nrow(data_eh))
      w[1] <- Inf
    }),
    covariate_infinite = within(data_eh, X[1] <- Inf)
  )

  for (fm in c(FALSE, TRUE)) {
    for (nm in names(cases)) {
      args <- list(yname = "Y", data = cases[[nm]], tname = "period",
                   idname = "id", gname = "G", bstrap = FALSE,
                   faster_mode = fm, panel = FALSE)
      if (nm == "weight_infinite") args$weightsname <- "w"
      if (nm == "covariate_infinite") args$xformla <- ~X
      expect_warning(
        res <- do.call(att_gt, args),
        "missing or non-finite data",
        info = paste(nm, "faster_mode", fm)
      )
      expect_s3_class(res, "MP")
    }
  }
})

test_that("att_gt errors on panel=TRUE without idname in both modes", {
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "Must provide idname when panel = TRUE"
    )
  }
})

test_that("att_gt still runs with panel=FALSE and no idname in both modes", {
  res <- list()
  for (fm in c(FALSE, TRUE)) {
    res[[as.character(fm)]] <- suppressWarnings(
      att_gt(yname = "Y", data = data_eh, tname = "period", gname = "G",
             panel = FALSE, bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res[[as.character(fm)]]$att)))
  }
  expect_equal(res[["FALSE"]]$att, res[["TRUE"]]$att, tolerance = 1e-10)
})

test_that("att_gt errors on invalid alp", {
  for (bad_alp in list(1.5, 0, 1, -0.05, c(0.05, 0.1), "0.05", NA_real_)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, alp = bad_alp),
      "alp must be a single number strictly between 0 and 1"
    )
  }
})

test_that("att_gt errors on invalid biters when bootstrapping", {
  for (bad_biters in list(-5, 0, 2.5, c(100, 200), "100", NA_real_, Inf)) {
    expect_error(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", bstrap = TRUE, biters = bad_biters),
      "biters must be a single positive whole number"
    )
  }
  # biters is not used (and so not validated) when bstrap = FALSE
  res <- suppressWarnings(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE, biters = -5)
  )
  expect_s3_class(res, "MP")
})

test_that("simulation helpers reject invalid scalar controls before raw R errors", {
  expect_error(did::reset.sim(time.periods = NA_integer_),
               "time.periods must be a single positive whole number")
  expect_error(did::reset.sim(n = 0),
               "n must be a single positive whole number")
  expect_error(did::reset.sim(ipw = NA),
               "ipw must be a single logical")
  expect_error(did::reset.sim(reg = c(TRUE, FALSE)),
               "reg must be a single logical")

  expect_error(did::build_sim_dataset(sp, panel = NA),
               "panel must be a single logical")
  expect_error(did::build_sim_dataset(1),
               "sp_list must be a list")
  sp_bad <- sp
  sp_bad$ipw <- NA
  expect_error(did::build_sim_dataset(sp_bad),
               "sp_list\\$ipw must be a single logical")
  sp_bad <- sp
  sp_bad$bett <- 1
  expect_error(did::build_sim_dataset(sp_bad),
               "sp_list\\$bett must be a numeric vector of length")
  sp_bad <- sp
  sp_bad$te.e[1] <- Inf
  expect_error(did::build_sim_dataset(sp_bad),
               "sp_list\\$te\\.e must be a numeric vector of length")
  sp_bad <- sp
  sp_bad$gamG <- sp_bad$gamG[-1]
  expect_error(did::build_sim_dataset(sp_bad),
               "sp_list\\$gamG must be a numeric vector of length")
  sp_bad <- sp
  sp_bad$te <- NA_real_
  expect_error(did::build_sim_dataset(sp_bad),
               "sp_list\\$te must be a single finite non-missing number")

  expect_error(did::sim(sp, ret = NA, bstrap = FALSE, cband = FALSE),
               "ret must be NULL or one of")
  expect_error(did::sim(sp, ret = c("Wpval", "cband"),
                        bstrap = FALSE, cband = FALSE),
               "ret must be NULL or one of")
  expect_error(did::sim(sp, bstrap = NA, cband = FALSE),
               "bstrap must be a single logical")
  expect_error(did::sim(sp, bstrap = FALSE, cband = NA),
               "cband must be a single logical")
})

test_that("trimmer rejects malformed exported utility arguments", {
  bad_control_groups <- list(NA_character_, c("notyettreated", "nevertreated"), "bad")
  for (bad_control_group in bad_control_groups) {
    expect_error(
      trimmer(3, "period", "id", "G", ~X, data_eh,
              control_group = bad_control_group),
      "control_group must be either"
    )
  }

  bad_thresholds <- list(NA_real_, c(0.9, 0.99), Inf, -1, "0.9")
  for (bad_threshold in bad_thresholds) {
    expect_error(
      trimmer(3, "period", "id", "G", ~X, data_eh,
              threshold = bad_threshold),
      "threshold must be a single finite number strictly between 0 and 1"
    )
  }

  for (bad_xformla in list(NA, "~X", 1, list(~X))) {
    expect_error(
      trimmer(3, "period", "id", "G", bad_xformla, data_eh),
      "xformla must be NULL or a formula"
    )
  }

  expect_error(trimmer(c(3, 4), "period", "id", "G", ~X, data_eh),
               "g must be a single finite non-missing number")
  expect_error(trimmer(NA_real_, "period", "id", "G", ~X, data_eh),
               "g must be a single finite non-missing number")
  expect_error(trimmer(3, c("period", "period"), "id", "G", ~X, data_eh),
               "tname must be a single non-missing character string")
  expect_error(trimmer(3, "period", NA_character_, "G", ~X, data_eh),
               "idname must be a single non-missing character string")
  expect_error(trimmer(3, "period", "id", "missing", ~X, data_eh),
               "gname must be a character scalar and a name of a column")
})

test_that("test.mboot rejects malformed bootstrap inputs before recycling", {
  inf_func <- array(rnorm(10 * 2 * 5), c(10, 2, 5))
  dp <- list(data = data.frame(id = 1:10, period = 1L, cl = rep(1:2, each = 5)),
             idname = "id", clustervars = NULL, biters = 20,
             tname = "period", alp = 0.05, panel = TRUE)

  expect_error(test.mboot(matrix(rnorm(20), 10, 2), dp),
               "inf.func must be a numeric three-dimensional array")
  expect_error(test.mboot(array(numeric(0), c(0, 2, 5)), dp),
               "inf.func must be a numeric three-dimensional array")
  expect_error(test.mboot("bad", dp),
               "inf.func must be a numeric three-dimensional array")
  expect_error(test.mboot(inf_func, 1),
               "DIDparams must be a list")

  dp_bad <- dp
  dp_bad$idname <- "missing"
  expect_error(test.mboot(inf_func, dp_bad),
               "DIDparams\\$idname must be a character scalar and a name of a column")
  dp_bad <- dp
  dp_bad$clustervars <- 1
  expect_error(test.mboot(inf_func, dp_bad),
               "DIDparams\\$clustervars must be NULL or a character vector")
  dp_bad <- dp
  dp_bad$data <- dp_bad$data[-1, ]
  expect_error(test.mboot(inf_func, dp_bad),
               "inf.func first dimension must match")
})

test_that("mboot rejects malformed direct helper inputs before raw errors", {
  inf_func <- matrix(rnorm(10 * 2), 10, 2)
  dp <- list(data = data.frame(id = 1:10, period = 1L, cl = rep(1:2, each = 5)),
             idname = "id", clustervars = NULL, biters = 20,
             tname = "period", alp = 0.05, panel = TRUE,
             true_repeated_cross_sections = FALSE,
             allow_unbalanced_panel = FALSE, faster_mode = FALSE)

  expect_error(mboot(numeric(0), dp, return_V = FALSE),
               "inf.func must be a numeric matrix")
  expect_error(mboot(matrix(numeric(0), 0, 2), dp, return_V = FALSE),
               "inf.func must be a numeric matrix")
  expect_error(mboot("bad", dp, return_V = FALSE),
               "inf.func must be a numeric matrix")

  dp_bad <- dp
  dp_bad$clustervars <- 1
  expect_error(mboot(inf_func, dp_bad, return_V = FALSE),
               "clustervars must be NULL or a character vector")
  dp_bad <- dp
  dp_bad$clustervars <- "missing"
  expect_error(mboot(inf_func, dp_bad, return_V = FALSE),
               "clustervars contains column name")
  dp_bad <- dp
  dp_bad$clustervars <- "cl"
  dp_bad$data <- dp_bad$data[-1, ]
  expect_error(mboot(inf_func, dp_bad, return_V = FALSE),
               "cluster vector length must match")
})

test_that("process_attgt rejects malformed group-time result lists", {
  expect_error(process_attgt(1),
               "attgt.list must be a non-empty list")
  expect_error(process_attgt(list()),
               "attgt.list must be a non-empty list")
  expect_error(process_attgt(list(1)),
               "Each attgt.list element must be a list-like")
  expect_error(process_attgt(list(list(group = c(1, 2), year = 1, att = 0))),
               "numeric scalar 'group'")
  expect_error(process_attgt(list(list(group = 1, year = NA_real_, att = 0))),
               "numeric scalar 'year'")

  out <- process_attgt(list(list(group = 1, year = 2, att = NA_real_)))
  expect_equal(out$group, 1)
  expect_equal(out$tt, 2)
  expect_true(is.na(out$att))
})

test_that("aggte rejects invalid scalar controls before base R errors", {
  mp <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  ))
  expect_error(aggte(1),
               "MP must be an MP object produced by att_gt")
  expect_error(aggte(mp, type = "simple", na.rm = NA),
               "na.rm must be a single logical")
  expect_error(aggte(mp, type = "simple", bstrap = NA),
               "bstrap must be a single logical")
  expect_error(aggte(mp, type = "simple", cband = NA),
               "cband must be a single logical")
  expect_error(aggte(mp, type = "simple", alp = NA_real_),
               "alp must be a single number strictly between 0 and 1")
  expect_error(aggte(mp, type = "simple", bstrap = TRUE, biters = 0),
               "biters must be a single positive whole number")
  expect_error(aggte(mp, type = "dynamic", min_e = NA_real_),
               "min_e must be a single non-missing number")
  expect_error(aggte(mp, type = "dynamic", balance_e = c(0, 1)),
               "balance_e must be a single non-negative whole number")
  expect_error(aggte(mp, type = "dynamic", balance_e = -1),
               "balance_e must be a single non-negative whole number")
  expect_error(aggte(mp, type = "dynamic", balance_e = 0.5),
               "balance_e must be a single non-negative whole number")
  expect_error(aggte(mp, type = "dynamic", balance_e = Inf),
               "balance_e must be a single non-negative whole number")

  mp_bad <- mp
  mp_bad$inffunc <- mp_bad$inffunc[, -1, drop = FALSE]
  expect_error(aggte(mp_bad, type = "simple"),
               "inconsistent influence-function columns")
  mp_bad <- mp
  mp_bad$inffunc <- mp_bad$inffunc[-1, , drop = FALSE]
  expect_error(aggte(mp_bad, type = "simple"),
               "inconsistent influence-function rows")
})

test_that("plotting helpers reject invalid scalar controls before ggplot errors", {
  mp <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  ))
  expect_error(ggdid(mp, legend = NA),
               "legend must be a single logical")
  expect_error(ggdid(mp, theming = NA),
               "theming must be a single logical")
  expect_error(ggdid(mp, xgap = NA_real_),
               "xgap must be a single positive finite number")
  expect_error(ggdid(mp, ncol = NA_real_),
               "ncol must be a single positive whole number")

  agg <- aggte(mp, type = "group", cband = FALSE)
  expect_error(ggdid(agg, legend = NA),
               "legend must be a single logical")
  expect_error(ggdid(agg, ref_line = c(0, 1)),
               "ref_line must be a single non-missing number")
})

test_that("mboot rejects invalid scalar controls before bootstrap internals", {
  mp <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  ))
  inf <- mp$inffunc[, 1, drop = FALSE]
  dp <- mp$DIDparams
  dp$biters <- 10

  expect_error(mboot(inf, dp, pl = NA),
               "pl must be a single logical")
  expect_error(mboot(inf, dp, cores = NA_real_),
               "cores must be a single positive whole number")
  expect_error(mboot(inf, dp, return_V = NA),
               "return_V must be a single logical")

  dp_bad <- dp
  dp_bad$biters <- Inf
  expect_error(mboot(inf, dp_bad),
               "biters must be a single positive whole number")
  dp_bad <- dp
  dp_bad$alp <- NA_real_
  expect_error(mboot(inf, dp_bad),
               "alp must be a single number strictly between 0 and 1")
  dp_bad <- dp
  dp_bad$panel <- NA
  expect_error(mboot(inf, dp_bad),
               "DIDparams\\$panel must be a single logical")
})

test_that("att_gt errors on fix_weights with panel=FALSE", {
  expect_error(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", fix_weights = "base_period", panel = FALSE,
           bstrap = FALSE),
    "not supported for repeated cross sections"
  )
})

test_that("att_gt warns on extra args with built-in est_method", {
  expect_warning(
    att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
           gname = "G", est_method = "reg", bstrap = FALSE, extra_arg = 1),
    "Extra arguments"
  )
})

test_that("att_gt messages about anticipation", {
  expect_message(
    suppressWarnings(
      att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
             gname = "G", anticipation = 1, bstrap = FALSE)
    ),
    "anticipation ="
  )
})

test_that("att_gt computes clustered SEs without the bootstrap (no 'requires bootstrap' warning)", {
  # Clustering at the individual level (idname) without the bootstrap returns the unit-level analytical
  # standard errors -- no "requires bootstrap" warning. A coarser cluster variable yields the analytical
  # cluster-robust SE (covered in test-cluster-analytic.R).
  w <- capture_warnings(
    res <- att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
                  gname = "G", clustervars = "id", bstrap = FALSE)
  )
  expect_false(any(grepl("Clustered standard errors require", w)))
  expect_false(any(grepl("could not be computed analytically", w)))
  expect_true(any(is.finite(res$se)))
})

# =============================================================================
# pre_process_did / pre_process_did2 validation errors
# =============================================================================

test_that("att_gt errors on missing column name (slower mode)", {
  expect_error(
    att_gt(yname = "nonexistent", data = data_eh, tname = "period",
           idname = "id", gname = "G", bstrap = FALSE, faster_mode = FALSE),
    "character scalar and a name of a column"
  )
})

test_that("att_gt errors on missing column name (faster_mode)", {
  expect_error(
    att_gt(yname = "nonexistent", data = data_eh, tname = "period",
           idname = "id", gname = "G", bstrap = FALSE, faster_mode = TRUE),
    "nonexistent"
  )
})

test_that("att_gt errors on non-numeric tname", {
  bad_data <- data_eh
  bad_data$period <- as.character(bad_data$period)
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "must be numeric"
  )
})

test_that("att_gt errors on non-numeric gname", {
  bad_data <- data_eh
  bad_data$G <- as.character(bad_data$G)
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "must be numeric"
  )
})

test_that("att_gt errors on non-numeric outcome variable in both modes", {
  bad_chr <- data_eh
  bad_chr$Y <- as.character(bad_chr$Y)
  bad_fac <- data_eh
  bad_fac$Y <- factor(ifelse(bad_fac$Y > 0, "hi", "lo"))
  for (fm in c(FALSE, TRUE)) {
    expect_error(
      att_gt(yname = "Y", data = bad_chr, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "The outcome variable 'Y' must be numeric"
    )
    expect_error(
      att_gt(yname = "Y", data = bad_fac, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm),
      "The outcome variable 'Y' must be numeric"
    )
  }
})

test_that("att_gt still accepts logical and integer outcomes in both modes", {
  log_data <- data_eh
  log_data$Y <- log_data$Y > median(log_data$Y)
  int_data <- data_eh
  int_data$Y <- as.integer(round(int_data$Y))
  for (fm in c(FALSE, TRUE)) {
    res_log <- suppressWarnings(
      att_gt(yname = "Y", data = log_data, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res_log$att)))
    res_int <- suppressWarnings(
      att_gt(yname = "Y", data = int_data, tname = "period", idname = "id",
             gname = "G", bstrap = FALSE, faster_mode = fm)
    )
    expect_true(all(is.finite(res_int$att)))
  }
})

test_that("att_gt errors on treatment reversals (faster_mode)", {
  bad_data <- data_eh
  # Make one unit switch groups across time
  target_id <- bad_data$id[1]
  periods <- sort(unique(bad_data$period))
  bad_data$G[bad_data$id == target_id & bad_data$period == periods[1]] <- 3
  bad_data$G[bad_data$id == target_id & bad_data$period == periods[2]] <- 4
  expect_error(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE, faster_mode = TRUE),
    "same across all periods|time-invariant|must be the same"
  )
})

test_that("att_gt warns on missing data dropped", {
  bad_data <- data_eh
  bad_data$Y[1:5] <- NA
  expect_warning(
    att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "dropped|missing"
  )
})

test_that("att_gt warns on small groups", {
  # Create data with a very small treated group
  small_data <- data_eh
  treated_ids <- unique(small_data$id[small_data$G > 0])
  # Keep only 2 treated units and some controls
  keep_ids <- c(treated_ids[1:2], unique(small_data$id[small_data$G == 0])[1:50])
  small_data <- small_data[small_data$id %in% keep_ids, ]
  expect_warning(
    att_gt(yname = "Y", data = small_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "very few observations"
  )
})

test_that("att_gt handles data with .w column (slower mode)", {
  # The .w column gets dropped during column selection in pre_process_did
  # before the collision check, so this should run without error
  bad_data <- data.frame(data_eh)
  bad_data$.w <- 1
  result <- att_gt(yname = "Y", data = bad_data, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE, faster_mode = FALSE)
  expect_s3_class(result, "MP")
})

test_that("att_gt messages on time-varying weights (panel)", {
  tv_data <- data_eh
  tv_data$tw <- tv_data$period + runif(nrow(tv_data))
  expect_message(
    att_gt(yname = "Y", data = tv_data, tname = "period", idname = "id",
           gname = "G", weightsname = "tw", bstrap = FALSE),
    "Time-varying weights"
  )
})

test_that("clustervars contract enforced in all faster_mode x bstrap combinations", {
  # ?att_gt promises an error for more than two cluster variables (one of which
  # must be idname). Both pre-processors must reject the input with identical
  # wording (also matching mboot()) regardless of bstrap -- the analytical
  # (bstrap = FALSE) slow path used to silently cluster on the first extra
  # variable only.
  d <- data_eh
  d$cl1 <- d$id %% 10  # time-invariant within unit
  d$cl2 <- d$id %% 7
  msg <- "At most one cluster variable \\(beyond 'idname'\\) is supported"
  for (fm in c(FALSE, TRUE)) {
    for (bs in c(FALSE, TRUE)) {
      # three cluster variables (idname + two extras)
      expect_error(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("id", "cl1", "cl2"),
               bstrap = bs, faster_mode = fm),
        msg
      )
      # two cluster variables, neither of which is idname
      expect_error(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("cl1", "cl2"),
               bstrap = bs, faster_mode = fm),
        msg
      )
    }
  }
})

test_that("one extra cluster variable still works in all faster_mode x bstrap combinations", {
  # clustered DGP: units within a cluster share a common per-period shock, so
  # the clustered SE genuinely differs from the unclustered one
  set.seed(20260609)
  n_cl <- 40L
  sz <- rep(c(2L, 4L), length.out = n_cl)
  N <- sum(sz)
  cl <- rep(seq_len(n_cl), times = sz)
  gC <- sample(c(2L, 3L, 0L), n_cl, replace = TRUE)[cl]
  eta <- matrix(rnorm(n_cl * 4L, 0, 1.5), n_cl, 4L)
  d <- data.frame(id = rep(seq_len(N), each = 4L), period = rep(1:4, N),
                  cl = rep(cl, each = 4L), G = rep(gC, each = 4L))
  d$Y <- eta[cbind(d$cl, d$period)] + 0.3 * d$period +
    (d$G != 0 & d$period >= d$G) + rnorm(nrow(d))
  for (fm in c(FALSE, TRUE)) {
    se_cl <- NULL
    for (bs in c(FALSE, TRUE)) {
      res <- suppressWarnings(
        att_gt(yname = "Y", data = d, tname = "period", idname = "id",
               gname = "G", clustervars = c("id", "cl"), bstrap = bs,
               biters = 100, cband = FALSE, faster_mode = fm)
      )
      expect_s3_class(res, "MP")
      expect_true(any(is.finite(res$se)))
      if (!bs) se_cl <- res$se
    }
    # the analytical (bstrap = FALSE) path actually clusters: the SE differs
    # from the unclustered analytical SE
    res_iid <- att_gt(yname = "Y", data = d, tname = "period", idname = "id",
                      gname = "G", bstrap = FALSE, faster_mode = fm)
    ok <- is.finite(se_cl) & is.finite(res_iid$se)
    expect_true(any(ok))
    expect_gt(max(abs(se_cl[ok] - res_iid$se[ok])), 1e-8)
  }
})

# =============================================================================
# aggte() validation errors
# =============================================================================

test_that("aggte errors on invalid type", {
  mp_tmp <- att_gt(yname = "Y", data = data_eh, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE)
  expect_error(
    aggte(mp_tmp, type = "invalid"),
    "must be one of"
  )
  expect_error(
    aggte(mp_tmp, type = c("simple", "group")),
    "must be one of"
  )
  expect_error(
    aggte(mp_tmp, type = NA_character_),
    "must be one of"
  )
})

test_that("aggte errors when ATTs contain NA and na.rm=FALSE", {
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  mp_na <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  ))
  # Inject NA directly — no need to rely on estimation failure
  mp_na$att[1] <- NA
  expect_error(
    aggte(mp_na, type = "dynamic", na.rm = FALSE),
    "Missing values"
  )
})

# =============================================================================
# Wald pre-test warnings
# =============================================================================

test_that("att_gt handles singular covariance for Wald test gracefully", {
  # When covariance matrix is singular, W and Wpval should be NULL
  # and the result should still be a valid MP object
  small_sp <- did::reset.sim()
  small_sp$n <- 50
  small_data <- did::build_sim_dataset(small_sp)
  result <- suppressWarnings(
    att_gt(yname = "Y", data = small_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
  # W should be either a valid number or NULL (when singular)
  expect_true(is.null(result$W) || is.numeric(result$W))
})

# =============================================================================
# compute.att_gt warnings
# =============================================================================

test_that("att_gt warns on overlap violations", {
  # Create data where propensity score is near 1 (near-perfect separation)
  sep_data <- data_eh
  # Make covariate perfectly predict treatment for some cells
  sep_data$X_sep <- ifelse(sep_data$G > 0, 100, -100)
  expect_warning(
    att_gt(yname = "Y", xformla = ~X_sep, data = sep_data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE),
    "overlap condition"
  )
})

test_that("slow path warns once per failed cell, matching fast-path wording, with accurate Wald diagnosis", {
  # Each failed (g,t) cell must produce exactly ONE warning with identical text in
  # both modes: the tryCatch wrapper must not re-warn the internal overlap/singular
  # sentinels after the diagnostic warning already fired. And when pre-treatment
  # cells existed but were all dropped for NA/zero variance, the Wald warning must
  # say so instead of claiming no pre-treatment cells exist.
  set.seed(20260610)
  d <- expand.grid(id = 1:200, period = 1:3)
  gvals <- c(0, 2, 3)
  d$G <- gvals[(d$id - 1) %% 3 + 1]
  d$Xsep <- 1 * (d$G > 0)  # separates treated from never-treated: overlap fails in all 4 cells
  d$Y <- 0.1 * d$period + (d$G > 0 & d$period >= d$G) + rnorm(nrow(d), 0, 0.5)

  w_slow <- capture_warnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~Xsep, data = d, tname = "period", idname = "id",
           gname = "G", est_method = "dr", faster_mode = FALSE, bstrap = FALSE)))
  w_fast <- capture_warnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~Xsep, data = d, tname = "period", idname = "id",
           gname = "G", est_method = "dr", faster_mode = TRUE, bstrap = FALSE)))

  # identical warning sets across modes (one per failed cell + the Wald warning)
  expect_identical(w_slow, w_fast)
  # exactly one warning per failed cell, with the fast-path wording
  expect_identical(sum(grepl("overlap condition violated for group", w_slow)), 4L)
  # no leaked internal sentinel via the tryCatch wrapper
  expect_false(any(grepl("Error computing internal 2x2 DiD", w_slow)))
  # accurate Wald diagnosis: pre-treatment cells existed but had NA/zero variance
  expect_true(any(grepl("missing or zero variance", w_slow)))
  expect_false(any(grepl("treated early in the panel", w_slow)))
})

test_that("att_gt warns when no pre-treatment periods for Wald test", {
  # Create data where a group is first treated in period 2 (only 1 pre-treatment period)
  # With base_period="varying", there may be no pre-treatment ATTs for the Wald test
  early_data <- data.frame(
    id = rep(1:100, each = 3),
    period = rep(1:3, 100),
    G = rep(c(rep(2, 50), rep(0, 50)), each = 3),
    Y = rnorm(300)
  )
  expect_warning(
    att_gt(yname = "Y", data = early_data, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "pre-treatment|Wald"
  )
})

test_that("empty-cell warning names the actually-empty base period under base_period='universal' in both modes", {
  # Group 3's universal base period is 2; group 3 has observations in every
  # OTHER period. The slow path used to name the cell's current period
  # (1/3/4) -- periods where the group does have units -- instead of the empty
  # base period (2), which the fast path reported correctly.
  set.seed(20260610)
  d <- expand.grid(rep = 1:30, period = 1:4, G = c(0, 3))
  d$Y <- rnorm(nrow(d)) + 0.1 * d$period + (d$G > 0 & d$period >= d$G)
  d <- d[!(d$G == 3 & d$period == 2), ]  # group 3 empty only in its base period
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", gname = "G",
             panel = FALSE, base_period = "universal", bstrap = FALSE,
             faster_mode = fm)
    ))
    # one warning per non-base cell (t = 1, 3, 4), each naming period 2
    expect_identical(
      sum(grepl("No units in group 3 in time period 2", w, fixed = TRUE)), 3L,
      info = paste("faster_mode =", fm)
    )
    expect_false(any(grepl("No units in group 3 in time period [134]", w)),
                 info = paste("faster_mode =", fm))
  }
})

# =============================================================================
# pre-processing warning parity across modes
# =============================================================================

test_that("'no never-treated group' warning is identical across modes and discloses the period filtering", {
  # Both modes drop all periods >= (latest cohort - anticipation); the fast-mode
  # warning used to omit that a chunk of the data was discarded.
  set.seed(20260609)
  d <- data.frame(
    id = rep(1:60, each = 4),
    period = rep(1:4, 60),
    G = rep(c(rep(2, 30), rep(4, 30)), each = 4)
  )
  d$Y <- rnorm(nrow(d))
  msgs <- list()
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", idname = "id",
             gname = "G", control_group = "nevertreated", bstrap = FALSE,
             faster_mode = fm)
    ))
    m <- grep("No never-treated group", w, value = TRUE)
    expect_length(m, 1L)
    expect_match(m, "filtered out", fixed = TRUE)
    msgs[[as.character(fm)]] <- m
  }
  expect_identical(msgs[["FALSE"]], msgs[["TRUE"]])
})

test_that("balanced-panel coercion warning counts dropped units identically in both modes", {
  # 5 units each miss one period's row: 5 UNITS (20 rows) are dropped while
  # coercing to a balanced panel. The slow path used to report the unit count
  # as 'observations' (a 4x undercount of the rows actually removed).
  drop_ids <- unique(data_eh$id)[1:5]
  d <- data_eh[!(data_eh$id %in% drop_ids & data_eh$period == max(data_eh$period)), ]
  expected <- "5 units are missing in some periods. Converting to balanced panel by dropping them."
  for (fm in c(FALSE, TRUE)) {
    w <- capture_warnings(suppressMessages(
      att_gt(yname = "Y", data = d, tname = "period", idname = "id",
             gname = "G", panel = TRUE, allow_unbalanced_panel = FALSE,
             bstrap = FALSE, faster_mode = fm)
    ))
    expect_true(any(w == expected), info = paste("faster_mode =", fm))
    expect_false(any(grepl("observations while converting", w)),
                 info = paste("faster_mode =", fm))
  }
})

# =============================================================================
# Systematic faster_mode=TRUE vs FALSE consistency tests
# =============================================================================

# Shared setup
set.seed(20260401)
sp <- did::reset.sim()
data_fm <- did::build_sim_dataset(sp)

# Unbalanced version
data_ub <- data_fm[-c(1, 5, 10), ]

# Helper to compare two att_gt results.
# NOTE: the fast and slow paths order units differently (the exported influence
# function is a ROW PERMUTATION between modes), so a naive expect_equal of the two
# inffunc matrices would falsely fail. We therefore compare permutation-INVARIANT
# summaries -- column sums, the Gram matrix t(IF) %*% IF (which is exactly what the
# analytic variance V = crossprod(IF)/n uses), and the resulting standard errors --
# which DO have to match and which catch any row-permutation / indexing bug.
compare_modes <- function(res_slow, res_fast, label) {
  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste(label, "ATT"))
  expect_equal(res_slow$group, res_fast$group, label = paste(label, "group"))
  expect_equal(res_slow$t, res_fast$t, label = paste(label, "t"))
  if (!is.null(res_slow$inffunc) && !is.null(res_fast$inffunc)) {
    ifs <- as.matrix(res_slow$inffunc); iff <- as.matrix(res_fast$inffunc)
    expect_equal(dim(ifs), dim(iff), label = paste(label, "IF dim"))
    expect_equal(colSums(ifs), colSums(iff), tolerance = 1e-8, label = paste(label, "IF colSums"))
    expect_equal(crossprod(ifs), crossprod(iff), tolerance = 1e-8, label = paste(label, "IF Gram"))
  }
  expect_equal(res_slow$se, res_fast$se, tolerance = 1e-9, label = paste(label, "SE"))
}

# =============================================================================
# Core grid: est_method x panel_type x control_group x base_period
# =============================================================================

# --- Balanced panel ---

for (em in c("dr", "ipw", "reg")) {
  for (cg in c("nevertreated", "notyettreated")) {
    for (bp in c("varying", "universal")) {
      label <- paste(em, "balanced", cg, bp)
      test_that(paste("consistency:", label), {
        res_slow <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 faster_mode = FALSE, bstrap = FALSE)
        ))
        res_fast <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 faster_mode = TRUE, bstrap = FALSE)
        ))
        compare_modes(res_slow, res_fast, label)
      })
    }
  }
}

# --- Unbalanced panel ---

for (em in c("dr", "ipw", "reg")) {
  for (cg in c("nevertreated", "notyettreated")) {
    for (bp in c("varying", "universal")) {
      label <- paste(em, "unbalanced", cg, bp)
      test_that(paste("consistency:", label), {
        res_slow <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_ub, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 allow_unbalanced_panel = TRUE,
                 faster_mode = FALSE, bstrap = FALSE)
        ))
        res_fast <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_ub, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 allow_unbalanced_panel = TRUE,
                 faster_mode = TRUE, bstrap = FALSE)
        ))
        compare_modes(res_slow, res_fast, label)
      })
    }
  }
}

# --- Repeated cross sections ---

for (em in c("dr", "ipw", "reg")) {
  for (cg in c("nevertreated", "notyettreated")) {
    for (bp in c("varying", "universal")) {
      label <- paste(em, "RC", cg, bp)
      test_that(paste("consistency:", label), {
        res_slow <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 panel = FALSE,
                 faster_mode = FALSE, bstrap = FALSE)
        ))
        res_fast <- suppressWarnings(suppressMessages(
          att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
                 idname = "id", gname = "G", est_method = em,
                 control_group = cg, base_period = bp,
                 panel = FALSE,
                 faster_mode = TRUE, bstrap = FALSE)
        ))
        compare_modes(res_slow, res_fast, label)
      })
    }
  }
}

# =============================================================================
# Additional consistency tests beyond the core grid
# =============================================================================

test_that("consistency with anticipation > 0", {
  res_slow <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
           idname = "id", gname = "G", est_method = "dr",
           anticipation = 1, faster_mode = FALSE, bstrap = FALSE)
  ))
  res_fast <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
           idname = "id", gname = "G", est_method = "dr",
           anticipation = 1, faster_mode = TRUE, bstrap = FALSE)
  ))
  compare_modes(res_slow, res_fast, "anticipation=1")
})

test_that("consistency without covariates", {
  res_slow <- att_gt(yname = "Y", data = data_fm, tname = "period",
                     idname = "id", gname = "G", est_method = "reg",
                     faster_mode = FALSE, bstrap = FALSE)
  res_fast <- att_gt(yname = "Y", data = data_fm, tname = "period",
                     idname = "id", gname = "G", est_method = "reg",
                     faster_mode = TRUE, bstrap = FALSE)
  compare_modes(res_slow, res_fast, "no covariates")
})

# =============================================================================
# aggte consistency across modes
# =============================================================================

test_that("aggte consistency across faster_mode for all types", {
  mp_slow <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
           idname = "id", gname = "G", est_method = "dr",
           faster_mode = FALSE, bstrap = FALSE)
  ))
  mp_fast <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data_fm, tname = "period",
           idname = "id", gname = "G", est_method = "dr",
           faster_mode = TRUE, bstrap = FALSE)
  ))

  for (tp in c("simple", "dynamic", "group", "calendar")) {
    agg_slow <- suppressWarnings(aggte(mp_slow, type = tp))
    agg_fast <- suppressWarnings(aggte(mp_fast, type = tp))
    expect_equal(agg_slow$overall.att, agg_fast$overall.att,
                 tolerance = 1e-8, label = paste("aggte", tp))
  }
})

# =============================================================================
# NA-propagation semantics in the unbalanced-panel influence aggregation
# =============================================================================

# With allow_unbalanced_panel = TRUE, NA/NaN entries in a 2x2 influence function
# (only reachable via a custom est_method returning a finite ATT alongside a
# partially-NA influence function) must propagate to the unit's aggregated
# influence value -- so that cell's SE is NA -- in BOTH modes. The fast path
# used to zero them silently (sum(na.rm = TRUE)), understating the SE and
# diverging from the slow path; this pins the propagating semantics.
test_that("partial-NA influence functions propagate identically across modes (unbalanced panel)", {
  set.seed(20260609)
  sp_na <- did::reset.sim(time.periods = 3, n = 400)
  data_na <- did::build_sim_dataset(sp_na, panel = TRUE)
  # Drop period 3 for 10 units so the panel is genuinely unbalanced (otherwise
  # pre-processing downgrades allow_unbalanced_panel to a balanced panel).
  drop_ids <- head(unique(data_na$id), 10)
  data_na <- data_na[!(data_na$id %in% drop_ids & data_na$period == 3), ]

  # Finite ATT, one NA injected into the influence function of every 2x2 cell
  na_inj <- function(y, post, D, covariates, i.weights, inffunc, ...) {
    res <- DRDID::reg_did_rc(y = y, post = post, D = D, covariates = covariates,
                             i.weights = i.weights, boot = FALSE, inffunc = TRUE)
    inf <- res$att.inf.func
    inf[1] <- NA_real_
    list(ATT = res$ATT, att.inf.func = inf)
  }

  run_na <- function(fm) suppressWarnings(suppressMessages(
    att_gt(yname = "Y", tname = "period", idname = "id", gname = "G",
           xformla = ~X, data = data_na, panel = TRUE,
           allow_unbalanced_panel = TRUE, est_method = na_inj,
           faster_mode = fm, bstrap = FALSE, cband = FALSE)
  ))

  res_fast <- run_na(TRUE)
  res_slow <- run_na(FALSE)

  # ATT point estimates stay finite and identical across modes
  expect_true(all(is.finite(res_fast$att)))
  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10)
  # The injected NA propagates: every (g,t) standard error is NA in both modes
  expect_true(all(is.na(res_fast$se)))
  expect_true(all(is.na(res_slow$se)))
  # Exactly one aggregated influence entry per (g,t) cell is NA, in both modes
  na_fast <- sum(is.na(as.matrix(res_fast$inffunc)))
  na_slow <- sum(is.na(as.matrix(res_slow$inffunc)))
  expect_identical(na_fast, na_slow)
  expect_identical(na_fast, length(res_fast$att))
})

# =============================================================================
# aggte() edge-case coverage: a single treated group exercises the single-column
# matrix-collapse paths in compute.aggte (where group/event-time matrices reduce to
# one column). All four aggregation types must still return a valid AGGTEobj.
# =============================================================================

test_that("aggte handles a single treated group for every aggregation type", {
  set.seed(5)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  g1 <- sort(unique(d$G[d$G > 0]))[1]
  d1 <- d[d$G == 0 | d$G == g1, ]   # one treated group + never-treated controls
  mp <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d1,
    tname = "period", idname = "id", gname = "G", bstrap = FALSE)))
  for (ty in c("simple", "group", "dynamic", "calendar")) {
    agg <- suppressWarnings(aggte(mp, type = ty))
    expect_s3_class(agg, "AGGTEobj")
    expect_true(is.finite(agg$overall.att))
    expect_true(is.finite(agg$overall.se))
  }
})

test_that("aggte dynamic with min_e / max_e windows returns a consistent event-time set", {
  set.seed(6)
  sp <- did::reset.sim(time.periods = 5)
  d <- did::build_sim_dataset(sp)
  mp <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", bstrap = FALSE)))
  agg <- suppressWarnings(aggte(mp, type = "dynamic", min_e = -1, max_e = 1))
  expect_true(all(agg$egt >= -1 & agg$egt <= 1))
  expect_equal(length(agg$att.egt), length(agg$egt))
})

# =============================================================================
# Tests for edge cases and boundary conditions
# =============================================================================

test_that("single treated group produces valid att_gt", {
  set.seed(20260401)
  data_sg <- data.frame(
    id = rep(1:200, each = 5),
    period = rep(1:5, 200),
    G = rep(c(rep(4, 50), rep(0, 150)), each = 5),
    Y = rnorm(1000)
  )
  data_sg$Y[data_sg$G == 4 & data_sg$period >= 4] <-
    data_sg$Y[data_sg$G == 4 & data_sg$period >= 4] + 1

  result <- att_gt(yname = "Y", data = data_sg, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE)
  expect_s3_class(result, "MP")
  expect_true(any(!is.na(result$att)))
})

test_that("single treated group works with all 4 aggte types", {
  set.seed(20260401)
  data_sg <- data.frame(
    id = rep(1:200, each = 5),
    period = rep(1:5, 200),
    G = rep(c(rep(4, 50), rep(0, 150)), each = 5),
    Y = rnorm(1000)
  )
  data_sg$Y[data_sg$G == 4 & data_sg$period >= 4] <-
    data_sg$Y[data_sg$G == 4 & data_sg$period >= 4] + 1

  mp <- att_gt(yname = "Y", data = data_sg, tname = "period", idname = "id",
               gname = "G", bstrap = FALSE)

  for (tp in c("simple", "dynamic", "group", "calendar")) {
    agg <- suppressWarnings(aggte(mp, type = tp))
    expect_s3_class(agg, "AGGTEobj")
    expect_false(is.na(agg$overall.att), label = tp)
  }
})

test_that("two-period data with universal base period", {
  set.seed(20260401)
  data_2p <- data.frame(
    id = rep(1:200, each = 2),
    period = rep(1:2, 200),
    G = rep(c(rep(2, 50), rep(0, 150)), each = 2),
    Y = rnorm(400)
  )
  data_2p$Y[data_2p$G == 2 & data_2p$period == 2] <-
    data_2p$Y[data_2p$G == 2 & data_2p$period == 2] + 1

  result <- suppressWarnings(
    att_gt(yname = "Y", data = data_2p, tname = "period", idname = "id",
           gname = "G", base_period = "universal", bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
  expect_true(any(!is.na(result$att)))
})

test_that("data with no never-treated group works with notyettreated", {
  set.seed(20260401)
  # All units are eventually treated (groups 3 and 5)
  data_nnt <- data.frame(
    id = rep(1:200, each = 6),
    period = rep(1:6, 200),
    G = rep(c(rep(3, 100), rep(5, 100)), each = 6),
    Y = rnorm(1200)
  )
  result <- suppressWarnings(
    att_gt(yname = "Y", data = data_nnt, tname = "period", idname = "id",
           gname = "G", control_group = "notyettreated", bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
})

test_that("data with no never-treated group warns with nevertreated", {
  set.seed(20260401)
  data_nnt <- data.frame(
    id = rep(1:200, each = 6),
    period = rep(1:6, 200),
    G = rep(c(rep(3, 100), rep(5, 100)), each = 6),
    Y = rnorm(1200)
  )
  expect_warning(
    att_gt(yname = "Y", data = data_nnt, tname = "period", idname = "id",
           gname = "G", control_group = "nevertreated", bstrap = FALSE),
    "No never-treated group|never-treated"
  )
})

test_that("groups treated in first period are dropped", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data_fp <- did::build_sim_dataset(sp)
  # Add units treated in the very first period
  first_per <- min(data_fp$period)
  extra <- data_fp[data_fp$G == sort(unique(data_fp$G[data_fp$G > 0]))[1], ]
  extra$G <- first_per
  extra$id <- extra$id + max(data_fp$id)
  data_fp <- rbind(data_fp, extra)

  expect_warning(
    att_gt(yname = "Y", data = data_fp, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE),
    "already treated|Dropped"
  )
})

test_that("non-consecutive time periods work", {
  set.seed(20260401)
  data_nc <- data.frame(
    id = rep(1:200, each = 4),
    period = rep(c(2000, 2003, 2007, 2010), 200),
    G = rep(c(rep(2007, 50), rep(0, 150)), each = 4),
    Y = rnorm(800)
  )
  data_nc$Y[data_nc$G == 2007 & data_nc$period >= 2007] <-
    data_nc$Y[data_nc$G == 2007 & data_nc$period >= 2007] + 1

  result <- suppressWarnings(
    att_gt(yname = "Y", data = data_nc, tname = "period", idname = "id",
           gname = "G", bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
  expect_true(any(!is.na(result$att)))

  agg <- suppressWarnings(aggte(result, type = "dynamic"))
  expect_s3_class(agg, "AGGTEobj")
})

test_that("non-consecutive group values work", {
  set.seed(20260401)
  data_ng <- data.frame(
    id = rep(1:200, each = 5),
    period = rep(1:5, 200),
    G = rep(c(rep(3, 50), rep(5, 50), rep(0, 100)), each = 5),
    Y = rnorm(1000)
  )
  result <- att_gt(yname = "Y", data = data_ng, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE)
  expect_s3_class(result, "MP")
  expect_true(length(unique(result$group)) >= 2)
})

test_that("allow_unbalanced_panel=TRUE with balanced data proceeds normally", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data_bal <- did::build_sim_dataset(sp)

  result <- suppressMessages(
    att_gt(yname = "Y", data = data_bal, tname = "period", idname = "id",
           gname = "G", allow_unbalanced_panel = TRUE, bstrap = FALSE)
  )
  expect_s3_class(result, "MP")
  expect_true(any(!is.na(result$att)))
})

test_that("allow_unbalanced_panel=TRUE with truly unbalanced data", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data_ub <- did::build_sim_dataset(sp)
  # Drop a few rows to make it unbalanced
  data_ub <- data_ub[-c(1, 5, 10), ]

  result <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", data = data_ub, tname = "period", idname = "id",
           gname = "G", allow_unbalanced_panel = TRUE, bstrap = FALSE)
  ))
  expect_s3_class(result, "MP")
  expect_true(any(!is.na(result$att)))
})

test_that("single post-treatment period produces valid results", {
  set.seed(20260401)
  data_spt <- data.frame(
    id = rep(1:200, each = 3),
    period = rep(1:3, 200),
    G = rep(c(rep(3, 50), rep(0, 150)), each = 3),
    Y = rnorm(600)
  )
  data_spt$Y[data_spt$G == 3 & data_spt$period == 3] <-
    data_spt$Y[data_spt$G == 3 & data_spt$period == 3] + 1

  result <- att_gt(yname = "Y", data = data_spt, tname = "period", idname = "id",
                   gname = "G", bstrap = FALSE)
  expect_s3_class(result, "MP")
  # Should have at least one post-treatment ATT
  post_atts <- result$att[result$group <= result$t]
  expect_true(any(!is.na(post_atts)))
})

# =============================================================================
# Tests for compute_inffunc = FALSE (point-estimates-only mode): skips influence
# functions, standard errors, bands, the pre-test, and the bootstrap. Point
# estimates must be identical to a full run.
# =============================================================================

test_that("compute_inffunc = FALSE gives point estimates identical to a full run", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  for (fm in c(TRUE, FALSE)) for (est in c("dr", "reg", "ipw")) for (pn in c(TRUE, FALSE)) {
    full <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = pn, bstrap = FALSE, faster_mode = fm)))
    pe <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = pn, faster_mode = fm,
             compute_inffunc = FALSE)))
    lab <- paste(est, "panel", pn, "fm", fm)
    expect_equal(full$att, pe$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_true(all(is.na(pe$se)), label = paste(lab, "se all NA"))
    expect_null(pe$inffunc, label = paste(lab, "inffunc NULL"))
    expect_null(pe$W, label = paste(lab, "Wald NULL"))
  }
})

test_that("compute_inffunc = FALSE forces bstrap/cband off and yields a valid MP object", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  pe <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
           gname = "G", bstrap = TRUE, cband = TRUE, compute_inffunc = FALSE)))
  expect_s3_class(pe, "MP")
  expect_false(pe$DIDparams$bstrap)
  expect_false(pe$DIDparams$cband)
  # print / summary / tidy / glance must work with NA SEs and no influence functions
  expect_output(print(pe))
  expect_output(summary(pe))
  td <- tidy(pe)
  expect_equal(nrow(td), length(pe$att))
  expect_true(all(is.na(td$std.error)))
  expect_s3_class(glance(pe), "data.frame")
})

test_that("aggte() errors clearly on a compute_inffunc = FALSE object", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  pe <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
           gname = "G", compute_inffunc = FALSE)))
  expect_error(suppressWarnings(suppressMessages(aggte(pe, type = "group"))),
               "compute_inffunc = FALSE")
})

test_that("compute_inffunc must be a single logical", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  expect_error(
    att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
           gname = "G", compute_inffunc = "yes"),
    "compute_inffunc must be a single logical")
})

# -----------------------------------------------------------------------------
# Non-default option matrix: the do_inf skips on the varying-weights, unbalanced,
# universal-base, clustered, and custom-est_method branches must leave point
# estimates identical to a full run in both modes.
# -----------------------------------------------------------------------------

test_that("compute_inffunc = FALSE matches a full run with fix_weights = 'varying'", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$tvw <- data$period + runif(nrow(data), -0.1, 0.1)
  for (fm in c(TRUE, FALSE)) {
    full <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", weightsname = "tvw", fix_weights = "varying",
             bstrap = FALSE, faster_mode = fm)))
    pe <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", weightsname = "tvw", fix_weights = "varying",
             faster_mode = fm, compute_inffunc = FALSE)))
    lab <- paste("fix_weights varying fm", fm)
    expect_equal(full$att, pe$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_true(identical(is.na(full$att), is.na(pe$att)), label = paste(lab, "NA pattern"))
    expect_null(pe$inffunc, label = paste(lab, "inffunc NULL"))
  }
})

test_that("compute_inffunc = FALSE matches a full run with anticipation, universal base period, not-yet-treated controls, and clustering", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  for (fm in c(TRUE, FALSE)) {
    full <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", anticipation = 1, base_period = "universal",
             control_group = "notyettreated", clustervars = "cluster",
             bstrap = FALSE, faster_mode = fm)))
    pe <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", anticipation = 1, base_period = "universal",
             control_group = "notyettreated", clustervars = "cluster",
             faster_mode = fm, compute_inffunc = FALSE)))
    lab <- paste("anticipation/universal/nyt/cluster fm", fm)
    expect_equal(full$att, pe$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_true(identical(is.na(full$att), is.na(pe$att)), label = paste(lab, "NA pattern"))
    expect_null(pe$inffunc, label = paste(lab, "inffunc NULL"))
  }
})

test_that("compute_inffunc = FALSE matches a full run on an unbalanced panel", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data_ub <- data[-sample(nrow(data), 150), ]
  for (fm in c(TRUE, FALSE)) {
    full <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data_ub, tname = "period", idname = "id",
             gname = "G", allow_unbalanced_panel = TRUE, bstrap = FALSE,
             faster_mode = fm)))
    pe <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data_ub, tname = "period", idname = "id",
             gname = "G", allow_unbalanced_panel = TRUE, faster_mode = fm,
             compute_inffunc = FALSE)))
    lab <- paste("unbalanced panel fm", fm)
    expect_equal(full$att, pe$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_true(identical(is.na(full$att), is.na(pe$att)), label = paste(lab, "NA pattern"))
    expect_null(pe$inffunc, label = paste(lab, "inffunc NULL"))
  }
})

test_that("custom est_method receives inffunc = FALSE and may return a NULL influence function", {
  set.seed(20260609)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  seen <- new.env(parent = emptyenv())
  my_panel_est <- function(y1, y0, D, covariates, i.weights, inffunc, ...) {
    seen$inffunc <- c(seen$inffunc, isTRUE(inffunc))
    delta <- y1 - y0
    att <- weighted.mean(delta[D == 1], i.weights[D == 1])
    inf <- if (isTRUE(inffunc)) (D / mean(D)) * (delta - att) else NULL
    list(ATT = att, att.inf.func = inf)
  }
  for (fm in c(TRUE, FALSE)) {
    seen$inffunc <- logical(0)
    full <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", est_method = my_panel_est, bstrap = FALSE,
             faster_mode = fm)))
    expect_true(length(seen$inffunc) > 0 && all(seen$inffunc),
                label = paste("full run passes inffunc = TRUE, fm", fm))
    seen$inffunc <- logical(0)
    pe <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X, data = data, tname = "period", idname = "id",
             gname = "G", est_method = my_panel_est, faster_mode = fm,
             compute_inffunc = FALSE)))
    expect_true(length(seen$inffunc) > 0 && !any(seen$inffunc),
                label = paste("point-estimate run passes inffunc = FALSE, fm", fm))
    lab <- paste("custom est_method fm", fm)
    expect_s3_class(pe, "MP")
    expect_equal(full$att, pe$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_null(pe$inffunc, label = paste(lab, "inffunc NULL"))
  }
})

# =============================================================================
# Tests for ggdid.MP and ggdid.AGGTEobj plotting functions
# =============================================================================

# Shared setup
set.seed(20260401)
sp <- did::reset.sim()
data_gg <- did::build_sim_dataset(sp)

mp_gg <- suppressWarnings(suppressMessages(
  att_gt(yname = "Y", xformla = ~X, data = data_gg, tname = "period",
         idname = "id", gname = "G", est_method = "dr",
         bstrap = FALSE)
))

agg_dyn <- suppressWarnings(aggte(mp_gg, type = "dynamic"))
agg_grp <- suppressWarnings(aggte(mp_gg, type = "group"))
agg_cal <- suppressWarnings(aggte(mp_gg, type = "calendar"))
agg_sim <- suppressWarnings(aggte(mp_gg, type = "simple"))

# =============================================================================
# ggdid.MP tests
# =============================================================================

test_that("ggdid.MP returns a ggplot object", {
  p <- ggdid(mp_gg)
  expect_s3_class(p, "gg")
})

test_that("ggdid.MP accepts group parameter", {
  groups <- unique(mp_gg$group)
  p <- ggdid(mp_gg, group = groups[1])
  expect_s3_class(p, "gg")
})

test_that("ggdid.MP warns for non-existent group values", {
  expect_warning(
    ggdid(mp_gg, group = c(9999)),
    "do not exist"
  )
})

test_that("ggdid.MP accepts custom labels", {
  p <- ggdid(mp_gg, xlab = "Time", ylab = "Effect", title = "Test Plot")
  expect_s3_class(p, "gg")
})

# =============================================================================
# ggdid.AGGTEobj tests
# =============================================================================

test_that("ggdid works for type = dynamic", {
  p <- ggdid(agg_dyn)
  expect_s3_class(p, "gg")
})

test_that("ggdid works for type = group", {
  p <- suppressWarnings(ggdid(agg_grp))
  expect_s3_class(p, "gg")
})

test_that("ggdid works for type = calendar", {
  p <- ggdid(agg_cal)
  expect_s3_class(p, "gg")
})

test_that("ggdid errors for type = simple", {
  expect_error(
    ggdid(agg_sim),
    "not available"
  )
})

test_that("ggdid.AGGTEobj accepts custom labels and theme settings", {
  p <- ggdid(agg_dyn, xlab = "Event Time", ylab = "ATT",
             title = "Event Study", theming = TRUE, legend = TRUE)
  expect_s3_class(p, "gg")
})

test_that("ggdid.AGGTEobj works with theming=FALSE", {
  p <- ggdid(agg_dyn, theming = FALSE)
  expect_s3_class(p, "gg")
})

test_that("ggdid.AGGTEobj works with ref_line=NULL", {
  p <- ggdid(agg_dyn, ref_line = NULL)
  expect_s3_class(p, "gg")
})

test_that("splot works for group-type and renders without deprecation warnings", {
  # splot() is the path that uses the ggplot2 version-gated errorbar layer.
  # It is called via ggdid.AGGTEobj() for type="group". Verify the output is
  # a valid ggplot object and no deprecation warnings leak through.
  expect_warning(p <- ggdid(agg_grp), regexp = NA)
  expect_s3_class(p, "gg")
  # Verify the plot actually builds (catches any layer construction errors)
  built <- ggplot2::ggplot_build(p)
  expect_s3_class(built, "ggplot_built")
})

# =============================================================================
# Tests for covariate handling in compute.att_gt.R (the slow path). The design
# matrix is built once over the full data and row-subset per (g,t) cell, giving:
#   * numeric formulae      -> bit-identical to a per-cell model.matrix()
#   * factor covariates     -> identical to manually-expanded dummy columns
#   * transform formulae    -> work (previously errored); slow == fast
# =============================================================================

# helper: add treatment-contrast dummy columns for a factor, named f_<level>
add_dummies <- function(d, fac = "fac") {
  lv <- levels(d[[fac]])
  for (l in lv[-1L]) d[[paste0("f_", l)]] <- as.numeric(d[[fac]] == l)
  d
}
dummy_formula <- function(d) {
  stats::as.formula(paste("~", paste(grep("^f_", names(d), value = TRUE), collapse = " + ")))
}

test_that("global design slice is bit-identical to per-subset model.matrix for numeric formulae", {
  set.seed(20260608)
  n <- 80
  df <- data.frame(X1 = rnorm(n), X2 = rnorm(n), Xp = runif(n, 1, 5))
  rows <- sample(n, 40)
  for (f in list(~X1, ~X1 + X2, ~X1 * X2, ~X1:X2, ~I(X1^2) + X2, ~log(Xp))) {
    a <- model.matrix(f, df[rows, ])
    b <- model.matrix(f, df)[rows, , drop = FALSE]
    expect_identical(as.vector(a), as.vector(b), label = deparse(f))
    expect_identical(colnames(a), colnames(b), label = deparse(f))
  }
})

test_that("slow path matches faster_mode for numeric / transform formulae", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  set.seed(13)
  data$X2 <- rnorm(nrow(data))
  data$Xpos <- runif(nrow(data), 1, 5)

  # bare numeric, interactions, and transform formulae (the latter previously errored)
  for (f in list(~X + X2, ~X * X2, ~X:X2, ~I(X^2) + X2, ~log(Xpos), ~poly(X, 2))) {
    res_slow <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period",
             idname = "id", gname = "G", est_method = "dr",
             bstrap = FALSE, faster_mode = FALSE)))
    res_fast <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period",
             idname = "id", gname = "G", est_method = "dr",
             bstrap = FALSE, faster_mode = TRUE)))
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste(deparse(f), "ATT"))
    expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10, label = paste(deparse(f), "se"))
  }
})

test_that("RC-path (positional global slice) matches faster_mode", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  set.seed(13)
  data$X2 <- rnorm(nrow(data))
  for (f in list(~X, ~X + X2)) {
    res_slow <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE, faster_mode = FALSE)))
    res_fast <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE, faster_mode = TRUE)))
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste("RC", deparse(f), "ATT"))
    expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10, label = paste("RC", deparse(f), "se"))
  }
})

test_that("unbalanced panel (positional global slice) matches faster_mode for factor / transform formulae", {
  # the allow_unbalanced_panel branch slices global_mm by POSITION via disdat_rows,
  # so cover it with formulae whose design depends on which rows are present
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  set.seed(42)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.05)), ]
  for (f in list(~fac, ~I(X^2), ~poly(X, 2), ~X * fac)) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             allow_unbalanced_panel = TRUE, faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             allow_unbalanced_panel = TRUE, faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("UB", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("UB", deparse(f), "se"))
  }
})

test_that("RC-path matches faster_mode for factor / data-dependent-basis formulae", {
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  set.seed(42)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.05)), ]
  for (f in list(~fac, ~poly(X, 2))) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE,
             faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE,
             faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("RC", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("RC", deparse(f), "se"))
  }
})

test_that("transform / factor formulae match faster_mode under universal base, notyettreated, anticipation", {
  # crosses the hoisted design with the non-default options the precompute path
  # reimplements (universal-base Ypre/Ypost swap, notyettreated C_full, anticipation)
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  for (f in list(~I(X^2) + X, ~fac)) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             control_group = "notyettreated", base_period = "universal",
             anticipation = 1, faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             control_group = "notyettreated", base_period = "universal",
             anticipation = 1, faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("cross", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("cross", deparse(f), "se"))
  }
})

test_that("global basis is reparameterization-invariant: ~poly(X, 2) equals ~X + I(X^2)", {
  # locks the claim justifying the GLOBAL basis for data-dependent terms (poly/scale/ns):
  # the 2x2 estimators are invariant to a full-rank reparameterization of the covariates,
  # so the orthogonal global basis must reproduce the raw basis spanning the same space
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  for (est in c("dr", "reg", "ipw")) for (fm in c(FALSE, TRUE)) {
    rp <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~poly(X, 2), data = d, tname = "period",
             idname = "id", gname = "G", est_method = est, bstrap = FALSE,
             faster_mode = fm)))
    rr <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X + I(X^2), data = d, tname = "period",
             idname = "id", gname = "G", est_method = est, bstrap = FALSE,
             faster_mode = fm)))
    lab <- paste(est, "fm", fm)
    expect_equal(rp$att, rr$att, tolerance = 1e-10, label = paste(lab, "ATT"))
    expect_equal(rp$se,  rr$se,  tolerance = 1e-10, label = paste(lab, "se"))
  }
})

test_that("a factor covariate equals manually-expanded dummies EXACTLY (dense levels)", {
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$fac <- factor(sample(c("a", "b", "c"), nrow(data), TRUE))
  data <- add_dummies(data)
  dform <- dummy_formula(data)
  for (est in c("dr", "reg", "ipw")) for (panel in c(TRUE, FALSE)) for (fm in c(TRUE, FALSE)) {
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~fac, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = panel, bstrap = FALSE, faster_mode = fm)))
    rd <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = dform, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = panel, bstrap = FALSE, faster_mode = fm)))
    lab <- paste(est, "panel", panel, "fm", fm)
    expect_equal(rf$att, rd$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_equal(rf$se,  rd$se,  tolerance = 1e-12, label = paste(lab, "se"))
    expect_equal(as.matrix(rf$inffunc), as.matrix(rd$inffunc), tolerance = 1e-12,
                 label = paste(lab, "inffunc"))
  }
})

test_that("transform formulae that evaluate to NaN drop those rows instead of crashing", {
  # Regression: the non-finite-row check must keep NA/NaN rows visible (model.frame
  # with na.pass) so they are dropped; model.matrix(na.pass) silently drops NaN rows,
  # which used to leave them in the data and crash att_gt with 'subscript out of bounds'.
  set.seed(404)
  sp <- did::reset.sim(time.periods = 4)
  data <- did::build_sim_dataset(sp)
  ids <- unique(data$id)
  xv <- stats::setNames(runif(length(ids), -0.5, 3), ids)   # some <= 0  -> log() is NaN
  data$Xneg <- xv[as.character(data$id)]
  logXneg <- suppressWarnings(log(data$Xneg))
  stopifnot(any(!is.finite(logXneg)))                        # the pathological rows exist

  rs <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = FALSE)))
  rf <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = TRUE)))
  expect_true(any(is.finite(rs$att)))
  expect_equal(rs$att, rf$att, tolerance = 1e-10)
  # equals manually pre-cleaning the NaN units
  d2 <- data[is.finite(logXneg), ]
  rref <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = d2, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = FALSE)))
  expect_equal(rs$att, rref$att, tolerance = 1e-12)
})

test_that("a globally-empty factor level is dropped instead of NA-failing every cell", {
  # Regression: factor(levels = c('a','b','c')) where 'c' never occurs in the data
  # (common after users subset their data, since R keeps empty levels) used to emit
  # an all-zero dummy column in EVERY (g,t) cell, NA-failing the whole estimation
  # with a misleading 'Not enough control units' warning. Globally-empty levels are
  # now dropped once up front on both paths, so results equal the droplevels()-ed
  # data exactly. Levels absent only from particular cells keep their documented
  # NA-that-cell-only behavior (see the sparse-factor test below).
  set.seed(22)
  nunit <- 80
  G <- rep(c(0, 2), each = 40)
  funit <- factor(sample(c("a", "b"), nunit, TRUE), levels = c("a", "b", "c"))
  dat <- expand.grid(id = 1:nunit, period = 1:3)
  dat$G <- G[dat$id]
  dat$flev <- funit[dat$id]
  eta <- rnorm(nunit)
  dat$Y <- eta[dat$id] + dat$period + 1 * (dat$G == 2 & dat$period >= 2) +
    rnorm(nrow(dat), sd = 0.1)
  dat2 <- dat
  dat2$flev <- droplevels(dat2$flev)

  run <- function(d, fm, est) suppressWarnings(suppressMessages(
    att_gt(yname = "Y", tname = "period", idname = "id", gname = "G", data = d,
           xformla = ~flev, est_method = est, bstrap = FALSE, faster_mode = fm)))

  for (fm in c(FALSE, TRUE)) for (est in c("dr", "ipw", "reg")) {
    lab <- paste("fm", fm, est)
    res <- run(dat, fm, est)    # empty level 'c' declared
    ref <- run(dat2, fm, est)   # empty level pre-dropped by the user
    expect_false(anyNA(res$att), label = paste(lab, "no NA"))
    expect_equal(res$att, ref$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_equal(res$se,  ref$se,  tolerance = 1e-12, label = paste(lab, "se"))
    expect_equal(as.matrix(res$inffunc), as.matrix(ref$inffunc), tolerance = 1e-12,
                 label = paste(lab, "inffunc"))
  }

  # known values (master's slow path estimated these with the empty level present)
  res_slow <- run(dat, FALSE, "dr")
  res_fast <- run(dat, TRUE, "dr")
  expect_equal(res_slow$att, c(1.0275, 0.9826), tolerance = 1e-3)
  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10)
  expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10)
})

test_that("a sparse factor (level absent from some cells) matches manual dummies, incl. warnings", {
  # 'b' appears only in group 2, so the never-treated control never has it; every cell
  # is rank-deficient and must return NA -- exactly as manual dummies do, with the same
  # warnings, and WITHOUT crashing (the old per-cell droplevels() raised a contrast error).
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  fac <- rep("a", nrow(data)); fac[data$G == 2] <- sample(c("a", "b"), sum(data$G == 2), TRUE)
  data$fac <- factor(fac)
  data$f_b <- as.numeric(data$fac == "b")

  collect <- function(f, fm) {
    ws <- character(0)
    res <- withCallingHandlers(
      suppressMessages(att_gt(yname = "Y", xformla = f, data = data, tname = "period",
                              idname = "id", gname = "G", est_method = "reg",
                              panel = TRUE, bstrap = FALSE, faster_mode = fm)),
      warning = function(w) { ws[[length(ws) + 1]] <<- conditionMessage(w); invokeRestart("muffleWarning") })
    list(att = res$att, se = res$se, warns = sort(unique(ws)))
  }
  for (fm in c(TRUE, FALSE)) {
    rf <- collect(~fac, fm)
    rd <- collect(~f_b, fm)
    expect_equal(rf$att, rd$att, label = paste("fm", fm, "ATT"))     # both all-NA, same cells
    expect_equal(rf$se,  rd$se,  label = paste("fm", fm, "se"))
    expect_identical(rf$warns, rd$warns, label = paste("fm", fm, "warnings"))
  }
})

test_that("RC sparse factor (level absent from some cells) matches manual dummies, incl. warnings", {
  # same construction as the panel test above, on the repeated-cross-sections path
  # (positional global_mm slicing): 'b' appears only in group 2, so every cell is
  # rank-deficient and must NA-fail exactly as manual dummies do, in both modes
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  fac <- rep("a", nrow(data)); fac[data$G == 2] <- sample(c("a", "b"), sum(data$G == 2), TRUE)
  data$fac <- factor(fac)
  data$f_b <- as.numeric(data$fac == "b")

  collect <- function(f, fm) {
    ws <- character(0)
    res <- withCallingHandlers(
      suppressMessages(att_gt(yname = "Y", xformla = f, data = data, tname = "period",
                              idname = "id", gname = "G", est_method = "reg",
                              panel = FALSE, bstrap = FALSE, faster_mode = fm)),
      warning = function(w) { ws[[length(ws) + 1]] <<- conditionMessage(w); invokeRestart("muffleWarning") })
    list(att = res$att, se = res$se, warns = sort(unique(ws)))
  }
  for (fm in c(TRUE, FALSE)) {
    rf <- collect(~fac, fm)
    rd <- collect(~f_b, fm)
    expect_equal(rf$att, rd$att, label = paste("RC fm", fm, "ATT"))
    expect_equal(rf$se,  rd$se,  label = paste("RC fm", fm, "se"))
    expect_equal(is.na(rf$att), is.na(rd$att), label = paste("RC fm", fm, "NA pattern"))
    expect_identical(rf$warns, rd$warns, label = paste("RC fm", fm, "warnings"))
  }
})

#-----------------------------------------------------------------------------
#
# These are tests for the conditional pre-tests in the conditional_did_pretest
# function.
#
#-----------------------------------------------------------------------------


test_that("conditional did pre-test", {
  skip_on_cran()  # bootstrap-based CvM test is time-consuming
  sp <- did::reset.sim(time.periods=3, n=1000)
  data <- did::build_sim_dataset(sp)

  cdp <- conditional_did_pretest(yname="Y",
                                 tname="period",
                                 idname="id",
                                 gname="G",
                                 xformla=~X,
                                 data=data)


  # check that test statistic and critical value take reasonable values
  expect_true((cdp$CvM > 0) & (cdp$CvM < 20))
  expect_true((cdp$CvMcval > 0) & (cdp$CvMcval < 100))
})

test_that("pretest setup-bundle/y-override path is bit-identical to the legacy data-copy loop", {
  skip_on_cran()
  # The pretest's per-unit loop reuses a y-invariant setup bundle and passes
  # only a modified outcome vector (dp$setup_precomp / dp$y_override), and
  # compute.att_gt skips post-treatment cells (dp$pretreatment_cells_only).
  # options(did.disable_precompute = TRUE) forces the legacy data-copy loop
  # (the bundle is only consumed on the panel precompute path), so the two
  # runs must agree EXACTLY: same kept cells, same estimator inputs, and no
  # RNG use in compute.att_gt, hence identical(), not just a tolerance.
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(20260401)
  sp <- did::reset.sim(time.periods = 3, n = 150)
  data <- did::build_sim_dataset(sp)
  run_pt <- function() {
    set.seed(42)  # same multiplier-bootstrap draws in both runs
    suppressWarnings(suppressMessages(
      conditional_did_pretest(yname = "Y", tname = "period", idname = "id",
                              gname = "G", xformla = ~X, data = data,
                              est_method = "ipw", biters = 100)))
  }
  options(did.disable_precompute = NULL)
  pt_bundle <- run_pt()
  options(did.disable_precompute = TRUE)
  pt_legacy <- run_pt()
  expect_identical(pt_bundle$CvM, pt_legacy$CvM)
  expect_identical(pt_bundle$CvMpval, pt_legacy$CvMpval)
  expect_identical(as.numeric(pt_bundle$CvMcval), as.numeric(pt_legacy$CvMcval))
  expect_identical(pt_bundle$CvMb, pt_legacy$CvMb)
})

test_that("conditional pre-test CvM is on the bootstrap scale (R>=4.0 orientation regression)", {
  skip_on_cran()
  # Regression guard for the class() == "matrix" length-2 bug (R >= 4.0): with MORE
  # THAN ONE pre-treatment (g,t) cell the observed CvM statistic was left in
  # (n_gt x nX) orientation and therefore scaled by n / n_gt relative to the bootstrap
  # null distribution (which is correctly (nX x n_gt) via test.mboot's t(apply(...))),
  # driving the p-value spuriously to ~0 so the pre-test almost always rejected.
  # mpdta (conditioning on lpop) has several pre-treatment cells and plausibly
  # satisfies conditional parallel trends, so the *fixed* test should NOT reject.
  data(mpdta, package = "did")
  set.seed(42)
  keep <- sample(unique(mpdta$countyreal), 150)
  d <- mpdta[mpdta$countyreal %in% keep, ]
  pt <- suppressWarnings(suppressMessages(
    conditional_did_pretest(yname = "lemp", tname = "year", idname = "countyreal",
                            gname = "first.treat", xformla = ~lpop, data = d,
                            est_method = "ipw", biters = 199)))
  expect_true(is.finite(pt$CvM))
  # the bug inflated CvM ~(n / n_gt)-fold above the critical value -> p ~ 0
  expect_lt(pt$CvM, pt$CvMcval)
  expect_gt(pt$CvMpval, 0.05)
})

#-----------------------------------------------------------------------------
#
# These are tests (primarily coming from github) related to bugs
# encountered by users.
#
#-----------------------------------------------------------------------------

test_that("having column named t1 causes code to crash", {
  data(mpdta, package="did")
  out <- att_gt(yname = "lemp",
                gname = "first.treat",
                idname = "countyreal",
                tname = "year",
                xformla = ~1,
                data = mpdta,
                est_method = "reg",
                control_group="notyettreated"
                )
  mpdta$t1 <- 1

  out <- att_gt(yname = "lemp",
                gname = "first.treat",
                idname = "countyreal",
                tname = "year",
                xformla = ~1,
                data = mpdta,
                est_method = "reg",
                control_group="notyettreated"
                )
  expect_false(is.null(out), "code crashed due to strange variable names")
})

test_that("missing covariates", {
  # should warn about missing data but otherwise run
  data(mpdta, package="did")

  mpdta[1, "lpop"] <- NA
  expect_warning(out <- att_gt(yname = "lemp",
                gname = "first.treat",
                idname = "countyreal",
                tname = "year",
                xformla = ~lpop,
                data = mpdta,
                est_method = "reg",
                control_group="notyettreated"
                ))

})


test_that("repeated cross sections small groups with covariates", {
  # from https://github.com/bcallaway11/did/issues/64
  sp <- did::reset.sim(time.periods=3)
  data <- build_sim_dataset(sp, panel=FALSE)
  data$X2 <- rnorm(nrow(data))
  data$X3 <- rnorm(nrow(data))
  dropids <- unique(subset(data, G==2)$id)
  # interestingly, this seems to return all NA's (which "works")
  # if you only keep one or two observations
  dropids <- dropids[-c(1,2,3)] # keep three observations from group 2
  data <- subset(data, !(id %in% dropids))

  expect_warning(res_dr <- att_gt(yname="Y", xformla=~X+X2+X3, data=data, tname="period", idname="id",
                   gname="G", est_method=DRDID::drdid_rc1, panel=FALSE), "very few observations")

  expect_true(is.numeric(res_dr$se[1]))

  skip("known bug, code crashes in this case, fix is probably in DRDID package")
})


test_that("fewer time periods than groups", {
  # from https://github.com/bcallaway11/did/issues/56
  # not sure if this is actually a bug,
  # can easily circumvent all of these issues by
  # manually recoding the groups
  time.periods <- 6
  sp <- did::reset.sim(time.periods=time.periods)
  sp$te <- 0
  sp$te.e <- 1:time.periods
  data <- build_sim_dataset(sp)
  data <- subset(data, !(period %in% c(2,5)))

  res_dr <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
                   gname="G", est_method="dr")
  res_idx <- which(res_dr$group==2 & res_dr$t==3)
  expect_equal(res_dr$att[res_idx], 2, tol=.5)

  dyn_agg <- aggte(res_dr, type="dynamic")
  dyn_idx <- which(dyn_agg$egt==3)
  expect_equal(dyn_agg$att.egt[dyn_idx], 4, tol=.5)
  expect_false(any(is.na(dyn_agg$att.egt)))

  group_agg <- aggte(res_dr, type="group")
  group_idx <- which(group_agg$egt==3)
  expect_equal(group_agg$att.egt[group_idx], mean(c(1,2,4)), tol=.5)
  expect_false(any(is.na(group_agg$att.egt)))

  cal_agg <- aggte(res_dr, type="calendar")
  expect_false(is.na(cal_agg$att.egt[1]))
})


test_that("0 pre-treatment estimates when outcomes are 0", {
  # from https://github.com/bcallaway11/did/issues/126
  sp <- did::reset.sim(time.periods=10)
  data <- build_sim_dataset(sp)
  data <- subset(data, G != 0) # drop never treated
  data <- subset(data, G > 6)
  data <- subset(data, period > 5)
  data$Y[(data$period < data$G)] <- 0 # set pre-treatment = 0
  res <- suppressWarnings(att_gt(yname="Y",
                tname="period",
                idname="id",
                gname="G",
                data=data,
                control_group = "notyettreated",
                base_period="universal"))
  res_idx <- which(res$group==9 & res$t==7)
  expect_equal(res$att[res_idx],0)

  res <- suppressWarnings(att_gt(yname="Y",
                tname="period",
                idname="id",
                gname="G",
                data=data,
                control_group = "notyettreated",
                base_period="varying"))
  res_idx <- which(res$group==9 & res$t==7)
  expect_equal(res$att[res_idx],0)
})

test_that("variables not in dataset", {
  sp <- did::reset.sim(time.periods=3)
  data <- build_sim_dataset(sp)

  X2  <- factor(data$cluster)

  expect_error(att_gt(yname="Y", xformla=~X2, data=data, tname="period", idname="id", control_group="notyettreated",
                      gname="G", est_method="dr", clustervars="cluster"), " variables are not in data")

})

test_that("groups treated after max(t) but within anticipation window are not coerced to never-treated", {
  # Bug: with anticipation > 0, groups treated shortly after the last observed

  # period may have anticipatory effects during the observed window.
  # These should NOT be coerced to never-treated (and used as controls).
  #
  # Example: tlist = [1,2,3,4,5], gname = 6, anticipation = 2
  # Anticipation effects start at period 6 - 2 = 4 (within observed data).
  # If coerced to never-treated, these contaminated units serve as controls.

  set.seed(20250228)
  n <- 600
  dt <- data.frame(
    id = rep(1:n, each = 5),
    time = rep(1:5, n)
  )
  # Groups: 0 (never-treated), 4 (treated at 4), 6 (treated at 6, after max(t)=5)
  group_vals <- c(rep(0, 200), rep(4, 200), rep(6, 200))
  dt$group <- group_vals[dt$id]
  dt$y <- rnorm(nrow(dt))

  # With anticipation = 0: group 6 is beyond max(t)=5, should be coerced to never-treated
  suppressWarnings({
    res_ant0 <- att_gt(yname = "y", gname = "group", idname = "id", tname = "time",
                       data = dt, anticipation = 0, bstrap = FALSE,
                       faster_mode = FALSE, print_details = FALSE,
                       control_group = "nevertreated")
  })

  # With anticipation = 2: group 6 anticipates at period 4, should NOT be coerced
  # to never-treated. Instead, group 6 should remain as a treated group.
  suppressWarnings({
    res_ant2 <- att_gt(yname = "y", gname = "group", idname = "id", tname = "time",
                       data = dt, anticipation = 2, bstrap = FALSE,
                       faster_mode = FALSE, print_details = FALSE,
                       control_group = "nevertreated")
  })

  # With anticipation = 0, group 6 is coerced to never-treated (only group 4 in results)
  expect_true(all(res_ant0$group == 4))

  # With anticipation = 2, group 6 should NOT be in the control group.
  # It may or may not appear in the results as a treated group (depending on
  # whether there are enough pre-treatment periods), but the key check is
  # that the number of control units differs between the two runs.
  # Group 6 units (200 units) should not be part of the never-treated pool.

  # Run the same test with faster_mode = TRUE for consistency
  suppressWarnings({
    res_ant0_fast <- att_gt(yname = "y", gname = "group", idname = "id", tname = "time",
                            data = dt, anticipation = 0, bstrap = FALSE,
                            faster_mode = TRUE, print_details = FALSE,
                            control_group = "nevertreated")
  })
  suppressWarnings({
    res_ant2_fast <- att_gt(yname = "y", gname = "group", idname = "id", tname = "time",
                            data = dt, anticipation = 2, bstrap = FALSE,
                            faster_mode = TRUE, print_details = FALSE,
                            control_group = "nevertreated")
  })

  # Both paths should agree
  expect_equal(res_ant0$att, res_ant0_fast$att)
  expect_equal(res_ant2$att, res_ant2_fast$att)

  # With anticipation = 2, group 6 should appear as a treated group in results
  # (it has enough pre-treatment periods: periods 1,2,3 are all before 6-2=4)
  expect_true(6 %in% res_ant2$group)
  expect_true(6 %in% res_ant2_fast$group)

  # The results should differ: with anticipation=2, group 6 is no longer
  # in the control pool, so there are more treated groups in the output
  expect_true(length(unique(res_ant2$group)) > length(unique(res_ant0$group)))
})

# =============================================================================
# Tests for covariate handling in compute.att_gt.R (the slow path). The design
# matrix is built once over the full data and row-subset per (g,t) cell, giving:
#   * numeric formulae      -> bit-identical to a per-cell model.matrix()
#   * factor covariates     -> identical to manually-expanded dummy columns
#   * transform formulae    -> work (previously errored); slow == fast
# =============================================================================

# helper: add treatment-contrast dummy columns for a factor, named f_<level>
add_dummies <- function(d, fac = "fac") {
  lv <- levels(d[[fac]])
  for (l in lv[-1L]) d[[paste0("f_", l)]] <- as.numeric(d[[fac]] == l)
  d
}
dummy_formula <- function(d) {
  stats::as.formula(paste("~", paste(grep("^f_", names(d), value = TRUE), collapse = " + ")))
}

test_that("global design slice is bit-identical to per-subset model.matrix for numeric formulae", {
  set.seed(20260608)
  n <- 80
  df <- data.frame(X1 = rnorm(n), X2 = rnorm(n), Xp = runif(n, 1, 5))
  rows <- sample(n, 40)
  for (f in list(~X1, ~X1 + X2, ~X1 * X2, ~X1:X2, ~I(X1^2) + X2, ~log(Xp))) {
    a <- model.matrix(f, df[rows, ])
    b <- model.matrix(f, df)[rows, , drop = FALSE]
    expect_identical(as.vector(a), as.vector(b), label = deparse(f))
    expect_identical(colnames(a), colnames(b), label = deparse(f))
  }
})

test_that("slow path matches faster_mode for numeric / transform formulae", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  set.seed(13)
  data$X2 <- rnorm(nrow(data))
  data$Xpos <- runif(nrow(data), 1, 5)

  # bare numeric, interactions, and transform formulae (the latter previously errored)
  for (f in list(~X + X2, ~X * X2, ~X:X2, ~I(X^2) + X2, ~log(Xpos), ~poly(X, 2))) {
    res_slow <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period",
             idname = "id", gname = "G", est_method = "dr",
             bstrap = FALSE, faster_mode = FALSE)))
    res_fast <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period",
             idname = "id", gname = "G", est_method = "dr",
             bstrap = FALSE, faster_mode = TRUE)))
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste(deparse(f), "ATT"))
    expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10, label = paste(deparse(f), "se"))
  }
})

test_that("RC-path (positional global slice) matches faster_mode", {
  set.seed(20260401)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  set.seed(13)
  data$X2 <- rnorm(nrow(data))
  for (f in list(~X, ~X + X2)) {
    res_slow <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE, faster_mode = FALSE)))
    res_fast <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = data, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE, faster_mode = TRUE)))
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste("RC", deparse(f), "ATT"))
    expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10, label = paste("RC", deparse(f), "se"))
  }
})

test_that("unbalanced panel (positional global slice) matches faster_mode for factor / transform formulae", {
  # the allow_unbalanced_panel branch slices global_mm by POSITION via disdat_rows,
  # so cover it with formulae whose design depends on which rows are present
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  set.seed(42)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.05)), ]
  for (f in list(~fac, ~I(X^2), ~poly(X, 2), ~X * fac)) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             allow_unbalanced_panel = TRUE, faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             allow_unbalanced_panel = TRUE, faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("UB", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("UB", deparse(f), "se"))
  }
})

test_that("RC-path matches faster_mode for factor / data-dependent-basis formulae", {
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  set.seed(42)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.05)), ]
  for (f in list(~fac, ~poly(X, 2))) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE,
             faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d_ub, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE, panel = FALSE,
             faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("RC", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("RC", deparse(f), "se"))
  }
})

test_that("transform / factor formulae match faster_mode under universal base, notyettreated, anticipation", {
  # crosses the hoisted design with the non-default options the precompute path
  # reimplements (universal-base Ypre/Ypost swap, notyettreated C_full, anticipation)
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), TRUE))
  for (f in list(~I(X^2) + X, ~fac)) {
    rs <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             control_group = "notyettreated", base_period = "universal",
             anticipation = 1, faster_mode = FALSE)))
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = f, data = d, tname = "period", idname = "id",
             gname = "G", est_method = "dr", bstrap = FALSE,
             control_group = "notyettreated", base_period = "universal",
             anticipation = 1, faster_mode = TRUE)))
    expect_equal(rs$att, rf$att, tolerance = 1e-10, label = paste("cross", deparse(f), "ATT"))
    expect_equal(rs$se,  rf$se,  tolerance = 1e-10, label = paste("cross", deparse(f), "se"))
  }
})

test_that("global basis is reparameterization-invariant: ~poly(X, 2) equals ~X + I(X^2)", {
  # locks the claim justifying the GLOBAL basis for data-dependent terms (poly/scale/ns):
  # the 2x2 estimators are invariant to a full-rank reparameterization of the covariates,
  # so the orthogonal global basis must reproduce the raw basis spanning the same space
  set.seed(20260401)
  sp <- did::reset.sim()
  d <- did::build_sim_dataset(sp)
  for (est in c("dr", "reg", "ipw")) for (fm in c(FALSE, TRUE)) {
    rp <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~poly(X, 2), data = d, tname = "period",
             idname = "id", gname = "G", est_method = est, bstrap = FALSE,
             faster_mode = fm)))
    rr <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~X + I(X^2), data = d, tname = "period",
             idname = "id", gname = "G", est_method = est, bstrap = FALSE,
             faster_mode = fm)))
    lab <- paste(est, "fm", fm)
    expect_equal(rp$att, rr$att, tolerance = 1e-10, label = paste(lab, "ATT"))
    expect_equal(rp$se,  rr$se,  tolerance = 1e-10, label = paste(lab, "se"))
  }
})

test_that("a factor covariate equals manually-expanded dummies EXACTLY (dense levels)", {
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  data$fac <- factor(sample(c("a", "b", "c"), nrow(data), TRUE))
  data <- add_dummies(data)
  dform <- dummy_formula(data)
  for (est in c("dr", "reg", "ipw")) for (panel in c(TRUE, FALSE)) for (fm in c(TRUE, FALSE)) {
    rf <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = ~fac, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = panel, bstrap = FALSE, faster_mode = fm)))
    rd <- suppressWarnings(suppressMessages(
      att_gt(yname = "Y", xformla = dform, data = data, tname = "period", idname = "id",
             gname = "G", est_method = est, panel = panel, bstrap = FALSE, faster_mode = fm)))
    lab <- paste(est, "panel", panel, "fm", fm)
    expect_equal(rf$att, rd$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_equal(rf$se,  rd$se,  tolerance = 1e-12, label = paste(lab, "se"))
    expect_equal(as.matrix(rf$inffunc), as.matrix(rd$inffunc), tolerance = 1e-12,
                 label = paste(lab, "inffunc"))
  }
})

test_that("transform formulae that evaluate to NaN drop those rows instead of crashing", {
  # Regression: the non-finite-row check must keep NA/NaN rows visible (model.frame
  # with na.pass) so they are dropped; model.matrix(na.pass) silently drops NaN rows,
  # which used to leave them in the data and crash att_gt with 'subscript out of bounds'.
  set.seed(404)
  sp <- did::reset.sim(time.periods = 4)
  data <- did::build_sim_dataset(sp)
  ids <- unique(data$id)
  xv <- stats::setNames(runif(length(ids), -0.5, 3), ids)   # some <= 0  -> log() is NaN
  data$Xneg <- xv[as.character(data$id)]
  logXneg <- suppressWarnings(log(data$Xneg))
  stopifnot(any(!is.finite(logXneg)))                        # the pathological rows exist

  rs <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = FALSE)))
  rf <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = data, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = TRUE)))
  expect_true(any(is.finite(rs$att)))
  expect_equal(rs$att, rf$att, tolerance = 1e-10)
  # equals manually pre-cleaning the NaN units
  d2 <- data[is.finite(logXneg), ]
  rref <- suppressWarnings(suppressMessages(
    att_gt(yname = "Y", xformla = ~log(Xneg), data = d2, tname = "period",
           idname = "id", gname = "G", est_method = "dr", bstrap = FALSE, faster_mode = FALSE)))
  expect_equal(rs$att, rref$att, tolerance = 1e-12)
})

test_that("matrix-valued transformed formula terms drop non-finite rows row-wise", {
  set.seed(20260619)
  data <- did::build_sim_dataset(did::reset.sim(n = 500))
  data$Xpos <- exp(data$X)
  data$Xpos[data$id == unique(data$id)[1]] <- 0
  f <- ~I(cbind(log(Xpos), X^2))

  for (fm in c(FALSE, TRUE)) {
    expect_warning(
      res <- att_gt(yname = "Y", xformla = f, data = data,
                    tname = "period", idname = "id", gname = "G",
                    bstrap = FALSE, faster_mode = fm),
      "missing or non-finite data",
      info = paste("faster_mode", fm)
    )
    expect_s3_class(res, "MP")
    expect_false(anyNA(res$att))
  }
})

test_that("a globally-empty factor level is dropped instead of NA-failing every cell", {
  # Regression: factor(levels = c('a','b','c')) where 'c' never occurs in the data
  # (common after users subset their data, since R keeps empty levels) used to emit
  # an all-zero dummy column in EVERY (g,t) cell, NA-failing the whole estimation
  # with a misleading 'Not enough control units' warning. Globally-empty levels are
  # now dropped once up front on both paths, so results equal the droplevels()-ed
  # data exactly. Levels absent only from particular cells keep their documented
  # NA-that-cell-only behavior (see the sparse-factor test below).
  set.seed(22)
  nunit <- 80
  G <- rep(c(0, 2), each = 40)
  funit <- factor(sample(c("a", "b"), nunit, TRUE), levels = c("a", "b", "c"))
  dat <- expand.grid(id = 1:nunit, period = 1:3)
  dat$G <- G[dat$id]
  dat$flev <- funit[dat$id]
  eta <- rnorm(nunit)
  dat$Y <- eta[dat$id] + dat$period + 1 * (dat$G == 2 & dat$period >= 2) +
    rnorm(nrow(dat), sd = 0.1)
  dat2 <- dat
  dat2$flev <- droplevels(dat2$flev)

  run <- function(d, fm, est) suppressWarnings(suppressMessages(
    att_gt(yname = "Y", tname = "period", idname = "id", gname = "G", data = d,
           xformla = ~flev, est_method = est, bstrap = FALSE, faster_mode = fm)))

  for (fm in c(FALSE, TRUE)) for (est in c("dr", "ipw", "reg")) {
    lab <- paste("fm", fm, est)
    res <- run(dat, fm, est)    # empty level 'c' declared
    ref <- run(dat2, fm, est)   # empty level pre-dropped by the user
    expect_false(anyNA(res$att), label = paste(lab, "no NA"))
    expect_equal(res$att, ref$att, tolerance = 1e-12, label = paste(lab, "ATT"))
    expect_equal(res$se,  ref$se,  tolerance = 1e-12, label = paste(lab, "se"))
    expect_equal(as.matrix(res$inffunc), as.matrix(ref$inffunc), tolerance = 1e-12,
                 label = paste(lab, "inffunc"))
  }

  # known values (master's slow path estimated these with the empty level present)
  res_slow <- run(dat, FALSE, "dr")
  res_fast <- run(dat, TRUE, "dr")
  expect_equal(res_slow$att, c(1.0275, 0.9826), tolerance = 1e-3)
  expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10)
  expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10)
})

test_that("a sparse factor (level absent from some cells) matches manual dummies, incl. warnings", {
  # 'b' appears only in group 2, so the never-treated control never has it; every cell
  # is rank-deficient and must return NA -- exactly as manual dummies do, with the same
  # warnings, and WITHOUT crashing (the old per-cell droplevels() raised a contrast error).
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  fac <- rep("a", nrow(data)); fac[data$G == 2] <- sample(c("a", "b"), sum(data$G == 2), TRUE)
  data$fac <- factor(fac)
  data$f_b <- as.numeric(data$fac == "b")

  collect <- function(f, fm) {
    ws <- character(0)
    res <- withCallingHandlers(
      suppressMessages(att_gt(yname = "Y", xformla = f, data = data, tname = "period",
                              idname = "id", gname = "G", est_method = "reg",
                              panel = TRUE, bstrap = FALSE, faster_mode = fm)),
      warning = function(w) { ws[[length(ws) + 1]] <<- conditionMessage(w); invokeRestart("muffleWarning") })
    list(att = res$att, se = res$se, warns = sort(unique(ws)))
  }
  for (fm in c(TRUE, FALSE)) {
    rf <- collect(~fac, fm)
    rd <- collect(~f_b, fm)
    expect_equal(rf$att, rd$att, label = paste("fm", fm, "ATT"))     # both all-NA, same cells
    expect_equal(rf$se,  rd$se,  label = paste("fm", fm, "se"))
    expect_identical(rf$warns, rd$warns, label = paste("fm", fm, "warnings"))
  }
})

test_that("RC sparse factor (level absent from some cells) matches manual dummies, incl. warnings", {
  # same construction as the panel test above, on the repeated-cross-sections path
  # (positional global_mm slicing): 'b' appears only in group 2, so every cell is
  # rank-deficient and must NA-fail exactly as manual dummies do, in both modes
  set.seed(11)
  sp <- did::reset.sim()
  data <- did::build_sim_dataset(sp)
  fac <- rep("a", nrow(data)); fac[data$G == 2] <- sample(c("a", "b"), sum(data$G == 2), TRUE)
  data$fac <- factor(fac)
  data$f_b <- as.numeric(data$fac == "b")

  collect <- function(f, fm) {
    ws <- character(0)
    res <- withCallingHandlers(
      suppressMessages(att_gt(yname = "Y", xformla = f, data = data, tname = "period",
                              idname = "id", gname = "G", est_method = "reg",
                              panel = FALSE, bstrap = FALSE, faster_mode = fm)),
      warning = function(w) { ws[[length(ws) + 1]] <<- conditionMessage(w); invokeRestart("muffleWarning") })
    list(att = res$att, se = res$se, warns = sort(unique(ws)))
  }
  for (fm in c(TRUE, FALSE)) {
    rf <- collect(~fac, fm)
    rd <- collect(~f_b, fm)
    expect_equal(rf$att, rd$att, label = paste("RC fm", fm, "ATT"))
    expect_equal(rf$se,  rd$se,  label = paste("RC fm", fm, "se"))
    expect_equal(is.na(rf$att), is.na(rd$att), label = paste("RC fm", fm, "NA pattern"))
    expect_identical(rf$warns, rd$warns, label = paste("RC fm", fm, "warnings"))
  }
})

# =============================================================================
# Slow-path (faster_mode = FALSE) panel precompute: the per-period-block fast path
# must be BIT-IDENTICAL to the original get_wide_data() reshape. The escape hatch
# options(did.disable_precompute = TRUE) forces the original path, so we can compare
# the two implementations directly. Also guards the precompute precondition (the
# panel is re-sorted by (id, period) in preprocessing, so input row order is
# irrelevant).
# =============================================================================

test_that("panel precompute is bit-identical to the original get_wide_data path", {
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(101)
  sp <- did::reset.sim(time.periods = 5)
  d <- did::build_sim_dataset(sp)
  for (cg in c("nevertreated", "notyettreated"))
  for (em in c("dr", "reg", "ipw"))
  for (bp in c("varying", "universal"))
  for (ant in c(0, 1)) {
    options(did.disable_precompute = TRUE)
    ref <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE,
      control_group = cg, est_method = em, base_period = bp, anticipation = ant)))
    options(did.disable_precompute = FALSE)
    new <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE,
      control_group = cg, est_method = em, base_period = bp, anticipation = ant)))
    lab <- paste(cg, em, bp, "ant", ant)
    expect_equal(ref$att, new$att, tolerance = 0, label = paste(lab, "ATT"))
    expect_equal(as.matrix(ref$inffunc), as.matrix(new$inffunc), tolerance = 0,
                 label = paste(lab, "IF"))
    expect_equal(is.na(ref$att), is.na(new$att), label = paste(lab, "NA pattern"))
  }
})

test_that("panel precompute also bit-identical for factor / transform covariates", {
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(202)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$fac <- factor(sample(c("a", "b", "c"), nrow(d), replace = TRUE))
  for (xf in c(~X, ~fac, ~I(X^2), ~X + fac)) {
    options(did.disable_precompute = TRUE)
    ref <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = xf, data = d,
      tname = "period", idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE)))
    options(did.disable_precompute = FALSE)
    new <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = xf, data = d,
      tname = "period", idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE)))
    expect_equal(ref$att, new$att, tolerance = 0, label = deparse(xf))
    expect_equal(as.matrix(ref$inffunc), as.matrix(new$inffunc), tolerance = 0,
                 label = deparse(xf))
  }
})

test_that("panel precompute also bit-identical with time-varying sampling weights", {
  # the precompute retains per-period weights (period_w[[min(t + tfac, pret)]]);
  # time-varying weights are the only sim quantity that can expose a wrong-period index
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(303)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$tvw <- d$period + runif(nrow(d), -0.1, 0.1)
  for (bp in c("varying", "universal")) {
    options(did.disable_precompute = TRUE)
    ref <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", weightsname = "tvw",
      faster_mode = FALSE, bstrap = FALSE, est_method = "dr", base_period = bp)))
    options(did.disable_precompute = FALSE)
    new <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
      tname = "period", idname = "id", gname = "G", weightsname = "tvw",
      faster_mode = FALSE, bstrap = FALSE, est_method = "dr", base_period = bp)))
    expect_equal(ref$att, new$att, tolerance = 0, label = paste("tvw", bp, "ATT"))
    expect_equal(as.matrix(ref$inffunc), as.matrix(new$inffunc), tolerance = 0,
                 label = paste("tvw", bp, "IF"))
  }
})

test_that("panel precompute bit-identical with fix_weights base_period/first_period", {
  # the precompute gate admits fix_weights = "base_period"/"first_period": only the
  # weight vector differs (weights_by_period[[pret_g]] / [[1L]] instead of
  # period_w[[earlier_idx]]), so the assembly must match the get_wide_data fallback
  # exactly; time-varying weights are the only sim quantity that can expose a
  # wrong-period weight index
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(808)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$tvw <- d$period + runif(nrow(d), -0.1, 0.1)
  configs <- list()
  for (fw in c("base_period", "first_period")) {
    for (bp in c("varying", "universal"))
    for (cg in c("nevertreated", "notyettreated")) {
      configs[[length(configs) + 1]] <- list(fw = fw, bp = bp, cg = cg, em = "dr")
    }
    for (em in c("reg", "ipw")) {
      configs[[length(configs) + 1]] <- list(fw = fw, bp = "varying",
                                             cg = "nevertreated", em = em)
    }
  }
  for (cfg in configs) {
    run <- function() suppressWarnings(suppressMessages(att_gt(yname = "Y",
      xformla = ~X, data = d, tname = "period", idname = "id", gname = "G",
      weightsname = "tvw", fix_weights = cfg$fw, faster_mode = FALSE, bstrap = FALSE,
      est_method = cfg$em, base_period = cfg$bp, control_group = cfg$cg)))
    options(did.disable_precompute = TRUE)
    ref <- run()
    options(did.disable_precompute = FALSE)
    new <- run()
    lab <- paste(cfg$fw, cfg$bp, cfg$cg, cfg$em)
    expect_equal(ref$att, new$att, tolerance = 0, label = paste(lab, "ATT"))
    expect_equal(ref$se,  new$se,  tolerance = 0, label = paste(lab, "se"))
    expect_equal(as.matrix(ref$inffunc), as.matrix(new$inffunc), tolerance = 0,
                 label = paste(lab, "IF"))
  }
})

test_that("RC/unbalanced precompute is bit-identical to the legacy subset path", {
  # the positional assembly (per-period row indices + plain vectors) must equal
  # the legacy rightids/%in%/data.table-subset construction exactly; the escape
  # hatch options(did.disable_precompute = TRUE) forces the legacy path
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(404)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  set.seed(99)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.10)), ]
  configs <- list(
    rc_nev_var_dr = list(data = d, args = list(panel = FALSE,
      control_group = "nevertreated", base_period = "varying", est_method = "dr")),
    rc_nyt_uni_dr = list(data = d, args = list(panel = FALSE,
      control_group = "notyettreated", base_period = "universal", est_method = "dr")),
    rc_ipw = list(data = d, args = list(panel = FALSE,
      control_group = "nevertreated", base_period = "varying", est_method = "ipw")),
    rc_nyt_ant = list(data = d, args = list(panel = FALSE,
      control_group = "notyettreated", base_period = "varying", est_method = "dr",
      anticipation = 1)),
    unb_nyt_var = list(data = d_ub, args = list(allow_unbalanced_panel = TRUE,
      control_group = "notyettreated", base_period = "varying", est_method = "dr")),
    unb_nev_uni = list(data = d_ub, args = list(allow_unbalanced_panel = TRUE,
      control_group = "nevertreated", base_period = "universal", est_method = "dr"))
  )
  for (nm in names(configs)) {
    cfg <- configs[[nm]]
    common <- c(list(yname = "Y", xformla = ~X, data = cfg$data, tname = "period",
                     idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE),
                cfg$args)
    options(did.disable_precompute = TRUE)
    ref <- suppressWarnings(suppressMessages(do.call(att_gt, common)))
    options(did.disable_precompute = FALSE)
    new <- suppressWarnings(suppressMessages(do.call(att_gt, common)))
    expect_equal(ref$att, new$att, tolerance = 0, label = paste(nm, "ATT"))
    expect_equal(ref$se,  new$se,  tolerance = 0, label = paste(nm, "se"))
    expect_equal(as.matrix(ref$inffunc), as.matrix(new$inffunc), tolerance = 0,
                 label = paste(nm, "IF"))
  }
})

test_that("RC/unbalanced precompute bit-identical with fix_weights base/first, incl. dropped units", {
  # the per-period .rowid/.w lookup vectors must reproduce the legacy per-cell
  # data[t_col == target_period, ] subset + match() exactly, including the
  # dropped-units branch (units missing from the target period) and its warnings
  old_opt <- getOption("did.disable_precompute")
  withr::defer(options(did.disable_precompute = old_opt))
  set.seed(505)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  d$tvw <- d$period + runif(nrow(d), -0.1, 0.1)
  set.seed(99)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.10)), ]
  collect <- function(fw) {
    ws <- character(0)
    res <- withCallingHandlers(
      suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d_ub, tname = "period",
        idname = "id", gname = "G", allow_unbalanced_panel = TRUE, weightsname = "tvw",
        fix_weights = fw, est_method = "dr", faster_mode = FALSE, bstrap = FALSE)),
      warning = function(w) { ws[[length(ws) + 1]] <<- conditionMessage(w); invokeRestart("muffleWarning") })
    list(att = res$att, se = res$se, inffunc = as.matrix(res$inffunc), warns = ws)
  }
  for (fw in c("base_period", "first_period")) {
    options(did.disable_precompute = TRUE)
    ref <- collect(fw)
    options(did.disable_precompute = FALSE)
    new <- collect(fw)
    expect_equal(ref$att, new$att, tolerance = 0, label = paste(fw, "ATT"))
    expect_equal(ref$se,  new$se,  tolerance = 0, label = paste(fw, "se"))
    expect_equal(ref$inffunc, new$inffunc, tolerance = 0, label = paste(fw, "IF"))
    expect_identical(ref$warns, new$warns, label = paste(fw, "warnings"))
    # the dropped-units branch must actually be exercised
    expect_true(any(grepl("^Dropped", new$warns)), label = paste(fw, "drop branch hit"))
  }
})

test_that("RC/unbalanced precompute keeps slow-fast parity", {
  set.seed(606)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  set.seed(99)
  d_ub <- d[-sample(nrow(d), floor(nrow(d) * 0.10)), ]
  for (cfg in list(list(data = d, extra = list(panel = FALSE)),
                   list(data = d_ub, extra = list(allow_unbalanced_panel = TRUE)))) {
    common <- c(list(yname = "Y", xformla = ~X, data = cfg$data, tname = "period",
                     idname = "id", gname = "G", est_method = "dr", bstrap = FALSE),
                cfg$extra)
    res_slow <- suppressWarnings(suppressMessages(do.call(att_gt, c(common, list(faster_mode = FALSE)))))
    res_fast <- suppressWarnings(suppressMessages(do.call(att_gt, c(common, list(faster_mode = TRUE)))))
    lab <- paste(names(cfg$extra)[1])
    expect_equal(res_slow$att, res_fast$att, tolerance = 1e-10, label = paste(lab, "ATT"))
    expect_equal(res_slow$se,  res_fast$se,  tolerance = 1e-10, label = paste(lab, "se"))
  }
})

test_that("input row order does not affect slow-path results (precompute precondition)", {
  set.seed(7)
  sp <- did::reset.sim(time.periods = 4)
  d <- did::build_sim_dataset(sp)
  ref <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X, data = d,
    tname = "period", idname = "id", gname = "G", faster_mode = FALSE, bstrap = FALSE)))
  shf <- suppressWarnings(suppressMessages(att_gt(yname = "Y", xformla = ~X,
    data = d[sample(nrow(d)), ], tname = "period", idname = "id", gname = "G",
    faster_mode = FALSE, bstrap = FALSE)))
  expect_equal(ref$att, shf$att, tolerance = 1e-12)
  expect_equal(ref$se, shf$se, tolerance = 1e-12)
})

test_that("att_gt does not mutate caller data in either implementation", {
  set.seed(20260619)
  sp <- did::reset.sim(n = 500)
  panel_data <- did::build_sim_dataset(sp)
  rc_data <- did::build_sim_dataset(sp, panel = FALSE)

  cases <- list(
    panel_df = list(data = as.data.frame(panel_data), panel = TRUE),
    panel_dt = list(data = data.table::as.data.table(panel_data), panel = TRUE),
    rc_df = list(data = as.data.frame(rc_data), panel = FALSE),
    rc_dt = list(data = data.table::as.data.table(rc_data), panel = FALSE)
  )

  for (case_name in names(cases)) {
    for (fm in c(FALSE, TRUE)) {
      d <- cases[[case_name]]$data
      before_names <- names(d)
      before_data <- as.data.frame(d)

      suppressWarnings(suppressMessages(
        att_gt(yname = "Y", tname = "period", idname = "id", gname = "G",
               xformla = ~X, data = d, panel = cases[[case_name]]$panel,
               bstrap = FALSE, faster_mode = fm)
      ))

      expect_identical(names(d), before_names,
                       info = paste(case_name, "faster_mode", fm))
      expect_equal(as.data.frame(d), before_data,
                   ignore_attr = TRUE,
                   info = paste(case_name, "faster_mode", fm))
    }
  }
})

#' @title gplot
#'
#' @description does the heavy lifting for making a plot of an group-time
#'  average treatment effect
#'
#' @inheritParams ggdid
#'
#' @return a `ggplot2` object
#'
#' @keywords internal
#'
#' @export
gplot <- function(ssresults, ylim=NULL, xlab=NULL, ylab=NULL, title="Group", xgap=1,
                  legend=TRUE, ref_line = 0, theming = TRUE) {
  unique_years <- sort(unique(as.numeric(as.character(ssresults$year))))
  xgap_int <- max(1L, as.integer(round(xgap)))
  dabreaks <- unique_years[seq(1, length(unique_years), by = xgap_int)]

  c.point <-  qnorm(1 - ssresults$alp/2)

  p <- ggplot(ssresults,
              aes(x=as.numeric(as.character(year)), y=att, ymin=(att-c*att.se),
                  ymax=(att+c*att.se))) +

    geom_point(aes(colour=post), size=1.5) +
    #geom_ribbon(aes(x=as.numeric(year)), alpha=0.2) +
    geom_errorbar(aes(colour=post), width=0.1) +
    scale_y_continuous(limits=ylim) +
    #scale_x_discrete(breaks=dabreaks, labels=as.character(dabreaks)) +
    scale_x_continuous(breaks=as.numeric(dabreaks), labels=as.character(dabreaks)) +
    scale_color_manual(drop=FALSE, values=c("#e87d72","#56bcc2"), breaks = c(0, 1), labels = c('Pre','Post')) +
    labs(x = xlab, y = ylab, title = title, color = NULL)

  if (!is.null(ref_line)) {
    p <- p + geom_hline(aes(yintercept = ref_line), linetype = 'dashed')
  }
  if (theming) {
    p <- p + theme_classic() +
      theme(plot.title = element_text(color="darkgray", face="bold", size=12),
            axis.title = element_text(color="darkgray", face="bold", size=12),
            strip.background = element_rect(fill = 'white', color = 'white'),
            strip.text = element_text(color = 'darkgray', face = 'bold', size = 12, hjust = 0),
            legend.position = 'bottom')
  }
  if (!legend) {
    p  <- p + theme(legend.position = "none")
  }

  p
}


#' @title splot
#'
#' @description alternative plotting function for the case with selective treatment
#'  timing
#'
#' @inheritParams ggdid
#' @inheritParams gplot
#'
#' @return a `ggplot2` object
#'
#' @keywords internal
#'
#' @export
splot <- function(ssresults, ylim=NULL, xlab=NULL, ylab=NULL, title="Group",
                  legend=TRUE, ref_line = 0, theming = TRUE) {

  # names of variables are "weird" for this function because this code builds
  # on the same infrastructure as for plotting group-time average treatment
  # effects and aggregations using event time or calendar time

  # the group variable is saved in year

  if (is.null(xlab)) {
    xlab <- 'ATT'
  }
  if (is.null(ylab)) {
    ylab <- 'Group'
  }

  errorbar_layer <- if (utils::packageVersion("ggplot2") >= "3.3.0") {
    geom_errorbar(aes(colour=post), width=0.1, orientation="y")
  } else {
    geom_errorbarh(aes(colour=post), height=0.1)
  }

  p <- ggplot(ssresults,
              aes(y=as.factor(year), x=att, xmin=(att-c*att.se),
                  xmax=(att+c*att.se))) +
    geom_point(aes(colour=post), size=1.5) +
    #geom_ribbon(aes(x=as.numeric(year)), alpha=0.2) +
    errorbar_layer +
    scale_y_discrete(breaks=as.factor(ssresults$year)) +
    #scale_x_discrete(breaks=dabreaks, labels=as.character(dabreaks)) +
    scale_x_continuous(limits=ylim) +
    scale_color_manual(drop=FALSE, values=c("#e87d72","#56bcc2"), breaks = c(0, 1), labels = c('Pre','Post')) +
    labs(x = xlab, y = ylab, title = title)

  if (!is.null(ref_line)) {
    p <- p + geom_vline(aes(xintercept = ref_line), linetype = 'dashed')
  }
  if (theming) {
    p <- p + theme_classic() +
      theme(plot.title = element_text(color="darkgray", face="bold", size=12),
            axis.title = element_text(color="darkgray", face="bold", size=12),
            legend.position = 'none')
  }

  if (!legend) {
    p  <- p + theme(legend.position = "none")
  }

  p
}

#' @title Plot `did` objects using `ggplot2`
#'
#' @description Function to plot objects from the `did` package
#'
#' @param object either a `MP` object or `AGGTEobj` object. See `help(ggdid.MP)` and `help(ggdid.AGGTEobj)`.
#' @param ... other arguments
#'
#' @examples
#' \dontrun{
#' data(mpdta)
#' out <- att_gt(yname = "lemp",
#'               gname = "first.treat",
#'               idname = "countyreal",
#'               tname = "year",
#'               xformla = ~1,
#'               data = mpdta)
#'
#' # plot all group-time average treatment effects
#' ggdid(out)
#'
#' # plot event study aggregation
#' es <- aggte(out, type = "dynamic")
#' ggdid(es)
#' }
#'
#' @export
ggdid <- function(object, ...) {
  UseMethod("ggdid", object)
}


## #' @param type the type of plot, should be one of "attgt", "dynamic",
## #'  "group", "calendar", "dynsel".  "attgt" is the default and plots
## #'  all group-time average treatment effects separately by group (including
## #'  pre-treatment time periods); "dynamic" plots dynamic treatment effects --
## #'  these are the same as event studies; "group" plots average effects
## #'  of the treatment separately by group (which allows for selective treatment
## #'  timing); "calendar" plots average treatment effects by time period; and
## #'  "dynsel" plots dynamic effects allowing for selective treatment timing
## #'  (this also requires setting the additional paramater e1)


#' @title Plot `MP` objects using `ggplot2`
#'
#' @description A function to plot `MP` objects
#'
#' @inheritParams ggdid
#' @param object an `MP` object, as returned by `att_gt`.
#' @param ylim optional y limits for the plot; setting here makes the y limits
#'  the same across different plots
#' @param xlab optional x-axis label
#' @param ylab optional y-axis label
#' @param title optional plot title
#' @param xgap optional gap between the labels on the x-axis.  For example,
#'  `xgap=3` indicates that the labels should show up for every third
#'  value on the x-axis.  The default is 1.
#' @param ncol The number of columns to include in the resulting plot.  The
#'  default is 1.
#' @param legend Whether or not to include a legend (which will indicate color
#'  of pre- and post-treatment estimates).  Default is `TRUE`.
#' @param group Vector for which groups to include in the plots of ATT(g,t).
#'   Default is NULL, and, in this case, plots for all groups will be included (`ggdid.MP` only).
#' @param ref_line A reference line at this value, usually to compare confidence
#'   intervals to 0. Set to NULL to omit.
#' @param theming Set to FALSE to skip all theming so you can do it yourself.
#' @param grtitle Title to append before each group name (`ggdid.MP` only).
#'
#' @export
ggdid.MP <- function(object,
                     ylim=NULL,
                     xlab=NULL,
                     ylab=NULL,
                     title="Group",
                     xgap=1,
                     ncol=1,
                     legend=TRUE,
                     group=NULL,
                     ref_line = 0,
                     theming = TRUE,
                     grtitle = "Group",
                     ...) {

  mpobj <- object

  G <- length(unique(mpobj$group))
  Y <- length(unique(mpobj$t))## drop 1 period bc DID
  g <- unique(mpobj$group)[order(unique(mpobj$group))] ## -1 to drop control group
  y <- unique(mpobj$t)

  results <- data.frame(year=rep(y,G))
  results$group <- unlist(lapply(g, function(x) { rep(x, Y) }))
  results$grtitle <- paste(grtitle,results$group)
  results$att <- mpobj$att
  n <- mpobj$n
  results$att.se <- mpobj$se #sqrt(diag(mpobj$V)/n)
  results$post <- as.factor(1*(results$year >= results$group))
  results$year <- as.factor(results$year)
  results$c <- mpobj$c
  #vcovatt <- mpobj$V/n
  alp <- mpobj$alp


  if(is.null(group)) group <- g
  if(any(!(group %in% g))){
    warning('Some of the specified groups do not exist in the data. Reporting all available groups.')
    group <- g
  }

  mplots <- gplot(results[results$group %in% group,],
                  ylim, xlab, ylab, title, xgap,
                  legend, ref_line, theming) +
    facet_wrap(~grtitle, ncol = ncol, scales = 'free')

  return(mplots)
}


#' @title Plot `AGGTEobj` objects
#'
#' @description A function to plot `AGGTEobj` objects
#'
#' @inheritParams ggdid.MP
#' @param object an `AGGTEobj` object, as returned by `aggte`.
#'
#' @export
ggdid.AGGTEobj <- function(object,
                           ylim=NULL,
                           xlab=NULL,
                           ylab=NULL,
                           title="",
                           xgap=1,
                           legend=TRUE,
                           ref_line = 0,
                           theming = TRUE,
                           ...) {

  if ( !(object$type %in% c("dynamic","group","calendar")) ) {
    stop(paste0("Plot method not available for this type of aggregation"))
  }

  post.treat <- 1*(object$egt >= 0)
  results <- cbind.data.frame(year=object$egt,
                              att=object$att.egt,
                              att.se=object$se.egt,
                              post=as.factor(post.treat))
  alp <- object$DIDparams$alp
  if (is.null(alp)) alp <- 0.05
  results$c <- if (is.null(object$crit.val.egt)) qnorm(1 - alp/2) else object$crit.val.egt

  if (title == "") {
    # get title right depending on which aggregation
    title <- if (object$type=="group") {
      "Average Effect by Group"
    } else if (object$type=="dynamic") {
      "Average Effect by Length of Exposure"
    } else {
      "Average Effect by Time Period"
    }
  }

  if (object$type == "group") {
    # alternative plot if selective/group treatment timing plot
    p <- splot(results, ylim, xlab, ylab, title, legend, ref_line, theming)
  } else {
    p <- gplot(results, ylim, xlab, ylab, title, xgap, legend, ref_line, theming)
  }

  p
}

#' @title validate_args
#' @description A utility function to validate arguments passed to a att_gt()
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return nothing, but throws an error if any of the arguments are not valid
#' @noRd
validate_args <- function(args, data){

  data_names <- names(data)

  # ---------------------- Error Checking ----------------------
  args$control_group <- args$control_group[1]
  # Flag for control group types
  control_group_message <- "control_group must be either 'nevertreated' or 'notyettreated'"
  dreamerr::check_set_arg(args$control_group, "match", .choices = c("nevertreated", "notyettreated"), .message = control_group_message, .up = 1)

  # Flag for tname, gname, yname
  name_message <- "__ARG__ must be a character scalar and a name of a column from the dataset."
  dreamerr::check_set_arg(args$tname, args$gname, args$yname, "match", .choices = data_names, .message = name_message, .up = 1)

  # Strip idname from clustervars before validation: users may naturally pass
  # clustervars = c(idname, extra_var), and idname clustering is implicit/redundant.
  # This must happen before the dreamerr check which enforces scalar length.
  if (!is.null(args$clustervars) && !is.null(args$idname) && (args$idname %in% args$clustervars)) {
    args$clustervars <- setdiff(args$clustervars, args$idname)
    if (length(args$clustervars) == 0L) args$clustervars <- NULL
  }

  # At most one cluster variable beyond idname is supported (clustering at the
  # unit level via idname is implicit). Checked before the dreamerr scalar check
  # below so the error wording matches the slow path (pre_process_did) and
  # mboot(), instead of a dreamerr message exposing internal argument names.
  if (length(args$clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # Flag for clustervars and weightsname
  checkvar_message <- "__ARG__ must be NULL or a character scalar that is a name of a column from the dataset."
  dreamerr::check_set_arg(args$weightsname, args$clustervars, "NULL | match", .choices = data_names, .message = checkvar_message, .up = 1)

  # check if time periods are numeric
  if(!data[, is.numeric(get(args$tname))]){stop("The time variable '", args$tname, "' must be numeric. Please convert it.")}

  # Check if gname is numeric
  if(!data[, is.numeric(get(args$gname))]){stop("The group variable '", args$gname, "' must be numeric. Please convert it.")}

  # Check if yname is numeric (logical 0/1 outcomes are also allowed)
  if(!data[, is.numeric(get(args$yname)) || is.logical(get(args$yname))]){stop("The outcome variable '", args$yname, "' must be numeric. Please convert it.")}

  # Flag for idname
  if(!is.null(args$idname)){
    # Check if idname is in the data
    name_message <- "__ARG__ must be a character scalar and a name of a column from the dataset."
    dreamerr::check_set_arg(args$idname, "match", .choices = data_names, .message = name_message, .up = 1)

    #  check if idname is numeric
    if(!data[, is.numeric(get(args$idname))]){stop("The id variable '", args$idname, "' must be numeric. Please convert it.")}

    # Check if gname is unique by idname: irreversibility of the treatment
    # Use direct column access instead of get() for speed
    id_g_unique <- unique(data[, c(args$idname, args$gname), with = FALSE])
    check_treatment_uniqueness <- anyDuplicated(id_g_unique[[1]]) == 0L
    if (!check_treatment_uniqueness) {
      stop("The value of gname (treatment variable) must be the same across all periods for each particular unit. The treatment must be irreversible.")
    }

    # Check if any combination of idname and tname is duplicated
    n_id_year <- anyDuplicated(data, by = c(args$idname, args$tname))
    # If any combination is duplicated, stop execution and throw an error
    if (n_id_year > 0) {
      stop("The value of idname must be unique (by tname). Some units are observed more than once in a period.")
    }
  }

  # Flag for base period: not in c("universal", "varying"), stop
  args$base_period <- args$base_period[1]
  base_period_message <- "base_period must be either 'universal' or 'varying'."
  dreamerr::check_set_arg(args$base_period, "match", .choices = c("universal", "varying"), .message = base_period_message, .up = 1)

  # Flags for cluster variable
  # Note: idname was already stripped from clustervars and the at-most-one check
  # was enforced above (before the dreamerr check)
  if (!is.null(args$clustervars)) {
    # Check that cluster variables do not vary over time within each unit. For true repeated cross-sections
    # the user may omit idname (an internal ".rowid" is created later, one observation per row), so there is
    # no within-unit time variation to check and idname is not yet available here -- skip the check then.
    if (length(args$clustervars) > 0 && !is.null(args$idname)) {
      # Efficiently check for time-varying cluster variables
      clust_tv <- data[, lapply(.SD, function(col) length(unique(col)) == 1), by = get(args$idname), .SDcols = args$clustervars]
      # If any cluster variable varies over time within any unit, stop execution
      # (same wording as the slow path and mboot())
      if (!all(unlist(clust_tv[, -1, with = FALSE]))) {
        stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
      }
    }
  }

  # Check if anticipation is numeric using
  if (!is.numeric(args$anticipation)) {
    stop("anticipation must be numeric. Please convert it.")
  }

  # Check if anticipation is positive
  if (args$anticipation < 0) {
    stop("anticipation must be non-negative. Please check your arguments.")
  }

}

#' @title DiD standardization
#' @description A utility function to coerce the data into standard format
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return A list containing ordered data and arguments
#' @noRd
did_standardization <- function(data, args){
  # keep relevant columns plus the RAW covariate variables (all.vars of xformla), so
  # model.matrix(args$xformla, .) can be rebuilt downstream. Keeping the evaluated
  # model.frame instead would lose the raw variable for transform formulae
  # (~I(X^2), ~poly(X, 2), ~log(X)) and create matrix-valued columns that break the
  # data.table coercion. For bare-variable / factor formulae this is identical.
  xvars <- all.vars(args$xformla)
  cols_to_keep <- unique(c(args$idname, args$tname, args$gname, args$yname,
                           args$weightsname, args$clustervars, xvars))
  # Subset the dataset to keep only the relevant columns
  data <- data[, ..cols_to_keep]

  # Check if any covariates were missing
  n_orig <- data[, .N]
  data <- data[complete.cases(data)]
  # also drop rows whose EVALUATED design is non-finite (e.g. log of a non-positive
  # covariate); safe now that raw-covariate NAs are removed (so poly()/ns()/... will
  # not error on NA input). Use model.frame (NOT model.matrix) with na.action =
  # na.pass: model.frame keeps every row -- including NA/NaN-valued terms -- so the
  # complete.cases() mask stays aligned with `data` (model.matrix would silently drop
  # NaN rows, shortening the mask and letting the offending rows survive). Inf-valued
  # terms are kept, matching the prior behavior.
  if (length(xvars) > 0L && data[, .N] > 0L) {
    mf_check <- suppressWarnings(model.frame(args$xformla, data = data, na.action = na.pass))
    finite_rows <- complete.cases(mf_check)
    if (!all(finite_rows)) data <- data[finite_rows]
  }
  n_new <- data[, .N]
  n_diff <- n_orig - n_new
  if (n_diff != 0) {
    warning(paste0("dropped ", n_diff, " rows from original data due to missing data"))
  }

  # Set weights
  if (is.null(args$weightsname)) weights <- rep(1, n_new) else weights <- data[[args$weightsname]]
  # Validate user-supplied weights: negative weights flip signs and a non-positive
  # mean divides by ~0 during normalization, silently producing NA/NaN ATTs.
  if (!is.null(args$weightsname) && (any(weights < 0, na.rm = TRUE) || isTRUE(mean(weights, na.rm = TRUE) <= 0)))
    stop("The weights variable '", args$weightsname, "' must be non-negative with a positive mean.")
  # enforcing normalization of weights. At this point we already drop any missing values in weights.
  weights <- weights/mean(weights)
  if (".w" %in% names(data)) stop("Your data already contains a column named '.w', which is reserved for internal use by `did`. Please rename this column before calling att_gt().")
  data.table::set(data, j = ".w", value = weights)

  # Check for time-varying weights in panel data
  if (!is.null(args$weightsname) && args$panel) {
    w_range <- data[, .(w_range = max(.w) - min(.w)), by = c(args$idname)]
    if (any(w_range$w_range > .Machine$double.eps^0.5, na.rm = TRUE)) {
      message(
        "Time-varying weights detected. For balanced panel data, the default ",
        "behavior uses the weight from the earlier of the two time periods in ",
        "each 2x2 comparison (the base period for post-treatment cells). ",
        "Use the 'fix_weights' argument to control this behavior. ",
        "See ?att_gt for details."
      )
    }
  }

  # get a list of dates from min to max
  tlist <- data[, sort(unique(get(args$tname)))]

  # Ensure gname column is numeric (double) so it can hold Inf
  if (is.integer(data[[args$gname]])) {
    data[, (args$gname) := as.numeric(get(args$gname))]
  }

  # Coerce control group identified with zero as Inf
  data[get(args$gname) == 0, (args$gname) := Inf]

  # Working out the dates
  # Identify groups with treatment time bigger than the maximum treatment time + anticipation.
  # We account for anticipation because units treated shortly after the last observed period
  # may already exhibit anticipatory effects during the observed time window, and should
  # not be used as controls.
  max_treatment_time <- max(tlist, na.rm = TRUE)
  # gname has no NAs here (complete cases were enforced above), so the comparison
  # cannot yield NA; keep it as a local vector instead of a temporary column
  asif_never_treated <- data[[args$gname]] > max_treatment_time + args$anticipation
  # set gname to Inf for those groups considered as never treated
  if (any(asif_never_treated)) {
    data.table::set(data, i = which(asif_never_treated), j = args$gname, value = Inf)
  }

  # get list of treated groups (by time) from min to max
  glist <- sort(unique(data[[args$gname]]))

  # Check if there is a never treated group in the data
  # if (!(Inf %in% glist)) {
  #   if (args$control_group == "nevertreated") {
  #     stop("There is no available never-treated group")
  #   } else {
  #     # Drop all time periods with time periods >= latest treated
  #     max_treated_time <- max(glist[is.finite(glist)], na.rm = TRUE)
  #     latest_treated_time <- max_treated_time - args$anticipation
  #     data <- data[get(args$tname) < latest_treated_time]
  #
  #     tlist <- sort(unique(data[[args$tname]]))
  #     glist <- sort(unique(data[[args$gname]]))
  #
  #     # don't compute ATT(g,t) for groups that are only treated at end
  #     # and only play a role as a comparison group
  #     glist <- glist[glist < max_treated_time]
  #   }
  # }

  if (!(Inf %in% glist)) {
    # Compute latest treated cohort once, and the cutoff time
    latest_g <- max(glist[is.finite(glist)], na.rm = TRUE)
    cutoff_t  <- latest_g - args$anticipation

    if (args$control_group == "nevertreated") {
      # Warn the user (same wording as the slow path, pre_process_did: the
      # filtering sentence matters -- periods >= cutoff_t are dropped just below)
      warning(
        "No never-treated group is available. ",
        "The last treated cohort is being coerced as 'never-treated' units, and data from periods after that is being filtered out (no available comparison groups)."
      )

      # Drop all periods ≥ (latest_g - anticipation)
      data <- data[get(args$tname) < cutoff_t]

      # For any row where gname == latest_g, set gname := Inf
      data[, (args$gname) := as.numeric(get(args$gname))] # Convert the column to numeric so Inf can be stored
      data[get(args$gname) == latest_g, (args$gname) := Inf]
    } else {
      # 3. If not "nevertreated", we simply drop those periods and leave gnames alone
      data <- data[get(args$tname) < cutoff_t]
    }

    # Recompute tlist and glist from the filtered/modified data
    tlist <- sort(unique(data[[args$tname]]))
    glist <- sort(unique(data[[args$gname]]))

    # 5. If control_group != "nevertreated", drop the max cohort from glist
    if (args$control_group != "nevertreated") {
      glist <- glist[glist < latest_g]
    }
  }


  # get only the first period
  first_period <- tlist[1]

  # drop groups treated in the first period or before; keep only treated groups
  glist <- glist[glist != Inf & glist > first_period + args$anticipation]

  # Check for groups treated in the first period (accounting for anticipation) and drop them
  # identify groups treated in the first period; gname has no NAs here, so keep a
  # local vector instead of a temporary column (it is only read before the subset below)
  gvec <- data[[args$gname]]
  treated_first_period <- is.finite(gvec) & gvec <= first_period + args$anticipation

  # count the number of units treated in the first period
  if (args$panel) {
    nfirstperiod <- length(unique(data[[args$idname]][treated_first_period]))
  } else {
    nfirstperiod <- sum(treated_first_period)
  }

  # handle units treated in the first period
  if (nfirstperiod > 0) {
    warning(paste0("Dropped ", nfirstperiod, " units that were already treated in the first period",
                    if (args$anticipation > 0) paste0(" (accounting for anticipation = ", args$anticipation, ")") else "",
                    "."))
    data <- data[get(args$gname) %in% c(glist, Inf)]

    # update tlist and glist
    tlist <- data[, sort(unique(get(args$tname)))]
    glist <- data[, sort(unique(get(args$gname)))]

    # Drop groups treated in the first period or before
    first_period <- tlist[1]
    glist <- glist[glist != Inf & glist > first_period + args$anticipation]
  }

  # If user specifies repeated cross sections,
  # set that it really is repeated cross sections
  args$true_repeated_cross_sections <- FALSE
  if (!args$panel) {
    args$true_repeated_cross_sections <- TRUE
  }

  #-----------------------------------------------------------------------------
  # setup data in panel case
  #-----------------------------------------------------------------------------
  # Check if data is a balanced panel if panel = TRUE and allow_unbalanced_panel = TRUE
  bal_panel_test <- (args$panel)*(args$allow_unbalanced_panel)

  if (bal_panel_test) {
    # First, focus on complete cases. Rows with missing data were already dropped
    # in did_standardization() above, so this is a guarded no-op in the common case.
    data_comp <- if (anyNA(data)) data[complete.cases(data)] else data

    # Check if the panel is balanced
    is_balanced <- check_balance(data_comp, id_col = args$idname, time_col = args$tname)

    args$allow_unbalanced_panel <- !is_balanced
    message(if (!is_balanced)
      "You have an unbalanced panel. Proceeding as such."
      else
        "You have a balanced panel. Setting allow_unbalanced_panel = FALSE.")
  }

  if (args$panel) {
    # Check for unbalanced panel
    if (args$allow_unbalanced_panel) {
      # Flag for true repeated cross sections
      args$panel <- FALSE
      args$true_repeated_cross_sections <- FALSE
    } else {
      # Coerce balanced panel

      # Focus on complete cases. Rows with missing data were already dropped in
      # did_standardization() above, so this is a guarded no-op in the common case.
      if (anyNA(data)) {
        keepers <- complete.cases(data)
        id_col <- data[[args$idname]]
        n <- length(unique(id_col))
        n_keep <- length(unique(id_col[keepers]))
        if (n_keep < n) {
          warning(paste0("Dropped ", (n - n_keep), " observations that had missing data."))
          data <- data[keepers, ]
        }
      }

      # Make balanced panel. validate_args() guarantees at most one row per (id, t),
      # so the panel is balanced iff nrow == n_ids * n_periods; the grouped count is
      # only needed in the unbalanced case to identify the offending units.
      raw_time_size <- length(unique(data[[args$tname]]))
      if (nrow(data) != length(unique(data[[args$idname]])) * raw_time_size) {
        unit_count <- data[, .(count = .N), by = c(args$idname)]
        mis_unit_ids <- unit_count[[1L]][unit_count[[2L]] < raw_time_size]
        warning(length(mis_unit_ids), " units are missing in some periods. Converting to balanced panel by dropping them.")
        data <- data[!(data[[args$idname]] %in% mis_unit_ids)]
      }

      # If all data is dropped, stop execution
      if (nrow(data) == 0) {
        stop("All observations dropped while converting data to balanced panel. Consider setting `panel = FALSE` and/or revisiting 'idname'.")
      }

      n <- sum(data[[args$tname]] == tlist[1])

      # Note: treatment irreversibility was already checked in validate_args()
    }
  }

  #-----------------------------------------------------------------------------
  # setup data in repeated cross section
  #-----------------------------------------------------------------------------

  if (!args$panel) {
    # Focus on complete cases
    data <- data[complete.cases(data)]

    if (nrow(data) == 0) {
      stop("All observations were dropped due to missing data. Check your outcome, group, time, and covariate variables for missing values.")
    }

    # n-row data.frame to hold the influence function
    if (args$true_repeated_cross_sections) {
      data[, .rowid := .I] # Create row index
      args$idname <- ".rowid"
    } else {
      # Set rowid to idname for repeated cross section/unbalanced
      data[, .rowid := get(args$idname)]
    }

    # Count unique number of cross section observations
    n <- length(unique(data[[args$idname]]))
  }

  # Check if groups is empty (usually a problem with the way people defined groups)
  if(length(glist)==0){
    stop("No valid groups. The variable in 'gname' should be expressed as the time a unit is first treated (0 if never-treated).")
  }

  # if there are only two time periods, then uniform confidence
  # bands are the same as pointwise confidence intervals
  if (length(tlist)==2) {
    args$cband <- FALSE
  }

  #-----------------------------------------------------------------------------
  # more error handling after we have balanced the panel

  # Check against very small groups
  # Calculate group sizes, dividing the count of each group by the length of tlist
  gsize <- data[, .N / length(tlist), by = get(args$gname)]

  # How many in each group before giving a warning
  reqsize <- length(BMisc::rhs_vars(args$xformla)) + 5

  # Filter groups smaller than reqsize
  gsize <- gsize[V1 < reqsize]

  # Warn if some groups are small
  if (nrow(gsize) > 0) {
    small_groups <- gsize[[1L]]
    gpaste <- paste(ifelse(is.infinite(small_groups), 0, small_groups), collapse = ",")
    warning(paste0("Some groups in your dataset have very few observations, which may cause estimation problems.\n  Check groups: ", gpaste, "."))

    # Check if the never treated group is too small
    if ((Inf %in% small_groups) & (args$control_group == "nevertreated")) {
      stop("The never-treated group is too small to serve as a reliable control. Try setting `control_group = 'notyettreated'` to include not-yet-treated units as controls.")
    }
  }

  #-----------------------------------------------------------------------------
  # Sort the data for easy access later on
  if (args$panel){
    setorderv(data, c(args$tname, args$gname, args$idname), c(1,1,1))
  } else {
    setorderv(data, c(args$tname, args$gname), c(1,1))
  }

  # Assign new args regarding number of time periods and groups
  # How many time periods
  args$time_periods_count <- length(tlist)
  # time periods
  args$time_periods <- tlist
  # How many treated groups
  args$treated_groups_count <- length(glist)
  # treated groups
  args$treated_groups <- glist
  # id size
  args$id_count <- n

  return(list(data = data, args = args))
}

#' @title DiD tensors
#' @description A utility function that splits the data into a list of outcome tensors and a list of arguments.
#' Tensors are objects of dimension id_count x 1 x time_periods_count and are used for faster filtering in the computation of the DiD estimator.
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return A list containing outcome tensors, time invariant data, cohort counts, covariates matrix, cluster vector, and weights vector.
#' @noRd
get_did_tensors <- function(data, args){

  # Drop globally-empty factor levels from the covariates referenced by xformla
  # before any design matrix is built: model.matrix() emits a dummy column for
  # every declared level, so a level absent from the ENTIRE estimation sample
  # (common after users subset their data) would produce an all-zero column in
  # every (g,t) cell and NA-fail the whole estimation. Done before invariant_data
  # is derived below so both model.matrix() calls see the same levels. Levels
  # absent only from a particular 2x2 cell are unaffected. Mirrors the identical
  # fix in compute.att_gt() (slow path).
  for (v in all.vars(args$xformla)) {
    if (is.factor(data[[v]])) set(data, j = v, value = droplevels(data[[v]]))
  }

  # Getting the outcome tensor: a vector of outcome variables per time period of dimension id_count x 1 x time_periods_count
  #if(!args$allow_unbalanced_panel){
  if(args$panel){
    # Extract the outcome column once as a plain vector, then split by time period
    y_vec <- data[[args$yname]]
    n <- args$id_count
    nT <- args$time_periods_count
    outcomes_tensor <- vector("list", nT)
    for(time in seq_len(nT)){
      start <- (time - 1L) * n + 1L
      outcomes_tensor[[time]] <- y_vec[start:(start + n - 1L)]
    }
    # Build weights tensor: one weight vector per time period
    w_vec <- data[[".w"]]
    weights_tensor <- vector("list", nT)
    for(time in seq_len(nT)){
      start <- (time - 1L) * n + 1L
      weights_tensor[[time]] <- w_vec[start:(start + n - 1L)]
    }
  } else {
    # for(time in args$time_periods){
    #   outcome_vector_time <- rep(NA, args$id_count)  # Initialize vector with NAs
    #   # Create a new column with Y values if tname == current_time, otherwise NA
    #   data[, outcome_vector_time := fifelse(get(args$tname) == time, get(args$yname), NA_real_)]
    #   # Store the vector in the outcomes_tensor
    #   outcomes_tensor[[time]] <- data$outcome_vector_time
    #   # drop the column created
    #   data[, outcome_vector_time := NULL]
    # }
    outcomes_tensor <- NULL
    weights_tensor <- NULL
  }

  # Getting the time invariant data
  time_invariant_cols <-  c(args$idname, args$gname, ".w", args$clustervars)
  #if(!args$allow_unbalanced_panel){
  if(args$panel){
    # We can do this filtering because the data is already sorted appropriately
    invariant_data <- data[1:args$id_count]
  } else {
    if (!args$allow_unbalanced_panel) {
      # get the first observation for each unit
      invariant_data <- data[data[, .I[1], by = get(args$idname)]$V1]
      # # order by idname
      # setorderv(invariant_data, c(args$idname), c(1))
    } else {
      invariant_data <- copy(data)
    }
  }

  # Get cohort, period and crosstable counts. In the balanced-panel case
  # invariant_data is exactly the first-period slice of the (t, g, id)-sorted data,
  # so the by-period and by-(period, cohort) tables are degenerate one-period shapes
  # derivable from cohort_counts without re-grouping the table; in the RCS/unbalanced
  # case the cohort and period margins are rollups of a single (t, g) pass.
  cohorts <- c(args$treated_groups, Inf)
  if (args$panel) {
    # Get cohort counts
    cohort_counts <- invariant_data[, .(cohort_size = .N) , by = get(args$gname)]
    names(cohort_counts)[1] <- "cohort" # changing the name

    # Get period counts: all invariant_data rows share the first time period
    period_counts <- data.table(period = args$time_periods[1],
                                period_size = nrow(invariant_data))

    # Get crosstable counts (long form): one row per cohort, all in the first period
    crosstable_counts <- data.table(period = args$time_periods[1],
                                    cohort = cohort_counts$cohort,
                                    N = cohort_counts$cohort_size)
  } else {
    # Get crosstable counts (long form): count occurrences by tname and gname
    crosstable_counts <- invariant_data[, .N, by = .(get(args$tname), get(args$gname))]
    names(crosstable_counts)[1] <- "period" # changing the name
    names(crosstable_counts)[2] <- "cohort" # changing the name

    # Get cohort and period counts as rollups of the (t, g) pass
    cohort_counts <- crosstable_counts[, .(cohort_size = sum(N)), by = cohort]
    period_counts <- crosstable_counts[, .(period_size = sum(N)), by = period]
  }
  crosstable_counts <- dcast(crosstable_counts, period ~ cohort, value.var = "N", fill = 0)  # Reshape


  # # Get covariates if any
  # if(args$xformla == ~1){
  #   covariates <- rep(1, args$id_count)
  # } else {
  #   covariates <- as.data.table(model.matrix(args$xformla, data = invariant_data, na.action = na.pass))
  # }

  # if (args$xformla == ~1) {
  #   covariates_tensor <- list(rep(1, args$id_count))
  # } else {
  #   if (args$panel){
  #     covariates_tensor <- vector("list", args$time_periods_count)
  #     for (tt in seq_along(args$time_periods)) {
  #       rng <- ((tt - 1) * args$id_count + 1):(tt * args$id_count)
  #       covariates_tensor[[tt]] <-
  #         model.matrix(args$xformla, data = data[rng], na.action = na.pass)
  #     }
  #   } else {
  #     covariates_tensor <- as.data.table(model.matrix(args$xformla, data = invariant_data, na.action = na.pass))
  #   }
  #
  # }
  if (args$xformla == ~1) {
    ones_vec <- rep(1, args$id_count)
    covariates_tensor <- replicate(args$time_periods_count, ones_vec,
                                   simplify = FALSE)
    covariates_matrix <- matrix(1, nrow = nrow(invariant_data), ncol = 1)
  } else {
    if (args$panel) {
      # Compute model.matrix once on the full stacked data, then split by time period
      full_mm <- model.matrix(args$xformla, data = data, na.action = na.pass)
      n <- args$id_count
      nT <- args$time_periods_count
      covariates_tensor <- vector("list", nT)
      for (tt in seq_len(nT)) {
        start <- (tt - 1L) * n + 1L
        covariates_tensor[[tt]] <- full_mm[start:(start + n - 1L), , drop = FALSE]
      }
      covariates_matrix <- NULL        # not needed in balanced panel
    } else {                             # RCS / unbalanced
      covariates_tensor <- NULL        # slices not needed
      covariates_matrix <- model.matrix(args$xformla,
                                        data = invariant_data,
                                        na.action = na.pass)
    }
  }

  # Get the cluster variable only (direct column access is faster than .SD/.SDcols)
  if(!is.null(args$clustervars)){
    cluster <- invariant_data[[args$clustervars]]
  } else {
    cluster <- NA
  }

  # Get the weights only
  weights <- invariant_data[[".w"]]

  # Gather all the arguments to return
  return(list(outcomes_tensor = outcomes_tensor,
              data = data,
              time_invariant_data = invariant_data,
              cohort_counts = cohort_counts,
              period_counts = period_counts,
              crosstable_counts = crosstable_counts,
              # covariates = covariates,
              covariates_matrix  = covariates_matrix,
              covariates_tensor = covariates_tensor,
              cluster = cluster,
              weights = weights,
              weights_tensor = weights_tensor))
}

#' @title Process `did` Function Arguments
#'
#' @description Function to process arguments passed to the main methods in the
#'  `did` package as well as conducting some tests to make sure
#'  data is in proper format / try to throw helpful error messages.
#'
#' @inheritParams att_gt
#' @param call Function call to att_gt
#'
#' @return a [`DIDparams`] object
#'
#' @export
pre_process_did2 <- function(yname,
                            tname,
                            idname,
                            gname,
                            xformla = NULL,
                            data,
                            panel = TRUE,
                            allow_unbalanced_panel,
                            control_group = c("nevertreated","notyettreated"),
                            anticipation = 0,
                            weightsname = NULL,
                            fix_weights = NULL,
                            alp = 0.05,
                            bstrap = FALSE,
                            cband = FALSE,
                            biters = 1000,
                            clustervars = NULL,
                            est_method = "dr",
                            base_period = "varying",
                            print_details = TRUE,
                            faster_mode=FALSE,
                            pl = FALSE,
                            cores = 1,
                            call = NULL) {


  # coerce data to data.table first, keeping only the columns the pipeline uses
  # (id/time/group/outcome/weights/cluster plus the raw xformla variables) so wide
  # inputs do not pay a full-table copy. If any referenced name is missing, fall
  # back to coercing everything so the downstream checks (dreamerr/validate_args/
  # xformla) fire with their usual messages and precedence.
  keep_cols <- unique(c(idname, tname, gname, yname, weightsname, clustervars,
                        if (!is.null(xformla)) all.vars(xformla)))
  if (is.data.frame(data) && is.character(keep_cols) && all(keep_cols %in% names(data))) {
    if (is.data.table(data)) {
      data <- data[, ..keep_cols]
    } else {
      # selecting columns from a data.frame is shallow; only kept columns are copied
      data <- as.data.table(data[keep_cols])
    }
  } else if (!is.data.table(data)) {
    data <- as.data.table(data)
  }

  # gathering all the arguments except data
  args_names <- setdiff(names(formals()), "data")
  args <- mget(args_names, sys.frame(sys.nframe()))

  # pick a control_group by default
  args$control_group <- control_group[1]
  if (!(args$control_group %in% c("nevertreated", "notyettreated"))) {
    stop("control_group must be either 'nevertreated' or 'notyettreated'")
  }
  args$base_period <- base_period[1]
  if (!(args$base_period %in% c("universal", "varying"))) {
    stop("base_period must be either 'universal' or 'varying'.")
  }
  check_reserved_did_names(yname = args$yname, tname = args$tname,
                           idname = args$idname, gname = args$gname,
                           xformla = args$xformla,
                           weightsname = args$weightsname,
                           clustervars = args$clustervars)

  # run error checking on arguments
  validate_args(args, data)

  # Strip idname from clustervars in the caller's copy of args.
  # validate_args works on a local copy, so we must repeat this here.
  # Clustering at the unit level (idname) is already done implicitly.
  if (!is.null(args$clustervars) && !is.null(args$idname) && (args$idname %in% args$clustervars)) {
    args$clustervars <- setdiff(args$clustervars, args$idname)
    if (length(args$clustervars) == 0L) args$clustervars <- NULL
  }

  # put in blank xformla if no covariates
  if (is.null(args$xformla)) {
    args$xformla <- ~1
  } else {
    # extract variable names from the formula
    formula_vars <- all.vars(args$xformla)

    # identify variables in xformla not in data
    missing_vars <- setdiff(formula_vars, names(data))

    # error checking for missing variables in data
    if (length(missing_vars) > 0) {
      stop(paste("The following variables are not in data:", paste(missing_vars, collapse = ", ")), call. = FALSE)
    }
  }

  # Put the data in a standard format after some validation
  cleaned_did <- did_standardization(data, args)

  # Partition staggered DiD into a 2x2 DiD for faster implementation
  did_tensors <- get_did_tensors(cleaned_did$data, cleaned_did$args)

  # store parameters for passing around later
  dp <- DIDparams2(did_tensors, cleaned_did$args, call=call)

  return(dp)
}

#' @title MP.TEST
#'
#' @description An object that holds results from computing pre-test of the
#'  conditional parallel trends assumption
#'
#' @param CvM Cramer von Mises test statistic
#' @param CvMb a vector of bootstrapped Cramer von Mises test statistics
#' @param CvMcval CvM critical value
#' @param CvMpval p-value for CvM test
#' @param KS Kolmogorov-Smirnov test statistic
#' @param KSb a vector of bootstrapped KS test statistics
#' @param KScval KS critical value
#' @param KSpval p-value for KS test
#' @param clustervars vector of which variables were clustered on for the test
#' @param xformla formla for the X variables used in the test
#'
#' @export
MP.TEST <- function(CvM=NULL, CvMb=NULL, CvMcval=NULL, CvMpval=NULL, KS=NULL, KSb=NULL, KScval=NULL, KSpval=NULL, clustervars=NULL, xformla=NULL) {
    out <- list(CvM=CvM, CvMb=CvMb, CvMcval=CvMcval, CvMpval=CvMpval, KS=KS, KSb=KSb, KScval=KScval, KSpval=KSpval, clustervars=clustervars, xformla=xformla)
    class(out) <- "MP.TEST"
    out
}

#' @title summary.MP.TEST
#'
#' @description print a summary of test results
#'
#' @param object an MP.TEST object
#' @param ... other variables
#'
#' @export
summary.MP.TEST <- function(object, ... ) {
    CvM <- object$CvM
    CvMcval <- object$CvMcval
    CvMpval <- object$CvMpval
    citation()
    cat("Cramer von Mises: \n")
    cat("  Test Statistic: ", CvM, "\n")
    cat("  Critical Value: ", CvMcval, "\n")
    cat("  P-value       : ", CvMpval, "\n \n")
    cat("Clustering on   : ", paste0(object$clustervars,sep=","), "\n")
    cat("X formula       : ", as.character(object$xformla), "\n")
}

#' @title reset.sim
#' @description a function to create a "reasonable" set of parameters
#'  to create simulated panel data that obeys a parallel trends assumption.
#'  In particular, it provides parameters where the the effect of participating
#'  in the treatment is equal to one in all post-treatment time periods.
#' 
#'  After calling this function, the user can change particular values of the
#'  parameters in order to generate dynamics, heterogeneous effects across
#'  groups, etc.
#'
#' @param time.periods The number of time periods to include
#' @param n The total number of observations
#' @param ipw If TRUE, sets parameters so that DGP is
#'  compatible with recovering ATT(g,t)'s using IPW (i.e.,
#'  where logit that just includes a linear term in X works).  If
#'  FALSE, sets parameters that will be incompatible with IPW.
#'  Either way, these parameters can be specified by the user
#'  if so desired.
#' @param reg If TRUE, sets parameters so that DGP is compatible
#'  with recovering ATT(g,t)'s using regressions on untreated
#'  untreated potential outcomes.  If FALSE, sets parameters that
#'  will be incompatible with using regressions (i.e., regressions
#'  that include only linear term in X).  Either way, these
#'  parameters can be specified by the user if so desired.
#'
#' @return list of simulation parameters
#' 
#' @export
reset.sim <- function(time.periods=4, n=5000, ipw=TRUE, reg=TRUE) {
  validate_positive_whole_number(time.periods, "time.periods")
  validate_positive_whole_number(n, "n")
  validate_logical_scalar(ipw, "ipw")
  validate_logical_scalar(reg, "reg")

  #-----------------------------------------------------------------------------
  # set parameters
  #-----------------------------------------------------------------------------
  # coefficient on X 
  bett <- seq(1:time.periods)
  # time fixed effect
  thet <- seq(1:time.periods)
  # time fixed effect
  theu <- thet # changing this creates violations of parallel trends
  # covariate effect
  betu <- bett # changing this creates violations of conditional parallel trends
  #-----------------------------------------------------------------------------
  # parameters for treated potential outcomes
  #-----------------------------------------------------------------------------
  te.bet.ind <- rep(1,time.periods) # no selective treatment timing
  te.bet.X <- bett #no heterogeneous effects by X
  te.t <- thet # no calendar time effects
  te.e <- rep(0,time.periods) # no dynamic effects
  te <- 1 # overall basic effect
  # parameters in generalized propensity score
  # don't make them too big otherwise can get divide by 0
  gamG <- c(0,1:time.periods)/(2*time.periods)

  # return list of parameters
  list(time.periods=time.periods,
       bett=bett,
       thet=thet,
       theu=theu,
       betu=betu,
       te.bet.ind=te.bet.ind,
       te.bet.X=te.bet.X,
       te.t=te.t,
       te.e=te.e,
       te=te,
       n=n,
       gamG=gamG,
       ipw=ipw,
       reg=reg
       )
}

#' @title build_sim_dataset
#'
#' @description A function for building simulated data
#'
#' @param sp_list A list of simulation parameters.  See `reset.sim` to generate
#'  some default values for parameters
#' @param panel whether to construct panel data (the default) or repeated
#'  cross sections data
#'
#' @return a data.frame with the following columns
#'   \itemize{
#'     \item G observations group
#'     \item X value of covariate
#'     \item id observation's id
#'     \item cluster observation's cluster (by construction there is no within-cluster correlation)
#'     \item period time period for current observation
#'     \item Y outcome
#'     \item treat whether or not this unit is ever treated
#'   }
#'
#' @export
build_sim_dataset <- function(sp_list, panel=TRUE) {
  validate_logical_scalar(panel, "panel")
  if (!is.list(sp_list)) {
    stop("sp_list must be a list of simulation parameters.")
  }

  #-----------------------------------------------------------------------------
  # build dataset
  #-----------------------------------------------------------------------------
  time.periods <- sp_list$time.periods
  nt <- sp_list$nt
  bett <- sp_list$bett
  thet=sp_list$thet
  nu <- sp_list$nu
  theu <- sp_list$theu
  betu <- sp_list$betu
  te.bet.ind <- sp_list$te.bet.ind
  te.bet.X <- sp_list$te.bet.X
  te.t <- sp_list$te.t
  te.e <- sp_list$te.e
  te <- sp_list$te
  n <- sp_list$n
  gamG <- sp_list$gamG
  ipw <- sp_list$ipw
  reg <- sp_list$reg
  validate_positive_whole_number(time.periods, "sp_list$time.periods")
  validate_positive_whole_number(n, "sp_list$n")
  validate_logical_scalar(ipw, "sp_list$ipw")
  validate_logical_scalar(reg, "sp_list$reg")
  validate_finite_numeric_vector(bett, "sp_list$bett", time.periods)
  validate_finite_numeric_vector(thet, "sp_list$thet", time.periods)
  validate_finite_numeric_vector(theu, "sp_list$theu", time.periods)
  validate_finite_numeric_vector(betu, "sp_list$betu", time.periods)
  validate_finite_numeric_vector(te.bet.ind, "sp_list$te.bet.ind", time.periods)
  validate_finite_numeric_vector(te.bet.X, "sp_list$te.bet.X", time.periods)
  validate_finite_numeric_vector(te.t, "sp_list$te.t", time.periods)
  validate_finite_numeric_vector(te.e, "sp_list$te.e", time.periods)
  validate_finite_numeric_vector(gamG, "sp_list$gamG", time.periods + 1L)
  validate_finite_numeric_scalar(te, "sp_list$te")

  X <- rnorm(n)

  # Build the exp(outer(...)) propensity term once and reuse it for the row-sum
  # normalization (it was previously computed twice). apply(., 1, sum) is kept
  # (rather than rowSums) so `pr` is bit-identical to before -- this matters because
  # the sample() draw for G below depends on the exact values of `pr`.
  if (ipw) {
    e <- exp(outer(X, gamG))
  } else {
    e <- exp(outer((pnorm(X) + 0.5)^2, gamG))
  }
  pr <- e / apply(e, 1, sum)

  G <- apply(pr, 1, function(pvec) sample(seq(0,time.periods), size=1, prob=pvec))

  Gt <- G[G>0]
  nt <- length(Gt)

  if (reg) {
    Xmodel <- X
  } else {
    Xmodel <- X^2
  }

  Xt <- Xmodel[G>0]

  # draw individual fixed effect
  Ct <- rnorm(nt, mean=G)

  # generate untreated potential outcomes in each time period
  Ynames <- paste0("Y",1:time.periods)
  #Ynames <- paste0(1:time.periods)
  Y0tmat <- sapply(1:time.periods, function(t) {
    thet[t] + Ct + Xt*bett[t] + rnorm(nt)
  })
  Y0tdf <- as.data.frame(Y0tmat)
  
  # generate treated potential outcomes
  # te.e[pmax(t-Gt+1, 1)] is the event-time treatment effect: event time
  # t - Gt is stored 1-based, so te.e[1] is the on-impact (t = Gt) effect.
  # pmax() clamps the index at 1 so it stays valid in pre-treatment periods
  # (t < Gt), where the (Gt <= t) factor zeroes the term anyway.
  Y1tdf <- sapply(1:time.periods, function(t) {
    te.t[t] + te.bet.ind[Gt]*Ct + Xt*te.bet.X[t] + (Gt <= t)*te.e[pmax(t-Gt+1, 1)] + te + rnorm(nt)
  })

  # generate observed data
  Ytdf <- sapply(1:time.periods, function(t) {
    (Gt<=t)*Y1tdf[,t] + (Gt>t)*Y0tdf[,t]
  })
  colnames(Ytdf) <- Ynames

  # store observed data for treated group
  dft <- cbind.data.frame(G=Gt,X=X[G>0],Ytdf)

  # untreated units

  # draw untreated covariate
  nu <- sum(G==0)
  Xu <- Xmodel[G==0]

  # draw untreated fixed effect
  Cu <- rnorm(nu, mean=0)


  # generate untreated potential outcomes
  Y0umat <- sapply(1:time.periods, function(t) {
    theu[t] + Cu + rnorm(nu) + Xu*betu[t]
  })
  Y0udf <- as.data.frame(Y0umat)
  colnames(Y0udf) <- Ynames

  # store dataset of observed outcomes for untreated units
  dfu <- cbind.data.frame(G=0,X=X[G==0],Y0udf)

  # store overall dataset
  df <- rbind.data.frame(dft, dfu)

  # generate id variable
  df$id <- 1:nrow(df)
  # generate clusters (there's no actual within-cluster correlation)
  df$cluster <- sample(1:50, size=nrow(df), replace=TRUE)

  # convert data from wide to long format
  ddf <- tidyr::pivot_longer(df,
                             cols=tidyr::starts_with("Y"),
                             names_to="period",
                             names_prefix="Y",
                             values_to="Y")
  ddf$period <- as.numeric(ddf$period)
  ddf$treat <- 1*(ddf$G > 0)
  ddf <- ddf[order(ddf$id, ddf$period),] # reorder data

  if (!panel) { # repeated cross sections
    n <- nt+nu
    Time <- sample(1:time.periods, size=n, replace=TRUE, prob=rep(1/time.periods, time.periods))
    right.row <- sapply( unique(ddf$id), function(i) {
      which(ddf$id==i & ddf$period==Time[i])
    })
    ddf <- ddf[right.row,]
  }

  ddf <- subset(ddf, G != 1)
  ddf
}


#' @title sim
#' @description An internal function that builds simulated data, computes
#'  ATT(g,t)'s and some aggregations.  It is useful for testing the inference
#'  procedures in the `did` function.
#'
#' @inheritParams reset.sim
#' @inheritParams build_sim_dataset
#'
#' @param ret which type of results to return.  The options are `Wpval` (returns
#'  1 if the p-value from a Wald test that all pre-treatment ATT(g,t)'s are equal
#'  is less than .05),
#'  `cband`  (returns 1 if a uniform confidence band covers 0 for groups and times),
#'  `simple` (returns 1 if, using the simple treatment effect aggregation results
#'  in rejecting that this aggregated treatment effect parameter is equal to 0),
#'  `dynamic` (returns 1 if the uniform confidence band from the dynamic treatment
#'  effect aggregation covers 0 in all pre- and post-treatment periods).  The default
#'  value is NULL, and in this case the function will just return the results from
#'  the call to `att_gt`.
#' @param bstrap whether or not to use the bootstrap to conduct inference (default is TRUE)
#' @param cband whether or not to compute uniform confidence bands in the call to `att_gt`
#'  (the default is TRUE)
#' @param control_group Whether to use the "nevertreated" comparison group (the default)
#'  or the "notyettreated" as the comparison group
#' @param xformla Formula for covariates in `att_gt` (default is `~X`)
#' @param est_method Which estimation method to use in `att_gt` (default is "dr")
#' @param clustervars Any additional variables which should be clustered on
#' @param panel whether to simulate panel data (the default) or otherwise repeated
#'  cross sections data
#' 
#' @return When `ret=NULL`, returns the results of the call to `att_gt`, otherwise returns
#'  1 if the specified test rejects or 0 if not.
#' 
#' @export
sim <- function(sp_list,
                ret=NULL,
                bstrap=TRUE,
                cband=TRUE,
                control_group="nevertreated",
                xformla=~X,
                est_method="dr",
                clustervars=NULL,
                panel=TRUE) {
  validate_logical_scalar(bstrap, "bstrap")
  validate_logical_scalar(cband, "cband")
  validate_logical_scalar(panel, "panel")
  validate_optional_choice_scalar(
    ret,
    "ret",
    c("Wpval", "cband", "simple", "dynamic", "notyettreated"),
    "ret must be NULL or one of 'Wpval', 'cband', 'simple', 'dynamic', or 'notyettreated'."
  )

  ddf <- build_sim_dataset(sp_list=sp_list,
                           panel=panel)

  time.periods <- sp_list$time.periods
  te.e <- sp_list$te.e
  te <- sp_list$te
  
  # get results
  res <- att_gt(yname="Y", xformla=xformla, data=ddf, tname="period", idname="id",
                gname="G", 
                bstrap=bstrap, cband=cband, control_group=control_group,
                est_method=est_method,
                clustervars=clustervars,
                panel=panel)


  if (is.null(ret)) {
    return(res)
  } else if (ret=="Wpval") {
    rej <- 1*(res$Wpval < .05)
    return(rej)
  } else if (ret=="cband") {
    cu <- res$att + res$c*res$se 
    cl <- res$att - res$c*res$se
    covers0 <- 1*(all( (cu > 0) & (cl < 0)))
    return(covers0)
  } else if (ret=="simple") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else if (ret=="dynamic") {
    truth <- c(rep(0,(time.periods-2)),te+te.e[1:(time.periods-1)])
    agg <- aggte(res, type="dynamic")
    cu <- agg$att.egt + agg$crit.val.egt * agg$se.egt
    cl <- agg$att.egt - agg$crit.val.egt * agg$se.egt
    coverstruth <- 1*(all( (cu > truth) & (cl < truth)))
    return(coverstruth)
  } else if (ret=="notyettreated") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else {
    return(res)
  }

}

## pretest_sim <- function(ret=NULL, bstrap=FALSE, cband=FALSE,
##                         control.group="nevertreated", panel=TRUE, xformla=~X, cores=1) {

##   ddf <- build_ipw_dataset(panel=panel)

##   # get results
##   res <- conditional_did_pretest(yname="Y", xformla=xformla, data=ddf,
##                                  tname="period", idname="id",
##                                  first.treat.name="G", estMethod="ipw",
##                                  printdetails=FALSE,
##                                  bstrap=bstrap, cband=cband,
##                                  control.group=control.group,
##                                  panel=panel, 
##                                  pl=TRUE, cores=cores)

##   res$CvMpval
## }

#' @title gplot
#'
#' @description does the heavy lifting for making a plot of an group-time
#'  average treatment effect
#'
#' @inheritParams ggdid
#'
#' @return a `ggplot2` object
#'
#' @keywords internal
#'
#' @export
gplot <- function(ssresults, ylim=NULL, xlab=NULL, ylab=NULL, title="Group", xgap=1,
                  legend=TRUE, ref_line = 0, theming = TRUE) {
  validate_positive_numeric_scalar(xgap, "xgap")
  validate_logical_scalar(legend, "legend")
  validate_logical_scalar(theming, "theming")
  validate_optional_numeric_scalar(ref_line, "ref_line")

  unique_years <- sort(unique(as.numeric(as.character(ssresults$year))))
  xgap_int <- max(1L, as.integer(round(xgap)))
  dabreaks <- unique_years[seq(1, length(unique_years), by = xgap_int)]

  c.point <-  qnorm(1 - ssresults$alp/2)

  p <- ggplot(ssresults,
              aes(x=as.numeric(as.character(year)), y=att, ymin=(att-c*att.se),
                  ymax=(att+c*att.se))) +

    geom_point(aes(colour=post), size=1.5) +
    #geom_ribbon(aes(x=as.numeric(year)), alpha=0.2) +
    geom_errorbar(aes(colour=post), width=0.1) +
    scale_y_continuous(limits=ylim) +
    #scale_x_discrete(breaks=dabreaks, labels=as.character(dabreaks)) +
    scale_x_continuous(breaks=as.numeric(dabreaks), labels=as.character(dabreaks)) +
    scale_color_manual(drop=FALSE, values=c("#e87d72","#56bcc2"), breaks = c(0, 1), labels = c('Pre','Post')) +
    labs(x = xlab, y = ylab, title = title, color = NULL)

  if (!is.null(ref_line)) {
    p <- p + geom_hline(aes(yintercept = ref_line), linetype = 'dashed')
  }
  if (theming) {
    p <- p + theme_classic() +
      theme(plot.title = element_text(color="darkgray", face="bold", size=12),
            axis.title = element_text(color="darkgray", face="bold", size=12),
            strip.background = element_rect(fill = 'white', color = 'white'),
            strip.text = element_text(color = 'darkgray', face = 'bold', size = 12, hjust = 0),
            legend.position = 'bottom')
  }
  if (!legend) {
    p  <- p + theme(legend.position = "none")
  }

  p
}


#' @title splot
#'
#' @description alternative plotting function for the case with selective treatment
#'  timing
#'
#' @inheritParams ggdid
#' @inheritParams gplot
#'
#' @return a `ggplot2` object
#'
#' @keywords internal
#'
#' @export
splot <- function(ssresults, ylim=NULL, xlab=NULL, ylab=NULL, title="Group",
                  legend=TRUE, ref_line = 0, theming = TRUE) {
  validate_logical_scalar(legend, "legend")
  validate_logical_scalar(theming, "theming")
  validate_optional_numeric_scalar(ref_line, "ref_line")

  # names of variables are "weird" for this function because this code builds
  # on the same infrastructure as for plotting group-time average treatment
  # effects and aggregations using event time or calendar time

  # the group variable is saved in year

  if (is.null(xlab)) {
    xlab <- 'ATT'
  }
  if (is.null(ylab)) {
    ylab <- 'Group'
  }

  errorbar_layer <- if (utils::packageVersion("ggplot2") >= "3.3.0") {
    geom_errorbar(aes(colour=post), width=0.1, orientation="y")
  } else {
    geom_errorbarh(aes(colour=post), height=0.1)
  }

  p <- ggplot(ssresults,
              aes(y=as.factor(year), x=att, xmin=(att-c*att.se),
                  xmax=(att+c*att.se))) +
    geom_point(aes(colour=post), size=1.5) +
    #geom_ribbon(aes(x=as.numeric(year)), alpha=0.2) +
    errorbar_layer +
    scale_y_discrete(breaks=as.factor(ssresults$year)) +
    #scale_x_discrete(breaks=dabreaks, labels=as.character(dabreaks)) +
    scale_x_continuous(limits=ylim) +
    scale_color_manual(drop=FALSE, values=c("#e87d72","#56bcc2"), breaks = c(0, 1), labels = c('Pre','Post')) +
    labs(x = xlab, y = ylab, title = title)

  if (!is.null(ref_line)) {
    p <- p + geom_vline(aes(xintercept = ref_line), linetype = 'dashed')
  }
  if (theming) {
    p <- p + theme_classic() +
      theme(plot.title = element_text(color="darkgray", face="bold", size=12),
            axis.title = element_text(color="darkgray", face="bold", size=12),
            legend.position = 'none')
  }

  if (!legend) {
    p  <- p + theme(legend.position = "none")
  }

  p
}

#' @title Plot `did` objects using `ggplot2`
#'
#' @description Function to plot objects from the `did` package
#'
#' @param object either a `MP` object or `AGGTEobj` object. See `help(ggdid.MP)` and `help(ggdid.AGGTEobj)`.
#' @param ... other arguments
#'
#' @examples
#' \dontrun{
#' data(mpdta)
#' out <- att_gt(yname = "lemp",
#'               gname = "first.treat",
#'               idname = "countyreal",
#'               tname = "year",
#'               xformla = ~1,
#'               data = mpdta)
#'
#' # plot all group-time average treatment effects
#' ggdid(out)
#'
#' # plot event study aggregation
#' es <- aggte(out, type = "dynamic")
#' ggdid(es)
#' }
#'
#' @export
ggdid <- function(object, ...) {
  UseMethod("ggdid", object)
}


## #' @param type the type of plot, should be one of "attgt", "dynamic",
## #'  "group", "calendar", "dynsel".  "attgt" is the default and plots
## #'  all group-time average treatment effects separately by group (including
## #'  pre-treatment time periods); "dynamic" plots dynamic treatment effects --
## #'  these are the same as event studies; "group" plots average effects
## #'  of the treatment separately by group (which allows for selective treatment
## #'  timing); "calendar" plots average treatment effects by time period; and
## #'  "dynsel" plots dynamic effects allowing for selective treatment timing
## #'  (this also requires setting the additional paramater e1)


#' @title Plot `MP` objects using `ggplot2`
#'
#' @description A function to plot `MP` objects
#'
#' @inheritParams ggdid
#' @param object an `MP` object, as returned by `att_gt`.
#' @param ylim optional y limits for the plot; setting here makes the y limits
#'  the same across different plots
#' @param xlab optional x-axis label
#' @param ylab optional y-axis label
#' @param title optional plot title
#' @param xgap optional gap between the labels on the x-axis.  For example,
#'  `xgap=3` indicates that the labels should show up for every third
#'  value on the x-axis.  The default is 1.
#' @param ncol The number of columns to include in the resulting plot.  The
#'  default is 1.
#' @param legend Whether or not to include a legend (which will indicate color
#'  of pre- and post-treatment estimates).  Default is `TRUE`.
#' @param group Vector for which groups to include in the plots of ATT(g,t).
#'   Default is NULL, and, in this case, plots for all groups will be included (`ggdid.MP` only).
#' @param ref_line A reference line at this value, usually to compare confidence
#'   intervals to 0. Set to NULL to omit.
#' @param theming Set to FALSE to skip all theming so you can do it yourself.
#' @param grtitle Title to append before each group name (`ggdid.MP` only).
#'
#' @export
ggdid.MP <- function(object,
                     ylim=NULL,
                     xlab=NULL,
                     ylab=NULL,
                     title="Group",
                     xgap=1,
                     ncol=1,
                     legend=TRUE,
                     group=NULL,
                     ref_line = 0,
                     theming = TRUE,
                     grtitle = "Group",
                     ...) {
  validate_positive_whole_number(ncol, "ncol")

  mpobj <- object

  G <- length(unique(mpobj$group))
  Y <- length(unique(mpobj$t))## drop 1 period bc DID
  g <- unique(mpobj$group)[order(unique(mpobj$group))] ## -1 to drop control group
  y <- unique(mpobj$t)

  results <- data.frame(year=rep(y,G))
  results$group <- unlist(lapply(g, function(x) { rep(x, Y) }))
  results$grtitle <- paste(grtitle,results$group)
  results$att <- mpobj$att
  n <- mpobj$n
  results$att.se <- mpobj$se #sqrt(diag(mpobj$V)/n)
  results$post <- as.factor(1*(results$year >= results$group))
  results$year <- as.factor(results$year)
  results$c <- mpobj$c
  #vcovatt <- mpobj$V/n
  alp <- mpobj$alp


  if(is.null(group)) group <- g
  if(any(!(group %in% g))){
    warning('Some of the specified groups do not exist in the data. Reporting all available groups.')
    group <- g
  }

  mplots <- gplot(results[results$group %in% group,],
                  ylim, xlab, ylab, title, xgap,
                  legend, ref_line, theming) +
    facet_wrap(~grtitle, ncol = ncol, scales = 'free')

  return(mplots)
}


#' @title Plot `AGGTEobj` objects
#'
#' @description A function to plot `AGGTEobj` objects
#'
#' @inheritParams ggdid.MP
#' @param object an `AGGTEobj` object, as returned by `aggte`.
#'
#' @export
ggdid.AGGTEobj <- function(object,
                           ylim=NULL,
                           xlab=NULL,
                           ylab=NULL,
                           title="",
                           xgap=1,
                           legend=TRUE,
                           ref_line = 0,
                           theming = TRUE,
                           ...) {

  if ( !(object$type %in% c("dynamic","group","calendar")) ) {
    stop(paste0("Plot method not available for this type of aggregation"))
  }

  post.treat <- 1*(object$egt >= 0)
  results <- cbind.data.frame(year=object$egt,
                              att=object$att.egt,
                              att.se=object$se.egt,
                              post=as.factor(post.treat))
  alp <- object$DIDparams$alp
  if (is.null(alp)) alp <- 0.05
  results$c <- if (is.null(object$crit.val.egt)) qnorm(1 - alp/2) else object$crit.val.egt

  if (title == "") {
    # get title right depending on which aggregation
    title <- if (object$type=="group") {
      "Average Effect by Group"
    } else if (object$type=="dynamic") {
      "Average Effect by Length of Exposure"
    } else {
      "Average Effect by Time Period"
    }
  }

  if (object$type == "group") {
    # alternative plot if selective/group treatment timing plot
    p <- splot(results, ylim, xlab, ylab, title, legend, ref_line, theming)
  } else {
    p <- gplot(results, ylim, xlab, ylab, title, xgap, legend, ref_line, theming)
  }

  p
}

#' @title validate_args
#' @description A utility function to validate arguments passed to a att_gt()
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return nothing, but throws an error if any of the arguments are not valid
#' @noRd
validate_args <- function(args, data){

  data_names <- names(data)

  # ---------------------- Error Checking ----------------------
  # Flag for control group types
  control_group_message <- "control_group must be either 'nevertreated' or 'notyettreated'"
  validate_choice_scalar(args$control_group, "control_group",
                         c("nevertreated", "notyettreated"),
                         control_group_message)

  # Flag for tname, gname, yname
  name_message <- "__ARG__ must be a character scalar and a name of a column from the dataset."
  dreamerr::check_set_arg(args$tname, args$gname, args$yname, "match", .choices = data_names, .message = name_message, .up = 1)

  # Strip idname from clustervars before validation: users may naturally pass
  # clustervars = c(idname, extra_var), and idname clustering is implicit/redundant.
  # This must happen before the dreamerr check which enforces scalar length.
  if (!is.null(args$clustervars) && !is.null(args$idname) && (args$idname %in% args$clustervars)) {
    args$clustervars <- setdiff(args$clustervars, args$idname)
    if (length(args$clustervars) == 0L) args$clustervars <- NULL
  }

  # At most one cluster variable beyond idname is supported (clustering at the
  # unit level via idname is implicit). Checked before the dreamerr scalar check
  # below so the error wording matches the slow path (pre_process_did) and
  # mboot(), instead of a dreamerr message exposing internal argument names.
  if (length(args$clustervars) > 1) {
    stop("At most one cluster variable (beyond 'idname') is supported. Please reduce to one.")
  }

  # Flag for clustervars and weightsname
  checkvar_message <- "__ARG__ must be NULL or a character scalar that is a name of a column from the dataset."
  dreamerr::check_set_arg(args$weightsname, args$clustervars, "NULL | match", .choices = data_names, .message = checkvar_message, .up = 1)

  # check if time periods are numeric
  if(!data[, is.numeric(get(args$tname))]){stop("The time variable '", args$tname, "' must be numeric. Please convert it.")}

  # Check if gname is numeric
  if(!data[, is.numeric(get(args$gname))]){stop("The group variable '", args$gname, "' must be numeric. Please convert it.")}

  # gname must be 0 (never-treated) or a positive treatment-timing value;
  # negative codes are not supported (0 is reserved for never-treated). Reject
  # them up front so the fast and slow paths behave identically (the fast path
  # previously accepted negative codes silently while the slow path errored).
  if (data[, any(get(args$gname) < 0, na.rm = TRUE)]) {
    stop("The group variable '", args$gname, "' must be 0 (never-treated) or a ",
         "positive treatment-timing value; negative values are not supported. ",
         "If your time periods are non-positive, shift them so the earliest ",
         "period is >= 1.")
  }

  # Check if yname is numeric (logical 0/1 outcomes are also allowed)
  if(!data[, is.numeric(get(args$yname)) || is.logical(get(args$yname))]){stop("The outcome variable '", args$yname, "' must be numeric. Please convert it.")}

  # Flag for idname
  if(!is.null(args$idname)){
    # Check if idname is in the data
    name_message <- "__ARG__ must be a character scalar and a name of a column from the dataset."
    dreamerr::check_set_arg(args$idname, "match", .choices = data_names, .message = name_message, .up = 1)

    #  check if idname is numeric
    if(!data[, is.numeric(get(args$idname))]){stop("The id variable '", args$idname, "' must be numeric. Please convert it.")}

    # Check if gname is unique by idname: irreversibility of the treatment
    # Use direct column access instead of get() for speed
    nonmissing_g <- !is.na(data[[args$idname]]) & !is.na(data[[args$gname]])
    id_g_unique <- unique(data[nonmissing_g, c(args$idname, args$gname), with = FALSE])
    check_treatment_uniqueness <- anyDuplicated(id_g_unique[[1]]) == 0L
    if (!check_treatment_uniqueness) {
      stop("The value of gname (treatment variable) must be the same across all periods for each particular unit. The treatment must be irreversible.")
    }

    # Check if any combination of idname and tname is duplicated
    nonmissing_id_time <- !is.na(data[[args$idname]]) & !is.na(data[[args$tname]])
    n_id_year <- anyDuplicated(data[nonmissing_id_time], by = c(args$idname, args$tname))
    # If any combination is duplicated, stop execution and throw an error
    if (n_id_year > 0) {
      stop("The value of idname must be unique (by tname). Some units are observed more than once in a period.")
    }
  }

  # Flag for base period: not in c("universal", "varying"), stop
  base_period_message <- "base_period must be either 'universal' or 'varying'."
  validate_choice_scalar(args$base_period, "base_period",
                         c("universal", "varying"),
                         base_period_message)

  # Flags for cluster variable
  # Note: idname was already stripped from clustervars and the at-most-one check
  # was enforced above (before the dreamerr check)
  if (!is.null(args$clustervars)) {
    # Check that cluster variables do not vary over time within each unit. For true repeated cross-sections
    # the user may omit idname (an internal ".rowid" is created later, one observation per row), so there is
    # no within-unit time variation to check and idname is not yet available here -- skip the check then.
    if (length(args$clustervars) > 0 && !is.null(args$idname)) {
      # Efficiently check for time-varying cluster variables
      clust_tv <- data[, lapply(.SD, function(col) length(unique(col)) == 1), by = get(args$idname), .SDcols = args$clustervars]
      # If any cluster variable varies over time within any unit, stop execution
      # (same wording as the slow path and mboot())
      if (!all(unlist(clust_tv[, -1, with = FALSE]))) {
        stop("Time-varying cluster variables are not supported. Please provide a time-invariant cluster variable.")
      }
    }
  }

  validate_logical_scalar(args$panel, "panel")
  validate_logical_scalar(args$allow_unbalanced_panel, "allow_unbalanced_panel")
  validate_logical_scalar(args$bstrap, "bstrap")
  validate_logical_scalar(args$cband, "cband")
  validate_logical_scalar(args$print_details, "print_details")
  validate_logical_scalar(args$pl, "pl")
  validate_positive_whole_number(args$cores, "cores")
  validate_anticipation(args$anticipation)
  validate_alp(args$alp)
  if (args$bstrap) validate_positive_whole_number(args$biters, "biters")

}

#' @title DiD standardization
#' @description A utility function to coerce the data into standard format
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return A list containing ordered data and arguments
#' @noRd
did_standardization <- function(data, args){
  # keep relevant columns plus the RAW covariate variables (all.vars of xformla), so
  # model.matrix(args$xformla, .) can be rebuilt downstream. Keeping the evaluated
  # model.frame instead would lose the raw variable for transform formulae
  # (~I(X^2), ~poly(X, 2), ~log(X)) and create matrix-valued columns that break the
  # data.table coercion. For bare-variable / factor formulae this is identical.
  xvars <- all.vars(args$xformla)
  cols_to_keep <- unique(c(args$idname, args$tname, args$gname, args$yname,
                           args$weightsname, args$clustervars, xvars))
  # Subset the dataset to keep only the relevant columns
  data <- data[, ..cols_to_keep]

  # Check if any covariates were missing
  n_orig <- data[, .N]
  # gname is excluded from the finite check because Inf is a valid never-treated
  # code there (see complete_finite_cases); missing/NaN gname is still dropped.
  data <- data[complete_finite_cases(data, finite_exclude = args$gname)]
  # also drop rows whose EVALUATED design is missing/non-finite (e.g. log of a non-positive
  # covariate); safe now that raw-covariate NAs are removed (so poly()/ns()/... will
  # not error on NA input). Use model.frame (NOT model.matrix) with na.action =
  # na.pass: model.frame keeps every row -- including NA/NaN-valued terms -- so the
  # complete.cases() mask stays aligned with `data` (model.matrix would silently drop
  # NaN rows, shortening the mask and letting the offending rows survive).
  if (length(xvars) > 0L && data[, .N] > 0L) {
    mf_check <- suppressWarnings(model.frame(args$xformla, data = data, na.action = na.pass))
    finite_rows <- complete_finite_cases(mf_check)
    if (!all(finite_rows)) data <- data[finite_rows]
  }
  n_new <- data[, .N]
  n_diff <- n_orig - n_new
  if (n_diff != 0) {
    warning(paste0("dropped ", n_diff, " rows from original data due to missing or non-finite data"))
  }

  # Set weights
  if (is.null(args$weightsname)) weights <- rep(1, n_new) else weights <- data[[args$weightsname]]
  # Validate user-supplied weights: negative weights flip signs and a non-positive
  # mean divides by ~0 during normalization, silently producing NA/NaN ATTs.
  if (!is.null(args$weightsname) && (any(weights < 0, na.rm = TRUE) || isTRUE(mean(weights, na.rm = TRUE) <= 0)))
    stop("The weights variable '", args$weightsname, "' must be non-negative with a positive mean.")
  # enforcing normalization of weights. At this point we already drop any missing values in weights.
  weights <- weights/mean(weights)
  if (".w" %in% names(data)) stop("Your data already contains a column named '.w', which is reserved for internal use by `did`. Please rename this column before calling att_gt().")
  data.table::set(data, j = ".w", value = weights)

  # Check for time-varying weights in panel data
  if (!is.null(args$weightsname) && args$panel) {
    w_range <- data[, .(w_range = max(.w) - min(.w)), by = c(args$idname)]
    if (any(w_range$w_range > .Machine$double.eps^0.5, na.rm = TRUE)) {
      message(
        "Time-varying weights detected. For balanced panel data, the default ",
        "behavior uses the weight from the earlier of the two time periods in ",
        "each 2x2 comparison (the base period for post-treatment cells). ",
        "Use the 'fix_weights' argument to control this behavior. ",
        "See ?att_gt for details."
      )
    }
  }

  # get a list of dates from min to max
  tlist <- data[, sort(unique(get(args$tname)))]

  # Ensure gname column is numeric (double) so it can hold Inf
  if (is.integer(data[[args$gname]])) {
    data[, (args$gname) := as.numeric(get(args$gname))]
  }

  # Coerce control group identified with zero as Inf
  data[get(args$gname) == 0, (args$gname) := Inf]

  # Working out the dates
  # Identify groups with treatment time bigger than the maximum treatment time + anticipation.
  # We account for anticipation because units treated shortly after the last observed period
  # may already exhibit anticipatory effects during the observed time window, and should
  # not be used as controls.
  max_treatment_time <- max(tlist, na.rm = TRUE)
  # gname has no NAs here (complete cases were enforced above), so the comparison
  # cannot yield NA; keep it as a local vector instead of a temporary column
  asif_never_treated <- data[[args$gname]] > max_treatment_time + args$anticipation
  # set gname to Inf for those groups considered as never treated
  if (any(asif_never_treated)) {
    data.table::set(data, i = which(asif_never_treated), j = args$gname, value = Inf)
  }

  # get list of treated groups (by time) from min to max
  glist <- sort(unique(data[[args$gname]]))

  # Check if there is a never treated group in the data
  # if (!(Inf %in% glist)) {
  #   if (args$control_group == "nevertreated") {
  #     stop("There is no available never-treated group")
  #   } else {
  #     # Drop all time periods with time periods >= latest treated
  #     max_treated_time <- max(glist[is.finite(glist)], na.rm = TRUE)
  #     latest_treated_time <- max_treated_time - args$anticipation
  #     data <- data[get(args$tname) < latest_treated_time]
  #
  #     tlist <- sort(unique(data[[args$tname]]))
  #     glist <- sort(unique(data[[args$gname]]))
  #
  #     # don't compute ATT(g,t) for groups that are only treated at end
  #     # and only play a role as a comparison group
  #     glist <- glist[glist < max_treated_time]
  #   }
  # }

  if (!(Inf %in% glist)) {
    # Compute latest treated cohort once, and the cutoff time
    latest_g <- max(glist[is.finite(glist)], na.rm = TRUE)
    cutoff_t  <- latest_g - args$anticipation

    if (args$control_group == "nevertreated") {
      # Warn the user (same wording as the slow path, pre_process_did: the
      # filtering sentence matters -- periods >= cutoff_t are dropped just below)
      warning(
        "No never-treated group is available. ",
        "The last treated cohort is being coerced as 'never-treated' units, and data from periods after that is being filtered out (no available comparison groups)."
      )

      # Drop all periods ≥ (latest_g - anticipation)
      data <- data[get(args$tname) < cutoff_t]

      # For any row where gname == latest_g, set gname := Inf
      data[, (args$gname) := as.numeric(get(args$gname))] # Convert the column to numeric so Inf can be stored
      data[get(args$gname) == latest_g, (args$gname) := Inf]
    } else {
      # 3. If not "nevertreated", we simply drop those periods and leave gnames alone
      data <- data[get(args$tname) < cutoff_t]
    }

    # Recompute tlist and glist from the filtered/modified data
    tlist <- sort(unique(data[[args$tname]]))
    glist <- sort(unique(data[[args$gname]]))

    # 5. If control_group != "nevertreated", drop the max cohort from glist
    if (args$control_group != "nevertreated") {
      glist <- glist[glist < latest_g]
    }
  }


  # get only the first period
  first_period <- tlist[1]

  # drop groups treated in the first period or before; keep only treated groups
  glist <- glist[glist != Inf & glist > first_period + args$anticipation]

  # Check for groups treated in the first period (accounting for anticipation) and drop them
  # identify groups treated in the first period; gname has no NAs here, so keep a
  # local vector instead of a temporary column (it is only read before the subset below)
  gvec <- data[[args$gname]]
  treated_first_period <- is.finite(gvec) & gvec <= first_period + args$anticipation

  # count the number of units treated in the first period
  if (args$panel) {
    nfirstperiod <- length(unique(data[[args$idname]][treated_first_period]))
  } else {
    nfirstperiod <- sum(treated_first_period)
  }

  # handle units treated in the first period
  if (nfirstperiod > 0) {
    warning(paste0("Dropped ", nfirstperiod, " units that were already treated in the first period",
                    if (args$anticipation > 0) paste0(" (accounting for anticipation = ", args$anticipation, ")") else "",
                    "."))
    # Drop ONLY the first-period-treated units, by row identity (see pre_process_did:
    # dropping by `gname %in% c(glist, Inf)` also deleted the latest control cohort --
    # removed from glist above -- whenever there was no never-treated group, silently
    # corrupting ATT(g,t) for the other groups).
    data <- data[!treated_first_period]

    # update tlist and glist
    tlist <- data[, sort(unique(get(args$tname)))]
    glist <- data[, sort(unique(get(args$gname)))]

    # Drop groups treated in the first period or before
    first_period <- tlist[1]
    glist <- glist[glist != Inf & glist > first_period + args$anticipation]

    # Keep the latest cohort in the data as a not-yet-treated control but excluded
    # from glist when there is still no never-treated group (mirrors the
    # `glist[glist < latest_g]` trim above for the nfirstperiod == 0 case).
    if (args$control_group != "nevertreated" && !any(is.infinite(data[[args$gname]]))) {
      glist <- glist[glist < latest_g]
    }
  }

  # If user specifies repeated cross sections,
  # set that it really is repeated cross sections
  args$true_repeated_cross_sections <- FALSE
  if (!args$panel) {
    args$true_repeated_cross_sections <- TRUE
  }

  #-----------------------------------------------------------------------------
  # setup data in panel case
  #-----------------------------------------------------------------------------
  # Check if data is a balanced panel if panel = TRUE and allow_unbalanced_panel = TRUE
  bal_panel_test <- (args$panel)*(args$allow_unbalanced_panel)

  if (bal_panel_test) {
    # First, focus on complete cases. Rows with missing data were already dropped
    # in did_standardization() above, so this is a guarded no-op in the common case.
    data_comp <- if (anyNA(data)) data[complete.cases(data)] else data

    # Check if the panel is balanced
    is_balanced <- check_balance(data_comp, id_col = args$idname, time_col = args$tname)

    args$allow_unbalanced_panel <- !is_balanced
    message(if (!is_balanced)
      "You have an unbalanced panel. Proceeding as such."
      else
        "You have a balanced panel. Setting allow_unbalanced_panel = FALSE.")
  }

  if (args$panel) {
    # Check for unbalanced panel
    if (args$allow_unbalanced_panel) {
      # Flag for true repeated cross sections
      args$panel <- FALSE
      args$true_repeated_cross_sections <- FALSE
    } else {
      # Coerce balanced panel

      # Focus on complete cases. Rows with missing data were already dropped in
      # did_standardization() above, so this is a guarded no-op in the common case.
      if (anyNA(data)) {
        keepers <- complete.cases(data)
        id_col <- data[[args$idname]]
        n <- length(unique(id_col))
        n_keep <- length(unique(id_col[keepers]))
        if (n_keep < n) {
          warning(paste0("Dropped ", (n - n_keep), " observations that had missing data."))
          data <- data[keepers, ]
        }
      }

      # Make balanced panel. validate_args() guarantees at most one row per (id, t),
      # so the panel is balanced iff nrow == n_ids * n_periods; the grouped count is
      # only needed in the unbalanced case to identify the offending units.
      raw_time_size <- length(unique(data[[args$tname]]))
      if (nrow(data) != length(unique(data[[args$idname]])) * raw_time_size) {
        unit_count <- data[, .(count = .N), by = c(args$idname)]
        mis_unit_ids <- unit_count[[1L]][unit_count[[2L]] < raw_time_size]
        warning(length(mis_unit_ids), " units are missing in some periods. Converting to balanced panel by dropping them.")
        data <- data[!(data[[args$idname]] %in% mis_unit_ids)]
      }

      # If all data is dropped, stop execution
      if (nrow(data) == 0) {
        stop("All observations dropped while converting data to balanced panel. Consider setting `panel = FALSE` and/or revisiting 'idname'.")
      }

      n <- sum(data[[args$tname]] == tlist[1])

      # Note: treatment irreversibility was already checked in validate_args()
    }
  }

  #-----------------------------------------------------------------------------
  # setup data in repeated cross section
  #-----------------------------------------------------------------------------

  if (!args$panel) {
    # Focus on complete cases
    data <- data[complete.cases(data)]

    if (nrow(data) == 0) {
      stop("All observations were dropped due to missing data. Check your outcome, group, time, and covariate variables for missing values.")
    }

    # n-row data.frame to hold the influence function
    if (args$true_repeated_cross_sections) {
      data[, .rowid := .I] # Create row index
      args$idname <- ".rowid"
    } else {
      # Set rowid to idname for repeated cross section/unbalanced
      data[, .rowid := get(args$idname)]
    }

    # Count unique number of cross section observations
    n <- length(unique(data[[args$idname]]))
  }

  # Check if groups is empty (usually a problem with the way people defined groups)
  if(length(glist)==0){
    stop("No valid groups. The variable in 'gname' should be expressed as the time a unit is first treated (0 if never-treated).")
  }

  # if there are only two time periods, then uniform confidence
  # bands are the same as pointwise confidence intervals
  if (length(tlist)==2) {
    args$cband <- FALSE
  }

  #-----------------------------------------------------------------------------
  # more error handling after we have balanced the panel

  # Check against very small groups
  # Calculate group sizes, dividing the count of each group by the length of tlist
  gsize <- data[, .N / length(tlist), by = get(args$gname)]

  # How many in each group before giving a warning
  reqsize <- length(BMisc::rhs_vars(args$xformla)) + 5

  # Filter groups smaller than reqsize
  gsize <- gsize[V1 < reqsize]

  # Warn if some groups are small
  if (nrow(gsize) > 0) {
    small_groups <- gsize[[1L]]
    gpaste <- paste(ifelse(is.infinite(small_groups), 0, small_groups), collapse = ",")
    warning(paste0("Some groups in your dataset have very few observations, which may cause estimation problems.\n  Check groups: ", gpaste, "."))

    # Check if the never treated group is too small
    if ((Inf %in% small_groups) & (args$control_group == "nevertreated")) {
      stop("The never-treated group is too small to serve as a reliable control. Try setting `control_group = 'notyettreated'` to include not-yet-treated units as controls.")
    }
  }

  #-----------------------------------------------------------------------------
  # Sort the data for easy access later on
  if (args$panel){
    setorderv(data, c(args$tname, args$gname, args$idname), c(1,1,1))
  } else {
    setorderv(data, c(args$tname, args$gname), c(1,1))
  }

  # Assign new args regarding number of time periods and groups
  # How many time periods
  args$time_periods_count <- length(tlist)
  # time periods
  args$time_periods <- tlist
  # How many treated groups
  args$treated_groups_count <- length(glist)
  # treated groups
  args$treated_groups <- glist
  # id size
  args$id_count <- n

  return(list(data = data, args = args))
}

#' @title DiD tensors
#' @description A utility function that splits the data into a list of outcome tensors and a list of arguments.
#' Tensors are objects of dimension id_count x 1 x time_periods_count and are used for faster filtering in the computation of the DiD estimator.
#' @param data data.table used in function
#' @param args list of arguments to validate
#'
#' @return A list containing outcome tensors, time invariant data, cohort counts, covariates matrix, cluster vector, and weights vector.
#' @noRd
get_did_tensors <- function(data, args){

  # Drop globally-empty factor levels from the covariates referenced by xformla
  # before any design matrix is built: model.matrix() emits a dummy column for
  # every declared level, so a level absent from the ENTIRE estimation sample
  # (common after users subset their data) would produce an all-zero column in
  # every (g,t) cell and NA-fail the whole estimation. Done before invariant_data
  # is derived below so both model.matrix() calls see the same levels. Levels
  # absent only from a particular 2x2 cell are unaffected. Mirrors the identical
  # fix in compute.att_gt() (slow path).
  for (v in all.vars(args$xformla)) {
    if (is.factor(data[[v]])) set(data, j = v, value = droplevels(data[[v]]))
  }

  # Getting the outcome tensor: a vector of outcome variables per time period of dimension id_count x 1 x time_periods_count
  #if(!args$allow_unbalanced_panel){
  if(args$panel){
    # Extract the outcome column once as a plain vector, then split by time period
    y_vec <- data[[args$yname]]
    n <- args$id_count
    nT <- args$time_periods_count
    outcomes_tensor <- vector("list", nT)
    for(time in seq_len(nT)){
      start <- (time - 1L) * n + 1L
      outcomes_tensor[[time]] <- y_vec[start:(start + n - 1L)]
    }
    # Build weights tensor: one weight vector per time period
    w_vec <- data[[".w"]]
    weights_tensor <- vector("list", nT)
    for(time in seq_len(nT)){
      start <- (time - 1L) * n + 1L
      weights_tensor[[time]] <- w_vec[start:(start + n - 1L)]
    }
  } else {
    # for(time in args$time_periods){
    #   outcome_vector_time <- rep(NA, args$id_count)  # Initialize vector with NAs
    #   # Create a new column with Y values if tname == current_time, otherwise NA
    #   data[, outcome_vector_time := fifelse(get(args$tname) == time, get(args$yname), NA_real_)]
    #   # Store the vector in the outcomes_tensor
    #   outcomes_tensor[[time]] <- data$outcome_vector_time
    #   # drop the column created
    #   data[, outcome_vector_time := NULL]
    # }
    outcomes_tensor <- NULL
    weights_tensor <- NULL
  }

  # Getting the time invariant data
  time_invariant_cols <-  c(args$idname, args$gname, ".w", args$clustervars)
  #if(!args$allow_unbalanced_panel){
  if(args$panel){
    # We can do this filtering because the data is already sorted appropriately
    invariant_data <- data[1:args$id_count]
  } else {
    if (!args$allow_unbalanced_panel) {
      # get the first observation for each unit
      invariant_data <- data[data[, .I[1], by = get(args$idname)]$V1]
      # # order by idname
      # setorderv(invariant_data, c(args$idname), c(1))
    } else {
      invariant_data <- copy(data)
    }
  }

  # Get cohort, period and crosstable counts. In the balanced-panel case
  # invariant_data is exactly the first-period slice of the (t, g, id)-sorted data,
  # so the by-period and by-(period, cohort) tables are degenerate one-period shapes
  # derivable from cohort_counts without re-grouping the table; in the RCS/unbalanced
  # case the cohort and period margins are rollups of a single (t, g) pass.
  cohorts <- c(args$treated_groups, Inf)
  if (args$panel) {
    # Get cohort counts
    cohort_counts <- invariant_data[, .(cohort_size = .N) , by = get(args$gname)]
    names(cohort_counts)[1] <- "cohort" # changing the name

    # Get period counts: all invariant_data rows share the first time period
    period_counts <- data.table(period = args$time_periods[1],
                                period_size = nrow(invariant_data))

    # Get crosstable counts (long form): one row per cohort, all in the first period
    crosstable_counts <- data.table(period = args$time_periods[1],
                                    cohort = cohort_counts$cohort,
                                    N = cohort_counts$cohort_size)
  } else {
    # Get crosstable counts (long form): count occurrences by tname and gname
    crosstable_counts <- invariant_data[, .N, by = .(get(args$tname), get(args$gname))]
    names(crosstable_counts)[1] <- "period" # changing the name
    names(crosstable_counts)[2] <- "cohort" # changing the name

    # Get cohort and period counts as rollups of the (t, g) pass
    cohort_counts <- crosstable_counts[, .(cohort_size = sum(N)), by = cohort]
    period_counts <- crosstable_counts[, .(period_size = sum(N)), by = period]
  }
  crosstable_counts <- dcast(crosstable_counts, period ~ cohort, value.var = "N", fill = 0)  # Reshape


  # # Get covariates if any
  # if(args$xformla == ~1){
  #   covariates <- rep(1, args$id_count)
  # } else {
  #   covariates <- as.data.table(model.matrix(args$xformla, data = invariant_data, na.action = na.pass))
  # }

  # if (args$xformla == ~1) {
  #   covariates_tensor <- list(rep(1, args$id_count))
  # } else {
  #   if (args$panel){
  #     covariates_tensor <- vector("list", args$time_periods_count)
  #     for (tt in seq_along(args$time_periods)) {
  #       rng <- ((tt - 1) * args$id_count + 1):(tt * args$id_count)
  #       covariates_tensor[[tt]] <-
  #         model.matrix(args$xformla, data = data[rng], na.action = na.pass)
  #     }
  #   } else {
  #     covariates_tensor <- as.data.table(model.matrix(args$xformla, data = invariant_data, na.action = na.pass))
  #   }
  #
  # }
  if (args$xformla == ~1) {
    ones_vec <- rep(1, args$id_count)
    covariates_tensor <- replicate(args$time_periods_count, ones_vec,
                                   simplify = FALSE)
    covariates_matrix <- matrix(1, nrow = nrow(invariant_data), ncol = 1)
  } else {
    if (args$panel) {
      # Compute model.matrix once on the full stacked data, then split by time period
      full_mm <- model.matrix(args$xformla, data = data, na.action = na.pass)
      n <- args$id_count
      nT <- args$time_periods_count
      covariates_tensor <- vector("list", nT)
      for (tt in seq_len(nT)) {
        start <- (tt - 1L) * n + 1L
        covariates_tensor[[tt]] <- full_mm[start:(start + n - 1L), , drop = FALSE]
      }
      covariates_matrix <- NULL        # not needed in balanced panel
    } else {                             # RCS / unbalanced
      covariates_tensor <- NULL        # slices not needed
      covariates_matrix <- model.matrix(args$xformla,
                                        data = invariant_data,
                                        na.action = na.pass)
    }
  }

  # Get the cluster variable only (direct column access is faster than .SD/.SDcols)
  if(!is.null(args$clustervars)){
    cluster <- invariant_data[[args$clustervars]]
  } else {
    cluster <- NA
  }

  # Get the weights only
  weights <- invariant_data[[".w"]]

  # Gather all the arguments to return
  return(list(outcomes_tensor = outcomes_tensor,
              data = data,
              time_invariant_data = invariant_data,
              cohort_counts = cohort_counts,
              period_counts = period_counts,
              crosstable_counts = crosstable_counts,
              # covariates = covariates,
              covariates_matrix  = covariates_matrix,
              covariates_tensor = covariates_tensor,
              cluster = cluster,
              weights = weights,
              weights_tensor = weights_tensor))
}

#' @title Process `did` Function Arguments
#'
#' @description Function to process arguments passed to the main methods in the
#'  `did` package as well as conducting some tests to make sure
#'  data is in proper format / try to throw helpful error messages.
#'
#' @inheritParams att_gt
#' @param call Function call to att_gt
#'
#' @return a [`DIDparams`] object
#'
#' @export
pre_process_did2 <- function(yname,
                            tname,
                            idname,
                            gname,
                            xformla = NULL,
                            data,
                            panel = TRUE,
                            allow_unbalanced_panel,
                            control_group = c("nevertreated","notyettreated"),
                            anticipation = 0,
                            weightsname = NULL,
                            fix_weights = NULL,
                            alp = 0.05,
                            bstrap = FALSE,
                            cband = FALSE,
                            biters = 1000,
                            clustervars = NULL,
                            est_method = "dr",
                            base_period = "varying",
                            print_details = TRUE,
                            faster_mode=FALSE,
                            pl = FALSE,
                            cores = 1,
                            call = NULL) {

  validate_xformla(xformla)

  # coerce data to data.table first, keeping only the columns the pipeline uses
  # (id/time/group/outcome/weights/cluster plus the raw xformla variables) so wide
  # inputs do not pay a full-table copy. If any referenced name is missing, fall
  # back to coercing everything so the downstream checks (dreamerr/validate_args/
  # xformla) fire with their usual messages and precedence.
  keep_cols <- unique(c(idname, tname, gname, yname, weightsname, clustervars,
                        if (!is.null(xformla)) all.vars(xformla)))
  if (is.data.frame(data) && is.character(keep_cols) && all(keep_cols %in% names(data))) {
    if (is.data.table(data)) {
      data <- data[, ..keep_cols]
    } else {
      # selecting columns from a data.frame is shallow; only kept columns are copied
      data <- as.data.table(data[keep_cols])
    }
  } else if (!is.data.table(data)) {
    data <- as.data.table(data)
  }

  # gathering all the arguments except data
  args_names <- setdiff(names(formals()), "data")
  args <- mget(args_names, sys.frame(sys.nframe()))

  # pick a control_group by default
  if (missing(control_group)) args$control_group <- "nevertreated"
  validate_choice_scalar(
    args$control_group,
    "control_group",
    c("nevertreated", "notyettreated"),
    "control_group must be either 'nevertreated' or 'notyettreated'"
  )
  validate_choice_scalar(
    args$base_period,
    "base_period",
    c("universal", "varying"),
    "base_period must be either 'universal' or 'varying'."
  )
  validate_logical_scalar(args$faster_mode, "faster_mode")
  check_reserved_did_names(yname = args$yname, tname = args$tname,
                           idname = args$idname, gname = args$gname,
                           xformla = args$xformla,
                           weightsname = args$weightsname,
                           clustervars = args$clustervars)

  # run error checking on arguments
  validate_args(args, data)

  # Strip idname from clustervars in the caller's copy of args.
  # validate_args works on a local copy, so we must repeat this here.
  # Clustering at the unit level (idname) is already done implicitly.
  if (!is.null(args$clustervars) && !is.null(args$idname) && (args$idname %in% args$clustervars)) {
    args$clustervars <- setdiff(args$clustervars, args$idname)
    if (length(args$clustervars) == 0L) args$clustervars <- NULL
  }

  # put in blank xformla if no covariates
  if (is.null(args$xformla)) {
    args$xformla <- ~1
  } else {
    # extract variable names from the formula
    formula_vars <- all.vars(args$xformla)

    # identify variables in xformla not in data
    missing_vars <- setdiff(formula_vars, names(data))

    # error checking for missing variables in data
    if (length(missing_vars) > 0) {
      stop(paste("The following variables are not in data:", paste(missing_vars, collapse = ", ")), call. = FALSE)
    }
  }

  # Put the data in a standard format after some validation
  cleaned_did <- did_standardization(data, args)

  # Partition staggered DiD into a 2x2 DiD for faster implementation
  did_tensors <- get_did_tensors(cleaned_did$data, cleaned_did$args)

  # store parameters for passing around later
  dp <- DIDparams2(did_tensors, cleaned_did$args, call=call)

  return(dp)
}

reset.sim <- function() {
  #-----------------------------------------------------------------------------
  # set parameters
  #-----------------------------------------------------------------------------
  # number of time periods
  # time.periods <<- 4
  # number of treated units
  nt <<- 4000
  # coefficient on X 
  bett <<- seq(1:time.periods)
  # time fixed effect
  thet <<- seq(1:time.periods)
  # number of untreated units
  nu <<- 4000
  # time fixed effect
  theu <<- thet # changing this creates violations of parallel trends
  # covariate effect
  betu <<- bett # changing this creates violations of conditional parallel trends
  #-----------------------------------------------------------------------------
  # parameters for treated potential outcomes
  #-----------------------------------------------------------------------------
  te.bet.ind <<- rep(1,time.periods) # no selective treatment timing
  te.bet.X <<- bett #no heterogeneous effects by X
  te.t <<- thet # no calendar time effects
  te.e <<- rep(0,time.periods) # no dynamic effects
  te <<- 1 # overall basic effect
  #-----------------------------------------------------------------------------
  # extra parameters for an ipw sim (otherwise, these are not used)
  #-----------------------------------------------------------------------------
  n <<- 4000
  # these are parameters in generalized propensity score
  # don't make them too big otherwise can get divide by 0
  gamG <<- c(0,1:time.periods)/(2*time.periods)
}


build_sim_dataset <- function(panel=TRUE) {
  #-----------------------------------------------------------------------------
  # build dataset
  #-----------------------------------------------------------------------------
  
  # randomly assign treated units to groups
  # random groups
  G <- sample(1:time.periods, size=nt, replace=TRUE)
  # fixed groups
  # G <- unlist(lapply(1:T, function(g) rep(g, nt/T)))
  
  # draw a single covariate
  Xt <- rnorm(nt)

  # draw individual fixed effect
  Ct <- rnorm(nt, mean=G)

  # generate untreated potential outcomes in each time period
  Ynames <- paste0("Y",1:time.periods)
  Ynames <- paste0(1:time.periods)
  Y0tmat <- sapply(1:time.periods, function(t) {
    thet[t] + Ct + Xt*bett[t] + rnorm(nt)
  })
  Y0tdf <- as.data.frame(Y0tmat)
  
  # generate treated potential outcomes
  Y1tdf <- sapply(1:time.periods, function(t) {
    te.t[t] + te.bet.ind[G]*Ct + Xt*te.bet.X[t] + (G <= t)*te.e[sapply(1:nt, function(i) max(t-G[i]+1,1))] + te + rnorm(nt) # hack for the dynamic effects but ok
  })

  # generate observed data
  Ytdf <- sapply(1:time.periods, function(t) {
    (G<=t)*Y1tdf[,t] + (G>t)*Y0tdf[,t]
  })
  colnames(Ytdf) <- Ynames

  # store observed data for treated group
  dft <- cbind.data.frame(G,X=Xt,Ytdf)

  # untreated units

  # draw untreated covariate
  Xu <- rnorm(nu, mean=1)

  # draw untreated fixed effect
  Cu <- rnorm(nu)


  # generate untreated potential outcomes
  Y0umat <- sapply(1:time.periods, function(t) {
    theu[t] + Cu + rnorm(nu) + Xu*betu[t]
  })
  Y0udf <- as.data.frame(Y0umat)
  colnames(Y0udf) <- Ynames

  # store dataset of observed outcomes for untreated units
  dfu <- cbind.data.frame(G=0,X=Xu,Y0udf)

  # store overall dataset
  df <- rbind.data.frame(dft, dfu)

  # generate id variable
  df$id <- 1:nrow(df)

  # convert data from wide to long format
  library(tidyr)
  ddf <- gather(df, period, Y, -G, -X, -id)
  ddf$period <- as.numeric(ddf$period)
  ddf$treat <- 1*(ddf$G > 0)
  ddf <- ddf[order(ddf$id, ddf$period),] # reorder data

  if (!panel) { # repeated cross sections
    n <- nt+nu
    Time <- sample(1:time.periods, size=n, replace=TRUE, prob=rep(1/time.periods, time.periods))
    right.row <- sapply( unique(ddf$id), function(i) {
      which(ddf$id==i & ddf$period==Time[i])
    })
    ddf <- ddf[right.row,]
  }

  ddf <- subset(ddf, G != 1)
  ddf
}



build_ipw_dataset <- function(panel=TRUE) {

  #-----------------------------------------------------------------------------
  # build dataset
  # build things up from the generalized propensity score
  #-----------------------------------------------------------------------------
  X <- rnorm(n)

  pr <- exp(outer(X,gamG)) / apply( exp(outer(X,gamG)), 1, sum)
  GG <- apply(pr, 1, function(pvec) sample(seq(0,time.periods), size=1, prob=pvec))

  G <- GG[GG>0]
  nt <- length(G)
  Xt <- X[GG>0]

  # draw individual fixed effect
  Ct <- rnorm(nt, mean=G)

  # generate untreated potential outcomes in each time period
  Ynames <- paste0("Y",1:time.periods)
  Ynames <- paste0(1:time.periods)
  Y0tmat <- sapply(1:time.periods, function(t) {
    thet[t] + Ct + Xt*bett[t] + rnorm(nt)
  })
  Y0tdf <- as.data.frame(Y0tmat)
  
  # generate treated potential outcomes
  Y1tdf <- sapply(1:time.periods, function(t) {
    te.t[t] + te.bet.ind[G]*Ct + Xt*te.bet.X[t] + (G <= t)*te.e[sapply(1:nt, function(i) max(t-G[i]+1,1))] + te + rnorm(nt) # hack for the dynamic effects but ok
  })

  # generate observed data
  Ytdf <- sapply(1:time.periods, function(t) {
    (G<=t)*Y1tdf[,t] + (G>t)*Y0tdf[,t]
  })
  colnames(Ytdf) <- Ynames

  # store observed data for treated group
  dft <- cbind.data.frame(G,X=Xt,Ytdf)

  # untreated units
  nu <- n-nt
  
  # draw untreated covariate
  Xu <- X[GG==0]

  # draw untreated fixed effect
  Cu <- rnorm(nu)


  # generate untreated potential outcomes
  Y0umat <- sapply(1:time.periods, function(t) {
    theu[t] + Cu + rnorm(nu) + Xu*betu[t]
  })
  Y0udf <- as.data.frame(Y0umat)
  colnames(Y0udf) <- Ynames

  # store dataset of observed outcomes for untreated units
  dfu <- cbind.data.frame(G=0,X=Xu,Y0udf)

  # store overall dataset
  df <- rbind.data.frame(dft, dfu)

  # generate id variable
  df$id <- 1:nrow(df)

  # convert data from wide to long format
  library(tidyr)
  ddf <- gather(df, period, Y, -G, -X, -id)
  ddf$period <- as.numeric(ddf$period)
  ddf$treat <- 1*(ddf$G > 0)
  ddf <- ddf[order(ddf$id, ddf$period),] # reorder data

  if (!panel) { # repeated cross sections
    Time <- sample(1:time.periods, size=n, replace=TRUE, prob=rep(1/time.periods, time.periods))
    right.row <- sapply( unique(ddf$id), function(i) {
      which(ddf$id==i & ddf$period==Time[i])
    })
    ddf <- ddf[right.row,]
  }
  
  ddf <- subset(ddf, G != 1)
  
  ddf
}


sim <- function(ret=NULL, bstrap=FALSE, cband=FALSE, control.group="nevertreated",
                panel=TRUE) {

  ddf <- build_sim_dataset(panel)
  
  # get results
  res <- att_gt(yname="Y", xformla=~X, data=ddf, tname="period", idname="id",
                gname="G", est_method="reg", 
                bstrap=bstrap, cband=cband, control_group=control.group,
                panel=panel)


  if (is.null(ret)) {
    return(res)
  } else if (ret=="Wpval") {
    rej <- 1*(res$Wpval < .05)
    return(rej)
  } else if (ret=="cband") {
    cu <- res$att + res$c * sqrt(diag(res$V))/sqrt(res$n)
    cl <- res$att - res$c * sqrt(diag(res$V))/sqrt(res$n)
    covers0 <- 1*(all( (cu > 0) & (cl < 0)))
    return(covers0)
  } else if (ret=="simple") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else if (ret=="dynamic") {
    truth <- c(rep(0,(time.periods-2)),te+te.e[1:(time.periods-1)])
    agg <- aggte(res, type="dynamic")
    cu <- agg$att.egt + agg$crit.val.egt * agg$se.egt
    cl <- agg$att.egt - agg$crit.val.egt * agg$se.egt
    coverstruth <- 1*(all( (cu > truth) & (cl < truth)))
    return(coverstruth)
  } else if (ret=="notyettreated") {
    agg <- aggte(res)
    rej <- 1*( abs(agg$overall.att / agg$overall.se) > qnorm(.975) )
    return(rej)
  } else {
    return(res)
  }

}

pretest_sim <- function(ret=NULL, bstrap=FALSE, cband=FALSE,
                        control.group="nevertreated", panel=TRUE, xformla=~X, cores=1) {

  ddf <- build_ipw_dataset(panel=panel)

  # get results
  res <- conditional_did_pretest(yname="Y", xformla=xformla, data=ddf,
                                 tname="period", idname="id",
                                 first.treat.name="G", estMethod="ipw",
                                 printdetails=FALSE,
                                 bstrap=bstrap, cband=cband,
                                 control.group=control.group,
                                 panel=panel, 
                                 pl=TRUE, cores=cores)

  res$CvMpval
}

#' @title Group-Time Average Treatment Effects
#'
#' @description `att_gt` computes average treatment effects in DID
#'  setups where there are more than two periods of data and allowing for
#'  treatment to occur at different points in time and allowing for
#'  treatment effect heterogeneity and dynamics.
#'  See Callaway and Sant'Anna (2021) for a detailed description.
#'
#' @param yname The name of the outcome variable
#' @param data The name of the data.frame that contains the data
#' @param tname The name of the column containing the time periods
#' @param idname The individual (cross-sectional unit) id name
#' @param gname The name of the variable in `data` that
#'  contains the first period when a particular observation is treated.
#'  This should be a positive number for all observations in treated groups.
#'  It defines which "group" a unit belongs to.  It should be 0 for units
#'  in the untreated group.
#' @param weightsname The name of the column containing the sampling weights.
#'  If not set, all observations have same weight. When weights are
#'  time-invariant (constant within each unit across periods), all
#'  \code{fix_weights} options produce identical results and no special
#'  handling is needed.
#'
#'  When weights vary across time (e.g., time-varying population sizes),
#'  the default behavior differs by panel type:
#'  \describe{
#'    \item{Balanced panel}{Each 2x2 DiD comparison uses the weight from the
#'      earlier of the two time periods involved. For post-treatment cells,
#'      this is the base period (g-1). For pre-treatment cells with
#'      \code{base_period="varying"}, this is the pre-treatment period itself.
#'      The panel DRDID estimators are used.}
#'    \item{Repeated cross sections and unbalanced panels}{Both periods'
#'      per-observation weights are passed directly to the RC DRDID estimators,
#'      so each observation carries its own period-specific weight.}
#'  }
#'  Use the \code{fix_weights} argument to override the default behavior.
#' @param fix_weights Controls how time-varying sampling weights are resolved.
#'  Only relevant when weights vary across time; with time-invariant weights,
#'  all options produce identical results. Options:
#'  \describe{
#'    \item{\code{NULL} (default)}{For balanced panel: uses the weight from
#'      the earlier of the two time periods in each 2x2 comparison. For
#'      post-treatment cells, this is the base period (g-1). For
#'      pre-treatment cells, this depends on the \code{base_period} setting.
#'      For RC/unbalanced panel: uses per-observation weights from both
#'      periods.}
#'    \item{\code{"varying"}}{Uses per-observation, period-specific weights
#'      for all panel types. For balanced panel data, this switches to the
#'      repeated cross-section DRDID estimators so that pre-period and
#'      post-period observations each carry their own weight. Covariates
#'      are held fixed at their pre-period values (same as the default
#'      panel estimator). This is the most flexible option for weights but
#'      sacrifices the efficiency of the panel estimator. For RC/unbalanced
#'      panel, this is identical to the default. Not supported with custom
#'      \code{est_method} functions.}
#'    \item{\code{"base_period"}}{Fixes weights at the base period (g-1) for
#'      all (g,t) cells within a group, for both pre-treatment and
#'      post-treatment comparisons. Ensures all cells within a group use the
#'      same weights. For unbalanced panels, units not observed in the base
#'      period are dropped with a warning. Not supported for repeated cross
#'      sections (\code{panel = FALSE}).}
#'    \item{\code{"first_period"}}{Fixes weights at the first time period in
#'      the dataset for all (g,t) cells. For unbalanced panels, units not
#'      observed in the first period are dropped with a warning. Not supported
#'      for repeated cross sections (\code{panel = FALSE}).}
#'  }
#' @param alp the significance level, default is 0.05
#' @param bstrap Boolean for whether or not to compute standard errors using
#'  the multiplier bootstrap.  Default is `TRUE` (in addition, cband
#'  is also by default `TRUE` indicating that uniform confidence bands
#'  will be returned).  If `bstrap=FALSE`, analytical standard errors are
#'  reported; these are cluster-robust when `clustervars` is supplied.
#' @param biters The number of bootstrap iterations to use.  The default is 1000,
#'  and this is only applicable if `bstrap=TRUE`.
#' @param clustervars A vector of variables names to cluster on.  At most, there
#'  can be two variables (otherwise will throw an error) and one of these
#'  must be the same as idname which allows for clustering at the individual
#'  level. Clustered standard errors are available with the multiplier bootstrap
#'  (`bstrap=TRUE`) or analytically (`bstrap=FALSE`).
#' @param cband Boolean for whether or not to compute a uniform confidence
#'  band that covers all of the group-time average treatment effects
#'  with fixed probability `1-alp`.  In order to compute uniform confidence
#'  bands, `bstrap` must also be set to `TRUE`.  The default is
#' `TRUE`.
#' @param print_details Whether or not to show details/progress of computations.
#'   Default is `FALSE`.
#' @param pl Whether or not to use parallel processing
#' @param cores The number of cores to use for parallel processing
#' @param compute_inffunc Whether or not to compute the influence functions. The
#'  default is `TRUE`. The influence functions are required for standard errors,
#'  uniform confidence bands, the parallel-trends pre-test, and for aggregating the
#'  group-time effects with [aggte()]. Set `compute_inffunc = FALSE` to obtain the
#'  group-time ATT *point estimates only*: this is faster and uses substantially less
#'  memory (no \eqn{n \times k} influence-function matrix is formed or bootstrapped),
#'  but the returned object has no standard errors / pre-test and cannot be passed to
#'  [aggte()]. The point estimates are identical to those from a full run. When
#'  `compute_inffunc = FALSE`, `bstrap` and `cband` are set to `FALSE` automatically.
#' @param est_method the method to compute group-time average treatment effects.  The default is "dr" which uses the doubly robust
#' approach in the `DRDID` package.  Other built-in methods
#' include "ipw" for inverse probability weighting and "reg" for
#' first step regression estimators.  The user can also pass their
#' own function for estimating group time average treatment
#' effects.  The required signature depends on the data structure:
#'
#' **Panel data** (`panel=TRUE`): `f(y1, y0, D, covariates,
#' i.weights, inffunc, ...)` where `y1` is an `n x 1` vector of
#' post-treatment outcomes, `y0` is an `n x 1` vector of
#' pre-treatment outcomes, `D` is a binary vector indicating
#' treatment group membership, `covariates` is an `n x k` matrix,
#' `i.weights` is a vector of sampling weights, and `inffunc` is a
#' logical requesting influence-function computation.
#'
#' **Repeated cross sections / unbalanced panel** (`panel=FALSE`):
#' `f(y, post, D, covariates, i.weights, inffunc, ...)` where `y` is
#' the outcome vector (length `n`), `post` is a binary indicator for
#' the post-treatment period, `D` is a binary treatment indicator,
#' `covariates` is an `n x k` matrix, `i.weights` is a vector of
#' sampling weights, and `inffunc` is a logical.
#'
#' In both cases the function should return a list that includes
#' `ATT` (the estimated group-time average treatment effect) and
#' `att.inf.func` (an `n x 1` influence function — one entry per
#' observation passed into the estimator).
#' The function can return other things as well, but these are
#' the only two that are required. With no covariates
#' (`xformla = NULL`), the built-in methods ("dr", "ipw", "reg") all
#' reduce to the unconditional difference-in-differences estimator, so
#' the choice among them is irrelevant; a custom `est_method` function
#' is still called (with an intercept-only `covariates` matrix) and
#' determines the estimates.
#' @param xformla A formula for the covariates to include in the
#'  model.  It should be of the form `~ X1 + X2`.  Default
#'  is NULL which is equivalent to `xformla=~1`.  This is
#'  used to create a matrix of covariates which is then passed
#'  to the 2x2 DID estimator chosen in `est_method`.
#'
#'  For time-varying covariates: (1) With balanced panel data,
#'  in each 2x2 comparison, the covariates
#'  are taken to be the value of the covariates in the earlier time
#'  period, and all of the underlying computations involve changes in Y
#'  as a function of those values of covariates.  (2) With repeated cross
#'  sections data and unbalanced panel data, the covariates are taken
#'  from each time period and computations involve Y_post conditional
#'  on X_post minus Y_pre conditional on X_pre.  A byproduct of this
#'  is that, with balanced panel data and in the presence of
#'  time-varying covariates, it is possible to get different numerical
#'  results according to whether or not `allow_unbalanced_panel=TRUE` or
#'  `FALSE`.
#' @param panel Whether or not the data is a panel dataset.
#'  The panel dataset should be provided in long format -- that
#'  is, where each row corresponds to a unit observed at a
#'  particular point in time.  The default is TRUE.  When
#'  using a panel dataset, the variable `idname` must
#'  be set.  When `panel=FALSE`, the data is treated
#'  as repeated cross sections.
#' @param allow_unbalanced_panel Whether or not function should
#'  "balance" the panel with respect to time and id.  The default
#'  value is `FALSE` which means that [att_gt()] will drop
#'  all units where data is not observed in all periods.
#'  The advantage of this is that the computations are faster
#'  (sometimes substantially).
#' @param control_group Which units to use as the control group.
#'  The default is "nevertreated" which sets the control group
#'  to be the group of units that never participate in the
#'  treatment.  This group does not change across groups or
#'  time periods.  The other option is to set
#'  `group="notyettreated"`.  In this case, the control group
#'  is set to the group of units that have not yet participated
#'  in the treatment in that time period.  This includes all
#'  never treated units, but it includes additional units that
#'  eventually participate in the treatment, but have not
#'  participated yet.
#' @param anticipation The number of time periods before participating
#'  in the treatment where units can anticipate participating in the
#'  treatment and therefore it can affect their untreated potential outcomes
#' @param faster_mode This option enables a faster version of `did`, optimizing
#' computation time for large datasets by improving data management within the package.
#' The default is set to `TRUE`. Both modes produce identical results up to
#' numerical precision; while the difference is minimal for small datasets,
#' the speedup is substantial for large ones.
#' @param base_period Whether to use a "varying" base period or a
#'  "universal" base period.  Either choice results in the same
#'  post-treatment estimates of ATT(g,t)'s.  In pre-treatment
#'  periods, using a varying base period amounts to computing a
#'  pseudo-ATT in each treatment period by comparing the change
#'  in outcomes for a particular group relative to its comparison
#'  group in the pre-treatment periods (i.e., in pre-treatment
#'  periods this setting computes changes from period t-1 to period
#'  t, but repeatedly changes the value of t)
#'
#'  A universal base period fixes the base period to always be
#'  (g-anticipation-1).  This does not compute
#'  pseudo-ATT(g,t)'s in pre-treatment periods, but rather
#'  reports average changes in outcomes from period t to
#'  (g-anticipation-1) for a particular group relative to its comparison
#'  group.  This is analogous to what is often reported in event
#'  study regressions.
#'
#'  Using a varying base period results in an estimate of
#'  ATT(g,t) being reported in the period immediately before
#'  treatment.  Using a universal base period normalizes the
#'  estimate in the period right before treatment (or earlier when
#'  the user allows for anticipation) to be equal to 0, but one
#'  extra estimate in an earlier period.
#'
#' @param ... Additional arguments to be passed to a custom `est_method`
#'  function. These are ignored when using built-in estimation methods
#'  (`"dr"`, `"ipw"`, `"reg"`).
#'
#' @section Options:
#'  When `faster_mode = FALSE`, setting `options(did.disable_precompute = TRUE)`
#'  disables the one-time positional precompute and assembles every 2x2
#'  comparison from the long data per cell, as in earlier versions. The
#'  covariate design matrix is still built once over the full sample under
#'  both settings, so results are identical either way; the option exists
#'  only as a debugging escape hatch.
#'
#' @references Callaway, Brantly and Pedro H.C. Sant'Anna.  \"Difference-in-Differences with Multiple Time Periods.\" Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. \doi{10.1016/j.jeconom.2020.12.001}, <https://arxiv.org/abs/1803.09015>
#'
#' @return an [`MP`] object containing all the results for group-time average
#'  treatment effects
#'
#'  The returned `inffunc` matrix collects the estimated influence functions,
#'  with one column per ATT(g,t) and one row per cross-sectional unit (one row
#'  per observation with repeated cross sections). Its rownames hold the unit
#'  ids (`idname`; an internal observation index for repeated cross sections)
#'  and are the authoritative link between rows and units: the row ORDER is
#'  mode-specific (`faster_mode = FALSE` sorts units by id, while
#'  `faster_mode = TRUE` uses an internal (period, cohort, id) ordering), so
#'  external consumers of the influence functions must align rows by rowname,
#'  never by position.
#'
#' @details # Examples:
#'
#' **Basic [att_gt()] call:**
#' ```{r, comment = "#>", collapse = TRUE}
#' # Example data
#' data(mpdta)
#' set.seed(09152024)
#' out1 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=NULL,
#'                data=mpdta)
#' summary(out1)
#' ```
#'
#' **Using covariates:**
#'
#' ```{r, comment = "#>", collapse = TRUE}
#' out2 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=~lpop,
#'                data=mpdta)
#' summary(out2)
#' ```
#'
#' **Specify comparison units:**
#'
#' ```{r, comment = "#>", collapse = TRUE}
#' out3 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=~lpop,
#'                control_group = "notyettreated",
#'                data=mpdta)
#' summary(out3)
#' ```
#'
#' @export


att_gt <- function(yname,
                   tname,
                   idname = NULL,
                   gname,
                   xformla = NULL,
                   data,
                   panel = TRUE,
                   allow_unbalanced_panel = FALSE,
                   control_group = c("nevertreated", "notyettreated"),
                   anticipation = 0,
                   weightsname = NULL,
                   fix_weights = NULL,
                   alp = 0.05,
                   bstrap = TRUE,
                   cband = TRUE,
                   biters = 1000,
                   clustervars = NULL,
                   est_method = "dr",
                   base_period = "varying",
                   faster_mode = TRUE,
                   print_details = FALSE,
                   pl = FALSE,
                   cores = 1,
                   compute_inffunc = TRUE,
                   ...) {
  # Capture extra arguments for custom est_method
  extra_args <- list(...)

  # Validate compute_inffunc (point-estimates-only switch)
  if (!is.logical(compute_inffunc) || length(compute_inffunc) != 1 || is.na(compute_inffunc)) {
    stop("compute_inffunc must be a single logical (TRUE or FALSE).")
  }
  # When influence functions are not computed there are no standard errors, no
  # uniform bands, and no parallel-trends pre-test, so the bootstrap is moot.
  if (!compute_inffunc) {
    bstrap <- FALSE
    cband <- FALSE
  }

  # Warn if extra arguments passed with built-in est_method
  if (length(extra_args) > 0 && !inherits(est_method, "function")) {
    warning("Extra arguments (", paste(names(extra_args), collapse = ", "),
            ") are ignored when using built-in est_method = \"", est_method,
            "\". Extra arguments are only passed to custom est_method functions.")
  }

  # Validate fix_weights
  if (!is.null(fix_weights)) {
    if (!is.character(fix_weights) || length(fix_weights) != 1 ||
        !(fix_weights %in% c("varying", "base_period", "first_period"))) {
      stop("fix_weights must be NULL or one of \"varying\", \"base_period\", or \"first_period\".")
    }
    if (!panel && fix_weights %in% c("base_period", "first_period")) {
      stop("fix_weights = \"", fix_weights, "\" is not supported for repeated cross sections ",
           "(panel = FALSE) because units are not tracked across periods. ",
           "Use fix_weights = \"varying\" or NULL instead.")
    }
    if (fix_weights == "varying" && panel && inherits(est_method, "function")) {
      stop("fix_weights = \"varying\" is not currently supported with custom est_method functions ",
           "on panel data. The \"varying\" option uses repeated cross-section estimators internally, ",
           "which require a different function signature (y, post, D) than the panel signature ",
           "(y1, y0, D). Use fix_weights = NULL, \"base_period\", or \"first_period\" instead.")
    }
  }

  # Validate est_method
  if (!inherits(est_method, "function")) {
    if (!is.character(est_method) || length(est_method) != 1) {
      stop("est_method must be a character string (\"dr\", \"ipw\", or \"reg\") or a custom function. ",
           "Received an object of class '", class(est_method)[1], "'.")
    }
    if (!(est_method %in% c("dr", "ipw", "reg"))) {
      stop("est_method must be one of \"dr\", \"ipw\", or \"reg\". Received: \"", est_method, "\".")
    }
  }

  # Require idname for panel data: without it the pre-processors fail with
  # confusing internal errors instead of pointing at the missing argument
  if (panel && is.null(idname)) {
    stop("Must provide idname when panel = TRUE. Set panel = FALSE for repeated cross sections.")
  }

  # Validate alp (significance level)
  if (!is.numeric(alp) || length(alp) != 1 || is.na(alp) || alp <= 0 || alp >= 1) {
    stop("alp must be a single number strictly between 0 and 1.")
  }

  # Validate biters (number of bootstrap iterations) when the bootstrap is used
  if (bstrap) {
    if (!is.numeric(biters) || length(biters) != 1 || is.na(biters) ||
        biters < 1 || biters != round(biters)) {
      stop("biters must be a single positive whole number.")
    }
  }

  # Warn users about anticipation and never-treated units
  if (anticipation > 0) {
    message("Note: anticipation = ", anticipation, ". Never-treated units (with group status 0 or Inf) ",
            "are assumed to never anticipate treatment. Anticipation only applies to eventually-treated units.")
  }

  # Check if user wants to run faster mode:
  if (faster_mode) {
    # this is a DIDparams2 object
    dp <- pre_process_did2(
      yname = yname,
      tname = tname,
      idname = idname,
      gname = gname,
      xformla = xformla,
      data = data,
      panel = panel,
      allow_unbalanced_panel = allow_unbalanced_panel,
      control_group = control_group,
      anticipation = anticipation,
      weightsname = weightsname,
      fix_weights = fix_weights,
      alp = alp,
      bstrap = bstrap,
      cband = cband,
      biters = biters,
      clustervars = clustervars,
      est_method = est_method,
      base_period = base_period,
      print_details = print_details,
      faster_mode = faster_mode,
      pl = pl,
      cores = cores,
      call = match.call()
    )

    # attach extra args for custom est_method
    dp$extra_args <- extra_args
    # whether to compute influence functions (FALSE = point estimates only)
    dp$compute_inffunc <- compute_inffunc

    #-----------------------------------------------------------------------------
    # Compute all ATT(g,t)
    #-----------------------------------------------------------------------------
    results <- compute.att_gt2(dp)
  } else {
    # this is a DIDparams object
    dp <- pre_process_did(
      yname = yname,
      tname = tname,
      idname = idname,
      gname = gname,
      xformla = xformla,
      data = data,
      panel = panel,
      allow_unbalanced_panel = allow_unbalanced_panel,
      control_group = control_group,
      anticipation = anticipation,
      weightsname = weightsname,
      fix_weights = fix_weights,
      alp = alp,
      bstrap = bstrap,
      cband = cband,
      biters = biters,
      clustervars = clustervars,
      est_method = est_method,
      base_period = base_period,
      print_details = print_details,
      pl = pl,
      cores = cores,
      call = match.call()
    )

    # attach extra args for custom est_method
    dp$extra_args <- extra_args
    # whether to compute influence functions (FALSE = point estimates only)
    dp$compute_inffunc <- compute_inffunc

    #-----------------------------------------------------------------------------
    # Compute all ATT(g,t)
    #-----------------------------------------------------------------------------
    results <- compute.att_gt(dp)
  }

  # extract ATT(g,t) and influence functions
  attgt.list <- results$attgt.list
  inffunc <- results$inffunc

  # Label the influence-function rows with the internal unit ids (the user's idname
  # for panels; ".rowid", one per observation, for repeated cross sections). The ROW
  # ORDER of inffunc is mode-specific: faster_mode = FALSE sorts units by id, while
  # faster_mode = TRUE keeps the internal (period, cohort, id) ordering -- so row
  # POSITIONS are not comparable across modes. The rownames attached here are the
  # authoritative unit labels; external consumers (e.g. sensitivity analyses or
  # custom cluster aggregation) must align rows by rowname, never by position.
  # Values are unchanged; only dimnames are added.
  if (!is.null(inffunc) && !is.null(dp$idname)) {
    inffunc_ids <- if (faster_mode) dp$time_invariant_data[[dp$idname]] else dp$data[[dp$idname]]
    # collapse per-observation ids (panel: one per period; unbalanced panel: one per
    # observation) to one label per influence-function row; unique() preserves the
    # first-appearance order, which is exactly each mode's inffunc row order.
    if (anyDuplicated(inffunc_ids)) inffunc_ids <- unique(inffunc_ids)
    if (length(inffunc_ids) == nrow(inffunc)) rownames(inffunc) <- as.character(inffunc_ids)
  }

  # process results
  # attgt.results <- process_attgt(attgt.list)
  tryCatch(
    {
      # Attempt to run this line for process results
      attgt.results <- process_attgt(attgt.list)
    },
    error = function(e) {
      # Handle the error
      if (faster_mode) {
        # If faster_mode is TRUE, send this stop message
        stop("An unexpected error occurred, normally associated with a singular matrix due to not enough control units. Try changing faster_mode=FALSE.")
      } else {
        # If faster_mode is FALSE, send this stop message
        stop("An unexpected error occurred, normally associated with a singular matrix due to not enough control units.")
      }
    }
  )
  group <- attgt.results$group
  att <- attgt.results$att
  tt <- attgt.results$tt



  # analytical standard errors
  # estimate variance. The i.i.d. form is clustered at the unit level; with a coarser cluster variable
  # the variance is formed from cluster sums instead (the cluster_analytic branch below).
  n <- if (faster_mode) dp$id_count else dp$n

  # All inference below requires the influence functions. With compute_inffunc = FALSE
  # (point estimates only) skip it entirely; see the else-branch before return().
  if (compute_inffunc) {
  # Analytical variance of the ATT(g,t)'s. With a cluster variable (beyond idname) and no bootstrap, use
  # the cluster-robust form -- the cluster sums of the influence function -- mirroring the cluster-sum
  # aggregation of the multiplier bootstrap (Callaway & Sant'Anna 2021, Remark 10). It is kept on the same
  # 1/n scaling as the i.i.d. V so that se = sqrt(diag(V)/n) and the Wald statistic n * b' V^{-1} b stay
  # valid for either form.
  # cluster variables beyond the unit id; clustering on the unit id alone is just the unit-level (i.i.d.)
  # SE. Use the *internal* unit id dp$idname -- the user's idname for panels, and ".rowid" for repeated
  # cross-sections / unbalanced panels (where the user may omit idname), mirroring how mboot() identifies units.
  unit_id <- dp$idname
  extra_clustervars <- clustervars[!(clustervars %in% c(unit_id, idname, ""))]
  # Per-unit cluster identifiers aligned with the rows of inffunc. faster_mode supplies dp$cluster_vector,
  # but for unbalanced panels (internally panel = FALSE, so the time-invariant data keeps every
  # observation) that vector is observation-length rather than unit-length and no longer aligns with the
  # unit-level influence function. Derive the vector from the data exactly as mboot() does -- one row per
  # cross-sectional unit, keyed on the internal unit id -- and store it back whenever it is missing OR does
  # not align with inffunc. This keeps the analytical clustered SE and aggte() working for faster_mode TRUE
  # and FALSE and for balanced and unbalanced panels; repeated cross sections (where the per-observation
  # vector already has one row per unit and aligns) fall through unchanged.
  if (length(extra_clustervars) > 0 && !is.null(dp$data) && !is.null(unit_id) &&
      (is.null(dp$cluster_vector) || length(dp$cluster_vector) != nrow(inffunc))) {
    cdat <- as.data.frame(dp$data)
    if (all(c(unit_id, extra_clustervars[1L]) %in% names(cdat))) {
      rebuilt_cv <- as.vector(unique(cdat[, c(unit_id, extra_clustervars[1L])])[, 2L])
      # only adopt the rebuilt vector if it lines up with the influence-function rows
      if (length(rebuilt_cv) == nrow(inffunc)) dp$cluster_vector <- rebuilt_cv
    }
  }
  cluster_vec <- dp$cluster_vector
  # Record which variable cluster_vec represents (when it is valid), so aggte() can tell whether an
  # inherited or overridden clustervars can be honored from this object -- and refuse to silently cluster
  # on the wrong variable.
  if (length(extra_clustervars) > 0 && !is.null(cluster_vec) && length(cluster_vec) == nrow(inffunc)) {
    dp$cluster_vector_var <- extra_clustervars[1L]
  }
  cluster_analytic <- (length(extra_clustervars) > 0) && !bstrap &&
    !is.null(cluster_vec) && length(cluster_vec) == nrow(inffunc)
  if (cluster_analytic) {
    Sc <- rowsum(as.matrix(inffunc), cluster_vec)   # n_clusters x k cluster sums
    V  <- Matrix::Matrix(crossprod(Sc) / n)
  } else {
    V <- Matrix::t(inffunc) %*% inffunc / (n)
  }
  se <- sqrt(Matrix::diag(V) / n)

  # Zero standard error replaced by NA
  se[se <= sqrt(.Machine$double.eps) * 10] <- NA

  # If a cluster variable beyond idname is requested without the bootstrap but the cluster-robust variance
  # could not be formed (e.g. cluster identifiers unavailable), fall back to i.i.d. and let the user know.
  if ((length(extra_clustervars) > 0) & !bstrap & !cluster_analytic) {
    warning("Clustered standard errors could not be computed analytically; the reported standard errors do NOT account for clustering. Set bstrap = TRUE for the cluster-robust multiplier bootstrap.")
  }

  # Identify entries of main diagonal V that are zero or NA
  zero_na_sd_entry <- unique(which(is.na(se)))

  # bootstrap variance matrix
  if (bstrap) {
    bout <- mboot(inffunc, DIDparams = dp, pl = pl, cores = cores, return_V = FALSE)
    bres <- bout$bres

    if (length(zero_na_sd_entry) > 0) {
      se[-zero_na_sd_entry] <- bout$se[-zero_na_sd_entry]
    } else {
      se <- bout$se
    }
  }
  # Zero standard error replaced by NA
  se[se <= sqrt(.Machine$double.eps) * 10] <- NA


  #-----------------------------------------------------------------------------
  # compute Wald pre-test
  #-----------------------------------------------------------------------------

  # Determine whether the variance matrix V is a valid basis for the Wald pre-test. The i.i.d. form
  # V = t(inffunc) %*% inffunc / n is valid for balanced panels, unbalanced panels (influence functions
  # are aggregated to the unit level), and repeated cross-sections (CLT applies per independent
  # observation). When a cluster variable beyond idname is specified and V is built from cluster sums
  # (cluster_analytic, above), V is cluster-robust and the Wald test remains valid. It is only skipped
  # when an extra cluster variable is present but V is the i.i.d. form (e.g. bstrap = TRUE), since then
  # V ignores between-cluster correlation; in that case we rely on the bootstrap confidence bands.

  wald_invalid <- NULL
  if (length(extra_clustervars) > 0 && !cluster_analytic) {
    wald_invalid <- paste0(
      "The Wald pre-test is not reported when clustering beyond the unit level ",
      "(clustervars = '", paste(extra_clustervars, collapse = "', '"), "') ",
      "because the analytical variance matrix does not account for between-cluster ",
      "correlation. Use the bootstrap confidence intervals to assess pre-trends."
    )
  }

  if (!is.null(wald_invalid)) {
    message(wald_invalid)
    W <- NULL
    Wpval <- NULL
  }

  if (is.null(wald_invalid)) {
    # select which periods are pre-treatment
    pre <- which(group > tt)
    # number of pre-treatment cells before the variance filter below, so the
    # no-pre-periods warning can distinguish "no pre-treatment cells exist"
    # from "all pre-treatment cells were dropped for NA/zero variance"
    n_pre_cells <- length(pre)

    # Drop group-periods that have variance equal to zero (singularity problems)
    if (length(zero_na_sd_entry) > 0) {
      pre <- pre[!(pre %in% zero_na_sd_entry)]
    }
    # pseudo-atts in pre-treatment periods
    preatt <- as.matrix(att[pre])

    # covariance matrix of pre-treatment atts
    preV <- as.matrix(V[pre, pre])

    # check if there are actually any pre-treatment periods
    if (length(preV) == 0) {
      if (n_pre_cells > 0) {
        # pre-treatment cells existed, but every one was dropped by the
        # zero/NA-variance filter above
        msg <- paste0(
          "No pre-treatment periods available for the Wald pre-test of parallel trends: ",
          "pre-treatment ATT(g,t) estimates exist, but all of them have missing or zero ",
          "variance (e.g., because the underlying 2x2 estimation failed for those cells), ",
          "so they were dropped from the pre-test."
        )
      } else {
        msg <- paste0(
          "No pre-treatment periods available for the Wald pre-test of parallel trends. ",
          "This can happen when all groups are first treated early in the panel ",
          "(e.g., in the second time period) so that no pre-treatment ATT(g,t) estimates exist."
        )
        if (anticipation > 0) {
          msg <- paste0(msg, " Note: anticipation=", anticipation, " further reduces the number of available pre-treatment periods.")
        }
      }
      warning(msg)
      W <- NULL
      Wpval <- NULL
    } else if (sum(is.na(preV))) {
      warning("Not returning pre-test Wald statistic due to NA pre-treatment values")
      W <- NULL
      Wpval <- NULL
    } else if (rcond(preV) <= .Machine$double.eps) {
      # singular covariance matrix for pre-treatment periods
      warning("Not returning pre-test Wald statistic due to singular covariance matrix")
      W <- NULL
      Wpval <- NULL
    } else {
      # everything is working...
      W <- n * t(preatt) %*% solve(preV) %*% preatt
      q <- length(pre) # number of restrictions
      Wpval <- round(1 - pchisq(W, q), 5)
    }
  }


  #-----------------------------------------------------------------------------
  # compute confidence intervals / bands
  #-----------------------------------------------------------------------------

  # critical value from N(0,1), for pointwise
  cval <- qnorm(1 - alp / 2)

  # in order to get uniform confidence bands
  # HAVE to use the bootstrap
  if (bstrap) {
    if (cband) {
      # for uniform confidence band
      # compute new critical value
      # see paper for details
      # Vectorized like mboot(): both quantiles of each bootstrap column are
      # taken in a single type = 1 call (one sort per column instead of two),
      # with the qnorm normalization hoisted out of the per-column loop.
      iqr_norm <- qnorm(.75) - qnorm(.25)
      bSigma <- apply(bres, 2, function(b) {
        qs <- quantile(b, c(.25, .75), type = 1, na.rm = TRUE)
        (qs[2L] - qs[1L]) / iqr_norm
      })

      bSigma[bSigma <= sqrt(.Machine$double.eps) * 10] <- NA

      # sup-t confidence band: per bootstrap draw, the max over columns of
      # |b / bSigma|, as a pmax column-fold (equivalent to the previous row-wise
      # apply(max, na.rm = TRUE): NA entries are ignored via -Inf, and an all-NA
      # row yields -Inf, exactly as before).
      if (ncol(bres) == 0L) {
        # every bootstrap dimension was degenerate (mboot dropped all columns):
        # the previous row-wise apply(max, na.rm = TRUE) returned -Inf per row
        bT <- rep(-Inf, nrow(bres))
      } else {
        scaled <- abs(bres / rep(bSigma, each = nrow(bres)))
        scaled[is.na(scaled)] <- -Inf
        bT <- scaled[, 1L]
        for (j in seq_len(ncol(scaled))[-1L]) bT <- pmax(bT, scaled[, j])
      }
      cval <- quantile(bT, 1 - alp, type = 1, na.rm = TRUE)
      if (cval >= 7) {
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when the number of observations per group is small and/or there is not much variation in outcomes.")
      }
    }
  }


  } else {
    # point estimates only (compute_inffunc = FALSE): no influence functions were
    # computed, so there are no standard errors, uniform bands, or pre-test.
    se <- rep(NA_real_, length(att))
    V <- NULL
    cval <- qnorm(1 - alp / 2)
    W <- NULL
    Wpval <- NULL
  }

  # Return this list
  return(MP(group = group, t = tt, att = att, V_analytical = V, se = se, c = cval, inffunc = inffunc, n = n, W = W, Wpval = Wpval, alp = alp, DIDparams = dp))
}

## -----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# test it works without covariates and only two gropups
#-----------------------------------------------------------------------------
test_that("att_gt works with 2 groups", {
  set.seed(20241017)
  n = 5000
  # Generate data with 4 periods
  p3_true = exp(0.5)
  denominator = 1 + p3_true

  #probability of each group: 2 groups in total (G=4)
  p3_true = p3_true / denominator
  pinf_true = 1 / denominator

  # Generate treatment status
  g <- as.numeric(runif(n) <= p3_true)
  g <- ifelse(g==1,3,0)


  #Potential outcomes: Yt_g:=Y_t(g)
  index_unobs_het = g
  nu <- stats::rnorm(n, mean = index_unobs_het, sd = 1)
  index_trend <- 1
  index_att <- 3
  att_rand <- stats::rnorm(n, mean = index_att, sd = 1)
  # Generate untreated potential outcomes
  Yt1_ginf = index_trend + nu + rnorm(n)
  Yt2_ginf = 2*index_trend + nu + rnorm(n)
  Yt3_ginf = 3*index_trend + nu + rnorm(n)
  Yt4_ginf = 4*index_trend + nu + rnorm(n)

  # Generate treated potential outcomes
  Yt1_g3 = index_trend + nu + rnorm(n)
  Yt2_g3 = 2*index_trend + nu + rnorm(n)
  Yt3_g3 = att_rand + 3*index_trend + nu + rnorm(n)
  Yt4_g3 = 1.5*att_rand + 4*index_trend + nu + rnorm(n)

  #Observed data
  y1 <- (g==3) * Yt1_g3 + (g==0) * Yt1_ginf
  y2 <- (g==3) * Yt2_g3 + (g==0) * Yt2_ginf
  y3 <- (g==3) * Yt3_g3 + (g==0) * Yt3_ginf
  y4 <- (g==3) * Yt4_g3 + (g==0) * Yt4_ginf

  data <- data.frame(y1,y2,y3,y4,g)
  data$id <- 1:dim(data)[1]
  # Have data in long format
  data <- data |>
    tidyr::pivot_longer(
      cols = starts_with("y"),
      names_to = "t",
      names_prefix = "y",
      values_to = "y",
    )
  data$t <- as.numeric(data$t)
  data$g <- as.numeric(data$g)

  # Run regression with never-treated and varying base period
  csdid_nt_varying <- did::att_gt(yname = "y",
                          idname = "id",
                          gname = "g",
                          tname = "t",
                          data = data,
                          control_group = "nevertreated",
                          panel = TRUE,
                          xformla = ~1,
                          bstrap = FALSE,
                          cband = FALSE,
                          est_method = "reg",
                          base_period = "varying"
  )

  # Run regression with never-treated and universal base period
  csdid_nt_universal <- did::att_gt(yname = "y",
                                  idname = "id",
                                  gname = "g",
                                  tname = "t",
                                  data = data,
                                  control_group = "nevertreated",
                                  panel = TRUE,
                                  xformla = ~1,
                                  bstrap = FALSE,
                                  cband = FALSE,
                                  est_method = "reg",
                                  base_period = "varying"
  )
  # Run regression with not-yet-treated and varying base period
  csdid_nyt_varying <- did::att_gt(yname = "y",
                                  idname = "id",
                                  gname = "g",
                                  tname = "t",
                                  data = data,
                                  control_group = "notyettreated",
                                  panel = TRUE,
                                  xformla = ~1,
                                  bstrap = FALSE,
                                  cband = FALSE,
                                  est_method = "reg",
                                  base_period = "varying"
  )

  # Run regression with not-yet-treated and universal base period
  csdid_nyt_universal <- did::att_gt(yname = "y",
                                    idname = "id",
                                    gname = "g",
                                    tname = "t",
                                    data = data,
                                    control_group = "notyettreated",
                                    panel = TRUE,
                                    xformla = ~1,
                                    bstrap = FALSE,
                                    cband = FALSE,
                                    est_method = "reg",
                                    base_period = "varying"
  )

  expect_equal(csdid_nyt_universal$att[2], 3, tol=.1)
  expect_equal(csdid_nt_universal$att[2], 3, tol=.1)
  expect_equal(csdid_nyt_varying$att[2], 3, tol=.1)
  expect_equal(csdid_nt_varying$att[2], 3, tol=.1)

  expect_equal(csdid_nyt_universal$att[2], csdid_nt_universal$att[2], tol=.0001)
  expect_equal(csdid_nyt_universal$att[2], csdid_nyt_varying$att[2], tol=.0001)
  expect_equal(csdid_nyt_universal$att[2], csdid_nt_varying$att[2], tol=.0001)


})

# Cluster-robust multiplier bootstrap (Callaway & Sant'Anna 2021, Remark 10): the clustered standard
# error aggregates the influence function to cluster *sums*. For equal-sized clusters this coincides with
# the previous cluster-mean aggregation; for unbalanced clusters and repeated cross-sections it follows
# the cluster-sum form that aligns with the (1/n) empirical average defining the estimator.
# small staggered panel with units grouped into clusters; bal controls equal vs unequal cluster sizes
.make_clustered <- function(seed, bal, G = 40L) {
  set.seed(seed)
  sz <- if (bal) rep(4L, G) else rep(c(1L, 2L, 3L, 8L), length.out = G)
  N <- sum(sz); cl <- rep(seq_len(G), times = sz)
  a <- rnorm(G, 0, 1)[cl]; nu <- rnorm(N, 0, 1); g <- ifelse(runif(N) < 0.5, 2L, 0L)
  d <- data.frame(id = rep(seq_len(N), each = 3L), t = rep(1:3, N), cl = rep(cl, each = 3L),
                  g = rep(g, each = 3L), a = rep(a, each = 3L), nu = rep(nu, each = 3L))
  d$y <- d$a + d$nu + 0.5 * d$t + (d$g == 2L & d$t >= 2L) + rnorm(nrow(d))
  d[order(d$id, d$t), ]
}

.cluster_targets <- function(res) {  # analytical cluster-SUM and cluster-MEAN for att(2,2)
  k <- which(res$group == 2L & res$t == 2L)
  cv <- res$DIDparams$cluster_vector
  inf <- as.matrix(res$inffunc)[, k]; n <- nrow(res$inffunc)
  S <- as.numeric(rowsum(matrix(inf, ncol = 1L), cv))
  nc <- as.numeric(table(cv)); Gc <- length(S)
  list(k = k, sum = sqrt(sum(S^2)) / n, mean = sqrt(sum((S / nc)^2)) / Gc, ok = length(cv) == n)
}

test_that("clustered mboot SE matches the cluster-sum (Remark 10) for UNBALANCED clusters", {
  skip_on_cran()  # bootstrap-heavy (biters = 5000)
  d <- .make_clustered(101L, bal = FALSE)
  res <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                control_group = "nevertreated", bstrap = TRUE, biters = 5000L,
                clustervars = "cl", pl = FALSE, cband = FALSE, base_period = "varying")
  tg <- .cluster_targets(res)
  expect_true(tg$ok)  # cluster_vector must be present and aligned with the influence-function rows
  # cluster-sum and cluster-mean genuinely differ when clusters are unbalanced
  expect_gt(abs(tg$sum - tg$mean) / tg$mean, 0.05)
  # reported SE tracks the cluster-SUM target (within bootstrap Monte Carlo tolerance) ...
  expect_equal(unname(res$se[tg$k]), tg$sum, tolerance = 0.08)
  # ... and is clearly NOT the cluster-mean
  expect_gt(abs(res$se[tg$k] - tg$mean) / tg$mean, 0.05)
})

test_that("clustered mboot SE is unchanged for BALANCED clusters (cluster-sum == cluster-mean)", {
  skip_on_cran()  # bootstrap-heavy (biters = 5000)
  d <- .make_clustered(202L, bal = TRUE)
  res <- att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
                control_group = "nevertreated", bstrap = TRUE, biters = 5000L,
                clustervars = "cl", pl = FALSE, cband = FALSE, base_period = "varying")
  tg <- .cluster_targets(res)
  expect_true(tg$ok)  # cluster_vector must be present and aligned with the influence-function rows
  expect_equal(tg$sum, tg$mean, tolerance = 1e-8)            # identical targets when balanced
  expect_equal(unname(res$se[tg$k]), tg$sum, tolerance = 0.08)
})

test_that("clustering validation is preserved (at most one cluster variable beyond idname)", {
  d <- .make_clustered(303L, bal = FALSE); d$cl2 <- d$cl
  # two clustering variables (neither is idname) must still be rejected after the cluster-sum change
  expect_error(
    att_gt(yname = "y", tname = "t", idname = "id", gname = "g", data = d,
           control_group = "nevertreated", bstrap = TRUE, biters = 99L,
           clustervars = c("cl", "cl2"), pl = FALSE, cband = FALSE, base_period = "varying"),
    regexp = "length 1|character scalar|clustervars|[Aa]t most one cluster"
  )
})

#' @title Get the cohort for the current g,t values
#' @description A utility function to get vector with index of units that will be included in the DiD estimation for the current g,t pair.
#'
#' @param group Group.
#' @param time Time period.
#' @param tfac Time factor, which is 1 for varying base period and 0 for universal base period.
#' @param pret Pre treatment period.
#' @param dp2 DiD parameters v2.0.
#'
#' @return A vector of 1s, 0s and NAs indicating the cohort for the current g,t pair.
#' @noRd
get_did_cohort_index <- function(group, time, tfac, pret, dp2){
  # return a vector of dimension id_size with 1, 0 or NA values
  time_periods <- dp2$time_periods

  if(dp2$panel){
    # --- Balanced panel: use pre-computed cumulative sizes ---
    cohort_vec <- dp2$cohort_counts$cohort
    cum_sizes <- dp2$.cohort_cum_sizes  # pre-computed in compute.att_gt2

    # find the first cohort after the cutoff for not-yet-treated
    if (dp2$control_group == "notyettreated") {
      cutoff <- time_periods[max(time, pret) + tfac] + dp2$anticipation
      min_idx <- match(TRUE, cohort_vec > cutoff)
      min_control_group <- if (is.na(min_idx)) Inf else cohort_vec[min_idx]
    } else {
      min_control_group <- Inf
    }

    # control group boundaries
    ctrl_start_idx <- match(TRUE, cohort_vec >= min_control_group)
    if (is.na(ctrl_start_idx)) {
      start_control <- dp2$id_count + 1L  # no control group found
    } else {
      start_control <- if (ctrl_start_idx == 1L) 1L else cum_sizes[ctrl_start_idx - 1L] + 1L
    }
    end_control <- cum_sizes[length(cum_sizes)]

    # treated group boundaries
    treat_idx <- match(dp2$treated_groups[group], cohort_vec)
    if (is.na(treat_idx)) {
      stop("Internal error: treated group ", dp2$treated_groups[group],
           " not found in cohort_vec. Please report this issue.")
    }
    start_treat <- if (treat_idx == 1L) 1L else cum_sizes[treat_idx - 1L] + 1L
    end_treat <- cum_sizes[treat_idx]

    # fill the cohort index vector
    did_cohort_index <- rep(NA_integer_, dp2$id_count)
    if (start_control <= end_control) did_cohort_index[start_control:end_control] <- 0L
    did_cohort_index[start_treat:end_treat] <- 1L

  } else {
    # getting the index to get units who will participate in the estimation for the (g,t) cell.
    # Note: This works because the data is already ordered in a specific way. Changing that order will break this.

    dat <- dp2$time_invariant_data          # shortcut

    # flags for treated and control units. The per-group flag, the
    # never-treated flag and the per-period masks are loop-invariant across
    # (g,t) cells, so they are pre-computed once in compute.att_gt2() instead
    # of rescanning every row per cell (complete.cases() ran upstream in
    # pre_process_did2, so == has no NA hazard relative to %in%).
    Gflag <- dp2$.gflag_by_group[[group]]

    if (dp2$control_group == "nevertreated") {
      Cflag <- dp2$.never_treated
    } else {  # not-yet-treated: the cutoff varies with t, so this comparison stays per cell
      Cflag <- dp2$.never_treated |
        (dat[[dp2$gname]] > dp2$time_periods[max(time, pret) + tfac] + dp2$anticipation &
           !Gflag)
    }

    # keep only rows observed in pret or t
    keep <- dp2$.period_masks[[time + tfac]] | dp2$.period_masks[[pret]]

    did_cohort_index <- rep(NA_integer_, nrow(dat))
    did_cohort_index[keep & Cflag] <- 0L
    did_cohort_index[keep & Gflag] <- 1L

  }

  return(did_cohort_index)
}

#' @title Wrapper to run DRDID package
#' @description A utility function to run the DRDID package for the current g,t pair.
#'
#' @param cohort_data Data table with the cohort data for the current g,t pair
#' @param covariates Matrix of covariates to be used in the estimation
#' @param dp2 DiD parameters v2.0.
#'
#' @return A list containing the estimated ATT and the influence function as
#'  sparse triplets (`if_i` row indices, `if_x` values), except on the force_rc
#'  path which returns the dense stacked vector (`inf_func`) for the caller's
#'  half-fold. `if_i`/`if_x` (or `inf_func`) are NULL for point estimates only.
#' @noRd
run_DRDID <- function(cohort_data, covariates, dp2, g_val = NULL, t_val = NULL,
                      pret_val = NULL, force_rc = FALSE, check_cache_key = NULL){

  extra_args <- if (is.null(dp2$extra_args)) list() else dp2$extra_args
  gt_label <- if (!is.null(g_val) && !is.null(t_val)) paste0(" for group ", g_val, " in time period ", t_val) else ""
  # whether to compute influence functions (default TRUE; FALSE = point estimates only)
  do_inf <- is.null(dp2$compute_inffunc) || isTRUE(dp2$compute_inffunc)

  if(dp2$panel && !force_rc){
    # --------------------------------------
    # Panel Data
    # --------------------------------------
    # cohort_data is a plain named list of vectors (D, y1, y0, i.weights), built in
    # run_att_gt_estimation(). Working on the vectors directly avoids per-cell
    # data.table construction/extraction; the estimator inputs are identical.

    # still total number of units (not just included in G or C)
    D_all <- cohort_data$D
    n <- length(D_all)

    # pick up the indices for units that will be used to compute ATT(g,t)
    valid_obs <- which(!is.na(D_all))
    D    <- D_all[valid_obs]
    y1   <- cohort_data$y1[valid_obs]
    y0   <- cohort_data$y0[valid_obs]
    i.weights <- cohort_data$i.weights[valid_obs]

    if(dp2$xformla != ~1){
      covariates <- covariates[valid_obs,]
      intercept_only <- FALSE
    } else {
      covariates <- rep(1, length(valid_obs))
      intercept_only <- TRUE
    }
    covariates <- as.matrix(covariates)

    # num obs. for computing ATT(g,t)
    n1 <- length(D)

    #-----------------------------------------------------------------------------
    # check for overlap and regression problems
    #-----------------------------------------------------------------------------
    custom_est_method <- inherits(dp2$est_method, "function")

    if (!custom_est_method) {
      D_vec <- D

      # Guard booleans: reuse the cached result when run_att_gt_estimation()
      # established that this cell's (covariates, D) inputs are bit-identical to
      # an earlier cell's (panel + nevertreated + non-varying weights; see
      # compute.att_gt2). Warnings are still emitted per cell below, so cached
      # failures keep the per-cell warning count and text unchanged.
      guard <- if (!is.null(check_cache_key)) dp2$.check_cache[[check_cache_key]] else NULL
      if (is.null(guard)) {
        # overlap check for pscore based methods; regression-feasibility check
        # for outcome-regression methods, run only if overlap passed (matching
        # the early return on overlap failure below)
        overlap_fail <- (dp2$est_method %in% c("dr", "ipw")) &&
          overlap_check_fail(covariates, D_vec, intercept_only)
        rcond_fail <- !overlap_fail && (dp2$est_method %in% c("dr", "reg")) &&
          rcond_check_fail(covariates, D_vec, intercept_only)
        guard <- list(overlap_fail = overlap_fail, rcond_fail = rcond_fail)
        if (!is.null(check_cache_key)) dp2$.check_cache[[check_cache_key]] <- guard
      }

      if (guard$overlap_fail) {
        warning(paste0("overlap condition violated", gt_label))
        return(list(att = NA, if_i = if (do_inf) seq_len(n) else NULL,
                    if_x = if (do_inf) rep(NA_real_, n) else NULL))
      }
      if (guard$rcond_fail) {
        warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned", gt_label, "; consider centering/rescaling covariates or removing collinear terms"))
        return(list(att = NA, if_i = if (do_inf) seq_len(n) else NULL,
                    if_x = if (do_inf) rep(NA_real_, n) else NULL))
      }
    }

    #-----------------------------------------------------------------------------
    # code for actually computing ATT(g,t)
    #-----------------------------------------------------------------------------


    if (inherits(dp2$est_method, "function")) {
      # user-specified function
      attgt <- do.call(dp2$est_method, c(list(
                          y1=y1,
                          y0=y0,
                          D=D,
                          covariates=covariates,
                          i.weights=i.weights,
                          inffunc=do_inf), extra_args))
    } else if (dp2$est_method == "ipw") {
      # inverse-probability weights
      attgt <- std_ipw_did_panel(y1=y1,
                                        y0=y0,
                                        D=D,
                                        covariates=covariates,
                                        i.weights=i.weights,
                                        boot=FALSE, inffunc=do_inf)
    } else if (dp2$est_method == "reg") {
      # regression
      attgt <- reg_did_panel(y1=y1,
                                    y0=y0,
                                    D=D,
                                    covariates=covariates,
                                    i.weights=i.weights,
                                    boot=FALSE, inffunc=do_inf)
    } else {
      # doubly robust, this is default
      attgt <- drdid_panel(y1=y1,
                                  y0=y0,
                                  D=D,
                                  covariates=covariates,
                                  i.weights=i.weights,
                                  boot=FALSE, inffunc=do_inf)
    }

    # adjust influence function to account for only using
    # subgroup to estimate att(g,t). Stored as sparse triplets (if_i row
    # indices / if_x values) rather than a dense length-n vector: the assembly
    # in compute.att_gt2() concatenates the triplets directly, mirroring the
    # slow path, instead of rescanning dense columns with which().
    if (do_inf) {
      if_i <- valid_obs
      if_x <- (n/n1)*attgt$att.inf.func
    } else {
      if_i <- NULL   # point estimates only
      if_x <- NULL
    }

  } else {
    # --------------------------------------
    # Repeated Cross-Section
    # --------------------------------------
    # cohort_data is a plain named list of vectors (D, y, post, i.weights, .rowid),
    # built in run_att_gt_estimation(). Working on the vectors directly avoids per-cell
    # data.table construction/extraction; the estimator inputs are identical.
    D_all <- cohort_data$D
    rowid_all <- cohort_data$.rowid   # all rows (needed for the unbalanced aggregation)

    # still total number of obs. (not just included in G or C)
    n <- length(D_all)

    # pick up the indices for units that will be used to compute ATT(g,t)
    valid_obs <- which(!is.na(D_all))
    D    <- D_all[valid_obs]
    y    <- cohort_data$y[valid_obs]
    post <- cohort_data$post[valid_obs]
    i.weights <- cohort_data$i.weights[valid_obs]

    # num obs. for computing ATT(g,t)
    n1 <- length(D)

    if(dp2$xformla != ~1){
      covariates <- covariates[valid_obs,]
      intercept_only <- FALSE
    } else {
      covariates <- covariates[valid_obs]
      intercept_only <- TRUE
    }

    covariates <- as.matrix(covariates)

    # number of rows of the final influence-function column for early returns:
    # the unbalanced-panel branch aggregates observations to units; force_rc
    # failure triplets are already at unit level (the caller's half-fold only
    # applies to the dense success vector); plain RC has one row per observation
    n_inf <- if (dp2$allow_unbalanced_panel || force_rc) dp2$id_count else n

    skip_this_att_gt <- FALSE
    if (sum(D * post) == 0) {
      warning(paste0("No units in group ", g_val, " in time period ", t_val))
      skip_this_att_gt <- TRUE
    }
    if (sum(D * (1 - post)) == 0) {
      warning(paste0("No units in group ", g_val, " in time period ", pret_val))
      skip_this_att_gt <- TRUE
    }
    if (sum((1 - D) * post) == 0) {
      warning(paste0("No available control units for group ", g_val,
                     " in time period ", t_val))
      skip_this_att_gt <- TRUE
    }
    if (sum((1 - D) * (1 - post)) == 0) {
      warning(paste0("No available control units for group ", g_val,
                     " in time period ", pret_val))
      skip_this_att_gt <- TRUE
    }
    if (skip_this_att_gt) {
      return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                  if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
    }

    #-----------------------------------------------------------------------------
    # check for overlap and regression problems
    #-----------------------------------------------------------------------------
    custom_est_method <- inherits(dp2$est_method, "function")

    if (!custom_est_method) {
      D_vec <- D

      # checks for pscore based methods (no caching on the RC branch: the
      # cohort index varies cell by cell here)
      if (dp2$est_method %in% c("dr", "ipw")) {
        if (overlap_check_fail(covariates, D_vec, intercept_only)) {
          warning(paste0("overlap condition violated", gt_label))
          return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                      if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
        }
      }

      # check if can run regression using control units
      if (dp2$est_method %in% c("dr", "reg")) {
        if (rcond_check_fail(covariates, D_vec, intercept_only)) {
          warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned", gt_label, "; consider centering/rescaling covariates or removing collinear terms"))
          return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                      if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
        }
      }
    }

    #-----------------------------------------------------------------------------
    # code for actually computing ATT(g,t)
    #-----------------------------------------------------------------------------

    if (inherits(dp2$est_method, "function")) {
      # user-specified function
      attgt <- do.call(dp2$est_method, c(list(
                          y=y,
                          post=post,
                          D=D,
                          covariates=covariates,
                          i.weights=i.weights,
                          inffunc=do_inf), extra_args))
    } else if (dp2$est_method == "ipw") {
      # inverse-probability weights
      attgt <- std_ipw_did_rc(y=y,
                             post=post,
                             D=D,
                             covariates=covariates,
                             i.weights=i.weights,
                             boot=FALSE, inffunc=do_inf)
    } else if (dp2$est_method == "reg") {
      # regression
      attgt <- reg_did_rc(y=y,
                         post=post,
                         D=D,
                         covariates=covariates,
                         i.weights=i.weights,
                         boot=FALSE, inffunc=do_inf)
    } else {
      # doubly robust, this is default
      attgt <- drdid_rc(y=y,
                       post=post,
                       D=D,
                       covariates=covariates,
                       i.weights=i.weights,
                       boot=FALSE, inffunc=do_inf)
    }

    # n/n1 adjusts for estimating the
    # att_gt only using observations from groups
    # G and C
    # adjust influence function to account for only using
    # subgroup to estimate att(g,t). Stored as sparse triplets (if_i/if_x),
    # mirroring the slow path; the dense vector is kept only where it is still
    # needed (unbalanced-panel rowsum aggregation, force_rc half-fold).
    if (!do_inf) {
      if_i <- NULL   # point estimates only
      if_x <- NULL
    } else if(dp2$allow_unbalanced_panel){
      # since this is technically a panel data but ran as RCS, we need to adjust the
      # influence function by aggregating influence value by .rowid (several obs of one
      # unit can be used to estimate ATT in each 2x2). rowsum(..., reorder = FALSE)
      # reproduces the data.table `by = .rowid` first-appearance group order exactly.
      # The aggregated dense vector is then converted to triplets with the same
      # is.na | != 0 scan the assembly previously applied to dense columns.
      inf_func_long <- numeric(n)
      inf_func_long[valid_obs] <- (dp2$id_count / n1) * attgt$att.inf.func
      agg <- dp2$.rowid_agg
      inf_func_vector <- if (!is.null(agg)) {
        # precomputed sparse unit-aggregation operator (see compute.att_gt2):
        # accumulates in row order within each unit, bit-identical to the
        # rowsum() fallback but without per-cell hashing/dimnames work
        as.vector(agg %*% inf_func_long)
      } else {
        rowsum(inf_func_long, rowid_all, reorder = FALSE)[, 1L]
      }
      if_i <- which(is.na(inf_func_vector) | inf_func_vector != 0)
      if_x <- inf_func_vector[if_i]

    } else if (force_rc) {
      # balanced panel run as RC (fix_weights = "varying"): return the dense
      # stacked [all pre, all post] vector; run_att_gt_estimation() half-folds
      # it to unit level and converts it to triplets there.
      inf_func_vector <- rep(0, n)
      inf_func_vector[valid_obs] <- (n/n1)*attgt$att.inf.func
      return(list(att = attgt$ATT, inf_func = inf_func_vector))
    } else {
      if_i <- valid_obs
      if_x <- (n/n1)*attgt$att.inf.func
    }

  }

  return(list(att = attgt$ATT, if_i = if_i, if_x = if_x))

}


#' @title Run ATT estimation for a given group-time pair
#'
#' @description `run_att_gt_estimation` does the main work for computing
#'  multiperiod group-time average treatment effects
#' @param g group of interest (treated group at time t)
#' @param t time period
#' @param dp2 A DIDparams object v2.0
#'
#' @return a list with the gt cell and the results after performing estimation
#'
#' @keywords internal
run_att_gt_estimation <- function(g, t, dp2){

  if(dp2$print_details){cat("\n", paste0("Evaluating (g,t) = (",dp2$treated_groups[g],",",dp2$time_periods[t],")"))}
  tfac <- if (dp2$base_period != "universal") 1L else 0L
  # set pret
  # varying base period
  pret <- t

  # Use pre-computed pret for universal base period and post-treatment periods
  pret_g <- dp2$.pret_by_group[g]  # pre-computed in compute.att_gt2

  # universal base period
  if (dp2$base_period == "universal") {
    pret <- pret_g
  }

  # check if in post-treatment period
  if ((dp2$treated_groups[g] <= dp2$time_periods[(t+tfac)])) {

    # update pre-period if in post-treatment period to
    # be  period (g-delta-1)
    pret <- pret_g

    # print a warning message if there are no pre-treatment periods
    if (is.na(pret)) {
      warning(paste0("There are no pre-treatment periods for the group first treated at ", dp2$treated_groups[g], "\nUnits from this group are dropped"))

      # if there are no pre-treatment periods, code will
      # jump out of this loop
      return(NULL)
    }
  }

  #-----------------------------------------------------------------------------
  # if we are in period (g-1), normalize results to be equal to 0
  # and break without computing anything
  if (dp2$base_period == "universal") {
    if (dp2$time_periods[pret] == dp2$time_periods[(t+tfac)]) {
      return(list(base_period_norm = TRUE))
    }
  }

  # get units in treatment and control group
  did_cohort_index <- get_did_cohort_index(group = g, time = t, tfac = tfac, pret = pret, dp2)
  # In case of no treatment or control group in the cohort, return NULL
  valid_did_cohort <- any(did_cohort_index == 1) & any(did_cohort_index == 0)
  if(!isTRUE(valid_did_cohort)){
    if(dp2$print_details){cat("\n Skipping (g,t) as no treatment group or control group found")}
    return(NULL)
  }



  # key into the overlap/rcond guard cache (set on the panel non-varying branch
  # below when the cache is active; NULL = no caching for this cell)
  check_cache_key <- NULL

  if(dp2$panel){
    # Determine which weight period to use based on fix_weights
    use_rc_for_weights <- (!is.null(dp2$fix_weights) && dp2$fix_weights == "varying")

    if (use_rc_for_weights) {
      # fix_weights = "varying": stack into RC format with per-period weights
      n_units <- length(did_cohort_index)
      cohort_data <- data.table(
        D = rep(did_cohort_index, 2),
        y = c(dp2$outcomes_tensor[[pret]], dp2$outcomes_tensor[[t+tfac]]),
        post = rep(c(0L, 1L), each = n_units),
        i.weights = c(dp2$weights_tensor[[pret]], dp2$weights_tensor[[t+tfac]])
      )
      # Use earlier-period covariates for both halves — fix_weights only
      # changes weights, not the covariate conditioning set.
      # Use min(pret, t) to match the panel estimator's convention:
      # with base_period="universal", pret can be later than t for placebo cells.
      cov_early <- dp2$covariates_tensor[[base::min(pret, t)]]
      if (is.matrix(cov_early)) {
        covariates <- rbind(cov_early, cov_early)
      } else {
        covariates <- c(cov_early, cov_early)
      }
    } else {
      # Default or fixed weight options: use panel estimator with single weight vector
      if (is.null(dp2$fix_weights)) {
        # Default: weight from earlier of the two periods
        w_idx <- base::min(pret, t)
      } else if (dp2$fix_weights == "base_period") {
        w_idx <- dp2$.pret_by_group[g]
      } else if (dp2$fix_weights == "first_period") {
        w_idx <- 1L
      }
      # Plain named list of the cohort vectors (the tensors already hold vectors);
      # this avoids constructing/extracting a data.table per (g,t) cell, which the
      # profiler showed is a large share of the run time. run_DRDID()'s panel branch
      # consumes these as vectors, so the result is identical.
      cohort_data <- list(D = did_cohort_index,
                          y1 = dp2$outcomes_tensor[[t + tfac]],
                          y0 = dp2$outcomes_tensor[[pret]],
                          i.weights = dp2$weights_tensor[[w_idx]])
      covariates <- dp2$covariates_tensor[[base::min(pret, t)]]
      # Key for the overlap/rcond guard cache (see compute.att_gt2): under
      # nevertreated the cohort index depends only on g, so the guards'
      # (covariates, D) inputs are bit-identical across all cells of a group
      # sharing a covariate (and weight) period.
      if (!is.null(dp2$.check_cache)) {
        check_cache_key <- paste(g, base::min(pret, t), w_idx, sep = "_")
      }
    }
  } else {

    # post indicator reuses the pre-computed period mask (see compute.att_gt2)
    log_vec <- dp2$.period_masks[[t + tfac]]
    # convert TRUE/FALSE to 1/0 in place (fastest)
    set(dp2$time_invariant_data, j = "post", value = as.integer(log_vec))

    # Handle fix_weights for RC/unbalanced panel
    if (!is.null(dp2$fix_weights) && dp2$fix_weights %in% c("base_period", "first_period")) {
      if (dp2$fix_weights == "base_period") {
        target_period <- dp2$time_periods[dp2$.pret_by_group[g]]
      } else {
        target_period <- dp2$time_periods[1]
      }
      tid <- dp2$time_invariant_data
      # Memoize the weight lookup per target period (env set in compute.att_gt2):
      # target_period depends only on g for "base_period" and is constant for
      # "first_period", and tid's id/time/.w columns are never mutated in the
      # (g,t) loop, so fixed_w/na_w are identical across all cells sharing a
      # target period. Keying by as.character(target_period) memoizes the NA
      # case (group with no pre-period, varying base, pre-treatment cell)
      # under "NA" with unchanged semantics. The warning and the
      # did_cohort_index mutation below stay per cell.
      memo <- dp2$.fixed_w_memo
      memo_key <- as.character(target_period)
      cached <- if (!is.null(memo)) memo[[memo_key]] else NULL
      if (is.null(cached)) {
        # Build weight lookup from target period
        target_mask <- tid[[dp2$tname]] == target_period
        target_ids <- tid[[dp2$idname]][target_mask]
        target_ws <- tid[[".w"]][target_mask]
        # Map each observation's weight from the target period by integer id matching.
        # Equivalent to the previous named-character lookup (first match wins on any
        # duplicate id; unmatched ids -> NA) but avoids coercing every id to character.
        fixed_w <- target_ws[match(tid[[dp2$idname]], target_ids)]
        cached <- list(fixed_w = fixed_w, na_w = is.na(fixed_w))
        if (!is.null(memo)) memo[[memo_key]] <- cached
      }
      fixed_w <- cached$fixed_w
      # Exclude units not observed in target period by setting D to NA
      # (run_DRDID filters on !is.na(D))
      na_w <- cached$na_w
      if (any(na_w & !is.na(did_cohort_index))) {
        did_cohort_index[na_w] <- NA_integer_
        warning(paste0("Some units not observed in ", dp2$fix_weights,
                       " (period ", target_period, ") for group ",
                       dp2$treated_groups[g], " in time period ",
                       dp2$time_periods[t+tfac], ". These units are excluded."))
      }
      # Plain named list of the cohort vectors (avoids per-cell data.table
      # construction/extraction; run_DRDID()'s RC branch consumes these as vectors).
      cohort_data <- list(D = did_cohort_index, y = tid[[dp2$yname]], post = tid$post,
                          i.weights = fixed_w, .rowid = tid$.rowid)
    } else {
      cohort_data <- list(D = did_cohort_index, y = dp2$time_invariant_data[[dp2$yname]],
                          post = dp2$time_invariant_data$post,
                          i.weights = dp2$time_invariant_data[[".w"]],
                          .rowid = dp2$time_invariant_data$.rowid)
    }
    covariates <- dp2$covariates_matrix
  }

  # run estimation
  force_rc <- !is.null(dp2$fix_weights) && dp2$fix_weights == "varying" && dp2$panel
  did_result <- tryCatch(run_DRDID(cohort_data, covariates, dp2, g_val = dp2$treated_groups[g],
                                   t_val = dp2$time_periods[t + tfac],
                                   pret_val = dp2$time_periods[pret],
                                   force_rc = force_rc,
                                   check_cache_key = check_cache_key),
                         error = function(e) {
                           warning("Error computing internal 2x2 DiD for (g, t) = (", dp2$treated_groups[g], ", ", dp2$time_periods[t+tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
                           return(NULL)
                         })

  # When force_rc on balanced panel, the influence function has 2*n_units rows.
  # Half-split is safe here: cohort_data is explicitly stacked as
  # [all pre, all post] via rep(c(0L, 1L), each = n_units) in construction above.
  # The folded dense vector is converted to sparse triplets (if_i/if_x) with the
  # same is.na | != 0 scan the assembly previously applied to dense columns.
  if (force_rc && !is.null(did_result) && dp2$panel && !is.null(did_result$inf_func)) {
    inf <- did_result$inf_func
    n_half <- length(inf) %/% 2L
    inf_folded <- inf[1:n_half] + inf[(n_half + 1):(2L * n_half)]
    did_result$if_i <- which(is.na(inf_folded) | inf_folded != 0)
    did_result$if_x <- inf_folded[did_result$if_i]
    did_result$inf_func <- NULL
  }

  return(did_result)

}

#' @title Compute Group-Time Average Treatment Effects
#'
#' @description `compute.att_gt2` does the (g,t) cell computation and sends it to estimation,
#' then does all the post-processing after estimation
#'
#' @param dp2 A DIDparams object v2.0
#'
#' @return a list with length equal to the number of groups times the
#'  number of time periods; each element of the list contains an
#'  object that contains group-time average treatment effect as well
#'  as which group it is for and which time period it is for. It also exports
#'  the influence function which is used externally to compute
#'  standard errors.
#'
#' @keywords internal
#' @export
compute.att_gt2 <- function(dp2) {

  n <- dp2$id_count  # Total number of units
  # whether to compute influence functions (default TRUE; FALSE = point estimates only)
  do_inf <- is.null(dp2$compute_inffunc) || isTRUE(dp2$compute_inffunc)
  time_periods <- dp2$time_periods # tlist

  tlist.length <- if (dp2$base_period != "universal") length(time_periods) - 1L else length(time_periods)
  tfac <- if (dp2$base_period != "universal") 1L else 0L

  # Pre-compute cumulative cohort sizes for fast indexing in get_did_cohort_index
  if (dp2$panel) {
    dp2$.cohort_cum_sizes <- cumsum(dp2$cohort_counts$cohort_size)
  } else {
    # Pre-compute the loop-invariant row masks used by get_did_cohort_index()
    # and the post indicator in run_att_gt_estimation(): per-period membership,
    # the never-treated flag, and per-group treated flags. Each was previously
    # recomputed with full O(rows) scans for every (g,t) cell on the RC /
    # unbalanced-panel path. Memory: (T + n_groups + 1) logical vectors of
    # length nrow(time_invariant_data).
    dat <- dp2$time_invariant_data
    dp2$.period_masks <- lapply(time_periods, function(tp) dat[[dp2$tname]] == tp)
    dp2$.never_treated <- dat[[dp2$gname]] == Inf
    dp2$.gflag_by_group <- lapply(dp2$treated_groups, function(gval) dat[[dp2$gname]] == gval)
    # Memo for the fix_weights = "base_period"/"first_period" weight lookup in
    # run_att_gt_estimation(): fixed_w/na_w depend only on the target period
    # (one per group for "base_period", a single constant for "first_period"),
    # so the full-table mask + match() is computed once per distinct period
    # instead of once per (g,t) cell.
    if (!is.null(dp2$fix_weights) && dp2$fix_weights %in% c("base_period", "first_period")) {
      dp2$.fixed_w_memo <- new.env(parent = emptyenv())
    }
    # Pre-build the unit-aggregation operator for the unbalanced-panel
    # influence function: sparse 0/1 matrix S (id_count x rows) with
    # S[u, r] = 1 iff row r belongs to unit u, units ordered by first
    # appearance. S %*% v accumulates v in row order within each unit,
    # reproducing rowsum(v, .rowid, reorder = FALSE) bit-identically, while
    # the grouping work is done once per att_gt() call instead of once per
    # (g,t) cell (rowsum re-hashes the full .rowid vector and builds
    # character dimnames on every call).
    if (dp2$allow_unbalanced_panel && do_inf) {
      rowid_all <- dat$.rowid
      gmap <- match(rowid_all, unique(rowid_all))
      dp2$.rowid_agg <- Matrix::sparseMatrix(i = gmap, j = seq_along(rowid_all),
                                             x = 1, dims = c(n, length(rowid_all)))
    }
  }

  # Cache for the per-cell overlap/rcond guard booleans. For panel data with
  # control_group = "nevertreated" and non-varying weights, the guards'
  # (covariates, D) inputs are bit-identical across all cells of a group that
  # share a covariate (and weight) period, so each boolean is computed once per
  # (g, covariate-period, weight-period) key (see run_DRDID). Not used for
  # notyettreated (the control cohort varies with t) or fix_weights = "varying"
  # (RC estimator path). Only the booleans are stored; failure warnings are
  # still emitted per cell. options(did.disable_check_cache = TRUE) forces the
  # per-cell checks (escape hatch / used to verify bit-identical equivalence).
  if (dp2$panel && dp2$control_group[1] == "nevertreated" &&
      (is.null(dp2$fix_weights) || dp2$fix_weights != "varying") &&
      !isTRUE(getOption("did.disable_check_cache"))) {
    dp2$.check_cache <- new.env(parent = emptyenv())
  }


  # Pre-compute pret for each group (used for post-treatment periods and universal base)
  # This avoids calling tail(which(...)) repeatedly inside run_att_gt_estimation
  pret_by_group <- vapply(seq_len(dp2$treated_groups_count), function(g) {
    idx <- which((time_periods + dp2$anticipation) < dp2$treated_groups[g])
    if (length(idx) == 0L) NA_integer_ else idx[length(idx)]
  }, integer(1))
  dp2$.pret_by_group <- pret_by_group

  # in terms of indexes, not calendar times.
  # Build the (g,t) index pairs directly in g-major / t-minor order. This is
  # identical to expand.grid(g, t) followed by order(g, t) but avoids both the
  # data.frame construction and the per-iteration single-row subset below.
  g_vec <- rep.int(seq_len(dp2$treated_groups_count), rep.int(tlist.length, dp2$treated_groups_count))
  t_vec <- rep.int(seq_len(tlist.length), dp2$treated_groups_count)
  total_gt_iterations <- length(g_vec)


  # Running estimation using run_att_gt_estimation() function
  # Helper function for processing each (g,t) pair
  process_gt <- function(g, t) {

    # Run estimation
    gt_result <- run_att_gt_estimation(g, t, dp2)

      # Compute post-treatment indicator
    post.treat <- as.integer(dp2$treated_groups[g] <= dp2$time_periods[t+tfac])

    # Check for NULL first (estimation failed or was skipped). When do_inf = FALSE
    # (point estimates only) no influence-function triplets are stored (saving memory).
    # Failed cells propagate NA to every row of their column: if_i = seq_len(n),
    # if_x = NA, exactly the triplets the dense-column scan used to produce.
    if (is.null(gt_result)) {
      gt_result <- list(att = NA, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) seq_len(n) else NULL,
                        if_x = if (do_inf) rep(NA_real_, n) else NULL)
      return(gt_result)
    }

    # Base period normalization: ATT is 0 by construction (empty triplet column)
    if (!is.null(gt_result$base_period_norm)) {
      gt_result <- list(att = 0, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) integer(0) else NULL,
                        if_x = if (do_inf) numeric(0) else NULL)
      return(gt_result)
    }

    if (is.null(gt_result$att)) {
      # Estimation returned a result but without an ATT
      gt_result <- list(att = NA, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) seq_len(n) else NULL,
                        if_x = if (do_inf) rep(NA_real_, n) else NULL)
      return(gt_result)
    } else {
      att <- gt_result$att
      if_i <- gt_result$if_i
      if_x <- gt_result$if_x

      # Handle NaN ATT: treat as estimation failure
      if (is.nan(att)) {
        att <- NA
        if (do_inf) {
          if_i <- seq_len(n)
          if_x <- rep(NA_real_, n)
        }
      }

      # Save ATT and influence triplets (if_i/if_x are NULL when do_inf = FALSE)
      gt_result <- list(att = att, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat, if_i = if_i, if_x = if_x)
      return(gt_result)
    }
  }

  # run the estimation for each (g,t) pair with process_gt
  gt_results <- lapply(seq_len(total_gt_iterations), function(idx) {
    process_gt(g_vec[idx], t_vec[idx])
  })

  # Filter out NULL results
  gt_results <- Filter(Negate(is.null), gt_results)

  if (length(gt_results) == 0L) {
    stop("No valid (g, t) cells found for att_gt() estimation. ",
         "Check treatment timing, control group definition, and anticipation settings.")
  }

  # Post processing: Build sparse influence function matrix directly from triplets.
  # Skipped entirely for point estimates only (do_inf = FALSE) -- no n x k matrix is
  # ever formed, which is the memory saving of compute_inffunc = FALSE.
  if (!do_inf) {
    inffunc <- NULL
  } else {
    n_cols <- length(gt_results)

    # Each cell already carries its influence column as sparse triplets
    # (if_i row indices / if_x values), mirroring the slow path; concatenate
    # them once instead of rescanning dense per-cell vectors with which().
    # The triplet order is unchanged (column-major, original within-column
    # order), so the resulting sparse matrix is identical. Every code path
    # produces columns with dp2$id_count rows (panel units, RC observations,
    # unbalanced-panel units after the rowsum aggregation, force_rc units
    # after the half-fold).
    nz_list <- lapply(gt_results, `[[`, "if_i")
    val_list <- lapply(gt_results, `[[`, "if_x")
    trip_i <- unlist(nz_list, use.names = FALSE)
    trip_x <- unlist(val_list, use.names = FALSE)
    trip_j <- rep.int(seq_len(n_cols), lengths(nz_list))
    inffunc <- Matrix::sparseMatrix(i = trip_i, j = trip_j, x = trip_x,
                                    dims = c(n, n_cols))
  }

  return(list(attgt.list=gt_results, inffunc=inffunc))
}

#' @title Compute Group-Time Average Treatment Effects
#'
#' @description `compute.att_gt` does the main work for computing
#'  multiperiod group-time average treatment effects
#'
#' @param dp A DIDparams object
#'
#' @return a list with length equal to the number of groups times the
#'  number of time periods; each element of the list contains an
#'  object that contains group-time average treatment effect as well
#'  as which group it is for and which time period it is for. It also exports
#'  the influence function which is used externally to compute
#'  standard errors.
#'
#' @keywords internal
#'
#' @export
compute.att_gt <- function(dp) {
  #-----------------------------------------------------------------------------
  # unpack DIDparams
  #-----------------------------------------------------------------------------
  # Internal pretest hook (set only by conditional_did_pretest): when a
  # y-invariant setup bundle is supplied (dp$setup_precomp, built once per
  # pretest call via dp$return_setup below), the bundle replaces the whole
  # per-call setup, including this data.table copy, so skip it.
  setup_precomp <- dp$setup_precomp
  if (is.null(setup_precomp)) data <- data.table::as.data.table(dp$data)
  yname <- dp$yname
  tname <- dp$tname
  idname <- dp$idname
  xformla <- dp$xformla
  weightsname <- dp$weightsname
  est_method <- dp$est_method
  extra_args <- if (is.null(dp$extra_args)) list() else dp$extra_args
  base_period <- dp$base_period
  panel <- dp$panel
  true_repeated_cross_sections <- dp$true_repeated_cross_sections
  print_details <- dp$print_details
  control_group <- dp$control_group
  anticipation <- dp$anticipation
  gname <- dp$gname
  n <- dp$n
  nT <- dp$nT
  nG <- dp$nG
  tlist <- dp$tlist
  glist <- dp$glist

  #-----------------------------------------------------------------------------
  # main computations
  #-----------------------------------------------------------------------------

  # will populate with all att(g,t)
  attgt.list <- list()

  # place holder in lists
  counter <- 1

  # number of time periods
  tlist.length <- if (base_period != "universal") length(tlist) - 1L else length(tlist)
  tfac <- if (base_period != "universal") 1L else 0L

  # influence function dimensions
  inffunc_nrow <- n

  # list to collect sparse matrix updates (built into sparse matrix after loops)
  inffunc_updates <- list()
  # counter for keeping track of updates
  update_counter <- 1

  # never treated option
  nevertreated <- (control_group[1] == "nevertreated")

  fix_weights <- dp$fix_weights

  # intercept-only design (xformla = ~1, the default): the per-cell overlap and
  # rcond guards have closed forms (see overlap_check_fail / rcond_check_fail),
  # so the preliminary logit fit and Gram matrix can be skipped.
  intercept_only <- (xformla == ~1)

  # Cache for the per-cell overlap/rcond guard booleans. For panel data with
  # control_group = "nevertreated" and non-varying weights, the guards'
  # (covariates, G) inputs are bit-identical across all cells of a group that
  # share a covariate (and weight) period, so each boolean is computed once per
  # (g, covariate-period, weight-period) key. Not used for notyettreated (the
  # control cohort varies with t) or fix_weights = "varying" (RC estimator
  # path). Only the booleans are stored; failure warnings are still emitted per
  # cell. options(did.disable_check_cache = TRUE) forces the per-cell checks
  # (escape hatch / used to verify bit-identical equivalence). Mirrors the fast
  # path (compute.att_gt2).
  use_check_cache <- panel && nevertreated &&
    (is.null(fix_weights) || fix_weights != "varying") &&
    !isTRUE(getOption("did.disable_check_cache"))
  if (use_check_cache) check_cache <- new.env(parent = emptyenv())

  # Whether to compute influence functions (default TRUE; FALSE = point estimates
  # only). When FALSE we skip requesting influence functions from the 2x2 estimators,
  # skip the per-cell influence-function scaling, and skip storing inffunc_updates --
  # so a point-estimates-only run does not pay the influence-function compute/memory
  # cost. The att estimates are identical either way.
  do_inf <- is.null(dp$compute_inffunc) || isTRUE(dp$compute_inffunc)

  # Internal pretest-only flag (set by conditional_did_pretest): skip
  # post-treatment (g,t) cells entirely -- the pretest immediately discards
  # them (its keepers filter retains only group > t cells), so computing them
  # is pure waste. attgt.list and the influence-function columns stay aligned
  # with the reduced cell set because the counters only increment when a cell
  # is appended.
  pretreatment_only <- isTRUE(dp$pretreatment_cells_only)

  if (!is.null(setup_precomp)) {
    # ---- internal pretest path: reuse the y-invariant setup bundle -----------
    # conditional_did_pretest() calls compute.att_gt once per unit with only
    # the outcome column changed, so the whole per-call setup below (column
    # extraction, design matrix, per-period slicing, alignment check) is built
    # once via dp$return_setup and reused; each per-unit call passes only
    # dp$y_override (that unit's modified outcome, aligned 1:1 with dp$data
    # rows). The bundle is consumed only when the balanced, id-aligned panel
    # precompute applies (conditional_did_pretest keeps its legacy data-copy
    # path otherwise), and that path never reads `data` / `g_col` / `t_col`
    # inside the loops -- they are deliberately left undefined here so any
    # accidental read fails loudly instead of silently using a stale outcome.
    use_precompute_rc <- FALSE
    use_precompute_panel <- isTRUE(setup_precomp$use_precompute_panel)
    weights_by_period <- setup_precomp$weights_by_period
    period_mm <- setup_precomp$period_mm
    period_id <- setup_precomp$period_id
    period_w  <- setup_precomp$period_w
    g_unit_panel <- setup_precomp$g_unit_panel
    if (nevertreated) C_full <- setup_precomp$C_full
    # only the outcome differs across pretest iterations: rebuild the
    # per-period outcome slices from the override (bit-identical to slicing a
    # full data copy whose outcome column was overwritten, which is what the
    # legacy pretest loop did)
    y_override <- dp$y_override
    period_y <- lapply(setup_precomp$rows_by_period, function(rr) y_override[rr])
  } else {
    # Pre-extract columns to avoid repeated get() inside data.table (which is slow)
    g_col <- data[[gname]]
    t_col <- data[[tname]]

    # Build weight lookup by period for fix_weights options (balanced panel only)
    if (!is.null(fix_weights) && panel) {
      weights_by_period <- list()
      for (tp in seq_along(tlist)) {
        weights_by_period[[tp]] <- data[t_col == tlist[tp], .w]
      }
    }

    # Positional precompute for the repeated-cross-section / unbalanced-panel path:
    # each (g,t) cell is assembled from plain vectors and per-period row indices
    # precomputed once per call, instead of two full data.table subsets plus a
    # full-data %in% per cell (the dominant per-cell cost on this path). Bit-identical
    # to the legacy construction (see the per-cell comments below);
    # options(did.disable_precompute = TRUE) forces the legacy per-cell assembly
    # (escape hatch / used to verify bit-identical cell assembly; both sub-paths
    # still slice the hoisted global design matrix), mirroring the panel precompute.
    use_precompute_rc <- !panel && !isTRUE(getOption("did.disable_precompute"))

    # The precompute paths never read data$.C / data$.G (cohort indicators are
    # rebuilt from plain vectors), so skip the full-length column writes there.
    if (nevertreated && !use_precompute_rc) {
      set(data, j = ".C", value = as.integer(g_col == 0))
    }

    # -- design matrix: built ONCE over the full data, then row-subset per (g,t) cell --
    # This (a) avoids rebuilding model.matrix() nG*nT times, and (b) gives every cell the
    # GLOBAL factor levels / GLOBAL transform basis -- exactly what the repeated-cross-
    # section path, the fast path, and manually-constructed dummy columns use. As a
    # result a factor covariate behaves identically to adding its dummies by hand: a
    # level absent from a particular 2x2 comparison becomes an all-zero column (handled
    # by the estimator's overlap/rank check, giving the same NA + warning as the manual
    # dummies) rather than a dropped contrast that changes the design across cells.
    #
    # For purely row-wise numeric formulae (e.g. ~X, ~X1+X2, ~X1*X2, ~I(X^2)) the slice
    # is bit-identical to a per-cell model.matrix(xformla, disdat); for factors it equals
    # the manual-dummy design; for data-dependent bases (poly/scale/ns/...) it uses the
    # global basis, which yields the same ATT/SE as a per-cell fit (the estimator is
    # invariant to a full-rank reparameterization of the covariates).
    #
    # `data` is already complete-cases and finite (see pre_process_did), so na.pass adds
    # no NA rows and full_mm aligns 1:1 (by row position) with `data`.
    #
    # Drop globally-empty factor levels first: model.matrix() emits a dummy column for
    # every declared level, so a level absent from the ENTIRE estimation sample (common
    # after users subset their data) would produce an all-zero column in every cell and
    # NA-fail the whole estimation. Levels absent only from a particular 2x2 cell are
    # unaffected (still all-zero in that cell only, as documented above). Mirrors the
    # identical fix in get_did_tensors() (fast path).
    for (v in all.vars(xformla)) {
      if (is.factor(data[[v]])) set(data, j = v, value = droplevels(data[[v]]))
    }
    full_mm <- model.matrix(xformla, data = data, na.action = na.pass)
    if (panel) {
      # per-period row slices keyed by idname, for the earlier-period covariates that
      # get_wide_data() retains. (idname is used rather than .rowid because .rowid is
      # only created on the RC/unbalanced path; in a balanced panel idname is unique
      # within each period.)
      period_mm <- vector("list", length(tlist))
      period_id <- vector("list", length(tlist))
      period_y  <- vector("list", length(tlist))
      period_w  <- vector("list", length(tlist))
      # per-period row indices, kept for the pretest setup bundle (dp$return_setup)
      rows_by_period <- vector("list", length(tlist))
      y_col <- data[[yname]]
      w_col <- data[[".w"]]
      for (tp in seq_along(tlist)) {
        rows_tp <- which(t_col == tlist[tp])
        rows_by_period[[tp]] <- rows_tp
        period_mm[[tp]] <- full_mm[rows_tp, , drop = FALSE]
        period_id[[tp]] <- data[[idname]][rows_tp]
        period_y[[tp]]  <- y_col[rows_tp]
        period_w[[tp]]  <- w_col[rows_tp]
      }
      # Precompute fast path: when the panel is balanced with every period sharing the
      # same units in the same (id-sorted) row order, each 2x2 cell can be assembled by
      # position from these per-period blocks -- skipping the per-cell data[time_mask]
      # subset + get_wide_data() reshape (the dominant cost on the slow path).
      # fix_weights = "base_period"/"first_period" only changes WHICH period's weight
      # vector each cell uses (selected from weights_by_period below, mirroring the
      # fallback), so the precompute applies there too. Falls through to the original
      # get_wide_data() code, bit-identically, for fix_weights = "varying" (which needs
      # the long-format RC data) / unaligned panels.
      panel_units_aligned <- all(vapply(period_id, function(x)
        length(x) == length(period_id[[1L]]) && all(x == period_id[[1L]]), logical(1)))
      g_unit_panel <- g_col[which(t_col == tlist[1L])]
      # options(did.disable_precompute = TRUE) disables only this per-cell positional
      # assembly, falling back to the per-cell get_wide_data() reshape below (escape
      # hatch / used to verify bit-identical cell assembly). It does NOT bypass the
      # design-matrix hoist: BOTH sub-paths slice the global full_mm/period_mm built
      # above, so a hoist regression would be invisible to this option -- the hoist
      # itself is validated by the master-comparison tests in
      # test-modelmatrix-hoist.R, not by this escape hatch.
      use_precompute_panel <- (is.null(fix_weights) ||
        fix_weights %in% c("base_period", "first_period")) && panel_units_aligned &&
        !isTRUE(getOption("did.disable_precompute"))
      # Never-treated control cohort is invariant across every (g,t) cell, so build
      # it once here; notyettreated rebuilds C_full per cell inside the loops.
      if (use_precompute_panel && nevertreated) {
        C_full <- as.numeric(g_unit_panel == 0)
      }
    } else {
      # repeated cross-section / unbalanced: row-subset the full design by POSITION
      # (.rowid is not unique across periods here).
      global_mm <- full_mm
      use_precompute_panel <- FALSE
      if (use_precompute_rc) {
        # One-time positional precompute: per-period row indices into `data` plus
        # plain copies of the columns each cell consumes. which() preserves data
        # order within each period, so per-period vectors are bit-identical to the
        # corresponding data.table period subsets.
        period_rows_rc <- vector("list", length(tlist))
        for (tp in seq_along(tlist)) {
          period_rows_rc[[tp]] <- which(t_col == tlist[tp])
        }
        rowid_col <- data$.rowid
        y_col_rc <- data[[yname]]
        w_col_rc <- data[[".w"]]
        # loop-invariant id universe for the influence-function index lookup
        unique_rowid <- unique(rowid_col)
        # fix_weights base/first: per-period .rowid/.w lookup vectors replace the
        # per-cell full-table data[t_col == target_period, ] subset; match()
        # semantics (first match wins, unmatched -> NA) are unchanged.
        if (!is.null(fix_weights) && fix_weights %in% c("base_period", "first_period")) {
          target_rowid_by_period <- lapply(period_rows_rc, function(rr) rowid_col[rr])
          target_w_by_period <- lapply(period_rows_rc, function(rr) w_col_rc[rr])
        }
      }
    }
  }

  # Internal (set only by conditional_did_pretest): return the y-invariant
  # setup bundle built above -- so the pretest can hoist it out of its
  # per-unit loop -- instead of estimating. Balanced panel only; the pretest
  # keeps its legacy data-copy loop for repeated cross sections, and only
  # consumes the bundle when use_precompute_panel is TRUE.
  if (isTRUE(dp$return_setup)) {
    if (!panel) return(NULL)
    return(list(
      use_precompute_panel = use_precompute_panel,
      weights_by_period = if (!is.null(fix_weights)) weights_by_period else NULL,
      period_mm = period_mm,
      period_id = period_id,
      period_w = period_w,
      rows_by_period = rows_by_period,
      g_unit_panel = g_unit_panel,
      C_full = if (use_precompute_panel && nevertreated) C_full else NULL
    ))
  }

  # loop over groups
  for (g in 1:nG) {
    # Set up .G once (use pre-extracted g_col to avoid get() inside data.table).
    # The precompute paths derive the treated cohort from g_unit_panel / g_col
    # directly and never read data$.G, so skip the full-length column write there.
    current_g <- glist[g]
    if (!use_precompute_panel && !use_precompute_rc) {
      set(data, j = ".G", value = as.numeric(g_col == current_g))
    }

    # pre-compute the universal/post-treatment base period for this group (used multiple times)
    idx_g <- which((tlist + anticipation) < glist[g])
    pret_g <- if (length(idx_g) == 0L) NA_integer_ else idx_g[length(idx_g)]

    # Precompute path: the treated-cohort indicator depends only on g, so build it
    # once per group. Under nevertreated the control cohort is also t-invariant
    # (C_full was hoisted above the loops), so the kept-unit index and the G/C
    # vectors are hoisted here too; notyettreated recomputes C_full/disidx/kept
    # per cell below because its control cohort varies with t.
    if (use_precompute_panel) {
      G_full <- as.numeric(g_unit_panel == current_g)
      if (nevertreated) {
        disidx <- G_full == 1 | C_full == 1
        kept <- which(disidx)
        n1 <- length(kept)
        G <- G_full[kept]
        C <- C_full[kept]
      }
    }

    # loop over time periods
    for (t in 1:tlist.length) {
      #-----------------------------------------------------------------------------
      # Set pret

      # varying base period
      pret <- t

      # universal base period
      if (base_period == "universal") {
        # use same base period as for post-treatment periods
        pret <- pret_g
      }

      # check if in post-treatment period
      if ((glist[g] <= tlist[(t + tfac)])) {
        # update pre-period if in post-treatment period to
        # be  period (g-delta-1)
        pret <- pret_g
      }

      # check if there are no pre-treatment periods
      if (is.na(pret)) {
        warning(paste0("There are no pre-treatment periods for the group first treated at ", glist[g], "\nUnits from this group are dropped"))
        break
      }

      # pretest-only: skip post-treatment cells entirely (see
      # pretreatment_only above; conditional_did_pretest discards these
      # cells unread, so none of the per-cell work below is needed)
      if (pretreatment_only && glist[g] <= tlist[t + tfac]) next

      # use "not yet treated as control"
      # that is, never treated + units that are eventually treated,
      # but not treated by the current period (+ anticipation)
      if (!nevertreated) {
        # Use pre-extracted g_col to avoid get() inside data.table.
        # The precompute paths recompute the control cohort from
        # g_unit_panel / g_col and never read data$.C, so skip the full-length
        # column write there (the time_threshold scalar is still needed below).
        time_threshold <- tlist[max(t, pret) + tfac] + anticipation
        if (!use_precompute_panel && !use_precompute_rc) {
          set(data, j = ".C", value = as.integer((g_col == 0) |
            ((g_col > time_threshold) &
              (g_col != current_g))))
        }
      }


      #-----------------------------------------------------------------------------
      # if we are in period (g-1), normalize results to be equal to 0
      # and break without computing anything
      if (base_period == "universal") {
        if (tlist[pret] == tlist[(t + tfac)]) {
          attgt.list[[counter]] <- list(att = 0, group = glist[g], year = tlist[(t + tfac)], post = 0)
          # inffunc[,counter] <- rep(0,n)
          # counter <- counter+1
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = integer(0), # No non-zero entries
            values = numeric(0)
          )

          # Update the counters
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      }

      # print the details of which iteration we are on
      if (print_details) {
        cat(paste("current period:", tlist[(t + tfac)]), "\n")
        cat(paste("current group:", glist[g]), "\n")
        cat(paste("set pretreatment period to be", tlist[pret]), "\n")
      }

      #-----------------------------------------------------------------------------
      # results for the case with panel data
      #-----------------------------------------------------------------------------

      # post treatment dummy variable
      post.treat <- 1 * (glist[g] <= tlist[t + tfac])

      # total number of units (not just included in G or C)
      # disdat <- data[data[,tname] == tlist[t+tfac] | data[,tname] == tlist[pret],]
      # The precompute paths (panel per-period blocks; RC/unbalanced positional
      # assembly) build each cell without this long 2-period subset (the dominant
      # per-cell cost); the legacy escape-hatch paths still build disdat as before.
      # time_mask is only read here and in the fix_weights == "varying" branch,
      # which always runs with use_precompute_panel == FALSE (the panel precompute
      # excludes fix_weights == "varying") and panel == TRUE (so use_precompute_rc
      # is FALSE), so it is computed on every path that consumes it.
      if (!(panel && use_precompute_panel) && !use_precompute_rc) {
        target_times <- c(tlist[t + tfac], tlist[pret])
        time_mask <- t_col %in% target_times
        disdat <- data[time_mask]
      }


      if (panel) {
        if (use_precompute_panel) {
          # ---- precompute fast path (balanced, id-aligned panel; non-varying weights) --
          # Build the 2x2 cell directly from the per-period blocks. g is time-invariant
          # so .G / .C are unit-level; Ypost is always period (t+tfac) and Ypre always
          # period pret because get_wide_data's .y1/.y0 (later/earlier) swap exactly
          # cancels the universal-base ordering used below. The earlier period (whose
          # weights + covariates get_wide_data retains) is min(t + tfac, pret).
          # G_full (and, under nevertreated, C_full/disidx/kept/n1/G/C) are hoisted
          # to the group/setup level above; only t-varying pieces are built here.
          if (!nevertreated) {
            # time_threshold was computed above (kept for the precompute path)
            C_full <- as.numeric((g_unit_panel == 0) |
              ((g_unit_panel > time_threshold) & (g_unit_panel != current_g)))
            disidx <- G_full == 1 | C_full == 1
            kept <- which(disidx)
            n1 <- length(kept)
            G <- G_full[kept]
            C <- C_full[kept]
          }
          n <- length(g_unit_panel)
          Ypost <- period_y[[t + tfac]][kept]
          Ypre  <- period_y[[pret]][kept]
          earlier_idx <- min(t + tfac, pret)
          # Select weights based on fix_weights, mirroring the fallback's
          # weights_by_period[[...]][disidx] selection bit-identically:
          # weights_by_period[[tp]] and period_w[[tp]] share the same within-period
          # data order, and kept == which(disidx) under panel_units_aligned.
          if (is.null(fix_weights)) {
            # Default: the earlier period's weights (what get_wide_data retains)
            w <- period_w[[earlier_idx]][kept]
          } else if (fix_weights == "base_period") {
            w <- weights_by_period[[pret_g]][kept]
          } else {
            # fix_weights == "first_period" ("varying" never reaches the precompute)
            w <- weights_by_period[[1L]][kept]
          }
          covariates <- period_mm[[earlier_idx]][kept, , drop = FALSE]
          dimnames(covariates) <- list(NULL, colnames(covariates))
        } else {
          # ---- original path: per-cell get_wide_data() reshape ----
          # transform  disdat it into "cross-sectional" data where one of the columns
          # contains the change in the outcome over time.
          disdat <- get_wide_data(disdat, yname, idname, tname)

          # still total number of units (not just included in G or C)
          n <- nrow(disdat)

          # pick up the indices for units that will be used to compute ATT(g,t)
          disidx <- disdat$.G == 1 | disdat$.C == 1

          # pick up the data that will be used to compute ATT(g,t)
          disdat <- disdat[disidx]

          n1 <- nrow(disdat) # num obs. for computing ATT(g,t)

          # drop missing factors
          disdat <- droplevels(disdat)

          # give short names for data in this iteration
          G <- disdat$.G
          C <- disdat$.C

          # handle pre-treatment universal base period differently
          # we need to do this because panel2cs2 just puts later period
          # in .y1, but if we are in a pre-treatment period with a universal
          # base period, then the "base period" is actually the later period
          Ypre <- if (tlist[(t + tfac)] > tlist[pret]) disdat$.y0 else disdat$.y1
          Ypost <- if (tlist[(t + tfac)] > tlist[pret]) disdat$.y1 else disdat$.y0

          # Select weights based on fix_weights
          if (is.null(fix_weights)) {
            # Default: .w from get_wide_data (earlier period)
            w <- disdat$.w
          } else if (fix_weights == "base_period") {
            w <- weights_by_period[[pret_g]][disidx]
          } else if (fix_weights == "first_period") {
            w <- weights_by_period[[1L]][disidx]
          } else if (fix_weights == "varying") {
            w <- disdat$.w  # will be overridden below when switching to RC estimator
          } else {
            w <- disdat$.w
          }

          # matrix of covariates: index the precomputed earlier-period design by this
          # cell's units (idname). get_wide_data retained the earlier period, whose
          # index is min(t + tfac, pret).
          earlier_idx <- min(t + tfac, pret)
          covariates <- period_mm[[earlier_idx]][match(disdat[[idname]], period_id[[earlier_idx]]), , drop = FALSE]
          dimnames(covariates) <- list(NULL, colnames(covariates))
        }

        #-----------------------------------------------------------------------------
        # code for actually computing att(g,t)
        #-----------------------------------------------------------------------------

        attgt <- tryCatch({
          if (!is.null(fix_weights) && fix_weights == "varying") {
            # fix_weights = "varying": use RC estimators with per-period weights
            # Go back to long-format data for this (g,t) cell
            disdat_long <- data[time_mask]
            disdat_long_idx <- disdat_long$.G == 1 | disdat_long$.C == 1
            disdat_long <- droplevels(disdat_long[disdat_long_idx])
            Y_rc <- disdat_long[[yname]]
            G_rc <- disdat_long$.G
            post_rc <- as.numeric(disdat_long[[tname]] == tlist[t + tfac])
            w_rc <- disdat_long$.w
            # Use earlier-period covariates for all observations — fix_weights
            # only changes weights, not the covariate conditioning set.
            # Use min(pret, t+tfac) to match the panel estimator's convention:
            # with base_period="universal", pret can be later than t for placebo cells.
            earlier_idx_v <- min(pret, t + tfac)
            earlier_period <- tlist[earlier_idx_v]
            early_mask <- disdat_long[[tname]] == earlier_period
            disdat_early <- disdat_long[early_mask]
            cov_early <- period_mm[[earlier_idx_v]][match(disdat_early[[idname]], period_id[[earlier_idx_v]]), , drop = FALSE]
            dimnames(cov_early) <- list(NULL, colnames(cov_early))
            # Map each row in disdat_long to its unit's earlier-period covariates
            early_ids <- disdat_early[[idname]]
            all_ids <- disdat_long[[idname]]
            id_map <- match(all_ids, early_ids)
            covariates_rc <- cov_early[id_map, , drop = FALSE]

            # Run overlap/rank checks on RC data (not wide panel data)
            if (!is.function(est_method)) {
              if (est_method %in% c("dr", "ipw")) {
                if (overlap_check_fail(covariates_rc, G_rc, intercept_only)) {
                  warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                  stop("overlap")
                }
              }
              if (est_method %in% c("dr", "reg")) {
                if (rcond_check_fail(covariates_rc, G_rc, intercept_only)) {
                  warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                  stop("singular")
                }
              }
            }

            if (inherits(est_method, "function")) {
              res <- do.call(est_method, c(list(
                y = Y_rc, post = post_rc,
                D = G_rc, covariates = covariates_rc,
                i.weights = w_rc, inffunc = do_inf
              ), extra_args))
            } else if (est_method == "ipw") {
              res <- DRDID::std_ipw_did_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            } else if (est_method == "reg") {
              res <- DRDID::reg_did_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            } else {
              res <- DRDID::drdid_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            }
          } else {
            # Panel path: run overlap/rank checks on panel data. Guard booleans
            # are reused from the cache when this cell's (covariates, G) inputs
            # are bit-identical to an earlier cell's (nevertreated + non-varying
            # weights; see the cache setup above); warnings are still emitted
            # per cell, so cached failures keep counts and texts unchanged.
            if (!is.function(est_method)) {
              guard <- NULL
              if (use_check_cache) {
                w_idx_key <- if (is.null(fix_weights)) earlier_idx
                             else if (fix_weights == "base_period") pret_g
                             else 1L
                check_key <- paste(g, earlier_idx, w_idx_key, sep = "_")
                guard <- check_cache[[check_key]]
              }
              if (is.null(guard)) {
                # overlap check for pscore based methods; regression-feasibility
                # check for outcome-regression methods, run only if overlap
                # passed (matching the stop("overlap") short-circuit below)
                overlap_fail <- (est_method %in% c("dr", "ipw")) &&
                  overlap_check_fail(covariates, G, intercept_only)
                rcond_fail <- !overlap_fail && (est_method %in% c("dr", "reg")) &&
                  rcond_check_fail(covariates, G, intercept_only)
                guard <- list(overlap_fail = overlap_fail, rcond_fail = rcond_fail)
                if (use_check_cache) check_cache[[check_key]] <- guard
              }
              if (guard$overlap_fail) {
                warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                stop("overlap")
              }
              if (guard$rcond_fail) {
                warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                stop("singular")
              }
            }

            if (inherits(est_method, "function")) {
              # user-specified function
              res <- do.call(est_method, c(list(
                y1 = Ypost, y0 = Ypre,
                D = G,
                covariates = covariates,
                i.weights = w,
                inffunc = do_inf
              ), extra_args))
            } else if (est_method == "ipw") {
              # inverse-probability weights
              res <- DRDID::std_ipw_did_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            } else if (est_method == "reg") {
              # regression
              res <- DRDID::reg_did_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            } else {
              # doubly robust, this is default
              res <- DRDID::drdid_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            }
          }

          # adjust influence function to account for only using
          # subgroup to estimate att(g,t) (skipped for point estimates only)
          if (do_inf) {
            if (!is.null(fix_weights) && fix_weights == "varying") {
              # RC influence function has one entry per obs in disdat_long
              # (2 per unit: pre + post). Aggregate to unit level by ID,
              # independent of row ordering.
              res$att.inf.func <- as.numeric(rowsum(res$att.inf.func,
                                                    disdat_long[[idname]],
                                                    reorder = FALSE))
              res$att.inf.func <- (n / n1) * res$att.inf.func
            } else {
              res$att.inf.func <- (n / n1) * res$att.inf.func
            }
          }

          # If ATT is NaN, replace it with NA, and mark influence function as missing.
          if (is.nan(res$ATT)) {
            res$ATT <- NA
            if (do_inf) res$att.inf.func <- rep(NA_real_, length(res$att.inf.func))
          }
          res
        }, error = function(e) {
          # The overlap/rank guards above already emitted a full diagnostic warning
          # before signalling the internal "overlap"/"singular" sentinels, so skip
          # the wrapper warning for those: each failed cell warns exactly once,
          # matching the fast path. Genuine estimator errors are still surfaced.
          if (!(conditionMessage(e) %in% c("overlap", "singular"))) {
            warning("Error computing internal 2x2 DiD for (g, t) = (", glist[g], ", ", tlist[t + tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
          }
          NULL
        })

        if (is.null(attgt)) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      } else { # repeated cross sections / unbalanced panel

        if (use_precompute_rc) {
          # ---- positional precompute path (default) ----
          # .G / .C are row-wise functions of g_col, and gname is time-invariant
          # within .rowid (pre_process_did stops otherwise; for true repeated
          # cross sections .rowid is the row number), so a unit satisfies
          # .G == 1 | .C == 1 either in BOTH of this cell's periods or in
          # NEITHER. The legacy rightids/%in% construction below therefore
          # reduces to a row-wise G|C filter over the two periods, assembled
          # here positionally from the vectors precomputed before the loops --
          # skipping the two full data.table subsets, the full-data %in%, and
          # droplevels() per cell (covariates come positionally from global_mm,
          # so factor levels in `data` are never read). sort() restores original
          # data order, so the rows match which(disidx) on the legacy path
          # bit-identically. (The two periods are always distinct here: the
          # universal-base cell with tlist[pret] == tlist[t + tfac] is skipped
          # before this point.)
          rows_pair <- sort(c(period_rows_rc[[t + tfac]], period_rows_rc[[pret]]))
          g_cell <- g_col[rows_pair]
          Gv <- as.numeric(g_cell == current_g)
          # same formulas and types as the .G/.C column writes on the legacy path
          Cv <- if (nevertreated) {
            as.integer(g_cell == 0)
          } else {
            as.integer((g_cell == 0) |
              ((g_cell > time_threshold) & (g_cell != current_g)))
          }
          sel <- Gv == 1 | Cv == 1
          # positional index into `data` for slicing the global design matrix,
          # kept in sync with any later fix_weights row drops below.
          disdat_rows <- rows_pair[sel]
          G <- Gv[sel]
          C <- Cv[sel]
          Y <- y_col_rc[disdat_rows]
          post <- 1 * (t_col[disdat_rows] == tlist[t + tfac])
          cell_rowid <- rowid_col[disdat_rows]
        } else {
          # ---- legacy path (options(did.disable_precompute = TRUE)) ----
          # pick up the indices for units that will be used to compute ATT(g,t)
          # these conditions are (1) you are observed in the right period and
          # (2) you are in the right group (it is possible to be observed in
          # the right period but still not be part of the treated or control
          # group in that period here
          rightids <- disdat$.rowid[disdat$.G == 1 | disdat$.C == 1]

          # this is the fix for unbalanced panels; 2nd criteria shouldn't do anything
          # with true repeated cross sections, but should pick up the right time periods
          # only with unbalanced panel
          disidx <- (data$.rowid %in% rightids) & ((t_col == tlist[t + tfac]) | (t_col == tlist[pret]))

          # pick up the data that will be used to compute ATT(g,t)
          disdat <- data[disidx, ]
          # positional index into `data` for slicing the global design matrix, kept in
          # sync with any later row drops below (droplevels does not reorder/drop rows).
          # global_mm[disdat_rows, ] is the design for the rows used in this cell.
          disdat_rows <- which(disidx)

          # drop missing factors
          disdat <- droplevels(disdat)

          # give short names for data in this iteration
          G <- disdat$.G
          C <- disdat$.C
          Y <- disdat[[yname]]
          post <- 1 * (disdat[[tname]] == tlist[t + tfac])
          cell_rowid <- disdat$.rowid
        }
        # num obs. for computing ATT(g,t), have to be careful here
        n1 <- sum(G + C)

        # Handle fix_weights for RC/unbalanced panel
        if (!is.null(fix_weights) && fix_weights %in% c("base_period", "first_period")) {
          # Determine which period's weight to use
          if (fix_weights == "base_period") {
            target_idx <- pret_g
          } else {
            target_idx <- 1L
          }
          target_period <- tlist[target_idx]
          # Map each observation's weight from the target period by integer .rowid
          # matching (equivalent to the previous named-character lookup: first match
          # wins, unmatched -> NA), without coercing every .rowid to character.
          if (use_precompute_rc) {
            # per-period lookup vectors precomputed before the loops; the period
            # subsets preserve data order, so the match() result is identical to
            # the legacy full-table subset below.
            w <- target_w_by_period[[target_idx]][match(cell_rowid, target_rowid_by_period[[target_idx]])]
          } else {
            # Build lookup: weight from target period per unit
            target_rows <- data[t_col == target_period, ]
            w <- target_rows$.w[match(cell_rowid, target_rows$.rowid)]
          }
          # Drop units not observed in the target period
          missing_w <- is.na(w)
          if (any(missing_w)) {
            n_dropped <- length(unique(cell_rowid[missing_w]))
            warning(paste0("Dropped ", n_dropped, " units not observed in ",
                           fix_weights, " (period ", target_period, ") ",
                           "for group ", glist[g], " in time period ", tlist[t + tfac]))
            disdat_rows <- disdat_rows[!missing_w]
            cell_rowid <- cell_rowid[!missing_w]
            if (use_precompute_rc) {
              G <- G[!missing_w]
              C <- C[!missing_w]
              Y <- Y[!missing_w]
              post <- post[!missing_w]
            } else {
              disdat <- disdat[!missing_w, ]
              G <- disdat$.G
              C <- disdat$.C
              Y <- disdat[[yname]]
              post <- 1 * (disdat[[tname]] == tlist[t + tfac])
            }
            n1 <- sum(G + C)
            w <- w[!missing_w]
          }
        } else {
          w <- if (use_precompute_rc) w_col_rc[disdat_rows] else disdat$.w
        }

        #-----------------------------------------------------------------------------
        # checks to make sure that we have enough observations
        skip_this_att_gt <- FALSE
        if (sum(G * post) == 0) {
          warning(paste0("No units in group ", glist[g], " in time period ", tlist[t + tfac]))
          skip_this_att_gt <- TRUE
        }
        if (sum(G * (1 - post)) == 0) {
          # the empty period is the base (pre) period tlist[pret], not tlist[t]
          # (they differ under base_period = "universal" and for post-treatment
          # cells under "varying"); matches the fast path's pret_val
          warning(paste0("No units in group ", glist[g], " in time period ", tlist[pret]))
          skip_this_att_gt <- TRUE
        }
        if (sum(C * post) == 0) {
          warning(paste0("No available control units for group ", glist[g], " in time period ", tlist[t + tfac]))
          skip_this_att_gt <- TRUE
        }
        if (sum(C * (1 - post)) == 0) {
          # same base-period indexing fix as the group check above
          warning(paste0("No available control units for group ", glist[g], " in time period ", tlist[pret]))
          skip_this_att_gt <- TRUE
        }

        if (skip_this_att_gt) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          # inffunc[,counter] <- NA
          # counter <- counter+1
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )

          # Update the counters
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }


        # matrix of covariates: row-subset the global design by position.
        covariates <- global_mm[disdat_rows, , drop = FALSE]
        dimnames(covariates) <- list(NULL, colnames(covariates))

        #-----------------------------------------------------------------------------
        # code for actually computing att(g,t)
        #-----------------------------------------------------------------------------

        attgt <- tryCatch({
          # more checks for enough observations in each group; run inside the
          # tryCatch (with the same stop("overlap")/stop("singular") sentinels
          # as the panel branch) so that a failing preliminary fit degrades to
          # an NA cell with a warning instead of aborting the entire att_gt()
          if (!is.function(est_method)) {
            # checks for pscore based methods (no caching on the RC branch: the
            # cohort index varies cell by cell here)
            if (est_method %in% c("dr", "ipw")) {
              if (overlap_check_fail(covariates, G, intercept_only)) {
                warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                stop("overlap")
              }
            }

            # check if can run regression using control units
            if (est_method %in% c("dr", "reg")) {
              if (rcond_check_fail(covariates, G, intercept_only)) {
                warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                stop("singular")
              }
            }
          }

          if (inherits(est_method, "function")) {
            # user-specified function
            res <- do.call(est_method, c(list(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              inffunc = do_inf
            ), extra_args))
          } else if (est_method == "ipw") {
            # inverse-probability weights
            res <- DRDID::std_ipw_did_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          } else if (est_method == "reg") {
            # regression
            res <- DRDID::reg_did_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          } else {
            # doubly robust, this is default
            res <- DRDID::drdid_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          }

          # n/n1 adjusts for estimating the att_gt only using observations from
          # groups G and C (skipped for point estimates only)
          if (do_inf) res$att.inf.func <- (n / n1) * res$att.inf.func

          # If ATT is NaN, replace it with NA, and mark influence function as missing
          if (is.nan(res$ATT)) {
            res$ATT <- NA
            if (do_inf) res$att.inf.func <- rep(NA_real_, length(res$att.inf.func))
          }
          res
        }, error = function(e) {
          # The overlap/rank guards above already emitted a full diagnostic warning
          # before signalling the internal "overlap"/"singular" sentinels, so skip
          # the wrapper warning for those: each failed cell warns exactly once,
          # matching the fast path. Genuine estimator errors are still surfaced.
          if (!(conditionMessage(e) %in% c("overlap", "singular"))) {
            warning("Error computing internal 2x2 DiD for (g, t) = (", glist[g], ", ", tlist[t + tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
          }
          NULL
        })

        if (is.null(attgt)) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      } # end panel if

      # save results for this att(g,t)
      attgt.list[[counter]] <- list(
        att = attgt$ATT, group = glist[g], year = tlist[(t + tfac)], post = post.treat
      )


      # populate the influence function in the right places (skipped for point
      # estimates only -- compute_inffunc = FALSE -- so we neither aggregate nor
      # store the per-cell influence function)
      if (do_inf) {
      if (panel) {
        # Store integer indices directly (avoids which() later). On the precompute
        # path `kept` is exactly which(disidx) (disidx is not modified in between),
        # so reuse it; the get_wide_data fallback derives the index from its
        # wide-data disidx as before.
        idx <- if (use_precompute_panel) kept else which(disidx)
        inffunc_updates[[update_counter]] <- list(
          indices = idx,
          values = attgt$att.inf.func
        )
      } else {
        # aggregate inf functions by id (order by id)
        # Use the current cell ids (may differ from the initial selection if
        # fix_weights dropped obs). rowsum(reorder = TRUE) sums per id and returns
        # rows in ascending-id order, matching stats::aggregate() but ~40x faster
        # (see test-slowpath-rowsum).
        if (use_precompute_rc) {
          # every row of this cell satisfies .G == 1 | .C == 1 by construction
          # (the positional assembly selects on exactly that condition, and gname
          # is time-invariant within .rowid -- pre_process_did stops otherwise),
          # so the legacy G|C re-filter below is vacuous here. unique_rowid was
          # precomputed before the loops (loop-invariant).
          current_ids <- cell_rowid
          rs <- rowsum(attgt$att.inf.func, current_ids, reorder = TRUE)
          rs_ids <- sort(unique(current_ids))
          idx <- match(rs_ids, unique_rowid)
        } else {
          current_ids <- disdat$.rowid[disdat$.G == 1 | disdat$.C == 1]
          rs <- rowsum(attgt$att.inf.func, current_ids, reorder = TRUE)
          rs_ids <- sort(unique(current_ids))
          idx <- match(rs_ids, unique(data$.rowid))
        }
        inffunc_updates[[update_counter]] <- list(
          indices = idx,
          values = as.numeric(rs[, 1])
        )
      }
      }


      # save it in influence function matrix
      # inffunc[g,t,] <- inf.func
      # inffunc[,counter] <- inf.func
      update_counter <- update_counter + 1

      # update counter
      counter <- counter + 1
    } # end looping over t
  } # end looping over g

  # Build the influence function sparse matrix directly from triplets.
  # Gather per-column indices/values into lists and concatenate once, rather than
  # growing the triplet vectors with c() on every iteration (which reallocates
  # O(ncol^2) in total). The triplet order is unchanged (column-major, original
  # within-column order), so the resulting sparse matrix is identical.
  # Skipped for point estimates only (compute_inffunc = FALSE): no matrix is returned.
  if (do_inf) {
    nz_list <- lapply(inffunc_updates, `[[`, "indices")
    val_list <- lapply(inffunc_updates, `[[`, "values")
    trip_i <- unlist(nz_list, use.names = FALSE)
    trip_x <- unlist(val_list, use.names = FALSE)
    trip_j <- rep.int(seq_along(inffunc_updates), lengths(nz_list))
    inffunc <- Matrix::sparseMatrix(
      i = trip_i,
      j = trip_j,
      x = trip_x,
      dims = c(inffunc_nrow, length(inffunc_updates))
    )
  } else {
    inffunc <- NULL
  }

  return(list(attgt.list = attgt.list, inffunc = inffunc))
}

## ----setup2, include = FALSE--------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  echo = TRUE,
  eval = FALSE
)

## ----DGP_viz, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # Load libraries and set baseline parameters
# library(tidyverse)
# library(lfe)
# library(fastDummies)
# library(ggthemes)
# library(did)
# theme_set(theme_clean() + theme(plot.background = element_blank()))
# #----------------------------------------------------------------------------
# iseed  = 20201221
# nrep <- 100
# true_mu <- 1
# set.seed(iseed)
# 
# #----------------------------------------------------------------------------
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobserved heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into four treatment groups G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse(year >= cohort_year, 1, 0),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error)
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data()
# 
# # plot
# plot1 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value", color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))  +
#   ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with all groups being eventually treated")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot1
# 
# #ggsave("plot_dgp1.png", plot1, width = 10, height = 5, dpi = 100)

## ----ES1, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse(rel_year < -5, 1, 0),
#            Post = ifelse(rel_year > 5, 1, 0))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical <- map_dfr(1:nrep, run_ES_DiD)
# 
# colors <- c("True Effect" = "red", "Estimated Effect" = "blue")
# 
# ES_plot_classical <- data_classical %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#   geom_line(aes(color = 'Estimated Effect'), size = 1) +
#   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("TWFE event-study regression with binned end-points")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical
# 
# #ggsave("es_plot_classical.png", ES_plot_classical, width = 10, height = 5, dpi = 150)

## ----ES3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year")
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat <- map_dfr(1:nrep, run_ES_DiD_sat)
# 
# ES_plot_sat <- data_sat %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# 
# ES_plot_sat
# 
# #ggsave("es_plot_sat.png", ES_plot_sat, width = 10, height = 5, dpi = 150)
# 

## ----CS, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS <- function(...) {
# 
#   # resimulate the data
#   data <- make_data()
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS <- map_dfr(1:nrep, run_CS)
# 
# ES_plot_CS <- data_CS %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated")+
#     scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS
# 
# #ggsave("es_plot_CS.png", ES_plot_CS, width = 10, height = 5, dpi = 150)
# 

## ----DGP_viz2, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data2 <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobserved heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into 1\{t \ge g \}G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
#     # Relabel 2004 cohort as never-treated
#     mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data2()
# 
# # plot
# plot2 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value",  color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
#                      values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3"))+
#   ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with a never-treated group")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot2
# 
# #ggsave("plot_dgp_never_treated.png",plot2, width = 10, height = 5, dpi = 100)

## ----ES1_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>%
#     dummy_cols(select_columns = "rel_year") %>%
#     mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
#            Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
#     mutate(Pre = ifelse(is.na(Pre), 0, Pre),
#            Post = ifelse(is.na(Post), 0, Post))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#  # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical_never <- map_dfr(1:nrep, run_ES_DiD_never)
# 
# ES_plot_classical_never <- data_classical_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with binned end-points")+
#     scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical_never
# 
# #ggsave("es_plot_classical_never.png", ES_plot_classical_never, width = 10, height = 5, dpi = 150)

## ----ES3_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year2 = rel_year) %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     select(-("rel_year_-Inf"))
# 
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
#  # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat_never <- map_dfr(1:nrep, run_ES_DiD_sat_never)
# 
# ES_plot_sat_never <- data_sat_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_sat_never
# 
# #ggsave("es_plot_sat_never.png", ES_plot_sat_never, width = 10, height = 5, dpi = 150)
# 

## ----CS_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS_never <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "nevertreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_never <- map_dfr(1:nrep, run_CS_never)
# 
# ES_plot_CS_never <- data_CS_never %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_never
# 
# #ggsave("es_plot_CS_never.png", ES_plot_CS_never, width = 10, height = 5, dpi = 150)

## ----CS_ny, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# # function to run ES DID
# run_CS_ny <- function(...) {
# 
#   # resimulate the data
#   data <- make_data2()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_ny <- map_dfr(1:nrep, run_CS_ny)
# 
# ES_plot_CS_ny <- data_CS_ny %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_ny
# 
# #ggsave("es_plot_CS_ny.png", ES_plot_CS_ny, width = 10, height = 5, dpi = 150)

## ----DGP_viz3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# ## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
# make_data3 <- function(nobs = 1000,
#                       nstates = 40) {
# 
#   # unit fixed effects (unobserved heterogeneity)
#   unit <- tibble(
#     unit = 1:nobs,
#     # generate state
#     state = sample(1:nstates, nobs, replace = TRUE),
#     unit_fe = rnorm(nobs, state/5, 1),
#     # generate instantaneous treatment effect
#     #mu = rnorm(nobs, true_mu, 0.2)
#     mu = true_mu
#   )
# 
#   # year fixed effects (first part)
#   year <- tibble(
#     year = 1980:2010,
#     year_fe = rnorm(length(year), 0, 1)
#   )
# 
#   # Put the states into treatment groups
#   treat_taus <- tibble(
#     # sample the states randomly
#     state = sample(1:nstates, nstates, replace = FALSE),
#     # place the randomly sampled states into 1\{t \ge g \}G_g
#     cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
#   )
# 
#   # make main dataset
#   # full interaction of unit X year
#   expand_grid(unit = 1:nobs, year = 1980:2010) %>%
#     left_join(., unit) %>%
#     left_join(., year) %>%
#     left_join(., treat_taus) %>%
#     # make error term and get treatment indicators and treatment effects
#     # Also get cohort specific trends (modify time FE)
#     mutate(error = rnorm(nobs*31, 0, 1),
#            treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
#            mu = ifelse(cohort_year==1992, 2, ifelse(cohort_year==1998, 1, 3)),
#            tau = ifelse(treat == 1, mu, 0),
#            year_fe = year_fe + 0.1*(year - cohort_year)
#     ) %>%
#     # calculate cumulative treatment effects
#     group_by(unit) %>%
#     mutate(tau_cum = cumsum(tau)) %>%
#     ungroup() %>%
#     # calculate the dep variable
#     mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
#     # Relabel 2004 cohort as never-treated
#     mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
# 
# }
# #----------------------------------------------------------------------------
# # make data
# data <- make_data3()
# 
# # plot
# plot3 <- data %>%
#   ggplot(aes(x = year, y = dep_var, group = unit)) +
#   geom_line(alpha = 1/8, color = "grey") +
#   geom_line(data = data %>%
#               group_by(cohort_year, year) %>%
#               summarize(dep_var = mean(dep_var)),
#             aes(x = year, y = dep_var, group = factor(cohort_year),
#                 color = factor(cohort_year)),
#             size = 2) +
#   labs(x = "", y = "Value",  color = "Treatment group   ") +
#   geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) +
#   geom_vline(xintercept = 1992, color = '#377EB8', size = 2) +
#   geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) +
#   #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) +
#   scale_color_brewer(palette = 'Set1') +
#   theme(legend.position = 'bottom',
#         #legend.title = element_blank(),
#         axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12)) +
#   scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
#                      values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3")) +
#   ggtitle("One draw of the DGP with heterogeneous treatment effect dynamics across cohorts \n and with a never-treated group")+
#   theme(plot.title = element_text(hjust = 0.5, size = 12))
# 
# plot3
# 
# #ggsave("plot_dgp_never_treated_het.png", plot3, width = 10, height = 5, dpi = 100)

## ----ES1_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# # variables we will use
# keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
#               "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")
# 
# run_ES_DiD_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   # make dummy columns
#   data <- data %>%
#     # make dummies
#     mutate(rel_year = year - cohort_year) %>%
#     mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>%
#     dummy_cols(select_columns = "rel_year") %>%
#     mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>%
#     # generate pre and post dummies
#     mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
#            Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
#     mutate(Pre = ifelse(is.na(Pre), 0, Pre),
#            Post = ifelse(is.na(Post), 0, Post))
# 
#   # estimate the model
#   mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` +
#                 `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` +
#                 `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
# # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_classical_never_het <- map_dfr(1:nrep, run_ES_DiD_never_het)
# 
# ES_plot_classical_never_het <- data_classical_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with binned end-points")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_classical_never_het
# 
# #ggsave("es_plot_classical_never_het.png", ES_plot_classical_never_het, width = 10, height = 5, dpi = 150)

## ----ES3_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_ES_DiD_sat_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
# 
#   # make dummy columns
#   data <- data %>%
#     # make relative year indicator
#     mutate(rel_year = year - cohort_year)
# 
#   # get the minimum relative year - we need this to reindex
#   min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
# 
#   # reindex the relative years
#   data <- data %>%
#     mutate(rel_year2 = rel_year) %>%
#     mutate(rel_year = rel_year - min_year) %>%
#     dummy_cols(select_columns = "rel_year") %>%
#     select(-("rel_year_-Inf"))
# 
# 
#   # make regression formula
#   indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
#   keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")
#   formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
# 
#   # run mod
#   mod <- felm(formula, data = data, exactDOF = TRUE)
# 
#   # grab the obs we need
# # grab the obs we need
#   mod2 <- tibble(
#     estimate = mod$coefficients,
#     term1 = rownames(mod$coefficients)
#     )
# 
#  es <-
#    mod2 %>%
#     filter(term1 %in% keepvars) %>%
#     mutate(t = c(-5:-2, 0:5)) %>%
#     select(t, estimate)
#  es
# }
# 
# data_sat_never_het <- map_dfr(1:nrep, run_ES_DiD_sat_never_het)
# 
# ES_plot_sat_never_het <- data_sat_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("TWFE event-study regression with 'all' leads and lags")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_sat_never_het
# 
# #ggsave("es_plot_sat_never_het.png", ES_plot_sat_never_het, width = 10, height = 5, dpi = 150)
# 

## ----CS_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# # function to run ES DID
# run_CS_never_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "nevertreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_never_het <- map_dfr(1:nrep, run_CS_never_het)
# 
# ES_plot_CS_never_het <- data_CS_never_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_never_het
# 
# #ggsave("es_plot_CS_never_het.png", ES_plot_CS_never_het, width = 10, height = 5)

## ----CS_ny_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200----
# 
# # function to run ES DID
# run_CS_ny_het <- function(...) {
# 
#   # resimulate the data
#   data <- make_data3()
#   data$cohort_year[data$cohort_year==Inf] <- 0
# 
#   mod <- did::att_gt(yname = "dep_var",
#                      tname = "year",
#                      idname = "unit",
#                      gname = "cohort_year",
#                      control_group = "notyettreated",
#                      bstrap = FALSE,
#                      data = data,
#                      print_details = FALSE)
#   event_std <- did::aggte(mod, type = "dynamic")
# 
#   att.egt <- event_std$att.egt
#   names(att.egt) <- event_std$egt
# 
#   # grab the obs we need
#   broom::tidy(att.egt) %>%
#     filter(names %in% -5:5) %>%
#     mutate(t = -5:5, estimate = x) %>%
#     select(t, estimate)
# }
# 
# data_CS_ny_het <- map_dfr(1:nrep, run_CS_ny_het)
# 
# ES_plot_CS_ny_het <- data_CS_ny_het %>%
#   group_by(t) %>%
#   summarize(avg = mean(estimate),
#             sd = sd(estimate),
#             lower.ci = avg - 1.96*sd,
#             upper.ci = avg + 1.96*sd) %>%
#   mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>%
#   ggplot(aes(x = t, y = avg)) +
#   #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) +
#   geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
#   geom_point(color = 'blue', size = 3) +
#    geom_line(aes(color = 'Estimated Effect'), size = 1) +
#    geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) +
#   geom_hline(yintercept = 0, linetype = "dashed") +
#   scale_x_continuous(breaks = -5:5) +
#   labs(x = "Relative Time", y = "Estimate") +
#   theme(axis.title = element_text(size = 14),
#         axis.text = element_text(size = 12))+
#   ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
#   scale_color_manual(values = colors) +
#   theme(plot.title = element_text(hjust = 0.5, size = 12),
#         legend.position = "bottom",
#         legend.title = element_blank())
# 
# ES_plot_CS_ny_het
# 
# #ggsave("es_plot_CS_ny_het.png", ES_plot_CS_ny_het, width = 10, height = 5, dpi = 150)

#' @title Group-Time Average Treatment Effects
#'
#' @description `att_gt` computes average treatment effects in DID
#'  setups where there are more than two periods of data and allowing for
#'  treatment to occur at different points in time and allowing for
#'  treatment effect heterogeneity and dynamics.
#'  See Callaway and Sant'Anna (2021) for a detailed description.
#'
#' @param yname The name of the outcome variable
#' @param data The name of the data.frame that contains the data
#' @param tname The name of the column containing the time periods
#' @param idname The individual (cross-sectional unit) id name
#' @param gname The name of the variable in `data` that
#'  contains the first period when a particular observation is treated.
#'  This should be a positive number for all observations in treated groups.
#'  It defines which "group" a unit belongs to.  It should be 0 for units
#'  in the untreated group.
#' @param weightsname The name of the column containing the sampling weights.
#'  If not set, all observations have same weight. When weights are
#'  time-invariant (constant within each unit across periods), all
#'  \code{fix_weights} options produce identical results and no special
#'  handling is needed.
#'
#'  When weights vary across time (e.g., time-varying population sizes),
#'  the default behavior differs by panel type:
#'  \describe{
#'    \item{Balanced panel}{Each 2x2 DiD comparison uses the weight from the
#'      earlier of the two time periods involved. For post-treatment cells,
#'      this is the base period (g-1). For pre-treatment cells with
#'      \code{base_period="varying"}, this is the pre-treatment period itself.
#'      The panel DRDID estimators are used.}
#'    \item{Repeated cross sections and unbalanced panels}{Both periods'
#'      per-observation weights are passed directly to the RC DRDID estimators,
#'      so each observation carries its own period-specific weight.}
#'  }
#'  Use the \code{fix_weights} argument to override the default behavior.
#' @param fix_weights Controls how time-varying sampling weights are resolved.
#'  Only relevant when weights vary across time; with time-invariant weights,
#'  all options produce identical results. Options:
#'  \describe{
#'    \item{\code{NULL} (default)}{For balanced panel: uses the weight from
#'      the earlier of the two time periods in each 2x2 comparison. For
#'      post-treatment cells, this is the base period (g-1). For
#'      pre-treatment cells, this depends on the \code{base_period} setting.
#'      For RC/unbalanced panel: uses per-observation weights from both
#'      periods.}
#'    \item{\code{"varying"}}{Uses per-observation, period-specific weights
#'      for all panel types. For balanced panel data, this switches to the
#'      repeated cross-section DRDID estimators so that pre-period and
#'      post-period observations each carry their own weight. Covariates
#'      are held fixed at their pre-period values (same as the default
#'      panel estimator). This is the most flexible option for weights but
#'      sacrifices the efficiency of the panel estimator. For RC/unbalanced
#'      panel, this is identical to the default. Not supported with custom
#'      \code{est_method} functions.}
#'    \item{\code{"base_period"}}{Fixes weights at the base period (g-1) for
#'      all (g,t) cells within a group, for both pre-treatment and
#'      post-treatment comparisons. Ensures all cells within a group use the
#'      same weights. For unbalanced panels, units not observed in the base
#'      period are dropped with a warning. Not supported for repeated cross
#'      sections (\code{panel = FALSE}).}
#'    \item{\code{"first_period"}}{Fixes weights at the first time period in
#'      the dataset for all (g,t) cells. For unbalanced panels, units not
#'      observed in the first period are dropped with a warning. Not supported
#'      for repeated cross sections (\code{panel = FALSE}).}
#'  }
#' @param alp the significance level, default is 0.05
#' @param bstrap Boolean for whether or not to compute standard errors using
#'  the multiplier bootstrap.  Default is `TRUE` (in addition, cband
#'  is also by default `TRUE` indicating that uniform confidence bands
#'  will be returned).  If `bstrap=FALSE`, analytical standard errors are
#'  reported; these are cluster-robust when `clustervars` is supplied.
#' @param biters The number of bootstrap iterations to use.  The default is 1000,
#'  and this is only applicable if `bstrap=TRUE`.
#' @param clustervars A vector of variables names to cluster on.  At most, there
#'  can be two variables (otherwise will throw an error) and one of these
#'  must be the same as idname which allows for clustering at the individual
#'  level. Clustered standard errors are available with the multiplier bootstrap
#'  (`bstrap=TRUE`) or analytically (`bstrap=FALSE`).
#' @param cband Boolean for whether or not to compute a uniform confidence
#'  band that covers all of the group-time average treatment effects
#'  with fixed probability `1-alp`.  In order to compute uniform confidence
#'  bands, `bstrap` must also be set to `TRUE`.  The default is
#' `TRUE`.
#' @param print_details Whether or not to show details/progress of computations.
#'   Default is `FALSE`.
#' @param pl Whether or not to use parallel processing
#' @param cores The number of cores to use for parallel processing
#' @param compute_inffunc Whether or not to compute the influence functions. The
#'  default is `TRUE`. The influence functions are required for standard errors,
#'  uniform confidence bands, the parallel-trends pre-test, and for aggregating the
#'  group-time effects with [aggte()]. Set `compute_inffunc = FALSE` to obtain the
#'  group-time ATT *point estimates only*: this is faster and uses substantially less
#'  memory (no \eqn{n \times k} influence-function matrix is formed or bootstrapped),
#'  but the returned object has no standard errors / pre-test and cannot be passed to
#'  [aggte()]. The point estimates are identical to those from a full run. When
#'  `compute_inffunc = FALSE`, `bstrap` and `cband` are set to `FALSE` automatically.
#' @param est_method the method to compute group-time average treatment effects.  The default is "dr" which uses the doubly robust
#' approach in the `DRDID` package.  Other built-in methods
#' include "ipw" for inverse probability weighting and "reg" for
#' first step regression estimators.  The user can also pass their
#' own function for estimating group time average treatment
#' effects.  The required signature depends on the data structure:
#'
#' **Panel data** (`panel=TRUE`): `f(y1, y0, D, covariates,
#' i.weights, inffunc, ...)` where `y1` is an `n x 1` vector of
#' post-treatment outcomes, `y0` is an `n x 1` vector of
#' pre-treatment outcomes, `D` is a binary vector indicating
#' treatment group membership, `covariates` is an `n x k` matrix,
#' `i.weights` is a vector of sampling weights, and `inffunc` is a
#' logical requesting influence-function computation.
#'
#' **Repeated cross sections / unbalanced panel** (`panel=FALSE`):
#' `f(y, post, D, covariates, i.weights, inffunc, ...)` where `y` is
#' the outcome vector (length `n`), `post` is a binary indicator for
#' the post-treatment period, `D` is a binary treatment indicator,
#' `covariates` is an `n x k` matrix, `i.weights` is a vector of
#' sampling weights, and `inffunc` is a logical.
#'
#' In both cases the function should return a list that includes
#' `ATT` (the estimated group-time average treatment effect) and
#' `att.inf.func` (an `n x 1` influence function — one entry per
#' observation passed into the estimator).
#' The function can return other things as well, but these are
#' the only two that are required. With no covariates
#' (`xformla = NULL`), the built-in methods ("dr", "ipw", "reg") all
#' reduce to the unconditional difference-in-differences estimator, so
#' the choice among them is irrelevant; a custom `est_method` function
#' is still called (with an intercept-only `covariates` matrix) and
#' determines the estimates.
#' @param xformla A formula for the covariates to include in the
#'  model.  It should be of the form `~ X1 + X2`.  Default
#'  is NULL which is equivalent to `xformla=~1`.  This is
#'  used to create a matrix of covariates which is then passed
#'  to the 2x2 DID estimator chosen in `est_method`.
#'
#'  For time-varying covariates: (1) With balanced panel data,
#'  in each 2x2 comparison, the covariates
#'  are taken to be the value of the covariates in the earlier time
#'  period, and all of the underlying computations involve changes in Y
#'  as a function of those values of covariates.  (2) With repeated cross
#'  sections data and unbalanced panel data, the covariates are taken
#'  from each time period and computations involve Y_post conditional
#'  on X_post minus Y_pre conditional on X_pre.  A byproduct of this
#'  is that, with balanced panel data and in the presence of
#'  time-varying covariates, it is possible to get different numerical
#'  results according to whether or not `allow_unbalanced_panel=TRUE` or
#'  `FALSE`.
#' @param panel Whether or not the data is a panel dataset.
#'  The panel dataset should be provided in long format -- that
#'  is, where each row corresponds to a unit observed at a
#'  particular point in time.  The default is TRUE.  When
#'  using a panel dataset, the variable `idname` must
#'  be set.  When `panel=FALSE`, the data is treated
#'  as repeated cross sections.
#' @param allow_unbalanced_panel Whether or not function should
#'  "balance" the panel with respect to time and id.  The default
#'  value is `FALSE` which means that [att_gt()] will drop
#'  all units where data is not observed in all periods.
#'  The advantage of this is that the computations are faster
#'  (sometimes substantially).
#' @param control_group Which units to use as the control group.
#'  The default is "nevertreated" which sets the control group
#'  to be the group of units that never participate in the
#'  treatment.  This group does not change across groups or
#'  time periods.  The other option is to set
#'  `group="notyettreated"`.  In this case, the control group
#'  is set to the group of units that have not yet participated
#'  in the treatment in that time period.  This includes all
#'  never treated units, but it includes additional units that
#'  eventually participate in the treatment, but have not
#'  participated yet.
#' @param anticipation The number of time periods before participating
#'  in the treatment where units can anticipate participating in the
#'  treatment and therefore it can affect their untreated potential outcomes
#' @param faster_mode This option enables a faster version of `did`, optimizing
#' computation time for large datasets by improving data management within the package.
#' The default is set to `TRUE`. Both modes produce identical results up to
#' numerical precision; while the difference is minimal for small datasets,
#' the speedup is substantial for large ones.
#' @param base_period Whether to use a "varying" base period or a
#'  "universal" base period.  Either choice results in the same
#'  post-treatment estimates of ATT(g,t)'s.  In pre-treatment
#'  periods, using a varying base period amounts to computing a
#'  pseudo-ATT in each treatment period by comparing the change
#'  in outcomes for a particular group relative to its comparison
#'  group in the pre-treatment periods (i.e., in pre-treatment
#'  periods this setting computes changes from period t-1 to period
#'  t, but repeatedly changes the value of t)
#'
#'  A universal base period fixes the base period to always be
#'  (g-anticipation-1).  This does not compute
#'  pseudo-ATT(g,t)'s in pre-treatment periods, but rather
#'  reports average changes in outcomes from period t to
#'  (g-anticipation-1) for a particular group relative to its comparison
#'  group.  This is analogous to what is often reported in event
#'  study regressions.
#'
#'  Using a varying base period results in an estimate of
#'  ATT(g,t) being reported in the period immediately before
#'  treatment.  Using a universal base period normalizes the
#'  estimate in the period right before treatment (or earlier when
#'  the user allows for anticipation) to be equal to 0, but one
#'  extra estimate in an earlier period.
#'
#' @param ... Additional arguments to be passed to a custom `est_method`
#'  function. These are ignored when using built-in estimation methods
#'  (`"dr"`, `"ipw"`, `"reg"`).
#'
#' @section Options:
#'  When `faster_mode = FALSE`, setting `options(did.disable_precompute = TRUE)`
#'  disables the one-time positional precompute and assembles every 2x2
#'  comparison from the long data per cell, as in earlier versions. The
#'  covariate design matrix is still built once over the full sample under
#'  both settings, so results are identical either way; the option exists
#'  only as a debugging escape hatch.
#'
#' @references Callaway, Brantly and Pedro H.C. Sant'Anna.  \"Difference-in-Differences with Multiple Time Periods.\" Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. \doi{10.1016/j.jeconom.2020.12.001}, <https://arxiv.org/abs/1803.09015>
#'
#' @return an [`MP`] object containing all the results for group-time average
#'  treatment effects
#'
#'  The returned `inffunc` matrix collects the estimated influence functions,
#'  with one column per ATT(g,t) and one row per cross-sectional unit (one row
#'  per observation with repeated cross sections). Its rownames hold the unit
#'  ids (`idname`; an internal observation index for repeated cross sections)
#'  and are the authoritative link between rows and units: the row ORDER is
#'  mode-specific (`faster_mode = FALSE` sorts units by id, while
#'  `faster_mode = TRUE` uses an internal (period, cohort, id) ordering), so
#'  external consumers of the influence functions must align rows by rowname,
#'  never by position.
#'
#' @details # Examples:
#'
#' **Basic [att_gt()] call:**
#' ```{r, comment = "#>", collapse = TRUE}
#' # Example data
#' data(mpdta)
#' set.seed(09152024)
#' out1 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=NULL,
#'                data=mpdta)
#' summary(out1)
#' ```
#'
#' **Using covariates:**
#'
#' ```{r, comment = "#>", collapse = TRUE}
#' out2 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=~lpop,
#'                data=mpdta)
#' summary(out2)
#' ```
#'
#' **Specify comparison units:**
#'
#' ```{r, comment = "#>", collapse = TRUE}
#' out3 <- att_gt(yname="lemp",
#'                tname="year",
#'                idname="countyreal",
#'                gname="first.treat",
#'                xformla=~lpop,
#'                control_group = "notyettreated",
#'                data=mpdta)
#' summary(out3)
#' ```
#'
#' @export


att_gt <- function(yname,
                   tname,
                   idname = NULL,
                   gname,
                   xformla = NULL,
                   data,
                   panel = TRUE,
                   allow_unbalanced_panel = FALSE,
                   control_group = c("nevertreated", "notyettreated"),
                   anticipation = 0,
                   weightsname = NULL,
                   fix_weights = NULL,
                   alp = 0.05,
                   bstrap = TRUE,
                   cband = TRUE,
                   biters = 1000,
                   clustervars = NULL,
                   est_method = "dr",
                   base_period = "varying",
                   faster_mode = TRUE,
                   print_details = FALSE,
                   pl = FALSE,
                   cores = 1,
                   compute_inffunc = TRUE,
                   ...) {
  # Capture extra arguments for custom est_method
  extra_args <- list(...)

  if (missing(control_group)) control_group <- "nevertreated"
  validate_choice_scalar(
    control_group,
    "control_group",
    c("nevertreated", "notyettreated"),
    "control_group must be either 'nevertreated' or 'notyettreated'"
  )
  validate_choice_scalar(
    base_period,
    "base_period",
    c("universal", "varying"),
    "base_period must be either 'universal' or 'varying'."
  )

  # Validate compute_inffunc (point-estimates-only switch)
  if (!is.logical(compute_inffunc) || length(compute_inffunc) != 1 || is.na(compute_inffunc)) {
    stop("compute_inffunc must be a single logical (TRUE or FALSE).")
  }
  validate_logical_scalar(panel, "panel")
  validate_logical_scalar(allow_unbalanced_panel, "allow_unbalanced_panel")
  validate_logical_scalar(bstrap, "bstrap")
  validate_logical_scalar(cband, "cband")
  validate_logical_scalar(faster_mode, "faster_mode")
  validate_logical_scalar(print_details, "print_details")
  validate_logical_scalar(pl, "pl")
  validate_positive_whole_number(cores, "cores")
  validate_anticipation(anticipation)
  validate_alp(alp)
  # When influence functions are not computed there are no standard errors, no
  # uniform bands, and no parallel-trends pre-test, so the bootstrap is moot.
  if (!compute_inffunc) {
    bstrap <- FALSE
    cband <- FALSE
  }

  # Warn if extra arguments passed with built-in est_method
  if (length(extra_args) > 0 && !inherits(est_method, "function")) {
    warning("Extra arguments (", paste(names(extra_args), collapse = ", "),
            ") are ignored when using built-in est_method = \"", est_method,
            "\". Extra arguments are only passed to custom est_method functions.")
  }

  # Validate fix_weights
  if (!is.null(fix_weights)) {
    if (!is.character(fix_weights) || length(fix_weights) != 1 ||
        !(fix_weights %in% c("varying", "base_period", "first_period"))) {
      stop("fix_weights must be NULL or one of \"varying\", \"base_period\", or \"first_period\".")
    }
    if (!panel && fix_weights %in% c("base_period", "first_period")) {
      stop("fix_weights = \"", fix_weights, "\" is not supported for repeated cross sections ",
           "(panel = FALSE) because units are not tracked across periods. ",
           "Use fix_weights = \"varying\" or NULL instead.")
    }
    if (fix_weights == "varying" && panel && inherits(est_method, "function")) {
      stop("fix_weights = \"varying\" is not currently supported with custom est_method functions ",
           "on panel data. The \"varying\" option uses repeated cross-section estimators internally, ",
           "which require a different function signature (y, post, D) than the panel signature ",
           "(y1, y0, D). Use fix_weights = NULL, \"base_period\", or \"first_period\" instead.")
    }
  }

  # Validate est_method
  if (!inherits(est_method, "function")) {
    if (!is.character(est_method) || length(est_method) != 1) {
      stop("est_method must be a character string (\"dr\", \"ipw\", or \"reg\") or a custom function. ",
           "Received an object of class '", class(est_method)[1], "'.")
    }
    if (!(est_method %in% c("dr", "ipw", "reg"))) {
      stop("est_method must be one of \"dr\", \"ipw\", or \"reg\". Received: \"", est_method, "\".")
    }
  }

  # Require idname for panel data: without it the pre-processors fail with
  # confusing internal errors instead of pointing at the missing argument
  if (panel && is.null(idname)) {
    stop("Must provide idname when panel = TRUE. Set panel = FALSE for repeated cross sections.")
  }

  # Validate biters (number of bootstrap iterations) when the bootstrap is used
  if (bstrap) {
    validate_positive_whole_number(biters, "biters")
  }

  # Warn users about anticipation and never-treated units
  if (anticipation > 0) {
    message("Note: anticipation = ", anticipation, ". Never-treated units (with group status 0 or Inf) ",
            "are assumed to never anticipate treatment. Anticipation only applies to eventually-treated units.")
  }

  # Check if user wants to run faster mode:
  if (faster_mode) {
    # this is a DIDparams2 object
    dp <- pre_process_did2(
      yname = yname,
      tname = tname,
      idname = idname,
      gname = gname,
      xformla = xformla,
      data = data,
      panel = panel,
      allow_unbalanced_panel = allow_unbalanced_panel,
      control_group = control_group,
      anticipation = anticipation,
      weightsname = weightsname,
      fix_weights = fix_weights,
      alp = alp,
      bstrap = bstrap,
      cband = cband,
      biters = biters,
      clustervars = clustervars,
      est_method = est_method,
      base_period = base_period,
      print_details = print_details,
      faster_mode = faster_mode,
      pl = pl,
      cores = cores,
      call = match.call()
    )

    # attach extra args for custom est_method
    dp$extra_args <- extra_args
    # whether to compute influence functions (FALSE = point estimates only)
    dp$compute_inffunc <- compute_inffunc

    #-----------------------------------------------------------------------------
    # Compute all ATT(g,t)
    #-----------------------------------------------------------------------------
    results <- compute.att_gt2(dp)
  } else {
    # this is a DIDparams object
    dp <- pre_process_did(
      yname = yname,
      tname = tname,
      idname = idname,
      gname = gname,
      xformla = xformla,
      data = data,
      panel = panel,
      allow_unbalanced_panel = allow_unbalanced_panel,
      control_group = control_group,
      anticipation = anticipation,
      weightsname = weightsname,
      fix_weights = fix_weights,
      alp = alp,
      bstrap = bstrap,
      cband = cband,
      biters = biters,
      clustervars = clustervars,
      est_method = est_method,
      base_period = base_period,
      print_details = print_details,
      pl = pl,
      cores = cores,
      call = match.call()
    )

    # attach extra args for custom est_method
    dp$extra_args <- extra_args
    # whether to compute influence functions (FALSE = point estimates only)
    dp$compute_inffunc <- compute_inffunc

    #-----------------------------------------------------------------------------
    # Compute all ATT(g,t)
    #-----------------------------------------------------------------------------
    results <- compute.att_gt(dp)
  }

  # extract ATT(g,t) and influence functions
  attgt.list <- results$attgt.list
  inffunc <- results$inffunc

  # Label the influence-function rows with the internal unit ids (the user's idname
  # for panels; ".rowid", one per observation, for repeated cross sections). The ROW
  # ORDER of inffunc is mode-specific: faster_mode = FALSE sorts units by id, while
  # faster_mode = TRUE keeps the internal (period, cohort, id) ordering -- so row
  # POSITIONS are not comparable across modes. The rownames attached here are the
  # authoritative unit labels; external consumers (e.g. sensitivity analyses or
  # custom cluster aggregation) must align rows by rowname, never by position.
  # Values are unchanged; only dimnames are added.
  if (!is.null(inffunc) && !is.null(dp$idname)) {
    inffunc_ids <- if (faster_mode) dp$time_invariant_data[[dp$idname]] else dp$data[[dp$idname]]
    # collapse per-observation ids (panel: one per period; unbalanced panel: one per
    # observation) to one label per influence-function row; unique() preserves the
    # first-appearance order, which is exactly each mode's inffunc row order.
    if (anyDuplicated(inffunc_ids)) inffunc_ids <- unique(inffunc_ids)
    if (length(inffunc_ids) == nrow(inffunc)) rownames(inffunc) <- as.character(inffunc_ids)
  }

  # process results
  # attgt.results <- process_attgt(attgt.list)
  tryCatch(
    {
      # Attempt to run this line for process results
      attgt.results <- process_attgt(attgt.list)
    },
    error = function(e) {
      # Handle the error
      if (faster_mode) {
        # If faster_mode is TRUE, send this stop message
        stop("An unexpected error occurred, normally associated with a singular matrix due to not enough control units. Try changing faster_mode=FALSE.")
      } else {
        # If faster_mode is FALSE, send this stop message
        stop("An unexpected error occurred, normally associated with a singular matrix due to not enough control units.")
      }
    }
  )
  group <- attgt.results$group
  att <- attgt.results$att
  tt <- attgt.results$tt



  # analytical standard errors
  # estimate variance. The i.i.d. form is clustered at the unit level; with a coarser cluster variable
  # the variance is formed from cluster sums instead (the cluster_analytic branch below).
  n <- if (faster_mode) dp$id_count else dp$n

  # All inference below requires the influence functions. With compute_inffunc = FALSE
  # (point estimates only) skip it entirely; see the else-branch before return().
  if (compute_inffunc) {
  # Analytical variance of the ATT(g,t)'s. With a cluster variable (beyond idname) and no bootstrap, use
  # the cluster-robust form -- the cluster sums of the influence function -- mirroring the cluster-sum
  # aggregation of the multiplier bootstrap (Callaway & Sant'Anna 2021, Remark 10). It is kept on the same
  # 1/n scaling as the i.i.d. V so that se = sqrt(diag(V)/n) and the Wald statistic n * b' V^{-1} b stay
  # valid for either form.
  # cluster variables beyond the unit id; clustering on the unit id alone is just the unit-level (i.i.d.)
  # SE. Use the *internal* unit id dp$idname -- the user's idname for panels, and ".rowid" for repeated
  # cross-sections / unbalanced panels (where the user may omit idname), mirroring how mboot() identifies units.
  unit_id <- dp$idname
  extra_clustervars <- clustervars[!(clustervars %in% c(unit_id, idname, ""))]
  # Per-unit cluster identifiers aligned with the rows of inffunc. faster_mode supplies dp$cluster_vector,
  # but for unbalanced panels (internally panel = FALSE, so the time-invariant data keeps every
  # observation) that vector is observation-length rather than unit-length and no longer aligns with the
  # unit-level influence function. Derive the vector from the data exactly as mboot() does -- one row per
  # cross-sectional unit, keyed on the internal unit id -- and store it back whenever it is missing OR does
  # not align with inffunc. This keeps the analytical clustered SE and aggte() working for faster_mode TRUE
  # and FALSE and for balanced and unbalanced panels; repeated cross sections (where the per-observation
  # vector already has one row per unit and aligns) fall through unchanged.
  if (length(extra_clustervars) > 0 && !is.null(dp$data) && !is.null(unit_id) &&
      (is.null(dp$cluster_vector) || length(dp$cluster_vector) != nrow(inffunc))) {
    cdat <- as.data.frame(dp$data)
    if (all(c(unit_id, extra_clustervars[1L]) %in% names(cdat))) {
      rebuilt_cv <- as.vector(unique(cdat[, c(unit_id, extra_clustervars[1L])])[, 2L])
      # only adopt the rebuilt vector if it lines up with the influence-function rows
      if (length(rebuilt_cv) == nrow(inffunc)) dp$cluster_vector <- rebuilt_cv
    }
  }
  cluster_vec <- dp$cluster_vector
  # Record which variable cluster_vec represents (when it is valid), so aggte() can tell whether an
  # inherited or overridden clustervars can be honored from this object -- and refuse to silently cluster
  # on the wrong variable.
  if (length(extra_clustervars) > 0 && !is.null(cluster_vec) && length(cluster_vec) == nrow(inffunc)) {
    dp$cluster_vector_var <- extra_clustervars[1L]
  }
  cluster_analytic <- (length(extra_clustervars) > 0) && !bstrap &&
    !is.null(cluster_vec) && length(cluster_vec) == nrow(inffunc)
  if (cluster_analytic) {
    Sc <- rowsum(as.matrix(inffunc), cluster_vec)   # n_clusters x k cluster sums
    V  <- Matrix::Matrix(crossprod(Sc) / n)
  } else {
    V <- Matrix::t(inffunc) %*% inffunc / (n)
  }
  se <- sqrt(Matrix::diag(V) / n)

  # Zero standard error replaced by NA
  se[se <= sqrt(.Machine$double.eps) * 10] <- NA

  # If a cluster variable beyond idname is requested without the bootstrap but the cluster-robust variance
  # could not be formed (e.g. cluster identifiers unavailable), fall back to i.i.d. and let the user know.
  if ((length(extra_clustervars) > 0) & !bstrap & !cluster_analytic) {
    warning("Clustered standard errors could not be computed analytically; the reported standard errors do NOT account for clustering. Set bstrap = TRUE for the cluster-robust multiplier bootstrap.")
  }

  # Identify entries of main diagonal V that are zero or NA
  zero_na_sd_entry <- unique(which(is.na(se)))

  # bootstrap variance matrix
  if (bstrap) {
    bout <- mboot(inffunc, DIDparams = dp, pl = pl, cores = cores, return_V = FALSE)
    bres <- bout$bres

    if (length(zero_na_sd_entry) > 0) {
      se[-zero_na_sd_entry] <- bout$se[-zero_na_sd_entry]
    } else {
      se <- bout$se
    }
  }
  # Zero standard error replaced by NA
  se[se <= sqrt(.Machine$double.eps) * 10] <- NA


  #-----------------------------------------------------------------------------
  # compute Wald pre-test
  #-----------------------------------------------------------------------------

  # Determine whether the variance matrix V is a valid basis for the Wald pre-test. The i.i.d. form
  # V = t(inffunc) %*% inffunc / n is valid for balanced panels, unbalanced panels (influence functions
  # are aggregated to the unit level), and repeated cross-sections (CLT applies per independent
  # observation). When a cluster variable beyond idname is specified and V is built from cluster sums
  # (cluster_analytic, above), V is cluster-robust and the Wald test remains valid. It is only skipped
  # when an extra cluster variable is present but V is the i.i.d. form (e.g. bstrap = TRUE), since then
  # V ignores between-cluster correlation; in that case we rely on the bootstrap confidence bands.

  wald_invalid <- NULL
  if (length(extra_clustervars) > 0 && !cluster_analytic) {
    wald_invalid <- paste0(
      "The Wald pre-test is not reported when clustering beyond the unit level ",
      "(clustervars = '", paste(extra_clustervars, collapse = "', '"), "') ",
      "because the analytical variance matrix does not account for between-cluster ",
      "correlation. Use the bootstrap confidence intervals to assess pre-trends."
    )
  }

  if (!is.null(wald_invalid)) {
    message(wald_invalid)
    W <- NULL
    Wpval <- NULL
  }

  if (is.null(wald_invalid)) {
    # select which periods are pre-treatment
    pre <- which(group > tt)
    # number of pre-treatment cells before the variance filter below, so the
    # no-pre-periods warning can distinguish "no pre-treatment cells exist"
    # from "all pre-treatment cells were dropped for NA/zero variance"
    n_pre_cells <- length(pre)

    # Drop group-periods that have variance equal to zero (singularity problems)
    if (length(zero_na_sd_entry) > 0) {
      pre <- pre[!(pre %in% zero_na_sd_entry)]
    }
    # pseudo-atts in pre-treatment periods
    preatt <- as.matrix(att[pre])

    # covariance matrix of pre-treatment atts
    preV <- as.matrix(V[pre, pre])

    # check if there are actually any pre-treatment periods
    if (length(preV) == 0) {
      if (n_pre_cells > 0) {
        # pre-treatment cells existed, but every one was dropped by the
        # zero/NA-variance filter above
        msg <- paste0(
          "No pre-treatment periods available for the Wald pre-test of parallel trends: ",
          "pre-treatment ATT(g,t) estimates exist, but all of them have missing or zero ",
          "variance (e.g., because the underlying 2x2 estimation failed for those cells), ",
          "so they were dropped from the pre-test."
        )
      } else {
        msg <- paste0(
          "No pre-treatment periods available for the Wald pre-test of parallel trends. ",
          "This can happen when all groups are first treated early in the panel ",
          "(e.g., in the second time period) so that no pre-treatment ATT(g,t) estimates exist."
        )
        if (anticipation > 0) {
          msg <- paste0(msg, " Note: anticipation=", anticipation, " further reduces the number of available pre-treatment periods.")
        }
      }
      warning(msg)
      W <- NULL
      Wpval <- NULL
    } else if (sum(is.na(preV))) {
      warning("Not returning pre-test Wald statistic due to NA pre-treatment values")
      W <- NULL
      Wpval <- NULL
    } else if (rcond(preV) <= .Machine$double.eps) {
      # singular covariance matrix for pre-treatment periods
      warning("Not returning pre-test Wald statistic due to singular covariance matrix")
      W <- NULL
      Wpval <- NULL
    } else {
      # everything is working...
      W <- n * t(preatt) %*% solve(preV) %*% preatt
      q <- length(pre) # number of restrictions
      Wpval <- round(1 - pchisq(W, q), 5)
    }
  }


  #-----------------------------------------------------------------------------
  # compute confidence intervals / bands
  #-----------------------------------------------------------------------------

  # critical value from N(0,1), for pointwise
  cval <- qnorm(1 - alp / 2)

  # in order to get uniform confidence bands
  # HAVE to use the bootstrap
  if (bstrap) {
    if (cband) {
      # for uniform confidence band
      # compute new critical value
      # see paper for details
      # Vectorized like mboot(): both quantiles of each bootstrap column are
      # taken in a single type = 1 call (one sort per column instead of two),
      # with the qnorm normalization hoisted out of the per-column loop.
      iqr_norm <- qnorm(.75) - qnorm(.25)
      bSigma <- apply(bres, 2, function(b) {
        qs <- quantile(b, c(.25, .75), type = 1, na.rm = TRUE)
        (qs[2L] - qs[1L]) / iqr_norm
      })

      bSigma[bSigma <= sqrt(.Machine$double.eps) * 10] <- NA

      # sup-t confidence band: per bootstrap draw, the max over columns of
      # |b / bSigma|, as a pmax column-fold (equivalent to the previous row-wise
      # apply(max, na.rm = TRUE): NA entries are ignored via -Inf, and an all-NA
      # row yields -Inf, exactly as before).
      if (ncol(bres) == 0L) {
        # every bootstrap dimension was degenerate (mboot dropped all columns):
        # the previous row-wise apply(max, na.rm = TRUE) returned -Inf per row
        bT <- rep(-Inf, nrow(bres))
      } else {
        scaled <- abs(bres / rep(bSigma, each = nrow(bres)))
        scaled[is.na(scaled)] <- -Inf
        bT <- scaled[, 1L]
        for (j in seq_len(ncol(scaled))[-1L]) bT <- pmax(bT, scaled[, j])
      }
      cval <- quantile(bT, 1 - alp, type = 1, na.rm = TRUE)
      if (cval >= 7) {
        warning("Simultaneous critical value is arguably `too large' to be reliable. This usually happens when the number of observations per group is small and/or there is not much variation in outcomes.")
      }
    }
  }


  } else {
    # point estimates only (compute_inffunc = FALSE): no influence functions were
    # computed, so there are no standard errors, uniform bands, or pre-test.
    se <- rep(NA_real_, length(att))
    V <- NULL
    cval <- qnorm(1 - alp / 2)
    W <- NULL
    Wpval <- NULL
  }

  # Return this list
  return(MP(group = group, t = tt, att = att, V_analytical = V, se = se, c = cval, inffunc = inffunc, n = n, W = W, Wpval = Wpval, alp = alp, DIDparams = dp))
}

#' @title Get the cohort for the current g,t values
#' @description A utility function to get vector with index of units that will be included in the DiD estimation for the current g,t pair.
#'
#' @param group Group.
#' @param time Time period.
#' @param tfac Time factor, which is 1 for varying base period and 0 for universal base period.
#' @param pret Pre treatment period.
#' @param dp2 DiD parameters v2.0.
#'
#' @return A vector of 1s, 0s and NAs indicating the cohort for the current g,t pair.
#' @noRd
get_did_cohort_index <- function(group, time, tfac, pret, dp2){
  # return a vector of dimension id_size with 1, 0 or NA values
  time_periods <- dp2$time_periods

  if(dp2$panel){
    # --- Balanced panel: use pre-computed cumulative sizes ---
    cohort_vec <- dp2$cohort_counts$cohort
    cum_sizes <- dp2$.cohort_cum_sizes  # pre-computed in compute.att_gt2

    # find the first cohort after the cutoff for not-yet-treated
    if (dp2$control_group == "notyettreated") {
      cutoff <- time_periods[max(time, pret) + tfac] + dp2$anticipation
      min_idx <- match(TRUE, cohort_vec > cutoff)
      min_control_group <- if (is.na(min_idx)) Inf else cohort_vec[min_idx]
    } else {
      min_control_group <- Inf
    }

    # control group boundaries
    ctrl_start_idx <- match(TRUE, cohort_vec >= min_control_group)
    if (is.na(ctrl_start_idx)) {
      start_control <- dp2$id_count + 1L  # no control group found
    } else {
      start_control <- if (ctrl_start_idx == 1L) 1L else cum_sizes[ctrl_start_idx - 1L] + 1L
    }
    end_control <- cum_sizes[length(cum_sizes)]

    # treated group boundaries
    treat_idx <- match(dp2$treated_groups[group], cohort_vec)
    if (is.na(treat_idx)) {
      stop("Internal error: treated group ", dp2$treated_groups[group],
           " not found in cohort_vec. Please report this issue.")
    }
    start_treat <- if (treat_idx == 1L) 1L else cum_sizes[treat_idx - 1L] + 1L
    end_treat <- cum_sizes[treat_idx]

    # fill the cohort index vector
    did_cohort_index <- rep(NA_integer_, dp2$id_count)
    if (start_control <= end_control) did_cohort_index[start_control:end_control] <- 0L
    did_cohort_index[start_treat:end_treat] <- 1L

  } else {
    # getting the index to get units who will participate in the estimation for the (g,t) cell.
    # Note: This works because the data is already ordered in a specific way. Changing that order will break this.

    dat <- dp2$time_invariant_data          # shortcut

    # flags for treated and control units. The per-group flag, the
    # never-treated flag and the per-period masks are loop-invariant across
    # (g,t) cells, so they are pre-computed once in compute.att_gt2() instead
    # of rescanning every row per cell (complete.cases() ran upstream in
    # pre_process_did2, so == has no NA hazard relative to %in%).
    Gflag <- dp2$.gflag_by_group[[group]]

    if (dp2$control_group == "nevertreated") {
      Cflag <- dp2$.never_treated
    } else {  # not-yet-treated: the cutoff varies with t, so this comparison stays per cell
      Cflag <- dp2$.never_treated |
        (dat[[dp2$gname]] > dp2$time_periods[max(time, pret) + tfac] + dp2$anticipation &
           !Gflag)
    }

    # keep only rows observed in pret or t
    keep <- dp2$.period_masks[[time + tfac]] | dp2$.period_masks[[pret]]

    did_cohort_index <- rep(NA_integer_, nrow(dat))
    did_cohort_index[keep & Cflag] <- 0L
    did_cohort_index[keep & Gflag] <- 1L

  }

  return(did_cohort_index)
}

#' @title Wrapper to run DRDID package
#' @description A utility function to run the DRDID package for the current g,t pair.
#'
#' @param cohort_data Data table with the cohort data for the current g,t pair
#' @param covariates Matrix of covariates to be used in the estimation
#' @param dp2 DiD parameters v2.0.
#'
#' @return A list containing the estimated ATT and the influence function as
#'  sparse triplets (`if_i` row indices, `if_x` values), except on the force_rc
#'  path which returns the dense stacked vector (`inf_func`) for the caller's
#'  half-fold. `if_i`/`if_x` (or `inf_func`) are NULL for point estimates only.
#' @noRd
run_DRDID <- function(cohort_data, covariates, dp2, g_val = NULL, t_val = NULL,
                      pret_val = NULL, force_rc = FALSE, check_cache_key = NULL){

  extra_args <- if (is.null(dp2$extra_args)) list() else dp2$extra_args
  gt_label <- if (!is.null(g_val) && !is.null(t_val)) paste0(" for group ", g_val, " in time period ", t_val) else ""
  # whether to compute influence functions (default TRUE; FALSE = point estimates only)
  do_inf <- is.null(dp2$compute_inffunc) || isTRUE(dp2$compute_inffunc)

  if(dp2$panel && !force_rc){
    # --------------------------------------
    # Panel Data
    # --------------------------------------
    # cohort_data is a plain named list of vectors (D, y1, y0, i.weights), built in
    # run_att_gt_estimation(). Working on the vectors directly avoids per-cell
    # data.table construction/extraction; the estimator inputs are identical.

    # still total number of units (not just included in G or C)
    D_all <- cohort_data$D
    n <- length(D_all)

    # pick up the indices for units that will be used to compute ATT(g,t)
    valid_obs <- which(!is.na(D_all))
    D    <- D_all[valid_obs]
    y1   <- cohort_data$y1[valid_obs]
    y0   <- cohort_data$y0[valid_obs]
    i.weights <- cohort_data$i.weights[valid_obs]

    if(dp2$xformla != ~1){
      covariates <- covariates[valid_obs,]
      intercept_only <- FALSE
    } else {
      covariates <- rep(1, length(valid_obs))
      intercept_only <- TRUE
    }
    covariates <- as.matrix(covariates)

    # num obs. for computing ATT(g,t)
    n1 <- length(D)

    #-----------------------------------------------------------------------------
    # check for overlap and regression problems
    #-----------------------------------------------------------------------------
    custom_est_method <- inherits(dp2$est_method, "function")

    if (!custom_est_method) {
      D_vec <- D

      # Guard booleans: reuse the cached result when run_att_gt_estimation()
      # established that this cell's (covariates, D) inputs are bit-identical to
      # an earlier cell's (panel + nevertreated + non-varying weights; see
      # compute.att_gt2). Warnings are still emitted per cell below, so cached
      # failures keep the per-cell warning count and text unchanged.
      guard <- if (!is.null(check_cache_key)) dp2$.check_cache[[check_cache_key]] else NULL
      if (is.null(guard)) {
        # overlap check for pscore based methods; regression-feasibility check
        # for outcome-regression methods, run only if overlap passed (matching
        # the early return on overlap failure below)
        overlap_fail <- (dp2$est_method %in% c("dr", "ipw")) &&
          overlap_check_fail(covariates, D_vec, intercept_only)
        rcond_fail <- !overlap_fail && (dp2$est_method %in% c("dr", "reg")) &&
          rcond_check_fail(covariates, D_vec, intercept_only)
        guard <- list(overlap_fail = overlap_fail, rcond_fail = rcond_fail)
        if (!is.null(check_cache_key)) dp2$.check_cache[[check_cache_key]] <- guard
      }

      if (guard$overlap_fail) {
        warning(paste0("overlap condition violated", gt_label))
        return(list(att = NA, if_i = if (do_inf) seq_len(n) else NULL,
                    if_x = if (do_inf) rep(NA_real_, n) else NULL))
      }
      if (guard$rcond_fail) {
        warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned", gt_label, "; consider centering/rescaling covariates or removing collinear terms"))
        return(list(att = NA, if_i = if (do_inf) seq_len(n) else NULL,
                    if_x = if (do_inf) rep(NA_real_, n) else NULL))
      }
    }

    #-----------------------------------------------------------------------------
    # code for actually computing ATT(g,t)
    #-----------------------------------------------------------------------------


    if (inherits(dp2$est_method, "function")) {
      # user-specified function
      attgt <- do.call(dp2$est_method, c(list(
                          y1=y1,
                          y0=y0,
                          D=D,
                          covariates=covariates,
                          i.weights=i.weights,
                          inffunc=do_inf), extra_args))
    } else if (dp2$est_method == "ipw") {
      # inverse-probability weights
      attgt <- std_ipw_did_panel(y1=y1,
                                        y0=y0,
                                        D=D,
                                        covariates=covariates,
                                        i.weights=i.weights,
                                        boot=FALSE, inffunc=do_inf)
    } else if (dp2$est_method == "reg") {
      # regression
      attgt <- reg_did_panel(y1=y1,
                                    y0=y0,
                                    D=D,
                                    covariates=covariates,
                                    i.weights=i.weights,
                                    boot=FALSE, inffunc=do_inf)
    } else {
      # doubly robust, this is default
      attgt <- drdid_panel(y1=y1,
                                  y0=y0,
                                  D=D,
                                  covariates=covariates,
                                  i.weights=i.weights,
                                  boot=FALSE, inffunc=do_inf)
    }

    # adjust influence function to account for only using
    # subgroup to estimate att(g,t). Stored as sparse triplets (if_i row
    # indices / if_x values) rather than a dense length-n vector: the assembly
    # in compute.att_gt2() concatenates the triplets directly, mirroring the
    # slow path, instead of rescanning dense columns with which().
    if (do_inf) {
      if_i <- valid_obs
      if_x <- (n/n1)*attgt$att.inf.func
    } else {
      if_i <- NULL   # point estimates only
      if_x <- NULL
    }

  } else {
    # --------------------------------------
    # Repeated Cross-Section
    # --------------------------------------
    # cohort_data is a plain named list of vectors (D, y, post, i.weights, .rowid),
    # built in run_att_gt_estimation(). Working on the vectors directly avoids per-cell
    # data.table construction/extraction; the estimator inputs are identical.
    D_all <- cohort_data$D
    rowid_all <- cohort_data$.rowid   # all rows (needed for the unbalanced aggregation)

    # still total number of obs. (not just included in G or C)
    n <- length(D_all)

    # pick up the indices for units that will be used to compute ATT(g,t)
    valid_obs <- which(!is.na(D_all))
    D    <- D_all[valid_obs]
    y    <- cohort_data$y[valid_obs]
    post <- cohort_data$post[valid_obs]
    i.weights <- cohort_data$i.weights[valid_obs]

    # num obs. for computing ATT(g,t)
    n1 <- length(D)

    if(dp2$xformla != ~1){
      covariates <- covariates[valid_obs,]
      intercept_only <- FALSE
    } else {
      covariates <- covariates[valid_obs]
      intercept_only <- TRUE
    }

    covariates <- as.matrix(covariates)

    # number of rows of the final influence-function column for early returns:
    # the unbalanced-panel branch aggregates observations to units; force_rc
    # failure triplets are already at unit level (the caller's half-fold only
    # applies to the dense success vector); plain RC has one row per observation
    n_inf <- if (dp2$allow_unbalanced_panel || force_rc) dp2$id_count else n

    skip_this_att_gt <- FALSE
    if (sum(D * post) == 0) {
      warning(paste0("No units in group ", g_val, " in time period ", t_val))
      skip_this_att_gt <- TRUE
    }
    if (sum(D * (1 - post)) == 0) {
      warning(paste0("No units in group ", g_val, " in time period ", pret_val))
      skip_this_att_gt <- TRUE
    }
    if (sum((1 - D) * post) == 0) {
      warning(paste0("No available control units for group ", g_val,
                     " in time period ", t_val))
      skip_this_att_gt <- TRUE
    }
    if (sum((1 - D) * (1 - post)) == 0) {
      warning(paste0("No available control units for group ", g_val,
                     " in time period ", pret_val))
      skip_this_att_gt <- TRUE
    }
    if (skip_this_att_gt) {
      return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                  if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
    }

    #-----------------------------------------------------------------------------
    # check for overlap and regression problems
    #-----------------------------------------------------------------------------
    custom_est_method <- inherits(dp2$est_method, "function")

    if (!custom_est_method) {
      D_vec <- D

      # checks for pscore based methods (no caching on the RC branch: the
      # cohort index varies cell by cell here)
      if (dp2$est_method %in% c("dr", "ipw")) {
        if (overlap_check_fail(covariates, D_vec, intercept_only)) {
          warning(paste0("overlap condition violated", gt_label))
          return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                      if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
        }
      }

      # check if can run regression using control units
      if (dp2$est_method %in% c("dr", "reg")) {
        if (rcond_check_fail(covariates, D_vec, intercept_only)) {
          warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned", gt_label, "; consider centering/rescaling covariates or removing collinear terms"))
          return(list(att = NA, if_i = if (do_inf) seq_len(n_inf) else NULL,
                      if_x = if (do_inf) rep(NA_real_, n_inf) else NULL))
        }
      }
    }

    #-----------------------------------------------------------------------------
    # code for actually computing ATT(g,t)
    #-----------------------------------------------------------------------------

    if (inherits(dp2$est_method, "function")) {
      # user-specified function
      attgt <- do.call(dp2$est_method, c(list(
                          y=y,
                          post=post,
                          D=D,
                          covariates=covariates,
                          i.weights=i.weights,
                          inffunc=do_inf), extra_args))
    } else if (dp2$est_method == "ipw") {
      # inverse-probability weights
      attgt <- std_ipw_did_rc(y=y,
                             post=post,
                             D=D,
                             covariates=covariates,
                             i.weights=i.weights,
                             boot=FALSE, inffunc=do_inf)
    } else if (dp2$est_method == "reg") {
      # regression
      attgt <- reg_did_rc(y=y,
                         post=post,
                         D=D,
                         covariates=covariates,
                         i.weights=i.weights,
                         boot=FALSE, inffunc=do_inf)
    } else {
      # doubly robust, this is default
      attgt <- drdid_rc(y=y,
                       post=post,
                       D=D,
                       covariates=covariates,
                       i.weights=i.weights,
                       boot=FALSE, inffunc=do_inf)
    }

    # n/n1 adjusts for estimating the
    # att_gt only using observations from groups
    # G and C
    # adjust influence function to account for only using
    # subgroup to estimate att(g,t). Stored as sparse triplets (if_i/if_x),
    # mirroring the slow path; the dense vector is kept only where it is still
    # needed (unbalanced-panel rowsum aggregation, force_rc half-fold).
    if (!do_inf) {
      if_i <- NULL   # point estimates only
      if_x <- NULL
    } else if(dp2$allow_unbalanced_panel){
      # since this is technically a panel data but ran as RCS, we need to adjust the
      # influence function by aggregating influence value by .rowid (several obs of one
      # unit can be used to estimate ATT in each 2x2). rowsum(..., reorder = FALSE)
      # reproduces the data.table `by = .rowid` first-appearance group order exactly.
      # The aggregated dense vector is then converted to triplets with the same
      # is.na | != 0 scan the assembly previously applied to dense columns.
      inf_func_long <- numeric(n)
      inf_func_long[valid_obs] <- (dp2$id_count / n1) * attgt$att.inf.func
      agg <- dp2$.rowid_agg
      inf_func_vector <- if (!is.null(agg)) {
        # precomputed sparse unit-aggregation operator (see compute.att_gt2):
        # accumulates in row order within each unit, bit-identical to the
        # rowsum() fallback but without per-cell hashing/dimnames work
        as.vector(agg %*% inf_func_long)
      } else {
        rowsum(inf_func_long, rowid_all, reorder = FALSE)[, 1L]
      }
      if_i <- which(is.na(inf_func_vector) | inf_func_vector != 0)
      if_x <- inf_func_vector[if_i]

    } else if (force_rc) {
      # balanced panel run as RC (fix_weights = "varying"): return the dense
      # stacked [all pre, all post] vector; run_att_gt_estimation() half-folds
      # it to unit level and converts it to triplets there.
      inf_func_vector <- rep(0, n)
      inf_func_vector[valid_obs] <- (n/n1)*attgt$att.inf.func
      return(list(att = attgt$ATT, inf_func = inf_func_vector))
    } else {
      if_i <- valid_obs
      if_x <- (n/n1)*attgt$att.inf.func
    }

  }

  return(list(att = attgt$ATT, if_i = if_i, if_x = if_x))

}


#' @title Run ATT estimation for a given group-time pair
#'
#' @description `run_att_gt_estimation` does the main work for computing
#'  multiperiod group-time average treatment effects
#' @param g group of interest (treated group at time t)
#' @param t time period
#' @param dp2 A DIDparams object v2.0
#'
#' @return a list with the gt cell and the results after performing estimation
#'
#' @keywords internal
run_att_gt_estimation <- function(g, t, dp2){

  if(dp2$print_details){cat("\n", paste0("Evaluating (g,t) = (",dp2$treated_groups[g],",",dp2$time_periods[t],")"))}
  tfac <- if (dp2$base_period != "universal") 1L else 0L
  # set pret
  # varying base period
  pret <- t

  # Use pre-computed pret for universal base period and post-treatment periods
  pret_g <- dp2$.pret_by_group[g]  # pre-computed in compute.att_gt2

  # universal base period
  if (dp2$base_period == "universal") {
    pret <- pret_g
  }

  # check if in post-treatment period
  if ((dp2$treated_groups[g] <= dp2$time_periods[(t+tfac)])) {

    # update pre-period if in post-treatment period to
    # be  period (g-delta-1)
    pret <- pret_g

    # print a warning message if there are no pre-treatment periods
    if (is.na(pret)) {
      warning(paste0("There are no pre-treatment periods for the group first treated at ", dp2$treated_groups[g], "\nUnits from this group are dropped"))

      # if there are no pre-treatment periods, code will
      # jump out of this loop
      return(NULL)
    }
  }

  #-----------------------------------------------------------------------------
  # if we are in period (g-1), normalize results to be equal to 0
  # and break without computing anything
  if (dp2$base_period == "universal") {
    if (dp2$time_periods[pret] == dp2$time_periods[(t+tfac)]) {
      return(list(base_period_norm = TRUE))
    }
  }

  # get units in treatment and control group
  did_cohort_index <- get_did_cohort_index(group = g, time = t, tfac = tfac, pret = pret, dp2)
  # In case of no treatment or control group in the cohort, return NULL
  valid_did_cohort <- any(did_cohort_index == 1) & any(did_cohort_index == 0)
  if(!isTRUE(valid_did_cohort)){
    if(dp2$print_details){cat("\n Skipping (g,t) as no treatment group or control group found")}
    return(NULL)
  }



  # key into the overlap/rcond guard cache (set on the panel non-varying branch
  # below when the cache is active; NULL = no caching for this cell)
  check_cache_key <- NULL

  if(dp2$panel){
    # Determine which weight period to use based on fix_weights
    use_rc_for_weights <- (!is.null(dp2$fix_weights) && dp2$fix_weights == "varying")

    if (use_rc_for_weights) {
      # fix_weights = "varying": stack into RC format with per-period weights
      n_units <- length(did_cohort_index)
      cohort_data <- data.table(
        D = rep(did_cohort_index, 2),
        y = c(dp2$outcomes_tensor[[pret]], dp2$outcomes_tensor[[t+tfac]]),
        post = rep(c(0L, 1L), each = n_units),
        i.weights = c(dp2$weights_tensor[[pret]], dp2$weights_tensor[[t+tfac]])
      )
      # Use earlier-period covariates for both halves — fix_weights only
      # changes weights, not the covariate conditioning set.
      # Use min(pret, t) to match the panel estimator's convention:
      # with base_period="universal", pret can be later than t for placebo cells.
      cov_early <- dp2$covariates_tensor[[base::min(pret, t)]]
      if (is.matrix(cov_early)) {
        covariates <- rbind(cov_early, cov_early)
      } else {
        covariates <- c(cov_early, cov_early)
      }
    } else {
      # Default or fixed weight options: use panel estimator with single weight vector
      if (is.null(dp2$fix_weights)) {
        # Default: weight from earlier of the two periods
        w_idx <- base::min(pret, t)
      } else if (dp2$fix_weights == "base_period") {
        w_idx <- dp2$.pret_by_group[g]
      } else if (dp2$fix_weights == "first_period") {
        w_idx <- 1L
      }
      # Plain named list of the cohort vectors (the tensors already hold vectors);
      # this avoids constructing/extracting a data.table per (g,t) cell, which the
      # profiler showed is a large share of the run time. run_DRDID()'s panel branch
      # consumes these as vectors, so the result is identical.
      cohort_data <- list(D = did_cohort_index,
                          y1 = dp2$outcomes_tensor[[t + tfac]],
                          y0 = dp2$outcomes_tensor[[pret]],
                          i.weights = dp2$weights_tensor[[w_idx]])
      covariates <- dp2$covariates_tensor[[base::min(pret, t)]]
      # Key for the overlap/rcond guard cache (see compute.att_gt2): under
      # nevertreated the cohort index depends only on g, so the guards'
      # (covariates, D) inputs are bit-identical across all cells of a group
      # sharing a covariate (and weight) period.
      if (!is.null(dp2$.check_cache)) {
        check_cache_key <- paste(g, base::min(pret, t), w_idx, sep = "_")
      }
    }
  } else {

    # post indicator reuses the pre-computed period mask (see compute.att_gt2)
    log_vec <- dp2$.period_masks[[t + tfac]]
    # convert TRUE/FALSE to 1/0 in place (fastest)
    set(dp2$time_invariant_data, j = "post", value = as.integer(log_vec))

    # Handle fix_weights for RC/unbalanced panel
    if (!is.null(dp2$fix_weights) && dp2$fix_weights %in% c("base_period", "first_period")) {
      if (dp2$fix_weights == "base_period") {
        target_period <- dp2$time_periods[dp2$.pret_by_group[g]]
      } else {
        target_period <- dp2$time_periods[1]
      }
      tid <- dp2$time_invariant_data
      # Memoize the weight lookup per target period (env set in compute.att_gt2):
      # target_period depends only on g for "base_period" and is constant for
      # "first_period", and tid's id/time/.w columns are never mutated in the
      # (g,t) loop, so fixed_w/na_w are identical across all cells sharing a
      # target period. Keying by as.character(target_period) memoizes the NA
      # case (group with no pre-period, varying base, pre-treatment cell)
      # under "NA" with unchanged semantics. The warning and the
      # did_cohort_index mutation below stay per cell.
      memo <- dp2$.fixed_w_memo
      memo_key <- as.character(target_period)
      cached <- if (!is.null(memo)) memo[[memo_key]] else NULL
      if (is.null(cached)) {
        # Build weight lookup from target period
        target_mask <- tid[[dp2$tname]] == target_period
        target_ids <- tid[[dp2$idname]][target_mask]
        target_ws <- tid[[".w"]][target_mask]
        # Map each observation's weight from the target period by integer id matching.
        # Equivalent to the previous named-character lookup (first match wins on any
        # duplicate id; unmatched ids -> NA) but avoids coercing every id to character.
        fixed_w <- target_ws[match(tid[[dp2$idname]], target_ids)]
        cached <- list(fixed_w = fixed_w, na_w = is.na(fixed_w))
        if (!is.null(memo)) memo[[memo_key]] <- cached
      }
      fixed_w <- cached$fixed_w
      # Exclude units not observed in target period by setting D to NA
      # (run_DRDID filters on !is.na(D))
      na_w <- cached$na_w
      if (any(na_w & !is.na(did_cohort_index))) {
        did_cohort_index[na_w] <- NA_integer_
        warning(paste0("Some units not observed in ", dp2$fix_weights,
                       " (period ", target_period, ") for group ",
                       dp2$treated_groups[g], " in time period ",
                       dp2$time_periods[t+tfac], ". These units are excluded."))
      }
      # Plain named list of the cohort vectors (avoids per-cell data.table
      # construction/extraction; run_DRDID()'s RC branch consumes these as vectors).
      cohort_data <- list(D = did_cohort_index, y = tid[[dp2$yname]], post = tid$post,
                          i.weights = fixed_w, .rowid = tid$.rowid)
    } else {
      cohort_data <- list(D = did_cohort_index, y = dp2$time_invariant_data[[dp2$yname]],
                          post = dp2$time_invariant_data$post,
                          i.weights = dp2$time_invariant_data[[".w"]],
                          .rowid = dp2$time_invariant_data$.rowid)
    }
    covariates <- dp2$covariates_matrix
  }

  # run estimation
  force_rc <- !is.null(dp2$fix_weights) && dp2$fix_weights == "varying" && dp2$panel
  did_result <- tryCatch(run_DRDID(cohort_data, covariates, dp2, g_val = dp2$treated_groups[g],
                                   t_val = dp2$time_periods[t + tfac],
                                   pret_val = dp2$time_periods[pret],
                                   force_rc = force_rc,
                                   check_cache_key = check_cache_key),
                         error = function(e) {
                           warning("Error computing internal 2x2 DiD for (g, t) = (", dp2$treated_groups[g], ", ", dp2$time_periods[t+tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
                           return(NULL)
                         })

  # When force_rc on balanced panel, the influence function has 2*n_units rows.
  # Half-split is safe here: cohort_data is explicitly stacked as
  # [all pre, all post] via rep(c(0L, 1L), each = n_units) in construction above.
  # The folded dense vector is converted to sparse triplets (if_i/if_x) with the
  # same is.na | != 0 scan the assembly previously applied to dense columns.
  if (force_rc && !is.null(did_result) && dp2$panel && !is.null(did_result$inf_func)) {
    inf <- did_result$inf_func
    n_half <- length(inf) %/% 2L
    # The RC influence function is normalized over the 2*n_units stacked rows;
    # folding pre+post per unit must divide by 2 so the unit-level influence
    # function (and hence the SE) matches the panel normalization. Without the
    # 1/2 the balanced-panel fix_weights = "varying" SE is exactly 2x too large.
    inf_folded <- 0.5 * (inf[1:n_half] + inf[(n_half + 1):(2L * n_half)])
    did_result$if_i <- which(is.na(inf_folded) | inf_folded != 0)
    did_result$if_x <- inf_folded[did_result$if_i]
    did_result$inf_func <- NULL
  }

  return(did_result)

}

#' @title Compute Group-Time Average Treatment Effects
#'
#' @description `compute.att_gt2` does the (g,t) cell computation and sends it to estimation,
#' then does all the post-processing after estimation
#'
#' @param dp2 A DIDparams object v2.0
#'
#' @return a list with length equal to the number of groups times the
#'  number of time periods; each element of the list contains an
#'  object that contains group-time average treatment effect as well
#'  as which group it is for and which time period it is for. It also exports
#'  the influence function which is used externally to compute
#'  standard errors.
#'
#' @keywords internal
#' @export
compute.att_gt2 <- function(dp2) {

  n <- dp2$id_count  # Total number of units
  # whether to compute influence functions (default TRUE; FALSE = point estimates only)
  do_inf <- is.null(dp2$compute_inffunc) || isTRUE(dp2$compute_inffunc)
  time_periods <- dp2$time_periods # tlist

  tlist.length <- if (dp2$base_period != "universal") length(time_periods) - 1L else length(time_periods)
  tfac <- if (dp2$base_period != "universal") 1L else 0L

  # Pre-compute cumulative cohort sizes for fast indexing in get_did_cohort_index
  if (dp2$panel) {
    dp2$.cohort_cum_sizes <- cumsum(dp2$cohort_counts$cohort_size)
  } else {
    # Pre-compute the loop-invariant row masks used by get_did_cohort_index()
    # and the post indicator in run_att_gt_estimation(): per-period membership,
    # the never-treated flag, and per-group treated flags. Each was previously
    # recomputed with full O(rows) scans for every (g,t) cell on the RC /
    # unbalanced-panel path. Memory: (T + n_groups + 1) logical vectors of
    # length nrow(time_invariant_data).
    dat <- dp2$time_invariant_data
    dp2$.period_masks <- lapply(time_periods, function(tp) dat[[dp2$tname]] == tp)
    dp2$.never_treated <- dat[[dp2$gname]] == Inf
    dp2$.gflag_by_group <- lapply(dp2$treated_groups, function(gval) dat[[dp2$gname]] == gval)
    # Memo for the fix_weights = "base_period"/"first_period" weight lookup in
    # run_att_gt_estimation(): fixed_w/na_w depend only on the target period
    # (one per group for "base_period", a single constant for "first_period"),
    # so the full-table mask + match() is computed once per distinct period
    # instead of once per (g,t) cell.
    if (!is.null(dp2$fix_weights) && dp2$fix_weights %in% c("base_period", "first_period")) {
      dp2$.fixed_w_memo <- new.env(parent = emptyenv())
    }
    # Pre-build the unit-aggregation operator for the unbalanced-panel
    # influence function: sparse 0/1 matrix S (id_count x rows) with
    # S[u, r] = 1 iff row r belongs to unit u, units ordered by first
    # appearance. S %*% v accumulates v in row order within each unit,
    # reproducing rowsum(v, .rowid, reorder = FALSE) bit-identically, while
    # the grouping work is done once per att_gt() call instead of once per
    # (g,t) cell (rowsum re-hashes the full .rowid vector and builds
    # character dimnames on every call).
    if (dp2$allow_unbalanced_panel && do_inf) {
      rowid_all <- dat$.rowid
      gmap <- match(rowid_all, unique(rowid_all))
      dp2$.rowid_agg <- Matrix::sparseMatrix(i = gmap, j = seq_along(rowid_all),
                                             x = 1, dims = c(n, length(rowid_all)))
    }
  }

  # Cache for the per-cell overlap/rcond guard booleans. For panel data with
  # control_group = "nevertreated" and non-varying weights, the guards'
  # (covariates, D) inputs are bit-identical across all cells of a group that
  # share a covariate (and weight) period, so each boolean is computed once per
  # (g, covariate-period, weight-period) key (see run_DRDID). Not used for
  # notyettreated (the control cohort varies with t) or fix_weights = "varying"
  # (RC estimator path). Only the booleans are stored; failure warnings are
  # still emitted per cell. options(did.disable_check_cache = TRUE) forces the
  # per-cell checks (escape hatch / used to verify bit-identical equivalence).
  if (dp2$panel && dp2$control_group[1] == "nevertreated" &&
      (is.null(dp2$fix_weights) || dp2$fix_weights != "varying") &&
      !isTRUE(getOption("did.disable_check_cache"))) {
    dp2$.check_cache <- new.env(parent = emptyenv())
  }


  # Pre-compute pret for each group (used for post-treatment periods and universal base)
  # This avoids calling tail(which(...)) repeatedly inside run_att_gt_estimation
  pret_by_group <- vapply(seq_len(dp2$treated_groups_count), function(g) {
    idx <- which((time_periods + dp2$anticipation) < dp2$treated_groups[g])
    if (length(idx) == 0L) NA_integer_ else idx[length(idx)]
  }, integer(1))
  dp2$.pret_by_group <- pret_by_group

  # in terms of indexes, not calendar times.
  # Build the (g,t) index pairs directly in g-major / t-minor order. This is
  # identical to expand.grid(g, t) followed by order(g, t) but avoids both the
  # data.frame construction and the per-iteration single-row subset below.
  g_vec <- rep.int(seq_len(dp2$treated_groups_count), rep.int(tlist.length, dp2$treated_groups_count))
  t_vec <- rep.int(seq_len(tlist.length), dp2$treated_groups_count)
  total_gt_iterations <- length(g_vec)


  # Running estimation using run_att_gt_estimation() function
  # Helper function for processing each (g,t) pair
  process_gt <- function(g, t) {

    # Run estimation
    gt_result <- run_att_gt_estimation(g, t, dp2)

      # Compute post-treatment indicator
    post.treat <- as.integer(dp2$treated_groups[g] <= dp2$time_periods[t+tfac])

    # Check for NULL first (estimation failed or was skipped). When do_inf = FALSE
    # (point estimates only) no influence-function triplets are stored (saving memory).
    # Failed cells propagate NA to every row of their column: if_i = seq_len(n),
    # if_x = NA, exactly the triplets the dense-column scan used to produce.
    if (is.null(gt_result)) {
      gt_result <- list(att = NA, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) seq_len(n) else NULL,
                        if_x = if (do_inf) rep(NA_real_, n) else NULL)
      return(gt_result)
    }

    # Base period normalization: ATT is 0 by construction (empty triplet column)
    if (!is.null(gt_result$base_period_norm)) {
      gt_result <- list(att = 0, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) integer(0) else NULL,
                        if_x = if (do_inf) numeric(0) else NULL)
      return(gt_result)
    }

    if (is.null(gt_result$att)) {
      # Estimation returned a result but without an ATT
      gt_result <- list(att = NA, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat,
                        if_i = if (do_inf) seq_len(n) else NULL,
                        if_x = if (do_inf) rep(NA_real_, n) else NULL)
      return(gt_result)
    } else {
      att <- gt_result$att
      if_i <- gt_result$if_i
      if_x <- gt_result$if_x

      # Handle NaN ATT: treat as estimation failure
      if (is.nan(att)) {
        att <- NA
        if (do_inf) {
          if_i <- seq_len(n)
          if_x <- rep(NA_real_, n)
        }
      }

      # Save ATT and influence triplets (if_i/if_x are NULL when do_inf = FALSE)
      gt_result <- list(att = att, group = dp2$treated_groups[g], year = dp2$time_periods[t+tfac], post = post.treat, if_i = if_i, if_x = if_x)
      return(gt_result)
    }
  }

  # run the estimation for each (g,t) pair with process_gt
  gt_results <- lapply(seq_len(total_gt_iterations), function(idx) {
    process_gt(g_vec[idx], t_vec[idx])
  })

  # Filter out NULL results
  gt_results <- Filter(Negate(is.null), gt_results)

  if (length(gt_results) == 0L) {
    stop("No valid (g, t) cells found for att_gt() estimation. ",
         "Check treatment timing, control group definition, and anticipation settings.")
  }

  # Post processing: Build sparse influence function matrix directly from triplets.
  # Skipped entirely for point estimates only (do_inf = FALSE) -- no n x k matrix is
  # ever formed, which is the memory saving of compute_inffunc = FALSE.
  if (!do_inf) {
    inffunc <- NULL
  } else {
    n_cols <- length(gt_results)

    # Each cell already carries its influence column as sparse triplets
    # (if_i row indices / if_x values), mirroring the slow path; concatenate
    # them once instead of rescanning dense per-cell vectors with which().
    # The triplet order is unchanged (column-major, original within-column
    # order), so the resulting sparse matrix is identical. Every code path
    # produces columns with dp2$id_count rows (panel units, RC observations,
    # unbalanced-panel units after the rowsum aggregation, force_rc units
    # after the half-fold).
    nz_list <- lapply(gt_results, `[[`, "if_i")
    val_list <- lapply(gt_results, `[[`, "if_x")
    trip_i <- unlist(nz_list, use.names = FALSE)
    trip_x <- unlist(val_list, use.names = FALSE)
    trip_j <- rep.int(seq_len(n_cols), lengths(nz_list))
    inffunc <- Matrix::sparseMatrix(i = trip_i, j = trip_j, x = trip_x,
                                    dims = c(n, n_cols))
  }

  return(list(attgt.list=gt_results, inffunc=inffunc))
}

#' @title Compute Group-Time Average Treatment Effects
#'
#' @description `compute.att_gt` does the main work for computing
#'  multiperiod group-time average treatment effects
#'
#' @param dp A DIDparams object
#'
#' @return a list with length equal to the number of groups times the
#'  number of time periods; each element of the list contains an
#'  object that contains group-time average treatment effect as well
#'  as which group it is for and which time period it is for. It also exports
#'  the influence function which is used externally to compute
#'  standard errors.
#'
#' @keywords internal
#'
#' @export
compute.att_gt <- function(dp) {
  #-----------------------------------------------------------------------------
  # unpack DIDparams
  #-----------------------------------------------------------------------------
  # Internal pretest hook (set only by conditional_did_pretest): when a
  # y-invariant setup bundle is supplied (dp$setup_precomp, built once per
  # pretest call via dp$return_setup below), the bundle replaces the whole
  # per-call setup, including this data.table copy, so skip it.
  setup_precomp <- dp$setup_precomp
  if (is.null(setup_precomp)) data <- data.table::as.data.table(dp$data)
  yname <- dp$yname
  tname <- dp$tname
  idname <- dp$idname
  xformla <- dp$xformla
  weightsname <- dp$weightsname
  est_method <- dp$est_method
  extra_args <- if (is.null(dp$extra_args)) list() else dp$extra_args
  base_period <- dp$base_period
  panel <- dp$panel
  true_repeated_cross_sections <- dp$true_repeated_cross_sections
  print_details <- dp$print_details
  control_group <- dp$control_group
  anticipation <- dp$anticipation
  gname <- dp$gname
  n <- dp$n
  nT <- dp$nT
  nG <- dp$nG
  tlist <- dp$tlist
  glist <- dp$glist

  #-----------------------------------------------------------------------------
  # main computations
  #-----------------------------------------------------------------------------

  # will populate with all att(g,t)
  attgt.list <- list()

  # place holder in lists
  counter <- 1

  # number of time periods
  tlist.length <- if (base_period != "universal") length(tlist) - 1L else length(tlist)
  tfac <- if (base_period != "universal") 1L else 0L

  # influence function dimensions
  inffunc_nrow <- n

  # list to collect sparse matrix updates (built into sparse matrix after loops)
  inffunc_updates <- list()
  # counter for keeping track of updates
  update_counter <- 1

  # never treated option
  nevertreated <- (control_group[1] == "nevertreated")

  fix_weights <- dp$fix_weights

  # intercept-only design (xformla = ~1, the default): the per-cell overlap and
  # rcond guards have closed forms (see overlap_check_fail / rcond_check_fail),
  # so the preliminary logit fit and Gram matrix can be skipped.
  intercept_only <- (xformla == ~1)

  # Cache for the per-cell overlap/rcond guard booleans. For panel data with
  # control_group = "nevertreated" and non-varying weights, the guards'
  # (covariates, G) inputs are bit-identical across all cells of a group that
  # share a covariate (and weight) period, so each boolean is computed once per
  # (g, covariate-period, weight-period) key. Not used for notyettreated (the
  # control cohort varies with t) or fix_weights = "varying" (RC estimator
  # path). Only the booleans are stored; failure warnings are still emitted per
  # cell. options(did.disable_check_cache = TRUE) forces the per-cell checks
  # (escape hatch / used to verify bit-identical equivalence). Mirrors the fast
  # path (compute.att_gt2).
  use_check_cache <- panel && nevertreated &&
    (is.null(fix_weights) || fix_weights != "varying") &&
    !isTRUE(getOption("did.disable_check_cache"))
  if (use_check_cache) check_cache <- new.env(parent = emptyenv())

  # Whether to compute influence functions (default TRUE; FALSE = point estimates
  # only). When FALSE we skip requesting influence functions from the 2x2 estimators,
  # skip the per-cell influence-function scaling, and skip storing inffunc_updates --
  # so a point-estimates-only run does not pay the influence-function compute/memory
  # cost. The att estimates are identical either way.
  do_inf <- is.null(dp$compute_inffunc) || isTRUE(dp$compute_inffunc)

  # Internal pretest-only flag (set by conditional_did_pretest): skip
  # post-treatment (g,t) cells entirely -- the pretest immediately discards
  # them (its keepers filter retains only group > t cells), so computing them
  # is pure waste. attgt.list and the influence-function columns stay aligned
  # with the reduced cell set because the counters only increment when a cell
  # is appended.
  pretreatment_only <- isTRUE(dp$pretreatment_cells_only)

  if (!is.null(setup_precomp)) {
    # ---- internal pretest path: reuse the y-invariant setup bundle -----------
    # conditional_did_pretest() calls compute.att_gt once per unit with only
    # the outcome column changed, so the whole per-call setup below (column
    # extraction, design matrix, per-period slicing, alignment check) is built
    # once via dp$return_setup and reused; each per-unit call passes only
    # dp$y_override (that unit's modified outcome, aligned 1:1 with dp$data
    # rows). The bundle is consumed only when the balanced, id-aligned panel
    # precompute applies (conditional_did_pretest keeps its legacy data-copy
    # path otherwise), and that path never reads `data` / `g_col` / `t_col`
    # inside the loops -- they are deliberately left undefined here so any
    # accidental read fails loudly instead of silently using a stale outcome.
    use_precompute_rc <- FALSE
    use_precompute_panel <- isTRUE(setup_precomp$use_precompute_panel)
    weights_by_period <- setup_precomp$weights_by_period
    period_mm <- setup_precomp$period_mm
    period_id <- setup_precomp$period_id
    period_w  <- setup_precomp$period_w
    g_unit_panel <- setup_precomp$g_unit_panel
    if (nevertreated) C_full <- setup_precomp$C_full
    # only the outcome differs across pretest iterations: rebuild the
    # per-period outcome slices from the override (bit-identical to slicing a
    # full data copy whose outcome column was overwritten, which is what the
    # legacy pretest loop did)
    y_override <- dp$y_override
    period_y <- lapply(setup_precomp$rows_by_period, function(rr) y_override[rr])
  } else {
    # Pre-extract columns to avoid repeated get() inside data.table (which is slow)
    g_col <- data[[gname]]
    t_col <- data[[tname]]

    # Build weight lookup by period for fix_weights options (balanced panel only)
    if (!is.null(fix_weights) && panel) {
      weights_by_period <- list()
      for (tp in seq_along(tlist)) {
        weights_by_period[[tp]] <- data[t_col == tlist[tp], .w]
      }
    }

    # Positional precompute for the repeated-cross-section / unbalanced-panel path:
    # each (g,t) cell is assembled from plain vectors and per-period row indices
    # precomputed once per call, instead of two full data.table subsets plus a
    # full-data %in% per cell (the dominant per-cell cost on this path). Bit-identical
    # to the legacy construction (see the per-cell comments below);
    # options(did.disable_precompute = TRUE) forces the legacy per-cell assembly
    # (escape hatch / used to verify bit-identical cell assembly; both sub-paths
    # still slice the hoisted global design matrix), mirroring the panel precompute.
    use_precompute_rc <- !panel && !isTRUE(getOption("did.disable_precompute"))

    # The precompute paths never read data$.C / data$.G (cohort indicators are
    # rebuilt from plain vectors), so skip the full-length column writes there.
    if (nevertreated && !use_precompute_rc) {
      set(data, j = ".C", value = as.integer(g_col == 0))
    }

    # -- design matrix: built ONCE over the full data, then row-subset per (g,t) cell --
    # This (a) avoids rebuilding model.matrix() nG*nT times, and (b) gives every cell the
    # GLOBAL factor levels / GLOBAL transform basis -- exactly what the repeated-cross-
    # section path, the fast path, and manually-constructed dummy columns use. As a
    # result a factor covariate behaves identically to adding its dummies by hand: a
    # level absent from a particular 2x2 comparison becomes an all-zero column (handled
    # by the estimator's overlap/rank check, giving the same NA + warning as the manual
    # dummies) rather than a dropped contrast that changes the design across cells.
    #
    # For purely row-wise numeric formulae (e.g. ~X, ~X1+X2, ~X1*X2, ~I(X^2)) the slice
    # is bit-identical to a per-cell model.matrix(xformla, disdat); for factors it equals
    # the manual-dummy design; for data-dependent bases (poly/scale/ns/...) it uses the
    # global basis, which yields the same ATT/SE as a per-cell fit (the estimator is
    # invariant to a full-rank reparameterization of the covariates).
    #
    # `data` is already complete-cases and finite (see pre_process_did), so na.pass adds
    # no NA rows and full_mm aligns 1:1 (by row position) with `data`.
    #
    # Drop globally-empty factor levels first: model.matrix() emits a dummy column for
    # every declared level, so a level absent from the ENTIRE estimation sample (common
    # after users subset their data) would produce an all-zero column in every cell and
    # NA-fail the whole estimation. Levels absent only from a particular 2x2 cell are
    # unaffected (still all-zero in that cell only, as documented above). Mirrors the
    # identical fix in get_did_tensors() (fast path).
    for (v in all.vars(xformla)) {
      if (is.factor(data[[v]])) set(data, j = v, value = droplevels(data[[v]]))
    }
    full_mm <- model.matrix(xformla, data = data, na.action = na.pass)
    if (panel) {
      # per-period row slices keyed by idname, for the earlier-period covariates that
      # get_wide_data() retains. (idname is used rather than .rowid because .rowid is
      # only created on the RC/unbalanced path; in a balanced panel idname is unique
      # within each period.)
      period_mm <- vector("list", length(tlist))
      period_id <- vector("list", length(tlist))
      period_y  <- vector("list", length(tlist))
      period_w  <- vector("list", length(tlist))
      # per-period row indices, kept for the pretest setup bundle (dp$return_setup)
      rows_by_period <- vector("list", length(tlist))
      y_col <- data[[yname]]
      w_col <- data[[".w"]]
      for (tp in seq_along(tlist)) {
        rows_tp <- which(t_col == tlist[tp])
        rows_by_period[[tp]] <- rows_tp
        period_mm[[tp]] <- full_mm[rows_tp, , drop = FALSE]
        period_id[[tp]] <- data[[idname]][rows_tp]
        period_y[[tp]]  <- y_col[rows_tp]
        period_w[[tp]]  <- w_col[rows_tp]
      }
      # Precompute fast path: when the panel is balanced with every period sharing the
      # same units in the same (id-sorted) row order, each 2x2 cell can be assembled by
      # position from these per-period blocks -- skipping the per-cell data[time_mask]
      # subset + get_wide_data() reshape (the dominant cost on the slow path).
      # fix_weights = "base_period"/"first_period" only changes WHICH period's weight
      # vector each cell uses (selected from weights_by_period below, mirroring the
      # fallback), so the precompute applies there too. Falls through to the original
      # get_wide_data() code, bit-identically, for fix_weights = "varying" (which needs
      # the long-format RC data) / unaligned panels.
      panel_units_aligned <- all(vapply(period_id, function(x)
        length(x) == length(period_id[[1L]]) && all(x == period_id[[1L]]), logical(1)))
      g_unit_panel <- g_col[which(t_col == tlist[1L])]
      # options(did.disable_precompute = TRUE) disables only this per-cell positional
      # assembly, falling back to the per-cell get_wide_data() reshape below (escape
      # hatch / used to verify bit-identical cell assembly). It does NOT bypass the
      # design-matrix hoist: BOTH sub-paths slice the global full_mm/period_mm built
      # above, so a hoist regression would be invisible to this option -- the hoist
      # itself is validated by the master-comparison tests in
      # test-modelmatrix-hoist.R, not by this escape hatch.
      use_precompute_panel <- (is.null(fix_weights) ||
        fix_weights %in% c("base_period", "first_period")) && panel_units_aligned &&
        !isTRUE(getOption("did.disable_precompute"))
      # Never-treated control cohort is invariant across every (g,t) cell, so build
      # it once here; notyettreated rebuilds C_full per cell inside the loops.
      if (use_precompute_panel && nevertreated) {
        C_full <- as.numeric(g_unit_panel == 0)
      }
    } else {
      # repeated cross-section / unbalanced: row-subset the full design by POSITION
      # (.rowid is not unique across periods here).
      global_mm <- full_mm
      use_precompute_panel <- FALSE
      if (use_precompute_rc) {
        # One-time positional precompute: per-period row indices into `data` plus
        # plain copies of the columns each cell consumes. which() preserves data
        # order within each period, so per-period vectors are bit-identical to the
        # corresponding data.table period subsets.
        period_rows_rc <- vector("list", length(tlist))
        for (tp in seq_along(tlist)) {
          period_rows_rc[[tp]] <- which(t_col == tlist[tp])
        }
        rowid_col <- data$.rowid
        y_col_rc <- data[[yname]]
        w_col_rc <- data[[".w"]]
        # loop-invariant id universe for the influence-function index lookup
        unique_rowid <- unique(rowid_col)
        # fix_weights base/first: per-period .rowid/.w lookup vectors replace the
        # per-cell full-table data[t_col == target_period, ] subset; match()
        # semantics (first match wins, unmatched -> NA) are unchanged.
        if (!is.null(fix_weights) && fix_weights %in% c("base_period", "first_period")) {
          target_rowid_by_period <- lapply(period_rows_rc, function(rr) rowid_col[rr])
          target_w_by_period <- lapply(period_rows_rc, function(rr) w_col_rc[rr])
        }
      }
    }
  }

  # Internal (set only by conditional_did_pretest): return the y-invariant
  # setup bundle built above -- so the pretest can hoist it out of its
  # per-unit loop -- instead of estimating. Balanced panel only; the pretest
  # keeps its legacy data-copy loop for repeated cross sections, and only
  # consumes the bundle when use_precompute_panel is TRUE.
  if (isTRUE(dp$return_setup)) {
    if (!panel) return(NULL)
    return(list(
      use_precompute_panel = use_precompute_panel,
      weights_by_period = if (!is.null(fix_weights)) weights_by_period else NULL,
      period_mm = period_mm,
      period_id = period_id,
      period_w = period_w,
      rows_by_period = rows_by_period,
      g_unit_panel = g_unit_panel,
      C_full = if (use_precompute_panel && nevertreated) C_full else NULL
    ))
  }

  # loop over groups
  for (g in 1:nG) {
    # Set up .G once (use pre-extracted g_col to avoid get() inside data.table).
    # The precompute paths derive the treated cohort from g_unit_panel / g_col
    # directly and never read data$.G, so skip the full-length column write there.
    current_g <- glist[g]
    if (!use_precompute_panel && !use_precompute_rc) {
      set(data, j = ".G", value = as.numeric(g_col == current_g))
    }

    # pre-compute the universal/post-treatment base period for this group (used multiple times)
    idx_g <- which((tlist + anticipation) < glist[g])
    pret_g <- if (length(idx_g) == 0L) NA_integer_ else idx_g[length(idx_g)]

    # Precompute path: the treated-cohort indicator depends only on g, so build it
    # once per group. Under nevertreated the control cohort is also t-invariant
    # (C_full was hoisted above the loops), so the kept-unit index and the G/C
    # vectors are hoisted here too; notyettreated recomputes C_full/disidx/kept
    # per cell below because its control cohort varies with t.
    if (use_precompute_panel) {
      G_full <- as.numeric(g_unit_panel == current_g)
      if (nevertreated) {
        disidx <- G_full == 1 | C_full == 1
        kept <- which(disidx)
        n1 <- length(kept)
        G <- G_full[kept]
        C <- C_full[kept]
      }
    }

    # loop over time periods
    for (t in 1:tlist.length) {
      #-----------------------------------------------------------------------------
      # Set pret

      # varying base period
      pret <- t

      # universal base period
      if (base_period == "universal") {
        # use same base period as for post-treatment periods
        pret <- pret_g
      }

      # check if in post-treatment period
      if ((glist[g] <= tlist[(t + tfac)])) {
        # update pre-period if in post-treatment period to
        # be  period (g-delta-1)
        pret <- pret_g
      }

      # check if there are no pre-treatment periods
      if (is.na(pret)) {
        warning(paste0("There are no pre-treatment periods for the group first treated at ", glist[g], "\nUnits from this group are dropped"))
        break
      }

      # pretest-only: skip post-treatment cells entirely (see
      # pretreatment_only above; conditional_did_pretest discards these
      # cells unread, so none of the per-cell work below is needed)
      if (pretreatment_only && glist[g] <= tlist[t + tfac]) next

      # use "not yet treated as control"
      # that is, never treated + units that are eventually treated,
      # but not treated by the current period (+ anticipation)
      if (!nevertreated) {
        # Use pre-extracted g_col to avoid get() inside data.table.
        # The precompute paths recompute the control cohort from
        # g_unit_panel / g_col and never read data$.C, so skip the full-length
        # column write there (the time_threshold scalar is still needed below).
        time_threshold <- tlist[max(t, pret) + tfac] + anticipation
        if (!use_precompute_panel && !use_precompute_rc) {
          set(data, j = ".C", value = as.integer((g_col == 0) |
            ((g_col > time_threshold) &
              (g_col != current_g))))
        }
      }


      #-----------------------------------------------------------------------------
      # if we are in period (g-1), normalize results to be equal to 0
      # and break without computing anything
      if (base_period == "universal") {
        if (tlist[pret] == tlist[(t + tfac)]) {
          attgt.list[[counter]] <- list(att = 0, group = glist[g], year = tlist[(t + tfac)], post = 0)
          # inffunc[,counter] <- rep(0,n)
          # counter <- counter+1
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = integer(0), # No non-zero entries
            values = numeric(0)
          )

          # Update the counters
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      }

      # print the details of which iteration we are on
      if (print_details) {
        cat(paste("current period:", tlist[(t + tfac)]), "\n")
        cat(paste("current group:", glist[g]), "\n")
        cat(paste("set pretreatment period to be", tlist[pret]), "\n")
      }

      #-----------------------------------------------------------------------------
      # results for the case with panel data
      #-----------------------------------------------------------------------------

      # post treatment dummy variable
      post.treat <- 1 * (glist[g] <= tlist[t + tfac])

      # total number of units (not just included in G or C)
      # disdat <- data[data[,tname] == tlist[t+tfac] | data[,tname] == tlist[pret],]
      # The precompute paths (panel per-period blocks; RC/unbalanced positional
      # assembly) build each cell without this long 2-period subset (the dominant
      # per-cell cost); the legacy escape-hatch paths still build disdat as before.
      # time_mask is only read here and in the fix_weights == "varying" branch,
      # which always runs with use_precompute_panel == FALSE (the panel precompute
      # excludes fix_weights == "varying") and panel == TRUE (so use_precompute_rc
      # is FALSE), so it is computed on every path that consumes it.
      if (!(panel && use_precompute_panel) && !use_precompute_rc) {
        target_times <- c(tlist[t + tfac], tlist[pret])
        time_mask <- t_col %in% target_times
        disdat <- data[time_mask]
      }


      if (panel) {
        if (use_precompute_panel) {
          # ---- precompute fast path (balanced, id-aligned panel; non-varying weights) --
          # Build the 2x2 cell directly from the per-period blocks. g is time-invariant
          # so .G / .C are unit-level; Ypost is always period (t+tfac) and Ypre always
          # period pret because get_wide_data's .y1/.y0 (later/earlier) swap exactly
          # cancels the universal-base ordering used below. The earlier period (whose
          # weights + covariates get_wide_data retains) is min(t + tfac, pret).
          # G_full (and, under nevertreated, C_full/disidx/kept/n1/G/C) are hoisted
          # to the group/setup level above; only t-varying pieces are built here.
          if (!nevertreated) {
            # time_threshold was computed above (kept for the precompute path)
            C_full <- as.numeric((g_unit_panel == 0) |
              ((g_unit_panel > time_threshold) & (g_unit_panel != current_g)))
            disidx <- G_full == 1 | C_full == 1
            kept <- which(disidx)
            n1 <- length(kept)
            G <- G_full[kept]
            C <- C_full[kept]
          }
          n <- length(g_unit_panel)
          Ypost <- period_y[[t + tfac]][kept]
          Ypre  <- period_y[[pret]][kept]
          earlier_idx <- min(t + tfac, pret)
          # Select weights based on fix_weights, mirroring the fallback's
          # weights_by_period[[...]][disidx] selection bit-identically:
          # weights_by_period[[tp]] and period_w[[tp]] share the same within-period
          # data order, and kept == which(disidx) under panel_units_aligned.
          if (is.null(fix_weights)) {
            # Default: the earlier period's weights (what get_wide_data retains)
            w <- period_w[[earlier_idx]][kept]
          } else if (fix_weights == "base_period") {
            w <- weights_by_period[[pret_g]][kept]
          } else {
            # fix_weights == "first_period" ("varying" never reaches the precompute)
            w <- weights_by_period[[1L]][kept]
          }
          covariates <- period_mm[[earlier_idx]][kept, , drop = FALSE]
          dimnames(covariates) <- list(NULL, colnames(covariates))
        } else {
          # ---- original path: per-cell get_wide_data() reshape ----
          # transform  disdat it into "cross-sectional" data where one of the columns
          # contains the change in the outcome over time.
          disdat <- get_wide_data(disdat, yname, idname, tname)

          # still total number of units (not just included in G or C)
          n <- nrow(disdat)

          # pick up the indices for units that will be used to compute ATT(g,t)
          disidx <- disdat$.G == 1 | disdat$.C == 1

          # pick up the data that will be used to compute ATT(g,t)
          disdat <- disdat[disidx]

          n1 <- nrow(disdat) # num obs. for computing ATT(g,t)

          # drop missing factors
          disdat <- droplevels(disdat)

          # give short names for data in this iteration
          G <- disdat$.G
          C <- disdat$.C

          # handle pre-treatment universal base period differently
          # we need to do this because panel2cs2 just puts later period
          # in .y1, but if we are in a pre-treatment period with a universal
          # base period, then the "base period" is actually the later period
          Ypre <- if (tlist[(t + tfac)] > tlist[pret]) disdat$.y0 else disdat$.y1
          Ypost <- if (tlist[(t + tfac)] > tlist[pret]) disdat$.y1 else disdat$.y0

          # Select weights based on fix_weights
          if (is.null(fix_weights)) {
            # Default: .w from get_wide_data (earlier period)
            w <- disdat$.w
          } else if (fix_weights == "base_period") {
            w <- weights_by_period[[pret_g]][disidx]
          } else if (fix_weights == "first_period") {
            w <- weights_by_period[[1L]][disidx]
          } else if (fix_weights == "varying") {
            w <- disdat$.w  # will be overridden below when switching to RC estimator
          } else {
            w <- disdat$.w
          }

          # matrix of covariates: index the precomputed earlier-period design by this
          # cell's units (idname). get_wide_data retained the earlier period, whose
          # index is min(t + tfac, pret).
          earlier_idx <- min(t + tfac, pret)
          covariates <- period_mm[[earlier_idx]][match(disdat[[idname]], period_id[[earlier_idx]]), , drop = FALSE]
          dimnames(covariates) <- list(NULL, colnames(covariates))
        }

        #-----------------------------------------------------------------------------
        # code for actually computing att(g,t)
        #-----------------------------------------------------------------------------

        attgt <- tryCatch({
          if (!is.null(fix_weights) && fix_weights == "varying") {
            # fix_weights = "varying": use RC estimators with per-period weights
            # Go back to long-format data for this (g,t) cell
            disdat_long <- data[time_mask]
            disdat_long_idx <- disdat_long$.G == 1 | disdat_long$.C == 1
            disdat_long <- droplevels(disdat_long[disdat_long_idx])
            Y_rc <- disdat_long[[yname]]
            G_rc <- disdat_long$.G
            post_rc <- as.numeric(disdat_long[[tname]] == tlist[t + tfac])
            w_rc <- disdat_long$.w
            # Use earlier-period covariates for all observations — fix_weights
            # only changes weights, not the covariate conditioning set.
            # Use min(pret, t+tfac) to match the panel estimator's convention:
            # with base_period="universal", pret can be later than t for placebo cells.
            earlier_idx_v <- min(pret, t + tfac)
            earlier_period <- tlist[earlier_idx_v]
            early_mask <- disdat_long[[tname]] == earlier_period
            disdat_early <- disdat_long[early_mask]
            cov_early <- period_mm[[earlier_idx_v]][match(disdat_early[[idname]], period_id[[earlier_idx_v]]), , drop = FALSE]
            dimnames(cov_early) <- list(NULL, colnames(cov_early))
            # Map each row in disdat_long to its unit's earlier-period covariates
            early_ids <- disdat_early[[idname]]
            all_ids <- disdat_long[[idname]]
            id_map <- match(all_ids, early_ids)
            covariates_rc <- cov_early[id_map, , drop = FALSE]

            # Run overlap/rank checks on RC data (not wide panel data)
            if (!is.function(est_method)) {
              if (est_method %in% c("dr", "ipw")) {
                if (overlap_check_fail(covariates_rc, G_rc, intercept_only)) {
                  warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                  stop("overlap")
                }
              }
              if (est_method %in% c("dr", "reg")) {
                if (rcond_check_fail(covariates_rc, G_rc, intercept_only)) {
                  warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                  stop("singular")
                }
              }
            }

            if (inherits(est_method, "function")) {
              res <- do.call(est_method, c(list(
                y = Y_rc, post = post_rc,
                D = G_rc, covariates = covariates_rc,
                i.weights = w_rc, inffunc = do_inf
              ), extra_args))
            } else if (est_method == "ipw") {
              res <- DRDID::std_ipw_did_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            } else if (est_method == "reg") {
              res <- DRDID::reg_did_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            } else {
              res <- DRDID::drdid_rc(Y_rc, post_rc, G_rc,
                covariates = covariates_rc,
                i.weights = w_rc, boot = FALSE, inffunc = do_inf)
            }
          } else {
            # Panel path: run overlap/rank checks on panel data. Guard booleans
            # are reused from the cache when this cell's (covariates, G) inputs
            # are bit-identical to an earlier cell's (nevertreated + non-varying
            # weights; see the cache setup above); warnings are still emitted
            # per cell, so cached failures keep counts and texts unchanged.
            if (!is.function(est_method)) {
              guard <- NULL
              if (use_check_cache) {
                w_idx_key <- if (is.null(fix_weights)) earlier_idx
                             else if (fix_weights == "base_period") pret_g
                             else 1L
                check_key <- paste(g, earlier_idx, w_idx_key, sep = "_")
                guard <- check_cache[[check_key]]
              }
              if (is.null(guard)) {
                # overlap check for pscore based methods; regression-feasibility
                # check for outcome-regression methods, run only if overlap
                # passed (matching the stop("overlap") short-circuit below)
                overlap_fail <- (est_method %in% c("dr", "ipw")) &&
                  overlap_check_fail(covariates, G, intercept_only)
                rcond_fail <- !overlap_fail && (est_method %in% c("dr", "reg")) &&
                  rcond_check_fail(covariates, G, intercept_only)
                guard <- list(overlap_fail = overlap_fail, rcond_fail = rcond_fail)
                if (use_check_cache) check_cache[[check_key]] <- guard
              }
              if (guard$overlap_fail) {
                warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                stop("overlap")
              }
              if (guard$rcond_fail) {
                warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                stop("singular")
              }
            }

            if (inherits(est_method, "function")) {
              # user-specified function
              res <- do.call(est_method, c(list(
                y1 = Ypost, y0 = Ypre,
                D = G,
                covariates = covariates,
                i.weights = w,
                inffunc = do_inf
              ), extra_args))
            } else if (est_method == "ipw") {
              # inverse-probability weights
              res <- DRDID::std_ipw_did_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            } else if (est_method == "reg") {
              # regression
              res <- DRDID::reg_did_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            } else {
              # doubly robust, this is default
              res <- DRDID::drdid_panel(Ypost, Ypre, G,
                covariates = covariates,
                i.weights = w,
                boot = FALSE, inffunc = do_inf
              )
            }
          }

          # adjust influence function to account for only using
          # subgroup to estimate att(g,t) (skipped for point estimates only)
          if (do_inf) {
            if (!is.null(fix_weights) && fix_weights == "varying") {
              # RC influence function has one entry per obs in disdat_long
              # (2 per unit: pre + post). Aggregate to unit level by ID,
              # independent of row ordering.
              res$att.inf.func <- as.numeric(rowsum(res$att.inf.func,
                                                    disdat_long[[idname]],
                                                    reorder = FALSE))
              # 1/2: the RC influence function is normalized over 2 obs/unit
              # (pre + post); folding to the unit level via rowsum() must divide
              # by 2 so the balanced-panel fix_weights = "varying" SE matches the
              # panel normalization (mirrors the force_rc fold in compute.att_gt2).
              res$att.inf.func <- 0.5 * (n / n1) * res$att.inf.func
            } else {
              res$att.inf.func <- (n / n1) * res$att.inf.func
            }
          }

          # If ATT is NaN, replace it with NA, and mark influence function as missing.
          if (is.nan(res$ATT)) {
            res$ATT <- NA
            if (do_inf) res$att.inf.func <- rep(NA_real_, length(res$att.inf.func))
          }
          res
        }, error = function(e) {
          # The overlap/rank guards above already emitted a full diagnostic warning
          # before signalling the internal "overlap"/"singular" sentinels, so skip
          # the wrapper warning for those: each failed cell warns exactly once,
          # matching the fast path. Genuine estimator errors are still surfaced.
          if (!(conditionMessage(e) %in% c("overlap", "singular"))) {
            warning("Error computing internal 2x2 DiD for (g, t) = (", glist[g], ", ", tlist[t + tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
          }
          NULL
        })

        if (is.null(attgt)) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      } else { # repeated cross sections / unbalanced panel

        if (use_precompute_rc) {
          # ---- positional precompute path (default) ----
          # .G / .C are row-wise functions of g_col, and gname is time-invariant
          # within .rowid (pre_process_did stops otherwise; for true repeated
          # cross sections .rowid is the row number), so a unit satisfies
          # .G == 1 | .C == 1 either in BOTH of this cell's periods or in
          # NEITHER. The legacy rightids/%in% construction below therefore
          # reduces to a row-wise G|C filter over the two periods, assembled
          # here positionally from the vectors precomputed before the loops --
          # skipping the two full data.table subsets, the full-data %in%, and
          # droplevels() per cell (covariates come positionally from global_mm,
          # so factor levels in `data` are never read). sort() restores original
          # data order, so the rows match which(disidx) on the legacy path
          # bit-identically. (The two periods are always distinct here: the
          # universal-base cell with tlist[pret] == tlist[t + tfac] is skipped
          # before this point.)
          rows_pair <- sort(c(period_rows_rc[[t + tfac]], period_rows_rc[[pret]]))
          g_cell <- g_col[rows_pair]
          Gv <- as.numeric(g_cell == current_g)
          # same formulas and types as the .G/.C column writes on the legacy path
          Cv <- if (nevertreated) {
            as.integer(g_cell == 0)
          } else {
            as.integer((g_cell == 0) |
              ((g_cell > time_threshold) & (g_cell != current_g)))
          }
          sel <- Gv == 1 | Cv == 1
          # positional index into `data` for slicing the global design matrix,
          # kept in sync with any later fix_weights row drops below.
          disdat_rows <- rows_pair[sel]
          G <- Gv[sel]
          C <- Cv[sel]
          Y <- y_col_rc[disdat_rows]
          post <- 1 * (t_col[disdat_rows] == tlist[t + tfac])
          cell_rowid <- rowid_col[disdat_rows]
        } else {
          # ---- legacy path (options(did.disable_precompute = TRUE)) ----
          # pick up the indices for units that will be used to compute ATT(g,t)
          # these conditions are (1) you are observed in the right period and
          # (2) you are in the right group (it is possible to be observed in
          # the right period but still not be part of the treated or control
          # group in that period here
          rightids <- disdat$.rowid[disdat$.G == 1 | disdat$.C == 1]

          # this is the fix for unbalanced panels; 2nd criteria shouldn't do anything
          # with true repeated cross sections, but should pick up the right time periods
          # only with unbalanced panel
          disidx <- (data$.rowid %in% rightids) & ((t_col == tlist[t + tfac]) | (t_col == tlist[pret]))

          # pick up the data that will be used to compute ATT(g,t)
          disdat <- data[disidx, ]
          # positional index into `data` for slicing the global design matrix, kept in
          # sync with any later row drops below (droplevels does not reorder/drop rows).
          # global_mm[disdat_rows, ] is the design for the rows used in this cell.
          disdat_rows <- which(disidx)

          # drop missing factors
          disdat <- droplevels(disdat)

          # give short names for data in this iteration
          G <- disdat$.G
          C <- disdat$.C
          Y <- disdat[[yname]]
          post <- 1 * (disdat[[tname]] == tlist[t + tfac])
          cell_rowid <- disdat$.rowid
        }
        # num obs. for computing ATT(g,t), have to be careful here
        n1 <- sum(G + C)

        # Handle fix_weights for RC/unbalanced panel
        if (!is.null(fix_weights) && fix_weights %in% c("base_period", "first_period")) {
          # Determine which period's weight to use
          if (fix_weights == "base_period") {
            target_idx <- pret_g
          } else {
            target_idx <- 1L
          }
          target_period <- tlist[target_idx]
          # Map each observation's weight from the target period by integer .rowid
          # matching (equivalent to the previous named-character lookup: first match
          # wins, unmatched -> NA), without coercing every .rowid to character.
          if (use_precompute_rc) {
            # per-period lookup vectors precomputed before the loops; the period
            # subsets preserve data order, so the match() result is identical to
            # the legacy full-table subset below.
            w <- target_w_by_period[[target_idx]][match(cell_rowid, target_rowid_by_period[[target_idx]])]
          } else {
            # Build lookup: weight from target period per unit
            target_rows <- data[t_col == target_period, ]
            w <- target_rows$.w[match(cell_rowid, target_rows$.rowid)]
          }
          # Drop units not observed in the target period
          missing_w <- is.na(w)
          if (any(missing_w)) {
            n_dropped <- length(unique(cell_rowid[missing_w]))
            warning(paste0("Dropped ", n_dropped, " units not observed in ",
                           fix_weights, " (period ", target_period, ") ",
                           "for group ", glist[g], " in time period ", tlist[t + tfac]))
            disdat_rows <- disdat_rows[!missing_w]
            cell_rowid <- cell_rowid[!missing_w]
            if (use_precompute_rc) {
              G <- G[!missing_w]
              C <- C[!missing_w]
              Y <- Y[!missing_w]
              post <- post[!missing_w]
            } else {
              disdat <- disdat[!missing_w, ]
              G <- disdat$.G
              C <- disdat$.C
              Y <- disdat[[yname]]
              post <- 1 * (disdat[[tname]] == tlist[t + tfac])
            }
            n1 <- sum(G + C)
            w <- w[!missing_w]
          }
        } else {
          w <- if (use_precompute_rc) w_col_rc[disdat_rows] else disdat$.w
        }

        #-----------------------------------------------------------------------------
        # checks to make sure that we have enough observations
        skip_this_att_gt <- FALSE
        if (sum(G * post) == 0) {
          warning(paste0("No units in group ", glist[g], " in time period ", tlist[t + tfac]))
          skip_this_att_gt <- TRUE
        }
        if (sum(G * (1 - post)) == 0) {
          # the empty period is the base (pre) period tlist[pret], not tlist[t]
          # (they differ under base_period = "universal" and for post-treatment
          # cells under "varying"); matches the fast path's pret_val
          warning(paste0("No units in group ", glist[g], " in time period ", tlist[pret]))
          skip_this_att_gt <- TRUE
        }
        if (sum(C * post) == 0) {
          warning(paste0("No available control units for group ", glist[g], " in time period ", tlist[t + tfac]))
          skip_this_att_gt <- TRUE
        }
        if (sum(C * (1 - post)) == 0) {
          # same base-period indexing fix as the group check above
          warning(paste0("No available control units for group ", glist[g], " in time period ", tlist[pret]))
          skip_this_att_gt <- TRUE
        }

        if (skip_this_att_gt) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          # inffunc[,counter] <- NA
          # counter <- counter+1
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )

          # Update the counters
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }


        # matrix of covariates: row-subset the global design by position.
        covariates <- global_mm[disdat_rows, , drop = FALSE]
        dimnames(covariates) <- list(NULL, colnames(covariates))

        #-----------------------------------------------------------------------------
        # code for actually computing att(g,t)
        #-----------------------------------------------------------------------------

        attgt <- tryCatch({
          # more checks for enough observations in each group; run inside the
          # tryCatch (with the same stop("overlap")/stop("singular") sentinels
          # as the panel branch) so that a failing preliminary fit degrades to
          # an NA cell with a warning instead of aborting the entire att_gt()
          if (!is.function(est_method)) {
            # checks for pscore based methods (no caching on the RC branch: the
            # cohort index varies cell by cell here)
            if (est_method %in% c("dr", "ipw")) {
              if (overlap_check_fail(covariates, G, intercept_only)) {
                warning(paste0("overlap condition violated for group ", glist[g], " in time period ", tlist[t + tfac]))
                stop("overlap")
              }
            }

            # check if can run regression using control units
            if (est_method %in% c("dr", "reg")) {
              if (rcond_check_fail(covariates, G, intercept_only)) {
                warning(paste0("Covariate matrix for control units is singular or numerically ill-conditioned for group ", glist[g], " in time period ", tlist[t + tfac], "; consider centering/rescaling covariates or removing collinear terms"))
                stop("singular")
              }
            }
          }

          if (inherits(est_method, "function")) {
            # user-specified function
            res <- do.call(est_method, c(list(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              inffunc = do_inf
            ), extra_args))
          } else if (est_method == "ipw") {
            # inverse-probability weights
            res <- DRDID::std_ipw_did_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          } else if (est_method == "reg") {
            # regression
            res <- DRDID::reg_did_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          } else {
            # doubly robust, this is default
            res <- DRDID::drdid_rc(
              y = Y,
              post = post,
              D = G,
              covariates = covariates,
              i.weights = w,
              boot = FALSE, inffunc = do_inf
            )
          }

          # n/n1 adjusts for estimating the att_gt only using observations from
          # groups G and C (skipped for point estimates only)
          if (do_inf) res$att.inf.func <- (n / n1) * res$att.inf.func

          # If ATT is NaN, replace it with NA, and mark influence function as missing
          if (is.nan(res$ATT)) {
            res$ATT <- NA
            if (do_inf) res$att.inf.func <- rep(NA_real_, length(res$att.inf.func))
          }
          res
        }, error = function(e) {
          # The overlap/rank guards above already emitted a full diagnostic warning
          # before signalling the internal "overlap"/"singular" sentinels, so skip
          # the wrapper warning for those: each failed cell warns exactly once,
          # matching the fast path. Genuine estimator errors are still surfaced.
          if (!(conditionMessage(e) %in% c("overlap", "singular"))) {
            warning("Error computing internal 2x2 DiD for (g, t) = (", glist[g], ", ", tlist[t + tfac], "): ", e$message, ". The ATT for this cell will be set to NA.")
          }
          NULL
        })

        if (is.null(attgt)) {
          attgt.list[[counter]] <- list(att = NA, group = glist[g], year = tlist[(t + tfac)], post = post.treat)
          if (do_inf) inffunc_updates[[update_counter]] <- list(
            indices = seq_len(n),
            values = rep(NA_real_, n)
          )
          update_counter <- update_counter + 1
          counter <- counter + 1
          next
        }
      } # end panel if

      # save results for this att(g,t)
      attgt.list[[counter]] <- list(
        att = attgt$ATT, group = glist[g], year = tlist[(t + tfac)], post = post.treat
      )


      # populate the influence function in the right places (skipped for point
      # estimates only -- compute_inffunc = FALSE -- so we neither aggregate nor
      # store the per-cell influence function)
      if (do_inf) {
      if (panel) {
        # Store integer indices directly (avoids which() later). On the precompute
        # path `kept` is exactly which(disidx) (disidx is not modified in between),
        # so reuse it; the get_wide_data fallback derives the index from its
        # wide-data disidx as before.
        idx <- if (use_precompute_panel) kept else which(disidx)
        inffunc_updates[[update_counter]] <- list(
          indices = idx,
          values = attgt$att.inf.func
        )
      } else {
        # aggregate inf functions by id (order by id)
        # Use the current cell ids (may differ from the initial selection if
        # fix_weights dropped obs). rowsum(reorder = TRUE) sums per id and returns
        # rows in ascending-id order, matching stats::aggregate() but ~40x faster
        # (see test-slowpath-rowsum).
        if (use_precompute_rc) {
          # every row of this cell satisfies .G == 1 | .C == 1 by construction
          # (the positional assembly selects on exactly that condition, and gname
          # is time-invariant within .rowid -- pre_process_did stops otherwise),
          # so the legacy G|C re-filter below is vacuous here. unique_rowid was
          # precomputed before the loops (loop-invariant).
          current_ids <- cell_rowid
          rs <- rowsum(attgt$att.inf.func, current_ids, reorder = TRUE)
          rs_ids <- sort(unique(current_ids))
          idx <- match(rs_ids, unique_rowid)
        } else {
          current_ids <- disdat$.rowid[disdat$.G == 1 | disdat$.C == 1]
          rs <- rowsum(attgt$att.inf.func, current_ids, reorder = TRUE)
          rs_ids <- sort(unique(current_ids))
          idx <- match(rs_ids, unique(data$.rowid))
        }
        inffunc_updates[[update_counter]] <- list(
          indices = idx,
          values = as.numeric(rs[, 1])
        )
      }
      }


      # save it in influence function matrix
      # inffunc[g,t,] <- inf.func
      # inffunc[,counter] <- inf.func
      update_counter <- update_counter + 1

      # update counter
      counter <- counter + 1
    } # end looping over t
  } # end looping over g

  # Build the influence function sparse matrix directly from triplets.
  # Gather per-column indices/values into lists and concatenate once, rather than
  # growing the triplet vectors with c() on every iteration (which reallocates
  # O(ncol^2) in total). The triplet order is unchanged (column-major, original
  # within-column order), so the resulting sparse matrix is identical.
  # Skipped for point estimates only (compute_inffunc = FALSE): no matrix is returned.
  if (do_inf) {
    nz_list <- lapply(inffunc_updates, `[[`, "indices")
    val_list <- lapply(inffunc_updates, `[[`, "values")
    trip_i <- unlist(nz_list, use.names = FALSE)
    trip_x <- unlist(val_list, use.names = FALSE)
    trip_j <- rep.int(seq_along(inffunc_updates), lengths(nz_list))
    inffunc <- Matrix::sparseMatrix(
      i = trip_i,
      j = trip_j,
      x = trip_x,
      dims = c(inffunc_nrow, length(inffunc_updates))
    )
  } else {
    inffunc <- NULL
  }

  return(list(attgt.list = attgt.list, inffunc = inffunc))
}

---
title: "Problems with two-way fixed-effects event-study regressions"
author: "Brantly Callaway and Pedro H. C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Problems with two-way fixed-effects event-study regressions} 
  %\VignetteEngine{knitr::rmarkdown} 
  %\VignetteEncoding{UTF-8}
---

```{r setup2, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  echo = TRUE,
  eval = FALSE
)
```


## Introduction

In this vignette, we want to illustrate the dangers of relying on traditional two-way fixed-effects (TWFE) regression models with leads and lags to conduct causal inference about treatment effect dynamics. In fact, we will show that *even* when treatment effect dynamics are homogeneous across different treatment groups/cohorts, some of these TWFE regression models can be severely biased. 


In order to effectively communicate these dangers, we consider a stylized simulation example where covariates do not play any major role. This example draws from several discussions we have had with Andrew Baker. We consider three setups. In the first one all units are eventually treated, and treatment effect dynamics are homogeneous across treatment groups. In the second, we have a group of units that remain untreated at the end of our sample, and treatment effect dynamics are homogeneous across treatment groups. The third one is like the second, but with heterogeneous treatment effect dynamics across treatment groups.


## Setup with all units being eventually treated and homogeneous treatment effect dynamics

Consider a staggered treatment adoption setup where 40 'states' ($state = \{1,2,\dots,40\}$) are randomly assigned into 4 treatment groups depending on the treatment starting year (1986, 1992, 1998, and 2004). We denote the treatment starting period as $g$ with ($g \in \{ 1986, 1992, 1998, 2004\}$). The 1000 units (e.g., 'counties' or 'firms') $i$ in the sample are randomly assigned to one of the 40 states. Let $G_i$ indicates the group/cohort unit $i$ belongs to, i.e., $G \subseteq \{ 1986, 1992, 1998, 2004\}$.


The data generating process (DGP) for the outcome $Y$ we consider is 
$$Y_{i,t} = (2010-g) + \alpha_i + \alpha_t + \tau_{i,t} + \epsilon_{i,t}$$
where $\alpha_i$ are unit fixed effects drawn from $\sim N(\mu_{state}, 1)$ with state-specific mean $\mu_{state} = state/5$, $\alpha_t$ are time fixed effects (cohort specific parallel time-trends) generated as $$\alpha_t = 0.1 * (t - g) + \epsilon^{time FE}_t, $$ with $\epsilon^{time FE}_t \sim N(0, 1)$, $\epsilon_{i,t} \sim N\left(0, \left(\frac{1}{2}\right)^2\right)$ is an idiosyncratic error term, and $\tau_{i,t}$ are the unit-specific treatment effects at time $t$ generated as 
$$ \tau_{i,t} = \mu \times (t - g + 1)\times 1\{t \ge g \},$$
where $\mu$ is the the instantaneous treatment effect; we set $\mu=1$.


From the above DGP, it is evident that the ATT at the time a unit is treated is equal to 1 for all groups, it is equal to 2 one period after treatment started for all groups, etc. Thus, the evolution of the treatment effects is homogeneous across treatment groups. This is an *a priori* strong restriction that should make things "easier" to estimate (but is by no means required!).

### Visualizing the DGP
Let's start the discussion by first visualizing how the DGP looks like.

```{r DGP_viz, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# Load libraries and set baseline parameters
library(tidyverse)
library(lfe)
library(fastDummies)
library(ggthemes)
library(did)
theme_set(theme_clean() + theme(plot.background = element_blank()))
#----------------------------------------------------------------------------
iseed  = 20201221
nrep <- 100  
true_mu <- 1
set.seed(iseed)

#----------------------------------------------------------------------------
## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobservd heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into four treatment groups G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse(year >= cohort_year, 1, 0),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error)
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data()

# plot
plot1 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value", color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))  +
  ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with all groups being eventually treated")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot1

#ggsave("plot_dgp1.png", plot1, width = 10, height = 5, dpi = 100)
```
![](plot_dgp1.png){ width=100% }

The above plot shows the 1,000 individual values of $Y_{i,t}$, as well as the average by treatment-group (the thicker colorful lines), and the vertical lines show the period when treatment begins for each group. Because treatment effects here grows linearly with time elapsed since treatment started, we can see that the earliest treated units end up with the highest value of Y, and that the difference grows by the gap between treatment years.


### Estimating dynamic treatment effects via TWFE event-study regressions

Given that we are interested in treatment effect dynamics, we then proceed to consider a classical two-way fixed-effects (TWFE) event study specification 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L} \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$ 
where $D_{i,t}^{k}=1\{t-G_i=k\}$ is an "event-study" dummy variable that takes value one if a unit $i$ is $k$ periods away from initial treatment at time $t$ and zero otherwise, $D_{i,t}^{<-K}=1\{t-G_i<-K\}$ and $D_{i,t}^{>L}=1\{t-G_i>L\}$ are defined analogously. For instance, $D_{i,t}^{0}$ is equal to one if the unit $i$ is first treated at time $t$, $D_{i,t}^{1}$ is equal to one if a one period has passed since treatment started (treatment lags), etc. Alternatively we have that $D_{i,t}^{-2}$ is equal to one if a unit $i$ will be treated in two periods from $t$ (treatment leads). In this exercise we set $K$ and $L$ to be equal to 5.


Up to today, it is customary to interpret estimates of $\gamma_k^{lags}$ as ``good'' measures of the average treatment effect for being exposed to treatment for $k$ periods, and estimates of $\gamma_k^{leads}$ as measures of pre-trends. Our first exercise here is to assess if this is OK-ish. 


Towards this end, we will a perform Monte Carlo exercise where we generate data according to the DGP described above and estimate all the parameters of the TWFE event-study regression above. We repeat this 100 times (same holds true if we do 1,000) and plot the Monte Carlo mean and 1.96 times the Monte-Carlo standard-deviation of the relative time indicator dummies, as well as the true treatment effect below.


```{r ES1, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse(rel_year < -5, 1, 0),
           Post = ifelse(rel_year > 5, 1, 0))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
  # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical <- map_dfr(1:nrep, run_ES_DiD)

colors <- c("True Effect" = "red", "Estimated Effect" = "blue")

ES_plot_classical <- data_classical %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
  geom_line(aes(color = 'Estimated Effect'), size = 1) + 
  geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("TWFE event-study regression with binned end-points")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical

#ggsave("es_plot_classical.png", ES_plot_classical, width = 10, height = 5, dpi = 150)
```

![](es_plot_classical.png){ width=100% }

The results above show that these TWFE event-study type estimates are *severely biased* for the true treatment effects. Furthermore, using the estimates of coefficient of treatment leads as a way to find evidence of "pre-trends" is very problematic, as illustrated above; see Sun and Abraham (2021) for formal results! Putting it simply, the results above highlight that such TWFE linear regression should not be used to highlight treatment effect dynamics!



Now, you may wonder: What happens if I include all possible leads and lags in the TWFE event study specification, i.e., to set K and L to the maximum allowable in the data, making inclusion of $D_{i,t}^{<-K}$
and of $D_{i,t}^{>L}$ unnecessary. In what follows, we do exactly that, though we need to omit one extra period apart from -1 to avoid problems of multicollinearity identified by Borusyak and Jaravel (2018). We drop the most negative time period, which is perhaps the most natural thing to do. In order words, we set $K=23$ and $L=24$, and drop the  $D_{i,t}^{<-K}$ and $D_{i,t}^{>L}$ from the TWFE specification.


The result of this exercise is displayed below. And, as you can see, this still leads to *very biased* estimates for the causal effects of interest!

```{r ES3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year")
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat <- map_dfr(1:nrep, run_ES_DiD_sat)

ES_plot_sat <- data_sat %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())
        

ES_plot_sat

#ggsave("es_plot_sat.png", ES_plot_sat, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat.png){ width=100% }


### Using the Callaway and Sant'Anna (2021) framework

Instead of using variants of the TWFE specification above, we advocate you to use the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our package **did**. 

In this specific DGP, as there is no "never-treated" units, we can use the "not-yet-treated" units as a valid comparison group for each cohort. Furthermore, since covariates are not available (and do not play any major role into the analysis), we do not need to worry about the underlying "est_method".

Below we show the results associated with using the Callaway and Sant'Anna (2021) procedure. As you can see, you get reliable and precise estimates for treatment effect dynamics. This is good news!
```{r CS, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS <- map_dfr(1:nrep, run_CS)

ES_plot_CS <- data_CS %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated")+
    scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS

#ggsave("es_plot_CS.png", ES_plot_CS, width = 10, height = 5, dpi = 150)

```
![](es_plot_CS.png){ width=100% }

## Setup with a group that remains untreated at the end of the sample

One of the features of the above DGP is that all units eventually get treated. In some applications, we have that a group of units remain untreated at the end of the sample. We call this group of units the "never-treated" group. To assess if the problems associated with TWFE specification highlighted above persist if such a group of units is available, we modify a little bit the DGP and repeat the same exercise as above.


### DGP with a "never-treated" group and homogeneous treatment effect dynamics

The DGP we consider is very similar to the one we have already considered. The only difference is that now we suppress treatment for the last cohort. Put it simply, the cohort with $g=2004$ is now "never-treated", implying that, in this DGP the unit-specific treatment effects at time $t$  $\tau_{i,t}$ are generated as 
$$ \tau_{i,t} = \mu \times (t - g + 1)\times 1\{t \ge g \}1\{g\not=2004 \},$$
where $\mu$ is the the instantaneous treatment effect; we set $\mu=1$.


Like before, the ATT at the time a unit is treated is 1, it is equal to 2 one period after treatment started, etc. 

### Visualizing the DGP
Let's start the discussion by first visualizing how the DGP looks like.

```{r DGP_viz2, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data2 <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobservd heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into 1\{t \ge g \}G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
    # Relabel 2004 cohort as never-treated
    mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data2()

# plot
plot2 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value",  color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
                     values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3"))+
  ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with a never-treated group")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot2 

#ggsave("plot_dgp_never_treated.png",plot2, width = 10, height = 5, dpi = 100)
```

![](plot_dgp_never_treated.png){ width=100% }

### Estimating dynamic treatment effects via TWFE event-study regressions

Like before, we start by estimating a TWFE event-study regression 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L}  \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$
with $K$ and $L$ set to 5. That is, binning the end points.

As you can see from the results below, the bias are still substantial. Thus, the problems with TWFE specifications persists even if you have a "never-treated" group.
```{r ES1_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>% 
    dummy_cols(select_columns = "rel_year") %>% 
    mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
           Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
    mutate(Pre = ifelse(is.na(Pre), 0, Pre),
           Post = ifelse(is.na(Post), 0, Post))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
 # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical_never <- map_dfr(1:nrep, run_ES_DiD_never)

ES_plot_classical_never <- data_classical_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with binned end-points")+
    scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical_never

#ggsave("es_plot_classical_never.png", ES_plot_classical_never, width = 10, height = 5, dpi = 150)
```
![](es_plot_classical_never.png){ width=100% }

Next, we consider the case where we one include almost all leads and lags ($K=17$ and $L=24$) but omit the furthest lead $K=18$. 


```{r ES3_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year2 = rel_year) %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    select(-("rel_year_-Inf"))
    
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
 # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat_never <- map_dfr(1:nrep, run_ES_DiD_sat_never)

ES_plot_sat_never <- data_sat_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_sat_never

#ggsave("es_plot_sat_never.png", ES_plot_sat_never, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat_never.png){ width=100% }

As the result above shows, when you have a never-treated group, treatment effects are homogeneous across cohorts, and you do not bin event-times, TWFE event-study regressions do recover the true underlying dynamic treatment effects. 


### Using the Callaway and Sant'Anna (2021) framework

Next, we show that the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our **did** can also be used in this setup.


In this specific DGP, as there is a "never-treated" group, we can use use different comparison groups: 1) the "never-treated" group, and 2) the "not-yet-treated" group. In this particular DGP, using any of these two different comparison groups should recover the true underlying dynamic treatment effect. Let's see if this is indeed the case.


First, let's estimate it using the "never-treated" units as the comparison group.

```{r CS_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "never_treated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_never <- map_dfr(1:nrep, run_CS_never)

ES_plot_CS_never <- data_CS_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_never

#ggsave("es_plot_CS_never.png", ES_plot_CS_never, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_never.png){ width=100% }
 
 As expected, the results look good!
 
 Now, let's consider the "not-yet-treated" units as comparison group.

```{r CS_ny, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

# function to run ES DID
run_CS_ny <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_ny <- map_dfr(1:nrep, run_CS_ny)

ES_plot_CS_ny <- data_CS_ny %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_ny

#ggsave("es_plot_CS_ny.png", ES_plot_CS_ny, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_ny.png){ width=100% }

As expected, the results look good, too


## Setup with Heterogeneous Treatment Effects

Now, we want to consider a setup with heterogeneous treatment effect dynamics and with a "never-treated" group. This DGP is a modification of the one we have considered above, where now the "unit-specific" treatment effect at time $t$ is given by
$$ \tau_{i,t} = \mu_g \times (t - g + 1)\times 1\{t \ge g \}1\{g\not=2004 \},$$
where we set $\mu_{1986} = 3$, $\mu_{1992} = 2$ and $\mu_{1998} = 1$. This implies that for the group who started treatment in 1986, its average treatment effects evolves with elapsed time as $3,6, 9,\dots$. For the group that started treatment in 1992 (1998), its average treatment effect evolves with elapsed time as $2,4, 6,\dots$ ($1,2, 3,\dots$). Of course, treatment effect is zero for the "never-treated" cohort ($g=2004$, which later we will relabel to $g=\infty$). This is a DGP where "early-treated" units benefit more from the treatment than "later-treated" units.

Given that each group has same size (on average), the true average treatment effect dynamic across treated groups is $(3+2+1)/3 = 2$ at the time a unit is treated, $2 \times (3+2+1)/3 = 4$ in the first period after treatment started, etc.

### Visualizing the DGP
Like before, let's start the discussion by visualizing how this DGP looks like.

```{r DGP_viz3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data3 <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobservd heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into 1\{t \ge g \}G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
           mu = ifelse(cohort_year==1992, 2, ifelse(cohort_year==1998, 1, 3)),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
    # Relabel 2004 cohort as never-treated
    mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data3()

# plot
plot3 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value",  color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
                     values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3")) +
  ggtitle("One draw of the DGP with heterogeneous treatment effect dynamics across cohorts \n and with a never-treated group")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot3 

#ggsave("plot_dgp_never_treated_het.png", plot3, width = 10, height = 5, dpi = 100)
```
![](plot_dgp_never_treated_het.png){ width=100% }

### Estimating dynamic treatment effects via TWFE event-study regressions

Like before, we start by estimating a TWFE event-study regression 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L}  \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$
with $K$ and $L$ set to 5. That is, binning the end points.

As you can see from the results below, the bias are still substantial. Given the results we have shown before, this should be not a surprise.
```{r ES1_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>% 
    dummy_cols(select_columns = "rel_year") %>% 
    mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
           Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
    mutate(Pre = ifelse(is.na(Pre), 0, Pre),
           Post = ifelse(is.na(Post), 0, Post))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
  # grab the obs we need
# grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical_never_het <- map_dfr(1:nrep, run_ES_DiD_never_het)

ES_plot_classical_never_het <- data_classical_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with binned end-points")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical_never_het

#ggsave("es_plot_classical_never_het.png", ES_plot_classical_never_het, width = 10, height = 5, dpi = 150)
```
![](es_plot_classical_never_het.png){ width=100% }

Next, we consider the case where one include almost all leads and lags ($K=17$ and $L=24$) but omit the furthest lead $K=18$. 


```{r ES3_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year2 = rel_year) %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    select(-("rel_year_-Inf"))
    
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
# grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat_never_het <- map_dfr(1:nrep, run_ES_DiD_sat_never_het)

ES_plot_sat_never_het <- data_sat_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_sat_never_het

#ggsave("es_plot_sat_never_het.png", ES_plot_sat_never_het, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat_never_het.png){ width=100% }

As the result above shows, when treatment effects dynamics are heterogeneous across groups, TWFE event-study regressions do not recover the true underlying dynamic treatment effects, even if a "never-treated" group is available and one does not bin extreme event-times. Again, this is a red flag for the usage of TWFE-type regressions in DiD setups. 

### Using the Callaway and Sant'Anna (2021) framework

Next, we show that the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our **did** can also be used in this setup. In other words, Callaway and Sant'Anna (2021) proposal *does not rely on restricting treatment effect heterogeneity*.


First, let's estimate the treatment effect dynamics using the "never-treated" units as the comparison group.

```{r CS_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "never_treated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_never_het <- map_dfr(1:nrep, run_CS_never_het)

ES_plot_CS_never_het <- data_CS_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_never_het

#ggsave("es_plot_CS_never_het.png", ES_plot_CS_never_het, width = 10, height = 5)
```
![](es_plot_CS_never_het.png){ width=100% }

As expected, the results look good!

 
Now, let's consider the "not-yet-treated" units as comparison group.

```{r CS_ny_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

# function to run ES DID
run_CS_ny_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_ny_het <- map_dfr(1:nrep, run_CS_ny_het)

ES_plot_CS_ny_het <- data_CS_ny_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_ny_het

#ggsave("es_plot_CS_ny_het.png", ES_plot_CS_ny_het, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_ny_het.png){ width=100% }

As expected, the results look good, too!

## Conclusion

In this vignette, we have shown that relying on TWFE event-study regressions can lead to very misleading conclusions about treatment effect dynamics. On the other hand, the approach put forward by Callaway and Sant'Anna (2021) and implement in our **did** package does not suffer from these drawbacks. 

## References

* [Borusyak, Kirill, and Xavier Jaravel. "Revisiting Event Study Designs". Available at SSRN 2826228 (2018)](https://scholar.harvard.edu/files/borusyak/files/borusyak_jaravel_event_studies.pdf)

* [Callaway, Brantly, and Pedro H. C. Sant'Anna. "Difference-in-differences with multiple time periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021](https://doi.org/10.1016/j.jeconom.2020.12.001).

* [Sun, Liyang, and Sarah Abraham. "Estimating dynamic treatment effects in event studies with heterogeneous treatment effects." Journal of Econometrics, Vol. 225, No. 2, pp. 175-199, 2021](https://doi.org/10.1016/j.jeconom.2020.09.006)

---
title: "Introduction to DiD with Multiple Time Periods"
author: "Brantly Callaway and Pedro H.C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to DiD with Multiple Time Periods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

Difference-in-differences is one of the most common approaches for identifying and estimating the causal effect of participating in a treatment on some outcome.

The "canonical" version of DiD involves two periods and two groups.  The *untreated group* never participates in the treatment, and the *treated group* becomes treated in the second period.

However, much applied work deals with cases where there are more than two time periods and different units can become treated at different points in time.  Regardless of the number of time periods, by far the leading approach in applied work is to try to estimate the effect of the treatment using a two-way fixed effects (TWFE) linear regression.  This works great in the case with two periods, but there are a number of recent methodological papers that suggest that there may be substantial drawbacks to using TWFE with multiple time periods.

This vignette briefly discusses the emerging literature on DiD with multiple time periods -- both issues with standard approaches as well as remedies for these potential problems.  The `did` package implements a number of these remedies.  A vignette for how to use the `did` package is available [here](did-basics.html). The background article for these vignettes is [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001).

## Background

To start with, we'll consider some background material in this section.  First, we'll discuss DiD with two time periods and two groups -- this is the "canonical" case of DiD.  Second, we briefly consider issues with TWFE linear regressions when there are multiple time periods.

### DiD with 2 Periods and 2 Groups

The baseline case for DiD is the one with two periods (let's call these periods $t$ and $t-1$) and two groups (a treated group and an untreated group).

**Notation / Setup**

* For $s \in \{t,t-1\}$, $Y_{is}(0)$ is unit $i$'s *untreated potential outcomes* -- this is the outcome that unit $i$ would experience in period $s$ if they *did not* participate in the treatment

* For $s \in \{t,t-1\}$, $Y_{is}(1)$ is unit $i$'s *treated potential outcome* -- this is the outcome that unit $i$ would experience in period $s$ if they *did* participate in the treatment.

* Set $D=1$ for units in the treated group and $D=0$ for units in the untreated group

* In the first period, no one participates in the treatment.  In the second period, units in the treated group become treated.  This means that observed outcomes are given by
$$
  Y_{it-1} = Y_{it-1}(0) \quad \textrm{and} \quad Y_{it} = D_i Y_{it}(1) + (1-D_i) Y_{it}(0)
$$
In other words, in the first period, we observe untreated potential outcomes for everyone (there is a no-anticipation assumption built in here).  In the second period, we observe treated potential outcomes for units that actually participate in the treatment and untreated potential outcomes for units that do not participate in the treatment.

* The main parameter of interest in most DiD designs is the Average Treatment Effect on the Treated (ATT).  It is given by
$$
  ATT = E[Y_t(1) - Y_t(0) | D=1]
$$
This is the difference between treated and untreated potential outcomes, on average, for units in the treated group.

The main assumption in DiD designs is called the parallel trends assumption:

**Parallel Trends Assumption**

$$
  E[Y_t(0) - Y_{t-1}(0)| D=1] = E[ Y_t(0)-Y_{t-1} | D=0]
$$

In words, this assumption says that the change (or "path") in outcomes over time that units in the treated group *would have experienced if they had not participated in the treatment* is the same as the path of outcomes that units in the untreated group actually experienced.  The parallel trends assumption allows for the level of untreated potential outcomes to differ across groups and is consistent with, for example, fixed effects models for untreated potential outcomes where the mean of the unobserved fixed effect can be different across groups.

This assumption is potentially useful because the path of untreated potential outcomes for units in the treated group (the term on the left in the above equation) is not known, but the researcher does observe the path of untreated potential outcomes for units in the untreated group (term on the right in the above equation).  In fact, it is straightforward to show that, under the parallel trends assumption, the $ATT$ is identified and given by
$$
  ATT = E[ Y_t - Y_{t-1}| D=1] - E[ Y_t - Y_{t-1}| D=0]
$$

That is, the $ATT$ is the difference between the mean change in outcomes over time experienced by units in the treated group adjusted by the mean change in outcomes over time experienced by units in the untreated group; the latter term, under the parallel trends assumption, is what the path of outcomes for units in the treated group would have been if they had not participated in the treatment.


### Two way fixed effects regressions

Now let's move to a more general case where there are $\mathcal{T}$ total time periods.  Denote particular time periods by $t$ where $t=1,\ldots,\mathcal{T}$.

By far the most common approach to *trying* to estimate the effect of a binary treatment in this setup is the TWFE linear regression.  This is a regression like
$$
Y_{it} = \theta_t + \eta_i + \alpha D_{it} + v_{it}
$$
where $\theta_t$ is a time fixed effect, $\eta_i$ is a unit fixed effect, $D_{it}$ is a treatment dummy variable, $v_{it}$ are time varying unobservables that are mean independent of everything else, and $\alpha$ is presumably the parameter of interest.  $\alpha$ is often interpreted as the *average effect* of participating in the treatment.
      
Although this is essentially a standard approach in applied work, there are a number of recent papers that point out potentially severe drawbacks of using the TWFE estimation procedure.  These include: Borusyak and Jaravel (2018), Goodman-Bacon (2021), de Chaisemartin and D'Haultfoeuille (2020), and Sun and Abraham (2021).

**When will TWFE work?**

1. Effects really aren't heterogeneous.  If the effect of participating in the treatment really is $\alpha$ for all units, TWFE will work great.  That being said, in many applications, treatment effects are very likely to be heterogeneous -- they may vary across different units or exhibit dynamics or change across different time periods.  In particular applications, this is worth thinking about, but, at least in our view, we think that heterogeneous effects of participating in some treatment is the leading case.

2. There are only two time periods.  This is the canonical case (2 periods, one group becomes treated in the second period, the other is never treated).  In this case, under parallel trends an no-anticipation, $\alpha$ is going to be numerically equal to the $ATT$.  In other words, in this case, even though it looks like you have restricted the effect of participating in the treatment to be the same across all units, TWFE exhibits *robustness* to treatment effect heterogeneity.  Unfortunately, this robustness to treatment effect heterogeneity does not continue to hold when there are more periods and groups become treated at different points in time.

**Why is TWFE not robust to treatment effect heterogeneity?**

There are entire papers written about this, see, e.g., Borusyak and Jaravel (2018), Goodman-Bacon (2021), de Chaisemartin and D'Haultfoeuille (2020), and Sun and Abraham (2021).  But here is the short version: in a TWFE regression, units whose treatment status doesn't change over time serve as the comparison group for units whose treatment status does change over time.  With multiple time periods and variation of treatment timing, some of these comparisons are:

* newly treated units relative to ``never treated'' units (good!)

* newly treated units relative to ``not-yet treated'' units (good!)

* newly treated units relative to already treated units (bad!!!)

The first of these two comparisons are good (or at least in the spirit of DiD) in that they take the path of outcomes experienced by units that become treated and adjust it by the path of outcomes experienced by units that are not participating in the treatment.  The third comparison is different though: it adjusts the path of outcomes for newly treated units by the path of outcomes for already treated units.  But this is not the path of untreated potential outcomes, it includes *treatment effect dynamics*.  Thus, these dynamics appear in $\alpha$, *making it very hard to give a clear causal interpretation*.

And this issue can have potentially severe consequences.  For example, it is possible to come up with examples where the effect of participating in the treatment is positive for all units in all time periods, but the TWFE estimation procedure leads to estimating a negative effect of participating in the treatment. Even in the case where ``negative weights'' can be ruled out, $\alpha$ recover a weighted average of $ATT's$, though these weights are hard to interpret.


## Treatment Effects in Difference in Differences Designs with Multiple Periods

In light of the potential problems with TWFE regressions in DiD designs with multiple periods, are there alternative approaches that can be used in this case?

Yes, and it turns out that it is not all that complicated! It is just a matter of using the ``good/desirable'' comparisons between groups instead of all possible comparisons.  

To fix ideas, let's provide some extended notation and be clear about the identifying assumptions that we are going to make.

**Notation**

* $Y_{it}(0)$ is unit $i$'s untreated potential outcome.  This is the outcome that unit $i$ would experience in period $t$ if they do not participate in the treatment.

* $Y_{it}(g)$ is unit $i$'s potential outcome in time period $t$ if they become treated in period $g$.

* $G_i$ is the time period when unit $i$ becomes treated (often *groups* are defined by the time period when a unit becomes treated; hence, the $G$ notation).

* $C_i$ is an indicator variable for whether unit $i$ is in a **never-treated** group.

* $D_{it}$ is an indicator variable for whether unit $i$ has been treated by time $t$.

* $Y_{it}$ is unit $i$'s observed outcome in time period $t$.  For units in the never-treated group, $Y_{it} = Y_{it}(0)$ in all time periods.  For units in other groups, we observe $Y_{it} = \mathbf{1}\{ G_i > t\} Y_{it}(0) + \mathbf{1}\{G_i \leq t \} Y_{it}(G_i)$.  The notation here is a bit complicated, but in words, we observe untreated potential outcomes for units that have not yet participated in the treatment, and we observe treated potential outcomes for units once they start to participate in the treatment (and these can depend on *when* they became treated). Implicit in this notation there is a **no treatment anticipation** assumption, which can be relaxed as discussed in [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001).

* $X_i$ vector of pre-treatment covariates.

### Main Assumptions
**Staggered Treatment Adoption Assumption**
Recall that $D_{it} = 1$ if a unit $i$ has been treated by time $t$ and $D_{it}=0$ otherwise.  Then, for $t=1,...,\mathcal{T}-1$, $D_{it} = 1 \implies D_{it+1} = 1$.

Staggered treatment adoption implies that once a unit participates in the treatment, they remain treated. In other words, units do not "forget" about their treatment experience. This is a leading case in many applications in economics.  For example, it would be the case for policies that roll out to different locations over some period of time.  It would also be the case for many unit-level treatments that have a "scarring" effect.  For example, in the context of job training, many applications consider participating in the treatment *ever* as defining treatment. 

Within the DiD context, we believe it is hard to analyze non-staggered treatment setups **without** further restricting treatment effect heterogeneity across time, groups, treatment sequences, etc. That is the main reason we focus on this leading case.

**Parallel Trends Assumption based on never-treated units**
For all $g=2,...,\mathcal{T}$, $t=2,...,\mathcal{T}$ with $t \ge g$,
$$
  E[ Y_t(0) - Y_{t-1}(0) | G=g] = E[ Y_t(0) - Y_{t-1}(0)| C=1]
$$

This is a natural extension of the parallel trends assumption in the two periods and two groups case.  It says that, in the absence of treatment, average untreated potential outcomes for the group first treated in time $g$ and for the "never treated" group would have followed parallel paths in all post-treatment periods $t \ge g$.

Note that the aforementioned parallel trend assumption rely on using the ``never treated'' units as comparison group for all "eventually treated" groups. This presumes that (i) a (large enough) "never-treated" group is available in the data, and (ii) these units are "similar enough" to the eventually treated units such that they can indeed be used as a valid comparison group. In situations where these conditions are not satisfied, one can use an alternative parallel trends assumption that uses the **not-yet treated** units as valid comparison groups. 

**Parallel Trends Assumption based on not-yet treated units**
For all $g=2,...,\mathcal{T}$, $s,t=2,...,\mathcal{T}$ with $t \ge g$ and $s \ge t$
$$
  E[ Y_t(0) - Y_{t-1}(0) | G=g] = E[ Y_t(0) - Y_{t-1}(0)| D_s=0, G\not=g]
$$
In plain English, this assumption states that one can use the not-yet-treated by time $s$ ($s \ge t$) units as valid comparison groups when computing the average treatment effect for the group first treated in time $g$. In general, this assumption uses more data when constructing comparison groups. However, as noted in [Marcus and Sant'Anna (2021)](https://www.journals.uchicago.edu/doi/full/10.1086/711509), this assumption does restrict some pre-treatment trends across different groups. In other words, there is no free-lunch.



## Group-Time Average Treatment Effects

The above assumptions are natural extensions of the identifying assumptions in the two periods and two groups case to the multiple periods case.  

Likewise, a natural way to generalize the parameter of interest (the ATT) from the two periods and two groups case to the multiple periods case is to define **group-time average treatment effects**:

$$
  ATT(g,t) = E[Y_t(g) - Y_t(0) | G=g]
$$

This is the average effect of participating in the treatment for units in group $g$ at time period $t$.  Notice that when there are two time periods and two groups (the canonical case), the average treatment effect on the treated is given by $ATT = ATT(g=2,t=2)$.

To give a couple more examples, suppose that a researcher has access to three time periods.  Then, $ATT(g=2,t=3)$ is the average effect of participating in the treatment for the group of units that become treated in time period 2, in time period 3.  Similarly, $ATT(g=3,t=3)$ is the average effect of participating in the treatment for the group of units that become treated in time period 3, in time period 3.

**Identification of Group-Time Average Treatment Effects**

Under either version of the parallel trends assumptions mentioned above, it is straightforward to show that group-time average treatment effects are identified. For instance, when one impose the parallel trends assumption based on "never-treated units", we have that, for all $t \ge g$
$$
ATT(g,t) = E[ Y_t - Y_{g-1}| G=g] - E[ Y_t - Y_{g-1}| C=1].
$$
Alternatively, when one impose the parallel trends assumption based on "not-yet-treated units", we have that, for all $t \ge g$
$$
ATT(g,t) = E[ Y_t - Y_{g-1}| G=g] - E[ Y_t - Y_{g-1}| D_t=0, G\not=g].
$$

These group-time average treatment effects are the building blocks of understanding the effect of participating in a treatment in DiD designs with multiple time periods.

## Parallel Trends Conditional on Covariates

In many cases, the parallel trends assumption is substantially more plausible if it holds after conditioning on observed pre-treatment covariates.  In other words, if the parallel trends assumptions are modified to be

**Conditional Parallel Trends Assumption based on never-treated units**
For all $g=2,...,\mathcal{T}$, $t=2,...,\mathcal{T}$ with $t \ge g$,
$$
  E[ Y_t(0) - Y_{t-1}(0) |X, G=g] = E[ Y_t(0) - Y_{t-1}(0)| X, C=1]
$$

**Parallel Trends Assumption based on not-yet treated units**
For all $g=2,...,\mathcal{T}$, $s,t=2,...,\mathcal{T}$ with $t \ge g$ and $s \ge t$
$$
  E[ Y_t(0) - Y_{t-1}(0) | X, G=g] = E[ Y_t(0) - Y_{t-1}(0)|X, D_s=0, G\not=g]
$$

These parallel trends assumptions are the conditional analogues of previous ones. Importantly, they allow for covariate-specific trends in outcomes across groups, which can be particularly important in setups where the distribution of covariates varies across groups.

An example of a case where this assumption is attractive is one where a researcher is interested in estimating the effect of participating in job training on earnings.  In that case, if the path of earnings (in the absence of participating in job training) depends on things like education, previous occupation, or years of experience (which it almost certainly does), then it would be important to condition on these types of variables in order to make parallel trends more credible.  

In this case, the parameter of interest is still often the $ATT(g,t)'s$ (or their aggregation).  It is still straightforward to identify and estimate the $ATT$ in this case.  Basically, one needs to estimate the change in outcomes for units in the untreated group conditional on $X$, but average out $X$ over the distribution of covariates for individuals in group $g$ to obtain $ATT(g,t)$ (see [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001) and references therein for many more details).  In practice, you can use different approaches to recover these parameters. More precisely, you can estimate the $ATT(g,t)'s$ using outcome-regressions, inverse probability weighting, or doubly-robust methods. But the **did** package automates all of this for the user.

## Aggregating Group-Time Average Treatment Effects

Group-time average treatment effects are natural parameters to identify in the context of DiD with multiple periods and multiple groups.  But in many applications, there may be a lot of them.  There are some benefits and costs here.  The main benefit is that it is relatively straightforward to think about heterogeneous effects across groups and time using group-time average treatment effects.  On the other hand, it can be hard to summarize them (e.g., they are not just a single number).

In our paper, [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001), we propose a number of ways to aggregate group-time average treatment effects.  Here, we will just consider a few important ones that we think applied researchers are most often interested in.  First, consider the average effect of participating in the treatment, separately for each group.  This is given by

$$
  \theta_S(g) = \frac{1}{\mathcal{T} - g + 1} \sum_{t=2}^{\mathcal{T}} \mathbf{1}\{g \leq t\} ATT(g,t).
$$

This parameter may be of interest in its own right, since it allows one to highlight treatment effect heterogeneity with respect to treatment adoption period. Furthermore, it is fairly straightforward to further aggregate $\theta_S(g)$ to get an easy-to-interpret overall effect parameter,

$$
  \theta^O_S := \sum_{g=2}^{\mathcal{T}} \theta_S(g) P(G=g).
$$

$\theta^O_S$ is the overall effect of participating in the treatment across all groups that have ever participated in the treatment.  In our view, this is close to being a multi-period analogue of the $ATT$ in the two period case. Thus, if a researcher is constrained to report a single treatment effect summary parameter, we recommend reporting $\theta^O_S$.


In DiD setups with multiple periods, it is natural to ask "How does treatment effects vary with elapsed treatment time?" Here, note that researchers are interested in understanding treatment effect dynamics. This is at the heart of event-study-type of analysis that is widespread in applied work. 

In this case, a natural way to aggregate the group-time average treatment effect to highlight treatment effect dynamics is given by

$$
  \theta_D(e) := \sum_{g=2}^{\mathcal{T}} \mathbf{1} \{ g + e \leq \mathcal{T} \} ATT(g,g+e) P(G=g | G+e \leq \mathcal{T}).
$$

This is the average effect of participating in the treatment for the group of units that have been exposed to the treatment for exactly $e$ time periods.  

All of these aggregations are available in the **did** package and examples with real data are available in our [Getting Started with the **did** Package](did-basics.html) vignette. In [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001), we also discuss additional aggregation schemes. We encourage you to take a look!

## Conclusion

This vignette has covered basic background issues on DiD with multiple periods.  [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001) discusses many extensions and these are all provided in the **did** package as well. [See our User Guides for more details.](index.html)

## References

* [Borusyak, Kirill, and Xavier Jaravel. "Revisiting Event Study Designs". Available at SSRN 2826228 (2018)](https://scholar.harvard.edu/files/borusyak/files/borusyak_jaravel_event_studies.pdf)

* [Callaway, Brantly, and Pedro H. C. Sant'Anna. "Difference-in-differences with multiple time periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021](https://doi.org/10.1016/j.jeconom.2020.12.001).

* [de Chaisemartin, Clement, and Xavier d'Haultfoeuille. "Two-way fixed effects estimators with heterogeneous treatment effects." American Economic Review 110.9 (2020): 2964-96.](https://doi.org/10.1257/aer.20181169)

* [Goodman-Bacon, Andrew. Difference-in-differences with variation in treatment timing." Journal of Econometrics, Vol. 225, No. 2, pp. 254-277, 2021](https://doi.org/10.1016/j.jeconom.2021.03.014)

* [Marcus, Michelle, and Pedro H. C. Sant'Anna. "The Role of Parallel Trends in Event Study Settings: An Application to Environmental Economics". Journal of the Association of Environmental and Resource Economists, Vol. 8, No. 2, pp. 235-275, 2021](https://www.journals.uchicago.edu/doi/10.1086/711509)

* [Sun, Liyang, and Sarah Abraham. "Estimating dynamic treatment effects in event studies with heterogeneous treatment effects." Journal of Econometrics, Vol. 225, No. 2, pp. 175-199, 2021](https://doi.org/10.1016/j.jeconom.2020.09.006)

---
title: "Writing Extensions to the did Package"
author: "Brantly Callaway and Pedro H.C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Writing Extensions to the did Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, echo=FALSE, results="hide", warning=FALSE, message=FALSE}
library(did)
# Source the currently version of the did package (based on our Dropbox)
# fldr <- here::here("R/")
# sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
source(here::here("vignettes/setup_sims.R"))
```

## Introduction

This vignette provides an example of extending the ideas of the **did** package to some other cases.

* We see the central idea of our approach to DiD to be for researchers to carefully make the comparisons that they want to make.  It can be tricky to operationalize this though.  And our code does not come close to covering all possible cases where one would like for identification to hold under some type of parallel trends assumption.

* In this vignette, we demonstrate how to do difference in differences when individuals anticipate participating in the treatment and potentially adjust their behavior before they participate.

* Although we now allow for treatment anticipation in the **did** package using the `anticipation` argument, we still think that it is helpful to show a simplified version of the code behind the **did** package so that researchers can potentially borrow/modify it for one-off DiD applications that do not conform exactly to the expected setup in the **did** package.  You will see that it's not too complicated!

* For this vignette, there's some math.  The notation may be self-explanatory, but see our [Introduction to DiD with Multiple Time Periods](multi-period-did.html) for precise explanations of the notation.

## DiD with Anticipation

The example we give next is one where individuals **anticipate** participating in the treatment and adjust their behavior (or at least their outcomes change) before they actually participate in the treatment.  This is a feature of some applications in economics -- for example, in labor applications such as job training or job displacement, individual's earnings often "dip" just before they actually participate in the treatment.  One solution to this is to impose that parallel trends holds, but adjust the periods over which it holds.  One version of this is to make the assumption that

**Parallel Trends with Anticipation** For all groups and time periods such that $t \ge g-1$
$$
  E[Y_t(0) - Y_{g-2}(0) | G=g] = E[Y_t(0) - Y_{g-2}(0) | C=1] 
$$

In other words, the path of outcomes, in the absence of participating in the treatment, between periods $(g-2)$ and $t$ is the same for individuals in group $g$ (which is not observed) as for untreated individuals (which is observed).

In this case, it is straightforward to show that, for all $t \ge g-1$,
$$
  ATT(g,t) = E[Y_t - Y_{g-2} | G=g] - E[Y_t - Y_{g-2} | C=1]
$$

This is very similar to the main case covered in the **did** package (see discussion in our [Introduction to DiD with Multiple Time Periods](multi-period-did.html) vignette), except that the "base period" here is $(g-2)$ (two periods before individuals in group $g$ become treated) rather than $(g-1)$ (one period before individuals in group $g$ become treated) which is the setup used in the **did** package.

If you are worried that that units anticipate treatment participation two periods before treatment actually starts, you could address this by setting the base period to be $(g-3)$ instead.  It is worth pointing out, however, that there is a tradeoff.  Allowing for more periods of anticipation will require that the parallel trends assumption hold over more periods in order to identify the group-time average treatment effects.  At any rate, for this vignette, we stick with the case where a researcher wants to allow anticipation exactly one period before units participate in the treatment.


### Computing Treatment Effects under DiD with Anticipation

To start with, we want to point out that the important part for this vignette is not so much this particular application, but rather just to demonstrate how to write the code for this particular case.  

* The code below should be instructive for extending these sorts of ideas to other cases that may show up in applications.

* We'll write a simplified version of the code -- it will not be as fast as the code in the **did** package, but it will demonstrate that it is relatively straightforward to write the code for modified versions of our approach.  [BTW, the code below is not too slow either, basically things will run with a few thousand observations in a couple minutes rather than a couple seconds.]

To start with, we've constructed a dataset (called `dta`) that has anticipation.  All groups follow parallel trends (in the absence of treatment) except in the period immediately before treatment (due to treatment anticipation). In those periods, outcomes for individuals who participate in the treatment "dip" -- here, they decrease by 1 in the pre-treatment period.  In post-treatment periods, the effect of participating in the treatment is equal to 1.  This is simulated data, but it has features that would be common in, for example, an application evaluating the effect of a job training program.

```{r echo=FALSE}
time.periods <- 5
reset.sim()
te <- 0
te.e <- c(-1,rep(1,time.periods-1))
bett <- betu <- rep(0,time.periods)
set.seed(12345)

# generate data set with these parameters
dta <- build_sim_dataset()
dta$G <- ifelse(dta$G==0, 0, dta$G+1)
dta$G <- ifelse(dta$G==6, 0, dta$G)
```

Here is what the data looks like

```{r}
nrow(dta)
head(dta)
```

where `G` defines the period when an individual first becomes treated, `X` is a covariate but we'll ignore it (parallel trends holds in this example without having to condition on `X`), and `Y` is the outcome.  

As a first step, we'll try estimating the effect of participating in the treatment (focus on dynamic effects as in an event study plot) while just ignoring the possibility of anticipation.

```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# estimate group-time average treatment effects using att_gt method
# (and ignoring pre-treatment "dip")
attgt.ignoredip <- att_gt(yname = "Y",
                          tname = "period",
                          idname = "id",
                          gname = "G",
                          xformla = ~1,
                          data = dta,
                          )

# summarize the results
summary(attgt.ignoredip)

# make dynamic effects plot
p <- ggdid(aggte(attgt.ignoredip, "dynamic"))

# add actual treatment effects to the plot
library(ggplot2)
truth <- cbind.data.frame(e = seq(-3,2), att.e = c(0,0,-1,1,1,1))
p <- p + geom_line(data = truth, aes(x = e, y = att.e), inherit.aes = FALSE, color = "blue")
p <- p + geom_point(data = truth, aes(x = e, y = att.e), inherit.aes = FALSE, color = "blue")
p
```

In the figure, the blue line is the "truth".  You can see that anticipation leads to two things. First, it leads to rejecting the parallel trends assumptions -- you can correctly visualize the anticipation effects here and this is quite helpful for understanding what is going on.  Second, however, the anticipation effects feed into the treatment effect estimates.  Here, we over-estimate the effect of participating in the treatment due to the anticipation.  Also, note that this example is really simple, and it would be easy to come up with cases where the results were much more complicated by anticipation.

Next we want to write our own code for getting around anticipation.  A few notes:

* The goal here is to write a simple version of the code -- it's possible to write faster code than what is given below, but there is a trade-off between code running fast and the amount of time spent writing it.  If we were writing a one-off project that had anticipation, this is how we'd do it.  The way we implemented this feature to the **did** package, on the other hand, is faster.

* The code below is not (too much) optimized for `R`.  We have tried to write code that is easy to understand and (close to being) portable to another programming language without too much trouble.

* Code below computes an event-study-type estimator, but we could easily compute some other aggregation or look at group-time average treatment effects if that's what we wanted.

Now to the code.  We'll start by writing a function to compute group-time average treatment effects and dynamic effects given some dataset (we'll rely on the dataset having variables called `G`, `id`, `period`, and `Y`, but this would be easy to modify).  Again, this function is a simplified version of what's going on behind the scenes in the **did** package.

```{r}
compute.attgt <- function(dta) {
  # pick up all groups
  groups <- unique(dta$G)

  # pick up all time periods
  time.periods <- unique(dta$period)

  # sort the groups and drop the untreated group
  groups <- sort(groups)[-1]

  # sort the time periods and drop the first two
  # (can't compute treatment effects for these two
  # periods with one-period anticipation -- w/o anticipation
  # we would just drop one period here)
  time.periods <- sort(time.periods)[-c(1,2)]

  # drop last time period (because we don't know if
  # these units are affected by anticipation or not
  # and we are being very careful)
  # (if you were worried about more than one anticipation
  # period here, would need to drop more time periods
  # from the end)
  time.periods <- time.periods[-length(time.periods)]

  # list to store all group-time average treatment effects
  # that we calculate
  attgt.list <- list()
  counter <- 1

  # loop over all groups
  for (g in groups) {

    # get the correct "base" period for this group
    # (subtract 2 to avoid anticipation)
    main.base.period <- g - 2
    
    # loop over all time periods
    for (tp in time.periods) {

      #----------------------------------------------------
      # if it's a pre-treatment time period (used for the
      # pre-test, we need to adjust the base period)

      # group not treated yet
      if (tp < g) {
        # move two periods before
        base.period <- tp - 2
      } else {
        # this is a post-treatment period
        base.period <- main.base.period
      }
      #----------------------------------------------------



      #----------------------------------------------------
      # now, all we need to do is collect the right subset
      # of the data and estimate a 2x2 DiD

      # get group g and untreated group
      this.data <- subset(dta, G==g | G==0)

      # get current period and base period data
      this.data <- subset(this.data, period==tp | period==base.period)

      # set up to compute 2x2 estimator
      Ypost <- subset(this.data, period==tp)$Y
      Ypre <- subset(this.data, period==base.period)$Y

      # dummy variable being in group g
      G <- 1*(subset(this.data, period==tp)$G == g)

      # compute 2x2 estimator using DRDID package
      # (in this unconditional case, it would be straightforward
      # to calculate the 2x2 att just using averages, but we
      # like the DRDID package as it will work for more complicated
      # cases as well)
      attgt <- DRDID::reg_did_panel(Ypost, Ypre, G, covariates=NULL)$ATT

      # save results
      attgt.list[[counter]] <- list(att=attgt, group=g, time.period=tp)
      counter <- counter+1
      #----------------------------------------------------

    }
  }

  #-----------------------------------------------------------------------------
  # aggregate into dynamic effects
  
  # turn results into a data.frame
  attgt.results <- do.call("rbind.data.frame", attgt.list)

  # add event time to the results
  attgt.results$e <- attgt.results$time.period - attgt.results$group
  
  # calculate relative sizes of each group
  # (will be used as weights)
  n.group <- sapply(groups, function(gg) nrow(subset(dta, G==gg)))
  # merge in group sizes
  ngroup.mat <- cbind(groups, n.group)
  attgt.results <- merge(attgt.results, ngroup.mat, by.x = "group", by.y = "groups")

  # event times to calculate dynamic effects
  eseq <- unique(attgt.results$e) 
  eseq <- sort(eseq)

  # calculate average effects by event time
  att.e <- c()
  counter <- 1
  for (this.e in eseq) {
    # get subset of results at this event time
    res.e <- subset(attgt.results, e==this.e)

    # calculate weights by group size
    res.e$weight <- res.e$n.group / sum(res.e$n.group)

    # calculate dynamic effect as weighted average
    att.e[counter] <- sum(res.e$att * res.e$weight)

    # on to the next one
    counter <- counter+1
  }

  # store dynamic effects results
  dyn.results <- cbind.data.frame(e = eseq, att.e = att.e)

  # return group-time average treatment effects and dynamic effects
  return(list(attgt.results=attgt.results[,c("group","att","time.period")],
              dyn.results=dyn.results))
}
```

Now, we have a function to compute group-time average treatment effects and dynamic effects.  Let's use it on the data that we have

```{r}
anticipation.results <- compute.attgt(dta)
anticipation.results
```

Finally, let's compute some standard errors using the nonparametric bootstrap (which is different and much slower than the multiplier bootstrap that we use in the *did* package); we'll just focus on the dynamic effects estimator

```{r bootstrap, eval=FALSE}
# the number of bootstrap iterations
biters <- 100

# list to store bootstrap results
boot.res <- list()

# loop for each nonparametric bootstrap iteration
for (b in 1:biters) {
  # draw a bootstrap sample; here, we'll call an outside function
  bdata <- BMisc::blockBootSample(dta, "id")

  # call our function for estimating dynamic effects on the
  # bootstrapped data
  boot.res[[b]] <- compute.attgt(bdata)$dyn.results$att.e
}

# use the bootstrapped results to compute standard errors
boot.res <- t(simplify2array(boot.res))
boot.se <- apply(boot.res, 2, sd)

# add the standard errors to the main results
dyn.results <- anticipation.results$dyn.results
dyn.results$att.se <- boot.se
```

Finally, we can translate these into a plot.  [One last thing here, to keep things very simple, we are going to plot pointwise confidence intervals, but we strongly suggest computing simultaneous confidence bands in practice.]

```{r, eval=FALSE}
p <- ggplot(data=dyn.results, aes(x=e, y=att.e)) +
  geom_line() +
  geom_point() +
  geom_errorbar(aes(ymin=(att.e-1.96*att.se), ymax=(att.e+1.96*att.se)), width=0.1) +
  theme_bw()

p <- p + geom_line(data=truth, aes(x=e, y=att.e), inherit.aes=FALSE, color="blue")
p <- p + geom_point(data=truth, aes(x=e, y=att.e), inherit.aes=FALSE, color="blue")
p

```

![](extensions_anticipation.png){ width=90% }

Here, the black line is our estimate and the blue line is the "truth".  You can immediately see that we get things right here when we account for anticipation.  The only other thing to notice is that we do not estimate dynamic effects at as many different lengths of exposure, but that is the price to pay for accounting for anticipation.

## Conclusion

In this vignette, we illustrated how one could apply the principles in the **did** package to the case where units anticipate participating in the treatment.  This is a relatively straightforward exercise, and we think that showing this sort of extension is potentially useful for other situations that do not fit exactly into the cases covered by the **did** package.

---
title: "Introduction to DiD with Multiple Time Periods"
author: "Brantly Callaway and Pedro H.C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to DiD with Multiple Time Periods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

Difference-in-differences is one of the most common approaches for identifying and estimating the causal effect of participating in a treatment on some outcome.

The "canonical" version of DiD involves two periods and two groups.  The *untreated group* never participates in the treatment, and the *treated group* becomes treated in the second period.

However, much applied work deals with cases where there are more than two time periods and different units can become treated at different points in time.  Regardless of the number of time periods, by far the leading approach in applied work is to try to estimate the effect of the treatment using a two-way fixed effects (TWFE) linear regression.  This works great in the case with two periods, but there are a number of recent methodological papers that suggest that there may be substantial drawbacks to using TWFE with multiple time periods.

This vignette briefly discusses the emerging literature on DiD with multiple time periods -- both issues with standard approaches as well as remedies for these potential problems.  The `did` package implements a number of these remedies.  A vignette for how to use the `did` package is available [here](did-basics.html). The background article for these vignettes is [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001).

## Background

To start with, we'll consider some background material in this section.  First, we'll discuss DiD with two time periods and two groups -- this is the "canonical" case of DiD.  Second, we briefly consider issues with TWFE linear regressions when there are multiple time periods.

### DiD with 2 Periods and 2 Groups

The baseline case for DiD is the one with two periods (let's call these periods $t$ and $t-1$) and two groups (a treated group and an untreated group).

**Notation / Setup**

* For $s \in \{t,t-1\}$, $Y_{is}(0)$ is unit $i$'s *untreated potential outcomes* -- this is the outcome that unit $i$ would experience in period $s$ if they *did not* participate in the treatment

* For $s \in \{t,t-1\}$, $Y_{is}(1)$ is unit $i$'s *treated potential outcome* -- this is the outcome that unit $i$ would experience in period $s$ if they *did* participate in the treatment.

* Set $D=1$ for units in the treated group and $D=0$ for units in the untreated group

* In the first period, no one participates in the treatment.  In the second period, units in the treated group become treated.  This means that observed outcomes are given by
$$
  Y_{it-1} = Y_{it-1}(0) \quad \textrm{and} \quad Y_{it} = D_i Y_{it}(1) + (1-D_i) Y_{it}(0)
$$
In other words, in the first period, we observe untreated potential outcomes for everyone (there is a no-anticipation assumption built in here).  In the second period, we observe treated potential outcomes for units that actually participate in the treatment and untreated potential outcomes for units that do not participate in the treatment.

* The main parameter of interest in most DiD designs is the Average Treatment Effect on the Treated (ATT).  It is given by
$$
  ATT = E[Y_t(1) - Y_t(0) | D=1]
$$
This is the difference between treated and untreated potential outcomes, on average, for units in the treated group.

The main assumption in DiD designs is called the parallel trends assumption:

**Parallel Trends Assumption**

$$
  E[Y_t(0) - Y_{t-1}(0)| D=1] = E[ Y_t(0)-Y_{t-1} | D=0]
$$

In words, this assumption says that the change (or "path") in outcomes over time that units in the treated group *would have experienced if they had not participated in the treatment* is the same as the path of outcomes that units in the untreated group actually experienced.  The parallel trends assumption allows for the level of untreated potential outcomes to differ across groups and is consistent with, for example, fixed effects models for untreated potential outcomes where the mean of the unobserved fixed effect can be different across groups.

This assumption is potentially useful because the path of untreated potential outcomes for units in the treated group (the term on the left in the above equation) is not known, but the researcher does observe the path of untreated potential outcomes for units in the untreated group (term on the right in the above equation).  In fact, it is straightforward to show that, under the parallel trends assumption, the $ATT$ is identified and given by
$$
  ATT = E[ Y_t - Y_{t-1}| D=1] - E[ Y_t - Y_{t-1}| D=0]
$$

That is, the $ATT$ is the difference between the mean change in outcomes over time experienced by units in the treated group adjusted by the mean change in outcomes over time experienced by units in the untreated group; the latter term, under the parallel trends assumption, is what the path of outcomes for units in the treated group would have been if they had not participated in the treatment.


### Two way fixed effects regressions

Now let's move to a more general case where there are $\mathcal{T}$ total time periods.  Denote particular time periods by $t$ where $t=1,\ldots,\mathcal{T}$.

By far the most common approach to *trying* to estimate the effect of a binary treatment in this setup is the TWFE linear regression.  This is a regression like
$$
Y_{it} = \theta_t + \eta_i + \alpha D_{it} + v_{it}
$$
where $\theta_t$ is a time fixed effect, $\eta_i$ is a unit fixed effect, $D_{it}$ is a treatment dummy variable, $v_{it}$ are time varying unobservables that are mean independent of everything else, and $\alpha$ is presumably the parameter of interest.  $\alpha$ is often interpreted as the *average effect* of participating in the treatment.
      
Although this is essentially a standard approach in applied work, there are a number of recent papers that point out potentially severe drawbacks of using the TWFE estimation procedure.  These include: Borusyak and Jaravel (2018), Goodman-Bacon (2021), de Chaisemartin and D'Haultfoeuille (2020), and Sun and Abraham (2021).

**When will TWFE work?**

1. Effects really aren't heterogeneous.  If the effect of participating in the treatment really is $\alpha$ for all units, TWFE will work great.  That being said, in many applications, treatment effects are very likely to be heterogeneous -- they may vary across different units or exhibit dynamics or change across different time periods.  In particular applications, this is worth thinking about, but, at least in our view, we think that heterogeneous effects of participating in some treatment is the leading case.

2. There are only two time periods.  This is the canonical case (2 periods, one group becomes treated in the second period, the other is never treated).  In this case, under parallel trends an no-anticipation, $\alpha$ is going to be numerically equal to the $ATT$.  In other words, in this case, even though it looks like you have restricted the effect of participating in the treatment to be the same across all units, TWFE exhibits *robustness* to treatment effect heterogeneity.  Unfortunately, this robustness to treatment effect heterogeneity does not continue to hold when there are more periods and groups become treated at different points in time.

**Why is TWFE not robust to treatment effect heterogeneity?**

There are entire papers written about this, see, e.g., Borusyak and Jaravel (2018), Goodman-Bacon (2021), de Chaisemartin and D'Haultfoeuille (2020), and Sun and Abraham (2021).  But here is the short version: in a TWFE regression, units whose treatment status doesn't change over time serve as the comparison group for units whose treatment status does change over time.  With multiple time periods and variation of treatment timing, some of these comparisons are:

* newly treated units relative to ``never treated'' units (good!)

* newly treated units relative to ``not-yet treated'' units (good!)

* newly treated units relative to already treated units (bad!!!)

The first of these two comparisons are good (or at least in the spirit of DiD) in that they take the path of outcomes experienced by units that become treated and adjust it by the path of outcomes experienced by units that are not participating in the treatment.  The third comparison is different though: it adjusts the path of outcomes for newly treated units by the path of outcomes for already treated units.  But this is not the path of untreated potential outcomes, it includes *treatment effect dynamics*.  Thus, these dynamics appear in $\alpha$, *making it very hard to give a clear causal interpretation*.

And this issue can have potentially severe consequences.  For example, it is possible to come up with examples where the effect of participating in the treatment is positive for all units in all time periods, but the TWFE estimation procedure leads to estimating a negative effect of participating in the treatment. Even in the case where ``negative weights'' can be ruled out, $\alpha$ recover a weighted average of $ATT's$, though these weights are hard to interpret.


## Treatment Effects in Difference in Differences Designs with Multiple Periods

In light of the potential problems with TWFE regressions in DiD designs with multiple periods, are there alternative approaches that can be used in this case?

Yes, and it turns out that it is not all that complicated! It is just a matter of using the ``good/desirable'' comparisons between groups instead of all possible comparisons.  

To fix ideas, let's provide some extended notation and be clear about the identifying assumptions that we are going to make.

**Notation**

* $Y_{it}(0)$ is unit $i$'s untreated potential outcome.  This is the outcome that unit $i$ would experience in period $t$ if they do not participate in the treatment.

* $Y_{it}(g)$ is unit $i$'s potential outcome in time period $t$ if they become treated in period $g$.

* $G_i$ is the time period when unit $i$ becomes treated (often *groups* are defined by the time period when a unit becomes treated; hence, the $G$ notation).

* $C_i$ is an indicator variable for whether unit $i$ is in a **never-treated** group.

* $D_{it}$ is an indicator variable for whether unit $i$ has been treated by time $t$.

* $Y_{it}$ is unit $i$'s observed outcome in time period $t$.  For units in the never-treated group, $Y_{it} = Y_{it}(0)$ in all time periods.  For units in other groups, we observe $Y_{it} = \mathbf{1}\{ G_i > t\} Y_{it}(0) + \mathbf{1}\{G_i \leq t \} Y_{it}(G_i)$.  The notation here is a bit complicated, but in words, we observe untreated potential outcomes for units that have not yet participated in the treatment, and we observe treated potential outcomes for units once they start to participate in the treatment (and these can depend on *when* they became treated). Implicit in this notation there is a **no treatment anticipation** assumption, which can be relaxed as discussed in [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001).

* $X_i$ vector of pre-treatment covariates.

### Main Assumptions
**Staggered Treatment Adoption Assumption**
Recall that $D_{it} = 1$ if a unit $i$ has been treated by time $t$ and $D_{it}=0$ otherwise.  Then, for $t=1,...,\mathcal{T}-1$, $D_{it} = 1 \implies D_{it+1} = 1$.

Staggered treatment adoption implies that once a unit participates in the treatment, they remain treated. In other words, units do not "forget" about their treatment experience. This is a leading case in many applications in economics.  For example, it would be the case for policies that roll out to different locations over some period of time.  It would also be the case for many unit-level treatments that have a "scarring" effect.  For example, in the context of job training, many applications consider participating in the treatment *ever* as defining treatment. 

Within the DiD context, we believe it is hard to analyze non-staggered treatment setups **without** further restricting treatment effect heterogeneity across time, groups, treatment sequences, etc. That is the main reason we focus on this leading case.

**Parallel Trends Assumption based on never-treated units**
For all $g=2,...,\mathcal{T}$, $t=2,...,\mathcal{T}$ with $t \ge g$,
$$
  E[ Y_t(0) - Y_{t-1}(0) | G=g] = E[ Y_t(0) - Y_{t-1}(0)| C=1]
$$

This is a natural extension of the parallel trends assumption in the two periods and two groups case.  It says that, in the absence of treatment, average untreated potential outcomes for the group first treated in time $g$ and for the "never treated" group would have followed parallel paths in all post-treatment periods $t \ge g$.

Note that the aforementioned parallel trend assumption rely on using the ``never treated'' units as comparison group for all "eventually treated" groups. This presumes that (i) a (large enough) "never-treated" group is available in the data, and (ii) these units are "similar enough" to the eventually treated units such that they can indeed be used as a valid comparison group. In situations where these conditions are not satisfied, one can use an alternative parallel trends assumption that uses the **not-yet treated** units as valid comparison groups. 

**Parallel Trends Assumption based on not-yet treated units**
For all $g=2,...,\mathcal{T}$, $s,t=2,...,\mathcal{T}$ with $t \ge g$ and $s \ge t$
$$
  E[ Y_t(0) - Y_{t-1}(0) | G=g] = E[ Y_t(0) - Y_{t-1}(0)| D_s=0, G\not=g]
$$
In plain English, this assumption states that one can use the not-yet-treated by time $s$ ($s \ge t$) units as valid comparison groups when computing the average treatment effect for the group first treated in time $g$. In general, this assumption uses more data when constructing comparison groups. However, as noted in [Marcus and Sant'Anna (2021)](https://www.journals.uchicago.edu/doi/full/10.1086/711509), this assumption does restrict some pre-treatment trends across different groups. In other words, there is no free-lunch.



## Group-Time Average Treatment Effects

The above assumptions are natural extensions of the identifying assumptions in the two periods and two groups case to the multiple periods case.  

Likewise, a natural way to generalize the parameter of interest (the ATT) from the two periods and two groups case to the multiple periods case is to define **group-time average treatment effects**:

$$
  ATT(g,t) = E[Y_t(g) - Y_t(0) | G=g]
$$

This is the average effect of participating in the treatment for units in group $g$ at time period $t$.  Notice that when there are two time periods and two groups (the canonical case), the average treatment effect on the treated is given by $ATT = ATT(g=2,t=2)$.

To give a couple more examples, suppose that a researcher has access to three time periods.  Then, $ATT(g=2,t=3)$ is the average effect of participating in the treatment for the group of units that become treated in time period 2, in time period 3.  Similarly, $ATT(g=3,t=3)$ is the average effect of participating in the treatment for the group of units that become treated in time period 3, in time period 3.

**Identification of Group-Time Average Treatment Effects**

Under either version of the parallel trends assumptions mentioned above, it is straightforward to show that group-time average treatment effects are identified. For instance, when one impose the parallel trends assumption based on "never-treated units", we have that, for all $t \ge g$
$$
ATT(g,t) = E[ Y_t - Y_{g-1}| G=g] - E[ Y_t - Y_{g-1}| C=1].
$$
Alternatively, when one impose the parallel trends assumption based on "not-yet-treated units", we have that, for all $t \ge g$
$$
ATT(g,t) = E[ Y_t - Y_{g-1}| G=g] - E[ Y_t - Y_{g-1}| D_t=0, G\not=g].
$$

These group-time average treatment effects are the building blocks of understanding the effect of participating in a treatment in DiD designs with multiple time periods.

## Parallel Trends Conditional on Covariates

In many cases, the parallel trends assumption is substantially more plausible if it holds after conditioning on observed pre-treatment covariates.  In other words, if the parallel trends assumptions are modified to be

**Conditional Parallel Trends Assumption based on never-treated units**
For all $g=2,...,\mathcal{T}$, $t=2,...,\mathcal{T}$ with $t \ge g$,
$$
  E[ Y_t(0) - Y_{t-1}(0) |X, G=g] = E[ Y_t(0) - Y_{t-1}(0)| X, C=1]
$$

**Parallel Trends Assumption based on not-yet treated units**
For all $g=2,...,\mathcal{T}$, $s,t=2,...,\mathcal{T}$ with $t \ge g$ and $s \ge t$
$$
  E[ Y_t(0) - Y_{t-1}(0) | X, G=g] = E[ Y_t(0) - Y_{t-1}(0)|X, D_s=0, G\not=g]
$$

These parallel trends assumptions are the conditional analogues of previous ones. Importantly, they allow for covariate-specific trends in outcomes across groups, which can be particularly important in setups where the distribution of covariates varies across groups.

An example of a case where this assumption is attractive is one where a researcher is interested in estimating the effect of participating in job training on earnings.  In that case, if the path of earnings (in the absence of participating in job training) depends on things like education, previous occupation, or years of experience (which it almost certainly does), then it would be important to condition on these types of variables in order to make parallel trends more credible.  

In this case, the parameter of interest is still often the $ATT(g,t)'s$ (or their aggregation).  It is still straightforward to identify and estimate the $ATT$ in this case.  Basically, one needs to estimate the change in outcomes for units in the untreated group conditional on $X$, but average out $X$ over the distribution of covariates for individuals in group $g$ to obtain $ATT(g,t)$ (see [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001) and references therein for many more details).  In practice, you can use different approaches to recover these parameters. More precisely, you can estimate the $ATT(g,t)'s$ using outcome-regressions, inverse probability weighting, or doubly-robust methods. But the **did** package automates all of this for the user.

## Aggregating Group-Time Average Treatment Effects

Group-time average treatment effects are natural parameters to identify in the context of DiD with multiple periods and multiple groups.  But in many applications, there may be a lot of them.  There are some benefits and costs here.  The main benefit is that it is relatively straightforward to think about heterogeneous effects across groups and time using group-time average treatment effects.  On the other hand, it can be hard to summarize them (e.g., they are not just a single number).

In our paper, [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001), we propose a number of ways to aggregate group-time average treatment effects.  Here, we will just consider a few important ones that we think applied researchers are most often interested in.  First, consider the average effect of participating in the treatment, separately for each group.  This is given by

$$
  \theta_S(g) = \frac{1}{\mathcal{T} - g + 1} \sum_{t=2}^{\mathcal{T}} \mathbf{1}\{g \leq t\} ATT(g,t).
$$

This parameter may be of interest in its own right, since it allows one to highlight treatment effect heterogeneity with respect to treatment adoption period. Furthermore, it is fairly straightforward to further aggregate $\theta_S(g)$ to get an easy-to-interpret overall effect parameter,

$$
  \theta^O_S := \sum_{g=2}^{\mathcal{T}} \theta_S(g) P(G=g).
$$

$\theta^O_S$ is the overall effect of participating in the treatment across all groups that have ever participated in the treatment.  In our view, this is close to being a multi-period analogue of the $ATT$ in the two period case. Thus, if a researcher is constrained to report a single treatment effect summary parameter, we recommend reporting $\theta^O_S$.


In DiD setups with multiple periods, it is natural to ask "How does treatment effects vary with elapsed treatment time?" Here, note that researchers are interested in understanding treatment effect dynamics. This is at the heart of event-study-type of analysis that is widespread in applied work. 

In this case, a natural way to aggregate the group-time average treatment effect to highlight treatment effect dynamics is given by

$$
  \theta_D(e) := \sum_{g=2}^{\mathcal{T}} \mathbf{1} \{ g + e \leq \mathcal{T} \} ATT(g,g+e) P(G=g | G+e \leq \mathcal{T}).
$$

This is the average effect of participating in the treatment for the group of units that have been exposed to the treatment for exactly $e$ time periods.  

All of these aggregations are available in the **did** package and examples with real data are available in our [Getting Started with the **did** Package](did-basics.html) vignette. In [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001), we also discuss additional aggregation schemes. We encourage you to take a look!

## Conclusion

This vignette has covered basic background issues on DiD with multiple periods.  [Callaway and Sant'Anna (2021)](https://doi.org/10.1016/j.jeconom.2020.12.001) discusses many extensions and these are all provided in the **did** package as well. [See our User Guides for more details.](index.html)

## References

* [Borusyak, Kirill, Jaravel, Xavier and Spiess, Jann. "Revisiting Event Study Designs: Robust and Efficient Estimation". The Review of Economic Studies, Vol. 91, No 6, pp. 3253-3285](https://doi.org/10.1093/restud/rdae007)

* [Callaway, Brantly, and Pedro H. C. Sant'Anna. "Difference-in-differences with multiple time periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021](https://doi.org/10.1016/j.jeconom.2020.12.001).

* [de Chaisemartin, Clement, and Xavier d'Haultfoeuille. "Two-way fixed effects estimators with heterogeneous treatment effects." American Economic Review 110.9 (2020): 2964-96.](https://doi.org/10.1257/aer.20181169)

* [Goodman-Bacon, Andrew. Difference-in-differences with variation in treatment timing." Journal of Econometrics, Vol. 225, No. 2, pp. 254-277, 2021](https://doi.org/10.1016/j.jeconom.2021.03.014)

* [Marcus, Michelle, and Pedro H. C. Sant'Anna. "The Role of Parallel Trends in Event Study Settings: An Application to Environmental Economics". Journal of the Association of Environmental and Resource Economists, Vol. 8, No. 2, pp. 235-275, 2021](https://www.journals.uchicago.edu/doi/10.1086/711509)

* [Sun, Liyang, and Sarah Abraham. "Estimating dynamic treatment effects in event studies with heterogeneous treatment effects." Journal of Econometrics, Vol. 225, No. 2, pp. 175-199, 2021](https://doi.org/10.1016/j.jeconom.2020.09.006)

---
title: "Getting Started with the did Package"
author: "Brantly Callaway and Pedro H.C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with the did Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

This vignette discusses the basics of using Difference-in-Differences (DiD) designs to identify and estimate the average effect of participating in a treatment with a particular focus on tools from the **did** package. The background article for it is [Callaway and Sant'Anna (2021), "Difference-in-Differences with Multiple Time Periods"](https://doi.org/10.1016/j.jeconom.2020.12.001).

* The **did** package allows for multiple periods and variation in treatment timing

* The **did** package allows the parallel trends assumption to hold conditional on covariates

* Treatment effect estimates coming from the **did** package do not suffer from any of the drawbacks associated with two-way fixed effects regressions or event study regressions when there are multiple periods / variation in treatment timing

* The **did** package can deliver disaggregated *group-time average treatment effects* as well as event-study type estimates (treatment effects parameters corresponding to different lengths of exposure to the treatment) and overall treatment effect estimates.


We use some notation in this vignette that is fully explained in our [Introduction to DiD with Multiple Time Periods](multi-period-did.html) vignette.

# Examples with simulated data

Let's start with a really simple example with simulated data.  Here, there are going to be 4 time periods.  There are 4000 units in the treated group that are randomly (with equal probability) assigned to first participate in the treatment (a *group*) in each time period.  And there are 4000 ``never treated'' units.  The data generating process for untreated potential outcomes

$$
  Y_{it}(0) = \theta_t + \eta_i + X_i'\beta_t + v_{it}
$$

This is an example of a very simple model for untreated potential outcomes that is compatible with a conditional parallel trends assumption.  In particular,

* We consider the case where $\eta_i$ can be distributed differently across groups.  This means that comparisons of outcomes in levels between treated and untreated units will not deliver an average treatment effect parameter.

* Next, notice that  $$ \Delta Y_{it}(0) = (\theta_t - \theta_{t-1}) + X_i'(\beta_t - \beta_{t-1}) + \Delta v_{it} $$  

    so that the path of outcomes depends on covariates.  And, in general, unconditional parallel trends is not valid in this setup unless either (i) the mean of the covariates is the same across groups (e.g., when treatment groups are independent of covariates), or (ii) $\beta_t = \beta_{t-1} = \cdots = \beta_{1}$ (this is the case when the path of untreated potential outcomes doesn't actually depend on covariates).
  
* In order to think about treatment effects, we also use the following stylized model for treated potential outcomes  $$
    Y_{it}(g) = Y_{it}(0)  + \mathbf{1}\{t \geq g\} (e+1) + ( u_{it} - v_{it})  $$  

    where $e := t-g$ is the event time (i.e., the difference between the current time period and the time when a unit becomes treated), and the last term just allows for the error terms to be different for treated and untreated potential outcomes.  It immediately follows that, in this particular example, $$ATT(g,t) = e+1$$ for all post-treatment periods $t \ge g$.

    In other words, the average effect of participating in the treatment for units in group $g$ is equal to their length of exposure plus one. Note that, for simplicity, we are considering the case where treatment effects are homogeneous across groups. Also, in this setup, units do not anticipate their treatment status, so the *no-anticipation assumption* is also satisfied.
  
* For the simulations, we also set 

  * $\theta_t = \beta_t = t$ for $t=1,\ldots,4$
	* $\eta_i \sim N(G_i, 1)$ where $G_i$ is which group an individual belongs to, 
	* $X_i \sim N( \mu_{D_i}, 1)$ where $\mu_{D_i} = 1$ for never treated units and 0 otherwise, 
	* $v_{it} \sim N(0,1)$, and $u_{it} \sim N(0,1)$.  
	* Also, all random variables are drawn independently of each other.  


## Estimating Group-Time Average Treatment Effects

**Building the dataset**

```{r, echo=FALSE, results="hide", warning=FALSE, message=FALSE}
library(did)
```

```{r, message = FALSE, warning = FALSE}
# set seed so everything is reproducible
set.seed(1814)

# generate dataset with 4 time periods
sp <- reset.sim()
sp$te <- 0
time.periods <- 4

# add dynamic effects
sp$te.e <- 1:time.periods

# generate data set with these parameters
# here, we dropped all units who are treated in time period 1 as they do not help us recover ATT(g,t)'s.
dta <- build_sim_dataset(sp)

# How many observations remained after dropping the ``always-treated'' units
nrow(dta)
#This is what the data looks like
head(dta)
```

**Estimating Group-Time Average Treatment Effects**

The main function to estimate group-time average treatment effects is the `att_gt` function.  [See documentation here](../reference/att_gt.html).  The most basic call to `att_gt` is given in the following

```{r}
# estimate group-time average treatment effects using att_gt method
example_attgt <- att_gt(yname = "Y",
                        tname = "period",
                        idname = "id",
                        gname = "G",
                        xformla = ~X,
                        data = dta
                        )

# summarize the results
summary(example_attgt)
```

The summary of `example_attgt` provides estimates of the group-time average treatment effects in the column labeled `att`, and the corresponding bootstrapped-based standard errors are given in the column `se`.  The corresponding groups and times are given in the columns `group` and `time`.  Under the no-anticipation and parallel trends assumptions, group-time average treatment effects are identified in periods when $t \geq g$ (i.e., post-treatment periods for each group).  The table also reports pseudo group-time average treatment effects when $t < g$ (i.e., pre-treatment periods for group $g$).  These can be used as a pre-test for the parallel trends assumption (as long as we assume that the no-anticipation assumption indeed holds).  In addition, the results of a Wald pre-test of the parallel trends assumption is reported in the summary of the results.  [A much more detailed discussion of using the **did** package for pre-testing is available here](pre-testing.html).

Next, we'll demonstrate how to plot group-time average treatment effects.  To plot these, use the `ggdid` function which builds off the `ggplot2` package.
```{r, fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200}
# plot the results
ggdid(example_attgt)
```

The resulting figure is one that contains separate plots for each group.  Notice in the figure above, the first plot is labeled "Group 2", the second "Group 3", etc.  Then, the figure contains estimates of group-time average treatment effects for each group in each time period along with a simultaneous confidence interval.  The red dots in the plots are pre-treatment pseudo group-time average treatment effects and are most useful for pre-testing the parallel trends assumption.  The blue dots are post-treatment group-time average treatment effects and should be interpreted as the average effect of participating in the treatment for units in a particular group at a particular point in time.

## Other features of the `did` package

The above discussion covered only the most basic case for using the `did` package.  There are a number of simple extensions that are useful in applications.  

### Adjustments for Multiple Hypothesis Testing

By default, the `did` package reports simultaneous confidence bands in plots of group-time average treatment effects with multiple time periods -- these are confidence bands that are robust to multiple hypothesis testing [essentially, the idea here is to use the same standard errors but make an adjustment to the critical value to account for multiple testing -- in the example in this section, the critical value for a 95\% uniform confidence band is `r round(example_attgt$c, 2)` instead of 1.96].  You can turn this off and compute analytical standard errors and corresponding figures with *pointwise* confidence intervals by setting `bstrap=FALSE, cband=FALSE` in the call to `att_gt`...**but we don't recommend it**!


### Aggregating group-time average treatment effects

In many applications, there can be a large number of groups and time periods.  In this case, it may be infeasible to interpret plots of group-time average treatment effects.  The `did` package provides a number of ways to aggregate group-time average treatment effects using the `aggte` function.

#### Simple Aggregation

One idea that is likely to immediately come to mind is to just return a weighted average of all group-time average treatment effects with weights proportional to the group size.  This is available by calling the `aggte` function with `type = simple`.

```{r}
agg.simple <- aggte(example_attgt, type = "simple")
summary(agg.simple)
```

This sort of aggregation immediately avoids the negative weights issue that two-way fixed effects regressions can suffer from, but we often think that *there are better alternatives*. In particular, this `simple` aggregation tends to overweight the effect of early-treated groups simply because we observe more of them during post-treatment periods.  We think there are likely to be better alternatives in most applications.

#### Dynamic Effects and Event Studies

One of the most common alternative approaches is to aggregate group-time effects into an event study plot.  Group-time average treatment effects can immediately be averaged into average treatment effects at different lengths of exposure to the treatment using the following code:

```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
agg.es <- aggte(example_attgt, type = "dynamic")
summary(agg.es)
ggdid(agg.es)
```

In this figure, the x-axis is the length of exposure to the treatment.  Length of exposure equal to 0 provides the average effect of participating in the treatment across groups in the time period when they first participate in the treatment (instantaneous treatment effect).  Length of exposure equal to -1 corresponds to the time period before groups first participate in the treatment, and length of exposure equal to 1 corresponds to the first time period after initial exposure to the treatment.

As we would expect based on the data that we generated, it looks like parallel trends holds in pre-treatment periods and the effect of participating in the treatment is increasing with length of exposure of the treatment.

The Overall ATT here averages the average treatment effects across all lengths of exposure to the treatment.

#### Group-Specific Effects

Another idea is to look at average effects specific to each group.  It is also straightforward to aggregate group-time average treatment effects into group-specific average treatment effects using the following code:
```{r, fig.width=6, fig.align='center', out.width="90%", dpi = 200}
agg.gs <- aggte(example_attgt, type = "group")
summary(agg.gs)
ggdid(agg.gs)
```

In this figure, the y-axis is categorized by group.  The x-axis provides estimates of the average effect of participating in the treatment for units in each group averaged across all time periods after that group becomes treated.  In our example, the average effect of participating in the treatment is largest for group 2 -- this is because the effect of treatment here increases with length of exposure to the treatment, and they are the group that is exposed to the treatment for the longest.

The Overall ATT averages the group-specific treatment effects across groups.  In our view, *this parameter is a leading choice as an overall summary effect of participating in the treatment*. It is the average effect of participating in the treatment that was experienced across all units that participate in the treatment in any period.  In this sense, it has a similar interpretation to the ATT in the textbook case where there are exactly two periods and two groups.

#### Calendar Time Effects

Finally, the `did` package allows aggregations across different time periods.  To do this

```{r, fig.width=8, fig.height=5, fig.align='center', out.width="90%", dpi = 200}
agg.ct <- aggte(example_attgt, type = "calendar")
summary(agg.ct)
ggdid(agg.ct)
```

In this figure, the x-axis is the time period and the estimates along the y-axis are the average effect of participating in the treatment in a particular time period for all groups that participated in the treatment in that time period.

### Small Group Sizes

Small group sizes can sometimes cause estimation problems in the `did` package.  If a group-time average treatment effect cannot be estimated (e.g., because a group has too few observations relative to the number of covariates in the model, the covariate matrix for some comparison is singular, or there are overlap violations), `att_gt` issues a warning for the affected $(g,t)$ cell and reports `NA` for that $ATT(g,t)$ rather than erroring.  (The exception is when the never-treated comparison group itself is too small, in which case `att_gt` stops and suggests setting `control_group = "notyettreated"`.)  The `did` package also reports a warning if any group has fewer observations than the number of covariates in the model plus five.

In addition, statistical inference, particularly on group-time average treatment may become more tenuous with small groups.  For example, the effective sample size for estimating the change in outcomes over time for individuals in a particular group is equal to the number of observations in that group and asymptotic results are unlikely to provide good approximations to the sampling distribution of group-time average treatment effects when the number of units in a group is small. In these cases, one should be very cautious about interpreting the $ATT(g,t)$ results for these small groups.

A reasonable alternative approach in this case is to just focus on aggregated treatment effect parameters (i.e., to run `aggte(..., type = "group")` or `aggte(...,type = "dynamic")`).  For each of these cases, the effective sample size is the total number of units that are ever treated.  As long as the total number of ever treated units is "large" (which should be the case for many DiD application), then the statistical inference results provided by the `did` package should be more stable.

### Selecting Alternative Control Groups

By default, the `did` package uses the group of units that never participate in the treatment as the control group.  In this case, if there is no group that never participates, then the `did` package will drop the last period and set units that do not become treated until the last period as the control group (this will also throw a warning).  The other option for the control group is to use the "not yet treated".  The "not yet treated" include the never treated as well as those units that, for a particular point in time, have not been treated yet (though they eventually become treated).  This group is at least as large as the never treated group though it changes across time periods.  To use the "not yet treated" as the control, set the option `control_group="notyettreated"`.

```{r eval=FALSE}
example_attgt_altcontrol <- att_gt(yname = "Y",
                                   tname = "period",
                                   idname = "id",
                                   gname = "G",
                                   xformla = ~X,
                                   data = dta,
                                   control_group = "notyettreated"			
                                   )
summary(example_attgt_altcontrol)
```

### Repeated cross sections

The `did` package can also work with repeated cross section rather than panel data.  If the data is repeated cross sections, simply set the option `panel = FALSE`.  In this case, `idname` may be omitted.  If `idname` is supplied, it is still validated -- it must be numeric, the treatment timing variable (`gname`) must not vary within unit, and each unit can appear at most once per time period -- so for genuine repeated cross sections (where, for example, the same household or region code may appear multiple times in the same period) it is best to omit `idname`.  Aside from this, usage is nearly identical to the case with panel data, although a few panel-only options (e.g., `fix_weights = "base_period"` or `fix_weights = "first_period"`) are not available when `panel = FALSE`.

### Unbalanced Panel Data

By default, the `did` package takes in panel data and, if it is not balanced, coerces it into being a balanced panel by dropping units with observations that are missing in any time period.  However, if the user specifies the option `allow_unbalanced_panel = TRUE`, then the `did` package will not coerce the data into being balanced.  Practically, the main cost here is that the computation time will increase in this case.  We also recommend that users think carefully about why they have an unbalanced panel before proceeding this direction.

### Alternative Estimation Methods

The `did` package implements all the $2 \times 2$ DiD estimators that are in the `DRDID` package.  By default, the `did` package uses "doubly robust" estimators that are based on first step linear regressions for the outcome variable and logit for the generalized propensity score.  The other options are "ipw" for inverse probability weighting and "reg" for regression.

```{r eval=FALSE}
example_attgt_reg <- att_gt(yname = "Y",
                            tname = "period",
                            idname = "id",
                            gname = "G",
                            xformla = ~X,
                            data = dta,
                            est_method = "reg"
                            )
summary(example_attgt_reg)
```

The argument `est_method` is also available to pass in a custom function for estimating DiD with 2 periods and 2 groups.  [See its documentation for more details](../reference/att_gt.html).  **Fair Warning:** this is very advanced use of the `did` package and should be done with caution.

# An example with real data

Next, we use a subset of data that comes from [Callaway and Sant'Anna (2020)](https://doi.org/10.1016/j.jeconom.2020.12.001).  This is a dataset that contains county-level teen employment rates from 2003-2007.  The data can be loaded by
```{r}
data(mpdta)
```
`mpdta` is a balanced panel with `r nrow(mpdta)` observations.  And the dataset looks like
```{r}
head(mpdta)
```

**Data Requirements**

In particular applications, the dataset should look like this with the key parts being:

* The dataset should be in *long* format -- each row corresponds to a particular unit at a particular point in time.  Sometimes panel data is in *wide* format -- each row contains all the information about a particular unit in all time periods.  To convert from wide to long in `R`, one can use the `tidyr::pivot_longer` function.  [Here is an example](http://www.cookbook-r.com/Manipulating_data/Converting_data_between_wide_and_long_format/) 

* There needs to be an id variable.  In `mpdta`, it is the variable `countyreal`.  This should not vary over time for particular units.  The name of this variable is passed to methods in the `did` package by setting, for example, `idname = "countyreal"`  

* There needs to be a time variable.  In `mpdta`, it is the variable `year`.  The name of this variable is passed to methods in the `did` package by setting, for example, `tname = "year"`

* In this application, the outcome is `lemp`.  The name of this variable is passed to methods in the `did` package by setting, for example, `yname = "lemp"`

* There needs to be a group variable.  In `mpdta`, it is the variable `first.treat`.  This is the time period when an individual first becomes treated.   *For individuals that are never treated, this variable should be set equal to 0.*  The name of this variable is passed to methods in the `did` package by setting, for example, `gname = "first.treat"`

* The `did` package allows for incorporating covariates so that the parallel trends assumption holds only after conditioning on these covariates.  In `mpdta`, `lpop` is the log of county population.  Covariates may be time-invariant or time-varying.  With balanced panel data, the `did` package sets the value of a time-varying covariate to be equal to the value of the covariate in the "base period" where, in post-treatment periods the base period is the period immediately before observations in a particular group become treated (when there is anticipation, it is before anticipation effects start too), and in pre-treatment periods the base period is the period right before the current period.  With repeated cross sections data or unbalanced panel data, the covariates are instead taken from each time period (see the discussion of `xformla` in the `att_gt` documentation for details).  Covariates are passed as a formula to the `did` package by setting, for example, `xformla = ~lpop`; the formula may also include transformations of covariates (e.g., `~ I(lpop^2)`, `~ log(lpop)`, interactions) as well as factor variables.  For estimators under unconditional parallel trends, the `xformla` argument can be left blank or can be set as `xformla = ~1` to only include a constant.

### The Effect of the Minimum Wage on Youth Employment

Next, we walk through a straightforward, but realistic way to use the `did` package to carry out an application.

**Side Comment:** This is just an example of how to use our method in a real-world setup.  To really evaluate the effect of the minimum wage on teen employment, one would need to be more careful along a number of dimensions. Thus, results displayed here should be interpreted as illustrative only.


We'll consider two cases.  For the first case, we will not condition on any covariates.  For the second, we will condition on the log of county population (in a "real" application, one might want to condition on more covariates).

```{r, fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200}
# estimate group-time average treatment effects without covariates
mw.attgt <- att_gt(yname = "lemp",
                   gname = "first.treat",
                   idname = "countyreal",
                   tname = "year",
                   xformla = ~1,
                   data = mpdta
                   )

# summarize the results
summary(mw.attgt)

# plot the results
# set ylim so that all plots have the same scale along y-axis
ggdid(mw.attgt, ylim = c(-.3, .3))
```

There are a few things to notice in this case

* There does not appear to be much evidence against the parallel trends assumption.  One fails to reject using the Wald test reported in `summary`; likewise the uniform confidence bands cover 0 in all pre-treatment periods.

* There is some evidence of negative effects of the minimum wage on employment.  Two group-time average treatment effects are negative and statistically different from 0.  These results also suggest that it may be helpful to aggregate the group-time average treatment effects.

```{r,  fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# aggregate the group-time average treatment effects
mw.dyn <- aggte(mw.attgt, type = "dynamic")
summary(mw.dyn)
ggdid(mw.dyn, ylim = c(-.3, .3))
```

These continue to be simultaneous confidence bands for dynamic effects. The results are broadly similar to the ones from the group-time average treatment effects:  one fails to reject parallel trends in pre-treatment periods and it looks like somewhat negative effects of the minimum wage on youth employment.

One potential issue with these dynamic effect estimators is that the composition of the groups changes with different lengths of exposure in the event study plots.  For example, for the group of states who increased their minimum wage in 2007, we can only identify the instantaneous average effect of the minimum wage ($e=0$), whereas for states that raised their  minimum wage in 2004 (2006), we can identify the average effect of the minimum wage on event-times $e=0,1,2,3$ ($e=0,1$). When computing the event-study plot for $e=0$, we would aggregate the effects for all three groups, but this is not the case when $e=1,2,3$. If the effects of the minimum wage are systematically different across groups (here, there is not much evidence of this as the effect for all groups seems to be close to 0 on impact and perhaps becoming more negative over time), then this can lead to confounding dynamics and selective treatment timing among different groups.  

One way to combat this is to balance the sample by (i) only including groups that are exposed to the treatment for at least a certain number of time periods and (ii) only look at dynamic effects in those time periods.  In the `did` package, one can do this by specifying the `balance_e` option.  Here, we set `balance_e = 1` -- what this does is to only consider groups of states that
are treated in 2004 and 2006 (so we can compute event-study-type parameters for them for $e=0,1$), drops the group treated in 2007 (as we can not compute the event-study-type parameter with $e=1$ for this group)), and then only looks at instantaneous average effects and the average effect one period after states raised the minimum wage.

```{r,  fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
mw.dyn.balance <- aggte(mw.attgt, type = "dynamic", balance_e = 1)
summary(mw.dyn.balance)
ggdid(mw.dyn.balance, ylim = c(-.3, .3))
```

Finally, we can run all the same results including a covariate.  In this application the results turn out to be nearly identical, and here we provide just the code for estimating the group-time average treatment effects while including covariates.  The other steps are otherwise the same.

```{r eval=FALSE}
mw.attgt.X <- att_gt(yname = "lemp",
                     gname = "first.treat",
                     idname = "countyreal",
                     tname = "year",
                     xformla = ~lpop,
                     data = mpdta
                     )
```


## Common Issues with the did package

We update this section with common issues that people run into when using the `did` package.  Please feel free to contact us with questions or comments.

* The `did` package is only built to handle staggered treatment adoption designs.  This means that once an individual becomes treated, they remain treated in all subsequent periods.

## Bugs

We also welcome (and encourage) users to report any bugs / difficulties they have using our code.

---
title: "Pre-Testing in a DiD Setup using the did Package"
author: "Brantly Callaway and Pedro H.C. Sant&apos;Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Pre-Testing in a DiD Setup using the did Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, echo=FALSE, results="hide", warning=FALSE, message=FALSE}
library(did)
# Source the current version of the did package (based on our Dropbox)
# fldr <- here::here("R/")
# sapply(paste0(fldr,list.files(fldr)), source)
# Source simulation designs
source(here::here("vignettes/setup_sims.R"))
```


## Introduction

This vignette provides a discussion of how to conduct pre-tests in DiD setups using the **did** package.

* One appealing feature of many DiD applications with multiple periods is that the researcher can pre-test the parallel trends assumptions.

    * The idea here is simple: although one cannot always test whether parallel trends itself holds, one can check if it holds in periods before treated units actually become treated.

    * Importantly, this is just a pre-test; it is different from an actual test.  Whether or not the parallel trends assumption holds in pre-treatment periods does not actually tell you if it holds in the current period (and this is when you need it to hold!).  It is certainly possible for the identifying assumptions to hold in previous periods but not hold in current periods; it is also possible for identifying assumptions to be violated in previous periods but for them to hold in current periods.  That being said, we view the pre-test as a piece of evidence on the credibility of the DiD design in a particular application.

* In this vignette, we demonstrate that the approach used in the **did** package for pre-testing may work substantially better than the more common "event study regression".

## Common Approaches to Pre-Testing in Applications

By far the most common approach to pre-testing in applications is to run an **event-study regression**.  Here, the idea is to run a regression that includes leads and lags of the treatment dummy variable such as
$$
  Y_{it} = \theta_t + \eta_i + \sum_{l=-\mathcal{T}}^{\mathcal{T}-1} D_{it}^l \mu_l + v_{it}
$$

where $D_{it}^l = 1$ if individual $i$ has been exposed to the treatment for $l$ periods in period $t$, and $D_{it}^l = 0$ otherwise.  To be clear here, it is helpful to give some examples.  Suppose individual $i$ becomes treated in period 3.  Then,

* $D_{it}^0 = 1$ when $t=3$ and is equal to 0 in other time periods

* $D_{it}^2 = 1$ when $t=5$ and is equal to 0 in other time periods

* $D_{it}^{-2} = 1$ when $t=1$ and is equal to 0 in other time periods.

And $\mu_l$ is interpreted as the effect of treatment for different lengths of exposure to the treatment.  Typically, $\mu_{-1}$ is normalized to be equal to 0, and we follow that convention here.  It is common to interpret estimated $\mu_l$'s with $l < 0$ as a way to pre-test the parallel trends assumption.

### Pitfalls with Event Study Regressions

```{r echo=FALSE}
time.periods <- 4
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)
```

#### Best Case Scenario for Pre-Testing

First, let's start with a case where an event study regression is going to work well for pre-testing the parallel trends assumption

```{r}
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()

head(data)
```

The main thing to notice here:

* The dynamics are common across all groups.  This is the case where an event-study regression will work.

Next, a bit more code

```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

head(data)

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)

# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways",
          index = c("id", "period"))

summary(es)

#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs=coefs, ses=ses, exposure=exposure)

library(ggplot2)

ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()
```

You will notice that everything looks good here.  The pre-test performs well (the caveat to this is that the standard errors are "pointwise" and would be better to have uniform confidence bands though this does not seem to be standard practice in applications).

We can compare this to what happens using the `did` package:
```{r, fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200}
# estimate group-time average treatment effects
did_att_gt <- att_gt(yname = "Y",
                     tname = "period",
                     idname = "id",
                     gname = "G",
                     data = data,
                     bstrap = FALSE,
                     cband = FALSE)
summary(did_att_gt)

# plot them
ggdid(did_att_gt)

```

```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# aggregate them into event study plot
did_es <- aggte(did_att_gt, type = "dynamic")

# plot the event study
ggdid(did_es)
```

Overall, everything looks good using either approach.  (Just to keep things fair, we report pointwise confidence intervals for group-time average treatment effects, but it is easy to get uniform confidence bands by setting the options `bstrap=TRUE, cband=TRUE` to the call to `att_gt`.)


#### Pitfall: Selective Treatment Timing

Sun and Abraham (2021) point out a major limitation of event study regressions:  when there is **selective treatment timing** the $\mu_l$ end up being weighted averages of treatment effects *across different lengths of exposures*.

**Selective treatment timing** means that individuals in different groups experience systematically different effects of participating in the treatment from individuals in other groups.  For example, there would be selective treatment timing if individuals choose to be treated in earlier periods if they tend to experience larger benefits from participating in the treatment.  This sort of selective treatment timing is likely to be present in many applications in economics / policy evaluation.

Contrary to event study regressions, pre-tests based on group-time average treatment effects (or based on group-time average treatment effects that are aggregated into an event study plot) **are still valid even in the presence of selective treatment timing**.

To see this in action, let's keep the same example as before, but add selective treatment timing.

```{r, echo=FALSE}
reset.sim()
bett <- betu <- rep(0, time.periods)
te <- 0
set.seed(1814)
```

```{r}
# generate dataset with 4 time periods
time.periods <- 4

# generate dynamic effects
te.e <- time.periods:1

# generate selective treatment timing
# (*** this is what is different here ***)
te.bet.ind <- time.periods:1 / (time.periods/2)

# generate data set with these parameters
# (main thing: it generates a dataset that satisfies
# parallel trends in all periods...including pre-treatment)
data <- build_sim_dataset()
```

```{r}
# run through same code as in earlier example...omitted
```

```{r, echo=FALSE}
#-----------------------------------------------------------------------------
# modify the dataset a bit so that we can run an event study
#-----------------------------------------------------------------------------

# generate leads and lags of the treatment
Dtl <- sapply(-(time.periods-1):(time.periods-2), function(l) {
    dtl <- 1*( (data$period == data$G + l) & (data$G > 0) )
    dtl
})
Dtl <- as.data.frame(Dtl)
cnames1 <- paste0("Dtmin", (time.periods-1):1)
colnames(Dtl) <- c(cnames1, paste0("Dt", 0:(time.periods-2)))
data <- cbind.data.frame(data, Dtl)
row.names(data) <- NULL

#-----------------------------------------------------------------------------
# run the event study regression
#-----------------------------------------------------------------------------

# load plm package
library(plm)
```

```{r}
# run event study regression
# normalize effect to be 0 in pre-treatment period
es <- plm(Y ~ Dtmin3 + Dtmin2 + Dt0 + Dt1 + Dt2, 
          data = data, model = "within", effect = "twoways", 
          index = c("id", "period"))

summary(es)
```


```{r, echo=FALSE}
#-----------------------------------------------------------------------------
# make an event study plot
#-----------------------------------------------------------------------------

# some housekeeping for making the plot
# add 0 at event time -1
coefs1 <- coef(es)
ses1 <- sqrt(diag(summary(es)$vcov))
idx.pre <- 1:(time.periods-2)
idx.post <- (time.periods-1):length(coefs1)
coefs <- c(coefs1[idx.pre], 0, coefs1[idx.post])
ses <- c(ses1[idx.pre], 0, ses1[idx.post])
exposure <- -(time.periods-1):(time.periods-2)

cmat <- data.frame(coefs = coefs, ses = ses, exposure = exposure)
```


```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# run through same code as before...omitted

# new event study plot
ggplot(data = cmat, mapping = aes(y = coefs, x = exposure)) +
  geom_line(linetype = "dashed") +
  geom_point() + 
  geom_errorbar(aes(ymin = (coefs-1.96*ses), ymax = (coefs+1.96*ses)), width = 0.2) +
  ylim(c(-2, 5)) +
  theme_bw()
```

In contrast to the last case, it is clear that things have gone wrong here.  **Parallel trends holds in all time periods and for all groups here**, but the event study regression incorrectly rejects that parallel trends holds -- this is due to the selective treatment timing.

We can compare this to what happens using the `did` package:
```{r, fig.width=8,fig.height=10, fig.align='center', out.width="90%", dpi = 200}
# estimate group-time average treatment effects
did.att.gt <- att_gt(yname = "Y",
                     tname = "period",
                     idname = "id",
                     gname = "G",
                     data = data
                     )
summary(did.att.gt)

# plot them
ggdid(did.att.gt)
```


```{r, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# aggregate them into event study plot
did.es <- aggte(did.att.gt, type = "dynamic")

# plot the event study
ggdid(did.es)
```

This is the correct performance (up to aforementioned caveats about multiple hypothesis testing).

## Conditional Moment Tests

Another main use case for the `did` package is when the parallel trends assumptions holds after conditioning on some covariates.  This is likely to be important in many applications.  For example, to evaluate the effect of participating in a job training program on earnings, it is likely to be important to condition on an individual's education.  This would be true if (i) the distribution of education is different for individuals that participate in job training relative to those that don't (this is very likely to hold as people that participate in job training tend to have less education than those who do not), and (ii) if the path of earnings (absent participating in job training) depends on an individual's education.  See Heckman, Ichimura, and Todd (1998) and Abadie (2005) for more discussion.

Even when one includes covariates to estimate group-time average treatment effects, pre-tests based only on group-time average treatment effects can fail to detect some violations of the parallel trends assumption.  To give an example, suppose that the only covariate is binary variable for an individual's sex.  Pre-tests based on group-time average treatment effects could fail to detect violations of the conditional parallel trends assumption in cases where it is violated in one direction for men and in the other direction for women.

The `did` package contains an additional pre-test for the conditional parallel trends assumption in the `conditional_did_pretest` function.

```{r eval=FALSE}
# not run (this code can be substantially slower)
reset.sim()
set.seed(1814)
nt <- 1000
nu <- 1000
cdp <- conditional_did_pretest("Y", "period", "id", "G", xformla = ~X, data = data)
cdp
```


## References

* [Abadie, Alberto. "Semiparametric difference-in-differences estimators." The Review of Economic Studies 72.1 (2005): 1-19.](https://doi.org/10.1111/0034-6527.00321)

* [Callaway, Brantly, and Pedro H. C. Sant'Anna. "Difference-in-differences with multiple time periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021](https://doi.org/10.1016/j.jeconom.2020.12.001).

* [Heckman, James J., Hidehiko Ichimura, and Petra Todd. "Matching as an econometric evaluation estimator." The Review of Economic Studies 65.2 (1998): 261-294.](https://doi.org/10.2307/2971733)

* [Sun, Liyang, and Sarah Abraham. "Estimating dynamic treatment effects in event studies with heterogeneous treatment effects." Journal of Econometrics, Vol. 225, No. 2, pp. 175-199, 2021](https://doi.org/10.1016/j.jeconom.2020.09.006)

---
title: "Problems with two-way fixed-effects event-study regressions"
author: "Brantly Callaway and Pedro H. C. Sant'Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Problems with two-way fixed-effects event-study regressions} 
  %\VignetteEngine{knitr::rmarkdown} 
  %\VignetteEncoding{UTF-8}
---

```{r setup2, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  echo = TRUE,
  eval = FALSE
)
```


## Introduction

In this vignette, we want to illustrate the dangers of relying on traditional two-way fixed-effects (TWFE) regression models with leads and lags to conduct causal inference about treatment effect dynamics. In fact, we will show that *even* when treatment effect dynamics are homogeneous across different treatment groups/cohorts, some of these TWFE regression models can be severely biased. 


In order to effectively communicate these dangers, we consider a stylized simulation example where covariates do not play any major role. This example draws from several discussions we have had with Andrew Baker. We consider three setups. In the first one all units are eventually treated, and treatment effect dynamics are homogeneous across treatment groups. In the second, we have a group of units that remain untreated at the end of our sample, and treatment effect dynamics are homogeneous across treatment groups. The third one is like the second, but with heterogeneous treatment effect dynamics across treatment groups.


## Setup with all units being eventually treated and homogeneous treatment effect dynamics

Consider a staggered treatment adoption setup where 40 'states' ($state = \{1,2,\dots,40\}$) are randomly assigned into 4 treatment groups depending on the treatment starting year (1986, 1992, 1998, and 2004). We denote the treatment starting period as $g$ with ($g \in \{ 1986, 1992, 1998, 2004\}$). The 1000 units (e.g., 'counties' or 'firms') $i$ in the sample are randomly assigned to one of the 40 states. Let $G_i$ indicates the group/cohort unit $i$ belongs to, i.e., $G \subseteq \{ 1986, 1992, 1998, 2004\}$.


The data generating process (DGP) for the outcome $Y$ we consider is 
$$Y_{i,t} = (2010-g) + \alpha_i + \alpha_t + \tau_{i,t} + \epsilon_{i,t}$$
where $\alpha_i$ are unit fixed effects drawn from $\sim N(\mu_{state}, 1)$ with state-specific mean $\mu_{state} = state/5$, $\alpha_t$ are time fixed effects (cohort specific parallel time-trends) generated as $$\alpha_t = 0.1 * (t - g) + \epsilon^{time FE}_t, $$ with $\epsilon^{time FE}_t \sim N(0, 1)$, $\epsilon_{i,t} \sim N\left(0, \left(\frac{1}{2}\right)^2\right)$ is an idiosyncratic error term, and $\tau_{i,t}$ are the unit-specific treatment effects at time $t$ generated as 
$$ \tau_{i,t} = \mu \times (t - g + 1)\times 1\{t \ge g \},$$
where $\mu$ is the the instantaneous treatment effect; we set $\mu=1$.


From the above DGP, it is evident that the ATT at the time a unit is treated is equal to 1 for all groups, it is equal to 2 one period after treatment started for all groups, etc. Thus, the evolution of the treatment effects is homogeneous across treatment groups. This is an *a priori* strong restriction that should make things "easier" to estimate (but is by no means required!).

### Visualizing the DGP
Let's start the discussion by first visualizing how the DGP looks like.

```{r DGP_viz, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# Load libraries and set baseline parameters
library(tidyverse)
library(lfe)
library(fastDummies)
library(ggthemes)
library(did)
theme_set(theme_clean() + theme(plot.background = element_blank()))
#----------------------------------------------------------------------------
iseed  = 20201221
nrep <- 100  
true_mu <- 1
set.seed(iseed)

#----------------------------------------------------------------------------
## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobserved heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into four treatment groups G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse(year >= cohort_year, 1, 0),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error)
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data()

# plot
plot1 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value", color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))  +
  ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with all groups being eventually treated")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot1

#ggsave("plot_dgp1.png", plot1, width = 10, height = 5, dpi = 100)
```
![](plot_dgp1.png){ width=100% }

The above plot shows the 1,000 individual values of $Y_{i,t}$, as well as the average by treatment-group (the thicker colorful lines), and the vertical lines show the period when treatment begins for each group. Because treatment effects here grows linearly with time elapsed since treatment started, we can see that the earliest treated units end up with the highest value of Y, and that the difference grows by the gap between treatment years.


### Estimating dynamic treatment effects via TWFE event-study regressions

Given that we are interested in treatment effect dynamics, we then proceed to consider a classical two-way fixed-effects (TWFE) event study specification 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L} \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$ 
where $D_{i,t}^{k}=1\{t-G_i=k\}$ is an "event-study" dummy variable that takes value one if a unit $i$ is $k$ periods away from initial treatment at time $t$ and zero otherwise, $D_{i,t}^{<-K}=1\{t-G_i<-K\}$ and $D_{i,t}^{>L}=1\{t-G_i>L\}$ are defined analogously. For instance, $D_{i,t}^{0}$ is equal to one if the unit $i$ is first treated at time $t$, $D_{i,t}^{1}$ is equal to one if a one period has passed since treatment started (treatment lags), etc. Alternatively we have that $D_{i,t}^{-2}$ is equal to one if a unit $i$ will be treated in two periods from $t$ (treatment leads). In this exercise we set $K$ and $L$ to be equal to 5.


Up to today, it is customary to interpret estimates of $\gamma_k^{lags}$ as ``good'' measures of the average treatment effect for being exposed to treatment for $k$ periods, and estimates of $\gamma_k^{leads}$ as measures of pre-trends. Our first exercise here is to assess if this is OK-ish. 


Towards this end, we will a perform Monte Carlo exercise where we generate data according to the DGP described above and estimate all the parameters of the TWFE event-study regression above. We repeat this 100 times (same holds true if we do 1,000) and plot the Monte Carlo mean and 1.96 times the Monte-Carlo standard-deviation of the relative time indicator dummies, as well as the true treatment effect below.


```{r ES1, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse(rel_year < -5, 1, 0),
           Post = ifelse(rel_year > 5, 1, 0))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
  # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical <- map_dfr(1:nrep, run_ES_DiD)

colors <- c("True Effect" = "red", "Estimated Effect" = "blue")

ES_plot_classical <- data_classical %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
  geom_line(aes(color = 'Estimated Effect'), size = 1) + 
  geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("TWFE event-study regression with binned end-points")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical

#ggsave("es_plot_classical.png", ES_plot_classical, width = 10, height = 5, dpi = 150)
```

![](es_plot_classical.png){ width=100% }

The results above show that these TWFE event-study type estimates are *severely biased* for the true treatment effects. Furthermore, using the estimates of coefficient of treatment leads as a way to find evidence of "pre-trends" is very problematic, as illustrated above; see Sun and Abraham (2021) for formal results! Putting it simply, the results above highlight that such TWFE linear regression should not be used to highlight treatment effect dynamics!



Now, you may wonder: What happens if I include all possible leads and lags in the TWFE event study specification, i.e., to set K and L to the maximum allowable in the data, making inclusion of $D_{i,t}^{<-K}$
and of $D_{i,t}^{>L}$ unnecessary. In what follows, we do exactly that, though we need to omit one extra period apart from -1 to avoid problems of multicollinearity identified by Borusyak and Jaravel (2018). We drop the most negative time period, which is perhaps the most natural thing to do. In order words, we set $K=23$ and $L=24$, and drop the  $D_{i,t}^{<-K}$ and $D_{i,t}^{>L}$ from the TWFE specification.


The result of this exercise is displayed below. And, as you can see, this still leads to *very biased* estimates for the causal effects of interest!

```{r ES3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year")
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat <- map_dfr(1:nrep, run_ES_DiD_sat)

ES_plot_sat <- data_sat %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())
        

ES_plot_sat

#ggsave("es_plot_sat.png", ES_plot_sat, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat.png){ width=100% }


### Using the Callaway and Sant'Anna (2021) framework

Instead of using variants of the TWFE specification above, we advocate you to use the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our package **did**. 

In this specific DGP, as there is no "never-treated" units, we can use the "not-yet-treated" units as a valid comparison group for each cohort. Furthermore, since covariates are not available (and do not play any major role into the analysis), we do not need to worry about the underlying "est_method".

Below we show the results associated with using the Callaway and Sant'Anna (2021) procedure. As you can see, you get reliable and precise estimates for treatment effect dynamics. This is good news!
```{r CS, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS <- function(...) {
  
  # resimulate the data
  data <- make_data()
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS <- map_dfr(1:nrep, run_CS)

ES_plot_CS <- data_CS %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated")+
    scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS

#ggsave("es_plot_CS.png", ES_plot_CS, width = 10, height = 5, dpi = 150)

```
![](es_plot_CS.png){ width=100% }

## Setup with a group that remains untreated at the end of the sample

One of the features of the above DGP is that all units eventually get treated. In some applications, we have that a group of units remain untreated at the end of the sample. We call this group of units the "never-treated" group. To assess if the problems associated with TWFE specification highlighted above persist if such a group of units is available, we modify a little bit the DGP and repeat the same exercise as above.


### DGP with a "never-treated" group and homogeneous treatment effect dynamics

The DGP we consider is very similar to the one we have already considered. The only difference is that now we suppress treatment for the last cohort. Put it simply, the cohort with $g=2004$ is now "never-treated", implying that, in this DGP the unit-specific treatment effects at time $t$  $\tau_{i,t}$ are generated as 
$$ \tau_{i,t} = \mu \times (t - g + 1)\times 1\{t \ge g \}1\{g\not=2004 \},$$
where $\mu$ is the the instantaneous treatment effect; we set $\mu=1$.


Like before, the ATT at the time a unit is treated is 1, it is equal to 2 one period after treatment started, etc. 

### Visualizing the DGP
Let's start the discussion by first visualizing how the DGP looks like.

```{r DGP_viz2, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data2 <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobserved heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into 1\{t \ge g \}G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
    # Relabel 2004 cohort as never-treated
    mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data2()

# plot
plot2 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value",  color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
                     values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3"))+
  ggtitle("One draw of the DGP with homogeneous effects across cohorts \n and with a never-treated group")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot2 

#ggsave("plot_dgp_never_treated.png",plot2, width = 10, height = 5, dpi = 100)
```

![](plot_dgp_never_treated.png){ width=100% }

### Estimating dynamic treatment effects via TWFE event-study regressions

Like before, we start by estimating a TWFE event-study regression 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L}  \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$
with $K$ and $L$ set to 5. That is, binning the endpoints.

As you can see from the results below, the biases are still substantial. Thus, the problems with TWFE specifications persist even if you have a "never-treated" group.
```{r ES1_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>% 
    dummy_cols(select_columns = "rel_year") %>% 
    mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
           Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
    mutate(Pre = ifelse(is.na(Pre), 0, Pre),
           Post = ifelse(is.na(Post), 0, Post))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
 # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical_never <- map_dfr(1:nrep, run_ES_DiD_never)

ES_plot_classical_never <- data_classical_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with binned end-points")+
    scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical_never

#ggsave("es_plot_classical_never.png", ES_plot_classical_never, width = 10, height = 5, dpi = 150)
```
![](es_plot_classical_never.png){ width=100% }

Next, we consider the case where we one include almost all leads and lags ($K=17$ and $L=24$) but omit the furthest lead $K=18$. 


```{r ES3_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year2 = rel_year) %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    select(-("rel_year_-Inf"))
    
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
 # grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat_never <- map_dfr(1:nrep, run_ES_DiD_sat_never)

ES_plot_sat_never <- data_sat_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_sat_never

#ggsave("es_plot_sat_never.png", ES_plot_sat_never, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat_never.png){ width=100% }

As the result above shows, when you have a never-treated group, treatment effects are homogeneous across cohorts, and you do not bin event-times, TWFE event-study regressions do recover the true underlying dynamic treatment effects. 


### Using the Callaway and Sant'Anna (2021) framework

Next, we show that the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our **did** can also be used in this setup.


In this specific DGP, as there is a "never-treated" group, we can use use different comparison groups: 1) the "never-treated" group, and 2) the "not-yet-treated" group. In this particular DGP, using any of these two different comparison groups should recover the true underlying dynamic treatment effect. Let's see if this is indeed the case.


First, let's estimate it using the "never-treated" units as the comparison group.

```{r CS_never, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS_never <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "nevertreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_never <- map_dfr(1:nrep, run_CS_never)

ES_plot_CS_never <- data_CS_never %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_never

#ggsave("es_plot_CS_never.png", ES_plot_CS_never, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_never.png){ width=100% }
 
 As expected, the results look good!
 
 Now, let's consider the "not-yet-treated" units as a comparison group.

```{r CS_ny, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

# function to run ES DID
run_CS_ny <- function(...) {
  
  # resimulate the data
  data <- make_data2()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_ny <- map_dfr(1:nrep, run_CS_ny)

ES_plot_CS_ny <- data_CS_ny %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* true_mu, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_ny

#ggsave("es_plot_CS_ny.png", ES_plot_CS_ny, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_ny.png){ width=100% }

As expected, the results look good, too


## Setup with Heterogeneous Treatment Effects

Now, we want to consider a setup with heterogeneous treatment effect dynamics and with a "never-treated" group. This DGP is a modification of the one we have considered above, where now the "unit-specific" treatment effect at time $t$ is given by
$$ \tau_{i,t} = \mu_g \times (t - g + 1)\times 1\{t \ge g \}1\{g\not=2004 \},$$
where we set $\mu_{1986} = 3$, $\mu_{1992} = 2$ and $\mu_{1998} = 1$. This implies that for the group who started treatment in 1986, its average treatment effects evolve with elapsed time as $3,6, 9,\dots$. For the group that started treatment in 1992 (1998), its average treatment effect evolves with elapsed time as $2,4, 6,\dots$ ($1,2, 3,\dots$). Of course, the treatment effect is zero for the "never-treated" cohort ($g=2004$, which later we will relabel to $g=\infty$). This is a DGP where "early-treated" units benefit more from the treatment than "later-treated" units.

Given that each group has the same size (on average), the true average treatment effect dynamic across treated groups is $(3+2+1)/3 = 2$ at the time a unit is treated, $2 \times (3+2+1)/3 = 4$ in the first period after treatment started, etc.

### Visualizing the DGP
Like before, let's start the discussion by visualizing what this DGP looks like.

```{r DGP_viz3, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

## Generate data - treated cohorts consist of 250 obs each, with the treatment effect still = true_mu on average
make_data3 <- function(nobs = 1000, 
                      nstates = 40) {
  
  # unit fixed effects (unobserved heterogeneity)
  unit <- tibble(
    unit = 1:nobs,
    # generate state
    state = sample(1:nstates, nobs, replace = TRUE),
    unit_fe = rnorm(nobs, state/5, 1),
    # generate instantaneous treatment effect
    #mu = rnorm(nobs, true_mu, 0.2)
    mu = true_mu
  )
  
  # year fixed effects (first part)
  year <- tibble(
    year = 1980:2010,
    year_fe = rnorm(length(year), 0, 1)
  )
  
  # Put the states into treatment groups
  treat_taus <- tibble(
    # sample the states randomly
    state = sample(1:nstates, nstates, replace = FALSE),
    # place the randomly sampled states into 1\{t \ge g \}G_g
    cohort_year = sort(rep(c(1986, 1992, 1998, 2004), 10))
  )
  
  # make main dataset
  # full interaction of unit X year 
  expand_grid(unit = 1:nobs, year = 1980:2010) %>% 
    left_join(., unit) %>% 
    left_join(., year) %>% 
    left_join(., treat_taus) %>% 
    # make error term and get treatment indicators and treatment effects
    # Also get cohort specific trends (modify time FE)
    mutate(error = rnorm(nobs*31, 0, 1),
           treat = ifelse((year >= cohort_year)* (cohort_year != 2004), 1, 0),
           mu = ifelse(cohort_year==1992, 2, ifelse(cohort_year==1998, 1, 3)),
           tau = ifelse(treat == 1, mu, 0),
           year_fe = year_fe + 0.1*(year - cohort_year)
    ) %>% 
    # calculate cumulative treatment effects
    group_by(unit) %>% 
    mutate(tau_cum = cumsum(tau)) %>% 
    ungroup() %>% 
    # calculate the dep variable
    mutate(dep_var = (2010 - cohort_year) + unit_fe + year_fe + tau_cum + error) %>%
    # Relabel 2004 cohort as never-treated
    mutate(cohort_year = ifelse(cohort_year == 2004, Inf, cohort_year))
  
}
#----------------------------------------------------------------------------
# make data
data <- make_data3()

# plot
plot3 <- data %>% 
  ggplot(aes(x = year, y = dep_var, group = unit)) + 
  geom_line(alpha = 1/8, color = "grey") + 
  geom_line(data = data %>% 
              group_by(cohort_year, year) %>% 
              summarize(dep_var = mean(dep_var)),
            aes(x = year, y = dep_var, group = factor(cohort_year),
                color = factor(cohort_year)),
            size = 2) + 
  labs(x = "", y = "Value",  color = "Treatment group   ") + 
  geom_vline(xintercept = 1986, color = '#E41A1C', size = 2) + 
  geom_vline(xintercept = 1992, color = '#377EB8', size = 2) + 
  geom_vline(xintercept = 1998, color = '#4DAF4A', size = 2) + 
  #geom_vline(xintercept = 2004, color = '#984EA3', size = 2) + 
  scale_color_brewer(palette = 'Set1') + 
  theme(legend.position = 'bottom',
        #legend.title = element_blank(), 
        axis.title = element_text(size = 14),
        axis.text = element_text(size = 12)) +
  scale_color_manual(labels = c("1986", "1992", "1998", "Never-treated"),
                     values = c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3")) +
  ggtitle("One draw of the DGP with heterogeneous treatment effect dynamics across cohorts \n and with a never-treated group")+
  theme(plot.title = element_text(hjust = 0.5, size = 12))

plot3 

#ggsave("plot_dgp_never_treated_het.png", plot3, width = 10, height = 5, dpi = 100)
```
![](plot_dgp_never_treated_het.png){ width=100% }

### Estimating dynamic treatment effects via TWFE event-study regressions

Like before, we start by estimating a TWFE event-study regression 
$$Y_{i,t} = \alpha_i + \alpha_t + \gamma_k^{-K}~ D_{i,t}^{<-K} + \sum_{k=-K}^{-2} \gamma_k^{lead} D_{i,t}^{k} +\sum_{k=0}^{L}  \gamma_k^{lag} D_{i,t}^{k} + \gamma_k^{L+} D_{i,t}^{>L} + \varepsilon_{i,t}$$
with $K$ and $L$ set to 5. That is, binning the endpoints.

As you can see from the results below, the biases are still substantial. Given the results we have shown before, this should not be a surprise.
```{r ES1_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
# variables we will use
keepvars <- c("`rel_year_-5`",  "`rel_year_-4`",  "`rel_year_-3`",  "`rel_year_-2`",
              "rel_year_0", "rel_year_1", "rel_year_2", "rel_year_3", "rel_year_4", "rel_year_5")

run_ES_DiD_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  # make dummy columns
  data <- data %>% 
    # make dummies
    mutate(rel_year = year - cohort_year) %>% 
    mutate(rel_year = ifelse(rel_year == -Inf, NA, rel_year))%>% 
    dummy_cols(select_columns = "rel_year") %>% 
    mutate(across(starts_with("rel_year_"), ~replace_na(., 0))) %>% 
    # generate pre and post dummies
    mutate(Pre = ifelse((rel_year < -5) * (!is.na(rel_year)), 1, 0),
           Post = ifelse((rel_year > 5) * (!is.na(rel_year)), 1, 0)) %>%
    mutate(Pre = ifelse(is.na(Pre), 0, Pre),
           Post = ifelse(is.na(Post), 0, Post))
  
  # estimate the model
  mod <- lfe::felm(dep_var ~ Pre + `rel_year_-5` + `rel_year_-4` + `rel_year_-3` + `rel_year_-2` + 
                `rel_year_0` + `rel_year_1` + `rel_year_2` + `rel_year_3` + `rel_year_4` + 
                `rel_year_5` + Post | unit + year | 0 | state, data = data, exactDOF = TRUE)
  
  # grab the obs we need
# grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_classical_never_het <- map_dfr(1:nrep, run_ES_DiD_never_het)

ES_plot_classical_never_het <- data_classical_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with binned end-points")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_classical_never_het

#ggsave("es_plot_classical_never_het.png", ES_plot_classical_never_het, width = 10, height = 5, dpi = 150)
```
![](es_plot_classical_never_het.png){ width=100% }

Next, we consider the case in which one includes almost all leads and lags ($K=17$ and $L=24$) but omits the furthest lead, $K=18$. 


```{r ES3_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_ES_DiD_sat_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  
  # make dummy columns
  data <- data %>% 
    # make relative year indicator
    mutate(rel_year = year - cohort_year)
  
  # get the minimum relative year - we need this to reindex
  min_year <- min(data$rel_year * (data$rel_year != -Inf), na.rm = T)
  
  # reindex the relative years
  data <- data %>% 
    mutate(rel_year2 = rel_year) %>% 
    mutate(rel_year = rel_year - min_year) %>% 
    dummy_cols(select_columns = "rel_year") %>% 
    select(-("rel_year_-Inf"))
    
  
  # make regression formula 
  indics <- paste("rel_year", (1:max(data$rel_year))[-(-1 - min_year)], sep = "_", collapse = " + ")
  keepvars <- paste("rel_year", c(-5:-2, 0:5) - min_year, sep = "_")  
  formula <- as.formula(paste("dep_var ~", indics, "| unit + year | 0 | state"))
  
  # run mod
  mod <- felm(formula, data = data, exactDOF = TRUE)
  
  # grab the obs we need
# grab the obs we need
  mod2 <- tibble(
    estimate = mod$coefficients,
    term1 = rownames(mod$coefficients)
    )
  
 es <-
   mod2 %>% 
    filter(term1 %in% keepvars) %>% 
    mutate(t = c(-5:-2, 0:5)) %>% 
    select(t, estimate)
 es
}

data_sat_never_het <- map_dfr(1:nrep, run_ES_DiD_sat_never_het)

ES_plot_sat_never_het <- data_sat_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  bind_rows(tibble(t = -1, avg = 0, sd = 0, lower.ci = 0, upper.ci = 0)) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("TWFE event-study regression with 'all' leads and lags")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_sat_never_het

#ggsave("es_plot_sat_never_het.png", ES_plot_sat_never_het, width = 10, height = 5, dpi = 150)

```
![](es_plot_sat_never_het.png){ width=100% }

As the result above shows, when treatment effects dynamics are heterogeneous across groups, TWFE event-study regressions do not recover the true underlying dynamic treatment effects, even if a "never-treated" group is available and one does not bin extreme event-times. Again, this is a red flag for the usage of TWFE-type regressions in DiD setups. 

### Using the Callaway and Sant'Anna (2021) framework

Next, we show that the procedure put forward by Callaway and Sant'Anna (2021) and implemented in our **did** can also be used in this setup. In other words, Callaway and Sant'Anna (2021) proposal *does not rely on restricting treatment effect heterogeneity*.


First, let's estimate the treatment effect dynamics using the "never-treated" units as the comparison group.

```{r CS_never_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}
# function to run ES DID
run_CS_never_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "nevertreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_never_het <- map_dfr(1:nrep, run_CS_never_het)

ES_plot_CS_never_het <- data_CS_never_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Never-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_never_het

#ggsave("es_plot_CS_never_het.png", ES_plot_CS_never_het, width = 10, height = 5)
```
![](es_plot_CS_never_het.png){ width=100% }

As expected, the results look good!

 
Now, let's consider the "not-yet-treated" units as a comparison group.

```{r CS_ny_het, message = FALSE, warning = FALSE, cache = TRUE, cache.lazy = TRUE, fig.width=8,fig.height=5, fig.align='center', out.width="90%", dpi = 200}

# function to run ES DID
run_CS_ny_het <- function(...) {
  
  # resimulate the data
  data <- make_data3()
  data$cohort_year[data$cohort_year==Inf] <- 0
  
  mod <- did::att_gt(yname = "dep_var", 
                     tname = "year",
                     idname = "unit",
                     gname = "cohort_year",
                     control_group = "notyettreated",
                     bstrap = FALSE,
                     data = data,
                     print_details = FALSE)
  event_std <- did::aggte(mod, type = "dynamic")
  
  att.egt <- event_std$att.egt
  names(att.egt) <- event_std$egt
  
  # grab the obs we need
  broom::tidy(att.egt) %>% 
    filter(names %in% -5:5) %>% 
    mutate(t = -5:5, estimate = x) %>% 
    select(t, estimate)
}

data_CS_ny_het <- map_dfr(1:nrep, run_CS_ny_het)

ES_plot_CS_ny_het <- data_CS_ny_het %>% 
  group_by(t) %>% 
  summarize(avg = mean(estimate),
            sd = sd(estimate),
            lower.ci = avg - 1.96*sd,
            upper.ci = avg + 1.96*sd) %>% 
  mutate(true_tau = ifelse(t >= 0, (t + 1)* 2, 0)) %>% 
  ggplot(aes(x = t, y = avg)) + 
  #geom_linerange(aes(ymin = lower.ci, ymax = upper.ci), color = 'darkgrey', size = 2) + 
  geom_ribbon(aes(ymin = lower.ci, ymax = upper.ci), color = "lightgrey", alpha = 0.2) +
  geom_point(color = 'blue', size = 3) + 
   geom_line(aes(color = 'Estimated Effect'), size = 1) + 
   geom_line(aes(x = t, y = true_tau, color = 'True Effect'), linetype = "dashed", size = 2) + 
  geom_hline(yintercept = 0, linetype = "dashed") + 
  scale_x_continuous(breaks = -5:5) + 
  labs(x = "Relative Time", y = "Estimate") + 
  theme(axis.title = element_text(size = 14),
        axis.text = element_text(size = 12))+
  ggtitle("Event-study-parameters estimated using Callaway and Sant'Anna (2021)\nComparison group: Not-yet-treated units")+
  scale_color_manual(values = colors) + 
  theme(plot.title = element_text(hjust = 0.5, size = 12),
        legend.position = "bottom", 
        legend.title = element_blank())

ES_plot_CS_ny_het

#ggsave("es_plot_CS_ny_het.png", ES_plot_CS_ny_het, width = 10, height = 5, dpi = 150)
```
![](es_plot_CS_ny_het.png){ width=100% }

As expected, the results look good, too!

## Conclusion

In this vignette, we have shown that relying on TWFE event-study regressions can lead to very misleading conclusions about treatment effect dynamics. On the other hand, the approach put forward by Callaway and Sant'Anna (2021) and implemented in our **did** package does not suffer from these drawbacks. 

## References

* [Borusyak, Kirill, Jaravel, Xavier and Spiess, Jann. "Revisiting Event Study Designs: Robust and Efficient Estimation". The Review of Economic Studies, Vol. 91, No 6, pp. 3253–3285](https://doi.org/10.1093/restud/rdae007)

* [Callaway, Brantly, and Pedro H. C. Sant'Anna. "Difference-in-differences with multiple time periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021](https://doi.org/10.1016/j.jeconom.2020.12.001).

* [Sun, Liyang, and Sarah Abraham. "Estimating dynamic treatment effects in event studies with heterogeneous treatment effects." Journal of Econometrics, Vol. 225, No. 2, pp. 175-199, 2021](https://doi.org/10.1016/j.jeconom.2020.09.006)

---
title: "Fast Test Cases for did Package"
author: "Brantly Callaway and Pedro H.C. Sant&apos;Anna"
date: "`r Sys.Date()`"
output: rmarkdown::html_document
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, echo=FALSE, results="hide", warning=FALSE, message=FALSE}
source("setup_sims.R")
fldr <- "~/Dropbox/did/R/"
#fldr <- "/Users/santanph/Dropbox/Co-authored Projects/did/R/"
sapply(paste0(fldr,list.files(fldr)), source)
library(DRDID)
library(BMisc)
library(ggplot2)
library(ggpubr)
```

# Fast Simulations

These are simulations that will run fast
and serve as basic checks that we have not
introduced any bugs into the code


## Tests with panel data

```{r}
#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, ipw model is incorectly specified here
#-----------------------------------------------------------------------------
time.periods <- 4
reset.sim()
data <- build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res

#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
reset.sim()
data <- build_ipw_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res
```


## Tests with 2 periods

```{r}
#-----------------------------------------------------------------------------
# test if 2 period case works (possible to introduce bugs that affect this
# case only)
# Expected results: warning about no pre-treatment periods to test
#-----------------------------------------------------------------------------
time.periods <- 2
reset.sim()
data <- build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw")
res

summary(aggte(res, type="simple"))
summary(aggte(res, type="group"))
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="calendar"))
```

## Test no covariates
```{r}
#-----------------------------------------------------------------------------
# test no covariates case
# Expected Result: te=1, p-value for pre-test uniformly distributed,
#  identical results for different estimation methods
#-----------------------------------------------------------------------------
time.periods <- 4
reset.sim()
bett <- betu <- rep(0,time.periods)
data <- build_sim_dataset()

res <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
              gname="G", est_method="dr")
res

res <- att_gt(yname="Y", xformla=~1, data=data, tname="period", idname="id",
              gname="G", est_method="reg")
res
```

## Test Repeated Cross Sections

```{r}
#-----------------------------------------------------------------------------
# test repeated cross sections, regression sims
# Expected result: te=1, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
reset.sim()
data <- build_sim_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg", panel=FALSE)
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", panel=FALSE)
res

#-----------------------------------------------------------------------------
# test repeated cross sections, ipw sims
# Expected result: te=1, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
reset.sim()
data <- build_ipw_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

# reg
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="reg", panel=FALSE)
res

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", panel=FALSE)
res

#-----------------------------------------------------------------------------
# test repeated cross sections, test aggregations
# Expected result: te=length of exposure, p-value for pre-test uniformly distributed
#-----------------------------------------------------------------------------
reset.sim()
te.e <- 1:time.periods
data <- build_sim_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=FALSE)
res

summary(aggte(res))
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))
```


## Unbalanced Panel Approach

```{r}
#-----------------------------------------------------------------------------
# these are same test cases as for panel data
# but estimate using allow_unbalanced_panel = TRUE
# but setting an id which gives a way to incorporate 
# unbalanced panel
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test uniform[0,1]
#-----------------------------------------------------------------------------
time.periods <- 4
reset.sim()
data <- build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", allow_unbalanced_panel=TRUE)
res


#-----------------------------------------------------------------------------
# test each estimation method with panel data
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
reset.sim()
data <- build_ipw_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="ipw", allow_unbalanced_panel=TRUE)
res

#-----------------------------------------------------------------------------
# try it with an actual unbalanced panel
# Expected results: treatment effects = 1, p-value for pre-test
# uniformly distributed, reg model is incorectly specified here
#-----------------------------------------------------------------------------
reset.sim()
data <- build_ipw_dataset()
data <- data[sample(1:nrow(data),  size=floor(.9*nrow(data))),]

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", panel=TRUE, allow_unbalanced_panel=TRUE)
res

#-----------------------------------------------------------------------------
# version that should error
# have to have an idname if you use an unbalanced panel
#-----------------------------------------------------------------------------
reset.sim()
data <- build_sim_dataset()
data <- data[sample(1:nrow(data),  size=floor(.9*nrow(data))),]

tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname=NULL,
                       gname="G", est_method="reg", panel=TRUE, allow_unbalanced_panel=TRUE),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         })

```


## Test not-yet-treated as control

```{r}
#-----------------------------------------------------------------------------
# test not yet treated as control
# Expected result: te=1, p-value for pre-test U[0,1]
#-----------------------------------------------------------------------------
reset.sim()
data <- build_ipw_dataset(panel=FALSE)

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated", 
              gname="G", est_method="dr", panel=FALSE)
res
```

```{r}
#-----------------------------------------------------------------------------
# test not yet treated as control in case w/o never treated group
# Expected result: te=1, p-value for pre-test U[0,1]
#-----------------------------------------------------------------------------
reset.sim()
data <- build_sim_dataset()
data <- subset(data, G > 0) # drop nevertreated

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated",
              gname="G", est_method="dr", panel=FALSE)
res
```


```{r}
#-----------------------------------------------------------------------------
# test nevertreated as control in case w/o never treated group
# Expected result: te=1, p-value for pre-test U[0,1], error on no never treated
#  units
#-----------------------------------------------------------------------------
reset.sim()
data <- build_sim_dataset()
data <- subset(data, G > 0) # drop nevertreated

# dr
tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         }
         )
```



## Test Aggregated Treatment Effects
```{r}
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
reset.sim()
te <- 0
te.e <- 1:time.periods
data <- build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic"))


#-----------------------------------------------------------------------------
# test group treatment timing
# Expected result: te constant within group / varies across groups
#-----------------------------------------------------------------------------
reset.sim()
te <- 0
te.bet.ind <- 1:time.periods
data <- build_ipw_dataset(panel=FALSE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="ipw", panel=FALSE)
res
summary(aggte(res, type="group"))


#-----------------------------------------------------------------------------
# test calendar time effects
# expected result: te=time
#-----------------------------------------------------------------------------
reset.sim()
te <- 0
te.t <- thet + 1:time.periods
data <- build_sim_dataset(panel=FALSE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="calendar"))

#-----------------------------------------------------------------------------
# test balancing with respect to length of exposure
# expected result: balancing fixes treatment effect dynamics
#-----------------------------------------------------------------------------
reset.sim()
te <- 0
te.e <- 1:time.periods
te.bet.ind <- 1:time.periods
data <- build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="dynamic", balance_e=1))
```


## Test Unequally Spaced Periods
```{r}
#-----------------------------------------------------------------------------
# test that att_gt and aggte work with unequally spaced periods
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
time.periods <- 8
reset.sim()
te <- 0
te.e <- 1:time.periods
data <- build_sim_dataset()
keep.periods <- c(1,2,5,7)
data <- subset(data, G %in% c(0, keep.periods))
data <- subset(data, period %in% keep.periods)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic"))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))
```

## Test Unequally Spaced First Treatment Timing
```{r}
#-----------------------------------------------------------------------------
# test that att_gt and aggte work with unequally spaced groups
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
time.periods <- 5
reset.sim()
te <- 0
te.e <- 1:time.periods
data <- build_sim_dataset()
keep.groups <- c(3,5)
data <- subset(data, G %in% c(0, keep.groups))

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="notyettreated",
              gname="G", est_method="reg", panel=FALSE)
res
summary(aggte(res, type="dynamic", balance_e=0))
summary(aggte(res, type="group"))
summary(aggte(res, type="calendar"))
```

## Test Some Units Treated in First Period
```{r}
#-----------------------------------------------------------------------------
# test that att_gt works when some units are treated in first period
# expected result: te=length of exposure, code runs with warning message about
#  dropped units
#-----------------------------------------------------------------------------
time.periods <- 4
reset.sim()
te <- 1
data <- build_sim_dataset()
data <- subset(data, period >= 2)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="reg", panel=FALSE)
res
```

## Test Dropping Some Event Times
```{r}
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure
#-----------------------------------------------------------------------------
reset.sim()
te <- 0
te.e <- 1:time.periods
data <- build_sim_dataset()

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              control_group="nevertreated",
              gname="G", est_method="dr", panel=FALSE)
res
summary(aggte(res, type="dynamic", min_e=-1, max_e=1))
```

## Test anticipation
```{r}
#-----------------------------------------------------------------------------
# *test dynamic effects*
# expected result: te=length of exposure - 1 (w/ one period -1 anticipation)
#-----------------------------------------------------------------------------
time.periods <- 5
reset.sim()
te <- 0
te.e <- -1:(time.periods-2)
data <- build_sim_dataset()
data$G <- data$G + 1 # add anticipation

#-----------------------------------------------------------------------------
# will get results wrong due to anticipation effect for g=6
# which shows up in the comparison group
# this will only affect g=5 with 1 period anticipation
#-----------------------------------------------------------------------------
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=1
              )
res
summary(aggte(res, type="dynamic"))

#-----------------------------------------------------------------------------
# drop last time period and results should be correct
#-----------------------------------------------------------------------------
data <- subset(data, period < time.periods)
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=1
              )
res
summary(aggte(res, type="dynamic"))

#-----------------------------------------------------------------------------
# incorrectly ignore anticipation, dynamic effects are incorrect due to ignoring
# anticipation
#-----------------------------------------------------------------------------
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period",
              idname="id",
              control_group="nevertreated",
              gname="G", est_method="dr",
              print_details=TRUE,
              anticipation=0
              )
res
summary(aggte(res, type="dynamic"))
```

## Change significance level
```{r}
time.periods <- 4
reset.sim()
data <- build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", alp=0.01)
res
```

## Report Pointwise Confidence Interval
```{r}
time.periods <- 4
reset.sim()
data <- build_sim_dataset()

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", alp=0.01, cband=FALSE)
res
```


```{r}
## some groups later than last treated period
## plus missing groups
time.periods <- 7
reset.sim()
data <- build_sim_dataset()
data <- subset(data, period <= 4)
missingG_ids <- sample(unique(data$id), size=10)
data[data$id %in% missingG_ids,"G"] <- NA

# dr
res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method="dr", cband=FALSE)
res
```



## Misc error handling tests

```{r}
#-----------------------------------------------------------------------------
# incorrectly specified id
#-----------------------------------------------------------------------------
time.periods <- 4
reset.sim()
data <- build_sim_dataset()

# dr
tryCatch(res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="brant",
                       gname="G", est_method="dr"),
         error=function(cond) {
           message("expected error:")
           message(cond)
           message("\n")
           return(NA)
         }
         )


#-----------------------------------------------------------------------------
# incorrectly specified time period
#-----------------------------------------------------------------------------
```


## Test custom estimation method
```{r}
#-----------------------------------------------------------------------------
# custom estimation method
# Expected results: te=1, pre-test p-value uniformly distributed, code runs
#-----------------------------------------------------------------------------
reset.sim()
data <- build_sim_dataset(panel=TRUE)

res <- att_gt(yname="Y", xformla=~X, data=data, tname="period", idname="id",
              gname="G", est_method=DRDID::drdid_imp_panel, panel=TRUE)
res
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AGGTEobj.R
\name{print.AGGTEobj}
\alias{print.AGGTEobj}
\title{print.AGGTEobj}
\usage{
\method{print}{AGGTEobj}(x, ...)
}
\arguments{
\item{x}{a \code{AGGTEobj} object}

\item{...}{extra arguments}
}
\description{
prints value of a \code{AGGTEobj} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AGGTEobj.R
\name{AGGTEobj}
\alias{AGGTEobj}
\title{AGGTEobj}
\usage{
AGGTEobj(
  overall.att = NULL,
  overall.se = NULL,
  type = "simple",
  egt = NULL,
  att.egt = NULL,
  se.egt = NULL,
  crit.val.egt = NULL,
  inf.function = NULL,
  min_e = NULL,
  max_e = NULL,
  balance_e = NULL,
  call = NULL,
  DIDparams = NULL
)
}
\arguments{
\item{overall.att}{The estimated overall ATT}

\item{overall.se}{Standard error for overall ATT}

\item{type}{Which type of aggregated treatment effect parameter to compute.
One option is "simple" (this just computes a weighted average of all
group-time average treatment effects with weights proportional to group
size).  Other options are "dynamic" (this computes average effects across
different lengths of exposure to the treatment and is similar to an
"event study"; here the overall effect averages the effect of the
treatment across all positive lengths of exposure); "group" (this
is the default option and
computes average treatment effects across different groups; here
the overall effect averages the effect across different groups); and
"calendar" (this computes average treatment effects across different
time periods; here the overall effect averages the effect across each
time period).}

\item{egt}{Holds the length of exposure (for dynamic effects), the
group (for selective treatment timing), or the time period (for calendar
time effects)}

\item{att.egt}{The ATT specific to egt}

\item{se.egt}{The standard error specific to egt}

\item{crit.val.egt}{A critical value for computing uniform confidence
bands for dynamic effects, selective treatment timing, or time period
effects.}

\item{inf.function}{The influence function of the chosen aggregated parameters}

\item{min_e}{For event studies, this is the smallest event time to compute
dynamic effects for.  By default, \code{min_e = -Inf} so that effects at
all lengths of exposure are computed.}

\item{max_e}{For event studies, this is the largest event time to compute
dynamic effects for.  By default, \code{max_e = Inf} so that effects at
all lengths of exposure are computed.}

\item{balance_e}{If set (and if one computes dynamic effects), it balances
the sample with respect to event time.  For example, if \code{balance_e=2},
\code{aggte} will drop groups that are not exposed to treatment for
at least three periods. (the initial period when \code{e=0} as well as the
next two periods when \code{e=1} and the \code{e=2}).  This ensures that
the composition of groups does not change when event time changes.}

\item{call}{The function call to aggte}

\item{DIDparams}{A DIDparams object}
}
\value{
an AGGTEobj
}
\description{
Objects of this class hold results on aggregated
group-time average treatment effects

An object for holding aggregated treatment effect parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{glance.AGGTEobj}
\alias{glance.AGGTEobj}
\title{glance model characteristics from AGGTEobj objects}
\usage{
\method{glance}{AGGTEobj}(x, ...)
}
\arguments{
\item{x}{a model of class AGGTEobj produced by the \code{\link[=aggte]{aggte()}} function}

\item{...}{other arguments passed to methods}
}
\description{
glance model characteristics from AGGTEobj objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_data.R
\name{build_sim_dataset}
\alias{build_sim_dataset}
\title{build_sim_dataset}
\usage{
build_sim_dataset(sp_list, panel = TRUE)
}
\arguments{
\item{sp_list}{A list of simulation parameters.  See \code{reset.sim} to generate
some default values for parameters}

\item{panel}{whether to construct panel data (the default) or repeated
cross sections data}
}
\value{
a data.frame with the following columns
\itemize{
\item G observations group
\item X value of covariate
\item id observation's id
\item cluster observation's cluster (by construction there is no within-cluster correlation)
\item period time period for current observation
\item Y outcome
\item treat whether or not this unit is ever treated
}
}
\description{
A function for building simulated data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AGGTEobj.R
\name{summary.AGGTEobj}
\alias{summary.AGGTEobj}
\title{Summary Aggregate Treatment Effect Parameter Objects}
\usage{
\method{summary}{AGGTEobj}(object, ...)
}
\arguments{
\item{object}{an \code{AGGTEobj} object}

\item{...}{other arguments}
}
\description{
A function to summarize aggregated treatment effect parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditional_did_pretest.R
\name{conditional_did_pretest}
\alias{conditional_did_pretest}
\title{Pre-Test of Conditional Parallel Trends Assumption}
\usage{
conditional_did_pretest(
  yname,
  tname,
  idname = NULL,
  gname,
  xformla = NULL,
  data,
  panel = TRUE,
  allow_unbalanced_panel = FALSE,
  control_group = c("nevertreated", "notyettreated"),
  weightsname = NULL,
  alp = 0.05,
  bstrap = TRUE,
  cband = TRUE,
  biters = 1000,
  clustervars = NULL,
  est_method = "ipw",
  print_details = FALSE,
  pl = FALSE,
  cores = 1
)
}
\arguments{
\item{yname}{The name of the outcome variable}

\item{tname}{The name of the column containing the time periods}

\item{idname}{The individual (cross-sectional unit) id name}

\item{gname}{The name of the variable in \code{data} that
contains the first period when a particular observation is treated.
This should be a positive number for all observations in treated groups.
It defines which "group" a unit belongs to.  It should be 0 for units
in the untreated group.}

\item{xformla}{A formula for the covariates to include in the
model.  It should be of the form \code{~ X1 + X2}.  Default
is NULL which is equivalent to \code{xformla=~1}.  This is
used to create a matrix of covariates which is then passed
to the 2x2 DID estimator chosen in \code{est_method}.

For time-varying covariates: (1) With balanced panel data,
in each 2x2 comparison, the covariates
are taken to be the value of the covariates in the earlier time
period, and all of the underlying computations involve changes in Y
as a function of those values of covariates.  (2) With repeated cross
sections data and unbalanced panel data, the covariates are taken
from each time period and computations involve Y_post conditional
on X_post minus Y_pre conditional on X_pre.  A byproduct of this
is that, with balanced panel data and in the presence of
time-varying covariates, it is possible to get different numerical
results according to whether or not \code{allow_unbalanced_panel=TRUE} or
\code{FALSE}.}

\item{data}{The name of the data.frame that contains the data}

\item{panel}{Whether or not the data is a panel dataset.
The panel dataset should be provided in long format -- that
is, where each row corresponds to a unit observed at a
particular point in time.  The default is TRUE.  When
using a panel dataset, the variable \code{idname} must
be set.  When \code{panel=FALSE}, the data is treated
as repeated cross sections.}

\item{allow_unbalanced_panel}{Whether or not function should
"balance" the panel with respect to time and id.  The default
value is \code{FALSE} which means that \code{\link[=att_gt]{att_gt()}} will drop
all units where data is not observed in all periods.
The advantage of this is that the computations are faster
(sometimes substantially).}

\item{control_group}{Which units to use as the control group.
The default is "nevertreated" which sets the control group
to be the group of units that never participate in the
treatment.  This group does not change across groups or
time periods.  The other option is to set
\code{group="notyettreated"}.  In this case, the control group
is set to the group of units that have not yet participated
in the treatment in that time period.  This includes all
never treated units, but it includes additional units that
eventually participate in the treatment, but have not
participated yet.}

\item{weightsname}{The name of the column containing the sampling weights.
If not set, all observations have same weight.}

\item{alp}{the significance level, default is 0.05}

\item{bstrap}{Not used by the pre-test.  Critical values and p-values
for the Cramer von Mises and Kolmogorov-Smirnov test statistics are
always computed with the multiplier bootstrap (using \code{biters}
iterations), regardless of this argument, and no standard errors are
reported.}

\item{cband}{Boolean for whether or not to compute a uniform confidence
band that covers all of the group-time average treatment effects
with fixed probability \code{1-alp}.  In order to compute uniform confidence
bands, \code{bstrap} must also be set to \code{TRUE}.  The default is
\code{TRUE}.}

\item{biters}{The number of multiplier bootstrap iterations used to
simulate the limiting distribution of the test statistics.  The
default is 1000.}

\item{clustervars}{A vector of variables names to cluster on (the
multiplier bootstrap then draws cluster-level multipliers).  At most,
there can be two variables (otherwise will throw an error) and one of
these must be the same as idname which allows for clustering at the
individual level.}

\item{est_method}{the method to compute group-time average treatment
effects.  The default for \code{conditional_did_pretest} is "ipw" for
inverse probability weighting.  Other built-in methods include "dr"
which uses the doubly robust approach in the \code{DRDID} package and "reg"
for first step regression estimators.}

\item{print_details}{Whether or not to show details/progress of computations.
Default is \code{FALSE}.}

\item{pl}{Whether or not to use parallel processing}

\item{cores}{The number of cores to use for parallel processing. This
parallelizes Step 1 (computing the test statistic); Step 2's multiplier
bootstrap is vectorized and runs in a single process.}
}
\value{
an \code{\link{MP.TEST}} object
}
\description{
An integrated moments test for the conditional parallel trends
assumption holding in all pre-treatment time periods for all groups
}
\examples{
\dontrun{
data(mpdta)
pre.test <- conditional_did_pretest(yname="lemp",
                                    tname="year",
                                    idname="countyreal",
                                    gname="first.treat",
                                    xformla=~lpop,
                                    data=mpdta)
summary(pre.test)
}

}
\references{
Callaway, Brantly and Sant'Anna, Pedro H. C.  "Difference-in-Differences with Multiple Time Periods and an Application on the Minimum Wage and Employment." Working Paper \url{https://arxiv.org/abs/1803.09015v2} (2018).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{mpdta}
\alias{mpdta}
\title{County Teen Employment Dataset}
\format{
A data frame with 2500 rows and 6 variables:
\describe{
\item{year}{the year of the observation}
\item{countyreal}{a unique identifier for a particular county}
\item{lpop}{the log of 1000s of population for the county}
\item{lemp}{the log of teen employment in the county}
\item{first.treat}{the year that the state where the county is located
raised its minimum wage, it is set equal to 0 for counties that have
minimum wages equal to the federal minimum wage over the entire
period.}
\item{treat}{whether or not a particular county is ever treated
during the sample period.  It is time-invariant within each county
and equal to 1 exactly when first.treat is positive; it is not a
year-specific treatment indicator.}
}
}
\source{
Callaway and Sant'Anna (2021)
}
\usage{
mpdta
}
\description{
A dataset containing (the log of) teen employment in 500 counties
in the U.S. from 2003 to 2007.  This is a subset of the dataset used in Callaway and
Sant'Anna (2021).  See that paper for additional descriptions.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DIDparams.R
\name{DIDparams}
\alias{DIDparams}
\title{DIDparams}
\usage{
DIDparams(
  yname,
  tname,
  idname = NULL,
  gname,
  xformla = NULL,
  data,
  control_group,
  anticipation = 0,
  weightsname = NULL,
  fix_weights = NULL,
  alp = 0.05,
  bstrap = TRUE,
  biters = 1000,
  clustervars = NULL,
  cband = TRUE,
  print_details = TRUE,
  faster_mode = FALSE,
  pl = FALSE,
  cores = 1,
  est_method = "dr",
  base_period = "varying",
  panel = TRUE,
  true_repeated_cross_sections,
  n = NULL,
  nG = NULL,
  nT = NULL,
  tlist = NULL,
  glist = NULL,
  call = NULL
)
}
\arguments{
\item{yname}{The name of the outcome variable}

\item{tname}{The name of the column containing the time periods}

\item{idname}{The individual (cross-sectional unit) id name}

\item{gname}{The name of the variable in \code{data} that
contains the first period when a particular observation is treated.
This should be a positive number for all observations in treated groups.
It defines which "group" a unit belongs to.  It should be 0 for units
in the untreated group.}

\item{xformla}{A formula for the covariates to include in the
model.  It should be of the form \code{~ X1 + X2}.  Default
is NULL which is equivalent to \code{xformla=~1}.  This is
used to create a matrix of covariates which is then passed
to the 2x2 DID estimator chosen in \code{est_method}.

For time-varying covariates: (1) With balanced panel data,
in each 2x2 comparison, the covariates
are taken to be the value of the covariates in the earlier time
period, and all of the underlying computations involve changes in Y
as a function of those values of covariates.  (2) With repeated cross
sections data and unbalanced panel data, the covariates are taken
from each time period and computations involve Y_post conditional
on X_post minus Y_pre conditional on X_pre.  A byproduct of this
is that, with balanced panel data and in the presence of
time-varying covariates, it is possible to get different numerical
results according to whether or not \code{allow_unbalanced_panel=TRUE} or
\code{FALSE}.}

\item{data}{The name of the data.frame that contains the data}

\item{control_group}{Which units to use as the control group.
The default is "nevertreated" which sets the control group
to be the group of units that never participate in the
treatment.  This group does not change across groups or
time periods.  The other option is to set
\code{group="notyettreated"}.  In this case, the control group
is set to the group of units that have not yet participated
in the treatment in that time period.  This includes all
never treated units, but it includes additional units that
eventually participate in the treatment, but have not
participated yet.}

\item{anticipation}{The number of time periods before participating
in the treatment where units can anticipate participating in the
treatment and therefore it can affect their untreated potential outcomes}

\item{weightsname}{The name of the column containing the sampling weights.
If not set, all observations have same weight. When weights are
time-invariant (constant within each unit across periods), all
\code{fix_weights} options produce identical results and no special
handling is needed.

When weights vary across time (e.g., time-varying population sizes),
the default behavior differs by panel type:
\describe{
\item{Balanced panel}{Each 2x2 DiD comparison uses the weight from the
earlier of the two time periods involved. For post-treatment cells,
this is the base period (g-1). For pre-treatment cells with
\code{base_period="varying"}, this is the pre-treatment period itself.
The panel DRDID estimators are used.}
\item{Repeated cross sections and unbalanced panels}{Both periods'
per-observation weights are passed directly to the RC DRDID estimators,
so each observation carries its own period-specific weight.}
}
Use the \code{fix_weights} argument to override the default behavior.}

\item{fix_weights}{Controls how time-varying sampling weights are resolved.
Only relevant when weights vary across time; with time-invariant weights,
all options produce identical results. Options:
\describe{
\item{\code{NULL} (default)}{For balanced panel: uses the weight from
the earlier of the two time periods in each 2x2 comparison. For
post-treatment cells, this is the base period (g-1). For
pre-treatment cells, this depends on the \code{base_period} setting.
For RC/unbalanced panel: uses per-observation weights from both
periods.}
\item{\code{"varying"}}{Uses per-observation, period-specific weights
for all panel types. For balanced panel data, this switches to the
repeated cross-section DRDID estimators so that pre-period and
post-period observations each carry their own weight. Covariates
are held fixed at their pre-period values (same as the default
panel estimator). This is the most flexible option for weights but
sacrifices the efficiency of the panel estimator. For RC/unbalanced
panel, this is identical to the default. Not supported with custom
\code{est_method} functions.}
\item{\code{"base_period"}}{Fixes weights at the base period (g-1) for
all (g,t) cells within a group, for both pre-treatment and
post-treatment comparisons. Ensures all cells within a group use the
same weights. For unbalanced panels, units not observed in the base
period are dropped with a warning. Not supported for repeated cross
sections (\code{panel = FALSE}).}
\item{\code{"first_period"}}{Fixes weights at the first time period in
the dataset for all (g,t) cells. For unbalanced panels, units not
observed in the first period are dropped with a warning. Not supported
for repeated cross sections (\code{panel = FALSE}).}
}}

\item{alp}{the significance level, default is 0.05}

\item{bstrap}{Boolean for whether or not to compute standard errors using
the multiplier bootstrap.  Default is \code{TRUE} (in addition, cband
is also by default \code{TRUE} indicating that uniform confidence bands
will be returned).  If \code{bstrap=FALSE}, analytical standard errors are
reported; these are cluster-robust when \code{clustervars} is supplied.}

\item{biters}{The number of bootstrap iterations to use.  The default is 1000,
and this is only applicable if \code{bstrap=TRUE}.}

\item{clustervars}{A vector of variables names to cluster on.  At most, there
can be two variables (otherwise will throw an error) and one of these
must be the same as idname which allows for clustering at the individual
level. Clustered standard errors are available with the multiplier bootstrap
(\code{bstrap=TRUE}) or analytically (\code{bstrap=FALSE}).}

\item{cband}{Boolean for whether or not to compute a uniform confidence
band that covers all of the group-time average treatment effects
with fixed probability \code{1-alp}.  In order to compute uniform confidence
bands, \code{bstrap} must also be set to \code{TRUE}.  The default is
\code{TRUE}.}

\item{print_details}{Whether or not to show details/progress of computations.
Default is \code{FALSE}.}

\item{faster_mode}{This option enables a faster version of \code{did}, optimizing
computation time for large datasets by improving data management within the package.
The default is set to \code{TRUE}. Both modes produce identical results up to
numerical precision; while the difference is minimal for small datasets,
the speedup is substantial for large ones.}

\item{pl}{Whether or not to use parallel processing}

\item{cores}{The number of cores to use for parallel processing}

\item{est_method}{the method to compute group-time average treatment effects.  The default is "dr" which uses the doubly robust
approach in the \code{DRDID} package.  Other built-in methods
include "ipw" for inverse probability weighting and "reg" for
first step regression estimators.  The user can also pass their
own function for estimating group time average treatment
effects.  The required signature depends on the data structure:

\strong{Panel data} (\code{panel=TRUE}): \code{f(y1, y0, D, covariates, i.weights, inffunc, ...)} where \code{y1} is an \verb{n x 1} vector of
post-treatment outcomes, \code{y0} is an \verb{n x 1} vector of
pre-treatment outcomes, \code{D} is a binary vector indicating
treatment group membership, \code{covariates} is an \verb{n x k} matrix,
\code{i.weights} is a vector of sampling weights, and \code{inffunc} is a
logical requesting influence-function computation.

\strong{Repeated cross sections / unbalanced panel} (\code{panel=FALSE}):
\code{f(y, post, D, covariates, i.weights, inffunc, ...)} where \code{y} is
the outcome vector (length \code{n}), \code{post} is a binary indicator for
the post-treatment period, \code{D} is a binary treatment indicator,
\code{covariates} is an \verb{n x k} matrix, \code{i.weights} is a vector of
sampling weights, and \code{inffunc} is a logical.

In both cases the function should return a list that includes
\code{ATT} (the estimated group-time average treatment effect) and
\code{att.inf.func} (an \verb{n x 1} influence function — one entry per
observation passed into the estimator).
The function can return other things as well, but these are
the only two that are required. With no covariates
(\code{xformla = NULL}), the built-in methods ("dr", "ipw", "reg") all
reduce to the unconditional difference-in-differences estimator, so
the choice among them is irrelevant; a custom \code{est_method} function
is still called (with an intercept-only \code{covariates} matrix) and
determines the estimates.}

\item{base_period}{Whether to use a "varying" base period or a
"universal" base period.  Either choice results in the same
post-treatment estimates of ATT(g,t)'s.  In pre-treatment
periods, using a varying base period amounts to computing a
pseudo-ATT in each treatment period by comparing the change
in outcomes for a particular group relative to its comparison
group in the pre-treatment periods (i.e., in pre-treatment
periods this setting computes changes from period t-1 to period
t, but repeatedly changes the value of t)

A universal base period fixes the base period to always be
(g-anticipation-1).  This does not compute
pseudo-ATT(g,t)'s in pre-treatment periods, but rather
reports average changes in outcomes from period t to
(g-anticipation-1) for a particular group relative to its comparison
group.  This is analogous to what is often reported in event
study regressions.

Using a varying base period results in an estimate of
ATT(g,t) being reported in the period immediately before
treatment.  Using a universal base period normalizes the
estimate in the period right before treatment (or earlier when
the user allows for anticipation) to be equal to 0, but one
extra estimate in an earlier period.}

\item{panel}{Whether or not the data is a panel dataset.
The panel dataset should be provided in long format -- that
is, where each row corresponds to a unit observed at a
particular point in time.  The default is TRUE.  When
using a panel dataset, the variable \code{idname} must
be set.  When \code{panel=FALSE}, the data is treated
as repeated cross sections.}

\item{true_repeated_cross_sections}{Whether or not the data really
is repeated cross sections.  (We include this because unbalanced
panel code runs through the repeated cross sections code)}

\item{n}{The number of observations.  This is equal to the
number of units (which may be different from the number
of rows in a panel dataset).}

\item{nG}{The number of groups}

\item{nT}{The number of time periods}

\item{tlist}{a vector containing each time period}

\item{glist}{a vector containing each group}

\item{call}{Function call to att_gt}
}
\description{
Object to hold did parameters that are passed across functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_data.R
\name{reset.sim}
\alias{reset.sim}
\title{reset.sim}
\usage{
reset.sim(time.periods = 4, n = 5000, ipw = TRUE, reg = TRUE)
}
\arguments{
\item{time.periods}{The number of time periods to include}

\item{n}{The total number of observations}

\item{ipw}{If TRUE, sets parameters so that DGP is
compatible with recovering ATT(g,t)'s using IPW (i.e.,
where logit that just includes a linear term in X works).  If
FALSE, sets parameters that will be incompatible with IPW.
Either way, these parameters can be specified by the user
if so desired.}

\item{reg}{If TRUE, sets parameters so that DGP is compatible
with recovering ATT(g,t)'s using regressions on untreated
untreated potential outcomes.  If FALSE, sets parameters that
will be incompatible with using regressions (i.e., regressions
that include only linear term in X).  Either way, these
parameters can be specified by the user if so desired.}
}
\value{
list of simulation parameters
}
\description{
a function to create a "reasonable" set of parameters
to create simulated panel data that obeys a parallel trends assumption.
In particular, it provides parameters where the the effect of participating
in the treatment is equal to one in all post-treatment time periods.

After calling this function, the user can change particular values of the
parameters in order to generate dynamics, heterogeneous effects across
groups, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.att_gt.R
\name{compute.att_gt}
\alias{compute.att_gt}
\title{Compute Group-Time Average Treatment Effects}
\usage{
compute.att_gt(dp)
}
\arguments{
\item{dp}{A DIDparams object}
}
\value{
a list with length equal to the number of groups times the
number of time periods; each element of the list contains an
object that contains group-time average treatment effect as well
as which group it is for and which time period it is for. It also exports
the influence function which is used externally to compute
standard errors.
}
\description{
\code{compute.att_gt} does the main work for computing
multiperiod group-time average treatment effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MP.R
\name{print.MP}
\alias{print.MP}
\title{print.MP}
\usage{
\method{print}{MP}(x, ...)
}
\arguments{
\item{x}{a \code{MP} object}

\item{...}{extra arguments}
}
\description{
prints value of a \code{MP} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{nobs.MP}
\alias{nobs.MP}
\title{Number of unique cross-sectional units in an MP object}
\usage{
\method{nobs}{MP}(object, ...)
}
\arguments{
\item{object}{a model of class MP produced by the \code{\link[=att_gt]{att_gt()}} function}

\item{...}{additional arguments (ignored)}
}
\value{
Integer. The number of unique cross-sectional units in the data.
}
\description{
Number of unique cross-sectional units in an MP object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gplot.R
\name{gplot}
\alias{gplot}
\title{gplot}
\usage{
gplot(
  ssresults,
  ylim = NULL,
  xlab = NULL,
  ylab = NULL,
  title = "Group",
  xgap = 1,
  legend = TRUE,
  ref_line = 0,
  theming = TRUE
)
}
\value{
a \code{ggplot2} object
}
\description{
does the heavy lifting for making a plot of an group-time
average treatment effect
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gplot.R
\name{splot}
\alias{splot}
\title{splot}
\usage{
splot(
  ssresults,
  ylim = NULL,
  xlab = NULL,
  ylab = NULL,
  title = "Group",
  legend = TRUE,
  ref_line = 0,
  theming = TRUE
)
}
\value{
a \code{ggplot2} object
}
\description{
alternative plotting function for the case with selective treatment
timing
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_data.R
\name{sim}
\alias{sim}
\title{sim}
\usage{
sim(
  sp_list,
  ret = NULL,
  bstrap = TRUE,
  cband = TRUE,
  control_group = "nevertreated",
  xformla = ~X,
  est_method = "dr",
  clustervars = NULL,
  panel = TRUE
)
}
\arguments{
\item{sp_list}{A list of simulation parameters.  See \code{reset.sim} to generate
some default values for parameters}

\item{ret}{which type of results to return.  The options are \code{Wpval} (returns
1 if the p-value from a Wald test that all pre-treatment ATT(g,t)'s are equal
is less than .05),
\code{cband}  (returns 1 if a uniform confidence band covers 0 for groups and times),
\code{simple} (returns 1 if, using the simple treatment effect aggregation results
in rejecting that this aggregated treatment effect parameter is equal to 0),
\code{dynamic} (returns 1 if the uniform confidence band from the dynamic treatment
effect aggregation covers 0 in all pre- and post-treatment periods).  The default
value is NULL, and in this case the function will just return the results from
the call to \code{att_gt}.}

\item{bstrap}{whether or not to use the bootstrap to conduct inference (default is TRUE)}

\item{cband}{whether or not to compute uniform confidence bands in the call to \code{att_gt}
(the default is TRUE)}

\item{control_group}{Whether to use the "nevertreated" comparison group (the default)
or the "notyettreated" as the comparison group}

\item{xformla}{Formula for covariates in \code{att_gt} (default is \code{~X})}

\item{est_method}{Which estimation method to use in \code{att_gt} (default is "dr")}

\item{clustervars}{Any additional variables which should be clustered on}

\item{panel}{whether to simulate panel data (the default) or otherwise repeated
cross sections data}
}
\value{
When \code{ret=NULL}, returns the results of the call to \code{att_gt}, otherwise returns
1 if the specified test rejects or 0 if not.
}
\description{
An internal function that builds simulated data, computes
ATT(g,t)'s and some aggregations.  It is useful for testing the inference
procedures in the \code{did} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mboot.R
\name{mboot}
\alias{mboot}
\title{Multiplier Bootstrap}
\usage{
mboot(inf.func, DIDparams, pl = FALSE, cores = 1, return_V = TRUE)
}
\arguments{
\item{inf.func}{an influence function}

\item{DIDparams}{DIDparams object}

\item{pl}{whether or not to use parallel processing in the multiplier
bootstrap, default=FALSE}

\item{cores}{the number of cores to use with parallel processing,
default=1}

\item{return_V}{whether to compute and return the bootstrap variance
matrix \code{V}. Default is \code{TRUE}. Internal callers that only consume
\code{bres}, \code{se}, or \code{crit.val} set this to \code{FALSE} to skip the computation
(it is the only O(biters x k^2) step in the function).}
}
\value{
list with elements
\item{bres}{results from each bootstrap iteration}
\item{V}{variance matrix (\code{NULL} when \code{return_V = FALSE})}
\item{se}{standard errors}
\item{crit.val}{a critical value for computing uniform confidence bands}
}
\description{
A function to take an influence function and use the
multiplier bootstrap to compute standard errors and critical values for
uniform confidence bands.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggte.R
\name{aggte}
\alias{aggte}
\title{Aggregate Group-Time Average Treatment Effects}
\usage{
aggte(
  MP,
  type = "group",
  balance_e = NULL,
  min_e = -Inf,
  max_e = Inf,
  na.rm = FALSE,
  bstrap = NULL,
  biters = NULL,
  cband = NULL,
  alp = NULL,
  clustervars = NULL
)
}
\arguments{
\item{MP}{an MP object (i.e., the results of the \code{\link[=att_gt]{att_gt()}} method)}

\item{type}{Which type of aggregated treatment effect parameter to compute.
One option is "simple" (this just computes a weighted average of all
group-time average treatment effects with weights proportional to group
size).  Other options are "dynamic" (this computes average effects across
different lengths of exposure to the treatment and is similar to an
"event study"; here the overall effect averages the effect of the
treatment across all positive lengths of exposure); "group" (this
is the default option and
computes average treatment effects across different groups; here
the overall effect averages the effect across different groups); and
"calendar" (this computes average treatment effects across different
time periods; here the overall effect averages the effect across each
time period).}

\item{balance_e}{If set (and if one computes dynamic effects), it balances
the sample with respect to event time.  For example, if \code{balance_e=2},
\code{aggte} will drop groups that are not exposed to treatment for
at least three periods. (the initial period when \code{e=0} as well as the
next two periods when \code{e=1} and the \code{e=2}).  This ensures that
the composition of groups does not change when event time changes.}

\item{min_e}{For event studies, this is the smallest event time to compute
dynamic effects for.  By default, \code{min_e = -Inf} so that effects at
all lengths of exposure are computed.}

\item{max_e}{For event studies, this is the largest event time to compute
dynamic effects for.  By default, \code{max_e = Inf} so that effects at
all lengths of exposure are computed.}

\item{na.rm}{Logical value if we are to remove missing values from analyses. Default is FALSE.}

\item{bstrap}{Boolean for whether or not to compute standard errors using
the multiplier bootstrap.  Default is value set in the MP object.  If \code{bstrap=FALSE}, then analytical
standard errors are reported; these are cluster-robust when a cluster variable was supplied to \code{att_gt}.}

\item{biters}{The number of bootstrap iterations to use.  The default is the value set in the MP object,
and this is only applicable if \code{bstrap=TRUE}.}

\item{cband}{Boolean for whether or not to compute a uniform confidence
band that covers all of the group-time average treatment effects
with fixed probability \code{1-alp}.  In order to compute uniform confidence
bands, \code{bstrap} must also be set to \code{TRUE}.  The default is
the value set in the MP object}

\item{alp}{the significance level, default is value set in the MP object.}

\item{clustervars}{A vector of variables to cluster on.  The default is
the cluster variables set in the MP object (i.e., the ones passed to
\code{\link[=att_gt]{att_gt()}}).  Clustering at the aggregation stage reuses the per-unit
cluster information stored by \code{\link[=att_gt]{att_gt()}}, so \code{aggte} can only cluster
on \code{idname} and/or the variable that was passed to \code{\link[=att_gt]{att_gt()}} via its
\code{clustervars} argument.  If a requested cluster variable cannot be
honored -- because \code{clustervars} was not supplied to \code{\link[=att_gt]{att_gt()}}, or
because it names a different variable than the one \code{\link[=att_gt]{att_gt()}} clustered
on -- \code{aggte} issues a warning and reports standard errors that do NOT
account for clustering.  To obtain clustered standard errors for the
aggregated parameters, re-run \code{\link[=att_gt]{att_gt()}} with the desired \code{clustervars}.}
}
\value{
An \code{\link{AGGTEobj}} object that holds the results from the
aggregation
}
\description{
A function to take group-time average treatment effects
and aggregate them into a smaller number of parameters.  There are
several possible aggregations including "simple", "dynamic", "group",
and "calendar."
}
\section{Examples}{


Initial ATT(g,t) estimates from \code{\link[=att_gt]{att_gt()}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{data(mpdta)
set.seed(09152024)
out <- att_gt(yname="lemp",
               tname="year",
               idname="countyreal",
               gname="first.treat",
               xformla=NULL,
               data=mpdta)
}\if{html}{\out{</div>}}

You can aggregate the ATT(g,t) in many ways.

\strong{Overall ATT:}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{aggte(out, type = "simple")
#> 
#> Call:
#> aggte(MP = out, type = "simple")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#>    ATT    Std. Error     [ 95\%  Conf. Int.]  
#>  -0.04        0.0119    -0.0633     -0.0166 *
#> 
#> 
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Doubly Robust
}\if{html}{\out{</div>}}

\strong{Dynamic ATT (Event-Study):}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{aggte(out, type = "dynamic")
#> 
#> Call:
#> aggte(MP = out, type = "dynamic")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on event-study/dynamic aggregation:  
#>      ATT    Std. Error     [ 95\%  Conf. Int.]  
#>  -0.0772        0.0208     -0.118     -0.0365 *
#> 
#> 
#> Dynamic Effects:
#>  Event time Estimate Std. Error [95\% Simult.  Conf. Band]  
#>          -3   0.0305     0.0154       -0.0101      0.0711  
#>          -2  -0.0006     0.0129       -0.0346      0.0334  
#>          -1  -0.0245     0.0145       -0.0628      0.0139  
#>           0  -0.0199     0.0119       -0.0513      0.0114  
#>           1  -0.0510     0.0160       -0.0934     -0.0085 *
#>           2  -0.1373     0.0393       -0.2412     -0.0333 *
#>           3  -0.1008     0.0356       -0.1949     -0.0067 *
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Doubly Robust
}\if{html}{\out{</div>}}

\strong{ATT for each group:}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{aggte(out, type = "group")
#> 
#> Call:
#> aggte(MP = out, type = "group")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on group/cohort aggregation:  
#>     ATT    Std. Error     [ 95\%  Conf. Int.]  
#>  -0.031        0.0126    -0.0558     -0.0063 *
#> 
#> 
#> Group Effects:
#>  Group Estimate Std. Error [95\% Simult.  Conf. Band]  
#>   2004  -0.0797     0.0300       -0.1445     -0.0150 *
#>   2006  -0.0229     0.0163       -0.0582      0.0123  
#>   2007  -0.0261     0.0167       -0.0622      0.0101  
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Doubly Robust
}\if{html}{\out{</div>}}

\strong{ATT for each calendar year:}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{aggte(out, type = "calendar")
#> 
#> Call:
#> aggte(MP = out, type = "calendar")
#> 
#> Reference: Callaway, Brantly and Pedro H.C. Sant'Anna.  "Difference-in-Differences with Multiple Time Periods." Journal of Econometrics, Vol. 225, No. 2, pp. 200-230, 2021. <https://doi.org/10.1016/j.jeconom.2020.12.001>, <https://arxiv.org/abs/1803.09015> 
#> 
#> 
#> Overall summary of ATT's based on calendar time aggregation:  
#>      ATT    Std. Error     [ 95\%  Conf. Int.]  
#>  -0.0417        0.0175    -0.0759     -0.0075 *
#> 
#> 
#> Time Effects:
#>  Time Estimate Std. Error [95\% Simult.  Conf. Band]  
#>  2004  -0.0105     0.0242       -0.0685      0.0475  
#>  2005  -0.0704     0.0319       -0.1467      0.0059  
#>  2006  -0.0488     0.0206       -0.0980      0.0004  
#>  2007  -0.0371     0.0133       -0.0689     -0.0052 *
#> ---
#> Signif. codes: `*' confidence band does not cover 0
#> 
#> Control Group:  Never Treated,  Anticipation Periods:  0
#> Estimation Method:  Doubly Robust
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imports.R
\docType{package}
\name{did-package}
\alias{did}
\alias{did-package}
\title{Difference in Differences}
\description{
\code{did} implements Difference in Differences with multiple periods and variation in
treatment timing.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://bcallaway11.github.io/did/}
  \item \url{https://github.com/bcallaway11/did/}
}

}
\author{
\strong{Maintainer}: Brantly Callaway \email{brantly.callaway@uga.edu}

Authors:
\itemize{
  \item Brantly Callaway \email{brantly.callaway@uga.edu}
  \item Pedro H. C. Sant'Anna \email{pedro.santanna@emory.edu}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MP.R
\name{MP}
\alias{MP}
\title{MP}
\usage{
MP(
  group,
  t,
  att,
  V_analytical,
  se,
  c,
  inffunc,
  n = NULL,
  W = NULL,
  Wpval = NULL,
  aggte = NULL,
  alp = 0.05,
  DIDparams = NULL
)
}
\arguments{
\item{group}{which group (defined by period first treated) an group-time average treatment effect is for}

\item{t}{which time period a group-time average treatment effect is for}

\item{att}{the group-average treatment effect for group \code{group} and time
period \code{t}}

\item{V_analytical}{Analytical estimator for the asymptotic variance-covariance matrix for group-time average treatment effects}

\item{se}{standard errors for group-time average treatment effects. If bootstrap is set to TRUE, this provides bootstrap-based se.}

\item{c}{simultaneous critical value if one is obtaining simultaneous confidence
bands. Otherwise it reports the critical value based on pointwise normal
approximation.}

\item{inffunc}{the influence function for estimating group-time average
treatment effects: one column per ATT(g,t) and one row per cross-sectional
unit (one row per observation with repeated cross sections). The rownames
hold the unit ids and are the authoritative link between rows and units --
the row order differs across \code{faster_mode = TRUE} (internal
(period, cohort, id) ordering) and \code{faster_mode = FALSE} (id-sorted), so
align rows by rowname, never by position.}

\item{n}{the number of unique cross-sectional units (unique values of idname)}

\item{W}{the Wald statistic for pre-testing the common trends assumption}

\item{Wpval}{the p-value of the Wald statistic for pre-testing the
common trends assumption}

\item{aggte}{an aggregate treatment effects object}

\item{alp}{the significance level, default is 0.05}

\item{DIDparams}{a \code{\link{DIDparams}} object.  A way to optionally return the parameters
of the call to \code{\link[=att_gt]{att_gt()}} or \code{\link[=conditional_did_pretest]{conditional_did_pretest()}}.}
}
\value{
MP object
}
\description{
Multi-period objects that hold results for group-time average treatment effects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{tidy.AGGTEobj}
\alias{tidy.AGGTEobj}
\title{Tidy an AGGTEobj into a data frame}
\usage{
\method{tidy}{AGGTEobj}(x, ...)
}
\arguments{
\item{x}{a model of class AGGTEobj produced by the \code{\link[=aggte]{aggte()}} function}

\item{...}{Additional arguments to tidying method.}
}
\value{
A data frame whose columns depend on \code{type}:
\describe{
\item{type}{the aggregation type: \code{"simple"}, \code{"dynamic"},
\code{"group"}, or \code{"calendar"}}
\item{term}{label for each estimate}
\item{estimate}{point estimate}
\item{std.error}{standard error}
\item{statistic}{t-statistic (\code{estimate / std.error})}
\item{p.value}{two-sided pointwise p-value
(\code{2 * (1 - pnorm(abs(statistic)))}).
Marginal per-estimate; does \strong{not} account for multiple testing
across event times or groups.}
\item{conf.low, conf.high}{simultaneous confidence band limits.  When
\code{bstrap=TRUE} and \code{cband=TRUE} these use the bootstrap uniform
critical value (\code{crit.val.egt}); otherwise they equal the pointwise
intervals.  For \code{type="simple"} and the overall average row of
\code{type="group"}, a single scalar is returned so simultaneous and
pointwise coincide.}
\item{point.conf.low, point.conf.high}{pointwise confidence interval limits
always using \code{qnorm(1 - alp/2)}.}
}
}
\description{
Returns a tidy data frame of aggregated treatment effect estimates from an
\code{\link[=aggte]{aggte()}} result.
}
\details{
The key distinction between \code{conf.low}/\code{conf.high} and
\code{point.conf.low}/\code{point.conf.high} is that the former accounts for
multiple testing across all estimates (simultaneous coverage), while the
latter provides marginal (per-estimate) coverage only.  Use the simultaneous
bands when you want to make joint inferences across all event times or groups.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggdid.R
\name{ggdid}
\alias{ggdid}
\title{Plot \code{did} objects using \code{ggplot2}}
\usage{
ggdid(object, ...)
}
\arguments{
\item{object}{either a \code{MP} object or \code{AGGTEobj} object. See \code{help(ggdid.MP)} and \code{help(ggdid.AGGTEobj)}.}

\item{...}{other arguments}
}
\description{
Function to plot objects from the \code{did} package
}
\examples{
\dontrun{
data(mpdta)
out <- att_gt(yname = "lemp",
              gname = "first.treat",
              idname = "countyreal",
              tname = "year",
              xformla = ~1,
              data = mpdta)

# plot all group-time average treatment effects
ggdid(out)

# plot event study aggregation
es <- aggte(out, type = "dynamic")
ggdid(es)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggdid.R
\name{ggdid.AGGTEobj}
\alias{ggdid.AGGTEobj}
\title{Plot \code{AGGTEobj} objects}
\usage{
\method{ggdid}{AGGTEobj}(
  object,
  ylim = NULL,
  xlab = NULL,
  ylab = NULL,
  title = "",
  xgap = 1,
  legend = TRUE,
  ref_line = 0,
  theming = TRUE,
  ...
)
}
\arguments{
\item{object}{an \code{AGGTEobj} object, as returned by \code{aggte}.}

\item{ylim}{optional y limits for the plot; setting here makes the y limits
the same across different plots}

\item{xlab}{optional x-axis label}

\item{ylab}{optional y-axis label}

\item{title}{optional plot title}

\item{xgap}{optional gap between the labels on the x-axis.  For example,
\code{xgap=3} indicates that the labels should show up for every third
value on the x-axis.  The default is 1.}

\item{legend}{Whether or not to include a legend (which will indicate color
of pre- and post-treatment estimates).  Default is \code{TRUE}.}

\item{ref_line}{A reference line at this value, usually to compare confidence
intervals to 0. Set to NULL to omit.}

\item{theming}{Set to FALSE to skip all theming so you can do it yourself.}

\item{...}{other arguments}
}
\description{
A function to plot \code{AGGTEobj} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MP.TEST.R
\name{summary.MP.TEST}
\alias{summary.MP.TEST}
\title{summary.MP.TEST}
\usage{
\method{summary}{MP.TEST}(object, ...)
}
\arguments{
\item{object}{an MP.TEST object}

\item{...}{other variables}
}
\description{
print a summary of test results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.aggte.R
\name{compute.aggte}
\alias{compute.aggte}
\title{Compute Aggregated Treatment Effect Parameters}
\usage{
compute.aggte(
  MP,
  type = "group",
  balance_e = NULL,
  min_e = -Inf,
  max_e = Inf,
  na.rm = FALSE,
  bstrap = NULL,
  biters = NULL,
  cband = NULL,
  alp = NULL,
  clustervars = NULL,
  call = NULL
)
}
\arguments{
\item{MP}{an MP object (i.e., the results of the \code{\link[=att_gt]{att_gt()}} method)}

\item{type}{Which type of aggregated treatment effect parameter to compute.
One option is "simple" (this just computes a weighted average of all
group-time average treatment effects with weights proportional to group
size).  Other options are "dynamic" (this computes average effects across
different lengths of exposure to the treatment and is similar to an
"event study"; here the overall effect averages the effect of the
treatment across all positive lengths of exposure); "group" (this
is the default option and
computes average treatment effects across different groups; here
the overall effect averages the effect across different groups); and
"calendar" (this computes average treatment effects across different
time periods; here the overall effect averages the effect across each
time period).}

\item{balance_e}{If set (and if one computes dynamic effects), it balances
the sample with respect to event time.  For example, if \code{balance_e=2},
\code{aggte} will drop groups that are not exposed to treatment for
at least three periods. (the initial period when \code{e=0} as well as the
next two periods when \code{e=1} and the \code{e=2}).  This ensures that
the composition of groups does not change when event time changes.}

\item{min_e}{For event studies, this is the smallest event time to compute
dynamic effects for.  By default, \code{min_e = -Inf} so that effects at
all lengths of exposure are computed.}

\item{max_e}{For event studies, this is the largest event time to compute
dynamic effects for.  By default, \code{max_e = Inf} so that effects at
all lengths of exposure are computed.}

\item{na.rm}{Logical value if we are to remove missing values from analyses. Default is FALSE.}

\item{bstrap}{Boolean for whether or not to compute standard errors using
the multiplier bootstrap.  Default is \code{TRUE} (in addition, cband
is also by default \code{TRUE} indicating that uniform confidence bands
will be returned).  If \code{bstrap=FALSE}, analytical standard errors are
reported; these are cluster-robust when \code{clustervars} is supplied.}

\item{biters}{The number of bootstrap iterations to use.  The default is 1000,
and this is only applicable if \code{bstrap=TRUE}.}

\item{cband}{Boolean for whether or not to compute a uniform confidence
band that covers all of the group-time average treatment effects
with fixed probability \code{1-alp}.  In order to compute uniform confidence
bands, \code{bstrap} must also be set to \code{TRUE}.  The default is
\code{TRUE}.}

\item{alp}{the significance level, default is 0.05}

\item{clustervars}{A vector of variables names to cluster on.  At most, there
can be two variables (otherwise will throw an error) and one of these
must be the same as idname which allows for clustering at the individual
level. Clustered standard errors are available with the multiplier bootstrap
(\code{bstrap=TRUE}) or analytically (\code{bstrap=FALSE}).}

\item{call}{The function call to aggte}
}
\value{
\code{\link{AGGTEobj}} object
}
\description{
Does the heavy lifting on computing aggregated group-time
average treatment effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utility_functions.R
\name{trimmer}
\alias{trimmer}
\title{trimmer}
\usage{
trimmer(
  g,
  tname,
  idname,
  gname,
  xformla,
  data,
  control_group = "notyettreated",
  threshold = 0.999
)
}
\arguments{
\item{g}{is a particular group (below I pass in 2009)}

\item{tname}{The name of the column containing the time periods}

\item{idname}{The individual (cross-sectional unit) id name}

\item{gname}{The name of the variable in \code{data} that
contains the first period when a particular observation is treated.
This should be a positive number for all observations in treated groups.
It defines which "group" a unit belongs to.  It should be 0 for units
in the untreated group.}

\item{xformla}{A formula for the covariates to include in the
model.  It should be of the form \code{~ X1 + X2}.  Default
is NULL which is equivalent to \code{xformla=~1}.  This is
used to create a matrix of covariates which is then passed
to the 2x2 DID estimator chosen in \code{est_method}.

For time-varying covariates: (1) With balanced panel data,
in each 2x2 comparison, the covariates
are taken to be the value of the covariates in the earlier time
period, and all of the underlying computations involve changes in Y
as a function of those values of covariates.  (2) With repeated cross
sections data and unbalanced panel data, the covariates are taken
from each time period and computations involve Y_post conditional
on X_post minus Y_pre conditional on X_pre.  A byproduct of this
is that, with balanced panel data and in the presence of
time-varying covariates, it is possible to get different numerical
results according to whether or not \code{allow_unbalanced_panel=TRUE} or
\code{FALSE}.}

\item{data}{The name of the data.frame that contains the data}

\item{control_group}{Which units to use as the control group.
The default is "nevertreated" which sets the control group
to be the group of units that never participate in the
treatment.  This group does not change across groups or
time periods.  The other option is to set
\code{group="notyettreated"}.  In this case, the control group
is set to the group of units that have not yet participated
in the treatment in that time period.  This includes all
never treated units, but it includes additional units that
eventually participate in the treatment, but have not
participated yet.}

\item{threshold}{the cutoff for which observations are flagged as
likely violators of the support condition.}
}
\value{
list of ids of observations that likely violate support conditions
}
\description{
A utility function to find observations that appear
to violate support conditions.  This function is not called
anywhere in the code, but it is just useful for debugging some
common issues that users run into.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{tidy.MP}
\alias{tidy.MP}
\title{Tidy an MP object into a data frame}
\usage{
\method{tidy}{MP}(x, ...)
}
\arguments{
\item{x}{a model of class MP produced by the \code{\link[=att_gt]{att_gt()}} function}

\item{...}{Additional arguments to tidying method.}
}
\value{
A data frame with one row per ATT(g,t) estimate and columns:
\describe{
\item{term}{ATT(g,t) label}
\item{group}{the treatment cohort g}
\item{time}{the time period t}
\item{estimate}{the ATT(g,t) point estimate}
\item{std.error}{standard error}
\item{statistic}{t-statistic (\code{estimate / std.error})}
\item{p.value}{two-sided pointwise p-value
(\code{2 * (1 - pnorm(abs(statistic)))}).
Marginal per-estimate; does \strong{not} account for multiple testing
across ATT(g,t) cells.}
\item{conf.low, conf.high}{simultaneous confidence band limits, using the
bootstrap uniform critical value when \code{bstrap=TRUE} and
\code{cband=TRUE}, otherwise pointwise}
\item{point.conf.low, point.conf.high}{pointwise confidence interval limits
using \code{qnorm(1 - alp/2)}, always}
}
}
\description{
Returns a tidy data frame of group-time average treatment effect estimates
from an \code{\link[=att_gt]{att_gt()}} result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_attgt.R
\name{process_attgt}
\alias{process_attgt}
\title{Process Results from \code{\link[=compute.att_gt]{compute.att_gt()}}}
\usage{
process_attgt(attgt.list)
}
\arguments{
\item{attgt.list}{list of results from \code{\link[=compute.att_gt]{compute.att_gt()}}}
}
\value{
list with elements:
\item{group}{which group a set of results belongs to}
\item{tt}{which time period a set of results belongs to}
\item{att}{the group time average treatment effect}
}
\description{
Process Results from \code{\link[=compute.att_gt]{compute.att_gt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MP.TEST.R
\name{MP.TEST}
\alias{MP.TEST}
\title{MP.TEST}
\usage{
MP.TEST(
  CvM = NULL,
  CvMb = NULL,
  CvMcval = NULL,
  CvMpval = NULL,
  KS = NULL,
  KSb = NULL,
  KScval = NULL,
  KSpval = NULL,
  clustervars = NULL,
  xformla = NULL
)
}
\arguments{
\item{CvM}{Cramer von Mises test statistic}

\item{CvMb}{a vector of bootstrapped Cramer von Mises test statistics}

\item{CvMcval}{CvM critical value}

\item{CvMpval}{p-value for CvM test}

\item{KS}{Kolmogorov-Smirnov test statistic}

\item{KSb}{a vector of bootstrapped KS test statistics}

\item{KScval}{KS critical value}

\item{KSpval}{p-value for KS test}

\item{clustervars}{vector of which variables were clustered on for the test}

\item{xformla}{formla for the X variables used in the test}
}
\description{
An object that holds results from computing pre-test of the
conditional parallel trends assumption
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.att_gt2.R
\name{run_att_gt_estimation}
\alias{run_att_gt_estimation}
\title{Run ATT estimation for a given group-time pair}
\usage{
run_att_gt_estimation(g, t, dp2)
}
\arguments{
\item{g}{group of interest (treated group at time t)}

\item{t}{time period}

\item{dp2}{A DIDparams object v2.0}
}
\value{
a list with the gt cell and the results after performing estimation
}
\description{
\code{run_att_gt_estimation} does the main work for computing
multiperiod group-time average treatment effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\name{glance.MP}
\alias{glance.MP}
\title{glance model characteristics from MP objects}
\usage{
\method{glance}{MP}(x, ...)
}
\arguments{
\item{x}{a model of class MP produced by the \code{\link[=att_gt]{att_gt()}} function}

\item{...}{other arguments passed to methods}
}
\description{
glance model characteristics from MP objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditional_did_pretest.R
\name{indicator}
\alias{indicator}
\title{indicator}
\usage{
indicator(X, u)
}
\arguments{
\item{X}{matrix of X's from the data}

\item{u}{a particular value to compare X's to}
}
\value{
numeric vector
}
\description{
indicator weighting function
}
\examples{
data(mpdta)
dta <- subset(mpdta, year==2007)
X <- model.matrix(~lpop, data=dta)
X <- indicator(X, X[1,])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MP.R
\name{summary.MP}
\alias{summary.MP}
\title{summary.MP}
\usage{
\method{summary}{MP}(object, ...)
}
\arguments{
\item{object}{an \code{MP} object}

\item{...}{extra arguments}
}
\description{
prints a summary of a \code{MP} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.aggte.R
\name{getSE}
\alias{getSE}
\title{Take influence function and return standard errors}
\usage{
getSE(thisinffunc, DIDparams = NULL)
}
\arguments{
\item{thisinffunc}{An influence function}
}
\value{
scalar standard error
}
\description{
Function to take an nx1 influence function and return
a standard error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditional_did_pretest.R
\name{test.mboot}
\alias{test.mboot}
\title{Multiplier Bootstrap for Conditional Moment Test}
\usage{
test.mboot(inf.func, DIDparams, cores = 1)
}
\arguments{
\item{inf.func}{an influence function}

\item{DIDparams}{DIDparams object}

\item{cores}{Unused; retained for backward compatibility. The
multiplier bootstrap is computed with vectorized matrix operations
in a single process, so this argument has no effect here. In
\code{conditional_did_pretest}, \code{cores} parallelizes only
Step 1 (computing the test statistic).}
}
\value{
list
\item{bres}{CvM test statistics for each bootstrap iteration}
\item{crit.val}{critical value for CvM test statistic}
}
\description{
A slightly modified multiplier bootstrap procedure
for the pre-test of the conditional parallel trends assumption
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{tidy}
\alias{glance}
\title{tidy results}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{generics}{\code{\link[generics:glance]{glance()}}, \code{\link[generics:tidy]{tidy()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggdid.R
\name{ggdid.MP}
\alias{ggdid.MP}
\title{Plot \code{MP} objects using \code{ggplot2}}
\usage{
\method{ggdid}{MP}(
  object,
  ylim = NULL,
  xlab = NULL,
  ylab = NULL,
  title = "Group",
  xgap = 1,
  ncol = 1,
  legend = TRUE,
  group = NULL,
  ref_line = 0,
  theming = TRUE,
  grtitle = "Group",
  ...
)
}
\arguments{
\item{object}{an \code{MP} object, as returned by \code{att_gt}.}

\item{ylim}{optional y limits for the plot; setting here makes the y limits
the same across different plots}

\item{xlab}{optional x-axis label}

\item{ylab}{optional y-axis label}

\item{title}{optional plot title}

\item{xgap}{optional gap between the labels on the x-axis.  For example,
\code{xgap=3} indicates that the labels should show up for every third
value on the x-axis.  The default is 1.}

\item{ncol}{The number of columns to include in the resulting plot.  The
default is 1.}

\item{legend}{Whether or not to include a legend (which will indicate color
of pre- and post-treatment estimates).  Default is \code{TRUE}.}

\item{group}{Vector for which groups to include in the plots of ATT(g,t).
Default is NULL, and, in this case, plots for all groups will be included (\code{ggdid.MP} only).}

\item{ref_line}{A reference line at this value, usually to compare confidence
intervals to 0. Set to NULL to omit.}

\item{theming}{Set to FALSE to skip all theming so you can do it yourself.}

\item{grtitle}{Title to append before each group name (\code{ggdid.MP} only).}

\item{...}{other arguments}
}
\description{
A function to plot \code{MP} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.aggte.R
\name{get_agg_inf_func}
\alias{get_agg_inf_func}
\title{Get an influence function for particular aggregate parameters}
\usage{
get_agg_inf_func(att, inffunc1, whichones, weights.agg, wif = NULL)
}
\arguments{
\item{att}{vector of group-time average treatment effects}

\item{inffunc1}{influence function for all group-time average treatment effects
(matrix)}

\item{whichones}{which elements of att will be used to compute the aggregated
treatment effect parameter}

\item{weights.agg}{the weights to apply to each element of att(whichones);
should have the same dimension as att(whichones)}

\item{wif}{extra influence function term coming from estimating the weights;
should be n x k matrix where k is dimension of whichones}
}
\value{
nx1 influence function
}
\description{
Get an influence function for particular aggregate parameters

This is a generic internal function for combining influence
functions across ATT(g,t)'s to return an influence function for
various aggregated treatment effect parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute.aggte.R
\name{wif}
\alias{wif}
\title{Compute extra term in influence function due to estimating weights}
\usage{
wif(keepers, pg, weights.ind, G, group)
}
\arguments{
\item{keepers}{a vector of indices for which group-time average
treatment effects are used to compute a particular aggregated parameter}

\item{pg}{a vector with same length as total number of group-time average
treatment effects that contains the probability of being in particular group}

\item{weights.ind}{additional sampling weights (nx1)}

\item{G}{vector containing which group a unit belongs to (nx1)}

\item{group}{vector of groups}
}
\value{
nxk influence function matrix
}
\description{
A function to compute the extra term that shows up in the
influence function for aggregated treatment effect parameters
due to estimating the weights
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{citation}
\alias{citation}
\title{citation}
\usage{
citation()
}
\description{
print the citation for the relevant paper
}
\keyword{internal}
